/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2003 Erik Bengtson - removed exist() operation
2003 Andy Jefferson - commented
2003 Andy Jefferson - added check on makePersistent that class is persistence capable
2003 Andy Jefferson - introduction of localiser
2003 Erik Bengtson - provided initial caching mechanism
2003 Erik Bengtson - added getObjectbyAID
2003 Erik Bengtson - removed unused variables
2004 Erik Bengtson - implemented evictAll    
2004 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - added LifecycleListener
2004 Andy Jefferson - rewritten Cache
2005 Andy Jefferson - moved ClassLoaderResolver handling to separate class
2005 Marco Schulze - implemented copying the lifecycle listeners in j2ee environment
     ...
 **********************************************************************/
package org.jpox;

import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.Extent;
import javax.jdo.FetchPlan;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOHelper;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOOptimisticVerificationException;
import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.Query;
import javax.jdo.datastore.Sequence;
import javax.jdo.identity.SingleFieldIdentity;
import javax.jdo.listener.InstanceLifecycleListener;
import javax.jdo.spi.Detachable;
import javax.jdo.spi.JDOImplHelper;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.cache.CachedPC;
import org.jpox.cache.Level1Cache;
import org.jpox.cache.Level2Cache;
import org.jpox.exceptions.ClassNotDetachableException;
import org.jpox.exceptions.ClassNotPersistenceCapableException;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.exceptions.CommitStateTransitionException;
import org.jpox.exceptions.MetaDataForPersistenceCapableClassNotReachableException;
import org.jpox.exceptions.ObjectDetachedException;
import org.jpox.exceptions.RollbackStateTransitionException;
import org.jpox.exceptions.TransactionActiveException;
import org.jpox.exceptions.TransactionNotActiveException;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.MetaDataManager;
import org.jpox.metadata.QueryLanguage;
import org.jpox.metadata.QueryMetaData;
import org.jpox.metadata.SequenceMetaData;
import org.jpox.plugin.ConfigurationElement;
import org.jpox.plugin.Extension;
import org.jpox.state.CallbackHandler;
import org.jpox.state.DetachState;
import org.jpox.state.FetchPlanState;
import org.jpox.state.StateManagerFactory;
import org.jpox.store.FieldValues;
import org.jpox.store.OID;
import org.jpox.store.OIDFactory;
import org.jpox.store.SCOID;
import org.jpox.store.StoreManager;
import org.jpox.store.StoreManagerFactory;
import org.jpox.store.query.QueryResult;
import org.jpox.store.rdbms.query.JPOXSQLQuery;
import org.jpox.store.rdbms.query.SQLQuery;
import org.jpox.util.AIDUtils;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.SoftValueMap;
import org.jpox.util.WeakValueMap;

/**
 * This class implements the functionality needed for a JDO persistence manager.
 * 
 * @version $Revision: 1.237.2.17 $
 */
public abstract class AbstractPersistenceManager implements PersistenceManager
{
    /** Localisation utility for output messages */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /** Owning PMF for this PM. */
    protected final AbstractPersistenceManagerFactory apmf;

    /** StoreManager that the PM relates to. */
    private StoreManager srm;

    /** Map of user objects attached to this PM. */
    private Map userObjectMap;

    /** User object attached to the PM. */
    private Object userObject;

    /** State variable for whether the PM is closed. */
    private boolean closed;

    //-- cache level 1 --//
    /** Level 1 Cache */
    private Level1Cache cache;

    //-- reachability --//
    /** Cache of ObjectIds of objects either persisted using makePersistent, or known as already persistent in the current transaction */
    private Set txKnownPersists = new HashSet(); 

    /** Cache of ObjectIds of objects newly persistent in the current transaction */
    private Set txFlushedNew = new HashSet();

    //-- convenience operations --//
    /** State variable used when searching for the StateManager for an object. */
    private PersistenceCapable lookingForStateManagerFor = null;

    /** State variable used when searching for the StateManager for an object. */
    private StateManager foundStateManager = null;

    //-- transaction --//
    /** Current transaction */
    protected Transaction tx;

    /** Cache of StateManagers enlisted in the current transaction, keyed by the object id. */
    private Map enlistedSMCache = new WeakValueMap();

    /** Identities of all objects enlisted in the current transaction. Used in reachability to define the extent of checks. */
    private Set enlistedIds = new HashSet();

    /** List of StateManagers for all current dirty objects managed by this PM. */
    private List dirtySMs = new ArrayList(10);

    //-- configuration --//
    /** Whether to ignore the cache */
    private boolean ignoreCache;

    /** Whether to detach objects on close of the PM. */
    private boolean detachOnClose;

    /** Whether to detach all objects on commit of the transaction. */
    private boolean detachAllOnCommit;

    /** Whether to operate multithreaded */
    private boolean multithreaded;

    //-- others --//
    /** Current FetchPlan for the PM. */
    private FetchPlan fetchPlan;

    /** The ClassLoader resolver to use for class loading issues. */
    protected ClassLoaderResolver clr = null;

    /** list of query results to be closed at close. Can be garbage collected. **/
    private Map queryResults = new SoftValueMap();

    /** List of all queries already run by this PM. The queries are used to hint the fetch plan usage. 
     * If the query was run before, we don't fetch all fields in fetch plan again **/
    private Map queriesRun = new SoftValueMap();

    private CallbackHandler callbacks;
    
    /**
     * Constructor.
     * @param apmf Persistence Manager Factory
     * @param userName Username for the datastore
     * @param password Password for the datastore
     */
    public AbstractPersistenceManager(AbstractPersistenceManagerFactory apmf, String userName, String password)
    {
        ClassLoader contextLoader = Thread.currentThread().getContextClassLoader();
        clr = apmf.getPMFContext().getClassLoaderResolver(contextLoader);
        try
        {
            clr.registerClassLoader(apmf.getPMFContext().getImplementationCreator().getClassLoader());
        }
        catch (Exception ex)
        {
            //do nothing
        }

        this.apmf = apmf;
        closed = false;
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            JPOXLogger.JDO.debug(LOCALISER.msg("PM.Opened", this, srm));
        }

        srm = getStoreManager(apmf, userName, password);

        setIgnoreCache(apmf.getPMFConfiguration().getIgnoreCache());
        setDetachOnClose(apmf.getPMFConfiguration().getDetachOnClose());
        setDetachAllOnCommit(apmf.getPMFConfiguration().getDetachAllOnCommit());
        setMultithreaded(apmf.getPMFConfiguration().getMultithreaded());
        fetchPlan = new FetchPlanImpl().setMaxFetchDepth(apmf.getMaxFetchDepth());
        userObject = null;
        userObjectMap = null;

        // Set up the Level 1 Cache
        initialiseLevel1Cache();
    }

    /**
     * Method to initialise the L1 cache.
     */
    protected void initialiseLevel1Cache()
    {
        // Find the L1 cache class name from its plugin name
        String level1Type = apmf.getPMFConfiguration().getJdoCacheLevel1Type();
        String level1ClassName = apmf.getPMFContext().getPluginManager().getAttributeValueForExtension("org.jpox.cache_level1", 
            "name", level1Type, "class-name");
        if (level1ClassName == null)
        {
            // Plugin of this name not found
            throw new JDOFatalUserException(LOCALISER.msg("Cache.Level1.PluginNotFound", level1Type));
        }

        try
        {
            // Create an instance of the L1 Cache
            Class level1CacheClass = Class.forName(level1ClassName);
            cache = (Level1Cache) level1CacheClass.newInstance();
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.Initialised", level1Type));
            }
        }
        catch (Exception e)
        {
            // Class name for this L1 cache plugin is not found!
            throw new JDOFatalUserException(LOCALISER.msg("Cache.Level1.PluginClassNotFound", level1Type, level1ClassName),e);
        }
    }

    /**
     * Get an StoreManager based on the ConnectionURL
     * @param apmf the AbstractPersistenceManagerFactory
     * @param userName the user name
     * @param password the password
     * @return the StoreManager
     */
    private StoreManager getStoreManager(AbstractPersistenceManagerFactory apmf, String userName, String password)
    {
        Extension[] exts = apmf.getPMFContext().getPluginManager().getExtensionPoint("org.jpox.store_manager").getExtensions();

        // get only the first part / protocol of the url. eg. "jdbc:derby:.." keeps only "jdbc"
        String url = apmf.getConnectionURL();
        if (url != null)
        {
            int idx = url.indexOf(':');
            if (idx > -1)
            {
                url = url.substring(0, idx);
            }
        }

        StoreManager srm = null;
        for (int e=0; srm == null && e<exts.length; e++)
        {
            ConfigurationElement[] confElm = exts[e].getConfigurationElements();
            for( int c=0; srm == null && c<confElm.length; c++)
            {
                String storeMgrClassName = confElm[c].getAttribute("class-name");
                String key = confElm[c].getAttribute("key");
                if (url == null || key.equalsIgnoreCase(url))
                {
                    srm = StoreManagerFactory.getStoreManager(storeMgrClassName, clr, apmf, userName, password);
                }
            }
        }
        if (srm == null)
        {
            // get the first available
            srm = StoreManagerFactory.getStoreManager(exts[0].getConfigurationElements()[0].getAttribute("class-name"), clr, 
                apmf, userName, password);
        }
        return srm;
    }

    /**
     * Accessor for the ClassLoaderResolver
     * @return the ClassLoaderResolver
     */
    public ClassLoaderResolver getClassLoaderResolver()
    {
        return clr;
    }

    /**
     * Accessor for the Store Manager.
     * @return StoreManager
     */
    public StoreManager getStoreManager()
    {
        return srm;
    }

    /**
     * Acessor for the current FetchPlan
     * @return FetchPlan
     * @since 1.1
     */
    public FetchPlan getFetchPlan()
    {
        //TODO this should be enabled, but with detachAllOnCommit=true it raises exception 
        //assertIsOpen();
        return fetchPlan;
    }

    /**
     * Accessor for the PersistenceManager Factory.
     * @return The PersistenceManagerFactory
     */
    public AbstractPersistenceManagerFactory getAbstractPersistenceManagerFactory()
    {
        return apmf;
    }

    /**
     * Accessor for the PersistenceManager Factory.
     * @return The PersistenceManagerFactory
     */
    public PersistenceManagerFactory getPersistenceManagerFactory()
    {
        // To be implemented by subclasses
        throw new JDOException(LOCALISER.msg("PM.GetPersistenceManagerFactoryNotImplemented"));
    }

    /**
     * Gets the context in which this PersistenceManager is running
     * @return Returns the pmfContext.
     */
    public PMFContext getPMFContext()
    {
        return apmf.getPMFContext();
    }

    /**
     * Accessor for the MetaDataManager for this PMF. This is used as the
     * interface to MetaData in the PM/PMF.
     * @return Returns the MetaDataManager.
     */
    public MetaDataManager getMetaDataManager()
    {
        return apmf.getPMFContext().getMetaDataManager();
    }

    /**
     * Accessor for a handle to this PersistenceManager.
     * @return The handle for this PM.
     */
    public abstract PersistenceManager getPMHandle();

    /**
     * Whether the datastore operations are delayed until commit. In optimistic
     * transactions this is automatically enabled. If we are committing then will
     * return false since the delay is no longer required.
     * @return true if datastore operations are delayed until commit
     */
    public boolean isDelayDatastoreOperationsEnabled()
    {
        if (tx.isCommitting())
        {
            // Already committing so no point delaying further!
            return false;
        }

        if (tx.getOptimistic())
        {
            return true;
        }
        else
        {
            return getPMFContext().getPmfConfiguration().getDatastoreDelayOperationsEnabled();
        }
    }

    /**
     * Accessor for the current transaction.
     * @return The transaction
     */
    public javax.jdo.Transaction currentTransaction()
    {
        assertIsOpen();
        return tx;
    }

    /**
     * Method to enlist the specified StateManager in the current transaction.
     * @param sm The StateManager
     */
    public synchronized void enlistInTransaction(StateManager sm)
    {
        assertActiveTransaction();
        if (JPOXLogger.TRANSACTION.isDebugEnabled())
        {
            JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.ObjectEnlistedInCache", sm.getInternalObjectId().toString()));
        }

        if (apmf.getPersistenceByReachabilityAtCommit())
        {
            if (sm.getObject().jdoIsNew())
            {
                // Add this object to the list of new objects in this transaction
                txFlushedNew.add(sm.getInternalObjectId());
            }
            else if (sm.getObject().jdoIsPersistent() && !sm.getObject().jdoIsDeleted())
            {
                // Add this object to the list of known valid persisted objects (unless it is a known new object)
                if (!txFlushedNew.contains(sm.getInternalObjectId()))
                {
                    txKnownPersists.add(sm.getInternalObjectId());
                }
            }
        }

        if (apmf.getPersistenceByReachabilityAtCommit() && !runningPBRAtCommit)
        {
        	// Note the id for when we run PBR at commit.
            enlistedIds.add(sm.getInternalObjectId());
        }
        enlistedSMCache.put(sm.getInternalObjectId(),sm);
    }

    /**
     * Method to evict the specified StateManager from the current transaction.
     * @param sm The StateManager
     */
    public synchronized void evictFromTransaction(StateManager sm)
    {
        if (JPOXLogger.TRANSACTION.isDebugEnabled())
        {
            JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.ObjectEvictedFromCache", sm.getInternalObjectId().toString()));
        }

        if (enlistedSMCache.remove(sm.getInternalObjectId())==null)
        {
            //probably because the object was garbage collected
            if (JPOXLogger.TRANSACTION.isDebugEnabled())
            {
                JPOXLogger.TRANSACTION.debug(LOCALISER.msg("PM.ObjectNotTransactional", sm.getInternalObjectId()));
            }
        }
    }

    /**
     * Accessor for whether the specified object is enlisted in the current transaction.
     * Only of use when running persistence-by-reachability-at-commit
     * @param id Identity of the object
     * @return Whether it is enlisted
     */
    public boolean isEnlistedInTransaction(Object id)
    {
    	if (id == null)
    	{
    		return false;
    	}
    	return enlistedIds.contains(id);
    }

    /**
     * Method to add the object managed by the specified StateManager to the cache.
     * @param sm The StateManager
     */
    public synchronized void addStateManager(StateManager sm)
    {
        // Add to the Level 1/2 Caches
        putObjectIntoCache(sm, true, true);
    }

    /**
     * Method to remove the object managed by the specified StateManager from the cache.
     * @param sm The StateManager
     */
    public synchronized void removeStateManager(StateManager sm)
    {
        PersistenceCapable pc = sm.getObject();

        // Remove from the Level 1 Cache
        removeObjectFromCache(pc, sm.getInternalObjectId(), true, false);

        // Remove it from any transaction
        enlistedSMCache.remove(sm.getInternalObjectId());
    }

    /**
     * Accessor for the StateManager of an object given the object id.
     * @param id Id of the object.
     * @return The StateManager
     */
    public synchronized StateManager getStateManagerById(Object id)
    {
        assertIsOpen();
        PersistenceCapable pc = getObjectFromCache(id);
        return findStateManager(pc);
    }

    /**
     * Accessor for whether this Persistence Manager is closed.
     * 
     * @return Whether this manager is closed.
     */
    public boolean isClosed()
    {
        return closed;
    }

    /**
     * Method to close the Persistence Manager.
     */
    public synchronized void close()
    {
        if (closed)
        {
            throw new JDOUserException(LOCALISER.msg("PM.IsClosed"));
        }
        if (tx.isActive())
        {
            throw new TransactionActiveException(this);
        }

        userObject = null;
        userObjectMap = null;

        if (detachOnClose)
        {
            // JPOX "detach-on-close" extension, detaching all currently cached objects.
            JPOXLogger.JDO.debug(LOCALISER.msg("PM.DetachOnCloseStart"));
            try
            {
                // Start a transaction in case we need to load any unloaded fields
                tx.begin();

                List toDetach = new ArrayList();

                toDetach.addAll(cache.values());
                Iterator iter = toDetach.iterator();
                while (iter.hasNext())
                {
                    Object obj = iter.next();
                    StateManager sm = (StateManager) obj;
                    if (sm != null)
                    {
                        //TODO why SM is in cache if null state ...
                        if (sm != null && sm.getObject() != null && !sm.isDeleted(sm.getObject()))
                        {
                            try
                            {
                                sm.detach(new DetachState());
                            }
                            catch (JDOObjectNotFoundException onfe)
                            {
                                // Catch exceptions for any objects that are deleted in other PMs whilst having this open
                            }
                        }
                    }
                }
                tx.commit();
            }
            finally
            {
                if (tx.isActive())
                {
                    tx.rollback();
                }
            }
            JPOXLogger.JDO.debug(LOCALISER.msg("PM.DetachOnCloseEnd"));
        }

        disconnectSMCache();
        disconnectQueryCache();
        disconnectLifecycleListener();
        reset();

        if (JPOXLogger.JDO.isDebugEnabled())
        {
            JPOXLogger.JDO.debug(LOCALISER.msg("PM.Closed", this));
        }
    }

    /**
     * Disconnect SM instances, clear cache and reset settings 
     */
    protected void disconnectSMCache()
    {
        // Clear out the cache (use separate list since sm.disconnect will remove the object from "cache" so we avoid
        // any ConcurrentModification issues)
        Collection cachedSMsClone = new HashSet(cache.values());
        Iterator iter = cachedSMsClone.iterator();
        while (iter.hasNext())
        {
            StateManager sm = (StateManager) iter.next();
            if (sm != null)
            {
                sm.disconnect();
            }
        }
        cache.clear();
        if (JPOXLogger.CACHE.isDebugEnabled())
        {
            JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.Cleared"));
        }

    }

    /**
     * Reset the PersistenceManager
     */
    protected void reset()
    {
        // Reset the Fetch Plan to its DFG setting
        getFetchPlan().clearGroups().addGroup(FetchPlan.DEFAULT);
        apmf.releasePersistenceManager(this);

        closed = true;
        srm = null;
        tx = null;
    }
    
    /**
     * Disconnect Query Results cache to close it, and disconnect the
     * list of queries that has been run 
     */
    protected void disconnectQueryCache()
    {
        // Clear out query results
        if (queryResults.size() > 0)
        {
            // Get array of results to avoid ConcurrentModificationException with results being GCed while closing
            Object[] results = queryResults.values().toArray();
            for (int i=0;i<results.length;i++)
            {
                QueryResult qr = (QueryResult)results[i];
                qr.close();
            }
            queryResults.clear();
        }
        queriesRun.clear();
    }

    /**
     * Disconnect the registered LifecycleListener
     */
    protected void disconnectLifecycleListener()
    {
        // Clear out lifecycle listeners that were registered
        if (callbacks != null)
        {
            callbacks.close();
        }
    }
    
    // ----------------------------- Eviction --------------------------------------------

    /**
     * Internal method to evict an object from L1 cache.
     * @param obj The object
     */
    private void internalEvict(Object obj)
    {
        if (obj == null)
        {
            return;
        }

        try
        {
            clr.setPrimary(obj.getClass().getClassLoader());
            assertPersistenceCapable(obj);
            assertNotDetached(obj);

            // we do not directly remove from cache level 1 here. The cache level 1 will be evicted 
            // automatically on garbage collection, if the object can be evicted. it means not all
            // jdo states allows the object to be evicted.
            PersistenceCapable pc = (PersistenceCapable) obj;
            StateManager sm = findStateManager(pc);
            if (sm == null)
            {
                throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
            }
            sm.evict();
        }
        finally
        {
            clr.unsetPrimary();
        }
    }

    /**
     * Method to evict an object from L1 cache.
     * @param pc The object
     */
    public synchronized void evict(Object pc)
    {
        assertIsOpen();
        internalEvict(pc);
    }

    /**
     * Method to evict an array of objects from L1 cache.
     * @param pcs The objects
     */
    public synchronized void evictAll(Object[] pcs)
    {
        evictAll(Arrays.asList(pcs));
    }

    /**
     * Method to evict a collection of objects from L1 cache.
     * @param pcs The objects
     * @throws JDOUserException thrown if some instances could not be evicted
     */
    public synchronized void evictAll(Collection pcs)
    {
        assertIsOpen();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalEvict(i.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.EvictError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Method to evict all current objects from L1 cache.
     */
    public synchronized void evictAll()
    {
        assertIsOpen();

        // All persistent non-transactional instances should be evicted here,
        // but we don't support those yet.
        ArrayList stateManagersToEvict = new ArrayList();
        stateManagersToEvict.addAll(cache.values());

        // Evict StateManagers and remove objects from cache
        // Performed in separate loop to avoid ConcurrentModificationException
        Iterator smIter = stateManagersToEvict.iterator();
        while (smIter.hasNext())
        {
            StateManager sm = (StateManager)smIter.next();
            PersistenceCapable pc = sm.getObject();
            sm.evict();

            // Evict from L1
            removeObjectFromCache(pc, pc.jdoGetObjectId(), true, false);
        }
    }

    // --------------------------------- Refresh ----------------------------------------

    /**
     * Internal method to do a refresh of an object, updating it from its
     * datastore representation. Also updates the object in the L1/L2 caches.
     * @param obj The Object
     */
    private void internalRefresh(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        try
        {
            clr.setPrimary(obj.getClass().getClassLoader());
            assertPersistenceCapable(obj);
            assertNotDetached(obj);
            PersistenceCapable pc = (PersistenceCapable) obj;
            StateManager sm = findStateManager(pc);
            if (sm == null)
            {
                throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
            }
            sm.refresh();
    
            // Refresh L2 cache object since it is a copy of this object
            putObjectIntoCache(sm, false, true);
        }
        finally
        {
            clr.unsetPrimary();
        }
    }

    /**
     * Method to do a refresh of an object.
     * @param pc The Object
     */
    public synchronized void refresh(Object pc)
    {
        assertIsOpen();
        internalRefresh(pc);
    }

    /**
     * Method to do a refresh of an array of objects.
     * @param pcs The Objects
     */
    public synchronized void refreshAll(Object[] pcs)
    {
        refreshAll(Arrays.asList(pcs));
    }

    /**
     * Method to do a refresh of a collection of objects.
     * @param pcs The Objects
     * @throws JDOUserException thrown if instances could not be refreshed.
     */
    public synchronized void refreshAll(Collection pcs)
    {
        assertIsOpen();
        ArrayList failures = new ArrayList();
        Iterator iter = pcs.iterator();
        while (iter.hasNext())
        {
            try
            {
                internalRefresh(iter.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.RefreshError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Method to do a refresh of all objects.
     * @throws JDOUserException thrown if instances could not be refreshed.
     */
    public synchronized void refreshAll()
    {
        assertIsOpen();

        List failures = new ArrayList();

        Set toRefresh = new HashSet();
        toRefresh.addAll(enlistedSMCache.values());
        toRefresh.addAll(dirtySMs);
        if (!tx.isActive())
        {
            toRefresh.addAll(cache.values());
        }

        Iterator iter = toRefresh.iterator();
        while (iter.hasNext())
        {
            try
            {
                Object obj = iter.next();
                StateManager sm;
                if (obj instanceof PersistenceCapable)
                {
                    PersistenceCapable pc = (PersistenceCapable)obj;
                    sm = findStateManager(pc);
                }
                else
                {
                    sm = (StateManager) obj;
                }
                sm.refresh();

                // Refresh L2 cache object since it is a copy of this object
                putObjectIntoCache(sm, false, true);
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.RefreshError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Method to do a refresh of objects that failed verification in the
     * exception.
     * @param exc The JDO exception containing the objects that failed
     * @since 1.1
     */
    public synchronized void refreshAll(JDOException exc)
    {
        Object obj = exc.getFailedObject();
        if (obj != null)
        {
            refresh(obj);
        }

        Throwable[] nested_excs = exc.getNestedExceptions();
        if (nested_excs != null)
        {
            for (int i = 0; i < nested_excs.length; i++)
            {
                if (nested_excs[i] instanceof JDOException)
                {
                    refreshAll((JDOException) nested_excs[i]);
                }
            }
        }
    }

    // ------------------------------- Retrieve ------------------------------------------

    /**
     * Method to retrieve an object (internally).
     * @param obj The object
     * @param fgOnly Whether to retrieve the current fetch group fields only
     */
    private void internalRetrieve(Object obj, boolean fgOnly)
    {
        if (obj == null)
        {
            return;
        }
        try
        {
            clr.setPrimary(obj.getClass().getClassLoader());
            assertPersistenceCapable(obj);
            assertNotDetached(obj);
            PersistenceCapable pc = (PersistenceCapable) obj;
            StateManager sm = findStateManager(pc);
            if (sm == null)
            {
                throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
            }
            sm.retrieve(fgOnly);
        }
        finally
        {
            clr.unsetPrimary();
        }
    }

    /**
     * Method to retrieve an object (internally).
     * @param obj The object
     * @param fetchPlan the fetch plan to load fields
     */
    public synchronized void internalRetrieve(Object obj, FetchPlan fetchPlan)
    {
        if (obj == null)
        {
            return;
        }
        try
        {
            clr.setPrimary(obj.getClass().getClassLoader());
            assertPersistenceCapable(obj);
            assertNotDetached(obj);
            PersistenceCapable pc = (PersistenceCapable) obj;
            StateManager sm = findStateManager(pc);
            if (sm == null)
            {
                throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
            }
            sm.retrieve(fetchPlan);
        }
        finally
        {
            clr.unsetPrimary();
        }
    }
    
    /**
     * Method to retrieve the fields of an object.
     * @param pc The object
     * @param fgOnly Whether to retrieve the current fetch group fields only
     */
    public synchronized void retrieve(Object pc, boolean fgOnly)
    {
        assertIsOpen();
        internalRetrieve(pc, fgOnly);
    }
    
    /**
     * Method to retrieve the fields of an object.
     * @param pc The object
     * @param fetchPlan the fetch plan to load fields
     */
    public synchronized void retrieve(Object pc, FetchPlan fetchPlan)
    {
        assertIsOpen();
        internalRetrieve(pc, fetchPlan);
    }
    

    /**
     * Method to retrieve the fields of an object.
     * @param pc The object
     */
    public synchronized void retrieve(Object pc)
    {
        retrieve(pc, false);
    }

    /**
     * Method to retrieve an array of objects.
     * @param pcs The objects
     */
    public synchronized void retrieveAll(Object[] pcs)
    {
        retrieveAll(Arrays.asList(pcs), false);
    }

    /**
     * Retrieve field values of instances from the store. This tells the
     * <code>PersistenceManager</code> that the application intends to use the
     * instances, and their field values should be retrieved. The fields in the
     * current fetch group must be retrieved, and the implementation might
     * retrieve more fields than the current fetch group.
     * <P>
     * The <code>PersistenceManager</code> might use policy information about
     * the class to retrieve associated instances.
     * @param pcs the instances
     * @param fgOnly whether to retrieve only the current fetch group fields
     * @since JDO1.0.1
     */
    public void retrieveAll(Object[] pcs, boolean fgOnly)
    {
        retrieveAll(Arrays.asList(pcs), fgOnly);
    }

    /**
     * Retrieve field values of instances from the store. This tells the
     * <code>PersistenceManager</code> that the application intends to use the
     * instances, and their field values should be retrieved. The fields in the
     * current fetch group must be retrieved, and the implementation might
     * retrieve more fields than the current fetch group.
     * <P>
     * The <code>PersistenceManager</code> might use policy information about
     * the class to retrieve associated instances.
     * @param pcs the instances
     * @param fgOnly whether to retrieve only the current fetch-group fields
     * @since JDO1.0.1
     */
    public void retrieveAll(Collection pcs, boolean fgOnly)
    {
        assertIsOpen();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalRetrieve(i.next(), fgOnly);
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.RetrieveError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Method to retrieve a collection of objects. Throws a JDOUserException if
     * instances could not be retrieved.
     * @param pcs The objects
     */
    public synchronized void retrieveAll(Collection pcs)
    {
        retrieveAll(pcs.toArray(), false);
    }

    // ---------------------------------- Make Persistent ---------------------------------------

    /**
     * Method to make an object persistent.
     * Throws a JDOUserException if the object is managed by a different PM.
     * @param obj The object
     * @param preInsertChanges Any changes to make before inserting
     * @return The persisted object
     */
    private Object internalMakePersistent(Object obj, FieldValues preInsertChanges)
    {
        if (obj == null)
        {
            return null;
        }
        try
        {
            clr.setPrimary(obj.getClass().getClassLoader());
            assertPersistenceCapable(obj);
            PersistenceCapable pc = (PersistenceCapable) obj;
            if (!pc.jdoIsDetached() && JPOXLogger.JDO.isDebugEnabled())
            {
                try
                {
                    JPOXLogger.JDO.debug(LOCALISER.msg("PM.MakePersistent", pc));
                }
                catch (Exception e)
                {
                    // we catch any exception that can be raised inside the PC.toString() method
                    JPOXLogger.JDO.debug(LOCALISER.msg("PM.MakePersistent", pc.getClass().getName()));
                }
            }
    
            javax.jdo.PersistenceManager pm = pc.jdoGetPersistenceManager();
            if (pm != null)
            {
                if (!(pm instanceof PersistenceManager) || ((PersistenceManager)pm).getPMHandle() != this.getPMHandle())
                {
                    // Object managed by a different PM
                    throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", obj));
                }
            }
    
            Object persistedPc = pc; // Persisted object is the passed in pc (unless being attached)
            if (pc.jdoIsDetached())
            {
                // Detached object so attach it
                assertDetachable(obj);
                persistedPc = attachCopy(pc, false);
            }
            else if (pc.jdoIsTransactional() && !pc.jdoIsPersistent())
            {
                // TransientTransactional object to be persisted
                StateManager sm = findStateManager(pc);
                if (sm == null)
                {
                    throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
                }
                sm.makePersistentTransactionalTransient();
            }
            else if (!pc.jdoIsPersistent())
            {
                // Transient object, so persist it
                if (this.multithreaded)
                {
                    synchronized (pc)
                    {
                        StateManager sm = findStateManager(pc);
                        if (sm == null)
                        {
                            sm = StateManagerFactory.newStateManager(this, pc, preInsertChanges);
                            sm.makePersistent();
                        }
                        else
                        {
                        	if (sm.getAttachedPC() == null)
                        	{
                                sm.makePersistent();
                        	}
                        	else
                        	{
                        		// Object is being attached right now
                        		persistedPc = sm.getAttachedPC();
                        	}
                        }
                    }
                }
                else
                {
                    StateManager sm = findStateManager(pc);
                    if (sm == null)
                    {
                        sm = StateManagerFactory.newStateManager(this, pc, preInsertChanges);
                        sm.makePersistent();
                    }
                    else
                    {
                    	if (sm.getAttachedPC() == null)
                    	{
                            sm.makePersistent();
                    	}
                    	else
                    	{
                    		// Object is being attached right now
                    		persistedPc = sm.getAttachedPC();
                    	}
                    }
                }
            }
            else if (pc.jdoIsPersistent() && pc.jdoGetObjectId() == null)
            {
                // Embedded/Serialised objects have a StateManager yet have no identity.
                // They can be persisted in their own right, so we allow this.
                // Should we be making a copy of the object here ?
                if (this.multithreaded)
                {
                    synchronized (pc)
                    {
                        StateManager sm = findStateManager(pc);
                        sm.makePersistent();
                    }
                }
                else
                {
                    StateManager sm = findStateManager(pc);
                    sm.makePersistent();
                }
            }
            return persistedPc;
        }
        finally
        {
            clr.unsetPrimary();
        }
    }

    /**
     * Internal method to make persistent an object.
     * <B>For internal use only</B>
     * @param obj The object
     * @return The persisted object
     */
    public synchronized Object makePersistentInternal(Object obj, FieldValues preInsertChanges)
    {
        assertIsOpen();
        assertActiveTransaction();
        return internalMakePersistent(obj, preInsertChanges);
    }

    /**
     * JDO method to make persistent an object.
     * @param obj The object
     */
    public synchronized Object makePersistent(Object obj)
    {
        assertIsOpen();
        assertActiveTransaction();
        if (obj == null)
        {
            return null;
        }

        // Persist the object
        boolean detached = JDOHelper.isDetached(obj);
        Object persistedPc = internalMakePersistent(obj, null);
        StateManager sm = findStateManager((PersistenceCapable)persistedPc);
        if (apmf.getPersistenceByReachabilityAtCommit())
        {
            if (sm != null)
            {
                if (detached || sm.isNew((PersistenceCapable)persistedPc))
                {
                    txKnownPersists.add(sm.getInternalObjectId());
                }
            }
        }
        return persistedPc;
    }

    /**
     * JDO method to make persistent an array of objects.
     * @param pcs The objects to persist
     */
    public synchronized Object[] makePersistentAll(Object[] pcs)
    {
        return makePersistentAll(Arrays.asList(pcs)).toArray();
    }

    /**
     * JDO method to make persistent a collection of objects.
     * Throws a JDOUserException if objects could not be made persistent.
     * @param pcs The objects to persist
     */
    public synchronized Collection makePersistentAll(Collection pcs)
    {
        assertIsOpen();
        assertActiveTransaction();

        List failures = new ArrayList();
        Iterator i = pcs.iterator();
        Collection persistedPcs = new ArrayList();
        while (i.hasNext())
        {
            try
            {
                Object persistedPc = makePersistent(i.next());
                persistedPcs.add(persistedPc);
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakePersistentError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }

        return persistedPcs;
    }

    /**
     * Tests whether this object is in the process of being inserted.
     * @param pc the object to verify the status
     * @return true if this instance is inserting.
     */
    public boolean isInserting(PersistenceCapable pc)
    {
        StateManager sm = findStateManager(pc);
        if (sm == null)
        {
            return false;
        }
        return sm.isInserting();
    }

    /**
     * Convenience method to find if the specified field of the specified object
     * has been inserted yet.
     * @param pc The object
     * @param fieldNumber The absolute field number
     * @return Whether it has been inserted into the datastore
     */
    public boolean isInserted(PersistenceCapable pc, int fieldNumber)
    {
        StateManager sm = findStateManager(pc);
        if (sm == null)
        {
            return false;
        }
        return sm.isInserted(fieldNumber);
    }

    /**
     * Method to attach a persistent detached object.
     * If the object is of class that is not detachable, a ClassNotDetachableException
     * will be thrown.
     * TODO Merge this with internalMakePersistent().
     * @param pc The object
     * @param embedded Whether the object is embedded
     * @return The attached object
     */
    public synchronized Object attachCopy(Object pc, boolean embedded)
    {
        assertIsOpen();
        assertPersistenceCapable(pc);
        assertDetachable(pc);

        Object id = ((PersistenceCapable) pc).jdoGetObjectId();
        if (id != null && isInserting((PersistenceCapable)pc))
        {
            // Object is being inserted in this transaction
            return pc;
        }
        else if (id == null && !embedded)
        {
            // Object was not persisted before so persist it
            return makePersistentInternal(pc, null);
        }
        else if (((PersistenceCapable)pc).jdoIsPersistent())
        {
            // Already persistent hence can't be attached
            return pc;
        }

        if (JPOXLogger.JDO.isDebugEnabled())
        {
            try
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("PM.MakePersistentAttach", pc));
            }
            catch (Exception e)
            {
                // we catch any exception that can be raised inside the PC.toString() method
                JPOXLogger.JDO.debug(LOCALISER.msg("PM.MakePersistentAttach", pc.getClass().getName()));
            }
        }

        // Object should exist in this datastore now
        PersistenceCapable pcTarget = null;
        if (embedded)
        {
            boolean detached = false;
            if (JDOHelper.isDetached(pc))
            {
                detached = true;
            }
            StateManager smTarget = StateManagerFactory.newStateManager((PersistenceCapable)pc, this, true);
            pcTarget = smTarget.getObject();
            if (detached)
            {
                // If the object had previously been detached, re-attach it
                smTarget.attachCopy((PersistenceCapable)pc, embedded);
            }
        }
        else
        {
            pcTarget = (PersistenceCapable) getObjectById(id, false, false, pc.getClass().getName());
            findStateManager(pcTarget).attachCopy((PersistenceCapable) pc, embedded);
        }

        return pcTarget;
    }

    // ------------------------------- Delete Persistent ------------------------------------------

    /**
     * Method to delete an object from the datastore (internally).
     * @param obj The object
     */
    private void internalDeletePersistent(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        try
        {
            clr.setPrimary(obj.getClass().getClassLoader());
            assertPersistenceCapable(obj);
            PersistenceCapable pc = (PersistenceCapable) obj;
            if (pc.jdoIsDetached())
            {
                // Load up the attached instance with this identity
                pc = (PersistenceCapable)getObjectById(pc.jdoGetObjectId());
            }
    
            if (JPOXLogger.JDO.isDebugEnabled())
            {
                try
                {
                    JPOXLogger.JDO.debug(LOCALISER.msg("PM.DeletePersistent", pc));
                }
                catch (Exception e)
                {
                    // we catch any exception that can be raised inside the PC.toString() method
                    JPOXLogger.JDO.debug(LOCALISER.msg("PM.DeletePersistent", pc.getClass().getName()));
                }
            }
    
            // Check that the object is valid for deleting
            if (!pc.jdoIsPersistent() && !pc.jdoIsTransactional())
            {
                throw new JDOUserException(LOCALISER.msg("PM.DeletePersistentOnTransientInstances"));
            }
            else if (!pc.jdoIsPersistent() && pc.jdoIsTransactional())
            {
                throw new JDOUserException(LOCALISER.msg("PM.DeletePersistentOnTransactionalInstances"));
            }
    
            // Delete it
            StateManager sm = findStateManager(pc);
            if (sm == null)
            {
                throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
            }
            sm.deletePersistent();
        }
        finally
        {
            clr.unsetPrimary();
        }
    }

    /**
     * JDO method to delete an object from the datastore.
     * @param obj The object
     */
    public synchronized void deletePersistent(Object obj)
    {
        assertIsOpen();
        assertActiveTransaction();
        internalDeletePersistent(obj);
    }

    /**
     * JDO method to delete an array of objects from the datastore.
     * @param pcs The objects
     */
    public synchronized void deletePersistentAll(Object[] pcs)
    {
        deletePersistentAll(Arrays.asList(pcs));
    }

    /**
     * JDO method to delete a collection of objects from the datastore. 
     * Throws a JDOUserException if objects could not be deleted.
     * @param pcs The objects
     */
    public synchronized void deletePersistentAll(Collection pcs)
    {
        assertIsOpen();
        assertActiveTransaction();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalDeletePersistent(i.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.DeletePersistentError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    // -------------------------------- Make Transient -----------------------------------------

    /**
     * Method to make transient an object (internally).
     * @param obj The object
     * @param state Object containing the state of the fetch plan process (if any)
     */
    public synchronized void internalMakeTransient(Object obj, FetchPlanState state)
    {
        if (obj == null)
        {
            return;
        }
        try
        {
            clr.setPrimary(obj.getClass().getClassLoader());
            assertPersistenceCapable(obj);
            assertNotDetached(obj);
            
            PersistenceCapable pc = (PersistenceCapable) obj;
            if (JPOXLogger.JDO.isDebugEnabled())
            {
                try
                {
                    JPOXLogger.JDO.debug(LOCALISER.msg("PM.MakeTransient", pc));
                }
                catch (Exception e)
                {
                    // we catch any exception that can be raised inside the PC.toString() method
                    JPOXLogger.JDO.debug(LOCALISER.msg("PM.MakeTransient", pc.getClass().getName()));
                }
            }
    
            if (pc.jdoIsPersistent())
            {
                StateManager sm = findStateManager(pc);
                sm.makeTransient(state);
            }
        }
        finally
        {
            clr.unsetPrimary();
        }
    }

    /**
     * Method to make transient an object allowing fetching using the fetch plan.
     * @param pc The object
     * @param useFetchPlan Whether to make transient all objects in the fetch plan
     */
    public synchronized void makeTransient(Object pc, boolean useFetchPlan)
    {
        assertIsOpen();

        FetchPlanState state = null;
        if (useFetchPlan)
        {
            // Create a state object to carry the processing state info
            state = new FetchPlanState();
        }
        internalMakeTransient(pc, state);
    }

    /**
     * Method to make transient an object.
     * This doesn't use the fetch plan and just makes the specified object transient.
     * @param pc The object
     */
    public synchronized void makeTransient(Object pc)
    {
        makeTransient(pc, false);
    }

    /**
     * Method to make transient an array of objects.
     * @param pcs The objects
     */
    public synchronized void makeTransientAll(Object[] pcs)
    {
        makeTransientAll(Arrays.asList(pcs));
    }

    /**
     * Method to make transient an array of objects.
     * @param pcs The objects
     * @param includeFetchPlan Whether to make transient all objects in the fetch plan
     */
    public synchronized void makeTransientAll(Object[] pcs, boolean includeFetchPlan)
    {
        makeTransientAll(Arrays.asList(pcs), includeFetchPlan);
    }

    /**
     * Method to make transient a collection of objects.
     * @param pcs The objects
     * @param useFetchPlan Whether to use the fetch plan when making transient
     * @throws JDOUserException thrown if objects could not be made transient.
     */
    public synchronized void makeTransientAll(Collection pcs, boolean useFetchPlan)
    {
        assertIsOpen();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        FetchPlanState state = null;
        if (useFetchPlan)
        {
            // Create a state object to carry the processing state info
            state = new FetchPlanState();
        }
        while (i.hasNext())
        {
            try
            {
                internalMakeTransient(i.next(), state);
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakeTransientError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Method to make transient a collection of objects.
     * @param pcs The objects
     * @throws JDOUserException thrown if objects could not be made transient.
     */
    public synchronized void makeTransientAll(Collection pcs)
    {
        makeTransientAll(pcs, false);
    }

    // ----------------------------------- Make Transactional --------------------------------------

    /**
     * Method to make transactional an object (internally).
     * @param obj The object
     */
    private void internalMakeTransactional(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        try
        {
            clr.setPrimary(obj.getClass().getClassLoader());
            assertPersistenceCapable(obj);
            assertNotDetached(obj);
            PersistenceCapable pc = (PersistenceCapable) obj;
            if (pc.jdoIsPersistent())
            {
                assertActiveTransaction();
            }
            StateManager sm = findStateManager(pc);
            if (sm == null)
            {
                sm = StateManagerFactory.newStateManager(this, pc);
            }
            sm.makeTransactional();
        }
        finally
        {
            clr.unsetPrimary();
        }
    }

    /**
     * Method to make transactional an object.
     * @param pc The object
     */
    public synchronized void makeTransactional(Object pc)
    {
        assertIsOpen();
        internalMakeTransactional(pc);
    }

    /**
     * Method to make transactional an array of objects.
     * @param pcs The objects
     */
    public synchronized void makeTransactionalAll(Object[] pcs)
    {
        makeTransactionalAll(Arrays.asList(pcs));
    }

    /**
     * Method to make transactional a collection of objects.
     * @param pcs The objects
     * @throws JDOUserException thrown if objects could not be made transactional
     */
    public synchronized void makeTransactionalAll(Collection pcs)
    {
        assertIsOpen();
        assertActiveTransaction();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalMakeTransactional(i.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakeTransactionalError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    // ------------------------------ Make NonTransactional -------------------------------------------

    /**
     * Method to make nontransactional an object (internally).
     * @param obj The object
     */
    private void internalMakeNontransactional(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        try
        {
            clr.setPrimary(obj.getClass().getClassLoader());
            assertPersistenceCapable(obj);
            PersistenceCapable pc = (PersistenceCapable) obj;
            if (!pc.jdoIsPersistent() && pc.jdoIsTransactional() && pc.jdoIsDirty())
            {
                throw new JDOUserException(LOCALISER.msg("PM.MakeNonTransactionalOnTransientDirtyInstances"));
            }
            StateManager sm = findStateManager(pc);
            sm.makeNontransactional();
        }
        finally
        {
            clr.unsetPrimary();
        }
    }

    /**
     * Method to make nontransactional an object.
     * @param pc The object
     */
    public synchronized void makeNontransactional(Object pc)
    {
        assertIsOpen();
        if (pc == null)
        {
            // JDO2 [12.6.0] Null means do nothing
            return;
        }

        // for non transactional read, tx might be not active
        // TODO add verification if is non transactional read = true
        if (tx.isActive())
        {
            assertActiveTransaction();
        }
        // if !transactional and !persistent
        if (!((PersistenceCapable) pc).jdoIsTransactional() && !((PersistenceCapable) pc).jdoIsPersistent())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakeNontransactionalOnNontransactionalTransientInstances"));
        }
        // if !transactional and persistent, do nothing
        if (!((PersistenceCapable) pc).jdoIsTransactional() && ((PersistenceCapable) pc).jdoIsPersistent())
        {
            return;
        }
        internalMakeNontransactional(pc);
    }

    /**
     * Method to make nontransactional an array of objects.
     * @param pcs The objects.
     */
    public synchronized void makeNontransactionalAll(Object[] pcs)
    {
        makeNontransactionalAll(Arrays.asList(pcs));
    }

    /**
     * Method to make nontransactional a collection of objects.
     * @param pcs The objects.
     * @throws JDOUserException thrown if objects could not be made nontransactional
     */
    public synchronized void makeNontransactionalAll(Collection pcs)
    {
        assertIsOpen();
        assertActiveTransaction();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalMakeNontransactional(i.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakeNonTransactionalError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    // ------------------------- Attach/Detach instances -----------------------

    /**
     * Method to detach a persistent object without making a copy. Note that 
     * also all the objects which are refered to from this object are detached.
     * If the object is of class that is not detachable a ClassNotDetachableException
     * will be thrown. If the object is not persistent a JDOUserException is thrown.
     * <B>For internal use only</B>
     * @param pc The object
     * @param state State for the detachment process
     */
    public synchronized void detachInternal(Object pc, FetchPlanState state)
    {
        assertIsOpen();
        assertPersistenceCapable(pc);
        assertDetachable(pc);

        if (JDOHelper.isDetached(pc))
        {
            return;
        }

        if (!JDOHelper.isPersistent(pc))
        {
            // Transient object passed so persist it before thinking about detaching
            if (tx.isActive())
            {
                internalMakePersistent(pc, null);
            }
        }

        StateManager sm = findStateManager((PersistenceCapable) pc);
        if (sm == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", ((PersistenceCapable) pc).jdoGetObjectId()));
        }
        sm.detach(state);
    }

    /**
     * Internal method to detach a persistent object using the provided detach state.
     * @param pc The object
     * @param state State for the detachment process
     * @return The detached object
     */
    private Object internalDetachCopy(Object pc, FetchPlanState state)
    {
        Object thePC = pc;
        try
        {
            clr.setPrimary(pc.getClass().getClassLoader());
            if (!JDOHelper.isPersistent(pc) && !JDOHelper.isDetached(pc))
            {
                // Transient object passed so persist it before thinking about detaching
                if (tx.isActive())
                {
                    thePC = internalMakePersistent(pc, null);
                }
                else
                {
                    // JDO2 [12.6.8] "If a detachCopy method is called outside an active transaction, the reachability algorithm
                    // will not be run; if any transient instances are reachable via persistent fields, a JDOUserException is thrown 
                    // for each persistent instance containing such fields.
                    throw new JDOUserException(LOCALISER.msg("PM.Detach.TransientOutsideTransaction"));
                }
            }

            if (((PersistenceCapable)thePC).jdoIsDetached())
            {
                // Passing in a detached (dirty or clean) instance, so get a persistent copy to detach
                thePC = getObjectById(JDOHelper.getObjectId(thePC), false);
            }
    
            PersistenceCapable detached = ((DetachState)state).getDetachedCopyObject(thePC);
            if (detached == null)
            {
                StateManager sm = findStateManager((PersistenceCapable) thePC);
                if (sm == null)
                {
                    throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", ((PersistenceCapable) thePC).jdoGetObjectId()));
                }

                detached = (PersistenceCapable) sm.detachCopy(state);
                ((DetachState)state).setDetachedCopyObject(detached, sm.getExternalObjectId(sm.getObject()));
            }
            else
            {
                // What if the fetch-group here implies a different depth through this object ? we need to continue detaching
                // TODO Check the detach and proceed further where necessary
            }
            return detached;
        }
        finally
        {
            clr.unsetPrimary();
        }
    }

    /**
     * Internal method to detach a persistent object using the passed state.
     * If the object is of class that is not detachable a ClassNotDetachableException 
     * will be thrown. If the object is not persistent a JDOUserException is thrown.
     * <B>For internal use only</B>
     * @param pc The object
     * @param state The detached state
     * @return The detached object
     */
    public synchronized Object detachCopyInternal(Object pc, FetchPlanState state)
    {
        assertIsOpen();
        assertPersistenceCapable(pc);

        return internalDetachCopy(pc, state);
    }

    /**
     * JDO method to detach a persistent object.
     * If the object is of class that is not detachable a ClassNotDetachableException 
     * will be thrown. If the object is not persistent a JDOUserException is thrown.
     * @param pc The object
     * @return The detached object
     */
    public synchronized Object detachCopy(Object pc)
    {
        assertIsOpen();
        if (pc == null)
        {
            // Stupid user
            return null;
        }
        assertPersistenceCapable(pc);
        assertActiveTransactionOrNontransactionRead("detachCopy");

        return internalDetachCopy(pc, new DetachState());
    }

    /**
     * Detach the specified objects from the <code>PersistenceManager</code>. 
     * The objects returned can be manipulated and re-attached with 
     * {@link #makePersistentAll(Object[])}.
     * The detached instances will be unmanaged copies of the specified parameters, 
     * and are suitable for serialization and manipulation outside of a JDO environment.
     * When detaching instances, only fields in the current {@link FetchPlan} will be 
     * traversed. Thus, to detach a graph of objects, relations to other persistent 
     * instances must either be in the <code>default-fetch-group</code>, or in the
     * current custom {@link FetchPlan}.
     * @param pcs the instances to detach
     * @return the detached instances
     * @throws JDOUserException if any of the instances do not
     * @throws ClassNotDetachableException If any of the objects are of a class that is not detachable
     * @since JDO 2.0
     */
    public synchronized Object[] detachCopyAll(Object[] pcs)
    {
        return detachCopyAll(Arrays.asList(pcs)).toArray();
    }

    /**
     * Detach the specified objects from the <code>PersistenceManager</code>.
     * @param pcs the instances to detach
     * @return the detached instances
     * @see #detachCopyAll(Object[])
     * @since JDO 2.0
     */
    public synchronized Collection detachCopyAll(Collection pcs)
    {
        assertIsOpen();
        assertActiveTransactionOrNontransactionRead("detachCopyAll");

        // Detach the objects
        FetchPlanState state = new DetachState();
        List detacheds = new ArrayList();
        for (Iterator it = pcs.iterator(); it.hasNext();)
        {
            Object obj = it.next();
            if (obj == null)
            {
                detacheds.add(null);
            }
            else
            {
                detacheds.add(detachCopyInternal(obj, state));
            }
        }

        return detacheds;
    }

    // --------------------------------- Queries ----------------------------------------

    /**
     * Construct an empty query instance.
     * @return The query
     */
    public synchronized Query newQuery()
    {
        return newQuery("javax.jdo.query.JDOQL", null);
    }

    /**
     * Construct a query instance from another query. The parameter might be a
     * serialized/restored Query instance from the same JDO vendor but a
     * different execution environment, or the parameter might be currently
     * bound to a PersistenceManager from the same JDO vendor. Any of the
     * elements Class, Filter, IgnoreCache flag, Import declarations, Variable
     * declarations, Parameter declarations, and Ordering from the parameter
     * Query are copied to the new Query instance, but a candidate Collection or
     * Extent element is discarded.
     * @param obj The object to use in the query
     * @return The query
     */
    public synchronized Query newQuery(Object obj)
    {
        // TODO Remove the references to SQLQuery/JPOXSQLQuery from here if possible so we have no links to things that may be RDBMS-specific
        if (obj instanceof SQLQuery)
        {
            return newQuery("javax.jdo.query.SQL", obj);
        }
        else if (obj instanceof JPOXSQLQuery)
        {
            return newQuery("javax.jdo.query.JPOXSQL", obj);
        }
        else
        {
            return newQuery("javax.jdo.query.JDOQL", obj);
        }
    }

    /**
     * Construct a query instance using the specified Single-String query.
     * @param query The single-string query
     * @return The Query
     */
    public synchronized Query newQuery(String query)
    {
        return newQuery("javax.jdo.query.JDOQL", query);
    }

    /**
     * Construct a query instance using the specified language and the specified
     * query. The query instance will be of a class defined by the query
     * language.
     * @param language The language parameter for the JDO Query language. This
     * is by default "javax.jdo.query.JDOQL", but in JDO 2.0 can also be
     * "javax.jdo.query.SQL", or vendor provided languages.
     * @param query The query object
     * @return The query
     */
    public synchronized Query newQuery(String language, Object query)
    {
        assertIsOpen();

        if (language == null)
        {
            language = "javax.jdo.query.JDOQL";
        }

        // Check that our store supports the language
        if (!srm.supportsQueryLanguage(language))
        {
            throw new JDOUserException("PM.Query.LanguageNotSupportedByStore");
        }

    //        throw new JDOUserException(LOCALISER.msg("PM.Query.LanguageUnknown", language));

        return srm.newQuery(language, getPMHandle(), query);
    }

    /**
     * Construct a query instance with the candidate class specified.
     * @param cls The class to query
     * @return The query
     */
    public synchronized Query newQuery(Class cls)
    {
        Query query = newQuery();
        query.setClass(cls);
        return query;
    }

    /**
     * Construct a query instance with the candidate Extent specified; the
     * candidate class is taken from the Extent.
     * @param cln The extent to query
     * @return The query
     */
    public synchronized Query newQuery(Extent cln)
    {
        Query query = newQuery();
        query.setClass(cln.getCandidateClass());
        query.setCandidates(cln);
        return query;
    }

    /**
     * Construct a query instance with the candidate class and candidate
     * Collection specified.
     * @param cls The class to query
     * @param cln The collection
     * @return The query
     */
    public synchronized Query newQuery(Class cls, Collection cln)
    {
        Query query = newQuery();
        query.setClass(cls);
        query.setCandidates(cln);
        return query;
    }

    /**
     * Construct a query instance with the candidate class and filter specified.
     * @param cls The class to query
     * @param filter A filter to apply
     * @return The query
     */
    public synchronized Query newQuery(Class cls, String filter)
    {
        Query query = newQuery();
        query.setClass(cls);
        query.setFilter(filter);
        return query;
    }

    /**
     * Construct a query instance with the candidate class, the candidate
     * Collection, and filter specified.
     * @param cls The class to query
     * @param cln A collection
     * @param filter A filter to apply
     * @return The query
     */
    public synchronized Query newQuery(Class cls, Collection cln, String filter)
    {
        Query query = newQuery();
        query.setClass(cls);
        query.setCandidates(cln);
        query.setFilter(filter);
        return query;
    }

    /**
     * Construct a query instance with the candidate Extent and filter
     * specified. The candidate class is taken from the Extent.
     * @param cln The extent to query
     * @param filter A filter to apply
     * @return The query
     */
    public synchronized Query newQuery(Extent cln, String filter)
    {
        Query query = newQuery();
        query.setClass(cln.getCandidateClass());
        query.setCandidates(cln);
        query.setFilter(filter);
        return query;
    }

    /**
     * Construct a query instance with the candidate class and the query name.
     * @param cls The class to query
     * @param queryName Name of the query.
     * @return The query
     */
    public synchronized Query newNamedQuery(Class cls, String queryName)
    {
        assertIsOpen();

        // Throw exception on incomplete input
        if (queryName == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.Query.NamedQueryNotFound", queryName, cls));
        }

        // Find the Query for the specified class
        QueryMetaData qmd = getMetaDataManager().getMetaDataForQuery(cls, clr, queryName);
        if (qmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.Query.NamedQueryNotFound", queryName, cls));
        }

        // Create the Query
        Query query = newQuery(qmd.getLanguage().toString(), qmd.getQuery());
        if (cls != null)
        {
            query.setClass(cls);
            if (!srm.managesClass(cls.getName()))
            {
                // Load the candidate class since not yet managed
                srm.addClass(cls.getName(), clr);
            }
        }

        // Optional args that should only be used with SQL
        if (qmd.getLanguage() == QueryLanguage.JDOQL && (qmd.isUnique() || qmd.getResultClass() != null))
        {
            throw new JDOUserException(LOCALISER.msg("PM.Query.NamedQueryJDOQLOnlySingleString", queryName));
        }
        if (qmd.isUnique())
        {
            query.setUnique(true);
        }
        if (qmd.getResultClass() != null)
        {
            // Set the result class, allowing for it being in the same package as the candidate
            Class resultCls = null;
            try
            {
                resultCls = clr.classForName(qmd.getResultClass());
            }
            catch (ClassNotResolvedException cnre)
            {
                try
                {
                    String resultClassName = cls.getPackage().getName() + "." + qmd.getResultClass();
                    resultCls = clr.classForName(resultClassName);
                }
                catch (ClassNotResolvedException cnre2)
                {
                    throw new JDOUserException(LOCALISER.msg("PM.Query.NamedQueryResultClassNotFound", queryName, qmd.getResultClass()));
                }
            }
            query.setResultClass(resultCls);
        }

        // Add any JPOX extensions
        if (qmd.getLanguage() == QueryLanguage.JPOXSQL)
        {
            // Apply any imports specified
            if (qmd.hasExtension("imports"))
            {
                query.declareImports(qmd.getValueForExtension("imports"));
            }
            // Apply any parameters specified
            if (qmd.hasExtension("parameters"))
            {
                query.declareParameters(qmd.getValueForExtension("parameters"));
            }
        }
        if (qmd.isUnmodifiable())
        {
            query.setUnmodifiable();
        }

        return query;
    }

    // ------------------------------- Extents ------------------------------------------

    /**
     * Extents are collections of datastore objects managed by the datastore,
     * not by explicit user operations on collections. Extent capability is a
     * boolean property of classes that are persistence capable. If an instance
     * of a class that has a managed extent is made persistent via reachability,
     * the instance is put into the extent implicitly.
     * @param pcClass The class to query
     * @param subclasses Whether to include subclasses in the query.
     * @return returns an Extent that contains all of the instances in the
     * parameter class, and if the subclasses flag is true, all of the instances
     * of the parameter class and its subclasses.
     */
    public synchronized Extent getExtent(Class pcClass, boolean subclasses)
    {
        assertIsOpen();
        try
        {
            clr.setPrimary(pcClass.getClassLoader());
            assertPersistenceCapableClass(pcClass);
            return srm.getExtent(getPMHandle(), pcClass, subclasses);
        }
        finally
        {
            clr.unsetPrimary();
        }
    }
    
    /**
     * Extents are collections of datastore objects managed by the datastore,
     * not by explicit user operations on collections. Extent capability is a
     * boolean property of classes that are persistence capable. If an instance
     * of a class that has a managed extent is made persistent via reachability,
     * the instance is put into the extent implicitly.
     * @param pcClass The class to query
     * @return returns an Extent that contains all of the instances in the
     * parameter class, and all of the instances of the parameter class and its
     * subclasses.
     * @since 1.1
     */
    public synchronized Extent getExtent(Class pcClass)
    {
        return getExtent(pcClass, true);
    }

    // ----------------------------- New Instances ----------------------------------

    /**
     * Method to generate an instance of an interface, abstract class, or concrete PC class.
     * @param persistenceCapable The class of the interface or abstract class, or concrete class defined in MetaData
     * @return The instance of this type
     */
    public Object newInstance(Class persistenceCapable)
    {
        assertIsOpen();
        return apmf.getImplementationCreator().newInstance(persistenceCapable, getMetaDataManager(), getClassLoaderResolver());
    }

    /**
     * This method returns an object id instance corresponding to the pcClass and key arguments.
     * It has 2 modes of operation. Where SingleFieldIdentity is being used the key is the
     * value of the key field. For all other cases the key is the String form of the
     * object id instance.
     * @param pcClass Class of the PersistenceCapable to create the OID for.
     * @param key Value of the key for SingleFieldIdentity
     * @return The new object-id instance
     */
    public Object newObjectIdInstance(Class pcClass, Object key)
    {
        assertIsOpen();
        if (pcClass == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.ObjectId.ClassIsNull"));
        }
        assertPersistenceCapableClass(pcClass);

        AbstractClassMetaData cmd = getMetaDataManager().getMetaDataForClass(pcClass, clr);
        if (cmd == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("PM.ObjectId.ClassHasNoMetaData", pcClass.getName()));
        }

        // If the class is not yet managed, manage it
        if (!srm.managesClass(cmd.getFullClassName()))
        {
            srm.addClass(cmd.getFullClassName(), getClassLoaderResolver());
        }

        Object id = null;
        if (cmd.usesSingleFieldIdentityClass())
        {
            // Single Field Identity
            Class idType = clr.classForName(cmd.getObjectidClass());
            id = AIDUtils.getNewSingleFieldIdentity(idType, pcClass, key);
        }
        else if (key instanceof java.lang.String)
        {
            // String-based PK (datastore identity or application identity)
            if (cmd.getIdentityType() == IdentityType.APPLICATION)
            {
                if (Modifier.isAbstract(pcClass.getModifiers()) && cmd.getObjectidClass() != null) 
                {
                    try
                    {
                        Constructor c = clr.classForName(cmd.getObjectidClass()).getDeclaredConstructor(new Class[] {java.lang.String.class});
                        id = c.newInstance(new Object[] {(String)key});
                    }
                    catch(Exception e) 
                    {
                        String msg = LOCALISER.msg("PM.NewObjectIdInstance.ConstructionError", cmd.getObjectidClass(), cmd.getFullClassName());
                        JPOXLogger.JDO.error(msg);
                        JPOXLogger.JDO.error(e);

                        throw new JDOUserException(msg);
                    }
                }
                else
                {
                    getClassLoaderResolver().classForName(pcClass.getName(), true);
                    id = JDOImplHelper.getInstance().newObjectIdInstance(pcClass, key);
                }
            }
            else
            {
                id = new OID((String)key);
            }
        }
        else
        {
            // Key is not a string, and is not SingleFieldIdentity
            throw new JDOUserException(LOCALISER.msg("PM.ObjectId.KeyValueNotSupported", pcClass.getName(), key.getClass().getName()));
        }

        return id;
    }

    // ----------------------------- Object Retrieval by Id ----------------------------------

    /**
     * Accessor for the StateManager of an object given the object AID.
     * @param pcClass The class of the PC object
     * @param fv The field values to be loaded
     * @param ignoreCache true if it must ignore the cache
     * @param checkInheritance Whether look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     * @return Object
     */
    public synchronized Object getObjectByAID(Class pcClass, FieldValues fv, boolean ignoreCache, boolean checkInheritance)
    {
        assertIsOpen();

        StateManager sm = StateManagerFactory.newStateManager(this, pcClass, fv);
        if (!ignoreCache)
        {
            Object oid = sm.getInternalObjectId();
            PersistenceCapable pc = getObjectFromCache(oid);
            if (pc != null)
            {
                sm = findStateManager(pc);
                sm.loadFieldValues(fv); // Load the values retrieved by the query
                return pc;
            }
            if (checkInheritance)
            {
                if (oid instanceof OID || oid instanceof SingleFieldIdentity)
                {
                    // Check if this id for any known subclasses is in the cache to save searching
                    String[] subclasses = getMetaDataManager().getSubclassesForClass(pcClass.getName(), true);
                    if (subclasses != null)
                    {
                        for (int i=0;i<subclasses.length;i++)
                        {
                            if (oid instanceof OID)
                            {
                                oid = OIDFactory.getInstance(subclasses[i], ((OID)oid).keyValue());
                            }
                            else if (oid instanceof SingleFieldIdentity)
                            {
                                oid = AIDUtils.getNewSingleFieldIdentity(oid.getClass(), clr.classForName(subclasses[i]), 
                                    ((SingleFieldIdentity)oid).getKeyAsObject());
                            }
                            pc = getObjectFromCache(oid);
                            if (pc != null)
                            {
                                sm = findStateManager(pc);
                                sm.loadFieldValues(fv); // Load the values retrieved by the query
                                return pc;
                            }
                        }
                    }
                }
            }
        }
        if (checkInheritance)
        {
            sm.checkInheritance(fv);
            if (!ignoreCache)
            {
                // Check the cache in case this updated object id is present (since we should use that if available)
                Object oid = sm.getInternalObjectId();
                PersistenceCapable pc = getObjectFromCache(oid);
                if (pc != null)
                {
                    // We have an object with this new object id already so return it with the retrieved field values imposed
                    sm = findStateManager(pc);
                    sm.loadFieldValues(fv); // Load the values retrieved by the query
                    return pc;
                }
            }
            
        }
        putObjectIntoCache(sm, true, true);

        return sm.getObject();
    }

    /**
     * Accessor for an object given the object id. 
     * @param id Id of the object.
     * @param fv Field values for the object
     * @param cls the type which the object is. This type will be used to instanciat the object
     * @param ignoreCache true if it must ignore the cache
     * @return The Object
     */
    public synchronized Object getObjectById(Object id, FieldValues fv, Class cls, boolean ignoreCache)
    {
        assertIsOpen();

        PersistenceCapable pc = null;
        if (!ignoreCache)
        {
            pc = getObjectFromCache(id);
        }

        if (pc == null)
        {
            // Object not found so load the requested fields into a new object
            StateManager sm = StateManagerFactory.newStateManager(this, cls, id, fv);
            // TODO verify it

            pc = sm.getObject();

            // Save the object
            putObjectIntoCache(sm, true, true);
        }
        else
        {
            // TODO We should make sure that the fields specified in "fv" are loaded in the cached object
            // The commented out code causes issues with AttachDetachTest currently
            // Object found in the cache so load the requested fields
            StateManager sm = findStateManager(pc);
            
            if (sm != null)
            {
                // Load the requested fields
                fv.fetchNonLoadedFields(sm);
            }
        }

        return pc;
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @return The Object
     * @since 1.1
     */
    public synchronized Object getObjectById(Object id)
    {
        return getObjectById(id, true);
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param validate Whether to validate the object state
     * @return The Object
     */
    public synchronized Object getObjectById(Object id, boolean validate)
    {
        assertIsOpen();
        if (id == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.GetObjectByIdNullId"));
        }

        // Use inheritance checks whatever the identity type. Really we should be able to
        // assume that SingleFieldIdentity and datastore identity cases are of the class defined
        // by the identity itself since it contains the class name.
        return getObjectById(id, validate, true, null);
    }

    /**
     * Accessor for an object given the object id. If validate is false, we return the object
     * if found in the cache, or otherwise a Hollow object with that id. If validate is true
     * we check with the datastore and return an object with the FetchPlan fields loaded (unless
     * the object doesnt exist in the datastore, where we throw an ObjectNotFoundException).
     * @param id Id of the object.
     * @param validate Whether to validate the object state
     * @param checkInheritance Whether look to the database to determine which
     * class this object is.
     * @param objectClassName Class name for the object with this id (if known, optional)
     * @return The Object with this id
     */
    public synchronized Object getObjectById(Object id, boolean validate, boolean checkInheritance, String objectClassName)
    {
        assertIsOpen();
        if (id == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.GetObjectByIdNullId"));
        }

        // try to find object in cache(s)
        PersistenceCapable pc = getObjectFromCache(id);

        StateManager sm = null;
        if (pc == null)
        {
            // Object not found in cache(s) with this identity
            String className = null;
            String originalClassName = null;
            boolean checkedClassName = false;
            if (id instanceof OID)
            {
                // OID, so check that the implied class is managed
                originalClassName = srm.manageClassForIdentity((OID)id, getClassLoaderResolver());
            }
            else if (id instanceof SingleFieldIdentity)
            {
                // SingleFieldIdentity, so check that the implied class is managed
                originalClassName = srm.manageClassForIdentity((SingleFieldIdentity)id, getClassLoaderResolver());
            }
            else if (objectClassName != null)
            {
                // Object class name specified so use that directly
                originalClassName = objectClassName;
            }
            else
            {
                // We dont know the object class so try to deduce it from what is known by the StoreManager
                originalClassName = srm.getClassNameForObjectID(id, clr, this);
                checkedClassName = true;
            }

            if (checkInheritance)
            {
	            // Verify if correct class inheritance level is set
                if (!checkedClassName)
                {
                    className = srm.getClassNameForObjectID(id, clr, this);
                }
                else
                {
                    // We just checked the name of the class in the section above so just use that
                    className = originalClassName;
                }

	            if (className == null)
	            {
	                throw new JDOObjectNotFoundException(LOCALISER.msg("PM.ObjectDoesntExist"), id);
                }

                if (originalClassName != null && !originalClassName.equals(className))
                {
                    // Inheritance checking has found a different inherited
                    // object with this identity so create new id
                    if (id instanceof OID)
                    {
                        // Create new OID using correct target class
                        id = OIDFactory.getInstance(className, ((OID) id).keyValue());

                        // try again to read object from cache with this id
                        pc = getObjectFromCache(id);
                    }
                    else if (id instanceof SingleFieldIdentity)
                    {
                        // Create new SingleFieldIdentity using correct targetClass
                        id = AIDUtils.getNewSingleFieldIdentity(id.getClass(), getClassLoaderResolver().classForName(className), 
                                ((SingleFieldIdentity)id).getKeyAsObject());

                        // try again to read object from cache with this id
                        pc = getObjectFromCache(id);
                    }
                }
            }
            else
            {
                className = originalClassName;
            }

            if (pc == null)
            {
                // Object not found in Cache even after inheritance updates so create a Hollow instance
                try
                {
                    Class pcClass = clr.classForName(className, (id instanceof OID) ? null : id.getClass().getClassLoader());
                    sm = StateManagerFactory.newStateManager(this, pcClass, id);
                    pc = sm.getObject();
                }
                catch (ClassNotResolvedException e)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PM.GetObjectByIdClassNotFound", id));
                    throw new JDOUserException(LOCALISER.msg("PM.GetObjectByIdClassNotFound", id), e);
                }
            }
        }

        if (validate)
        {
            // User requests validation of the instance so go to the datastore to validate it
            // loading any fetchplan fields that are needed in the process.
            if (sm == null)
            {
                sm = findStateManager(pc);
            }
            sm.validate();
        }

        if (sm != null)
        {
            // Cache the object (update it if already present)
            putObjectIntoCache(sm, true, true);
        }

        return pc;
    }

    /**
     * Accessor for the objects given the object ids.
     * @param oids Ids of the objects.
     * @param validate Whether to validate the object state
     * @return The Objects with these ids (in the same order)
     * @since 1.1
     */
    public Collection getObjectsById(Collection oids, boolean validate)
    {
        assertIsOpen();
        if (oids == null || oids.size() == 0)
        {
            throw new JDOUserException(LOCALISER.msg("PM.GetObjectsByIdNullId"));
        }

        Collection objects = new ArrayList(oids.size());
        Iterator iter = oids.iterator();
        while (iter.hasNext())
        {
            Object oid = iter.next();
            // Just use getObjectById to get the object
            objects.add(getObjectById(oid, validate));
        }
        return objects;
    }

    /**
     * Accessor for the objects given the object ids.
     * @param oids Ids of the objects.
     * @param validate Whether to validate the object state
     * @return The Objects with these ids (in the same order)
     * @since 1.1
     */
    public Object[] getObjectsById(Object[] oids, boolean validate)
    {
        assertIsOpen();
        if (oids == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.GetObjectsByIdNullId"));
        }
        Object[] objects = new Object[oids.length];
        for (int i = 0; i < oids.length; i++)
        {
            // Just use getObjectById to get the object
            objects[i] = getObjectById(oids[i], validate);
        }
        return objects;
    }

    /**
     * Accessor for the objects given the object ids, validating the objects.
     * @param oids Ids of the objects.
     * @return The Objects with these ids (in the same order)
     * @since 1.1
     */
    public Collection getObjectsById(Collection oids)
    {
        return getObjectsById(oids, true);
    }

    /**
     * Accessor for the objects given the object ids, validating the objects.
     * @param oids Ids of the objects.
     * @return The Objects with these ids (in the same order)
     * @since 1.1
     */
    public Object[] getObjectsById(Object[] oids)
    {
        return getObjectsById(oids, true);
    }

    /**
     * Convenience method that exactly matches the behavior of calling 
     * pm.getObjectById (pm.newObjectIdInstance (cls, key), true).
     * @param cls Class of the PersistenceCapable
     * @param key Value of the key field for SingleFieldIdentity, or the string value of the key otherwise
     * @return The object for this id.
     * @since 1.1
     */
    public Object getObjectById(Class cls, Object key)
    {
        return getObjectById(newObjectIdInstance (cls, key), true);
    }

    /**
     * Accessor for an object given the object id. Checks the inheritance level of the object
     * @param id Id of the object.
     * @param fv Field values for the object
     * @return The Object
     */
    public synchronized Object getObjectById(Object id, FieldValues fv)
    {
        assertIsOpen();

        PersistenceCapable pc = getObjectFromCache(id);
        if (pc == null)
        {
            String className = srm.getClassNameForObjectID(id, clr, this);
            // className is null when id class exists, and object has been
            // validated and doesn't exist.
            if (className == null)
            {
                throw new JDOObjectNotFoundException(LOCALISER.msg("PM.ObjectDoesntExist"), id);
            }

            if (id instanceof OID)
            {
                id = OIDFactory.getInstance(className,((OID) id).keyValue());

                // try again to read object from cache
                pc = getObjectFromCache(id);
            }
            // not found in cache, will get a Hollow instance
            if (pc == null)
            {
                try
                {
                    Class pcClass = clr.classForName(className, id.getClass().getClassLoader());
                    StateManager sm = StateManagerFactory.newStateManager(this, pcClass, id, fv);
                    pc = sm.getObject();
                    putObjectIntoCache(sm, true, true);
                }
                catch (ClassNotResolvedException e)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PM.GetObjectByIdClassNotFound", id));
                    throw new JDOUserException(LOCALISER.msg("PM.GetObjectByIdClassNotFound", id), e);
                }
            }
        }
        return pc;
    }

    /**
     * Accessor for an object id given the object.
     * @param pc The object
     * @return The Object id
     */
    public Object getObjectId(Object pc)
    {
        assertIsOpen();
        if (pc != null && pc instanceof PersistenceCapable)
        {
            PersistenceCapable p = (PersistenceCapable) pc;
            if (p.jdoIsPersistent() || p.jdoIsDetached())
            {
                return p.jdoGetObjectId();
            }
        }
        return null;
    }

    /**
     * Accessor for the object id of a transactional object given the object.
     * @param pc The object
     * @return The Object id
     */
    public Object getTransactionalObjectId(Object pc)
    {
        assertIsOpen();
        return ((PersistenceCapable) pc).jdoGetTransactionalObjectId();
    }

    /**
     * Method to put a user object into the PersistenceManager. This is so that
     * multiple users can each have a user object for example. <I>The parameter
     * is not inspected or used in any way by the JDO implementation. </I>
     * @param key The key to store the user object under
     * @param value The object to store
     * @return The previous value for this key
     * @since 1.1
     */
    public synchronized Object putUserObject(Object key, Object value)
    {
        assertIsOpen();
        if (key == null)
        {
            return null;
        }
        if ( userObjectMap == null)
        {
            userObjectMap = new HashMap();
        }
        if (value == null)
        {
            // Remove the object
            return userObjectMap.remove(key);
        }
        else
        {
            // Put the object
            return userObjectMap.put(key, value);
        }
    }

    /**
     * Method to get a user object from the PersistenceManager. This is for user
     * objects which are stored under a key. <I>The parameter is not inspected
     * or used in any way by the JDO implementation. </I>
     * @param key The key to store the user object under
     * @return The user object for that key
     * @since 1.1
     */
    public synchronized Object getUserObject(Object key)
    {
        assertIsOpen();
        if (key == null)
        {
            return null;
        }
        if (userObjectMap==null)
        {
            return null;
        }
        return userObjectMap.get(key);
    }

    /**
     * Method to remove a user object from the PersistenceManager. This is for
     * user objects which are stored under a key. <I>The parameter is not
     * inspected or used in any way by the JDO implementation. </I>
     * @param key The key whose uder object is to be removed.
     * @return The user object that was removed
     * @since 1.1
     */
    public synchronized Object removeUserObject(Object key)
    {
        assertIsOpen();
        if (key == null)
        {
            return null;
        }
        if (userObjectMap==null)
        {
            return null;
        }
        return userObjectMap.remove(key);
    }

    /**
     * The application might manage PersistenceManager instances by using an
     * associated object for bookkeeping purposes. These methods allow the user
     * to manage the associated object. <I>The parameter is not inspected or
     * used in any way by the JDO implementation. </I>
     * @param userObject The object
     */
    public synchronized void setUserObject(Object userObject)
    {
        assertIsOpen();
        this.userObject = userObject;
    }

    /**
     * The application might manage PersistenceManager instances by using an
     * associated object for bookkeeping purposes. These methods allow the user
     * to manage the associated object. <I>The parameter is not inspected or
     * used in any way by the JDO implementation. </I>
     * @return The user object
     */
    public synchronized Object getUserObject()
    {
        assertIsOpen();
        return userObject;
    }

    /**
     * Accessor for the class of the object id given the class of object.
     * @param cls The class name of the object
     * @return The class name of the object id
     */
    public Class getObjectIdClass(Class cls)
    {
        assertIsOpen();
        if (!ClassUtils.isPersistenceCapableClass(cls) || !hasMetaDataForPersistenceCapableClass(cls))
        {
            return null;
        }
        AbstractClassMetaData cmd = getMetaDataManager().getMetaDataForClass(cls, clr);
        if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            return OID.class;
        }
        else if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            try
            {
                return this.getClassLoaderResolver().classForName(getMetaDataManager().getMetaDataForClass(cls, clr).getObjectidClass(), null);
            }
            catch (ClassNotResolvedException e)
            {
                String msg = LOCALISER.msg("PM.ObjectIdClassNotFound", cls.getName());
                JPOXLogger.GENERAL.error(msg);
                throw new JDOException(msg);
            }
        }
        else
        {
            if (cmd.isRequiresExtent())
            {
                return OID.class;
            }
            else
            {
                return SCOID.class;
            }
        }
    }

    /**
     * Mutator for whether the Persistence Manager is multithreaded.
     * @param flag Whether to run multithreaded.
     */
    public void setMultithreaded(boolean flag)
    {
        assertIsOpen();
        this.multithreaded = flag;
    }

    /**
     * Accessor for whether the Persistence Manager is multithreaded.
     * @return Whether to run multithreaded.
     */
    public boolean getMultithreaded()
    {
        assertIsOpen();
        return this.multithreaded;
    }

    /**
     * Mutator for whether to ignore the cache.
     * @param flag Whether to ignore the cache.
     */
    public void setIgnoreCache(boolean flag)
    {
        assertIsOpen();
        this.ignoreCache = flag;
    }

    /**
     * Accessor for whether to ignore the cache.
     * @return Whether to ignore the cache.
     */
    public boolean getIgnoreCache()
    {
        assertIsOpen();
        return ignoreCache;
    }
    
    /**
     * Mutator for whether to detach objects on close of the PM.
     * <B>This is a JPOX extension and will not work when in JCA mode.</B>
     * @param flag Whether to detach on close.
     */
    public void setDetachOnClose(boolean flag)
    {
        assertIsOpen();
        this.detachOnClose = flag;
    }

    /**
     * Accessor for whether to detach objects on close of the PM.
     * <B>This is a JPOX extension and will not work when in JCA mode.</B>
     * @return Whether to detach on close.
     */
    public boolean getDetachOnClose()
    {
        assertIsOpen();
        return detachOnClose;
    }

    /**
     * Mutator for whether to detach all objects on commit of the transaction.
     * @param flag Whether to detach all on commit.
     */
    public void setDetachAllOnCommit(boolean flag)
    {
        assertIsOpen();
        this.detachAllOnCommit = flag;
    }

    /**
     * Accessor for whether to detach all objects on commit of the transaction.
     * @return Whether to detach all on commit.
     */
    public boolean getDetachAllOnCommit()
    {
        assertIsOpen();
        return detachAllOnCommit;
    }

    /**
     * Method to find the StateManager for an object.
     * @param pc The object
     * @return The StateManager, null if StateManager not found. see JDO spec
     * for the calling behavior when null is returned
     */
    public synchronized StateManager findStateManager(PersistenceCapable pc)
    {
        StateManager sm = null;
        PersistenceCapable previousLookingFor = lookingForStateManagerFor;
        StateManager previousFound = foundStateManager;
        try
        {
            lookingForStateManagerFor = pc;
            foundStateManager = null;
            // this method will call the
            // StateManager.getPersistenceManager(PersistenceCapable pc)
            // StateManager.getPersistenceManager will call hereIsStateManager
            // that will set the foundStateManager
            PersistenceManager pm = (PersistenceManager) pc.jdoGetPersistenceManager();
            if (pm != null && this.getPMHandle() != pm.getPMHandle())
            {
                throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
            }
            sm = foundStateManager;
        }
        finally
        {
            lookingForStateManagerFor = previousLookingFor;
            foundStateManager = previousFound;
        }
        return sm;
    }

    /**
     * Method to add the StateManager for an object to this PersistenceManager's
     * list.
     * @param sm The StateManager
     * @param pc The object managed by the StateManager
     */
    public synchronized void hereIsStateManager(StateManager sm, PersistenceCapable pc)
    {
        if (lookingForStateManagerFor == pc)
        {
            foundStateManager = sm;
        }
    }

    /**
     * Method to mark an object (StateManager) as dirty.
     * @param sm The StateManager
     */
    public synchronized void clearDirty(StateManager sm)
    {
        dirtySMs.remove(sm);
    }

    /**
     * Method to mark an object (StateManager) as dirty.
     * @param sm The StateManager
     */
    public synchronized void markDirty(StateManager sm)
    {
        if (!this.isDelayDatastoreOperationsEnabled() && dirtySMs.size() > 0 && !dirtySMs.contains(sm))
        {
            flush();
        }
        if (!dirtySMs.contains(sm))
        {
            dirtySMs.add(sm);
        }
    }

    /**
     * Method to mark as clean the StateManager.
     */
    public synchronized void clearDirty()
    {
        dirtySMs.clear();
    }

    /**
     * This method flushes all dirty, new, and deleted instances to the
     * datastore. It has no effect if a transaction is not active. If a
     * datastore transaction is active, this method synchronizes the cache with
     * the datastore and reports any exceptions. If an optimistic transaction is
     * active, this method obtains a datastore connection and synchronizes the
     * cache with the datastore using this connection. The connection obtained
     * by this method is held until the end of the transaction.
     */
    public synchronized void flush()
    {
        assertIsOpen();

        if (tx.isActive())
        {
            List optimisticFailures = new ArrayList();

            // Flush all dirty, new, deleted instances to the datastore when transaction is active
            List toFlush = new ArrayList(dirtySMs);
            clearDirty();
            for (int i = 0; i < toFlush.size(); i++)
            {
                StateManager sm = (StateManager) toFlush.get(i);
                try
                {
                    sm.flush();
                }
                catch (JDOOptimisticVerificationException ove)
                {
                    optimisticFailures.add(ove);
                }
            }
            toFlush.clear();

            if (optimisticFailures.size() > 0)
            {
                // Throw a single OptimisticVerificationException containing all failures
                throw new JDOOptimisticVerificationException(LOCALISER.msg("PM.OptimisticVerificationError"), 
                        (Throwable[])optimisticFailures.toArray(new Throwable[optimisticFailures.size()]));
            }
        }
    }

    /**
     * This method validates the cache with the datastore. It has no effect if a transaction 
     * is not active. If a datastore transaction is active, this method verifies the consistency 
     * of instances in the cache against the datastore. An implementation might flush instances 
     * as if flush() were called, but it is not required to do so. 
     * If an optimistic transaction is active, this method obtains a datastore connection 
     * and verifies the consistency of the instances in the cache against the datastore. If any 
     * inconsistencies are detected, a JDOOptimisticVerificationException is thrown. This 
     * exception contains a nested JDOOptimisticVerificationException for each object 
     * that failed the consistency check. No datastore resources acquired during the execution 
     * of this method are held beyond the scope of this method.
     */
    public void checkConsistency()
    {
        assertIsOpen();

        // If transaction is not active do nothing
        if (!tx.isActive())
        {
            return;
        }

        if (tx.getOptimistic())
        {
            // TODO Implement checkConsistency() for optimistic transactions
            throw new JDOUserException("checkConsistency() not yet implemented for optimistic transactions");
        }
        else
        {
            flush();
        }
    }

    /**
     * Method to detach all objects in the PM.
     * <B>For internal use only</B>
     */
    public void detachAll()
    {
        Object[] sms = this.enlistedSMCache.values().toArray();
        FetchPlanState fps = new FetchPlanState();
        for( int i=0; i<sms.length; i++)
        {
            ((StateManager)sms[i]).detach(fps);
        }
    }

    /** Flag whether we are currently runnign persistence-by-reachability-at-commit */
    private boolean runningPBRAtCommit = false;

    /**
     * Method to perform any pre-commit checks.
     */
    public synchronized void preCommit()
    {
        // Inform all query results that we are closing the connection 
        // (so they have the opportunity to read all remaining results now)
        notifyQueriesOfConnectionClose();

        // Make sure all is flushed before reachability
        flush();

        try
        {
            if (apmf.getPersistenceByReachabilityAtCommit())
            {
            	runningPBRAtCommit = true;
                if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
                {
                    JPOXLogger.JDO_REACHABILITY.debug(LOCALISER.msg("PM.PersistenceByReachability.RunningCommitCheck"));
                }

                // If we have some new objects in this transaction, and we have some known persisted objects (either
                // from makePersistent in this txn, or enlisted existing objects) then run reachability checks
                if (txKnownPersists.size() > 0 && txFlushedNew.size() > 0)
                {
                    Set currentReachables = new HashSet();

                    // Run "reachability" on all known persistent objects for this txn
                    Object ids[] = txKnownPersists.toArray();
                    Set objectNotFound = new HashSet();
                    for (int i=0; i<ids.length; i++)
                    {
                        if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
                        {
                            JPOXLogger.JDO_REACHABILITY.debug("Performing reachability algorithm on object with id \""+ids[i]+"\"");
                        }
                        try
                        {
                            StateManager sm = findStateManager((PersistenceCapable)getObjectById(ids[i]));
                            sm.runReachability(currentReachables);
                            if (i % 10000 == 0 || i == ids.length-1)
                            {
                                // Flush every 10000 or on last one to make sure tx cache is always empty
                                flush();
                            }
                        }
                        catch (JDOObjectNotFoundException ex)
                        {
                            objectNotFound.add(ids[i]);
                        }
                    }

                    // Remove any of the "reachable" instances that are no longer "reachable"
                    txFlushedNew.removeAll(currentReachables);

                    // For all remaining "new" instances, nullify their fields and delete them since no longer reachable
                    Object nonReachableIds[] = txFlushedNew.toArray();
                    for (int i=0; i<nonReachableIds.length; i++)
                    {
                        // Remove any of the instances made persistent by reachability that now aren't
                        if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
                        {
                            JPOXLogger.JDO_REACHABILITY.debug(LOCALISER.msg("PM.PersistenceByReachability.ObjectNoLongerReachable", nonReachableIds[i]));
                        }
                        try
                        {
                            if (!objectNotFound.contains(nonReachableIds[i]))
                            {
                                StateManager sm = findStateManager((PersistenceCapable)getObjectById(nonReachableIds[i]));
                                sm.nullifyFields();
                                if (i % 10000 == 0 || i == nonReachableIds.length-1)
                                {
                                    // Flush every 10000 or on last one to clear out dirties
                                    flush();
                                }
                            }
                        }
                        catch (JDOObjectNotFoundException ex)
                        {
                            //just ignore if the file does not exist anymore  
                        }
                    }

                    for (int i=0; i<nonReachableIds.length; i++)
                    {
                        // Remove any of the instances made persistent by reachability that now aren't
                        try
                        {
                            if (!objectNotFound.contains(nonReachableIds[i]))
                            {
                                StateManager sm = findStateManager((PersistenceCapable)getObjectById(nonReachableIds[i]));
                                sm.deletePersistent();
                                if (i % 10000 == 0 || i == nonReachableIds.length-1)
                                {
                                    // Flush every 10000 or on last one to clear out dirties
                                    flush();
                                }
                            }
                        }
                        catch (JDOObjectNotFoundException ex)
                        {
                            //just ignore if the file does not exist anymore  
                        }
                    }
                }
                if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
                {
                    JPOXLogger.JDO_REACHABILITY.debug(LOCALISER.msg("PM.PersistenceByReachability.CompletedCommitCheck"));
                }
            }
        }
        catch (Throwable t)
        {
            JPOXLogger.JDO.error(t);
            if (t instanceof JDOException)
            {
                throw (JDOException) t;
            }
            else
            {
                throw new JDOException("Unexpected error during precommit",t);
            }
        }
        finally
        {
        	runningPBRAtCommit = false;
        }

        // Flush in case the PBR at commit wasnt run
        flush();

        if (detachAllOnCommit)
        {
            // "detach-on-commit"
            // Load all unloaded fetch-plan fields of the objects that will be needed at commit time so we 
            // can safely detach it when we commit (the connection is still open here, whereas it is closed
            // in the postCommit when doing the detachment).
            smsToDetachAtCommit = getRootStateManagersForDetachAllOnCommit();
            for (int i = 0; i < smsToDetachAtCommit.length; ++i)
            {
                PersistenceCapable pc = smsToDetachAtCommit[i].getObject();
                if (pc != null && !pc.jdoIsDetached() && !pc.jdoIsDeleted())
                {
                    // Load all fields (and sub-objects) in the FetchPlan
                    FetchPlanState state = new FetchPlanState();
                    smsToDetachAtCommit[i].loadFieldsInFetchPlan(state);
                }
            }
        }
    }

    /**
     * Temporary array of StateManagers to detach at commit (to prevent garbage collection). 
     * Set up in preCommit() and used in postCommit().
     */
    private StateManager[] smsToDetachAtCommit = null;

    /**
     * Method to return the StateManagers for all root objects detached with DetachAllOnCommit.
     * @return SMs for root objects for DetachAllOnCommit
     */
    private StateManager[] getRootStateManagersForDetachAllOnCommit()
    {
        // JDO2 spec 12.7.3 "Root instances"
        // "Root instances are parameter instances for retrieve, detachCopy, and refresh; result
        // instances for queries. Root instances for DetachAllOnCommit are defined explicitly by
        // the user via the FetchPlan property DetachmentRoots or DetachmentRootClasses. 
        // If not set explicitly, the detachment roots consist of the union of all root instances of
        // methods executed since the last commit or rollback."
        StateManager[] sms = null;
        Collection roots = fetchPlan.getDetachmentRoots();
        Class[] rootClasses = fetchPlan.getDetachmentRootClasses();
        if (roots != null && roots.size() > 0)
        {
            // Detachment roots specified
            sms = new StateManager[roots.size()];
            Iterator rootsIter = roots.iterator();
            int i = 0;
            while (rootsIter.hasNext())
            {
                Object obj = rootsIter.next();
                sms[i++] = findStateManager((PersistenceCapable)obj);
            }
        }
        else if (rootClasses != null && rootClasses.length > 0)
        {
            // Detachment root classes specified
            ArrayList smList = new ArrayList();
            StateManager[] txSMs = (StateManager[])enlistedSMCache.values().toArray(new StateManager[enlistedSMCache.size()]);
            for (int i=0;i<txSMs.length;i++)
            {
                for (int j=0;j<rootClasses.length;j++)
                {
                    // Check if object is of this root type
                    if (txSMs[i].getObject().getClass() == rootClasses[j])
                    {
                        // This SM is for a valid root object
                        smList.add(txSMs[i]);
                        break;
                    }
                }
            }
            sms = (StateManager[]) smList.toArray(new StateManager[smList.size()]);
        }
        else
        {
            // Detach all objects in the L1 cache
            sms = (StateManager[])cache.values().toArray(new StateManager[cache.values().size()]);
        }

        return sms;
    }

    /**
     * Commit any changes made to objects managed by the persistence manager to
     * the database.
     */
    public synchronized void postCommit()
    {
        Collection smsToMakeTransient = null;
        if (detachAllOnCommit)
        {
            // Detach all detachment root objects (causes recursion through the fetch plan)
            StateManager[] smsToDetach = smsToDetachAtCommit;
            DetachState state = new DetachState();
            for (int i=0;i<smsToDetach.length;i++)
            {
                PersistenceCapable pc = smsToDetach[i].getObject();
                if (pc != null && !pc.jdoIsDetached() && !pc.jdoIsDeleted())
                {
                    if (pc instanceof Detachable)
                    {
                        // JDO2 spec 13.4.2, Detachable instances transition to DETACHED state
                        smsToDetach[i].detach(state);
                    }
                    else
                    {
                        // JDO2 spec 13.4.2, non-Detachable instances transition to TRANSIENT state
                        // Just collect the list of StateManagers for processing after postCommit is called on the SM
                        if (smsToMakeTransient == null)
                        {
                            smsToMakeTransient = new ArrayList();
                        }
                        smsToMakeTransient.add(smsToDetach[i]);
                    }
                }
            }
            smsToDetachAtCommit = null; // No longer need these references
        }

        List failures = new ArrayList();
        try
        {
            // Commit all enlisted StateManagers
            StateManager[] sms = (StateManager[]) enlistedSMCache.values().toArray(new StateManager[enlistedSMCache.size()]);
            for (int i = 0; i < sms.length; ++i)
            {
                try
                {
                    // Perform any operations required after committing
                    //TODO this if is due to sms that can have lc == null, why?, should not be here then
                    if( sms[i] != null && sms[i].getObject() != null && (sms[i].getObject().jdoIsPersistent() || sms[i].getObject().jdoIsTransactional()))
                    {
                        sms[i].postCommit(currentTransaction());
    
                        // TODO Change this check so that we remove all objects that are no longer suitable for caching
                        if (detachAllOnCommit && sms[i].getObject() instanceof Detachable)
                        {
                            // "DetachAllOnCommit" - Remove the object from the L1 cache since it is now detached
                            removeStateManager(sms[i]);
                        }
                    }
                }
                catch (RuntimeException e)
                {
                    failures.add(e);
                }
            }

            if (detachAllOnCommit && smsToMakeTransient != null)
            {
                // "DetachAllOnCommit" - JDO2 spec 13.4.2, non-Detachable instances transition to TRANSIENT state
                // This is done *after* SM.postCommit() so that these objects are now in HOLLOW state and we can move to transient safely
                Iterator makeTransientIter = smsToMakeTransient.iterator();
                while (makeTransientIter.hasNext())
                {
                    ((StateManager)makeTransientIter.next()).makeTransient(null);
                }
            }
        }
        finally
        {
            enlistedSMCache.clear();
            enlistedIds.clear();
            txKnownPersists.clear();
            txFlushedNew.clear();
            ((FetchPlanImpl)fetchPlan).resetDetachmentRoots();
        }
        if (!failures.isEmpty())
        {
            throw new CommitStateTransitionException((Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Convenience method to notify all queries for this PersistenceManager that
     * the current transaction is closing. Provides the queries with a way of
     * loading all remaining instances not yet accessed so that the query is
     * still accessible.
     */
    private void notifyQueriesOfConnectionClose()
    {
        Iterator queryResultIter = queryResults.values().iterator();
        while (queryResultIter.hasNext())
        {
            // This has the disadvantage that it will read in all result sets that are still open
            // and so could bring in many objects.
            // TODO Think of a better strategy, but bear-in-mind that we must allow the QueryResult
            // to be available after the commit of the transaction
            QueryResult queryResult = (QueryResult)queryResultIter.next();
            queryResult.closingConnection();
        }
    }

    /**
     * Rollback any changes made to objects managed by the persistence manager
     * to the database.
     */
    public synchronized void preRollback()
    {
        // Inform all query results that we are closing the connection 
        // (so they have the opportunity to read all remaining results now)
        notifyQueriesOfConnectionClose();

        ArrayList failures = new ArrayList();
        try
        {
            Collection sms = enlistedSMCache.values();
            Iterator smsIter = sms.iterator();
            while (smsIter.hasNext())
            {
                StateManager sm = (StateManager)smsIter.next();
                try
                {
                    sm.preRollback(currentTransaction());
                }
                catch (RuntimeException e)
                {
                    failures.add(e);
                }
            }
            clearDirty();
        }
        finally
        {
            enlistedSMCache.clear();
            enlistedIds.clear();
            txKnownPersists.clear();
            txFlushedNew.clear();
        }
        if (!failures.isEmpty())
        {
            throw new RollbackStateTransitionException((Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Diagnostic method to dump out the state of the passed object to the passed PrintWriter.
     * @param obj The PersistenceCapable object
     * @param out The PrintWriter
     */
    public void dump(Object obj, PrintWriter out)
    {
        findStateManager((PersistenceCapable) obj).dump(out);
    }

    // ------------------------------------- Sequence Management --------------------------------------

    /**
     * Method to retrieve a sequence by name. As per JDO2 spec section 12.14.
     * If the named sequence is not known, throws a JDOUserException.
     * @param sequenceName Fully qualified name of the sequence
     * @return The sequence
     */
    public Sequence getSequence(String sequenceName)
    {
        assertIsOpen();

        SequenceMetaData seqmd = getMetaDataManager().getMetaDataForSequence(getClassLoaderResolver(),sequenceName);
        if (seqmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("Sequence.SequenceNotFound", sequenceName));
        }

        Sequence seq = null;
        if (seqmd.getFactoryClass() != null)
        {
            // User has specified a factory class
            seq = apmf.getSequenceForFactoryClass(seqmd.getFactoryClass());
            if (seq == null)
            {
                // Create a new instance of the factory class and obtain the Sequence
                Class factory = clr.classForName(seqmd.getFactoryClass());
                if (factory == null)
                {
                    throw new JDOUserException(LOCALISER.msg("Sequence.FactoryClassNotFound", sequenceName, seqmd.getFactoryClass()));
                }

                Class[] argTypes = null;
                Object[] arguments = null;
                if (seqmd.getStrategy() != null)
                {
                    argTypes = new Class[2];
                    argTypes[0] = String.class;
                    argTypes[1] = String.class;
                    arguments = new Object[2];
                    arguments[0] = seqmd.getName();
                    arguments[1] = seqmd.getStrategy().toString();
                }
                else
                {
                    argTypes = new Class[1];
                    argTypes[0] = String.class;
                    arguments = new Object[1];
                    arguments[0] = seqmd.getName();
                }

                Method newInstanceMethod;
                try 
                {
                    // Obtain the sequence from the static "newInstance(...)" method
                    newInstanceMethod = factory.getMethod("newInstance", argTypes);
                    seq = (Sequence) newInstanceMethod.invoke(null, arguments);
                }
                catch (Exception e)
                {
                    throw new JDOUserException(LOCALISER.msg("Sequence.FactoryClassInvocationError", seqmd.getFactoryClass(), e.getMessage()));
                }

                // Register the sequence with the PMF
                apmf.addSequenceForFactoryClass(seqmd.getFactoryClass(), seq);
            }
        }
        else
        {
            seq = srm.getSequence(this, seqmd);
        }
        return seq;
    }

    // -------------------------------------- Cache Management ---------------------------------------

    /**
     * Replace the previous object id for a PC object to a new
     * @param pc The Persistence Capable object
     * @param oldID the old id
     * @param newID the new id
     */
    public synchronized void replaceObjectId(PersistenceCapable pc, Object oldID, Object newID)
    {
        if (pc == null || pc.jdoGetObjectId() == null)
        {
            JPOXLogger.CACHE.warn(LOCALISER.msg("Cache.Level1.AddingNullObjectError"));
            return;
        }

        if (JPOXLogger.CACHE.isDebugEnabled())
        {
            JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.AddingObject", newID));
        }
        Object o = cache.get(oldID); //use get here, because a cache.remove operation returns a weakReference instance
        if (o != null)
        {
            cache.remove(oldID); //this returns a weak reference
            cache.put(newID, o);
        }

        // Put into Level 2 Cache
        if (this.apmf.getJdoCacheLevel2())
        {
            StateManager sm = findStateManager(pc);

            putObjectIntoCache(sm, false, true);
        }

        if (apmf.getPersistenceByReachabilityAtCommit())
        {
            if (txFlushedNew.remove(oldID))
            {
                txFlushedNew.add(newID);
            }
            if (txKnownPersists.remove(oldID))
            {
                txKnownPersists.add(newID);
            }
        }
    }

    /**
     * Convenience method to add an object to the cache(s).
     * @param sm The State Manager
     * @param level1 Whether to put it in the L1 cache
     * @param level2 Whether to put it in the L2 cache
     */
    public synchronized void putObjectIntoCache(StateManager sm, boolean level1, boolean level2)
    {
        if( sm.getObject() == null )
        {
            JPOXLogger.CACHE.warn(LOCALISER.msg("Cache.Level1.AddingNullObjectError"));
            return;
        }
        Object id = sm.getObject().jdoGetObjectId();
        if (id == null)
        {
            JPOXLogger.CACHE.warn(LOCALISER.msg("Cache.Level1.AddingNullObjectError"));
            return;
        }

        // Put into Level 1 Cache
        if (level1)
        {
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.AddingObject", sm.getInternalObjectId()));
            }
            cache.put(sm.getInternalObjectId(), sm);
        }

        // Put into Level 2 Cache
        if (level2 && this.apmf.getJdoCacheLevel2())
        {
            boolean storeInL2Cache = true;
            AbstractClassMetaData acmd = getMetaDataManager().getMetaDataForClass(sm.getObject().getClass(), clr);
            if (acmd != null && acmd.getIdentityType() == IdentityType.APPLICATION)
            {
                // If using compound identity dont put it in the L2 Cache (the id uses a PC which we can't link to)
                int[] pkFieldNumbers = acmd.getPrimaryKeyFieldNumbers();
                for (int i=0;i<pkFieldNumbers.length;i++)
                {
                    AbstractPropertyMetaData fmd = acmd.getManagedFieldAbsolute(pkFieldNumbers[i]);
                    if (PersistenceCapable.class.isAssignableFrom(fmd.getType()))
                    {
                        storeInL2Cache = false;
                    }
                }
            }

            if (storeInL2Cache)
            {
                if (JPOXLogger.CACHE.isDebugEnabled())
                {
                    JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level2.AddingObject",id));
                }
                // Store an L2 cacheable form of this object
                CachedPC pcCopy = sm.getL2CacheableObject();
                ((Level2Cache)this.apmf.getDataStoreCache()).put(id, pcCopy);
            }
        }
    }

    /**
     * Convenience method to evict an object from the cache(s).
     * @param pc The PersistenceCapable object
     * @param id The PersistenceCapable object id
     * @param level1 Whether to evict from the L1 cache
     * @param level2 Whether to evict from the L2 cache
     */
    public synchronized void removeObjectFromCache(PersistenceCapable pc, Object id, boolean level1, boolean level2)
    {
        // Evict from the L1 cache
        if (level1 && id != null)
        {
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.RemovingObject", id, String.valueOf(cache.size())));
            }
            Object pcRemoved = cache.remove(id);
            if (pcRemoved == null && JPOXLogger.CACHE.isDebugEnabled())
            {
                // For some reason the object isn't in the L1 cache - garbage collected maybe ?
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.RemovalOfObjectFailed", id));
            }
        }

        // Evict from the L2 cache
        if (level2 && apmf.getJdoCacheLevel2() && pc.jdoGetObjectId() != null )
        {
            Level2Cache l2Cache = ((Level2Cache)apmf.getDataStoreCache());
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level2.RemovingObject", pc.jdoGetObjectId(), String.valueOf(l2Cache.getSize())));
            }
            l2Cache.evict(pc.jdoGetObjectId());
        }
    }

    /**
     * Convenience method to access an object in the cache.
     * Firstly looks in the L1 cache for this PM, and if not found looks in the L2 cache.
     * @param id Id of the object
     * @return Persistence Capable object (with connected StateManager).
     */
    public synchronized PersistenceCapable getObjectFromCache(Object id)
    {
        PersistenceCapable pc = null;

        // Try Level 1 first
        StateManager sm = (StateManager)cache.get(id);
        if (sm != null)
        {
            pc = sm.getObject();
            
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.RetrievalOfObject", id, "" + cache.size()));
            }
            // Wipe the detach state that may have been added if the object has been serialised in the meantime
            sm.resetDetachState();

            return pc;
        }
        else
        {
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.RetrievalFailed", id, "" + cache.size()));
            }
        }

        // Try Level 2 since not in Level 1
        if (apmf.getJdoCacheLevel2())
        {
            Level2Cache l2Cache = ((Level2Cache)apmf.getDataStoreCache());
            CachedPC cachedPC = l2Cache.get(id);

            // Copy the cached object and connect to a StateManager, with the same object id
            if (cachedPC != null)
            {
                sm = StateManagerFactory.newStateManager(this, id, cachedPC.getPersistenceCapable(), cachedPC.getLoadedFields());
                pc = sm.getObject();
                if (JPOXLogger.CACHE.isDebugEnabled())
                {
                    JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level2.RetrievalOfObject", id, "" + l2Cache.getSize()));
                }
                return pc;
            }
            else
            {
                if (JPOXLogger.CACHE.isDebugEnabled())
                {
                    JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level2.RetrievalFailed", id, "" + l2Cache.getSize()));
                }
            }
        }

        return null;
    }

    // ------------------------------------- Lifecycle Listeners --------------------------------------

    /**
     * This method removes all previously registered lifecycle listeners.
     * It is necessary to make sure, that a cached PM (in j2ee environment)
     * will have no listener before the listeners are copied from the PMF.
     * Otherwise they might be registered multiple times.
     */
    public void removeAllInstanceLifecycleListeners()
    {
        if (callbacks != null)
        {
            callbacks.close();
        }
    }

    /**
     * Method to register a lifecycle listener as per JDO 2.0 spec 12.15.
     * @param listener The instance lifecycle listener to sends events to
     * @param classes The classes that it is interested in
     * @since 1.1
     */
    public void addInstanceLifecycleListener(InstanceLifecycleListener listener, Class[] classes)
    {
        assertIsOpen();
        if (listener == null)
        {
            return;
        }
        getCallbackHandler().addListener(listener, classes);
    }
    
    /**
     * Retrieve the callback handler for this PM
     * @return the callback handler
     */
    public CallbackHandler getCallbackHandler()
    {
        if (callbacks != null)
        {
            return callbacks;
        }

        String callbackHandlerClassName = getPMFContext().getPluginManager().getAttributeValueForExtension("org.jpox.callbackhandler",
            "name", getPMFContext().getApi(), "class-name");
        if (callbackHandlerClassName != null)
        {
            try
            {
                callbacks = (CallbackHandler) clr.classForName(callbackHandlerClassName).newInstance();
                return callbacks;
            }
            catch (InstantiationException e)
            {
                JPOXLogger.RDBMS.error(LOCALISER.msg("CallbackHandler.InstantiationError", callbackHandlerClassName, e));
            }
            catch (IllegalAccessException e)
            {
                JPOXLogger.RDBMS.error(LOCALISER.msg("CallbackHandler.InstantiationError", callbackHandlerClassName, e));
            }
        }

        return null;
    }

    /**
     * Method to remove a currently registered lifecycle listener, as per 
     * JDO 2.0 spec 12.15.
     * @param listener The instance lifecycle listener to remove.
     * @since 1.1
     */
    public void removeInstanceLifecycleListener(InstanceLifecycleListener listener)
    {
        assertIsOpen();
        if (listener != null)
        {
            getCallbackHandler().removeListener(listener);
        }
    }

    /** Number of the current query result. Used as key in the queryResults cache (temporary) */
    private int queryResultNumber= 0;

    /**
     * Method to register a queryResult with the PM.
     * @param queryResult The results
     * @return The results number in this PM
     */
    public int addQueryResult(QueryResult queryResult)
    {
        int resultsNumber = queryResultNumber;
        queryResults.put(new Integer(resultsNumber), queryResult);
        queryResultNumber++;
        return resultsNumber;
    }

    /**
     * Method to deregister a queryResult from the PM.
     * @param resultsNumber Number of the query results
     */
    public void removeQueryResult(int resultsNumber)
    {
        queryResults.remove(new Integer(resultsNumber));
    }

    /* (non-Javadoc)
     * @see org.jpox.PersistenceManager#addQueryRun(java.lang.String)
     */
    public void addQueryRun(String singleStringQuery)
    {
        // TODO What do we save the SingleString query for? Why put them in a SoftValueMap?
        queriesRun.put(singleStringQuery, singleStringQuery);
    }

    /* (non-Javadoc)
     * @see org.jpox.PersistenceManager#hasQueryRun(java.lang.String)
     */
    public boolean hasQueryRun(String singleStringQuery)
    {
        // TODO What do we save the SingleString query for? Why put them in a SoftValueMap?
        return queriesRun.get(singleStringQuery) != null;
    }

    /* (non-Javadoc)
     * @see org.jpox.PersistenceManager#removeQueryRun(java.lang.String)
     */
    public void removeQueryRun(String singleStringQuery)
    {
        // TODO What do we save the SingleString query for? Why put them in a SoftValueMap?
        queriesRun.remove(singleStringQuery);
    }

    // ------------------------------- Assert Utilities ---------------------------------

    /**
     * Method to assert if this Persistence Manager is open. Throws a
     * JDOFatalUserException if it has been closed.
     */
    private void assertIsOpen()
    {
        if (isClosed())
        {
            throw new JDOFatalUserException(LOCALISER.msg("PM.IsClosed"));
        }
    }

    /**
     * Utility method to check if the specified class has reachable metadata.
     * @param cls The class to check
     * @return Whether the class has reachable metadata.
     */
    private boolean hasMetaDataForPersistenceCapableClass(Class cls)
    {
        boolean hasMetaData = cls != null && getMetaDataManager().getMetaDataForClass(cls, clr) != null;
        if( !hasMetaData && cls != null && cls.isInterface() )
        {
            //try to create an implementation of the interface at runtime. It will register the MetaData 
            //and make an implementation available
            try
            {
                newInstance(cls);
            }
            catch(RuntimeException ex)
            {
                //do nothing
            }
            hasMetaData = getMetaDataManager().getMetaDataForClass(cls, clr) != null;
        }
        return hasMetaData;
    }

    /**
     * Method to assert if the specified class is Persistence Capable.
     * @param cls The class to check
     * @throws ClassNotPersistenceCapableException if class is not persistence capable
     * @throws MetaDataForPersistenceCapableClassNotReachableException if no metadata found for class
     */
    private void assertPersistenceCapableClass(Class cls)
    {
        if (!ClassUtils.isPersistenceCapableClass(cls) && !cls.isInterface())
        {
            throw new ClassNotPersistenceCapableException(cls.getName());
        }
        if (!hasMetaDataForPersistenceCapableClass(cls))
        {
            throw new MetaDataForPersistenceCapableClassNotReachableException(cls.getName());
        }
    }

    /**
     * Method to assert if the specified object is Persistence Capable.
     * @param object The object to check
     * @throws ClassNotPersistenceCapableException if not persistence capable
     * @throws MetaDataForPersistenceCapableClassNotReachableException if no metadata found for class
     */
    private void assertPersistenceCapable(Object object)
    {
        if (!ClassUtils.isPersistenceCapable(object))
        {
            throw new ClassNotPersistenceCapableException(object.getClass().getName());
        }
        if (!hasMetaDataForPersistenceCapableClass(object.getClass()))
        {
            throw new MetaDataForPersistenceCapableClassNotReachableException(object.getClass().getName());
        }
    }

    /**
     * Method to assert if the specified object is Detachable. 
     * Throws a ClassNotDetachableException if not capable
     * @param object The object to check
     */
    private void assertDetachable(Object object)
    {
        if (!ClassUtils.isDetachable(object))
        {
            throw new ClassNotDetachableException(object.getClass().getName());
        }
    }

    /**
     * Method to assert if the specified object is detached.
     * Throws a ObjectDetachedException if it is detached.
     * @param object The object to check
     */
    private void assertNotDetached(Object object)
    {
        if (JDOHelper.isDetached(object))
        {
            throw new ObjectDetachedException(object.getClass().getName());
        }
    }

    /**
     * Method to assert if the current transaction is active. Throws a
     * TransactionNotActiveException if not active
     */
    private void assertActiveTransaction()
    {
        if (!tx.isActive())
        {
            throw new TransactionNotActiveException();
        }
    }

    /**
     * Method to assert if no active transaction and nontransactionalRead is not set.
     */
    private void assertActiveTransactionOrNontransactionRead(String operation)
    {
        if (!tx.isActive() && !tx.getNontransactionalRead())
        {
            throw new JDOUserException(LOCALISER.msg("PM.NonTransactionalOperationInvalid", operation));
        }
    }
    
    /**
     * Drop any loaded state for the given field of the given PersistenceCapable, so upon next read 
     * access of that field it will get reloaded. Only possible if pc is not embedded.
     * @param pc
     * @param fieldName
     */
    public void unloadField(PersistenceCapable pc, String fieldName) 
    {
        findStateManager(pc).unloadField(fieldName);
    }    
}
