/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2003 Erik Bengtson - added getObjectByAID
2004 Andy Jefferson - added MetaDataManager
2005 Andy Jefferson - javadocs
   ...
**********************************************************************/
package org.jpox;

import java.io.PrintWriter;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.metadata.MetaDataManager;
import org.jpox.state.CallbackHandler;
import org.jpox.state.FetchPlanState;
import org.jpox.store.FieldValues;
import org.jpox.store.StoreManager;
import org.jpox.store.query.QueryResult;

/**
 * Extension of JDO PersistenceManager. 
 * Adds aspects particular to JPOX.
 *
 * @version $Revision: 1.36.2.1 $
 **/
public interface PersistenceManager extends javax.jdo.PersistenceManager
{
    /**
     * Accessor for the Store Manager.
     * @return Store Manager
     */
    StoreManager getStoreManager();

    /**
     * Accessor for the MetaData Manager.
     * @return The MetaData Manager
     */
    MetaDataManager getMetaDataManager();

    /**
     * Accessor for the context in which this PersistenceManager is running.
     * @return Returns the pmfContext.
     */
    PMFContext getPMFContext();

    /**
     * Accessor for a handle to this PersistenceManager.
     * @return The handle.
     **/
    PersistenceManager getPMHandle();

    /**
     * Accessor for the ClassLoader resolver to use in class loading issues.
     * @return The ClassLoader resolver
     */
    ClassLoaderResolver getClassLoaderResolver();

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param fv the FieldValues
     * @return the Object
     */
	Object getObjectById(Object id, FieldValues fv);

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param validate Whether to validate the object state
     * @param checkInheritance Whether look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     * @param objectClassName Class name for the object with this id (if known, optional)
     * @return The Object
     */
    Object getObjectById(Object id, boolean validate, boolean checkInheritance, String objectClassName);

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param fv the FieldValues
     * @param pcClass the type which the object is. This type will be used to instanciat the object
     * @param ignoreCache true if the cache is ignored
     * @return the Object
     */
    Object getObjectById(Object id, FieldValues fv,Class pcClass, boolean ignoreCache);
    
    /**
     * Accessor for an object given the object id. The object uses application identity
     * @param fv the FieldValues containing at minimum the primary key fields
     * @param pcClass the type which the object is
     * @param ignoreCache true if the cache is ignored
     * @param checkInheritance true if the cache is ignored
     * @return the Object
     */
    Object getObjectByAID(Class pcClass, FieldValues fv, boolean ignoreCache, boolean checkInheritance);

    /**
     * Method to persist the passed object.
     * <B>For internal use only</B>
     * @param pc The object
     * @param preInsertChanges Changes to be made before inserting
     * @return The persisted object
     */
    Object makePersistentInternal(Object pc, FieldValues preInsertChanges);

    /**
     * Method to make transient the passed object.
     * <B>For internal use only</B>
     * @param pc The object
     * @param state Object containing the state of the fetchplan processing
     */
    void internalMakeTransient(Object pc, FetchPlanState state);

    /**
     * Method to detach the passed object.
     * <B>For internal use only</B>
     * @param pc The object to detach
     * @param state State for the detachment process.
     */
    void detachInternal(Object pc, FetchPlanState state);

    /**
     * Method to detach all objects in the PM.
     * <B>For internal use only</B>
     */
    void detachAll();

    /**
     * Method to mark as clean the StateManager.
     */
    void clearDirty();

    /**
     * Method to attach a copy of the passed object (and related objects).
     * @param pc The object
     * @param embedded Whether it is embedded
     * @return The attached copy of the input object
     */
    Object attachCopy(Object pc, boolean embedded);

    /**
     * Method to detach the passed object using the provided state.
     * <B>For internal use only</B>
     * @param pc The object
     * @param state State for the detachment process
     * @return The detached object
     */
    Object detachCopyInternal(Object pc, FetchPlanState state);

    /**
     * Method to enlist the specified StateManager in the current transaction.
     * @param sm The StateManager
     */
    void enlistInTransaction(StateManager sm);

    /**
     * Method to evict the specified StateManager from the current transaction.
     * @param sm The StateManager
     */
    void evictFromTransaction(StateManager sm);

    /**
     * Accessor for whether the specified object is enlisted in the current transaction.
     * @param id Identity of the object
     * @return Whether it is enlisted
     */
    boolean isEnlistedInTransaction(Object id);

    /**
     * Method to find the StateManager for the passed PersistenceCapable object
     * when it is managed by this PM.
     * @param pc The PersistenceCapable object
     * @return The StateManager
     */
    StateManager findStateManager(PersistenceCapable pc);

    /**
     * Method to add the StateManager for an object to this PersistenceManager's
     * list.
     * @param sm The StateManager
     * @param pc The object managed by the StateManager
     */
    void hereIsStateManager(StateManager sm, PersistenceCapable pc);

    /**
     * Method to add the object managed by the specified StateManager to the cache.
     * @param sm The StateManager
     */    
    void addStateManager(StateManager sm);

    /**
     * Method to remove the object managed by the specified StateManager from the cache.
     * @param sm The StateManager
     */
    void removeStateManager(StateManager sm);

    /**
     * Accessor for the StateManager of an object given the object id.
     * @param myID Id of the object.
     * @return The StateManager
     */
    StateManager getStateManagerById(Object myID);

    /**
     * Mark the specified StateManager as dirty
     * @param sm The StateManager
     */
    void markDirty(StateManager sm);

    /**
     * Mark the specified StateManager as clean.
     * @param sm The StateManager
     */
    void clearDirty(StateManager sm);

    /**
     * Tests whether this object is being inserted.
     * @param pc the object to verify the status
     * @return true if this instance is inserting.
     */
    boolean isInserting(PersistenceCapable pc);

    /**
     * Convenience method to find if the specified field of the specified object
     * has been inserted yet.
     * @param pc The object
     * @param fieldNumber The absolute field number
     * @return Whether it has been inserted into the datastore
     */
    boolean isInserted(PersistenceCapable pc, int fieldNumber);

    /**
	 * Whether the datastore operations are delayed until commit.
     * In optimistic transactions this is automatically enabled.
	 * @return true if datastore operations are delayed until commit
     */
    boolean isDelayDatastoreOperationsEnabled();

    /**
     * Retrieve the callback handler for this PM
     * @return the callback handler
     */
    CallbackHandler getCallbackHandler();

    // -------------------------------- Query Support -------------------------------------

	/**
	 * Add opened query results to the pm
	 * @param queryResult The result to add
     * @return The results number in this PM
	 */
	int addQueryResult(QueryResult queryResult);

	/**
	 * Remove query results from the pm
	 * @param resultsNumber The number of the result to remove
	 */
	void removeQueryResult(int resultsNumber);
    
    /**
     * Add queries already run to the pm
     * @param singleStringQuery
     */
    void addQueryRun(String singleStringQuery);

    /**
     * Remove query results from the pm
     * @param singleStringQuery
     */
    void removeQueryRun(String singleStringQuery);

    /**
     * Whether the query was already run
     * @param singleStringQuery
     * @return true if the query was run previously by this PersistenceManager
     */
    boolean hasQueryRun(String singleStringQuery);

    // --------------------------------- Caching ---------------------------------------

    /**
     * Replace the previous object id for a PC object to a new
     * @param pc The Persistence Capable object
     * @param oldID the old id
     * @param newID the new id
     */
    void replaceObjectId(PersistenceCapable pc, Object oldID, Object newID);

    /**
     * Method to put a PersistenceCapable object associated to the StateManager into the respective cache(s).
     * @param sm The State Manager
     * @param level1 Whether to put in the L1 cache
     * @param level2 Whether to put in the L2 cache
     */
    void putObjectIntoCache(StateManager sm, boolean level1, boolean level2);

    /**
     * Method to remove an object from the respective cache(s).
     * @param pc The object
     * @param id The id of the object
     * @param level1 Whether to remove from the L1 cache
     * @param level2 Whether to remove from the L2 cache
     */
    void removeObjectFromCache(PersistenceCapable pc, Object id, boolean level1, boolean level2);
    
    /**
     * Drop any loaded state for the given field of the given PersistenceCapable, so upon next read 
     * access of that field it will get reloaded. Only possible if pc is not embedded.
     * @param pc
     * @param fieldName
     */
    public void unloadField(PersistenceCapable pc, String fieldName); 
    

    // ----------------------------------- Utilities ----------------------------------

    /**
     * Diagnostic method to dump the state of the PC object to the passed
     * PrintWriter.
     * @param obj The PersistenceCapable object
     * @param out The PrintWriter
     */
    void dump(Object obj, PrintWriter out);


}