/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2002 Mike Martin - unknown changes
2003 Erik Bengtson - refactored the persistent id generator System property
2003 Andy Jefferson - introduction of LOCALISER, and comments
2003 Erik Bengtson - removed unused variables
2004 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - removed autoCreate properties
2004 Andy Jefferson - added readOnly
 	...
 **********************************************************************/
package org.jpox;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import javax.jdo.PersistenceManager;
import javax.jdo.PersistenceManagerFactory;
import javax.naming.Context;
import javax.naming.Name;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import javax.naming.spi.ObjectFactory;

import org.jpox.util.JPOXLogger;

/**
 * A factory used to obtain {@link javax.jdo.PersistenceManager} instances.
 *
 * @version $Revision: 1.54 $
 */
public class PersistenceManagerFactoryImpl extends AbstractPersistenceManagerFactory 
    implements PersistenceManagerFactory, ObjectFactory, Referenceable
{

    /**
     * Return a new PersistenceManagerFactoryImpl with options set according
     * to the given Properties.
     * @param props The Properties to initialize the PersistenceManagerFactory with.
     * @return  A PersistenceManagerFactoryImpl with options set according to
     *          the given Properties.
     * @see javax.jdo.JDOHelper#getPersistenceManagerFactory(java.util.Map)
     * @exception org.jpox.exceptions.UnsupportedConnectionFactoryException Thrown if the
     * configured connection factory is not a javax.sql.DataSource 
     */
    public synchronized static PersistenceManagerFactory getPersistenceManagerFactory(Properties props)
    {
        final PersistenceManagerFactoryImpl pmf = new PersistenceManagerFactoryImpl();

        // Read properties
        pmf.setOptions(props);

        // Freeze the PMF config now that we are handing out PM's
        // see JDO 1.0.1 section 11.7. The returned object from this method has to be read-only
        pmf.freezeConfiguration();
        
        return pmf;
    }

    /**
     * Return a new PersistenceManagerFactoryImpl with options set according
     * to the given Properties.
     * @param props The Map of properties to initialize the PersistenceManagerFactory with.
     * @return  A PersistenceManagerFactoryImpl with options set according to
     *          the given Properties.
     * @see javax.jdo.JDOHelper#getPersistenceManagerFactory(java.util.Map)
     * @exception org.jpox.exceptions.UnsupportedConnectionFactoryException Thrown if the
     * configured connection factory is not a javax.sql.DataSource 
     */
    public synchronized static PersistenceManagerFactory getPersistenceManagerFactory(Map props)
    {
        final PersistenceManagerFactoryImpl pmf = new PersistenceManagerFactoryImpl();

        // Read properties from the Map
        pmf.setOptions(props);

        // Freeze the PMF config now that we are handing out PM's
        // see JDO 1.0.1 section 11.7. The returned object from this method has to be read-only
        pmf.freezeConfiguration();

        return pmf;
    }

    /**
     * Constructs a new PersistenceManagerFactoryImpl.
     */
    public PersistenceManagerFactoryImpl()
    {
        String tip = null;
        try
        {
            tip = System.getProperty(PMFConfiguration.TRANSACTION_ISOLATION_PROPERTY);
        }
        catch (SecurityException e)
        {
            JPOXLogger.GENERAL.warn(e.getMessage(), e);
        }

        if (tip != null)
        {
            try
            {
                setTransactionIsolation(tip);
            }
            catch (IllegalArgumentException e)
            {
                JPOXLogger.TRANSACTION.warn(LOCALISER.msg("PMF.TransactionIsolationInvalid", PMFConfiguration.TRANSACTION_ISOLATION_PROPERTY, tip));
            }
        }

        // Set default transaction type for J2SE operation
        setTransactionType("ResourceLocal");
    }

    /**
     * Get an instance of <tt>PersistenceManager</tt> from this factory.  The
     * instance has default values for options.
     *
     * <p>After the first use of getPersistenceManager, no "set" methods will
     * succeed.</p>
     *
     * @return a <tt>PersistenceManager</tt> instance with default options.
     * @exception org.jpox.exceptions.UnsupportedConnectionFactoryException Thrown if the
     * configured connection factory is not a javax.sql.DataSource 
     */
    public synchronized PersistenceManager getPersistenceManager()
    {
        assertIsOpen();
        
        // Freeze the PMF config now that we are handing out PM's
        freezeConfiguration();

        return getPersistenceManager(this.getPMFConfiguration().getConnectionUserName(), this.getPMFConfiguration().getConnectionPassword());
    }

    /**
     * Get an instance of <tt>PersistenceManager</tt> from this factory.  The
     * instance has default values for options.  The parameters userid and
     * password are used when obtaining datastore connections from the
     * connection pool.
     *
     * <p>After the first use of getPersistenceManager, no "set" methods will
     * succeed.</p>
     *
     * @param userName  the user name for the connection
     * @param password  the password for the connection
     *
     * @return a <tt>PersistenceManager</tt> instance with default options.
     * @exception org.jpox.exceptions.UnsupportedConnectionFactoryException Thrown if the
     * configured connection factory is not a javax.sql.DataSource 
     */
    public synchronized PersistenceManager getPersistenceManager(String userName, String password)
    {
        assertIsOpen();

        // Freeze the PMF config now that we are handing out PM's
        freezeConfiguration();
        PersistenceManager pm = new PersistenceManagerImpl(this, userName, password);

        // Add any lifecycle listeners to the PM
        if (lifecycleListeners != null)
        {
            Iterator listenerIter = lifecycleListeners.iterator();
            while (listenerIter.hasNext())
            {
                LifecycleListenerForClass listener = (LifecycleListenerForClass) listenerIter.next();
                pm.addInstanceLifecycleListener(listener.getListener(), listener.getClasses());
            }
        }

        getPmCache().add(pm);

        return pm;
    }

    /**
     * Equality operator.
     * @param obj Object to compare against
     * @return Whether the objects are the same.
     **/
    public synchronized boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof PersistenceManagerFactoryImpl))
        {
            return false;
        }

        return super.equals(obj);
    }

    /**
     * Uses StringRefAddr's to store the information
     * @param obj The object
     * @param name Name
     * @param ctx Context
     * @param env properties
     * @return The instance
     * @throws Exception
     */
    public Object getObjectInstance(Object obj, Name name, Context ctx, Hashtable env) throws Exception
    {
        PersistenceManagerFactoryImpl rc = null;

        if (obj instanceof Reference)
        {
            Reference ref = (Reference) obj;
            if (ref.getClassName().equals(ClassNameConstants.PersistenceManagerFactoryImpl))
            {
                Properties p = new Properties();
                for (Enumeration e = ref.getAll(); e.hasMoreElements();)
                {
                    StringRefAddr sra = (StringRefAddr) e.nextElement();
                    p.setProperty(sra.getType(), (String) sra.getContent());
                }

                rc = new PersistenceManagerFactoryImpl();
                rc.setOptions(p);
            }
        }
        return rc;
    }

    /**
     * Uses StringRefAddr's to store the information
     * @return The reference
     */
    public Reference getReference()
    {
        Reference rc = null;
        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        try
        {
            ObjectOutputStream oos = new ObjectOutputStream(baos);
            oos.writeObject(this);
            rc = new Reference(ClassNameConstants.PersistenceManagerFactoryImpl, ClassNameConstants.PersistenceManagerFactoryImpl, null);
            
            Map p = getOptions();
            for (Iterator i = p.keySet().iterator(); i.hasNext();)
            {
                String key = (String) i.next();
                if( p.get(key) instanceof String )
                {
                    String value = (String) p.get(key);
                    rc.add(new StringRefAddr(key, value));
                }
            }
        }
        catch (IOException ex)
        {
            // do something
        }
        return rc;
    }
}