/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
2004 Erik Bengtson - added interfaces methods
    ...
**********************************************************************/
package org.jpox;

import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.jdo.JDOException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.plugin.ConfigurationElement;
import org.jpox.plugin.ExtensionPoint;
import org.jpox.plugin.PluginManager;
import org.jpox.sco.SCOList;
import org.jpox.sco.SCOMap;
import org.jpox.util.JPOXLogger;
import org.jpox.util.JavaUtils;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Registry of type support within JPOX. Any new types will be added to the registry
 * here. Provides global registry information. <B>MappingManager</B> provides a similar
 * type of registry however is for a particular PersistenceManagerFactory and so can,
 * in principle, include differencies for particular datastores.
 *
 * @version $Revision: 1.42 $
 */
public class TypeManager
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.Localisation");

    /** The types, keyed by the class name. */
    Map types = new HashMap();

    private static String[] DEFAULT_PERSISTENT_TYPE;
    private static String[] DEFAULT_FETCH_GROUP_TRUE;
    private static String[] DEFAULT_EMBEDDED_TYPE;

    /**
     * Constructor
     * initialisation block to load up JPOX's support for type mappings.
     * This adds entries for the majority of types that JDO requires, or allows
     * as optional.
     * @param clr the ClassLaoderResolver
     * @param mgr the PluginManager
     **/
    public TypeManager(PluginManager mgr, ClassLoaderResolver clr)
    {
        loadUserTypes(mgr, clr);
    }

    /**
     * Method to add support for a Java class (to some degree).
     * @param mgr The PluginManager
     * @param pluginId the plug-in id
     * @param className Name of the class to add
     * @param mappingClassName The Java mapping type
     * @param scoMutableWrapperType The SCO wrapper type (for mutable SCOs)
     * @param dfg whether this type should be on the default fetch group
     * @param javaVersion the minimum java version required at runtime to add this type
     * @param dftPersistent if this type is persistent by default. If not, the metadata must explicitily set persistent-modifier="persistent"
     * @param javaRestricted if this type is restricted only to the specified java version
     * @param dftEmbedded If this type is by default embedded
     * @param clr the ClassLoaderResolver
     */
    public void addType(PluginManager mgr,
                        String pluginId,
                        String className,
                        String mappingClassName,
                        String scoMutableWrapperType,
                        boolean dfg,
                        String javaVersion,
                        boolean dftPersistent,
                        boolean javaRestricted,
                        boolean dftEmbedded,
                        ClassLoaderResolver clr)
    {
        if (className == null)
        {
            return;
        }
        if ((JavaUtils.isGreaterEqualsThan(javaVersion) && !javaRestricted) ||
            (JavaUtils.isEqualsThan(javaVersion) && javaRestricted))
        {
            Class wrapperType = null;
            Class mappingType = null;
            if (mappingClassName != null && !StringUtils.isWhitespace(mappingClassName))
            {
                try
                {
                    mappingType = mgr.loadClass(pluginId,mappingClassName);
                }
                catch (JDOException jdoe)
                {
                    JPOXLogger.METADATA.error(LOCALISER.msg("Types.User.MappingNotFound", mappingClassName));
                    return;
                }
            }
            if (scoMutableWrapperType != null && !StringUtils.isWhitespace(scoMutableWrapperType))
            {
                try
                {
                    wrapperType = mgr.loadClass(pluginId,scoMutableWrapperType);
                }
                catch (JDOException jdoe)
                {
                    JPOXLogger.METADATA.error(LOCALISER.msg("Types.User.SCOWrapperNotFound", scoMutableWrapperType));
                    return;
                }
            }
            JPOXType type = new JPOXType(className, mappingType, wrapperType, dfg, javaVersion, dftPersistent,
                javaRestricted, dftEmbedded);
            if (dftPersistent)
            {
                addDefaultPersistent(className);
            }
            if (dfg)
            {
                addDefaultFetchGroup(className);
            }
            if (dftEmbedded)
            {
                addDefaultEmbedded(className);
            }
            types.put(className, type);
            if (JPOXLogger.GENERAL.isDebugEnabled())
            {
                JPOXLogger.GENERAL.debug(LOCALISER.msg("Type.AddType", className, "" + dftPersistent, "" + dfg, "" + dftEmbedded));
            }
        }
    }

    /**
     * Convenience method returning whether the passed type is a reference type.
     * This typically means whether it is an interface or java.lang.Object.
     * @param cls The class
     * @return Whether it is a reference type
     */
    public boolean isReferenceType(Class cls)
    {
        if (cls == null)
        {
            return false;
        }
        return cls.isInterface() || cls.getName().equals("java.lang.Object");
    }

    /**
     * Accessor for whether a class is supported.
     * @param className The class name
     * @return Whether the class is supported (to some degree)
     */
    public boolean isSupportedType(String className)
    {
        if (className == null)
        {
            return false;
        }

        JPOXType type = getType(className);
        if (type == null)
        {
            return false;
        }
        else
        {
            return (type.getSCOMappingType() != null);
        }
    }

    /**
     * Accessor for the Supported Types.
     * @return Set of supported types (containing the fully qualified class names (String)).
     **/
    public Set getSupportedTypes()
    {
        return new HashSet(types.keySet());
    }

    /**
     * Accessor for whether a class is a supported second class mutable type
     * @param class_name The class name
     * @return Whether it is a second class mutable type
     */
    public boolean isSecondClassMutableType(String class_name)
    {
        return (getSecondClassWrapper(class_name) != null);
    }

    /**
     * Accessor for whether the type is a List type that uses a SCOList wrapper type.
     * @param c The class
     * @return if c is a List type
     */
    public boolean isSCOList(Class c)
    {
        if (c == null)
        {
            return false;
        }
        if (c.isInterface() && !isSupportedType(c.getName()))
        {
        	return false;
        }

        JPOXType type = getType(c.getName());
        return type != null ? SCOList.class.isAssignableFrom(type.getSCOMutableWrapperType()) : false;
    }

    /**
     * Accessor for whether the type is a Map type that uses a SCOMap wrapper type.
     * @param c The class
     * @return true if c is a Map type
     */
    public boolean isSCOMap(Class c)
    {
        if (c == null)
        {
            return false;
        }
        if (c.isInterface() && !isSupportedType(c.getName()))
        {
        	return false;
        }

        JPOXType type = getType(c.getName());
        return type != null ? SCOMap.class.isAssignableFrom(type.getSCOMutableWrapperType()) : false;
    }

    /**
     * Accessor for the Java Mapping type class for the supplied class.
     * @param className The class name
     * @return The Java mapping type
     */
    public Class getMappingType(String className)
    {
        if (className == null)
        {
            return null;
        }

        JPOXType type = getType(className);
        return type == null ? null : type.getSCOMappingType();
    }

    /**
     * Accessor for the Second Class Wrapper class for the supplied class.
     * A type will have a SCO wrapper if it is SCO supported and is mutable.
     * @param className The class name
     * @return The second class wrapper
     */
    public Class getSecondClassWrapper(String className)
    {
        if (className == null)
        {
            return null;
        }

        JPOXType type = getType(className);
        return type == null ? null : type.getSCOMutableWrapperType();
    }

    /**
     * Accessor for whether a class is a Second Class Wrapper type.
     * @param className The name of the class
     * @return Whether it is a second class wrapper type
     */
    public boolean isSecondClassWrapper(String className)
    {
        if (className == null)
        {
            return false;
        }

        Iterator iter = types.values().iterator();
        while (iter.hasNext())
        {
            JPOXType type = (JPOXType)iter.next();
            if (type.getSCOMutableWrapperType() != null &&
                type.getSCOMutableWrapperType().getName().equals(className))
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Utility class to retrieve a supported type
     * @param className The class name
     * @return The internal type information for the class
     */
    protected JPOXType getType(String className)
    {
        if (className == null)
        {
            return null;
        }

        return (JPOXType)types.get(className);
    }

    /**
     * Inner class defining a supported type. The type is supported to some
     * degree by JPOX, maybe FCO, SCO, or DFG.
     *
     * Usually only has SCO wrapper classes, if the mapped class
     */
    static class JPOXType
    {
        /** the mapped class */
        final String className;
        /** mapping class. An extension of {@link org.jpox.store.mapping.JavaTypeMapping}*/
        final Class scoMappingType;
        /** sco wrapper for mutable types. An implementation of {@link org.jpox.sco.SCO} */
        final Class scoMutableWrapperType;
        /** default-fetch-group */
        final boolean dfg;
        /** minimum required java version **/
        final String javaVersion;
        /** if this type is persistent by default. If not, the metadata must explicitily set persistent-modifier="persistent" **/
        final boolean dftPersistent;
        /** if this type is embedded by default. If not, the metadata must explicitily set embedded="true" **/
        final boolean dftEmbedded;
        /** if this type is restricted only to the specified java version */
        final boolean javaRestricted;

        /**
         * Constructor.
         * @param className the java type class name
         * @param sco_mapping_type the mapping class. An extension of {@link org.jpox.store.mapping.JavaTypeMapping}
         * @param sco_wrapper_type the SCO wrapper class name. An implementation of {@link org.jpox.sco.SCO}
         * @param dfg default fetch group
         * @param javaVersion minimum required java version
         * @param dftPersistent if this type is persistent by default
         * @param javaRestricted if this type is restricted only to the specified java version by <code>javaVersion</code>
         * @param dftEmbedded if this type is embedded by default
         */
        public JPOXType(String className,
                        Class sco_mapping_type,
                        Class sco_wrapper_type,
                        boolean dfg,
                        String javaVersion,
                        boolean dftPersistent,
                        boolean javaRestricted,
                        boolean dftEmbedded)
        {
            this.className = className;
            this.scoMappingType = sco_mapping_type;
            this.scoMutableWrapperType = sco_wrapper_type;
            this.dfg = dfg;
            this.javaVersion = javaVersion;
            this.dftPersistent = dftPersistent;
            this.javaRestricted = javaRestricted;
            this.dftEmbedded = dftEmbedded;
        }

        /**
         * Accessor for the mapping class. An extension of {@link org.jpox.store.mapping.JavaTypeMapping}
         * @return the mapping class. An extension of {@link org.jpox.store.mapping.JavaTypeMapping}
         */
        public Class getSCOMappingType()
        {
            return scoMappingType;
        }

        /**
         * Accessor for the SCO wrapper class name. An implementation of {@link org.jpox.sco.SCO}
         * @return the SCO wrapper class name. An implementation of {@link org.jpox.sco.SCO}
         */
        public Class getSCOMutableWrapperType()
        {
            return scoMutableWrapperType;
        }

        public String toString()
        {
            StringBuffer str = new StringBuffer(className + " [");
            if (scoMappingType != null)
            {
                str.append(" SCO : mapping=" + scoMappingType);
            }
            if (scoMutableWrapperType != null)
            {
                str.append(" SCO : mutable wrapper=" + scoMutableWrapperType);
            }
            str.append(" default-fetch-group : " + dfg);
            str.append(" java version : " + javaVersion + (javaRestricted ? " restricted" : ""));
            str.append(" default persisted : " + dftPersistent);
            str.append(" default embedded : " + dftEmbedded);
            str.append("]");
            return str.toString();
        }
    }

    /**
     * Add a default persistent type
     * @param className the type name
     */
    private synchronized void addDefaultPersistent(String className)
    {
        if (DEFAULT_PERSISTENT_TYPE == null)
        {
            DEFAULT_PERSISTENT_TYPE = new String[1];
            DEFAULT_PERSISTENT_TYPE[0] = className;
        }
        else
        {
            String[] temp = new String[DEFAULT_PERSISTENT_TYPE.length+1];
            System.arraycopy(DEFAULT_PERSISTENT_TYPE,0,temp,0,DEFAULT_PERSISTENT_TYPE.length);
            temp[temp.length-1] = className;
            Arrays.sort(temp);
            DEFAULT_PERSISTENT_TYPE = temp;
        }
    }

    /**
     * Add a default fetch group type
     * @param className the type name
     */
    private synchronized void addDefaultFetchGroup(String className)
    {
        if (DEFAULT_FETCH_GROUP_TRUE == null)
        {
            DEFAULT_FETCH_GROUP_TRUE = new String[1];
            DEFAULT_FETCH_GROUP_TRUE[0] = className;
        }
        else
        {
            String[] temp = new String[DEFAULT_FETCH_GROUP_TRUE.length+1];
            System.arraycopy(DEFAULT_FETCH_GROUP_TRUE,0,temp,0,DEFAULT_FETCH_GROUP_TRUE.length);
            temp[temp.length-1] = className;
            Arrays.sort(temp);
            DEFAULT_FETCH_GROUP_TRUE = temp;
        }
    }

    /**
     * Add a default embedded type
     * @param className the type name
     */
    private synchronized void addDefaultEmbedded(String className)
    {
        if (DEFAULT_EMBEDDED_TYPE == null)
        {
            DEFAULT_EMBEDDED_TYPE = new String[1];
            DEFAULT_EMBEDDED_TYPE[0] = className;
        }
        else
        {
            String[] temp = new String[DEFAULT_EMBEDDED_TYPE.length+1];
            System.arraycopy(DEFAULT_EMBEDDED_TYPE,0,temp,0,DEFAULT_EMBEDDED_TYPE.length);
            temp[temp.length-1] = className;
            Arrays.sort(temp);
            DEFAULT_EMBEDDED_TYPE = temp;
        }
    }

    /**
     * Accessor for the default "persistence-modifier" for a field given the
     * class, its modifier and whether it is a PersistentCapable class.
     * @param c The class
     * @param modifier The modifiers for the field
     * @param isPCclass Whether it is persistence capable.
     * @return The default field PersistenceModifier.
     */
    public final FieldPersistenceModifier getDefaultFieldPersistenceModifier(Class c, int modifier, boolean isPCclass)
    {
        // Static, Transient and Final fields are all not persistent
        if (Modifier.isStatic(modifier) ||
            Modifier.isTransient(modifier) ||
            Modifier.isFinal(modifier))
        {
            // Static, transient and final fields are not persistent
            return FieldPersistenceModifier.NONE;
        }

        if (isPCclass)
        {
            // All PC class fields are persistent by default
            return FieldPersistenceModifier.PERSISTENT;
        }

        if (c == null)
        {
            throw new RuntimeException("class is null");
        }

        if (Arrays.binarySearch(DEFAULT_PERSISTENT_TYPE, c.getName()) >= 0)
        {
            // Use the specified "jpoxtypes" default persistence modifier for this type (if specified)
            return FieldPersistenceModifier.PERSISTENT;
        }
        if (c.isArray() && PersistenceCapable.class.isAssignableFrom(c.getComponentType()))
        {
            // Arrays of PC types are, by default, persistent
            return FieldPersistenceModifier.PERSISTENT;
        }

        return FieldPersistenceModifier.NONE;
    }

    /**
     * Accessor for whether the class is by default in the default fetch group.
     * @param c The class
     * @return return true if is part of default fetch group by default
     */
    public boolean isDefaultFetchGroup(Class c)
    {
        if (c == null)
        {
            return false;
        }
        if (Arrays.binarySearch(DEFAULT_FETCH_GROUP_TRUE, c.getName()) >= 0)
        {
            return true;
        }

        return false;
    }

    /**
     * Utility to return whether a type defaults to embedded or not in JDO.
     * See JDO 2 Section 18.14 &lt;field&gt; for definition. This only provides for the
     * basic types, and ignores arrays/collections.
     * @param type The type
     * @return Whether it defaults to embedded or not.
     */
    public boolean isDefaultEmbeddedType(Class type)
    {
        if (type == null)
        {
            return false;
        }
        if (Arrays.binarySearch(DEFAULT_EMBEDDED_TYPE,type.getName()) >= 0)
        {
            return true;
        }

        return false;
    }
    
    /**
     * Method to load the user type mappings that are currently registered in the PluginManager.
     * @param mgr the PluginManager
     * @param clr the ClassLoaderResolver
     */
    private void loadUserTypes(PluginManager mgr, ClassLoaderResolver clr)
    {
        ExtensionPoint exPoint = mgr.getExtensionPoint("org.jpox.store_mapping");
        for( int i=0; i<exPoint.getExtensions().length; i++)
        {
            ConfigurationElement[] elms = exPoint.getExtensions()[i].getConfigurationElements();
            for( int e=0; e<elms.length; e++)
            {
                String javaName = elms[e].getAttribute("java-type").trim();
                String mappingClassName = elms[e].getAttribute("mapping-class");
                String scoWrapperClassName = elms[e].getAttribute("sco-wrapper");
                String defaultFetchGroup = elms[e].getAttribute("default-fetch-group");
                String defaultPersistent = elms[e].getAttribute("default-persistent");
                String defaultEmbedded = elms[e].getAttribute("default-embedded");
                String javaVersionRestrict = elms[e].getAttribute("java-version-restrict");
                boolean dfg = false;
                if( defaultFetchGroup != null )
                {
                    if (defaultFetchGroup.equalsIgnoreCase("true"))
                    {
                        dfg = Boolean.TRUE.booleanValue();
                    }
                }
                boolean dftPersistent = false;
                if( defaultPersistent != null )
                {
                    if (defaultPersistent.equalsIgnoreCase("true"))
                    {
                        dftPersistent = Boolean.TRUE.booleanValue();
                    }
                }
                boolean dftEmbedded = false;
                if (defaultEmbedded != null)
                {
                    if (defaultEmbedded.equalsIgnoreCase("true"))
                    {
                        dftEmbedded = Boolean.TRUE.booleanValue();
                    }
                }
                boolean javaRestricted = false;
                if( javaVersionRestrict != null )
                {
                    if (javaVersionRestrict.equalsIgnoreCase("true"))
                    {
                        javaRestricted = Boolean.TRUE.booleanValue();
                    }
                }                    
                String javaVersion = elms[e].getAttribute("java-version");
                if( javaVersion == null || javaVersion.length()<1)
                {
                    javaVersion = "1.3";
                }

                addType(mgr, elms[e].getExtension().getPlugin().getId(), javaName, mappingClassName, scoWrapperClassName, dfg, javaVersion, 
                        dftPersistent, javaRestricted, dftEmbedded, clr);
            }
        }
    }    
}