/**********************************************************************
 Copyright (c) 2006 Erik Bengtson and others. All rights reserved. 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
 

 Contributors:
 ...
 **********************************************************************/
package org.jpox.api;

import java.util.HashMap;
import java.util.Map;

import org.jpox.AbstractPersistenceManagerFactory;
import org.jpox.ClassLoaderResolver;
import org.jpox.plugin.ConfigurationElement;
import org.jpox.plugin.Extension;
import org.jpox.plugin.PluginManager;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Factory of API adapters for JPOX. 
 *
 * @version $Revision: 1.5 $ 
 */
public class ApiAdapterFactory
{
    /** Localisation of messages */
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /** The adapter instances, mapped by naming string. */
    HashMap adapters = new HashMap();

    /** Singleton instance */
    static ApiAdapterFactory adapterMgr = new ApiAdapterFactory();

    /**
     * Accessor for the ApiAdapterFactory for JPOX (singleton)
     * @return The manager of type information
     */
    public static ApiAdapterFactory getInstance()
    {
        return adapterMgr;
    }

    /**
     * Protected constructor to prevent outside instantiation
     */
    protected ApiAdapterFactory()
    {
    }

    /**
     * Method to add support for an API via an adapter.
     * @param name name of the API
     * @param apiAdapter instance of adapter
     */
    private void addAdapter(String name, ApiAdapter apiAdapter)
    {
        if (name == null || apiAdapter == null)
        {
            return;
        }

        adapters.put(name, apiAdapter);
    }

    /**
     * Accessor for the registered Api Adapters
     * @return A Map with ApiAdapter instances
     */
    public Map getRegisteredApiAdapters()
    {
        return adapters;
    }

    /**
     * Accessor for an adapter, given the api name.
     * @param name the adapter name
     * @return The ApiAdapter
     **/
    public ApiAdapter getApiAdapter(String name)
    {
        return (ApiAdapter) adapters.get(name);
    }

    /**
     * Accessor for the adapter for a specified api name.
     * @param clr ClassLoader resolver for resolving the adapter class
     * @param className the class name of the ApiAdapter
     * @return Instance of the ApiAdapter
     */
    private ApiAdapter newApiAdapter(ClassLoaderResolver clr, String className)
    {
        if (className == null)
        {
            return null;
        }

        final Object adapter_obj;
        Class adapterClass;

        // Create an instance of the datastore adapter
        try
        {
            adapterClass = clr.classForName(className, AbstractPersistenceManagerFactory.class.getClassLoader());
            adapter_obj = adapterClass.newInstance();
        }
        catch (Exception e)
        {
            JPOXLogger.RDBMS.error(LOCALISER.msg("Api.Adapter.InstantiationError", className, e));
            return null;
        }

        return (ApiAdapter) adapter_obj;
    }

    /**
     * Load "org.jpox.api.adapter" extensions 
     * @param clr ClassLoader resolver for resolving the adapter class
     * @param pluginMgr the Plug-in Manager
     */
    public void initializeApiAdapterExtensions(ClassLoaderResolver clr, PluginManager pluginMgr)
    {
        if (adapters.size() > 0)
        {
            //already loaded
            return;
        }
        Extension[] ex = pluginMgr.getExtensionPoint("org.jpox.api_adapter").getExtensions();
        for (int i = 0; i < ex.length; i++)
        {
            ConfigurationElement[] confElm = ex[i].getConfigurationElements();
            for (int c = 0; c < confElm.length; c++)
            {

                adapterMgr.addAdapter(confElm[c].getAttribute("name"), newApiAdapter(clr, confElm[c].getAttribute("class-name")));
            }
        }
    }
}