/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
2004 Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigArray
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.List;
import java.util.Set;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.ClassNameConstants;

/**
 * Representation of the Meta-Data for an Array.
 * <P>Represents the following MetaData Element 
 * <PRE>
 * &lt;!ELEMENT array (extension)*&gt;
 * &lt;!ATTLIST array element-type CDATA #IMPLIED&gt;
 * &lt;!ATTLIST array embedded-element (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST array serialized-element (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST array dependent-element (true|false) #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.22 $
 */
public class ArrayMetaData extends ContainerMetaData
{
    /** Implementation type(s) of elements that can be stored (when it is of a reference type). */
    protected String elementType;

    /** Element-type ClassMetaData */
    protected AbstractClassMetaData elementClassMetaData;

    /** Whether the array elements are embedded. */
    protected Boolean embeddedElement;

    /** Whether the array elements are serialized. */
    protected Boolean serializedElement;

    /** Whether the array elements are to be treated as dependent when removing from the array. */
    protected Boolean dependentElement;

    /**
     * Constructor to create a copy of the passed metadata using the passed parent.
     * @param parent The parent
     * @param arrmd The metadata to copy
     */
    public ArrayMetaData(AbstractPropertyMetaData parent, ArrayMetaData arrmd)
    {
        super(parent);
        this.embeddedElement = arrmd.embeddedElement;
        this.serializedElement = arrmd.serializedElement;
        this.dependentElement = arrmd.dependentElement;
        this.elementType = arrmd.elementType;
        this.elementClassMetaData = arrmd.elementClassMetaData;
    }

    /**
     * Constructor.
     * @param parent The parent Field
     * @param elementType Implementation type(s) of element possible (when it is a reference type).
     * @param embeddedElement Whether the element is embedded into the join table
     * @param dependentElement Whether the element is dependent for deletion purposes
     * @param serializedElement Whether the element is serialised into the join table
     */
    public ArrayMetaData(AbstractPropertyMetaData parent,
                         String elementType,
                         String embeddedElement,
                         String dependentElement,
                         String serializedElement)
    {
        super(parent);

        this.elementType = elementType;

        if (embeddedElement != null)
        {
            if (embeddedElement.equalsIgnoreCase("true"))
            {
                this.embeddedElement = Boolean.TRUE;
            }
            else if (embeddedElement.equalsIgnoreCase("false"))
            {
                this.embeddedElement = Boolean.FALSE;
            }
        }

        if (serializedElement != null)
        {
            if (serializedElement.equalsIgnoreCase("true"))
            {
                this.serializedElement = Boolean.TRUE;
            }
            else if (serializedElement.equalsIgnoreCase("false"))
            {
                this.serializedElement = Boolean.FALSE;
            }
        }

        if (dependentElement != null)
        {
            if (dependentElement.equalsIgnoreCase("true"))
            {
                this.dependentElement = Boolean.TRUE;
            }
            else if (dependentElement.equalsIgnoreCase("false"))
            {
                this.dependentElement = Boolean.FALSE;
            }
        }
    }

    /**
     * Method to populate any defaults, and check the validity of the MetaData.
     * @param clr ClassLoaderResolver to use in loading any classes 
     */
    public void populate(ClassLoaderResolver clr)
    {
        // Check the field type and see if it is an array type
        Class field_type = getAbstractPropertyMetaData().getType();
        if (!field_type.isArray())
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Array.IsNotArrayError",
                    getFieldName(),getAbstractPropertyMetaData().getClassName(false));
        }

        // "embedded-element"
        if (embeddedElement == null)
        {
            // Assign default for "embedded-element" based on 18.13.1 of JDO 2 spec
            // Note : this fails when using in the enhancer since not yet PC
            Class component_type = field_type.getComponentType();
            if (getMetaDataManager().getPMFContext().getTypeManager().isDefaultEmbeddedType(component_type))
            {
                embeddedElement = Boolean.TRUE;
            }
            else if (PersistenceCapable.class.isAssignableFrom(component_type) ||
                Object.class.isAssignableFrom(component_type) ||
                component_type.isInterface())
            {
                embeddedElement = Boolean.FALSE;
            }
            else
            {
                embeddedElement = Boolean.TRUE;
            }
        }
        if (embeddedElement == Boolean.FALSE)
        {
            // If the user has set a non-PC/non-Interface as not embedded, correct it since not supported.
            // Note : this fails when using in the enhancer since not yet PC
            Class component_type = field_type.getComponentType();
            if (!(PersistenceCapable.class.isAssignableFrom(component_type)) &&
                !(component_type.isInterface()) &&
                component_type != java.lang.Object.class)
            {
                embeddedElement = Boolean.TRUE;
            }
        }

        if (!getAbstractPropertyMetaData().getAbstractClassMetaData().getMetaDataManager().isEnhancing() && !getAbstractPropertyMetaData().isSerialized())
        {
            // Catch situations that JPOX doesnt support
            if (getAbstractPropertyMetaData().getJoinMetaData() == null &&
                !PersistenceCapable.class.isAssignableFrom(getAbstractPropertyMetaData().getType().getComponentType()))
            {
                // JPOX only supports persisting particular array types as byte-streams (non-Java-serialised)
                String arrayComponentType = getAbstractPropertyMetaData().getType().getComponentType().getName();
                if (!arrayComponentType.equals(ClassNameConstants.BOOLEAN) &&
                    !arrayComponentType.equals(ClassNameConstants.BYTE) &&
                    !arrayComponentType.equals(ClassNameConstants.CHAR) &&
                    !arrayComponentType.equals(ClassNameConstants.DOUBLE) &&
                    !arrayComponentType.equals(ClassNameConstants.FLOAT) &&
                    !arrayComponentType.equals(ClassNameConstants.INT) &&
                    !arrayComponentType.equals(ClassNameConstants.LONG) &&
                    !arrayComponentType.equals(ClassNameConstants.SHORT) &&
                    !arrayComponentType.equals(ClassNameConstants.JAVA_LANG_BOOLEAN) &&
                    !arrayComponentType.equals(ClassNameConstants.JAVA_LANG_BYTE) &&
                    !arrayComponentType.equals(ClassNameConstants.JAVA_LANG_CHARACTER) &&
                    !arrayComponentType.equals(ClassNameConstants.JAVA_LANG_DOUBLE) &&
                    !arrayComponentType.equals(ClassNameConstants.JAVA_LANG_FLOAT) &&
                    !arrayComponentType.equals(ClassNameConstants.JAVA_LANG_INTEGER) &&
                    !arrayComponentType.equals(ClassNameConstants.JAVA_LANG_LONG) &&
                    !arrayComponentType.equals(ClassNameConstants.JAVA_LANG_SHORT) &&
                    !arrayComponentType.equals("BigDecimal") &&
                    !arrayComponentType.equals("BigInteger"))
                {
                    // Impossible to persist an array of a non-PC element without a join table or without serialising the array
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Array.NonPCWithoutJoin",
                        getFieldName(), getAbstractPropertyMetaData().getType().getComponentType().getName());
                }
            }
        }

        // Keep a reference to the MetaData for the element
        MetaDataManager mmgr = getAbstractPropertyMetaData().getAbstractClassMetaData().getMetaDataManager();
        if (elementType != null)
        {
            Class elementCls = clr.classForName(elementType);
            elementClassMetaData = mmgr.getMetaDataForClassInternal(elementCls, clr);
        }
        else
        {
            elementType = field_type.getComponentType().getName();
            elementClassMetaData = mmgr.getMetaDataForClassInternal(field_type.getComponentType(), clr);
        }

        setPopulated();
    }

    /**
     * Accessor for the element implementation types (when element is a reference type).
     * The return can contain comma-separated values.
     * @return element implementation types
     */
    public String getElementType()
    {
        return elementType;
    }

    /**
     * Accessor for the Element ClassMetaData
     * @return element ClassMetaData
     */
    public AbstractClassMetaData getElementClassMetaData()
    {
        if (elementClassMetaData != null && !elementClassMetaData.isInitialised())
        {
            elementClassMetaData.initialise();
        }
        return elementClassMetaData;
    }

    /**
     * Accessor for the embedded-element value
     * @return embedded-element value
     */
    public boolean isEmbeddedElement()
    {
        if (embeddedElement == null)
        {
            return false;
        }
        else
        {
            return embeddedElement.booleanValue();
        }
    }

    /**
     * Accessor for the serialized-element tag value
     * @return serialized-element tag value
     */
    public boolean isSerializedElement()
    {
        if (serializedElement == null)
        {
            return false;
        }
        else
        {
            return serializedElement.booleanValue();
        }
    }

    /**
     * Accessor for The dependent-element attribute indicates that the
     * collection's element contains a reference that is to be deleted if the
     * referring instance is deleted.
     * 
     * @return dependent-element tag value
     */
    public boolean isDependentElement()
    {
        if (dependentElement == null)
        {
            return false;
        }
        else
        {
            return dependentElement.booleanValue();
        }
    }

    /**
     * Accessor for all AbstractClassMetaData referenced by this array.
     * @param orderedCMDs List of ordered AbstractClassMetaData objects (added to).
     * @param referencedCMDs Set of all AbstractClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param clr the ClassLoaderResolver 
     **/
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr)
    {
        AbstractClassMetaData element_cmd = getMetaDataManager().getMetaDataForClass(getAbstractPropertyMetaData().getType().getComponentType(), clr);
        if (element_cmd != null)
        {
            element_cmd.getReferencedClassMetaData(orderedCMDs, referencedCMDs, dba_vendor_id, clr);
        }
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * @param prefix The prefix string 
     * @param indent The indent string 
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<array");
        if (elementType != null)
        {
            sb.append(" element-type=\"").append(elementType).append("\"");
        }
        if (embeddedElement != null)
        {
            sb.append(" embedded-element=\"").append(embeddedElement).append("\"");
        }
        if (serializedElement != null)
        {
            sb.append(" serialized-element=\"").append(serializedElement).append("\"");
        }
        if (dependentElement != null)
        {
            sb.append(" dependent-element=\"").append(dependentElement).append("\"");
        }
        
        if (getNoOfExtensions() > 0)
        {
            sb.append(">\n");

            // Add extensions
            sb.append(super.toString(prefix + indent,indent));

            sb.append(prefix).append("</array>\n");
        }
        else
        {
            sb.append(prefix).append("/>\n");
        }

        return sb.toString();
    }
}