/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2004 Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigCollection
2004 Erik Bengtson - add dependent elements
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.List;
import java.util.Set;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * Representation of the MetaData of a collection.
 *
 * <H3>MetaData Element</H3>
 * Represents the following MetaData Element 
 * <PRE>
 * &lt;!ELEMENT collection (extension)*&gt;
 * &lt;!ATTLIST collection element-type CDATA #IMPLIED&gt;
 * &lt;!ATTLIST collection embedded-element (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST collection dependent-element (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST collection serialized-element (true|false) #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.26 $
 */
public class CollectionMetaData extends ContainerMetaData
{
    /** Element-type tag value. Default is Object (JDO2 Spec [18.13.1]). */
    protected String elementType = "java.lang.Object";

    /** Element-type ClassMetaData */
    protected AbstractClassMetaData elementClassMetaData;
    
    /** Embedded-element tag value. */
    protected Boolean embeddedElement;

    /** Serialized-element tag value. */
    protected Boolean serializedElement;

    /**
     * dependent-element tag value.
     * 
     * The dependent-element attribute indicates that the collection's element
     * contains a reference that is to be deleted if the referring instance is
     * deleted.
     */
    protected Boolean dependentElement;

    /**
     * Constructor to create a copy of the passed metadata using the passed parent.
     * @param parent The parent
     * @param colmd The metadata to copy
     */
    public CollectionMetaData(AbstractPropertyMetaData parent, CollectionMetaData colmd)
    {
        super(parent);
        this.embeddedElement = colmd.embeddedElement;
        this.serializedElement = colmd.serializedElement;
        this.dependentElement = colmd.dependentElement;
        this.elementType = colmd.elementType;
        this.elementClassMetaData = colmd.elementClassMetaData;
    }

    /**
     * Constructor.
     * @param parent The parent Field element
     * @param elementType The type of element
     * @param embeddedElement Whether elements are embedded 
     * @param dependentElement attribute dependent-element value
     * @param serializedElement attribute serialized-element value
     */
    public CollectionMetaData(AbstractPropertyMetaData parent,
                              String elementType,
                              String embeddedElement,
                              String dependentElement,
                              String serializedElement)
    {
        super(parent);

        if (!StringUtils.isWhitespace(elementType))
        {
            // Apply JDO naming rules - if no package, then assumed to be in same package as this class
            this.elementType = ClassUtils.createFullClassName(parent.getAbstractClassMetaData().getPackageName(), elementType);
        }

        if (embeddedElement != null)
        {
            if (embeddedElement.equalsIgnoreCase("true"))
            {
                this.embeddedElement = Boolean.TRUE;
            }
            else if (embeddedElement.equalsIgnoreCase("false"))
            {
                this.embeddedElement = Boolean.FALSE;
            }
        }

        if (dependentElement != null)
        {
            if (dependentElement.equalsIgnoreCase("true"))
            {
                this.dependentElement = Boolean.TRUE;
            }
            else if (dependentElement.equalsIgnoreCase("false"))
            {
                this.dependentElement = Boolean.FALSE;
            }
        }

        if (serializedElement != null)
        {
            if (serializedElement.equalsIgnoreCase("true"))
            {
                this.serializedElement = Boolean.TRUE;
            }
            else if (serializedElement.equalsIgnoreCase("false"))
            {
                this.serializedElement = Boolean.FALSE;
            }
        }
    }

    /**
     * Method to populate any defaults, and check the validity of the MetaData.
     * @param clr ClassLoaderResolver to use for any loading operations
     */
    public void populate(ClassLoaderResolver clr)
    {
        // Check the field type and see if it is castable to a Collection
        Class field_type=getAbstractPropertyMetaData().getType();
        if (!java.util.Collection.class.isAssignableFrom(field_type))
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Collection.IsNotCollectionError",
                    getFieldName(),getAbstractPropertyMetaData().getClassName(false));
        }

        // "element-type"
        if (elementType == null)
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Collection.ElementTypeNotSpecifiedError",
                    getFieldName(),getAbstractPropertyMetaData().getClassName(false));
        }

        // Check that the key type exists
        Class elementTypeClass = null;
        try
        {
            elementTypeClass = clr.classForName(elementType);
        }
        catch (ClassNotResolvedException cnre)
        {
            try
            {
                // Maybe the user specified a java.lang class without fully-qualifying it
                // This is beyond the scope of the JDO spec which expects java.lang cases to be fully-qualified
                elementTypeClass = clr.classForName(ClassUtils.getJavaLangClassForType(elementType));
            }
            catch (ClassNotResolvedException cnre2)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Collection.ElementTypeNotFoundError",
                    getFieldName(),getAbstractPropertyMetaData().getClassName(false),
                    elementType);
            }
        }

        if (!elementTypeClass.getName().equals(elementType))
        {
            // The element-type has been resolved from what was specified in the MetaData - update to the fully-qualified name
            JPOXLogger.METADATA.info(LOCALISER.msg("MetaData.Collection.ElementTypeResolved", getFieldName(), getAbstractPropertyMetaData().getClassName(false), 
                elementType, elementTypeClass.getName()));
            elementType = elementTypeClass.getName();
        }

		// "embedded-element"
        if (embeddedElement == null)
        {
            // Assign default for "embedded-element" based on 18.13.1 of JDO 2 spec
            // Note : this fails when using in the enhancer since not yet PC
            if (getMetaDataManager().getPMFContext().getTypeManager().isDefaultEmbeddedType(elementTypeClass))
            {
                embeddedElement = Boolean.TRUE;
            }
            else if (PersistenceCapable.class.isAssignableFrom(elementTypeClass) ||
                Object.class.isAssignableFrom(elementTypeClass) ||
                elementTypeClass.isInterface())
            {
                embeddedElement = Boolean.FALSE;
            }
            else
            {
                embeddedElement = Boolean.TRUE;
            }
        }
        if (embeddedElement == Boolean.FALSE)
        {
            // If the user has set a non-PC/non-Interface as not embedded, correct it since not supported.
            // Note : this fails when using in the enhancer since not yet PC
            if (!(PersistenceCapable.class.isAssignableFrom(elementTypeClass)) &&
                !(elementTypeClass.isInterface()) &&
                elementTypeClass != java.lang.Object.class)
            {
                embeddedElement = Boolean.TRUE;
            }
        }

        ElementMetaData elemmd = ((AbstractPropertyMetaData)parent).getElementMetaData();
        if (elemmd != null && elemmd.getEmbeddedMetaData() != null)
        {
            embeddedElement = Boolean.TRUE;
        }

        if (dependentElement == Boolean.TRUE)
        {
            // If the user has set a non-PC/non-reference as dependent, correct it since not valid.
            // Note : this fails when using in the enhancer since not yet PC
            if (!(PersistenceCapable.class.isAssignableFrom(elementTypeClass)) &&
                !(elementTypeClass.isInterface()) &&
                elementTypeClass != java.lang.Object.class)
            {
                dependentElement = Boolean.FALSE;
            }
        }

        // Keep a reference to the MetaData for the element
        elementClassMetaData = getAbstractPropertyMetaData().getAbstractClassMetaData().getMetaDataManager().getMetaDataForClassInternal(elementTypeClass, clr);

        setPopulated();
    }

    // ------------------------------ Accessors --------------------------------

    /**
     * Accessor for the element-type tag value.
     * This can contain comma-separated values.
     * @return element-type tag value
     */
    public String getElementType()
    {
        return elementType;
    }

    /**
     * Accessor for the Element ClassMetaData
     * @return element ClassMetaData
     */
    public AbstractClassMetaData getElementClassMetaData()
    {
        if (elementClassMetaData != null && !elementClassMetaData.isInitialised())
        {
            elementClassMetaData.initialise();
        }
        return elementClassMetaData;
    }

    /**
     * Accessor for the embedded-element tag value
     * @return embedded-element tag value
     */
    public boolean isEmbeddedElement()
    {
        if (embeddedElement == null)
        {
            return false;
        }
        else
        {
            return embeddedElement.booleanValue();
        }
    }

    /**
     * Accessor for The dependent-element attribute indicates that the
     * collection's element contains a reference that is to be deleted if the
     * referring instance is deleted.
     * 
     * @return dependent-element tag value
     */
    public boolean isDependentElement()
    {
        if (dependentElement == null)
        {
            return false;
        }
        else
        {
            return dependentElement.booleanValue();
        }
    }

    /**
     * Accessor for the serialized-element tag value
     * @return serialized-element tag value
     */
    public boolean isSerializedElement()
    {
        if (serializedElement == null)
        {
            return false;
        }
        else
        {
            return serializedElement.booleanValue();
        }
    }

    // ------------------------------- Utilities -------------------------------

    /**
     * Accessor for all ClassMetaData referenced by this array.
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of all ClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param clr the ClassLoaderResolver 
     **/
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr)
    { 
        AbstractClassMetaData element_cmd=getMetaDataManager().getMetaDataForClass(elementType,clr);
        if (element_cmd != null)
        {
            element_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
        }
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }
 
    /**
     * Returns a string representation of the object.
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<collection element-type=\"").append(elementType).append("\"");
        if (embeddedElement != null)
        {
            sb.append(" embedded-element=\"").append(embeddedElement).append("\"");
        }
        if (dependentElement != null)
        {
            sb.append(" dependent-element=\"").append(dependentElement).append("\"");
        }
        if (serializedElement != null)
        {
            sb.append(" serialized-element=\"").append(serializedElement).append("\"");
        }
        sb.append(">\n");

        // Add extensions
        sb.append(super.toString(prefix + indent,indent)); 
 
        sb.append(prefix).append("</collection>\n");
        return sb.toString();
    }
}