/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2004 Andy Jefferson - toString(), MetaData, javadocs
2004 Andy Jefferson - nullIndicatorColumn/Value, ownerField
    ...
**********************************************************************/
package org.jpox.metadata;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.jpox.ClassLoaderResolver;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;

/**
 * This element specifies the mapping for an embedded type. It contains multiple
 * field elements, one for each field in the type.
 * 
 * The null-indicator-column optionally identifies the name of the column used
 * to indicate whether the embedded instance is null. By default, if the value
 * of this column is null, then the embedded instance is null. This column might
 * be mapped to a field of the embedded instance but might be a synthetic column
 * for the sole purpose of indicating a null reference.
 * 
 * The null-indicator-value specifies the value to indicate that the embedded
 * instance is null. This is only used for non-nullable columns.
 * 
 * If null-indicator-column is omitted, then the embedded instance is assumed
 * always to exist.
 *
 * <H3>MetaData Element</H3>
 * The MetaData element represented here is as follows
 * <PRE>
 * &lt;!ELEMENT embedded (extension*, (field|property)*, extension*)&gt;
 * <!ATTLIST embedded owner-field CDATA>
 * <!ATTLIST embedded null-indicator-column CDATA #IMPLIED>
 * <!ATTLIST embedded null-indicator-value CDATA #IMPLIED>
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.27 $
 */
public class EmbeddedMetaData extends ExtendableMetaData
{
    /** Fields of the embedded object. */
    protected final List fields = new ArrayList();

    /** Name of the owner field in the embedded object. */
    protected final String ownerField;

    /** Name of a column used for determining if the embedded object is null */
    protected final String nullIndicatorColumn;

    /** Value in the null column indicating that the embedded object is null */
    protected final String nullIndicatorValue;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    protected AbstractPropertyMetaData fieldMetaData[];
    
    /**
     * Constructor to create a copy of the passed metadata using the provided parent.
     * @param parent The parent
     * @param embmd The metadata to copy
     */
    public EmbeddedMetaData(MetaData parent, EmbeddedMetaData embmd)
    {
        super(parent);
        this.ownerField = embmd.ownerField;
        this.nullIndicatorColumn = embmd.nullIndicatorColumn;
        this.nullIndicatorValue = embmd.nullIndicatorValue;
        for (int i=0;i<embmd.fields.size();i++)
        {
            if( embmd.fields.get(i) instanceof FieldMetaData )
            {
                addField(new FieldMetaData(this,(AbstractPropertyMetaData)embmd.fields.get(i)));
            }
            else
            {
                addField(new PropertyMetaData(this,(PropertyMetaData)embmd.fields.get(i)));
            }
        }
    }

    /**
     * Constructor.
     * @param parent Owning metadata
     * @param ownerField The owner field
     * @param nullColumn Column to use for determining null embedded object
     * @param nullValue Value of column to use for determining null embedded object
     */
    public EmbeddedMetaData(ExtendableMetaData parent,
                            String ownerField,
                            String nullColumn,
                            String nullValue)
    {
        super(parent);
        this.ownerField = ownerField;
        nullIndicatorColumn = nullColumn;
        nullIndicatorValue = nullValue;
    }

    /**
     * Method to populate the embedded MetaData.
     * This performs checks on the validity of the field types for embedding.
     * @param clr The class loader to use where necessary
     */
    public void populate(ClassLoaderResolver clr)
    {
        // Find the class that the embedded fields apply to
        MetaData md = getParent();
        AbstractPropertyMetaData apmd = null; // Field that has <embedded>
        AbstractClassMetaData embCmd = null; // The embedded class
        String embeddedType = null; // Name of the embedded type
        if (md instanceof AbstractPropertyMetaData)
        {
            // PC embedded in PC object
            apmd = (AbstractPropertyMetaData)md;
            embeddedType = apmd.getTypeName();
            embCmd = apmd.getAbstractClassMetaData().getMetaDataManager().getMetaDataForClassInternal(apmd.getType(), clr);
            if (embCmd == null)
            {
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Embedded.PC.UnsupportedType", apmd.getFullFieldName(), apmd.getTypeName()));
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Embedded.PC.UnsupportedType", apmd.getFullFieldName(), apmd.getTypeName());
            }
        }
        else if (md instanceof ElementMetaData)
        {
            // PC element embedded in collection
            ElementMetaData elemmd = (ElementMetaData)md;
            apmd = (AbstractPropertyMetaData)elemmd.getParent();
            embeddedType = apmd.getCollection().getElementType();
            try
            {
                Class cls = clr.classForName(embeddedType);
                embCmd = apmd.getAbstractClassMetaData().getMetaDataManager().getMetaDataForClassInternal(cls, clr);
            }
            catch (ClassNotResolvedException cnre)
            {
                // Should be caught by populating the Collection
            }
            if (embCmd == null)
            {
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Embedded.Collection.UnsupportedType", 
                    apmd.getFullFieldName(), apmd.getCollection().getElementType()));
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Embedded.Collection.UnsupportedType", apmd.getFullFieldName(), apmd.getTypeName());
            }
        }
        else if (md instanceof KeyMetaData)
        {
            // PC key embedded in Map
            KeyMetaData keymd = (KeyMetaData)md;
            apmd = (AbstractPropertyMetaData)keymd.getParent();
            embeddedType = apmd.getMap().getKeyType();
            try
            {
                Class cls = clr.classForName(embeddedType);
                embCmd = apmd.getAbstractClassMetaData().getMetaDataManager().getMetaDataForClassInternal(cls, clr);
            }
            catch (ClassNotResolvedException cnre)
            {
                // Should be caught by populating the Map
            }
            if (embCmd == null)
            {
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Embedded.MapKey.UnsupportedType", 
                    apmd.getFullFieldName(), apmd.getMap().getKeyType()));
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Embedded.MapKey.UnsupportedType", apmd.getFullFieldName(), embeddedType);
            }
        }
        else if (md instanceof ValueMetaData)
        {
            // PC value embedded in Map
            ValueMetaData valuemd = (ValueMetaData)md;
            apmd = (AbstractPropertyMetaData)valuemd.getParent();
            embeddedType = apmd.getMap().getValueType();
            try
            {
                Class cls = clr.classForName(embeddedType);
                embCmd = apmd.getAbstractClassMetaData().getMetaDataManager().getMetaDataForClassInternal(cls, clr);
            }
            catch (ClassNotResolvedException cnre)
            {
                // Should be caught by populating the Map
            }
            if (embCmd == null)
            {
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Embedded.MapValue.UnsupportedType", 
                    apmd.getFullFieldName(), apmd.getMap().getValueType()));
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Embedded.MapValue.UnsupportedType", apmd.getFullFieldName(), embeddedType);
            }
        }

        // Add fields for the class that aren't in the <embedded> block using Reflection.
        // NOTE 1 : We ignore fields in superclasses
        // NOTE 2 : We ignore "enhanced" fields (added by the JDO enhancer)
        // NOTE 3 : We ignore inner class fields (containing "$") 
        // NOTE 4 : We sort the fields into ascending alphabetical order
        Class embeddedClass = null;
        Collections.sort(fields);
        try
        {
            // Load the embedded class
            embeddedClass = clr.classForName(embeddedType);

            // Get all (reflected) fields in the populating class
            Field[] cls_fields=embeddedClass.getDeclaredFields();
            for (int i=0;i<cls_fields.length;i++)
            {
                // Limit to fields in this class, that aren't enhanced fields
                // that aren't inner class fields, and that aren't static
                if (cls_fields[i].getDeclaringClass().getName().equals(embeddedType) &&
                    !cls_fields[i].getName().startsWith("jdo") &&
                    !ClassUtils.isInnerClass(cls_fields[i].getName()) &&
                    !Modifier.isStatic(cls_fields[i].getModifiers()))
                {
                    // Find if there is a AbstractPropertyMetaData for this field.
                    // This is possible as AbstractPropertyMetaData implements Comparable
                    if (Collections.binarySearch(fields, cls_fields[i].getName()) < 0)
                    {
                        // Add a default AbstractPropertyMetaData for this field.
                        JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Embedded.AddingUnspecifiedField", 
                            cls_fields[i].getName(), embeddedType, apmd.getClassName(true)));
                        FieldMetaData omittedFmd = new FieldMetaData(this, cls_fields[i].getName());
                        fields.add(omittedFmd);
                        Collections.sort(fields);
                    }
                }
            }
        }
        catch (Exception e)
        {
            JPOXLogger.METADATA.error(e.getMessage(), e);
            throw new RuntimeException(e.getMessage());
        }
        // add properties of interface, only if interface
        if (embCmd instanceof InterfaceMetaData)
        {
            try
            {
                // Get all (reflected) fields in the populating class
                Method[] clsMethods=embeddedClass.getDeclaredMethods();
                for (int i=0;i<clsMethods.length;i++)
                {
                    // Limit to methods in this class, that aren't enhanced fields
                    // that aren't inner class fields, and that aren't static
                    if (clsMethods[i].getDeclaringClass().getName().equals(embeddedType) &&
                        ( clsMethods[i].getName().startsWith("get") || clsMethods[i].getName().startsWith("is") ) &&
                        !ClassUtils.isInnerClass(clsMethods[i].getName()))
                    {
                        String fieldName = ClassUtils.getFieldNameForJavaBeanGetter(clsMethods[i].getName() );
                        // Find if there is a PropertyMetaData for this field.
                        // This is possible as PropertyMetaData implements Comparable
                        if (Collections.binarySearch(fields,fieldName) < 0)
                        {
                            // Add a default AbstractPropertyMetaData for this field.
                            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Class.AddingUnspecifiedField",fieldName,apmd.getClassName(true)));
                            PropertyMetaData fmd=new PropertyMetaData(this, fieldName);
                            fields.add(fmd);
                            Collections.sort(fields);
                        }
                    }
                }
            }
            catch (Exception e)
            {
                JPOXLogger.METADATA.error(e.getMessage(), e);
                throw new RuntimeException(e.getMessage());
            }
        }
        Collections.sort(fields);

        Iterator fieldsIter = fields.iterator();
        while (fieldsIter.hasNext())
        {
            AbstractPropertyMetaData fieldFmd = (AbstractPropertyMetaData)fieldsIter.next();
            if (fieldFmd instanceof FieldMetaData)
            {
                Field cls_field = null;
                try
                {
                    cls_field = embeddedClass.getDeclaredField(fieldFmd.getName());
                }
                catch (Exception e)
                {
                    // MetaData field doesn't exist in the class!
                    throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Class.DeclaredFieldNotFoundError",
                        embeddedClass.getName(), fieldFmd.getFullFieldName());
                }
                fieldFmd.populate(clr, cls_field.getType(), cls_field.getModifiers());
            }
            else
            {
                Method cls_field=null;
                try
                {
                    try
                    {
                        cls_field = embeddedClass.getDeclaredMethod(ClassUtils.getJavaBeanGetterName(fieldFmd.getName(),true),null);
                    }
                    catch(Exception e)
                    {
                        cls_field = embeddedClass.getDeclaredMethod(ClassUtils.getJavaBeanGetterName(fieldFmd.getName(),false),null);
                    }
                }
                catch (Exception e)
                {
                    // MetaData field doesn't exist in the class!
                    throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Class.DeclaredFieldNotFoundError",
                        embeddedClass.getName(),fieldFmd.getFullFieldName());
                }
                fieldFmd.populate(clr, cls_field.getReturnType(), cls_field.getModifiers());
            }
        }
    }

    /**
     * Method to initialise the object, creating all internal convenience
     * arrays.
     */
    public void initialise()
    {
        fieldMetaData = new AbstractPropertyMetaData[fields.size()];
        for (int i=0; i<fieldMetaData.length; i++)
        {
            fieldMetaData[i] = (AbstractPropertyMetaData) fields.get(i);
            fieldMetaData[i].initialise();
        }

        setInitialised();
    }

    // ----------------------------- Accessors ---------------------------------

    /**
     * Accessor for fieldMetaData
     * @return Returns the fieldMetaData.
     */
    public final AbstractPropertyMetaData[] getFieldMetaData()
    {
        return fieldMetaData;
    }
    
    /**
     * Accessor for the owner field to contain the FK back to the owner
     * @return The owner field
     */
    public final String getOwnerField()
    {
        return ownerField;
    }

    /**
     * Accessor for the column to check for null values.
     * @return The column name
     */
    public final String getNullIndicatorColumn()
    {
        return nullIndicatorColumn;
    }

    /**
     * Accessor for the value of the null indicator column when null
     * @return The value
     */
    public final String getNullIndicatorValue()
    {
        return nullIndicatorValue;
    }

    // ----------------------------- Mutators ----------------------------------

    /**
     * Method to add a field to the embedded definition.
     * Rejects the addition of duplicate named fields.
     * @param fmd Meta-Data for the field.
     */
    public void addField(AbstractPropertyMetaData fmd)
    {
        if (fmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException(LOCALISER.msg("MetaData.Field.AlreadyInitialisedError", 
                fmd.getName(), fmd.getAbstractClassMetaData().getFullClassName()));
        }
        Iterator iter=fields.iterator();
        while (iter.hasNext())
        {
            AbstractPropertyMetaData md = (AbstractPropertyMetaData)iter.next();
            if (fmd.getName().equals(md.getName()))
            {
                throw new RuntimeException(LOCALISER.msg("MetaData.Field.DuplicateField", 
                    fmd.getName(), fmd.getAbstractClassMetaData().getFullClassName()));
            }
        }
        fields.add(fmd);
    }
    
    // ------------------------------- Utilities -------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        // Field needs outputting so generate metadata
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<embedded");
        if (ownerField != null)
        {
            sb.append(" owner-field=\"" + ownerField + "\"");
        }
        if (nullIndicatorColumn != null)
        {
            sb.append(" null-indicator-column=\"" + nullIndicatorColumn + "\"");
        }
        if (nullIndicatorValue != null)
        {
            sb.append(" null-indicator-value=\"" + nullIndicatorValue + "\"");
        }
        sb.append(">\n");

        // Add fields
        for (int i=0; i<fields.size(); i++)
        {
            AbstractPropertyMetaData f = (AbstractPropertyMetaData)fields.get(i);
            sb.append(f.toString(prefix + indent,indent));
        }
        
        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix + "</embedded>\n");
        return sb.toString();
    }
}