/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * Representation of the Meta-Data for an identity of a class.
 * <br>
 * The datastore-identity element declares the strategy for implementing
 * datastore identity for the class, including the mapping of the identity
 * columns of the relational table.
 * <br>
 * &lt;!ELEMENT datastore-identity ((extension)*, (column)*, (extension)*)?&gt;<br>
 * &lt;!ATTLIST datastore-identity column CDATA #IMPLIED&gt;<br>
 * &lt;!ATTLIST datastore-identity strategy CDATA #IMPLIED&gt;<br>
 * &lt;!ATTLIST datastore-identity sequence CDATA #IMPLIED&gt;<br>
 *
 * @since 1.1 
 * @version $Revision: 1.12 $
 */
public class IdentityMetaData extends ExtendableMetaData implements ColumnMetaDataContainer
{
    /**
     * column name value. 
     */
    protected String column;

    /**
     * strategy tag value.
     */
    protected IdentityStrategy strategy=null;

    /**
     * sequence tag value.
     */
    protected String sequence;

    /**
     * the columns
     */
    final List columns = new ArrayList();

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * Contains the metadata for column
     */
    protected ColumnMetaData[] columnMetaData;
 
    /**
     * Constructor 
     * @param parent parent AbstractClassMetaData instance
     * @param column column
     * @param strategy strategy
     * @param sequence sequence name
     */
    public IdentityMetaData(AbstractClassMetaData parent,
                            final String column,
                            final String strategy,
                            final String sequence)
    {
        super(parent);

        if (strategy != null &&
            IdentityStrategy.getIdentityStrategy(strategy) == null)
        {
            throw new RuntimeException(LOCALISER.msg("MetaData.Identity.StrategyNotSpecifiedError"));
        }

        this.column = column;
        this.sequence = sequence;
        if (sequence != null && strategy == null)
        {
            // JDO2 Section 18.6.1. No strategy, but sequence defined means use "sequence"
            this.strategy = IdentityStrategy.SEQUENCE;
        }
        else
        {
            this.strategy = IdentityStrategy.getIdentityStrategy(strategy);
        }
    }

    /**
     * Method to initialise all internal convenience arrays needed.
     */
    public void initialise()
    {
        // Cater for user specifying column name, or columns
        if (columns.size() == 0 && column != null)
        {
            columnMetaData = new ColumnMetaData[1];
            columnMetaData[0] = new ColumnMetaData(this, column);
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        setInitialised();
    }
 
    // -------------------------------- Mutators -------------------------------

    /**
     * Add a new ColumnMetaData element
     * @param colmd The ColumnMetaData to add
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
        colmd.parent = this;
        columnMetaData = new ColumnMetaData[columns.size()];
        for (int i=0; i<columnMetaData.length; i++)
        {
            columnMetaData[i] = (ColumnMetaData) columns.get(i);
        }
    }
 
    // -------------------------------- Accessors ------------------------------

    /**
     * Accessor for columnMetaData
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for the column
     * @return column
     */
    public String getColumn()
    {
        return column;
    }

    /**
     * Accessor for the strategy tag value
     * @return strategy tag value
     */
    public IdentityStrategy getStrategyValue()
    {
        return strategy;
    }

    /**
     * Accessor for the sequence name
     * @return sequence name
     */
    public String getSequence()
    {
        return sequence;
    }

    // ------------------------------ Utilities --------------------------------
 
    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        if (strategy != null)
        {
            sb.append(prefix).append("<datastore-identity strategy=\"" + strategy + "\"");
        }
        else
        {
            sb.append(prefix).append("<datastore-identity");
        }

        if (column != null)
        {
            sb.append("\n").append(prefix).append("        column=\"" + column + "\"");
        }
        if (sequence != null)
        {
            sb.append("\n").append(prefix).append("        sequence=\"" + sequence + "\"");
        }

        if ((columnMetaData != null && columnMetaData.length > 0) ||
            getNoOfExtensions() > 0)
        {
            sb.append(">\n");

            // Add columns
            if (columnMetaData != null)
            {
                for (int i=0; i<columnMetaData.length; i++)
                {
                    sb.append(columnMetaData[i].toString(prefix + indent,indent));
                }
            }
     
            // Add extensions
            sb.append(super.toString(prefix + indent,indent));

            sb.append(prefix).append("</datastore-identity>\n");
        }
        else
        {
            sb.append("/>\n");
        }

        return sb.toString();
    }
}