/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

import org.jpox.ClassLoaderResolver;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;

/**
 * The implements element declares a persistence-capable interface implemented by the 
 * persistence-capable class that contains this element. An extent of persistence-capable 
 * classes that implement this interface is managed by the JDO implementation. The extent 
 * can be used for queries or for iteration just like an extent of persistence-capable 
 * instances. The attribute name is required, and is the name of the interface. The java 
 * class naming rules apply: if the interface name is unqualified, the package is the name 
 * of the enclosing package.
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows
 * <PRE>
 * <!ELEMENT implements ((property)+, (extension)*)>
 * <!ATTLIST implements name CDATA #REQUIRED>
 * </PRE>
 *
 * @version $Revision: 1.8 $
 */
public class ImplementsMetaData extends ExtendableMetaData
{
    /** Name of the interface implemented. */
    protected String name;

    /** Properties implemented. */
    protected final List properties = new ArrayList();

    /**
     * Constructor.
     * @param parent Parent class
     * @param name Name of the interface being implemented
     */
    public ImplementsMetaData(ClassMetaData parent,
                              final String name)
    {
        super(parent);
        this.name = ClassUtils.createFullClassName(parent.getPackageName(), name);
    }

    /**
     * Accessor for name.
     * @return Returns the name.
     */
    public String getName()
    {
        return name;
    }

    /**
     * Method to add a property to this interface.
     * @param pmd Property
     */
    public void addProperty(PropertyMetaData pmd)
    {
        if (pmd == null)
        {
            return;
        }
        
        properties.add(pmd);
    }

    /**
     * Method to populate the details of the implements
     *
     * @param clr ClassLoaderResolver to use in loading any classes
     */
    public synchronized void populate(ClassLoaderResolver clr)
    {
        // Check the class that we're modelling exists
        try
        {
            clr.classForName(name).getName();
        }
        catch (ClassNotResolvedException cnre)
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Implements.ClassNotFoundError",((ClassMetaData)parent).getFullClassName(),name));
            throw new InvalidMetaDataException(LOCALISER, "MetaData.Implements.ClassNotFoundError",
                ((ClassMetaData)parent).getFullClassName(),name);
        }
        
        super.populate();        
    }
    
    // ------------------------------ Utilities --------------------------------
 
    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<implements name=\"" + name + "\">\n");

        // Add properties
        for (int i=0;i<properties.size();i++)
        {
            PropertyMetaData pmd = (PropertyMetaData)properties.get(i);
            sb.append(pmd.toString(prefix + indent, indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent, indent)); 

        sb.append(prefix + "</implements>\n");
        return sb.toString();
    }
}