/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFContext;
import org.jpox.util.JPOXLogger;

/**
 * Manager of JPA MetaData information in JPOX.
 * Manages the MetaData for a particular "persistence-unit".
 * 
 * @version $Revision: 1.8 $
 */
public class JPAMetaDataManager extends MetaDataManager
{
    /** Map of ClassMetaData, keyed by the JPA "entity name". */
    protected Map classMetaDataByEntityName = new HashMap();

    /** MetaData for the PersistenceUnit being managed. */
    protected PersistenceUnitMetaData unitMetaData = null;

    /** Flag for initialisation state. */
    protected boolean initialised = false;

    /**
     * Constructor.
     * @param pmfContext PMF Context that this metadata manager operates in
     */
    public JPAMetaDataManager(PMFContext pmfContext)
    {
        super(pmfContext);
    }

    /**
     * Initialisation method to define the "persistence-unit" being handled.
     * @param pumd The MetaData for this "persistence-unit"
     */
    public void initialise(PersistenceUnitMetaData pumd, ClassLoaderResolver clr)
    {
        if (initialised)
        {
            // TODO Throw exception
            return;
        }

        unitMetaData = pumd;

        HashSet exceptions = new HashSet();

        // Load annotations for all <class>, <jar-file> specifications
        HashSet classNames = pumd.getClassNames();
        if (classNames != null)
        {
            Iterator iter = classNames.iterator();
            while (iter.hasNext())
            {
                // Load MetaData via any annotations for the class
                String className = (String)iter.next();
                try
                {
                    Class cls = clr.classForName(className);
                    loadAnnotationsForClass(cls, clr, true);
                }
                catch (Exception e)
                {
                    exceptions.add(e);
                }
            }
        }
        // TODO Extract classes for jar files and load up annotations for classes

        // Load metadata for all <mapping-file> specifications
        HashSet mappingFiles = pumd.getMappingFiles();
        if (mappingFiles != null)
        {
            Iterator iter = mappingFiles.iterator();
            while (iter.hasNext())
            {
                String mappingFileName = (String)iter.next();
                try
                {
                    // TODO Use the ClassLoaderResolver ?
                    Enumeration files = Thread.currentThread().getContextClassLoader().getResources(mappingFileName);
                    while (files.hasMoreElements())
                    {
                        URL url = (URL)files.nextElement();
                        FileMetaData filemd = parseFile(url);
                        if (filemd != null)
                        {
                            registerFile(url.getFile(), filemd);

                            // Populate all classes in this file we've just parsed
                            populateFileMetaData(filemd, clr, null);
                        }
                    }
                }
                catch (InvalidMetaDataException imde)
                {
                    // Error in the metadata for this file
                    JPOXLogger.JPA.error(imde.getMessage());
                    exceptions.add(imde);
                }
                catch (IOException ioe)
                {
                    JPOXLogger.JPA.error("mapping-file " + mappingFileName + " could not be read");
                }
            }
        }

        if (exceptions.size() > 0)
        {
            // Errors were encountered so throw them all now
            throw new JDOUserException("Errors encountered whilst initialising MetaData for JPA. See the nested exceptions for details",
                (Throwable[])exceptions.toArray(new Throwable[exceptions.size()]));
        }
    }

    /**
     * Utility to parse a file, using the "jpa" MetaData handler.
     * @param fileURL URL of the file
     * @return The FileMetaData for this file
     */
    protected FileMetaData parseFile(URL fileURL)
    {
        if (metaDataParser == null)
        {
            metaDataParser = new MetaDataParser(this, validateMetaData);
        }
        return (FileMetaData)metaDataParser.parseMetaDataURL(fileURL, "jpa");
    }

    /**
     * Method to take the FileMetaData and register the relevant parts of it with the various
     * convenience collections/maps that we use for access.
     * @param fileURLString URL of the metadata file
     * @param filemd The File MetaData
     */
    protected void registerFile(String fileURLString, FileMetaData filemd)
    {
        if (fileURLString == null)
        {
            // Null file
            return;
        }
        if (fileMetaDataByURLString.get(fileURLString) != null)
        {
            // Already registered!
            return;
        }

        fileMetaDataByURLString.put(fileURLString, filemd);

        registerQueriesForFile(filemd);
        registerSequencesForFile(filemd);
        registerTableGeneratorsForFile(filemd);

        // Register the classes and interfaces for later use
        if (filemd.getType() == FileMetaData.JPA_FILE ||
            filemd.getType() == FileMetaData.ANNOTATIONS)
        {
            for (int i = 0; i < filemd.getNoOfPackages(); i++)
            {
                PackageMetaData pmd = filemd.getPackage(i);

                // Register all classes into the respective lookup maps
                for (int j = 0; j < pmd.getNoOfClasses(); j++)
                {
                    ClassMetaData cmd = pmd.getClass(j);
                    if (cmd.getEntityName() != null)
                    {
                        // Register the ClassMetaData for the implementation
                        classMetaDataByEntityName.put(cmd.getEntityName(), cmd);
                    }
                    classMetaDataByClass.put(cmd.getFullClassName(), cmd);
                }
            }
        }
    }

    /**
     * Convenience method to register all sequences found in the passed file.
     * @param filemd MetaData for the file
     */
    protected void registerSequencesForFile(FileMetaData filemd)
    {
        // Register all sequences for the packages in this file
        for (int i=0;i<filemd.getNoOfPackages();i++)
        {
            PackageMetaData pmd = filemd.getPackage(i);
            SequenceMetaData[] seqmds = pmd.getSequences();
            if (seqmds != null)
            {
                for (int j=0;j<seqmds.length;j++)
                {
                    // Register using its basic sequence name
                    sequenceMetaDataByPackageSequence.put(seqmds[j].getName(), seqmds[j]);
                }
            }
        }
    }

    /**
     * Convenience method to register all table generators found in the passed file.
     * @param filemd MetaData for the file
     */
    protected void registerTableGeneratorsForFile(FileMetaData filemd)
    {
        // Register all table generators for the packages in this file
        for (int i=0;i<filemd.getNoOfPackages();i++)
        {
            PackageMetaData pmd = filemd.getPackage(i);
            TableGeneratorMetaData[] tgmds = pmd.getTableGenerators();
            if (tgmds != null)
            {
                for (int j=0;j<tgmds.length;j++)
                {
                    // Register using its basic sequence name
                    tableGeneratorMetaDataByPackageSequence.put(tgmds[j].getName(), tgmds[j]);
                }
            }
        }
    }

    /**
     * Accessor for the JPA MetaData for a class.
     * With JPA we either register the classes in "persistence.xml" (via "class", "jar-file"
     * or "mapping-file") or we have them annotated. If they havent been loaded when we
     * loaded "persistence.xml" then we only check for annotations in that class.
     * @param c The class to find MetaData for
     * @return The ClassMetaData for this class (or null if not found)
     **/
    public synchronized AbstractClassMetaData getMetaDataForClassInternal(Class c, ClassLoaderResolver clr)
    {
        // Check for MetaData loaded when we loaded the "persistence-unit"
        AbstractClassMetaData acmd = (AbstractClassMetaData)classMetaDataByClass.get(c.getName());
        if (acmd != null)
        {
            return acmd;
        }

        // No MetaData so check for annotations
        FileMetaData annFilemd = loadAnnotationsForClass(c, clr, true);
        if (annFilemd != null)
        {
            // No MetaData but annotations present so use that
            return annFilemd.getPackage(0).getClass(0);
        }

        return null;
    }

    /**
     * Accessor for the MetaData for a class.
     * @param entityName The entity name to find MetaData for
     * @return The ClassMetaData for this entity name (or null if not found)
     */
    public synchronized AbstractClassMetaData getMetaDataForEntityName(String entityName)
    {
        return (AbstractClassMetaData) classMetaDataByEntityName.get(entityName);
    }

    /**
     * Convenience method to check if the specified class has metadata.
     * @param class_name The name of the class to check
     * @return Whether the metadata is already registered for this class
     */
    public boolean hasMetaDataForClass(String class_name)
    {
        if (class_name == null)
        {
            return false;
        }

        return (classMetaDataByClass.get(class_name) != null);
    }

    /**
     * Accessor for whether a class doesn't have MetaData.
     * @param className Name of the class
     * @return Whether it has no metadata.
     */
    protected boolean isClassWithoutMetaData(String className)
    {
        // Just return the same as hasMetaDataForClass sicne we load all at startup
        return !hasMetaDataForClass(className);
    }
}