/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2004 Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigMap
2004 Erik Bengtson - add dependent keys and values
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.List;
import java.util.Set;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * Representation of the Meta-Data for a Map.
 * 
 * <H3>MetaData Element</H3>
 * Represents the following MetaData Element
 * <PRE>
 * &lt;!ELEMENT map (extension)*&gt;
 * &lt;!ATTLIST map key-type CDATA #IMPLIED&gt;
 * &lt;!ATTLIST map embedded-key (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST map dependent-key (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST map serialized-key (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST map value-type CDATA #IMPLIED&gt;
 * &lt;!ATTLIST map embedded-value (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST map dependent-value (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST map serialized-value (true|false) #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1 
 * @version $Revision: 1.28 $
 */
public class MapMetaData extends ContainerMetaData
{
    /** embedded-key tag value. */
    protected Boolean embeddedKey;

    /** embedded-value tag value. */
    protected Boolean embeddedValue;

    /** serialized-key tag value. */
    protected Boolean serializedKey;

    /** serialized-value tag value. */
    protected Boolean serializedValue;

    /** key-type tag value. */
    protected String keyType = "java.lang.Object";

    /** value-type tag value. */
    protected String valueType = "java.lang.Object";

    /** Key-type ClassMetaData. */
    protected AbstractClassMetaData keyClassMetaData;

    /** Value-type ClassMetaData. */
    protected AbstractClassMetaData valueClassMetaData;

    /**
     * The dependent-value attribute indicates that the collection's value
     * contains references that are to be deleted if the referring instance is
     * deleted.
     */
    protected Boolean dependentValue; 

    /**
     * The dependent-key attribute indicates that the collection's key contains
     * references that are to be deleted if the referring instance is deleted.
     */
    protected Boolean dependentKey;

    /**
     * Constructor to create a copy of the passed metadata using the passed parent.
     * @param parent The parent
     * @param mapmd The metadata to copy
     */
    public MapMetaData(AbstractPropertyMetaData parent, MapMetaData mapmd)
    {
        super(parent);
        this.embeddedKey = mapmd.embeddedKey;
        this.embeddedValue = mapmd.embeddedValue;
        this.serializedKey = mapmd.serializedKey;
        this.serializedValue = mapmd.serializedValue;
        this.dependentKey = mapmd.dependentKey;
        this.dependentValue = mapmd.dependentValue;
        this.keyType = mapmd.keyType;
        this.valueType = mapmd.valueType;
        this.keyClassMetaData = mapmd.keyClassMetaData;
        this.valueClassMetaData = mapmd.valueClassMetaData;
    }

    /**
     * Constructor.
     * @param parent parent Field
     * @param keyType key-type tag value
     * @param embeddedKey embedded-key tag value
     * @param dependentKey dependent-key tag value
     * @param serializedKey serialized-key tag value
     * @param valueType value-type  tag value
     * @param embeddedValue embedded-value tag value
     * @param dependentValue dependent-value tag value
     * @param serializedValue serialized-value tag value
     */
    public MapMetaData(AbstractPropertyMetaData parent,
                       String keyType,
                       String embeddedKey,
                       String dependentKey,
                       String serializedKey,
                       String valueType,
                       String embeddedValue,
                       String dependentValue,
                       String serializedValue)
    {
        super(parent);

        if (!StringUtils.isWhitespace(keyType))
        {
            // Apply JDO naming rules - if no package, then assumed to be in same package as this class
            this.keyType = ClassUtils.createFullClassName(parent.getAbstractClassMetaData().getPackageName(), keyType);
        }
        if (!StringUtils.isWhitespace(valueType))
        {
            // Apply JDO naming rules - if no package, then assumed to be in same package as this class
            this.valueType = ClassUtils.createFullClassName(parent.getAbstractClassMetaData().getPackageName(), valueType);
        }

        if (embeddedKey != null)
        {
            if (embeddedKey.equalsIgnoreCase("true"))
            {
                this.embeddedKey = Boolean.TRUE;
            }
            else if (embeddedKey.equalsIgnoreCase("false"))
            {
                this.embeddedKey = Boolean.FALSE;
            }
        }
        if (embeddedValue != null)
        {
            if (embeddedValue.equalsIgnoreCase("true"))
            {
                this.embeddedValue = Boolean.TRUE;
            }
            else if (embeddedValue.equalsIgnoreCase("false"))
            {
                this.embeddedValue = Boolean.FALSE;
            }
        }
        // "dependent-value"
        if (dependentValue != null)
        {
            if (dependentValue.equalsIgnoreCase("true"))
            {
                this.dependentValue = Boolean.TRUE;
            }
            else if (dependentValue.equalsIgnoreCase("false"))
            {
                this.dependentValue = Boolean.FALSE;
            }
        }
        // "dependent-key"
        if (dependentKey != null)
        {
            if (dependentKey.equalsIgnoreCase("true"))
            {
                this.dependentKey = Boolean.TRUE;
            }
            else if (dependentKey.equalsIgnoreCase("false"))
            {
                this.dependentKey = Boolean.FALSE;
            }
        }

        if (serializedKey != null)
        {
            if (serializedKey.equalsIgnoreCase("true"))
            {
                this.serializedKey = Boolean.TRUE;
            }
            else if (serializedKey.equalsIgnoreCase("false"))
            {
                this.serializedKey = Boolean.FALSE;
            }
        }
        if (serializedValue != null)
        {
            if (serializedValue.equalsIgnoreCase("true"))
            {
                this.serializedValue = Boolean.TRUE;
            }
            else if (serializedValue.equalsIgnoreCase("false"))
            {
                this.serializedValue = Boolean.FALSE;
            }
        }
    }

    /**
     * Method to populate any defaults, and check the validity of the MetaData.
     * @param clr ClassLoaderResolver to use for loading any key/value types 
     */
    public void populate(ClassLoaderResolver clr)
    {
        // Check the field type and see if it is castable to a Map
        Class field_type=getAbstractPropertyMetaData().getType();
        if (!java.util.Map.class.isAssignableFrom(field_type))
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Map.IsNotMapError",
                    getFieldName(),getAbstractPropertyMetaData().getClassName(false));
        }

        // "key-type"
        if (keyType == null)
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Map.KeyTypeNotSpecifiedError",
                    getFieldName(),getAbstractPropertyMetaData().getClassName(false));
        }

        // Check that the key type exists
        Class keyTypeClass = null;
        try
        {
            keyTypeClass = clr.classForName(keyType);
        }
        catch (ClassNotResolvedException cnre)
        {
            try
            {
                // Maybe the user specified a java.lang class without fully-qualifying it
                // This is beyond the scope of the JDO spec which expects java.lang cases to be fully-qualified
                keyTypeClass = clr.classForName(ClassUtils.getJavaLangClassForType(keyType));
            }
            catch (ClassNotResolvedException cnre2)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Map.KeyTypeNotFoundError",
                    getFieldName(),getAbstractPropertyMetaData().getClassName(false),
                    keyType);
            }
        }

        if (!keyTypeClass.getName().equals(keyType))
        {
            // The value-type has been resolved from what was specified in the MetaData - update to the fully-qualified name
            JPOXLogger.METADATA.info(LOCALISER.msg("MetaData.Map.KeyTypeResolved", getFieldName(), getAbstractPropertyMetaData().getClassName(false), 
                keyType, keyTypeClass.getName()));
            keyType = keyTypeClass.getName();
        }

        // "embedded-key"
        if (embeddedKey == null)
        {
            // Assign default for "embedded-key" based on 18.13.2 of JDO 2 spec
            if (getMetaDataManager().getPMFContext().getTypeManager().isDefaultEmbeddedType(keyTypeClass))
            {
                embeddedKey = Boolean.TRUE;
            }
            else if (PersistenceCapable.class.isAssignableFrom(keyTypeClass) ||
                Object.class.isAssignableFrom(keyTypeClass) ||
                keyTypeClass.isInterface())
            {
                embeddedKey = Boolean.FALSE;
            }
            else
            {
                embeddedKey = Boolean.TRUE;
            }
        }
        if (embeddedKey == Boolean.FALSE)
        {
            // If the user has set a non-PC/non-Interface as not embedded, correct it since not supported.
            // Note : this fails when using in the enhancer since not yet PC
            if (!(PersistenceCapable.class.isAssignableFrom(keyTypeClass)) &&
                !(keyTypeClass.isInterface()) &&
                keyTypeClass != java.lang.Object.class)
            {
                embeddedKey = Boolean.TRUE;
            }
        }
        KeyMetaData keymd = ((AbstractPropertyMetaData)parent).getKeyMetaData();
        if (keymd != null && keymd.getEmbeddedMetaData() != null)
        {
            // If the user has specified <embedded>, set to true
            embeddedKey = Boolean.TRUE;
        }

        // "value-type"
        if (valueType == null)
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Map.ValueTypeNotSpecifiedError",
                    getFieldName(),getAbstractPropertyMetaData().getClassName(false));
        }

        // Check that the value-type exists
        Class valueTypeClass = null;
        try
        {
            valueTypeClass = clr.classForName(valueType);
        }
        catch (ClassNotResolvedException cnre)
        {
            try
            {
                // Maybe the user specified a java.lang class without fully-qualifying it
                // This is beyond the scope of the JDO spec which expects java.lang cases to be fully-qualified
                valueTypeClass = clr.classForName(ClassUtils.getJavaLangClassForType(valueType));
            }
            catch (ClassNotResolvedException cnre2)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Map.ValueTypeNotFoundError",
                    getFieldName(),getAbstractPropertyMetaData().getClassName(false),
                    valueType);
            }
        }

        if (!valueTypeClass.getName().equals(valueType))
        {
            // The value-type has been resolved from what was specified in the MetaData - update to the fully-qualified name
            JPOXLogger.METADATA.info(LOCALISER.msg("MetaData.Map.ValueTypeResolved", getFieldName(), getAbstractPropertyMetaData().getClassName(false),
                valueType, valueTypeClass.getName()));
            valueType = valueTypeClass.getName();
        }

        // "embedded-value"
        if (embeddedValue == null)
        {
            // Assign default for "embedded-value" based on 18.13.2 of JDO 2 spec
            if (getMetaDataManager().getPMFContext().getTypeManager().isDefaultEmbeddedType(valueTypeClass))
            {
                embeddedValue = Boolean.TRUE;
            }
            else if (PersistenceCapable.class.isAssignableFrom(valueTypeClass) ||
                Object.class.isAssignableFrom(valueTypeClass) ||
                valueTypeClass.isInterface())
            {
                embeddedValue = Boolean.FALSE;
            }
            else
            {
                embeddedValue = Boolean.TRUE;
            }
        }
        if (embeddedValue == Boolean.FALSE)
        {
            // If the user has set a non-PC/non-Interface as not embedded, correct it since not supported.
            // Note : this fails when using in the enhancer since not yet PC
            if (!(PersistenceCapable.class.isAssignableFrom(valueTypeClass)) &&
                !(valueTypeClass.isInterface()) &&
                valueTypeClass != java.lang.Object.class)
            {
                embeddedValue = Boolean.TRUE;
            }
        }
        ValueMetaData valuemd = ((AbstractPropertyMetaData)parent).getValueMetaData();
        if (valuemd != null && valuemd.getEmbeddedMetaData() != null)
        {
            // If the user has specified <embedded>, set to true
            embeddedValue = Boolean.TRUE;
        }

        keyClassMetaData = getAbstractPropertyMetaData().getAbstractClassMetaData().getMetaDataManager().getMetaDataForClassInternal(keyTypeClass, clr);
        valueClassMetaData = getAbstractPropertyMetaData().getAbstractClassMetaData().getMetaDataManager().getMetaDataForClassInternal(valueTypeClass, clr);

        setPopulated();
    }

    // ----------------------------- Accessors ---------------------------------

    /**
     * Accessor for the key-type tag value.
     * May be comma-separated if several key types are possible.
     * @return key-type tag value
     */
    public String getKeyType()
    {
        return keyType;
    }

    /**
     * Accessor for the Key ClassMetaData
     * @return key ClassMetaData
     */
    public AbstractClassMetaData getKeyClassMetaData()
    {
        if (keyClassMetaData != null && !keyClassMetaData.isInitialised())
        {
            keyClassMetaData.initialise();
        }
        return keyClassMetaData;
    }

    /**
     * Accessor for the value-type tag value.
     * May be comma-separated if several value types are possible.
     * @return value-type tag value
     */
    public String getValueType()
    {
        return valueType;
    }

    /**
     * Accessor for the Value ClassMetaData
     * @return value ClassMetaData
     */
    public AbstractClassMetaData getValueClassMetaData()
    {
        if (valueClassMetaData != null && !valueClassMetaData.isInitialised())
        {
            valueClassMetaData.initialise();
        }
        return valueClassMetaData;
    }

    /**
     * Accessor for the embedded-key tag value.
     * @return embedded-key tag value
     */
    public boolean isEmbeddedKey()
    {
        if (embeddedKey == null)
        {
            return false;
        }
        else
        {
            return embeddedKey.booleanValue();
        }
    }

    /**
     * Accessor for the embedded-value tag value.
     * @return embedded-value tag value
     */
    public boolean isEmbeddedValue()
    {
        if (embeddedValue == null)
        {
            return false;
        }
        else
        {
            return embeddedValue.booleanValue();
        }
    }

    /**
     * Accessor for the serialized-key tag value.
     * @return serialized-key tag value
     */
    public boolean isSerializedKey()
    {
        if (serializedKey == null)
        {
            return false;
        }
        else
        {
            return serializedKey.booleanValue();
        }
    }

    /**
     * Accessor for the serialized-value tag value.
     * @return serialized-value tag value
     */
    public boolean isSerializedValue()
    {
        if (serializedValue == null)
        {
            return false;
        }
        else
        {
            return serializedValue.booleanValue();
        }
    }

    /**
     * Accessor for the dependent-key attribute indicates that the map's
     * key contains references that are to be deleted if the referring instance
     * is deleted.
     * @return dependent-key tag value
     */
    public boolean isDependentKey()
    {
        if (dependentKey == null)
        {
            return false;
        }
        else
        {
            return dependentKey.booleanValue();
        }
    }

    /**
     * Accessor for the dependent-value attribute indicates that the
     * map's value contains references that are to be deleted if the
     * referring instance is deleted.
     * @return dependent-value tag value
     */
    public boolean isDependentValue()
    {
        if (dependentValue == null)
        {
            return false;
        }
        else
        {
            return dependentValue.booleanValue();
        }
    }

    // ----------------------------- Utilities ---------------------------------
 
    /**
     * Accessor for all ClassMetaData referenced by this array.
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of all ClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param clr the ClassLoaderResolver 
     **/
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr) 
    {
        AbstractClassMetaData key_cmd=getMetaDataManager().getMetaDataForClass(keyType,clr);
        if (key_cmd != null)
        {
            key_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
        }

        AbstractClassMetaData value_cmd=getMetaDataManager().getMetaDataForClass(valueType,clr);
        if (value_cmd != null)
        {
            value_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
        }
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("", "");
    }

    /**
     * Returns a string representation of the object.
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<map key-type=\"").append(keyType).append("\" value-type=\"").append(valueType).append("\"");
        if (embeddedKey != null)
        {
            sb.append(" embedded-key=\"").append(embeddedKey).append("\"");
        }
        if (embeddedValue != null)
        {
            sb.append(" embedded-value=\"").append(embeddedValue).append("\"");
        }
        if (dependentKey != null)
        {
            sb.append(" dependent-key=\"").append(dependentKey).append("\"");
        }
        if (dependentValue != null)
        {
            sb.append(" dependent-value=\"").append(dependentValue).append("\"");
        }
        if (serializedKey != null)
        {
            sb.append(" serialized-key=\"").append(serializedKey).append("\"");
        }
        if (serializedValue != null)
        {
            sb.append(" serialized-value=\"").append(serializedValue).append("\"");
        }
        sb.append(">\n");

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</map>\n");
        return sb.toString();
    }
}