/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2005 Andy Jefferson - addition of states
    ...
**********************************************************************/
package org.jpox.metadata;

import java.io.Serializable;

import org.jpox.util.Localiser;

/**
 * Base class for all MetaData. Provides state management for the MetaData.
 * The states represent the lifecycle of a MetaData object. The lifecycle goes
 * as follows :
 * <OL>
 * <LI>MetaData object is created (values passed in from a parsed file, or
 * manually generated)</LI>
 * <LI>MetaData object is populated (maybe pass in a class that it represents,
 * creating any additional information that wasn't in the initial data).</LI>
 * <LI>MetaData object is initialised (any internal arrays are set up, and
 * additions of data is blocked from this point).
 * <LI>MetaData object is added to with runtime information like actual column
 * names and types in use.</LI> 
 * </OL>
 *
 * @version $Revision: 1.7 $
 */
public class MetaData implements Serializable
{
    protected static Localiser LOCALISER=Localiser.getInstance("org.jpox.metadata.Localisation");

    /**
     * State representing the start state of MetaData, representing the initial
     * values passed in.
     */
    public static final int METADATA_CREATED_STATE=0;

    /**
     * State reflecting that MetaData object has been populated with the real
     * class definition, adding in extra fields that weren't in the startup.
     */
    public static final int METADATA_POPULATED_STATE=1;

    /**
     * State reflecting that MetaData object has been initialised, setting up
     * any internal information that is required.
     */
    public static final int METADATA_INITIALISED_STATE=2;

    /**
     * State reflecting that MetaData object has been modified with usage
     * information (e.g actual column names for the fields).
     */
    public static final int METADATA_USED_STATE=3;

    /**
     * State indicator for this MetaData object.
     */
    protected int metaDataState=METADATA_CREATED_STATE;

    /**
     * Parent MetaData object.
     */
    protected MetaData parent;

    /**
     * Constructor. Taking the parent MetaData object (if any).
     * @param parent The parent MetaData object.
     */
    public MetaData(MetaData parent)
    {
        this.parent = parent;
    }

    // -------------------------------- Mutators -------------------------------
 
    /**
     * Method to populate the object.
     * The state changes to "POPULATED" after this call.
     */
    public void populate()
    {
        setPopulated();
    }

    /**
     * Method to initialise the object.
     * The state changes to "INITIALISED" after this call.
     */
    public void initialise()
    {
        setInitialised();
    }

    /**
     * Utility to set the state as initialised.
     */
    public void setInitialised()
    {
        metaDataState = METADATA_INITIALISED_STATE;
    }

    /**
     * Utility to set the state as populated.
     */
    public void setPopulated()
    {
        metaDataState = METADATA_POPULATED_STATE;
    }

    /**
     * Utility to set the state as used.
     */
    public void setUsed()
    {
        metaDataState = METADATA_USED_STATE;
    }

    // -------------------------------- Accessors ------------------------------
 
    /**
     * Accessor for the parent MetaData object.
     * @return Parent MetaData object.
     */
    public MetaData getParent()
    {
        return parent;
    }

    /**
     * Accessor for whether the object state is "created".
     * @return Whether it is in "created" state
     */
    public boolean isCreated()
    {
        return (metaDataState == METADATA_CREATED_STATE);
    }

    /**
     * Accessor for whether the object state is "populated".
     * @return Whether it is populated.
     */
    public boolean isPopulated()
    {
        return (metaDataState == METADATA_POPULATED_STATE);
    }

    /**
     * Accessor for whether the object state is "initialised".
     * @return Whether it is initialised
     */
    public boolean isInitialised()
    {
        return (metaDataState == METADATA_INITIALISED_STATE);
    }

    /**
     * Accessor for whether the object state is "used".
     * @return Whether it is used.
     */
    public boolean isUsed()
    {
        return (metaDataState == METADATA_USED_STATE);
    }
}