/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.jpox.ClassLoaderResolver;
import org.jpox.ClassNameConstants;
import org.jpox.store.DatastoreField;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Utilities needed for the processing of MetaData.
 *
 * @version $Revision: 1.39.2.1 $
 */
public class MetaDataUtils
{
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.metadata.Localisation");

    private static MetaDataUtils instance;

    /**
     * Gets an instance of MetaDataUtils
     * @return a singleton instance of MetaDataUtils
     */
    public static synchronized MetaDataUtils getInstance()
    {
        if (instance == null)
        {
            instance = new MetaDataUtils();
        }
        return instance;
    }

    /**
     * Protected constructor to prevent outside instantiation
     */
    protected MetaDataUtils()
    {
    }

    /**
     * Utility to return a list of classes declared in provided metadata
     * files. The list of classes are ordered so that all superclasses occur
     * before their subclasses. Additionally populates the provided map
     * with entries mapping the class names to the JDO MetaData file name that
     * the class was found in. Used by SchemaTool.
     * @param metaDataFileURLs The list of JDO metadata file names.
     * @param classToMetaDataMap A Map that will be populated mapping class
     *   name to the MetaData file name. This argument can be null.
     * @return An ordered list of class name strings.
     * @throws Exception
     */
    public List getOrderedClassNamesFromFileList(URL[] metaDataFileURLs, Map classToMetaDataMap)
    throws Exception
    {
        HashMap superclassesByClass = new HashMap();
        final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        final DocumentBuilder db = factory.newDocumentBuilder();
        db.setEntityResolver(new JDOEntityResolver());

        for (int i=0; i<metaDataFileURLs.length; ++i)
        {
            Element docElement = db.parse(metaDataFileURLs[i].openStream()).getDocumentElement();

            NodeList nodes = docElement.getElementsByTagName("class");
            for (int j=0; j<nodes.getLength(); ++j)
            {
                Element clsElement = (Element)nodes.item(j);
                String packageName = ((Element)clsElement.getParentNode()).getAttribute("name");
                // Note : this attribute is now deprecated and so we cant rely on it being there
                String superclassName = clsElement.getAttribute("persistence-capable-superclass");
                if (superclassName.length() == 0)
                {
                    superclassName = null;
                }
                else
                {
                    superclassName = ClassUtils.createFullClassName(packageName, superclassName);
                }
                String className = clsElement.getAttribute("name");
                if (StringUtils.isWhitespace(packageName))
                {
                    JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Package.NameNotSpecified", metaDataFileURLs[i]));
                }
                else if (StringUtils.isWhitespace(className))
                {
                    JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.NameNotSpecified", packageName, metaDataFileURLs[i]));
                }
                else
                {
                    className = ClassUtils.createFullClassName(packageName, className);
                    superclassesByClass.put(className,superclassName);
                    if (classToMetaDataMap != null)
                    {
                        classToMetaDataMap.put(className,metaDataFileURLs[i]);
                    }
                }
            }
        }

        int numClasses = superclassesByClass.size();
        ArrayList orderedNames = new ArrayList(numClasses);
        ArrayList sortedNames = new ArrayList(numClasses);

        sortedNames.addAll(superclassesByClass.keySet());
        Collections.sort(sortedNames);
        Iterator i=sortedNames.iterator();
        while (i.hasNext())
        {
            addClassToOrderedList(orderedNames,(String)i.next(),superclassesByClass);
        }

        return orderedNames;
    }

    /**
     * Utility to add a class to a list of class names. Recursive to include
     * superclasses.
     * @param names List of classnames (updated by this operation)
     * @param className Class name to add
     * @param superclassesByClass Map of superclasses
     * @throws Exception
     */
    private static void addClassToOrderedList(List names, String className, Map superclassesByClass)
    throws Exception
    {
        String superclassName=(String)superclassesByClass.get(className);
        if (superclassName != null)
        {
            addClassToOrderedList(names,superclassName,superclassesByClass);
        }

        if (!names.contains(className))
        {
            names.add(className);
        }
    }

    /**
     * Convenience method to determine if an array is storable in a single column as a byte
     * array.
     * @param fmd The field
     * @return Whether this is an array that can be stored in a single column as non-serialised
     */
    public static boolean arrayStorableAsByteArrayInSingleColumn(AbstractPropertyMetaData fmd)
    {
        if (fmd == null || !fmd.hasArray())
        {
            return false;
        }

        String arrayComponentType = fmd.getType().getComponentType().getName();
        if (arrayComponentType.equals(ClassNameConstants.BOOLEAN) ||
            arrayComponentType.equals(ClassNameConstants.BYTE) ||
            arrayComponentType.equals(ClassNameConstants.CHAR) ||
            arrayComponentType.equals(ClassNameConstants.DOUBLE) ||
            arrayComponentType.equals(ClassNameConstants.FLOAT) ||
            arrayComponentType.equals(ClassNameConstants.INT) ||
            arrayComponentType.equals(ClassNameConstants.LONG) ||
            arrayComponentType.equals(ClassNameConstants.SHORT) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_LANG_BOOLEAN) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_LANG_BYTE) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_LANG_CHARACTER) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_LANG_DOUBLE) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_LANG_FLOAT) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_LANG_INTEGER) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_LANG_LONG) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_LANG_SHORT) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_MATH_BIGDECIMAL) ||
            arrayComponentType.equals(ClassNameConstants.JAVA_MATH_BIGINTEGER))
        {
            // These types can be stored as a single-column but setting the bytes only
            return true;
        }

        // All other arrays must be serialised into a single column
        return false;
    }

    /**
     * Convenience method that returns if a field stores a First-Class object (FCO).
     * @param fmd MetaData for the field
     * @param clr ClassLoaderResolver resolver
     * @return Whether it stores a FCO
     */
    public boolean storesFCO(AbstractPropertyMetaData fmd, ClassLoaderResolver clr)
    {
        if (fmd == null)
        {
            return false;
        }

        MetaDataManager mgr = fmd.getAbstractClassMetaData().getMetaDataManager();
        if (fmd.isSerialized() || fmd.isEmbedded())
        {
            // Serialised or embedded fields have no FCO
            return false;
        }
        else if (fmd.hasCollection() && !fmd.getCollection().isSerializedElement() && !fmd.getCollection().isEmbeddedElement())
        {
            if (fmd.getCollection().getElementClassMetaData() != null)
            {
                // Collection of PC elements
                return true;
            }
            else
            {
                String elementType = fmd.getCollection().getElementType();
                Class elementCls = clr.classForName(elementType);
                if (elementCls != null && mgr.getPMFContext().getTypeManager().isReferenceType(elementCls) &&
                    mgr.getMetaDataForImplementationOfReference(elementCls, null, clr) != null)
                {
                    // Collection of reference type for FCOs
                    return true;
                }
            }
        }
        else if (fmd.hasMap())
        {
            if (fmd.getMap().getKeyClassMetaData() != null &&
                !fmd.getMap().isEmbeddedKey() &&
                !fmd.getMap().isSerializedKey())
            {
                // Map of PC keys
                return true;
            }
            else
            {
                String keyType = fmd.getMap().getKeyType();
                Class keyCls = clr.classForName(keyType);
                if (keyCls != null && mgr.getPMFContext().getTypeManager().isReferenceType(keyCls) &&
                    mgr.getMetaDataForImplementationOfReference(keyCls, null, clr) != null)
                {
                    // Map with keys of reference type for FCOs
                    return true;
                }
            }
                
            if (fmd.getMap().getValueClassMetaData() != null &&
                !fmd.getMap().isEmbeddedValue() &&
                !fmd.getMap().isSerializedValue())
            {
                // Map of PC values
                return true;
            }
            else
            {
                String valueType = fmd.getMap().getValueType();
                Class valueCls = clr.classForName(valueType);
                if (valueCls != null && mgr.getPMFContext().getTypeManager().isReferenceType(valueCls) &&
                    mgr.getMetaDataForImplementationOfReference(valueCls, null, clr) != null)
                {
                    // Map with values of reference type for FCOs
                    return true;
                }
            }
        }
        else if (fmd.hasArray() && !fmd.getArray().isSerializedElement() && !fmd.getArray().isEmbeddedElement())
        {
            if (PersistenceCapable.class.isAssignableFrom(fmd.getType().getComponentType()))
            {
                // PersistenceCapable[]
                return true;
            }
        }
        else
        {
            // 1-1 relation with PC
            if (mgr.getPMFContext().getTypeManager().isReferenceType(fmd.getType()) &&
                mgr.getMetaDataForImplementationOfReference(fmd.getType(), null, clr) != null)
            {
                // Reference type for an FCO
                return true;
            }
            if (mgr.getMetaDataForClass(fmd.getType(), clr) != null)
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Convenience method that splits a comma-separated list of values into a String array (removing whitespace).
     * @param attr The attribute value
     * @return The string components
     */
    public String[] getValuesForCommaSeparatedAttribute(String attr)
    {
        if (attr == null || attr.length() == 0)
        {
            return null;
        }

        String[] values=StringUtils.split(attr, ",");

        // Remove any whitespace around the values
        if (values != null)
        {
            for (int i=0;i<values.length;i++)
            {
                values[i] = values[i].trim();
            }
        }
        return values;
    }

    /**
     * Convenience method to return the class names of the available implementation types for 
     * an interface/Object field, given its required role. Removes all duplicates from the list.
     * @param fmd MetaData for the field
     * @param fieldRole The role of the field
     * @param clr the ClassLoaderResolver
     * @return Names of the classes of the possible implementations of this interface/Object
     * @throws JDOUserException if no implementation types are found for the reference type field
     */
    public String[] getImplementationNamesForReferenceField(AbstractPropertyMetaData fmd, int fieldRole, ClassLoaderResolver clr)
    {
        // Check the JDO2 standard attribute for implementation types
        String[] implTypes = null;
        String implTypeStr = null;
        // TODO Support specification of collection/map implementation types using element-type,key-type,value-type
        /*if (DatastoreField.ROLE_COLLECTION_ELEMENT == role)
        {
            implTypeStr = fmd.getCollection().getElementType();
        }
        else if (DatastoreField.ROLE_MAP_KEY == role)
        {
            implTypeStr = fmd.getMap().getKeyImplementationType();
        }
        else if (DatastoreField.ROLE_MAP_VALUE == role)
        {
            implTypeStr = fmd.getMap().getValueImplementationType();
        }
        else */if (DatastoreField.ROLE_ARRAY_ELEMENT == fieldRole)
        {
            implTypeStr = fmd.getArray().getElementType();
        }
        else
        {
            implTypeStr = fmd.getFieldType();
        }

        if (implTypeStr != null)
        {
            implTypes = getValuesForCommaSeparatedAttribute(implTypeStr);
        }

        // JPOX extension that does the same thing as the JDO2 standard locations (when we adopt them fully)
        if (implTypes == null)
        {
            if (DatastoreField.ROLE_COLLECTION_ELEMENT == fieldRole)
            {
                implTypes = fmd.getValuesForExtension("implementation-classes");
            }
            else if (DatastoreField.ROLE_MAP_KEY == fieldRole)
            {
                implTypes = fmd.getValuesForExtension("key-implementation-classes");
            }
            else if (DatastoreField.ROLE_MAP_VALUE == fieldRole)
            {
                implTypes = fmd.getValuesForExtension("value-implementation-classes");
            }
            else if (DatastoreField.ROLE_ARRAY_ELEMENT == fieldRole)
            {
                implTypes = fmd.getValuesForExtension("implementation-classes");
            }
            else
            {
                implTypes = fmd.getValuesForExtension("implementation-classes");
            }
        }

        if (implTypes == null)
        {
            // Nothing specified, so check if it is an interface and if so use the <implements> definition to get some types
            String type = null;
            if (fmd.hasCollection() && fieldRole == DatastoreField.ROLE_COLLECTION_ELEMENT)
            {
                type = fmd.getCollection().getElementType();
            }
            else if (fmd.hasMap() && fieldRole == DatastoreField.ROLE_MAP_KEY)
            {
                type = fmd.getMap().getKeyType();
            }
            else if (fmd.hasMap() && fieldRole == DatastoreField.ROLE_MAP_VALUE)
            {
                type = fmd.getMap().getValueType();
            }
            else if (fmd.hasArray() && fieldRole == DatastoreField.ROLE_ARRAY_ELEMENT)
            {
                type = fmd.getType().getComponentType().getName();
            }
            else
            {
                type = fmd.getTypeName();
            }

            if (!type.equals(ClassNameConstants.OBJECT))
            {
                implTypes = fmd.getAbstractClassMetaData().getMetaDataManager().getClassesImplementingInterface(type,clr);
            }

            if (implTypes == null)
            {
                // Generate error since no implementations available
                throw new JDOUserException(LOCALISER.msg("MetaData.Reference.NoImplementationClassesFound", fmd.getFullFieldName(), type));
            }
        }

        // Remove all duplicates from the list but retain the original ordering
        int noOfDups = 0;
        for (int i=0;i<implTypes.length;i++)
        {
            for (int j=0;j<i;j++)
            {
                if (implTypes[j].equals(implTypes[i]))
                {
                    noOfDups++;
                    break;
                }
            }
        }
        String[] impls = new String[implTypes.length-noOfDups];
        int n = 0;
        for (int i=0;i<implTypes.length;i++)
        {
            boolean dup = false;
            for (int j=0;j<i;j++)
            {
                if (implTypes[j].equals(implTypes[i]))
                {
                    dup = true;
                    break;
                }
            }
            if (!dup)
            {
                impls[n++] = implTypes[i];
            }
        }

        return impls;
    }
}
