/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2004 Andy Jefferson - added toString(), MetaData docs, javadocs.
2004 Andy Jefferson - added unique, indexed
    ...
**********************************************************************/
package org.jpox.metadata;

import javax.jdo.JDOFatalUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.util.ClassUtils;

/**
 * This element specifies the mapping for the element component of arrays and
 * collections.
 * 
 * If only one column is mapped, and no additional information is needed for the
 * column, then the column attribute can be used. Otherwise, the column
 * element(s) are used.
 * 
 * The serialized attribute specifies that the key values are to be serialized
 * into the named column.
 * 
 * The foreign-key attribute specifies the name of a foreign key to be
 * generated.
 *
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows
 * <PRE> 
 * &lt;!ELEMENT value (extension*,embedded?,column*,foreign-key?,index?,unique?,extension*)&gt;
 * &lt;!ATTLIST value column CDATA #IMPLIED&gt;
 * &lt;!ATTLIST value delete-action (restrict|cascade|null|default|none) #IMPLIED&gt;
 * &lt;!ATTLIST value indexed (true|false|unique) #IMPLIED&gt;
 * &lt;!ATTLIST value unique (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST value mapped-by CDATA #IMPLIED&gt;
 * </PRE> 
 * 
 * @since 1.1
 * @version $Revision: 1.21 $
 */
public class ValueMetaData extends AbstractElementMetaData
{
    /**
     * Constructor to create a copy of the passed metadata using the provided parent.
     * @param parent The parent
     * @param vmd The metadata to copy
     */
    public ValueMetaData(MetaData parent, ValueMetaData vmd)
    {
        super(parent, vmd);
    }

    /**
     * Constructor.
     * @param parent MetaData of parent element
     * @param column The column name
     * @param deleteAction attribute delete-action value
     * @param updateAction attribute update-action value
     * @param indexed Whether to index this
     * @param unique Whether to add a unique constraint
     * @param mappedBy Field the value is mapped by in the key object
     */
    public ValueMetaData(MetaData parent,
                         String column,
                         String deleteAction,
                         String updateAction,
                         String indexed,
                         String unique,
                         String mappedBy)
    {
        super(parent, column, deleteAction, updateAction, indexed, unique, mappedBy);
    }

    /**
     * Populate the MetaData.
     * @param clr Class loader to use
     */
    public void populate(ClassLoaderResolver clr)
    {
        AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)parent;
        if( fmd.getMap() == null )
        {
            throw new JDOFatalUserException("The field "+fmd.getFullFieldName()+" is defined with <value>, however no <map> definition was found.");
        }
        String valueType = fmd.getMap().getValueType();
        Class valueTypeClass = null;
        try
        {
            valueTypeClass = clr.classForName(valueType);
        }
        catch (ClassNotResolvedException cnre)
        {
            try
            {
                // Maybe the user specified a java.lang class without fully-qualifying it
                // This is beyond the scope of the JDO spec which expects java.lang cases to be fully-qualified
                valueTypeClass = clr.classForName(ClassUtils.getJavaLangClassForType(valueType), false);
                fmd.getMap().valueType = valueTypeClass.getName(); // Update the MapMetaData since not yet populated
            }
            catch (ClassNotResolvedException cnre2)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Map.ValueTypeNotFoundError",
                    fmd.getFullFieldName(),fmd.getClassName(false),
                    valueType);
            }
        }
        if (embeddedMetaData != null &&
            (valueTypeClass.isInterface() || valueTypeClass.getName().equals("java.lang.Object")))
        {
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Map.KeyTypeReferenceCannotBeEmbedded",
                fmd.getFullFieldName(), valueTypeClass.getName());
        }

        // TODO This will not work currently since MapMetaData is populated *after* ValueMetaData and so the
        // valueClassMetaData is not yet populated. What we should do is provide a postPopulate() method here
        // that MapMetaData can call when it is populated
        if (embeddedMetaData == null && 
            ((AbstractPropertyMetaData)parent).hasMap() && 
            ((AbstractPropertyMetaData)parent).getMap().isEmbeddedValue() &&
            ((AbstractPropertyMetaData)parent).getJoinMetaData() != null &&
            ((AbstractPropertyMetaData)parent).getMap().getValueClassMetaData() != null)
        {
            // User has specified that the value is embedded in a join table but not how we embed it
            // so add a dummy definition
            embeddedMetaData = new EmbeddedMetaData(this, null, null, null);
        }

        super.populate(clr);
    }

    // ------------------------------- Utilities -------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        // Field needs outputting so generate metadata
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<value");
        if (mappedBy != null)
        {
            sb.append(" mapped-by=\"" + mappedBy + "\"");
        }
        if (column != null)
        {
            sb.append("\n");
            sb.append(prefix).append("       column=\"" + column + "\"");
        }
        sb.append(">\n");

        // Add columns
        for (int i=0;i<columns.size();i++)
        {
            ColumnMetaData colmd=(ColumnMetaData)columns.get(i);
            sb.append(colmd.toString(prefix + indent,indent));
        }

        // Add index metadata
        if (indexMetaData != null)
        {
            sb.append(indexMetaData.toString(prefix + indent,indent));
        }

        // Add unique metadata
        if (uniqueMetaData != null)
        {
            sb.append(uniqueMetaData.toString(prefix + indent,indent));
        }

        // Add embedded metadata
        if (embeddedMetaData != null)
        {
            sb.append(embeddedMetaData.toString(prefix + indent,indent));
        }

        // Add foreign-key metadata
        if (foreignKeyMetaData != null)
        {
            sb.append(foreignKeyMetaData.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</value>\n");
        return sb.toString();
    }
}