/**********************************************************************
Copyright (c) 2006 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.plugin;

import org.jpox.AbstractPersistenceManagerFactory;
import org.jpox.ClassLoaderResolver;
import org.jpox.PMFConfiguration;
import org.jpox.util.JPOXLogger;
import org.jpox.util.JavaUtils;
import org.jpox.util.Localiser;

/**
 * Factory for PluginRegistry. 
 * Creates an instance of PluginRegistry based on the available
 * PluginRegistry implementation in the JPOX Core classpath
 */
public class PluginRegistryFactory
{
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /**
     * Instantiates a PluginRegistry. If the JPOX Core is deployed as an
     * eclipse plugin, uses the Eclipse OSGI Registry to find other JPOX plug-ins
     * @param conf the PMFConfiguration
     * @param clr the ClassLoaderResolver
     * @return instance of the PluginRegistry
     */
    public static PluginRegistry newPluginRegistry(PMFConfiguration conf, ClassLoaderResolver clr)
    {
        PluginRegistry registry = null;
        if (conf.getPluginRegistryClassName() != null)
        {
            registry = newInstance(conf.getPluginRegistryClassName(), conf.getPluginRegistryClassName(), clr);
        }
        if (registry != null)
        {
            if (JPOXLogger.PLUGIN.isDebugEnabled())
            {
                JPOXLogger.PLUGIN.debug("Using PluginRegistry " + registry.getClass().getName());
            }
            return registry;
        }
        registry = newInstance("org.eclipse.core.runtime.RegistryFactory", "org.jpox.plugin.EclipsePluginRegistry", clr);
        if (registry != null)
        {
            if (JPOXLogger.PLUGIN.isDebugEnabled())
            {
                JPOXLogger.PLUGIN.debug("Using PluginRegistry " + registry.getClass().getName());
            }
            return registry;
        }

        if (JPOXLogger.PLUGIN.isDebugEnabled())
        {
            JPOXLogger.PLUGIN.debug("Using PluginRegistry " + NonManagedPluginRegistry.class.getName());
        }
        return new NonManagedPluginRegistry(clr);
    }

    /**
     * Instantiates a PluginRegistry. Only proceed if the testClass is found in
     * the classpath
     * @param testClass A test class
     * @param pluginRegistryClass The class that implements
     * {@link PluginRegistry}
     * @return instance of the PluginRegistry
     */
    private static PluginRegistry newInstance(String testClass, String pluginRegistryClass, ClassLoaderResolver clr)
    {
        try
        {
            if (clr.classForName(testClass, AbstractPersistenceManagerFactory.class.getClassLoader()) == null)
            {
                // Just treat all exceptions the same since this registry doesnt work either way
                if (JPOXLogger.PLUGIN.isDebugEnabled())
                {
                    JPOXLogger.PLUGIN.debug(LOCALISER.msg("Plugin.RegistryNotFound", pluginRegistryClass));
                }
            }
            return (PluginRegistry) clr.classForName(pluginRegistryClass,AbstractPersistenceManagerFactory.class.getClassLoader()).getConstructor(new Class[] { ClassLoaderResolver.class }).newInstance(new Object[] { clr });
        }
        catch (Exception e)
        {
            // Just treat all exceptions the same since this registry doesnt work either way
            if (JPOXLogger.PLUGIN.isDebugEnabled())
            {
                if (JavaUtils.isJRE1_4OrAbove())
                {
                    // Exception.getCause() JDK1.4+
                    JPOXLogger.PLUGIN.debug(LOCALISER.msg("Plugin.RegistryNotFoundWithReason", pluginRegistryClass,
                        e.getCause() != null ? e.getCause().getMessage() : e.getMessage()));
                }
                else
                {
                    JPOXLogger.PLUGIN.debug(LOCALISER.msg("Plugin.RegistryNotFoundWithReason", pluginRegistryClass,
                        e.getMessage()));
                }
            }
        }
        return null;
    }
}