/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2004 Andy Jefferson - addition of rollbackOnly
    ...
**********************************************************************/
package org.jpox.resource;

import java.sql.Connection;
import java.sql.SQLException;

import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.PersistenceManager;
import javax.jdo.datastore.JDOConnection;
import javax.resource.cci.LocalTransaction;
import javax.transaction.Synchronization;

import org.jpox.store.rdbms.RDBMSTransaction;

/**
 * JdoManagedTransaction.
 * Implementations of the methods simply hand off to the transaction of the PM.
 * TODO This is RDBMS-specific now. Why is it here ?
 *
 * @version $Revision: 1.10.2.1 $
 */
public class JdoTransactionHandle implements RDBMSTransaction, LocalTransaction
{
    private final PersistenceManagerImpl pm;

    /**
     * Constructor
     * @param pm the PersistenceManager
     */
    public JdoTransactionHandle(PersistenceManagerImpl pm) 
    {
        this.pm = pm;
    }

    // implementation of javax.jdo.Transaction interface

    /**
     * Begin the transaction.
     */
    public void begin()
    {
        getTx().begin();
        pm.mc.notifyTxBegin(pm);
    }

    /**
     * Commit the transaction.
     */
    public void commit()
    {
        getTx().commit();
        pm.mc.notifyTxCommit(pm);
    }

    /**
     * Perform rollback of the transaction
     */
    public void rollback()
    {
        getTx().rollback();
        pm.mc.notifyTxRollback(pm);
    }

    /**
     * Accessor for whether the transaction is active.
     * @return Whether the txn is active
     */
    public boolean isActive()
    {
        return getTx().isActive();
    }

    /**
     * Accessor for whether the transaction is committing.
     * @return Whether the txn is committing
     */
    public boolean isCommitting()
    {
        return getTx().isCommitting();
    }

    /**
     * Accessor for the "rollback only" flag.
     * @return The rollback only flag
     */
    public boolean getRollbackOnly()
    {
        return getTx().getRollbackOnly();
    }

    /**
     * Mutator for the "rollback only" flag.
     */
    public void setRollbackOnly()
    {
        getTx().setRollbackOnly();
    }

    /**
     * Mutator for non-transactional reads 
     *
     * @param ntr Whether to use nontransactional reads
     */
    public void setNontransactionalRead(boolean ntr)
    {
        getTx().setNontransactionalRead(ntr);
    }

    /**
     * Accessor for non-transactional reads 
     *
     * @return Whether to use nontransactional reads
     */
    public boolean getNontransactionalRead()
    {
        return getTx().getNontransactionalRead();
    }

    /**
     * Mutator for non-transactional writes 
     *
     * @param ntw Whether to use nontransactional writes
     */
    public void setNontransactionalWrite(boolean ntw)
    {
        getTx().setNontransactionalWrite(ntw);
    }

    /**
     *
     * @return Whether to use nontransactional writes
     */
    public boolean getNontransactionalWrite()
    {
        return getTx().getNontransactionalWrite();
    }

    /**
     * Set retainValues flag
     * @param retain_values <description>
     */
    public void setRetainValues(boolean retain_values)
    {
        getTx().setRetainValues(retain_values);
    }

    /**
     * Accessor for whether to retain values 
     *
     * @return Whether to retain values
     */
    public boolean getRetainValues()
    {
        return getTx().getRetainValues();
    }

    /**
     * Mutator for whether to restore values 
     *
     * @param restore_values Whether to restore values
     */
    public void setRestoreValues(boolean restore_values)
    {
        getTx().setRestoreValues(restore_values);
    }

    /**
     * Accessor for whether to restore values 
     *
     * @return Whether to restore values
     */
    public boolean getRestoreValues()
    {
        return getTx().getRestoreValues();
    }

    /**
     * set the optimistic transactions flag
     * @param optimistic <description>
     */
    public void setOptimistic(boolean optimistic)
    {
        getTx().setOptimistic(optimistic);
    }

    /**
     * Accessor for the optimistic transactions flag
     * @return <description>
     */
    public boolean getOptimistic()
    {
        return getTx().getOptimistic();
    }

    /**
     * Set the synchronization.
     * @param synch <description>
     */
    public void setSynchronization(Synchronization synch)
    {
        getTx().setSynchronization(synch);
    }

    /**
     * Accessor for the synchronization
     * @return <description>
     */
    public Synchronization getSynchronization()
    {
        return getTx().getSynchronization();
    }

    /**
     * Accessor for the persistence manager.
     * @return <description>
     */
    public PersistenceManager getPersistenceManager()
    {
        return pm;
    }

    private RDBMSTransaction getTx()
    {
        return (RDBMSTransaction)pm.mc.getJdoTransaction();
    }

    public Connection getConnection(boolean forWriting, boolean forQuery) throws SQLException
    {
        return getTx().getConnection(forWriting, forQuery);
    }

    public Connection getConnection(boolean forWriting, boolean forQuery, boolean useOpenedConnection) throws SQLException
    {
        return getTx().getConnection(forWriting, forQuery, useOpenedConnection);
    }
    
    public boolean isConnectionOpen()
    {
        return getTx().isConnectionOpen();
    }
    
    public void releaseConnection(Connection conn) throws SQLException
    {
        getTx().releaseConnection(conn);
    }

    public int getTransactionIsolation()
    {
        return getTx().getTransactionIsolation();
    }

    public void setTransactionIsolation(int isolationLevel)
    {
        getTx().setTransactionIsolation(isolationLevel);
    }

    public boolean useUpdateLockOnFetch()
    {
        return getTx().useUpdateLockOnFetch();
    }

    public void setUseUpdateLock(boolean lock)
    {
        getTx().setUseUpdateLock(lock);
    }

    /* (non-Javadoc)
     * @see org.jpox.Transaction#getJDOConnection()
     */
    public JDOConnection getJDOConnection()
    {
        // TODO Unsupported feature
        throw new JDOUnsupportedOptionException("TODO");
    }
}