/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Andy Jefferson - coding standards
2005 Marco Schulze - fixed: notifyTxBegin/Commit/Rollback was never called
    ...
**********************************************************************/
package org.jpox.resource;

import java.io.PrintWriter;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import javax.jdo.JDOException;
import javax.jdo.Transaction;
import javax.jdo.datastore.JDOConnection;
import javax.resource.ResourceException;
import javax.resource.spi.ConnectionEvent;
import javax.resource.spi.ConnectionEventListener;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.LocalTransaction;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionFactory;
import javax.resource.spi.ManagedConnectionMetaData;
import javax.resource.spi.security.PasswordCredential;
import javax.security.auth.Subject;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import org.jpox.AbstractPersistenceManager;
import org.jpox.LifecycleListenerForClass;
import org.jpox.PersistenceManager;
import org.jpox.store.rdbms.RDBMSManagedTransaction;
import org.jpox.util.JPOXLogger;

/**
 * Implementation of ManagedConnection persistence manager
 *
 * @version $Revision: 1.18 $
 */
public class ManagedConnectionImpl extends AbstractPersistenceManager 
    implements LocalTransaction, ManagedConnection, XAResource
{
    private final PasswordCredential credential;

    /**
     * If {@link #notifyClosed(PersistenceManagerImpl)} is called during a
     * transaction, the handles are copied here in order to be able to call
     * their {@link ConnectionEventListener}s during
     * {@link #notifyTxCommit(PersistenceManagerImpl)} and
     * {@link #notifyTxRollback(PersistenceManagerImpl)}. These methods will
     * remove the handles here. 
     */
    private final List closedHandles = new ArrayList();

    /**
     * Those instances of {@link PersistenceManagerImpl}, which have been
     * opened for this managed connection.
     */
    private final List handles = new ArrayList();

    private final Collection cels = new ArrayList();

    private PrintWriter logWriter;
    
    /**
     * Constructor
     * @param mcf the ManagedConnectionFactory
     * @param credential the PasswordCredential
     * @throws ResourceException
     */
    public ManagedConnectionImpl(ManagedConnectionFactory mcf, PasswordCredential credential)
    throws ResourceException
    {
        super((ManagedConnectionFactoryImpl)mcf, 
            credential == null ? null : credential.getUserName(),  
            credential == null ? null : new String(credential.getPassword()));
        this.credential = credential;
        tx = getStoreManager().getManagedTransaction(this, credential);
    }

    PasswordCredential getPasswordCredential()
    {
        return credential;
    }

    ManagedConnectionFactoryImpl getManagedConnectionFactory()
    {
        return (ManagedConnectionFactoryImpl)apmf;
    }

    // implementation of javax.resource.spi.ManagedConnection interface

    /**
     * Destroy method
     * @exception javax.resource.ResourceException <description>
     */
    public void destroy()
    throws ResourceException
    {
        // if the user has not closed it, we close it
        if( !handles.isEmpty() )
        {
            List handlesToClose = new ArrayList(handles);
            for( Iterator it = handlesToClose.iterator(); it.hasNext(); )
            {
                PersistenceManager pm = ((PersistenceManager)it.next());
                if( !pm.isClosed() )
                {
                    pm.close();
                }
            }
        }
        if (tx != null) 
        {
            try 
            {
                // TODO Abstract out this call into the managed transaction
                ((RDBMSManagedTransaction)tx).closeConnection();
            }
            catch (SQLException e)
            {
                throw new ResourceException("Problem destroying ManagedConnection: " + e);
            }
        }
        close();
    }

    /**
     * Cleanup method
     * @exception javax.resource.ResourceException <description>
     */
    public synchronized void cleanup()
    throws ResourceException
    {
        for (Iterator i = handles.iterator(); i.hasNext();)
        {
            ((PersistenceManagerImpl)i.next()).setManagedConnection(null);
        }
        handles.clear();
    }

    /**
     * Accessor for the connection
     * @param subject <description>
     * @param cri <description>
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     */
    public Object getConnection(Subject subject, ConnectionRequestInfo cri)
    throws ResourceException
    {
        PasswordCredential pc = getManagedConnectionFactory().getPasswordCredential(subject);
        if (credential != pc && credential != null && pc!=null && !credential.equals(pc))
        {
            throw new ResourceException("Wrong subject: "+subject+" MCF credentials: "+pc+" MC credentials: "+credential);
        }
        
        PersistenceManagerImpl o = new PersistenceManagerImpl(this);
        // copy the lifecycle listeners from the PMF into the new (or cached?) PM
        o.removeAllInstanceLifecycleListeners(); // make sure there's none, in case the PM was cached
        List lifecycleListeners = getManagedConnectionFactory().getLifecycleListenerSpecifications();
        if (lifecycleListeners != null)
        {
            Iterator listenerIter = lifecycleListeners.iterator();
            while (listenerIter.hasNext())
            {
                LifecycleListenerForClass listener = (LifecycleListenerForClass) listenerIter.next();
                o.addInstanceLifecycleListener(listener.getListener(), listener.getClasses());
            }
        }
        
        handles.add(0, o);
        return o;
    }

    /**
     * Accessor for the Log
     * @return The Log writer
     * @exception javax.resource.ResourceException <description>
     */
    public PrintWriter getLogWriter()
    throws ResourceException
    {
        return logWriter;
    }

    /**
     * Mutator for the Log
     * @param writer PrintWriter to use for Log
     * @exception javax.resource.ResourceException <description>
     */
    public void setLogWriter(PrintWriter writer)
    throws ResourceException
    {
        this.logWriter = writer;
    }

    /**
     * Mutator to add a connection listener
     * @param cel <description>
     */
    public void addConnectionEventListener(ConnectionEventListener cel)
    {
        synchronized (cels)
        {
            cels.add(cel);
        }
    }

    /**
     * Mutator to remove a connection listener
     * @param cel <description>
     */
    public void removeConnectionEventListener(ConnectionEventListener cel)
    {
        synchronized (cels)
        {
            cels.remove(cel);
        }
    }

    /**
     * Mutator to associate a connection
     * @param c <description>
     * @exception javax.resource.ResourceException <description>
     */
    public void associateConnection(Object c)
    throws ResourceException
    {
        if (!(c instanceof PersistenceManagerImpl)) 
        {
            throw new ResourceException("wrong Connection type!");
        }
        ((PersistenceManagerImpl)c).setManagedConnection(this);
        handles.add(0, c);
    }

    /**
     * Accessor for the local transaction
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     */
    public LocalTransaction getLocalTransaction()
    throws ResourceException
    {
        return this;
    }

    /**
     * Accessor for the JDO transaction
     * @return The JDO transaction
     **/
    public Transaction getJdoTransaction() 
    {
        return tx;
    }

    /**
     * Accessor for the connection MetaData
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     */
    public ManagedConnectionMetaData getMetaData()
    throws ResourceException
    {
        throw new ResourceException("Not Yet Implemented");
    }

    /**
     * Accessor for the XA resource
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     */
    public XAResource getXAResource()
    throws ResourceException
    {
        return this;
    }

    // implementation of javax.resource.spi.LocalTransaction interface

    /**
     * Method to start the Transaction
     * @exception javax.resource.ResourceException <description>
     */
    public void begin()
    throws ResourceException
    {
        if( tx == null )
        {
            JPOXLogger.JDO.error("Invalid state during begin invoke. Transaction is closed.");
            return;
        }
        try 
        {
            tx.begin();
        }
        catch (JDOException e)
        {
            throw new ResourceException("JDOException: " + e);
        }

        for (Iterator it = handles.iterator(); it.hasNext(); ) 
        {
            notifyTxBegin((PersistenceManagerImpl)it.next());
        }
    }

    /**
     * Method to commit the Transaction
     * @exception javax.resource.ResourceException <description>
     */
    public void commit()
    throws ResourceException
    {
        if( tx == null )
        {
            JPOXLogger.JDO.error("Invalid state during commit invoke. Transaction is closed.");
            return;
        }
        try 
        {
            tx.commit();
        }
        catch (JDOException e)
        {
            if (JPOXLogger.JDO.isInfoEnabled())
            {
                JPOXLogger.JDO.info("Exception during commit: ", e);
            }
            throw new ResourceException("JDOException: " + e);
        }

        List h = new ArrayList(handles);
        for (Iterator it = h.iterator(); it.hasNext(); )
        {
            notifyTxCommit((PersistenceManagerImpl)it.next());
        }

        List ch = new ArrayList(closedHandles);
        ch.removeAll(h);
        for (Iterator it = ch.iterator(); it.hasNext(); )
        {
            notifyTxCommit((PersistenceManagerImpl)it.next());
        }

        closedHandles.clear();
    }

    /**
     * Method to rollback the Transaction
     * @exception javax.resource.ResourceException <description>
     */
    public void rollback()
    throws ResourceException
    {
        if( tx == null )
        {
            JPOXLogger.JDO.error("Invalid state during rollback invoke. Transaction is closed.");
            return;
        }
        try 
        {
            tx.rollback();
        }
        catch (JDOException e)
        {
            if (JPOXLogger.JDO.isInfoEnabled())
            {
                JPOXLogger.JDO.info("Exception during rollback: ", e);
            }
            throw new ResourceException("JDOException: " + e);
        }

        List h = new ArrayList(handles);
        for (Iterator it = h.iterator(); it.hasNext(); )
        {
            notifyTxRollback((PersistenceManagerImpl)it.next());
        }

        List ch = new ArrayList(closedHandles);
        ch.removeAll(h);
        for (Iterator it = ch.iterator(); it.hasNext(); )
        {
            notifyTxRollback((PersistenceManagerImpl)it.next());
        }

        closedHandles.clear();
    }

    //ConnectionEvent management

    void notifyClosed(PersistenceManagerImpl handle)
    {
        disconnectLifecycleListener();
        disconnectQueryCache();

        //do not disconnect SM objects, since we may have to run reachability later during commit/rollback
        
        ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.CONNECTION_CLOSED, null);
        ce.setConnectionHandle(handle);
        Collection localCels = null;
        synchronized (cels)
        {
            localCels = new ArrayList(cels);
        }

        for (Iterator i = localCels.iterator(); i.hasNext(); )
        {
            ((ConnectionEventListener)i.next()).connectionClosed(ce);

        }
        if (tx != null && tx.isActive())
        {
            closedHandles.add(handle);
        }
    }

    public synchronized void preCommit()
    {
        super.preCommit();
        //flush before disconnect sm cache
        flush();
        disconnectSMCache();
    }
    
    void notifyTxBegin(PersistenceManagerImpl handle)
    {
        ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.LOCAL_TRANSACTION_STARTED, null);
        ce.setConnectionHandle(handle);
        Collection localCels = null;
        synchronized (cels)
        {
            localCels = new ArrayList(cels);
        }

        for (Iterator i = localCels.iterator(); i.hasNext(); )
        {
            ((ConnectionEventListener)i.next()).localTransactionStarted(ce);

        }
    }

    void notifyTxCommit(PersistenceManagerImpl handle)
    {
        ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.LOCAL_TRANSACTION_COMMITTED, null);
        ce.setConnectionHandle(handle);
        Collection localCels = null;
        synchronized (cels)
        {
            localCels = new ArrayList(cels);
        }
        for (Iterator i = localCels.iterator(); i.hasNext(); )
        {
            ((ConnectionEventListener)i.next()).localTransactionCommitted(ce);

        }
    }

    void notifyTxRollback(PersistenceManagerImpl handle)
    {
        ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.LOCAL_TRANSACTION_ROLLEDBACK, null);
        ce.setConnectionHandle(handle);
        Collection localCels = null;
        synchronized (cels)
        {
            localCels = new ArrayList(cels);
        }

        for (Iterator i = localCels.iterator(); i.hasNext(); )
        {
            ((ConnectionEventListener)i.next()).localTransactionRolledback(ce);

        }
    }

    /**
     * Accessor for the Persistence Manager handle
     * @return The PM handle
     **/
    public PersistenceManager getPMHandle()
    {
        if (handles.size() != 1) 
        {
            if (JPOXLogger.JDO.isInfoEnabled())
            {
                JPOXLogger.JDO.info(LOCALISER.msg("PM.HandleInvalid", "" + handles.size()));
            }
        }
        return (PersistenceManager)handles.iterator().next();
    }

    /**
     * Accessor for whether this PM and the passed on are the same.
     * @param pm The other PM
     * @return Whether they are the same
     **/
    public boolean isSamePM(javax.jdo.PersistenceManager pm)
    {
        return handles.contains(pm);
    }

    /* (non-Javadoc)
     * @see javax.jdo.PersistenceManager#getDataStoreConnection()
     */
    public JDOConnection getDataStoreConnection()
    {
        return tx.getJDOConnection();
    }

    //------ implementation of XAResource
    
    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#commit(javax.transaction.xa.Xid, boolean)
     */
    public void commit(Xid xid, boolean flags) throws XAException
    {
        try 
        {
            tx.commit();
        }
        catch (JDOException e)
        {
            if (JPOXLogger.JDO.isInfoEnabled())
            {
                JPOXLogger.JDO.info("Exception during commit: ", e);
            }
            throw new XAException("JDOException: " + e);
        }    
        
        List h = new ArrayList(handles);
        for (Iterator it = h.iterator(); it.hasNext(); )
        {
            notifyTxCommit((PersistenceManagerImpl)it.next());
        }

        List ch = new ArrayList(closedHandles);
        ch.removeAll(h);
        for (Iterator it = ch.iterator(); it.hasNext(); )
        {
            notifyTxCommit((PersistenceManagerImpl)it.next());
        }

        closedHandles.clear();
        
    }

    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#end(javax.transaction.xa.Xid, int)
     */
    public void end(Xid xid, int flags) throws XAException
    {
		if ( flags == XAResource.TMSUCCESS )
		{
			flush();
		}
    }

    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#forget(javax.transaction.xa.Xid)
     */
    public void forget(Xid xid) throws XAException
    {
    }

    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#getTransactionTimeout()
     */
    public int getTransactionTimeout() throws XAException
    {
        return 0;
    }

    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#isSameRM(javax.transaction.xa.XAResource)
     */
    public boolean isSameRM(XAResource xares) throws XAException
    {
        return false;
    }

    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#prepare(javax.transaction.xa.Xid)
     */
    public int prepare(Xid flags) throws XAException
    {
		return XAResource.XA_OK;
		
    }

    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#recover(int)
     */
    public Xid[] recover(int flags) throws XAException
    {
        return new Xid[0];
    }

    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#rollback(javax.transaction.xa.Xid)
     */
    public void rollback(Xid xid) throws XAException
    {
        try 
        {
            tx.rollback();
        }
        catch (JDOException e)
        {
            if (JPOXLogger.JDO.isInfoEnabled())
            {
                JPOXLogger.JDO.info("Exception during rollback: ", e);
            }
            throw new XAException("JDOException: " + e);
        }       
        
        List h = new ArrayList(handles);
        for (Iterator it = h.iterator(); it.hasNext(); )
        {
            notifyTxRollback((PersistenceManagerImpl)it.next());
        }

        List ch = new ArrayList(closedHandles);
        ch.removeAll(h);
        for (Iterator it = ch.iterator(); it.hasNext(); )
        {
            notifyTxRollback((PersistenceManagerImpl)it.next());
        }

        closedHandles.clear();
        
    }

    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#setTransactionTimeout(int)
     */
    public boolean setTransactionTimeout(int seconds) throws XAException
    {
        return false;
    }

    /* (non-Javadoc)
     * @see javax.transaction.xa.XAResource#start(javax.transaction.xa.Xid, int)
     */
    public void start(Xid xid, int flags) throws XAException
    {
        if( !tx.isActive() )
        {
            try 
            {
                tx.begin();
            }
            catch (JDOException e)
            {
                throw new XAException("JDOException: " + e);
            }  
            for (Iterator it = handles.iterator(); it.hasNext(); ) 
            {
                notifyTxBegin((PersistenceManagerImpl)it.next());
            }
            
        }
    }
}