/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Erik Bengtson - added getObjectbyAID
2004 Andy Jefferson - coding standards and javadocs
2005 Marco Schulze - implemented copying the lifecycle listeners in j2ee environment
2005 Marco Schulze - added delegate methods to add/remove a ConnectionEventListener
    ...
**********************************************************************/
package org.jpox.resource;

import java.io.PrintWriter;
import java.util.Collection;

import javax.jdo.Extent;
import javax.jdo.JDOException;
import javax.jdo.datastore.JDOConnection;
import javax.jdo.datastore.Sequence;
import javax.jdo.listener.InstanceLifecycleListener;
import javax.jdo.spi.PersistenceCapable;
import javax.jdo.FetchPlan;
import javax.jdo.JDOFatalUserException;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.Query;
import javax.jdo.Transaction;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionMetaData;
import javax.resource.cci.Interaction;
import javax.resource.cci.LocalTransaction;
import javax.resource.cci.ResultSetInfo;
import javax.resource.spi.ConnectionEventListener;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFContext;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.MetaDataManager;
import org.jpox.state.CallbackHandler;
import org.jpox.state.FetchPlanState;
import org.jpox.store.FieldValues;
import org.jpox.store.StoreManager;
import org.jpox.store.query.QueryResult;
import org.jpox.util.Localiser;

/**
 * Implementation of the PersistenceManager for use in JCA environments.
 * An application handle associated to the {@link org.jpox.resource.ManagedConnectionImpl}.
 * @version $Revision: 1.61.2.1 $
 */
public class PersistenceManagerImpl
    implements Connection, PersistenceManager
{
    /** Localisation utility for output messages */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    private final JdoTransactionHandle tx;

    ManagedConnectionImpl mc;

    //private PersistenceManagerFactory pmf;

    private boolean closed;

    /**
     * Method to dump to output an object for diagnostic purposes. 
     * @param obj The object to dump
     * @param out The PrintWriter to use 
     * @see org.jpox.PersistenceManager#dump(java.lang.Object, java.io.PrintWriter)
     */
    public void dump(Object obj, PrintWriter out)
    {
        mc.dump(obj,out);
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param fv Field values for the object
     * @param clazz Class name of the object
     * @param ignoreCache true if it must ignore the cache
     * @return The Object with this id
     */
    public Object getObjectById(Object id, FieldValues fv, Class clazz, boolean ignoreCache)
    {
        return mc.getObjectById(id, fv, clazz, ignoreCache);
    }

    /**
     * Accessor for the StateManager of an object given the object AID.
     * @param pcClass <description>
     * @param fv <description>
     * @param ignoreCache true if it must ignore the cache*
     * @param checkInheritance Whether look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     * @return Object with this AID
     */
    public synchronized Object getObjectByAID(Class pcClass, FieldValues fv,boolean ignoreCache, boolean checkInheritance)
    {
        return mc.getObjectByAID(pcClass,fv,ignoreCache,checkInheritance);
    }
 
    /**
     * Constructor.
     * @param mc The connection implementation.
     **/
    public PersistenceManagerImpl(final ManagedConnectionImpl mc) 
    {
        tx = new JdoTransactionHandle(this);
        setManagedConnection(mc);
    }

    /**
     * Mutator for the connection implementation.
     * @param mc The managed connection implementation
     **/
    void setManagedConnection(final ManagedConnectionImpl mc)
    {
        this.mc = mc;
    }

    // implementation of javax.resource.cci.Connection interface

    /**
     * Accessor for the local transaction.
     * @return The local transaction
     * @exception javax.resource.ResourceException Thrown when an error occurs
     **/
    public LocalTransaction getLocalTransaction()
    throws ResourceException
    {
        return tx;
    }

    /**
     * Accessor for Meta-Data.
     * @return The Connection Meta-Data.
     * @exception javax.resource.ResourceException <description>
     **/
    public ConnectionMetaData getMetaData() throws ResourceException
    {
        throw new ResourceException("Not Yet Implemented");
    }

    /**
     *
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     **/
    public Interaction createInteraction() throws ResourceException
    {
        throw new ResourceException("Not Yet Implemented");
    }

    /**
     *
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     **/
    public ResultSetInfo getResultSetInfo() throws ResourceException
    {
        throw new ResourceException("Not Yet Implemented");
    }

    // implementation of javax.jdo.PersistenceManager interface

    /**
     * Method to assert if this Persistence Manager is open. Throws a
     * JDOFatalUserException if it has been closed.
     */
    private void assertIsOpen()
    {
        if (isClosed())
        {
            throw new JDOFatalUserException(LOCALISER.msg("PM.IsClosed"));
        }
    }
    
    /**
     * Method to close the Persistence Manager.
     **/
    public void close()
    {
        assertIsOpen();
        
        if (mc != null) 
        {
            mc.flush();
            mc.notifyClosed(this);
        }
        closed = true;
    }

    /**
     * Accessor for whether the Persistence manager is closed.
     * @return Whether the Persistence Manager is closed.
     **/
    public boolean isClosed()
    {
        return closed;
    }

    /**
     * Mutator for whether to detach objects on close of the PM's
     * @param detach Whether to detach objects on close of the PM.
     */
    public void setDetachOnClose(boolean detach)
    {
        mc.setDetachOnClose(detach);
    }

    /**
     * Accessor for whether to detach objects on close of the PM's
     * @return Whether to detach objects on close of the PM's
     */
    public boolean getDetachOnClose()
    {
        return mc.getDetachOnClose();
    }

    /**
     * Mutator for whether to detach all objects on commit of the txn.
     * @param detach Whether to detach objects on commit of the txn.
     */
    public void setDetachAllOnCommit(boolean detach)
    {
        mc.setDetachAllOnCommit(detach);
    }

    /**
     * Accessor for whether to detach objects on commit of the txn.
     * @return Whether to detach objects on commit of the txn.
     */
    public boolean getDetachAllOnCommit()
    {
        return mc.getDetachAllOnCommit();
    }
    
    /**
     * Whether the datastore operations are delayed until commit.
     * In optimistic transactions this is automatically enabled.
     * @return true if datastore operations are delayed until commit
     */
    public boolean isDelayDatastoreOperationsEnabled()
    {
        return mc.isDelayDatastoreOperationsEnabled();
    }
    
    /**
     * Method to refresh an object 
     * @param o Object to refresh
     **/
    public void refresh(Object o)
    {
        checkStatus();
        mc.refresh(o);
    }

    /**
     * Method to retrieve an object 
     * @param o Object to retrieve
     * @param fgOnly Whether to include only fields in current fetch group
     **/
    public void retrieve(Object o, boolean fgOnly)
    {
        checkStatus();
        mc.retrieve(o, fgOnly);
    }

    /**
     * Method to retrieve an object 
     * @param o Object to retrieve
     **/
    public void retrieve(Object o)
    {
        checkStatus();
        mc.retrieve(o);
    }

    /**
     * Retrieve a Collection of Persistence-Capable objects 
     * @param pcs Collection of Persistence-Capable objects
     */
    public void retrieveAll(Collection pcs)
    {
        checkStatus();
        mc.retrieveAll(pcs);
    }

    /**
     * Retrieve an array of Persistence-Capable objects 
     * @param pcs Array of Persistence-Capable objects
     */
    public void retrieveAll(Object[] pcs)
    {
        checkStatus();
        mc.retrieveAll(pcs);
    }

    /**
     * Retrieve field values of instances from the store. This tells
     * the <code>PersistenceManager</code> that the application intends to use
     * the instances, and their field values should be retrieved. The fields
     * in the current fetch group must be retrieved, and the implementation
     * might retrieve more fields than the current fetch-group.
     * <P>The <code>PersistenceManager</code> might use policy information
     * about the class to retrieve associated instances.
     * @param pcs the instances
     * @param fgOnly whether to retrieve only the current fetch-group fields
     * @since JDO 1.0.1
     */
    public void retrieveAll (Object[] pcs, boolean fgOnly)
    {
        mc.retrieveAll(pcs, fgOnly);
    }

    /** Retrieve field values of instances from the store.  This tells
     * the <code>PersistenceManager</code> that the application intends to use
     * the instances, and their field values should be retrieved. The fields
     * in the current fetch group must be retrieved, and the implementation
     * might retrieve more fields than the current fetch-group.
     * <P>The <code>PersistenceManager</code> might use policy information
     * about the class to retrieve associated instances.
     * @param pcs the instances
     * @param fgOnly whether to retrieve only the current fetch group fields
     * @since JDO 1.0.1
     */
    public void retrieveAll (Collection pcs, boolean fgOnly)
    {
        mc.retrieveAll(pcs, fgOnly);
    }

    /**
     * Accessor for the current transaction 
     * @return The current transaction
     **/
    public Transaction currentTransaction()
    {
        return tx;
    }

    /**
     * Method to evict an object 
     * @param o Object to evict
     **/
    public void evict(Object o)
    {
        checkStatus();
        mc.evict(o);
    }

    /**
     * Method to evict an array of objects 
     * @param os Array of objects to evict
     **/
    public void evictAll(Object[] os)
    {
        checkStatus();
        mc.evictAll(os);
    }

    /**
     * Method to evict all of the specified objects from the PM.
     * @param os Collection of objects to evict
     **/
    public void evictAll(Collection os)
    {
        checkStatus();
        mc.evictAll(os);
    }

    /**
     * Method to evict all objects from the PM.
     **/
    public void evictAll()
    {
       checkStatus();
       mc.evictAll();
    }

    /**
     * Method to refresh an array of objects.
     * @param os Array of objects to refresh
     **/
    public void refreshAll(Object[] os)
    {
        checkStatus();
        mc.refreshAll(os);
    }

    /**
     * Method to refresh a collection of objects.
     * @param os Collection of objects to refresh
     */
    public void refreshAll(Collection os)
    {
        checkStatus();
        mc.refreshAll(os);
    }

    /**
     * Method to refresh objects that failed verification in a JDOException
     * @param exc Exception containing objects that failed verification
     */
    public void refreshAll(JDOException exc)
    {
        checkStatus();
        mc.refreshAll(exc);
    }

    /**
     * Method to refresh all objects in the Persistence Manager.
     */
    public void refreshAll()
    {
        checkStatus();
        mc.refreshAll();
    }

    /**
     * Accessor for a new query.
     * @return The query.
     */
    public Query newQuery()
    {
        checkStatus();
        return mc.newQuery();
    }

    /**
     * Accessor for a new query.
     * @param obj object
     * @return The new query
     */
    public Query newQuery(Object obj)
    {
        checkStatus();
        return mc.newQuery(obj);
    }

    /**
     * Accessor for a single-string query.
     * @param query The single-string query
     * @return The Query
     */
    public Query newQuery(String query)
    {
        checkStatus();
        return mc.newQuery(query);
    }

    /**
     * Accessor for a new query in the specified query language etc.
     * @param language The query language
     * @param query The query definition
     * @return The new query
     */
    public Query newQuery(String language, Object query)
    {
        checkStatus();
        return mc.newQuery(language, query);
    }

    /**
     * Accessor for a new query using this candidate class.
     * @param cls The candidate class to use.
     * @return The new query
     */
    public Query newQuery(Class cls)
    {
        checkStatus();
        return mc.newQuery(cls);
    }

    /**
     * Accessor for a new query based on the provided extent.
     * @param ext The extent to use
     * @return The new query
     */
    public Query newQuery(Extent ext)
    {
        checkStatus();
        return mc.newQuery(ext);
    }

    /**
     * Accessor for a new query.
     * @param cls The candidate class.
     * @param cln The collection to use.
     * @return The new query
     */
    public Query newQuery(Class cls, Collection cln)
    {
        checkStatus();
        return mc.newQuery(cls, cln);
    }

    /**
     * Accessor for a new query.
     * @param cls The candidate class
     * @param filter The filter to use
     * @return The new query
     */
    public Query newQuery(Class cls, String filter)
    {
        checkStatus();
        return mc.newQuery(cls, filter);
    }

    /**
     * Accessor for a new query.
     * @param cls The candidate class
     * @param cln The collection
     * @param filter The filter to use
     * @return The new query
     */
    public Query newQuery(Class cls, Collection cln, String filter)
    {
        checkStatus();
        return mc.newQuery(cls, cln, filter);
    }

    /**
     * Accessor for a new query.
     * @param cln The collection
     * @param filter The filter to use
     * @return The new query
     */
    public Query newQuery(Extent cln, String filter)
    {
        checkStatus();
        return mc.newQuery(cln, filter);
    }

    /**
     * Accessor for a new named query.
     * @param cls The candidate class
     * @param queryName The name of the query
     * @return The new query
     */
    public Query newNamedQuery(Class cls, String queryName)
    {
        checkStatus();
        return mc.newNamedQuery(cls, queryName);
    }

    /**
     * Accessor for the extent of a candidate class.
     * @param cls The candidate class
     * @param subclasses Whether to include subclasses
     * @return The extent for these classes
     */
    public Extent getExtent(Class cls, boolean subclasses)
    {
        checkStatus();
        return mc.getExtent(cls, subclasses);
    }

    /**
     * Accessor for the extent of a candidate class.
     * @param cls The candidate class
     * @return The extent for these classes
     * @since 1.1
     */
    public Extent getExtent(Class cls)
    {
        checkStatus();
        return mc.getExtent(cls);
    }
    
    /**
     * Acessor for the current FetchPlan
     * @return FetchPlan
     * @since 1.1
     */
    public FetchPlan getFetchPlan()
    {
        //TODO this should be enabled, but with detachAllOnCommit=true it raises exception 
        //checkStatus();
        return mc.getFetchPlan();
    }

    /**
     * Accessor for an object given its id.
     * @param id Id of the object.
     * @return The object with this id
     */
    public Object getObjectById(Object id)
    {
        return getObjectById(id, true);
    }

    /**
     * Accessor for an object given its id.
     * @param id Id of the object.
     * @param validate Whether to validate the object before returning it
     * @return The object with this id
     */
    public Object getObjectById(Object id, boolean validate)
    {
        checkStatus();
        return mc.getObjectById(id, validate);
    }

    public Collection getObjectsById (Collection oids)
    {
        checkStatus();
        return mc.getObjectsById(oids);
    }

    public Object[] getObjectsById (Object[] oids)
    {
        checkStatus();
        return mc.getObjectsById(oids);
    }

    public Collection getObjectsById (Collection oids, boolean validate)
    {
        checkStatus();
        return mc.getObjectsById(oids, validate);
    }

    public Object[] getObjectsById (Object[] oids, boolean validate)
    {
        checkStatus();
        return mc.getObjectsById(oids, validate);
    }

    public Object getObjectById (Class cls, Object key)
    {
        checkStatus();
        return mc.getObjectById(cls, key);
    }

    public Object newObjectIdInstance (Class pcClass, Object key)
    {
        checkStatus();
        return mc.newObjectIdInstance(pcClass, key);
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param validate Whether to validate the object state
     * @param checkInheritance Whether look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     * @param objectClassName Class name for the object with this id (if known, optional)
     * @return The Object
     */
    public Object getObjectById(Object id, boolean validate, boolean checkInheritance, String objectClassName)
    {
        checkStatus();
        return mc.getObjectById(id, validate, checkInheritance, objectClassName);
    }

    /**
     * Accessor for the id of an object.
     * @param pc The object
     * @return The objects id
     */
    public Object getObjectId(Object pc)
    {
        checkStatus();
        return mc.getObjectId(pc);
    }

    /**
     *
     * @param pc <description>
     * @return <description>
     */
    public Object getTransactionalObjectId(Object pc)
    {
        checkStatus();
        return mc.getTransactionalObjectId(pc);
    }

    /**
     *
     * @param clazz <description>
     * @param str <description>
     * @return <description>
     */
    public Object newObjectIdInstance(Class clazz, String str)
    {
        checkStatus();
        return mc.newObjectIdInstance(clazz, str);
    }

    /**
     * Method to create an instance of an interface or abstract class
     * @param persistenceCapable interface/abstract class declared in metadata
     * @return Instance of the interface / abstract class
     */
    public Object newInstance(Class persistenceCapable)
    {
        checkStatus();
        return mc.newInstance(persistenceCapable);
    }

    /**
     * Method to make an object persistent during the persisting process.
     * <B>For internal use only</B>
     * @param obj The object to persist
     * @param preInsertChanges Changes to make before inserting
     * @return The persisted object
     */
    public Object makePersistentInternal(Object obj, FieldValues preInsertChanges)
    {
        checkStatus();
        return mc.makePersistentInternal(obj, preInsertChanges);
    }

    /**
     * Method to make an object persistent.
     * @param pc The object to persist
     */
    public Object makePersistent(Object pc)
    {
        checkStatus();
        return mc.makePersistent(pc);
    }

    /**
     * Make an array of Persistent Capable objects persistent 
     * @param pcs Array of Persistent Capable objects
     */
    public Object[] makePersistentAll(Object[] pcs)
    {
        checkStatus();
        return mc.makePersistentAll(pcs);
    }

    /**
     * Make a Collection of Persistence Capable objects persistent 
     * @param pcs Collection of Persistence Capable objects
     */
    public Collection makePersistentAll(Collection pcs)
    {
        checkStatus();
        return mc.makePersistentAll(pcs);
    }

    /**
     * JDO method to delete a Persistence Capable object 
     * @param pc Persistence Capable object
     */
    public void deletePersistent(Object pc)
    {
        checkStatus();
        mc.deletePersistent(pc);
    }

    /**
     * JDO method to delete an array of Persistence Capable objects 
     * @param pcs Array of Persistence Capable objects
     */
    public void deletePersistentAll(Object[] pcs)
    {
        checkStatus();
        mc.deletePersistentAll(pcs);
    }

    /**
     * JDO method to delete a Collection of Persistence Capable objects 
     * @param pcs Collection of Persistence Capable objects
     */
    public void deletePersistentAll(Collection pcs)
    {
        checkStatus();
        mc.deletePersistentAll(pcs);
    }

    /**
     * Make a Persistence-Capable object transient , optionally using the fetch plan.
     * @param pc Persistence-Capable object
     * @param useFetchPlan Whether to use the fetch plan
     */
    public void makeTransient(Object pc, boolean useFetchPlan)
    {
        checkStatus();
        mc.makeTransient(pc, useFetchPlan);
    }

    /**
     * Make an array of Persistence-Capable objects transient 
     * @param pcs Array of Persistence-Capable objects
     * @param useFetchPlan Whether to use the fetch plan
     */
    public void makeTransientAll(Object[] pcs, boolean useFetchPlan)
    {
        checkStatus();
        mc.makeTransientAll(pcs, useFetchPlan);
    }

    /**
     * Make a Collection of Persistence-Capable objects transient 
     * @param pcs Collection of Persistence-Capable objects
     * @param useFetchPlan Whether to use the fetch plan
     */
    public void makeTransientAll(Collection pcs, boolean useFetchPlan)
    {
        checkStatus();
        mc.makeTransientAll(pcs, useFetchPlan);
    }

    /**
     * Make a Persistence-Capable object transient 
     * @param pc Persistence-Capable object
     */
    public void makeTransient(Object pc)
    {
        checkStatus();
        mc.makeTransient(pc);
    }

    /**
     * Make an array of Persistence-Capable objects transient 
     * @param pcs Array of Persistence-Capable objects
     */
    public void makeTransientAll(Object[] pcs)
    {
        checkStatus();
        mc.makeTransientAll(pcs);
    }

    /**
     * Make a Collection of Persistence-Capable objects transient 
     * @param pcs Collection of Persistence-Capable objects
     */
    public void makeTransientAll(Collection pcs)
    {
        checkStatus();
        mc.makeTransientAll(pcs);
    }

    /**
     * Method to make an object transient.
     * <B>For internal use only</B>
     * @param obj The object to persist
     * @param state Object containing the state of the fetchplan process
     * @return The persisted object
     */
    public void internalMakeTransient(Object obj, FetchPlanState state)
    {
        checkStatus();
        mc.internalMakeTransient(obj, state);
    }

    /**
     * Make a Persistence-Capable object transient 
     * @param pc Persistence-Capable object
     */
    public void makeTransactional(Object pc)
    {
        checkStatus();
        mc.makeTransactional(pc);
    }

    /**
     * Make an array of Persistence-Capable objects transactional 
     * @param pcs Array of Persistence-Capable objects
     */
    public void makeTransactionalAll(Object[] pcs)
    {
        checkStatus();
        mc.makeTransactionalAll(pcs);
    }

    /**
     * Make a collection of PersistenceCapable objects transactional
     * @param pcs Collection of Persistence-Capable objects
     */
    public void makeTransactionalAll(Collection pcs)
    {
        checkStatus();
        mc.makeTransactionalAll(pcs);
    }

    /**
     * Make a Persistence-Capable object non-transactional 
     * @param pc Persistence-Capable object
     */
    public void makeNontransactional(Object pc)
    {
        checkStatus();
        mc.makeNontransactional(pc);
    }

    /**
     * Make an array of Persistence-Capable objects non-transactional 
     * @param pcs Array of Persistence-Capable objects
     */
    public void makeNontransactionalAll(Object[] pcs)
    {
        checkStatus();
        mc.makeNontransactionalAll(pcs);
    }

    /**
     * Make a Collection of Persistence-Capable objects non-transactional 
     * @param pcs Collection of Persistence-Capable objects
     */
    public void makeNontransactionalAll(Collection pcs)
    {
        checkStatus();
        mc.makeNontransactionalAll(pcs);
    }

    /**
     * Detach the specified object from the <code>PersistenceManager</code>.
     * @param pc the instance to detach
     * @param state State for the detachment process
     * @since JDO 2.0
     */
    public synchronized void detachInternal(Object pc, FetchPlanState state)
    {
        checkStatus();
        mc.detachInternal(pc, state);
    }

    /**
     * Detach the specified object from the <code>PersistenceManager</code>.
     * @param pc the instance to detach
     * @return the detached instance
     * @see #detachCopyAll(Object[])
     * @since JDO 2.0
     */
    public synchronized Object detachCopy(Object pc)
    {
        checkStatus();
        return mc.detachCopy(pc);
    }

    /**
     * Method to detach a copy of the passed object using the detached state.
     * @param pc The Object
     * @param state Detach state to use
     * @return The detached object copy of the original
     */
    public Object detachCopyInternal(Object pc, FetchPlanState state)
    {
        checkStatus();
        return mc.detachCopyInternal(pc, state);
    }

    /**
     * Detach the specified objects from the
     * <code>PersistenceManager</code>. The objects returned can be
     * manipulated and re-attached with 
     * {@link #makePersistentAll(Object[])}. 
     * The detached instances will be
     * unmanaged copies of the specified parameters, and are suitable
     * for serialization and manipulation outside of a JDO
     * environment. When detaching instances, only fields in the
     * current {@link FetchPlan} will be traversed. Thus, to detach a
     * graph of objects, relations to other persistent instances must
     * either be in the <code>default-fetch-group</code>, or in the
     * current custom {@link FetchPlan}.
     * @param pcs the instances to detach
     * @return the detached instances
     * @throws javax.jdo.JDOUserException if any of the instances do not
     * @see #makePersistentAll(Object[])
     * @see #getFetchPlan
     * @since JDO 2.0
     */ 
    public synchronized Object[] detachCopyAll(Object[] pcs)
    {
        checkStatus();
        return mc.detachCopyAll(pcs);
    }

    /**
     * Detach the specified objects from the <code>PersistenceManager</code>.
     * @param pcs the instances to detach
     * @return the detached instances
     * @see #detachCopyAll(Object[])
     * @since JDO 2.0
     */ 
    public synchronized Collection detachCopyAll(Collection pcs)
    {
        checkStatus();
        return mc.detachCopyAll(pcs);
    }

    /**
     * Method to attach a persistent object.
     * @param pc The object
     * @param embedded Whether the object is embedded
     * @return The attached object
     **/ 
    public synchronized Object attachCopy(Object pc, boolean embedded)
    {
        checkStatus();
        return mc.attachCopy(pc, embedded);
    }

    /**
     * Method to put a user object into the PersistenceManager. This is so that
     * multiple users can each have a user object for example. <I>The parameter
     * is not inspected or used in any way by the JDO implementation. </I>
     * @param key The key to store the user object under
     * @param value The object to store
     * @return The previous value for this key
     * @since 1.1
     */
    public synchronized Object putUserObject(Object key, Object value)
    {
        checkStatus();
        return mc.putUserObject(key, value);
    }

    /**
     * Method to get a user object from the PersistenceManager. This is for user
     * objects which are stored under a key. <I>The parameter is not inspected
     * or used in any way by the JDO implementation. </I>
     * @param key The key to store the user object under
     * @return The user object for that key
     * @since 1.1
     */
    public synchronized Object getUserObject(Object key)
    {
        checkStatus();
        return mc.getUserObject(key);
    }

    /**
     * Method to remove a user object from the PersistenceManager. This is for
     * user objects which are stored under a key. <I>The parameter is not
     * inspected or used in any way by the JDO implementation. </I>
     * @param key The key whose uder object is to be removed.
     * @return The user object that was removed
     * @since 1.1
     */
    public synchronized Object removeUserObject(Object key)
    {
        checkStatus();
        return mc.removeUserObject(key);
    }

    /**
     * The application might manage PersistenceManager instances by using an
     * associated object for bookkeeping purposes. These methods allow the user
     * to manage the associated object.
     * The parameter is not inspected or used in any way by the JDO
     * implementation.
     *
     * @param obj User Object
     */
    public void setUserObject(Object obj)
    {
        checkStatus();
        mc.setUserObject(obj);
    }

    /**
     * The application might manage PersistenceManager instances by using an
     * associated object for bookkeeping purposes. These methods allow the user
     * to manage the associated object.
     * The parameter is not inspected or used in any way by the JDO
     * implementation.*
     * 
     * @return User object
     */
    public Object getUserObject()
    {
        checkStatus();
        return mc.getUserObject();
    }

    /**
     * Retrieve the PersistenceManagerFactory for this manager 
     * @return The PersistenceManagerFactory
     */
    public PersistenceManagerFactory getPersistenceManagerFactory()
    {
        return this.mc.getPersistenceManagerFactory();
    }

    /**
     * Retrieve the class for the objectid 
     * @param clazz The class to retrieve
     * @return The Class of the ObjectId
     */
    public Class getObjectIdClass(Class clazz)
    {
        checkStatus();
        return mc.getObjectIdClass(clazz);
    }

    /**
     * Mutator for the multithreaded capability of the manager 
     * @param multithreaded Whether to run multithreaded or not
     */
    public void setMultithreaded(boolean multithreaded)
    {
        checkStatus();
        mc.setMultithreaded(multithreaded);
    }

    /**
     * Accessor for the multithreaded capability of the manager 
     * @return Whether the PersistenceManager is multithreaded or not
     */
    public boolean getMultithreaded()
    {
        return false;
    }

    /**
     * Mutator for whether to ignore the cache or not 
     * @param ignore Whether to ignore the cache or not
     */
    public void setIgnoreCache(boolean ignore)
    {
        checkStatus();
        mc.setIgnoreCache(ignore);
    }

    /**
     * Accessor for whether to ignore the cache or not 
     * @return Whether to ignore the cache or not
     */
    public boolean getIgnoreCache()
    {
        return false;
    }

    /**
     * Accessor for the Store Manager.
     * @return The Store Manager.
     **/ 
    public StoreManager getStoreManager()
    {
        //TODO this should be enabled, but with reachability enabled it fails
        //checkStatus();
        return mc.getStoreManager();
    }

    /**
     * Accessor for the MetaData Manager.
     * @return The MetaData Manager.
     **/ 
    public MetaDataManager getMetaDataManager()
    {
        checkStatus();
        return mc.getMetaDataManager();
    }

    public void enlistInTransaction(StateManager sm)
    {
        checkStatus();
        mc.enlistInTransaction(sm);
    }

    public void evictFromTransaction(StateManager sm)
    {
        checkStatus();
        mc.evictFromTransaction(sm);
    }

    public boolean isEnlistedInTransaction(Object id)
    {
    	checkStatus();
    	return mc.isEnlistedInTransaction(id);
    }

    public void removeStateManager(StateManager sm)
    {
        checkStatus();
        mc.removeStateManager(sm);
    }

    public Object getObjectById(Object id, FieldValues fv)
    {
        checkStatus();
        return mc.getObjectById(id, fv);
    }

    public StateManager findStateManager(PersistenceCapable pc)
    {
        checkStatus();
        return mc.findStateManager(pc);
    }

    public void hereIsStateManager(StateManager sm, PersistenceCapable pc)
    {
        checkStatus();
        mc.hereIsStateManager(sm, pc);
    }

    public void markDirty(StateManager sm)
    {
        checkStatus();
        mc.markDirty(sm);
    }

    /**
     * This method flushes all dirty, new, and deleted instances to the datastore. It has no effect
     * if a transaction is not active.
     * If a datastore transaction is active, this method synchronizes the cache with the datastore
     * and reports any exceptions.
     * If an optimistic transaction is active, this method obtains a datastore connection and synchronizes
     * the cache with the datastore using this connection. The connection obtained by
     * this method is held until the end of the transaction.
     */
    public void flush()
    {
        checkStatus();
        mc.flush();
    }

    /**
     * Method to check the consistency of the cache.
     */
    public void checkConsistency()
    {
        checkStatus();
        mc.checkConsistency();
    }

    /**
     * Tests whether this object is being inserted.
     * @param pc the object to verify the status
     * @return true if this instance is inserting.
     */
    public boolean isInserting(PersistenceCapable pc)
    {
        checkStatus();
        return mc.isInserting(pc);
    }

    /**
     * Convenience method to find if the specified field of the specified object
     * has been inserted yet.
     * @param pc The object
     * @param fieldNumber The absolute field number
     * @return Whether it has been inserted into the datastore
     */
    public boolean isInserted(PersistenceCapable pc, int fieldNumber)
    {
        checkStatus();
        return mc.isInserted(pc, fieldNumber);
    }

    public void clearDirty(StateManager sm)
    {
        checkStatus();
        mc.clearDirty(sm);
    }    

    /**
     * Method to check the status.
     **/
    private void checkStatus()
    {
        if (closed || mc == null) 
        {
            throw new JDOException("Invalid state, closed or no mc");
        }
    }

    /**
     * Accessor for the ClassLoaderResolver
     * @return the ClassLoaderResolver
     */
    public ClassLoaderResolver getClassLoaderResolver()
    {
        return mc.getClassLoaderResolver();
    }

    /**
     * Accessor for a handle to the Persistence Manager
     * @return The PM handle.
     **/
    public PersistenceManager getPMHandle()
    {
        return mc.getPMHandle();
    }

    public void addStateManager(StateManager sm)
    {
        mc.addStateManager(sm);
    }

    public StateManager getStateManagerById(Object myID)
    {
        return mc.getStateManagerById(myID);
    }

    /**
     * Gets the context which this PersistenceManager is running
     * @return Returns the pmfContext.
     */
    public PMFContext getPMFContext()
    {
        return this.mc.getPMFContext();
    }

    /**
     * (non-Javadoc)
     * @see javax.jdo.PersistenceManager#getDataStoreConnection()
     * @since 1.1
     */
    public JDOConnection getDataStoreConnection()
    {
        checkStatus();
        return mc.getDataStoreConnection();
    }

    /**
     * Accessor for a Sequence.
     * @param sequenceName Name of the sequence
     * @return The sequence
     */
    public Sequence getSequence(String sequenceName)
    {
        checkStatus();
        return mc.getSequence(sequenceName);
    }

    /**
     * @see org.jpox.AbstractPersistenceManager#removeAllInstanceLifecycleListeners()
     */
    public void removeAllInstanceLifecycleListeners()
    {
        checkStatus();
        mc.removeAllInstanceLifecycleListeners();
    }

    /**
     * JDO 2.0 spec 12.15 "LifecycleListeners".
     * @param listener The instance lifecycle listener to sends events to
     * @param classes The classes that it is interested in
     * @since 1.1
     */
    public void addInstanceLifecycleListener(InstanceLifecycleListener listener, Class[] classes)
    {
        checkStatus();
        mc.addInstanceLifecycleListener(listener, classes);
    }
    
    /**
     * JDO 2.0 spec 12.15 "LifecycleListeners".
     * @param listener The instance lifecycle listener to remove.
     * @since 1.1
     */
    public void removeInstanceLifecycleListener(InstanceLifecycleListener listener)
    {
        checkStatus();
        mc.removeInstanceLifecycleListener(listener);
    }
    
    /**
     * Method to register a queryResult with the PM.
     * @param queryResult The results
     * @return The results number in this PM
     */
    public int addQueryResult(QueryResult queryResult)
    {
        return mc.addQueryResult(queryResult);
    }

    /**
     * Method to deregister a queryResult from the PM.
     * @param resultsNumber Number of the query results
     */
    public void removeQueryResult(int resultsNumber)
    {
        mc.removeQueryResult(resultsNumber);
    }

    /* (non-Javadoc)
     * @see org.jpox.PersistenceManager#addQueryRun(java.lang.String)
     */
    public void addQueryRun(String singleStringQuery)
    {
        mc.addQueryRun(singleStringQuery);
    }

    /* (non-Javadoc)
     * @see org.jpox.PersistenceManager#hasQueryRun(java.lang.String)
     */
    public boolean hasQueryRun(String singleStringQuery)
    {
        return mc.hasQueryRun(singleStringQuery);
    }

    /* (non-Javadoc)
     * @see org.jpox.PersistenceManager#removeQueryRun(java.lang.String)
     */
    public void removeQueryRun(String singleStringQuery)
    {
        mc.removeQueryRun(singleStringQuery);
    }

    /**
     * Replace the previous object id for a PC object to a new
     * @param pc The Persistence Capable object
     * @param oldID the old id
     * @param newID the new id
     */
    public void replaceObjectId(PersistenceCapable pc, Object oldID, Object newID)
    {
        mc.replaceObjectId(pc, oldID, newID);
    }

    /**
     * Method to put an object into the respective cache.
     * @param sm The state manager
     * @param level1 Whether to put into L1 cache
     * @param level2 Whether to put into L2 cache
     */
    public void putObjectIntoCache(StateManager sm, boolean level1, boolean level2)
    {
        mc.putObjectIntoCache(sm, level1, level2);
    }

    /**
     * Method to remove an object from the respective cache.
     * @param pc The object
     * @param id The id of the object
     * @param level1 Whether to remove from the L1 cache
     * @param level2 Whether to remove from the L2 cache
     */
    public void removeObjectFromCache(PersistenceCapable pc, Object id, boolean level1, boolean level2)
    {
        mc.removeObjectFromCache(pc, id, level1, level2);
    }

    /**
     * @param listener
     */
    public void addConnectionEventListener(ConnectionEventListener listener)
    {
        mc.addConnectionEventListener(listener);
    }
    /**
     * @param listener
     */
    public void removeConnectionEventListener(ConnectionEventListener listener)
    {
        mc.removeConnectionEventListener(listener);
    }

    public CallbackHandler getCallbackHandler()
    {
        return mc.getCallbackHandler();
    }

    public void detachAll()
    {
        mc.detachAll();
    }

    public void clearDirty()
    {
        mc.clearDirty();
    }
    
    /**
     * Drop any loaded state for the given field of the given PersistenceCapable, so upon next read 
     * access of that field it will get reloaded. Only possible if pc is not embedded.
     * @param pc
     * @param fieldName
     */
    public void unloadField(PersistenceCapable pc, String fieldName) 
    {
        findStateManager(pc).unloadField(fieldName);
    }
}