/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;
import java.util.AbstractQueue;
import java.util.Comparator;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOException;
import javax.jdo.spi.Detachable;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.sco.exceptions.IncompatibleElementTypeException;
import org.jpox.sco.exceptions.IncompatibleFieldTypeException;
import org.jpox.sco.exceptions.NullsNotAllowedException;
import org.jpox.sco.exceptions.QueryUnownedSCOException;
import org.jpox.state.FetchPlanState;
import org.jpox.state.StateManagerFactory;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.mapping.CollectionMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Queryable;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.store.scostore.CollectionStore;
import org.jpox.store.scostore.ListStore;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * A mutable second-class PriorityQueue object, representing a field in a class,
 * and containing elements of a particular type.
 * Uses a "delegate" as a local store for the PriorityQueue. Uses a "backing store"
 * (ListStore) to represent the datastore. The "delegate" is updated with the
 * "backing store" information at necessary intervals. 
 *
 * <H3>Modes of Operation</H3>
 * The user can operate the collection in 2 modes.
 * The <B>cached</B> mode will use an internal cache of the elements (in the
 * "delegate") reading them at the first opportunity and then using the cache
 * thereafter.
 * The <B>non-cached</B> mode will just go direct to the "backing store" each
 * call.
 *
 * <H3>Mutators</H3>
 * When the backing store is present any updates are passed direct to the
 * datastore as well as to the "delegate". If the "backing store" isn't present
 * the changes are made to the "delegate" only.
 *
 * <H3>Accessors</H3>
 * When any accessor method is invoked, it typically checks whether the
 * collection has been loaded from its backing store and does this as
 * necessary. Some methods (<B>size()</B>, <B>contains()</B>) just check if
 * everything is loaded and use the delegate if possible, otherwise going
 * direct to the datastore.
 *
 * @version $Revision: 1.2.2.1 $
 */
public class PriorityQueue extends AbstractQueue
    implements SCOCollection, SCOMtoN, Cloneable, Queryable, java.io.Serializable
{
    /** Localiser for messages. */
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation",PriorityQueue.class.getClassLoader());

    protected Object owner;
    protected StateManager ownerSM;
    protected String fieldName;
    protected int fieldNumber;
    protected Class elementType;
    protected boolean allowNulls;

    /** The "backing store" (for use when not serialised). */
    protected ListStore backingStore; // Really need a List since the Queue needs ordering

    /** The internal "delegate". */
    protected java.util.PriorityQueue delegate;

    /** Whether to use "delegate" caching. */
    protected boolean useCache=true;

    /** Status flag whether the collection is loaded into the cache. */
    protected boolean isCacheLoaded=false;

    /**
     * Constructor. 
     * @param ownerSM The State Manager for this set.
     * @param fieldName Name of the field
     **/
    public PriorityQueue(StateManager ownerSM, String fieldName)
    {
        init(ownerSM, fieldName, false, null);
    }

    /**
     * Initialisation method.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     * @param allowNulls Whether to allow nulls
     * @param backingStore Backing Store The backing store
     */
    protected void init(StateManager ownerSM, 
            String fieldName,
            boolean allowNulls,
            CollectionStore backingStore)
    {
        this.owner = ownerSM.getObject();
        this.ownerSM = ownerSM;
        this.fieldName = fieldName;
        this.allowNulls = false;

        Class c = ownerSM.getObject().getClass();
        StoreManager storeMgr = ownerSM.getStoreManager();
        AbstractClassMetaData cmd = storeMgr.getMetaDataManager().getMetaDataForClass(c, ownerSM.getPersistenceManager().getClassLoaderResolver());
        boolean serialised = SCOUtils.collectionHasSerialisedElements(cmd.getField(fieldName));
        this.fieldNumber = cmd.getFieldNumberAbsolute(fieldName);

        if (backingStore != null)
        {
            this.backingStore = (ListStore)backingStore;
            this.elementType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(backingStore.getElementType());
        }
        else if (!serialised && cmd.getManagedFieldAbsolute(fieldNumber).getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
        {
            DatastoreClass ownerTable = storeMgr.getDatastoreClass(c.getName(), ownerSM.getPersistenceManager().getClassLoaderResolver());
            JavaTypeMapping m = ownerTable.getFieldMapping(ownerSM.getClassMetaData().getField(fieldName));
            if (!(m instanceof CollectionMapping))
            {
                throw new IncompatibleFieldTypeException(ownerSM, fieldName, java.util.List.class.getName(), 
                    ownerSM.getMetaDataManager().getMetaDataForField(c, ownerSM.getPersistenceManager().getClassLoaderResolver(),fieldName).getTypeName());
            }
            this.backingStore = (ListStore)((CollectionMapping)m).getBackingStore(ownerSM.getPersistenceManager().getClassLoaderResolver());
            this.elementType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(this.backingStore.getElementType());
        }

        if (elementType != null && !this.elementType.isAssignableFrom(elementType))
        {
            throw new IncompatibleElementTypeException(ownerSM, fieldName, this.elementType.getName(), elementType.getName());
        }

        // Generate a suitable comparator where required (JPOX extension to JDO)
        Comparator comparator = null;
        AbstractPropertyMetaData fmd = cmd.getField(fieldName);
        String comparatorName = null;
        if (fmd.getCollection().hasExtension("comparatorName"))
        {
            comparatorName = fmd.getCollection().getValueForExtension("comparatorName");
            Class comparatorCls = null;
            try
            {
                comparatorCls = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(comparatorName);
                comparator = (Comparator)ClassUtils.newInstance(comparatorCls, null, null);
            }
            catch (JDOException jdoe)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("SCO.Comparator.ClassNotFound", fmd.getFullFieldName(), comparatorName));
            }
        }

        // Set up our delegate
        if (comparator != null)
        {
            this.delegate = new java.util.PriorityQueue(5, comparator);
        }
        else
        {
            this.delegate = new java.util.PriorityQueue();
        }

        // Set caching policy
        useCache = SCOUtils.useContainerCache(ownerSM, fieldName, cmd);
        if (useCache && !SCOUtils.useCachedLazyLoading(ownerSM, fieldName, cmd))
        {
            // Load up the collection now if not using lazy loading
            loadFromStore();
        }
    }

    // ----------------------- Implementation of SCO methods -------------------

    /**
     * Accessor for the element type.
     * @return the element type contained in the collection
     */
    public Class getElementType()
    {
        return elementType;
    }

    /**
     * Method to effect the load of the data in the SCO.
     * Used when the SCO supports lazy-loading to tell it to load all now.
     */
    public void load()
    {
        if (useCache)
        {
            loadFromStore();
        }
    }

    /**
     * Method to update an embedded element in this collection.
     * @param element The element
     * @param fieldNumber Number of field in the element
     * @param value New value for this field
     */
    public void updateEmbeddedElement(Object element, int fieldNumber, Object value)
    {
        if (backingStore != null)
        {
            backingStore.updateEmbeddedElement(ownerSM, element, fieldNumber, value);
        }
    }

    /**
     * Mutator for setting using an object.
     * @param o The object to set from
     * @param forUpdate Whether to update the datastore with this value
     */
    public void setValueFrom(Object o, boolean forUpdate)
    {
        java.util.Collection c = (java.util.Collection)o;
        if (c != null)
        {
            // Check for the case of serialised PC elements, and assign StateManagers to the elements without
            AbstractPropertyMetaData fmd = ownerSM.getClassMetaData().getField(fieldName);
            if (SCOUtils.collectionHasSerialisedElements(fmd) && fmd.getCollection().getElementClassMetaData() != null)
            {
                PersistenceManager pm = ownerSM.getPersistenceManager();
                Iterator iter = c.iterator();
                while (iter.hasNext())
                {
                    PersistenceCapable pc = (PersistenceCapable)iter.next();
                    StateManager objSM = pm.findStateManager(pc);
                    if (objSM == null)
                    {
                        objSM = StateManagerFactory.newStateManager(pc, pm, false);
                        objSM.addEmbeddedOwner(ownerSM, fieldNumber);
                    }
                }
            }
        }

        if (forUpdate)
        {
            SCOUtils.updateCollectionWithCollection(this, c);
        }
        else
        {
            delegate.clear();
            delegate.addAll(c);
        }
    }

    /**
     * Accessor for the field name.
     * @return The field name
     */
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Accessor for the owner object.
     * @return The owner object
     */
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Method to unset the owner and field information.
     */
    public synchronized void unsetOwner()
    {
        if (ownerSM != null)
        {
            owner = null;
            ownerSM = null;
            fieldName = null;
            backingStore = null;
        }
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        /*
         * Although we are never really "dirty", the owning object must be
         * marked dirty so that the proper state change occurs and its
         * jdoPreStore() gets called properly.
         */
        if (ownerSM != null)
        {
            ownerSM.makeDirty(fieldNumber);
        }
    }

    /**
     * Method to run reachability on this SCO.
     * @param reachables List of StateManagers reachable so far
     */
    public void runReachability(java.util.Set reachables)
    {
        Object[] values = toArray();
        for (int i = 0; i < values.length; i++)
        {
            if (values[i] != null && values[i] instanceof PersistenceCapable)
            {
                ownerSM.getPersistenceManager().findStateManager((PersistenceCapable)values[i]).runReachability(reachables);
            }
        }
    }

    /**
     * Method to make transient all elements of this container.
     * All elements that are PersistenceCapable will be made transient.
     * @param state State for the fetchplan process
     */
    public void makeTransient(FetchPlanState state)
    {
        Object[] values = toArray();
        for (int i = 0; i < values.length; i++)
        {
            if (values[i] != null && values[i] instanceof PersistenceCapable)
            {
                ownerSM.getPersistenceManager().internalMakeTransient(values[i], state);
            }
        }
        ownerSM = null;
        backingStore = null;
    }

    /**
     * Method to load all fields in the fetch plan (for subobjects).
     * @param state State for the fetchplan process
     */
    public void loadFieldsInFetchPlan(FetchPlanState state)
    {
        Object[] values = toArray();
        for (int i = 0; i < values.length; i++)
        {
            if (values[i] != null && values[i] instanceof PersistenceCapable)
            {
                ownerSM.getPersistenceManager().findStateManager((PersistenceCapable)values[i]).loadFieldsInFetchPlan(state);
            }
        }
    }

    /**
     * Method to detach the PersistenceCapable elements of this container.
     * @param state State for the detachment process
     */
    public void detach(FetchPlanState state)
    {
        for (Iterator it = iterator(); it.hasNext();)
        {
            Object object = it.next();
            if (object != null && object instanceof PersistenceCapable)
            {
               ownerSM.getPersistenceManager().detachInternal(object, state);
            }
        }
        ownerSM = null;
        backingStore = null;
    }

    /**
     * Method to return a detached copy of the container.
     * Recurses through the elements so that they are likewise detached.
     * @param state State for detachment process
     * @return The detached container
     */
    public Object detachCopy(FetchPlanState state)
    {
        java.util.Collection detached = new java.util.HashSet();
        for (Iterator it = iterator(); it.hasNext();)
        {
            Object object = it.next();
            if (object instanceof PersistenceCapable)
            {
                detached.add(ownerSM.getPersistenceManager().detachCopyInternal(object, state));
            }
            else
            {
                detached.add(object);
            }
        }
        return detached;
    }

    /**
     * Method to return an attached copy of the passed (detached) value. The returned attached copy
     * is a SCO wrapper. Goes through the existing elements in the store for this owner field and
     * removes ones no longer present, and adds new elements. All elements in the (detached)
     * value are attached.
     * @param value The new (collection) value
     */
    public void attachCopy(Object value)
    {
        java.util.Collection c = (java.util.Collection) value;

        // Attach all of the elements in the new collection
        AbstractPropertyMetaData fmd = ownerSM.getClassMetaData().getField(fieldName);
        boolean elementsWithoutIdentity = SCOUtils.collectionHasElementsWithoutIdentity(fmd);

        java.util.Collection attachedElements = new java.util.HashSet(c.size());
        Iterator iter = c.iterator();
        while (iter.hasNext())
        {
            Object detachedElement = iter.next();
            if (ClassUtils.isPersistenceCapable(detachedElement) && detachedElement instanceof Detachable)
            {
                attachedElements.add(ownerSM.getPersistenceManager().attachCopy(detachedElement, elementsWithoutIdentity));
            }
            else
            {
                attachedElements.add(detachedElement);
            }
        }

        // Update the attached collection with the detached elements
        SCOUtils.updateCollectionWithCollectionElements(this, attachedElements);
    }

    // ------------------------ Query Statement methods ------------------------

    /**
     * Method to generate a QueryStatement for the SCO.
     * @return The QueryStatement
     */
    public synchronized QueryExpression newQueryStatement()
    {
        return newQueryStatement(elementType);
    }

    /**
     * Method to return a QueryStatement, using the specified candidate class.
     * @param candidate_class the candidate class
     * @return The QueryStatement
     */
    public synchronized QueryExpression newQueryStatement(Class candidate_class)
    {
        if (backingStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return backingStore.newQueryStatement(ownerSM, candidate_class.getName());
    }

    /**
     * Method to return a ResultObjectFactory for the SCO.
     * @param stmt The QueryStatement
     * @param ignoreCache Whether to ignore the cache
     * @param resultClass Whether to create objects of a particular class
     * @param useFetchPlan whether to use the fetch plan to retrieve fields in the same query
     * @return The ResultObjectFactory
     */
    public synchronized ResultObjectFactory newResultObjectFactory(QueryExpression stmt,boolean ignoreCache, Class resultClass, boolean useFetchPlan)
    {
        if (backingStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return backingStore.newResultObjectFactory(ownerSM, stmt,ignoreCache,useFetchPlan);
    }

    // ---------------- Implementation of Queue methods -------------------

    /**
     * Creates and returns a copy of this object.
     * <P>
     * Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * @return A clone of the object
     */
    public Object clone()
    {
        if (useCache)
        {
            loadFromStore();
        }

//        return ((java.util.PriorityQueue)delegate).clone();
        // TODO Implement cloning
        return null;
    }

    /**
     * Accessor for the comparator.
     * @return The comparator
     */
    public Comparator comparator()
    {
        return delegate.comparator();
    }

    /**
     * Accessor for whether an element is contained in the Collection.
     * @param element The element
     * @return Whether the element is contained here
     **/
    public synchronized boolean contains(Object element)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.contains(element);
        }
        else if (backingStore != null)
        {
            return backingStore.contains(ownerSM,element);
        }

        return delegate.contains(element);
    }

    /**
     * Accessor for whether a collection of elements are contained here.
     * @param c The collection of elements.
     * @return Whether they are contained.
     **/
    public synchronized boolean containsAll(java.util.Collection c)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            java.util.HashSet h=new java.util.HashSet(c);
            Iterator iter=iterator();
            while (iter.hasNext())
            {
                h.remove(iter.next());
            }

            return h.isEmpty();
        }

        return delegate.containsAll(c);
    }

    /**
     * Equality operator.
     * @param o The object to compare against.
     * @return Whether this object is the same.
     **/
    public synchronized boolean equals(Object o)
    {
        if (useCache)
        {
            loadFromStore();
        }

        if (o == this)
        {
            return true;
        }
        if (!(o instanceof java.util.Set))
        {
            return false;
        }
        java.util.Set c = (java.util.Set)o;

        return c.size() == size() && containsAll(c);
    }

    /**
     * Hashcode operator.
     * @return The Hash code.
     **/
    public synchronized int hashCode()
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.hashCode();
    }

    /**
     * Accessor for whether the Collection is empty.
     * @return Whether it is empty.
     **/
    public synchronized boolean isEmpty()
    {
        return (size() == 0);
    }

    /**
     * Accessor for an iterator for the Collection.
     * @return The iterator
     **/
    public synchronized Iterator iterator()
    {
        // Populate the cache if necessary
        if (useCache)
        {
            loadFromStore();
        }
        return new SCOCollectionIterator(this, ownerSM, delegate, backingStore, useCache);
    }

    /**
     * Accessor for the size of the Collection.
     * @return The size
     **/
    public synchronized int size()
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.size();
        }
        else if (backingStore != null)
        {
            return backingStore.size(ownerSM);
        }

        return delegate.size();
    }

    /**
     * Method to return the Collection as an array.
     * @return The array
     **/
    public synchronized Object[] toArray()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return SCOUtils.toArray(backingStore,ownerSM);
        }  
        return delegate.toArray();
    }

    /**
     * Method to return the Collection as an array.
     * @param a The array to write the results to
     * @return The array
     **/
    public synchronized Object[] toArray(Object a[])
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return SCOUtils.toArray(backingStore,ownerSM,a);
        }  
        return delegate.toArray(a);
    }

    /**
     * Method to return the Collection as a String.
     * @return The string form
     **/
    public String toString()
    {
        StringBuffer s = new StringBuffer("[");
        int i=0;
        Iterator iter=iterator();
        while (iter.hasNext())
        {
            if (i > 0)
            {
                s.append(',');
            }
            s.append(iter.next());
            i++;
        }
        s.append("]");

        return s.toString();
    }

    // ----------------------------- Mutator methods ---------------------------

    /**
     * Method to add an element to the Collection.
     * @param element The element to add
     * @return Whether it was added successfully.
     **/
    public synchronized boolean add(Object element)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        if (useCache)
        {
            loadFromStore();
        }

        boolean backingSuccess = false;
        if (backingStore != null)
        {
            try
            {
                backingSuccess = backingStore.add(ownerSM,element);
            }
            catch (JDODataStoreException dse)
            {
                backingSuccess = false;
            }
        }

        // Only make it dirty after adding the field to the datastore so we give it time
        // to be inserted - otherwise jdoPreStore on this object would have been called before completing the addition
        makeDirty();

        boolean delegateSuccess = delegate.add(element);
        return (backingStore != null ? backingSuccess : delegateSuccess);
    }

    /**
     * Method to add a collection of elements.
     * @param c The collection of elements to add.
     * @return Whether they were added successfully.
     **/
    public synchronized boolean addAll(java.util.Collection c)
    {
        if (useCache)
        {
            loadFromStore();
        }

        boolean backingSuccess = false;
        if (backingStore != null)
        {
            try
            {
                backingSuccess = backingStore.addAll(ownerSM,c);
            }
            catch (JDODataStoreException dse)
            {
                backingSuccess = false;
            }
        }

        // Only make it dirty after adding the field to the datastore so we give it time
        // to be inserted - otherwise jdoPreStore on this object would have been called before completing the addition
        makeDirty();

        boolean delegateSuccess = delegate.addAll(c);
        return (backingStore != null ? backingSuccess : delegateSuccess);
    }

    /**
     * Method to clear the Collection.
     **/
    public synchronized void clear()
    {
        makeDirty();

        if (backingStore != null)
        {
            backingStore.clear(ownerSM);
        }
        delegate.clear();
    }

    /**
     * Method to offer an element to the Queue.
     * @param element The element to offer
     * @return Whether it was added successfully.
     **/
    public synchronized boolean offer(Object element)
    {
        return add(element);
    }

    /**
     * Method to peek at the next element in the Queue.
     * @return The element
     **/
    public synchronized Object peek()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return backingStore.get(ownerSM, 0);
        }

        return delegate.peek();
    }

    /**
     * Method to poll the next element in the Queue.
     * @return The element (now removed)
     **/
    public synchronized Object poll()
    {
        makeDirty();
 
        if (useCache)
        {
            loadFromStore();
        }

        Object delegateObject = delegate.poll();

        Object backingObject = null;
        if (backingStore != null)
        {
            try
            {
                backingObject = backingStore.remove(ownerSM, 0);
            }
            catch (JDODataStoreException dse)
            {
                backingObject = null;
            }
        }

        return (backingStore != null ? backingObject : delegateObject);
    }

    /**
     * Method to remove an element from the Collection.
     * @param element The Element to remove
     * @return Whether it was removed successfully.
     **/
    public synchronized boolean remove(Object element)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        boolean delegateSuccess = delegate.remove(element);

        boolean backingSuccess = false;
        if (backingStore != null)
        {
            try
            {
                backingSuccess = backingStore.remove(ownerSM, element);
            }
            catch (JDODataStoreException dse)
            {
                backingSuccess = false;
            }
        }

        return (backingStore != null ? backingSuccess : delegateSuccess);
    }

    /**
     * Method to remove a Collection of elements.
     * @param c The collection to remove
     * @return Whether they were removed successfully.
     **/
    public synchronized boolean removeAll(java.util.Collection c)
    {
        makeDirty();
 
        if (useCache)
        {
            loadFromStore();
        }

        boolean delegateSuccess = delegate.removeAll(c);

        boolean backingSuccess = false;
        if (backingStore != null)
        {
            try
            {
                backingSuccess = backingStore.removeAll(ownerSM, c);
            }
            catch (JDODataStoreException dse)
            {
                backingSuccess = false;
            }
        }

        return (backingStore != null ? backingSuccess : delegateSuccess);
    }

    /**
     * Method to retain a Collection of elements (and remove all others).
     * @param c The collection to retain
     * @return Whether they were retained successfully.
     **/
    public synchronized boolean retainAll(java.util.Collection c)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        boolean modified = false;
        Iterator iter = iterator();
        while (iter.hasNext())
        {
            Object element = iter.next();
            if (!c.contains(element))
            {
                iter.remove();
                modified = true;
            }
        }
        return modified;
    }

    /**
     * The writeReplace method is called when ObjectOutputStream is preparing
     * to write the object to the stream. The ObjectOutputStream checks whether
     * the class defines the writeReplace method. If the method is defined, the
     * writeReplace method is called to allow the object to designate its
     * replacement in the stream. The object returned should be either of the
     * same type as the object passed in or an object that when read and
     * resolved will result in an object of a type that is compatible with all
     * references to the object.
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        if (useCache)
        {
            loadFromStore();
            return new java.util.PriorityQueue(delegate);
        }
        else
        {
            // TODO Cater for non-cached collection, load elements in a DB call.
            return new java.util.PriorityQueue(delegate);
        }
    }

    /**
     * Method to load all elements from the "backing store" where appropriate.
     */
    protected void loadFromStore()
    {
        if (backingStore != null && !isCacheLoaded)
        {
            delegate.clear();
            Iterator iter=backingStore.iterator(ownerSM);
            while (iter.hasNext())
            {
                delegate.add(iter.next());
            }

            isCacheLoaded = true;
        }
    }
}