/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;
import java.util.Collection;
import java.util.Iterator;
import java.util.ListIterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.spi.Detachable;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.sco.exceptions.IncompatibleElementTypeException;
import org.jpox.sco.exceptions.IncompatibleFieldTypeException;
import org.jpox.sco.exceptions.NullsNotAllowedException;
import org.jpox.sco.exceptions.QueryUnownedSCOException;
import org.jpox.state.FetchPlanState;
import org.jpox.state.StateManagerFactory;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.mapping.CollectionMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Queryable;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.store.scostore.ListStore;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * A mutable second-class Stack object.
 * <P> 
 * Extends java.util.Stack since it needs to be castable to the users type.
 * Backed by a ListStore object. Uses a delegate internally to store its
 * internal state rather than the superclass. The reason for this is that the
 * java.util collections call their own methods internally and so things like
 * removeAll calls remove on each item in the list. We want control since we
 * want to remove the datastore entries in one call rather than mutiple.
 * All mutator methods are intercepted and sent to the backing store, so it
 * is always consistent with the current object. 
 * </P>
 * 
 * @version $Revision: 1.59.2.1 $
 */
public class Stack extends java.util.Stack 
    implements SCOList, Cloneable, Queryable
{
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    private transient Object owner;
    private transient StateManager ownerSM;
    private transient String fieldName;
    private transient int fieldNumber;
    private transient ListStore backingStore;
    private transient Class elementType;
    private transient boolean allowNulls;

    /** The internal "delegate". */
    private java.util.Stack delegate;

    /** Whether to use "delegate" caching. */
    protected boolean useCache=true;

    /** Status flag whether the collection is loaded into the cache. */
    protected boolean isCacheLoaded=false;
    
    /**
     * Constructor, using the StateManager of the "owner" and the field name.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     **/
    public Stack(StateManager ownerSM, String fieldName)
    {
        init(ownerSM, fieldName);
    }

    /**
     * Initialisation method.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     */
    private void init(StateManager ownerSM, 
                      String fieldName)
    {
        this.owner = ownerSM.getObject();
        this.ownerSM = ownerSM;
        this.fieldName = fieldName;
        this.allowNulls = false;

        Class c = ownerSM.getObject().getClass();
        StoreManager storeMgr = ownerSM.getStoreManager();
        AbstractClassMetaData cmd = storeMgr.getMetaDataManager().getMetaDataForClass(c, ownerSM.getPersistenceManager().getClassLoaderResolver());
        boolean serialised = SCOUtils.collectionHasSerialisedElements(cmd.getField(fieldName));
        this.fieldNumber = cmd.getFieldNumberAbsolute(fieldName);

        if (!serialised && cmd.getManagedFieldAbsolute(fieldNumber).getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
        {
            DatastoreClass ownerTable = storeMgr.getDatastoreClass(c.getName(), ownerSM.getPersistenceManager().getClassLoaderResolver());
            JavaTypeMapping m = ownerTable.getFieldMapping(ownerSM.getClassMetaData().getField(fieldName));
            if (!(m instanceof CollectionMapping))
            {
                throw new IncompatibleFieldTypeException(ownerSM, fieldName, java.util.Stack.class.getName(), ownerSM.getMetaDataManager().getMetaDataForField(c, ownerSM.getPersistenceManager().getClassLoaderResolver(),fieldName).getTypeName());
            }
            this.backingStore = (ListStore)((CollectionMapping)m).getBackingStore(ownerSM.getPersistenceManager().getClassLoaderResolver());
            this.elementType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(backingStore.getElementType());
        }

        if (elementType != null && !this.elementType.isAssignableFrom(elementType))
        {
            throw new IncompatibleElementTypeException(ownerSM, fieldName, this.elementType.getName(), elementType.getName());
        }

        // Set up our delegate
        this.delegate = new java.util.Stack();

        // Set caching policy
        useCache = SCOUtils.useContainerCache(ownerSM, fieldName, cmd);
        if (useCache && !SCOUtils.useCachedLazyLoading(ownerSM, fieldName, cmd))
        {
            // Load up the collection now if not using lazy loading
            loadFromStore();
        }
    }

    // ----------------------- Implementation of SCO methods -------------------

    /**
     * Accessor for the element type.
     * @return the element type contained in the collection
     */
    public Class getElementType()
    {
        return elementType;
    }

    /**
     * Method to effect the load of the data in the SCO.
     * Used when the SCO supports lazy-loading to tell it to load all now.
     */
    public void load()
    {
        if (useCache)
        {
            loadFromStore();
        }
    }

    /**
     * Method to update an embedded element in this collection.
     * @param element The element
     * @param fieldNumber Number of field in the element
     * @param value New value for this field
     */
    public void updateEmbeddedElement(Object element, int fieldNumber, Object value)
    {
        if (backingStore != null)
        {
            backingStore.updateEmbeddedElement(ownerSM, element, fieldNumber, value);
        }
    }

    /**
     * Mutator for setting using an object.
     * @param o The object to set from
     * @param forUpdate Whether to update the datastore with this value
     **/
    public void setValueFrom(Object o, boolean forUpdate)
    {
        Collection c = (Collection)o;
        if (c != null)
        {
            // Check for the case of serialised PC elements, and assign StateManagers to the elements without
            AbstractPropertyMetaData fmd = ownerSM.getClassMetaData().getField(fieldName);
            if (SCOUtils.collectionHasSerialisedElements(fmd) && fmd.getCollection().getElementClassMetaData() != null)
            {
                PersistenceManager pm = ownerSM.getPersistenceManager();
                Iterator iter = c.iterator();
                while (iter.hasNext())
                {
                    PersistenceCapable pc = (PersistenceCapable)iter.next();
                    StateManager objSM = pm.findStateManager(pc);
                    if (objSM == null)
                    {
                        objSM = StateManagerFactory.newStateManager(pc, pm, false);
                        objSM.addEmbeddedOwner(ownerSM, fieldNumber);
                    }
                }
            }
        }

        if (forUpdate)
        {
            clear();
            addAll(c);
        }
        else
        {
            delegate.clear();
            delegate.addAll(c);
        }
    }

    /**
     * Accessor for the field name.
     * @return The field name
     **/
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Accessor for the owner object.
     * @return The owner object
     **/
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Method to unset the owner and field information.
     **/
    public synchronized void unsetOwner()
    {
        if (ownerSM != null)
        {
            owner = null;
            ownerSM = null;
            fieldName = null;
            backingStore = null;
        }
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        // Although we are never really "dirty", the owning object must be
        // marked dirty so that the proper state change occurs and its
        // jdoPreStore() gets called properly.
        if (ownerSM != null)
        {
            ownerSM.makeDirty(fieldNumber);
        }
    }

    /**
     * Method to run reachability on this SCO.
     * @param reachables List of StateManagers reachable so far
     */
    public void runReachability(java.util.Set reachables)
    {
        Object[] values = toArray();
        for (int i = 0; i < values.length; i++)
        {
            if (values[i] != null && values[i] instanceof PersistenceCapable)
            {
                ownerSM.getPersistenceManager().findStateManager((PersistenceCapable)values[i]).runReachability(reachables);
            }
        }
    }

    /**
     * Method to make transient all elements of this container.
     * All elements that are PersistenceCapable will be made transient.
     * @param state State for the fetchplan process
     */
    public void makeTransient(FetchPlanState state)
    {
        Object[] values = toArray();
        for (int i = 0; i < values.length; i++)
        {
            if (values[i] != null && values[i] instanceof PersistenceCapable)
            {
                ownerSM.getPersistenceManager().internalMakeTransient(values[i], state);
            }
        }
        ownerSM = null;
        backingStore = null;
    }

    /**
     * Method to load all fields in the fetch plan (for subobjects).
     * @param state State for the fetchplan process
     */
    public void loadFieldsInFetchPlan(FetchPlanState state)
    {
        Object[] values = toArray();
        for (int i = 0; i < values.length; i++)
        {
            if (values[i] != null && values[i] instanceof PersistenceCapable)
            {
                ownerSM.getPersistenceManager().findStateManager((PersistenceCapable)values[i]).loadFieldsInFetchPlan(state);
            }
        }
    }

    /**
     * Method to detach the PersistenceCapable elements of this container.
     * @param state State for the detachment process
     */
    public void detach(FetchPlanState state)
    {
        for (Iterator it = iterator(); it.hasNext();)
        {
            Object object = it.next();
            if (object != null && object instanceof PersistenceCapable)
            {
               ownerSM.getPersistenceManager().detachInternal(object, state);
            }
        }
        ownerSM = null;
        backingStore = null;
    }

    /**
     * Method to return a detached copy of the container.
     * Recurse sthrough the elements so that they are likewise detached.
     * @param state State of detachment state
     * @return The detached container
     */
    public Object detachCopy(FetchPlanState state)
    {
        java.util.Collection detached = new java.util.Stack();
        Object[] values = toArray();
        for (int i = 0; i < values.length; i++)
        {
            if (values[i] == null)
            {
                detached.add(null);
            }
            else
            {
                Object object = values[i];
                if (object instanceof PersistenceCapable)
                {
                    detached.add(ownerSM.getPersistenceManager().detachCopyInternal(object, state));
                }
                else
                {
                    detached.add(object);
                }
            }
        }
        return detached;
    }

    /**
     * Method to return an attached copy of the passed (detached) value. The returned attached copy
     * is a SCO wrapper. Goes through the existing elements in the store for this owner field and
     * removes ones no longer present, and adds new elements. All elements in the (detached)
     * value are attached.
     * @param value The new (collection) value
     */
    public void attachCopy(Object value)
    {
        java.util.Collection c = (java.util.Collection) value;

        // Attach all of the elements in the new list
        AbstractPropertyMetaData fmd = ownerSM.getClassMetaData().getField(fieldName);
        boolean elementsWithoutIdentity = SCOUtils.collectionHasElementsWithoutIdentity(fmd);

        java.util.List attachedElements = new java.util.ArrayList(c.size());
        Iterator iter = c.iterator();
        while (iter.hasNext())
        {
            Object detachedElement = iter.next();
            if (ClassUtils.isPersistenceCapable(detachedElement) && detachedElement instanceof Detachable)
            {
                attachedElements.add(ownerSM.getPersistenceManager().attachCopy(detachedElement, elementsWithoutIdentity));
            }
            else
            {
                attachedElements.add(detachedElement);
            }
        }

        // Update the attached list with the detached elements
        SCOUtils.updateListWithListElements(this, attachedElements);
    }

    // ------------------------ Query Statement methods ------------------------
    /**
     * Method to generate a QueryStatement for the SCO.
     * @return The QueryStatement
     **/
    public synchronized QueryExpression newQueryStatement()
    {
        return newQueryStatement(elementType);
    }

    /**
     * Method to return a QueryStatement, using the specified candidate class.
     * @param candidate_class the candidate class
     * @return The QueryStatement
     **/
    public synchronized QueryExpression newQueryStatement(Class candidate_class)
    {
        if (backingStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return backingStore.newQueryStatement(ownerSM, candidate_class.getName());
    }

    /**
     * Method to return a ResultObjectFactory for the SCO.
     * @param stmt The QueryStatement
     * @param ignoreCache Whether to ignore the cache
     * @param resultClass Whether to create objects of a particular class
     * @param useFetchPlan whether to use the fetch plan to retrieve fields in the same query
     * @return The ResultObjectFactory
     **/
    public synchronized ResultObjectFactory newResultObjectFactory(QueryExpression stmt,boolean ignoreCache,Class resultClass,boolean useFetchPlan)
    {
        if (backingStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return backingStore.newResultObjectFactory(ownerSM, stmt,ignoreCache,useFetchPlan);
    }

    // ------------------- Implementation of Stack methods ---------------------
 
    /**
     * Clone operator to return a copy of this object.
     * <p>
     * Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * </p>
     *
     * @return The cloned object
     */
    public Object clone()
    {
        if (useCache)
        {
            loadFromStore();
        }

        return delegate.clone();
    }

    /**
     * Method to return if the list contains this element.
     * @param element The element
     * @return Whether it is contained
     **/
    public boolean contains(Object element)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.contains(element);
        }
        else if (backingStore != null)
        {
            return backingStore.contains(ownerSM,element);
        }

        return delegate.contains(element);
    }

    /**
     * Accessor for whether the Stack is empty.
     * @return Whether it is empty.
     **/
    public boolean empty()
    {
        return isEmpty();
    }

    /**
     * Equality operator.
     * @param o The object to compare against.
     * @return Whether this object is the same.
     **/
    public synchronized boolean equals(Object o)
    {
        if (useCache)
        {
            loadFromStore();
        }

        if (o == this)
        {
            return true;
        }

        if (!(o instanceof java.util.List))
        {
            return false;
        }
        java.util.List l = (java.util.List)o;
        if (l.size() != size())
        {
            return false;
        }
        Object[] elements = toArray();
        Object[] otherElements = l.toArray();
        for (int i=0;i<elements.length;i++)
        {
            if (!elements[i].equals(otherElements[i]))
            {
                return false;
            }
        }
        return true;
    }

    /**
     * Method to retrieve an element no.
     * @param index The item to retrieve
     * @return The element at that position.
     **/
    public Object get(int index)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return backingStore.get(ownerSM, index);
        }

        return delegate.get(index);
    }

    /**
     * Method to the position of an element.
     * @param element The element.
     * @return The position.
     **/
    public int indexOf(Object element)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return backingStore.indexOf(ownerSM, element);
        }
 
        return delegate.indexOf(element);
    }

    /**
     * Accessor for whether the Stack is empty.
     * @return Whether it is empty.
     **/
    public boolean isEmpty()
    {
        return size() == 0;
    }

    /**
     * Method to retrieve an iterator for the list.
     * @return The iterator
     **/
    public Iterator iterator()
    {
        // Populate the cache if necessary
        if (useCache)
        {
            loadFromStore();
        }

        return new SCOListIterator(this, ownerSM, delegate, backingStore, useCache, -1);
    }

    /**
     * Method to retrieve a List iterator for the list.
     * @return The iterator
     **/
    public ListIterator listIterator()
    {
        // Populate the cache if necessary
        if (useCache)
        {
            loadFromStore();
        }

        return new SCOListIterator(this, ownerSM, delegate, backingStore, useCache, -1);
    }

    /**
     * Method to retrieve a List iterator for the list from the index.
     * @param index The start point 
     * @return The iterator
     **/
    public ListIterator listIterator(int index)
    {
        // Populate the cache if necessary
        if (useCache)
        {
            loadFromStore();
        }

        return new SCOListIterator(this, ownerSM, delegate, backingStore, useCache, index);
    }

    /**
     * Method to retrieve the last position of the element.
     * @param element The element
     * @return The last position of this element in the List.
     **/
    public int lastIndexOf(Object element)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return backingStore.lastIndexOf(ownerSM, element);
        }
 
        return delegate.lastIndexOf(element);
    }

    /**
     * Method to retrieve the element at the top of the stack.
     *
     * @return The element at the top of the stack
     **/
    public Object peek()
    {
        return get(0);
    }

    /**
     * Accessor for the size of the Stack.
     * @return The size.
     **/
    public int size()
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.size();
        }
        else if (backingStore != null)
        {
            return backingStore.size(ownerSM);
        }

        return delegate.size();
    }

    /**
     * Accessor for the subList of elements between from and to of the List
     * @param from Start index (inclusive)
     * @param to End index (exclusive) 
     * @return The subList
     **/
    public synchronized java.util.List subList(int from,int to)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return backingStore.subList(ownerSM,from,to);
        }

        return delegate.subList(from,to);
    }

    /**
     * Method to return the list as an array.
     * @return The array
     **/
    public synchronized Object[] toArray()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return SCOUtils.toArray(backingStore,ownerSM);
        }  
        return delegate.toArray();
    }

    /**
     * Method to return the list as an array.
     * @param a The runtime types of the array being defined by this param
     * @return The array
     **/
    public synchronized Object[] toArray(Object a[])
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return SCOUtils.toArray(backingStore,ownerSM,a);
        }  
        return delegate.toArray(a);
    }

    // ------------------------------ Mutator methods --------------------------
 
    /**
     * Method to add an element to a position in the Stack
     *
     * @param index The position
     * @param element The new element
     **/
    public void add(int index,Object element)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        if (useCache)
        {
            loadFromStore();
        }

        if (backingStore != null)
        {
            try
            {
                backingStore.add(ownerSM, index, element);
            }
            catch (JDODataStoreException dse)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("SCO.ErrorExecutingMethod", "add", fieldName, dse));
            }
        }

        // Only make it dirty after adding the element(s) to the datastore so we give it time
        // to be inserted - otherwise jdoPreStore on this object would have been called before completing the addition
        makeDirty();

        delegate.add(index, element);
    }

    /**
     * Method to add an element to the Stack
     *
     * @param element The new element
     * @return Whether it was added ok.
     **/
    public boolean add(Object element)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        if (useCache)
        {
            loadFromStore();
        }

        boolean backingSuccess = false;
        if (backingStore != null)
        {
            try
            {
                backingSuccess = backingStore.add(ownerSM,element);
            }
            catch (JDODataStoreException dse)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("SCO.ErrorExecutingMethod", "add", fieldName, dse));
                backingSuccess = false;
            }
        }

        // Only make it dirty after adding the element(s) to the datastore so we give it time
        // to be inserted - otherwise jdoPreStore on this object would have been called before completing the addition
        makeDirty();

        boolean delegateSuccess = delegate.add(element);
        return (backingStore != null ? backingSuccess : delegateSuccess);
    }

    /**
     * Method to add an element to the Stack
     *
     * @param element The new element
     **/
    public void addElement(Object element)
    {
        // This is a historical wrapper to the Collection method
        add(element);
    }

    /**
     * Method to add a Collection to the Stack
     * @param elements The collection
     * @return Whether it was added ok.
     **/
    public boolean addAll(Collection elements)
    {
        if (useCache)
        {
            loadFromStore();
        }

        boolean backingSuccess = false;
        if (backingStore != null)
        {
            try
            {
                backingSuccess = backingStore.addAll(ownerSM, elements);
            }
            catch (JDODataStoreException dse)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("SCO.ErrorExecutingMethod", "addAll", fieldName, dse));
                backingSuccess = false;
            }
        }

        // Only make it dirty after adding the element(s) to the datastore so we give it time
        // to be inserted - otherwise jdoPreStore on this object would have been called before completing the addition
        makeDirty();

        boolean delegateSuccess = delegate.addAll(elements);
        return (backingStore != null ? backingSuccess : delegateSuccess);
    }

    /**
     * Method to add a Collection to a position in the Stack
     * @param index Position to insert the collection.
     * @param elements The collection
     * @return Whether it was added ok.
     **/
    public boolean addAll(int index,Collection elements)
    {
        if (useCache)
        {
            loadFromStore();
        }

        boolean backingSuccess = false;
        if (backingStore != null)
        {
            try
            {
                backingSuccess = backingStore.addAll(ownerSM, index, elements);
            }
            catch (JDODataStoreException dse)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("SCO.ErrorExecutingMethod", "addAll", fieldName, dse));
                backingSuccess = false;
            }
        }

        // Only make it dirty after adding the element(s) to the datastore so we give it time
        // to be inserted - otherwise jdoPreStore on this object would have been called before completing the addition
        makeDirty();

        boolean delegateSuccess = delegate.addAll(index, elements);
        return (backingStore != null ? backingSuccess : delegateSuccess);
    }

    /**
     * Method to clear the Stack
     **/
    public synchronized void clear()
    {
        makeDirty();

        if (backingStore != null)
        {
            backingStore.clear(ownerSM);
        }
        delegate.clear();
    }

    /**
     * Method to remove the top element in the stack and return it.
     * @return The top element that was in the Stack (now removed).
     **/
    public Object pop()
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (backingStore != null)
        {
            backingStore.remove(ownerSM,0);
        }
        return delegate.remove(0);
    }

    /**
     * Method to push an element onto the stack and return it.
     *
     * @param element The element to push onto the stack.
     * @return The element that was pushed onto the Stack
     **/
    public Object push(Object element)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        if (useCache)
        {
            loadFromStore();
        }

        if (backingStore != null)
        {
            backingStore.add(ownerSM,0,element);
        }

        // Only make it dirty after adding the element(s) to the datastore so we give it time
        // to be inserted - otherwise jdoPreStore on this object would have been called before completing the addition
        makeDirty();

        delegate.add(0,element);
        return element;
    }

    /**
     * Method to remove an element from the Stack
     * @param element The element
     * @return Whether the element was removed
     **/
    public boolean remove(Object element)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        boolean delegateSuccess = delegate.remove(element);

        boolean backingSuccess = false;
        if (backingStore != null)
        {
            try
            {
                backingSuccess = backingStore.remove(ownerSM, element);
            }
            catch (JDODataStoreException dse)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("SCO.ErrorExecutingMethod", "remove", fieldName, dse));
                backingSuccess = false;
            }
        }

        return (backingStore != null ? backingSuccess : delegateSuccess);
    }

    /**
     * Method to remove a Collection of objects from the Stack
     * @param elements The Collection
     * @return Whether the collection of elements were removed
     **/
    public boolean removeAll(Collection elements)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        boolean delegateSuccess = delegate.removeAll(elements);

        boolean backingSuccess = false;
        if (backingStore != null)
        {
            try
            {
                backingSuccess = backingStore.removeAll(ownerSM, elements);
            }
            catch (JDODataStoreException dse)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("SCO.ErrorExecutingMethod", "removeAll", fieldName, dse));
                backingSuccess = false;
            }
        }

        return (backingStore != null ? backingSuccess : delegateSuccess);
    }

    /**
     * Method to remove an element from the Stack
     * @param element The element
     * @return Whether the element was removed
     **/
    public boolean removeElement(Object element)
    {
        // This is a historical wrapper to the Collection method
        return remove(element);
    }

    /**
     * Method to remove an element from the Stack
     * @param index The element position.
     * @return The object that was removed
     **/
    public Object remove(int index)
    {
        makeDirty();
 
        if (useCache)
        {
            loadFromStore();
        }

        Object delegateObject = delegate.remove(index);

        Object backingObject = null;
        if (backingStore != null)
        {
            try
            {
                backingObject = backingStore.remove(ownerSM, index);
            }
            catch (JDODataStoreException dse)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("SCO.ErrorExecutingMethod", "remove", fieldName, dse));
                backingObject = null;
            }
        }

        return (backingStore != null ? backingObject : delegateObject);
    }

    /**
     * Method to remove an element from the Stack
     * @param index The element position.
     **/
    public void removeElementAt(int index)
    {
        // This is a historical wrapper to the Collection method
        remove(index);
    }

    /**
     * Method to remove all elements from the Stack.
     * Same as clear().
     **/
    public void removeAllElements()
    {
        clear();
    }

    /**
     * Method to retain a Collection of elements (and remove all others).
     * @param c The collection to retain
     * @return Whether they were retained successfully.
     **/
    public synchronized boolean retainAll(java.util.Collection c)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }
        
        boolean modified = false;
        Iterator iter=iterator();
        while (iter.hasNext())
        {
            Object element = iter.next();
            if (!c.contains(element))
            {
                iter.remove();
                modified = true;
            }
        }
        return modified;
    }

    /**
     * JPOX wrapper addition that allows turning off of the dependent-field checks
     * when doing the position setting. This means that we can prevent the deletion of
     * the object that was previously in that position. This particular feature is used
     * when attaching a list field and where some elements have changed positions.
     * @param index The position
     * @param element The new element
     * @return The element previously at that position
     */
    public Object set(int index, Object element, boolean allowDependentField)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (backingStore != null)
        {
            backingStore.set(ownerSM, index, element, allowDependentField);
        }
        return delegate.set(index, element);
    }

    /**
     * Method to set the element at a position in the Stack
     *
     * @param index The position
     * @param element The new element
     * @return The element previously at that position
     **/
    public Object set(int index,Object element)
    {
        return set(index, element, true);
    }

    /**
     * Method to set the element at a position in the Stack
     *
     * @param element The new element
     * @param index The position
     **/
    public void setElementAt(Object element,int index)
    {
        // This is a historical wrapper to the Collection method
        set(index,element);
    }

    /**
     * The writeReplace method is called when ObjectOutputStream is preparing
     * to write the object to the stream. The ObjectOutputStream checks whether
     * the class defines the writeReplace method. If the method is defined, the
     * writeReplace method is called to allow the object to designate its
     * replacement in the stream. The object returned should be either of the
     * same type as the object passed in or an object that when read and
     * resolved will result in an object of a type that is compatible with all
     * references to the object.
     * 
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        if (useCache)
        {
            loadFromStore();
            java.util.Stack stack = new java.util.Stack();
            stack.addAll(delegate);
            return stack;
        }
        else
        {
            // TODO Cater for non-cached collection, load elements in a DB call.
            java.util.Stack stack = new java.util.Stack();
            stack.addAll(delegate);
            return stack;
        }
    } 

    /**
     * Method to load all elements from the "backing store" where appropriate.
     */
    protected void loadFromStore()
    {
        if (backingStore != null && !isCacheLoaded)
        {
            delegate.clear();
            Iterator iter=backingStore.iterator(ownerSM);
            while (iter.hasNext())
            {
                delegate.add(iter.next());
            }

            isCacheLoaded = true;
        }
    }
}