/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
    ...
**********************************************************************/
package org.jpox.state;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.FetchPlanImpl.FetchPlanForClass;
import org.jpox.sco.SCO;
import org.jpox.sco.SCOContainer;
import org.jpox.store.fieldmanager.SingleValueFieldManager;

/**
 * Utility class to handle loading all fields of all objects in the fetch plan.
 * The method in StateManagerImpl only loads the fields for that object and so
 * will only load the DFG fields for objects (hence omitting any non-DFG fields
 * that are in the FetchPlan that have been omitted due to lazy-loading).
 *
 * @version $Revision: 1.1 $
 **/
public class LoadFieldManager extends AbstractFetchFieldManager
{
    /**
     * Constructor for a field manager for make transient process.
     * @param sm the StateManager of the instance being loaded
     * @param secondClassMutableFields The second class mutable fields for the class of this object
     * @param fpClass Fetch Plan for the class of this instance
     * @param state State object to hold any pertinent controls for the fetchplan process
     */
    public LoadFieldManager(StateManager sm, boolean[] secondClassMutableFields, FetchPlanForClass fpClass, FetchPlanState state)
    {
        super(sm, secondClassMutableFields, fpClass, state);
    }

    /**
     * Method to fetch an object field whether it is SCO collection, PC, or whatever for the fetchplan process.
     * @param fieldNumber Number of the field
     * @return The object
     */
    protected Object internalFetchObjectField(int fieldNumber)
    {
        SingleValueFieldManager sfv = new SingleValueFieldManager();
        sm.provideFields(new int[]{fieldNumber}, sfv);
        Object value = sfv.fetchObjectField(fieldNumber);

        if (value == null)
        {
            // Do nothing
        }
        else if (secondClassMutableFields[fieldNumber])
        {
            SCO sco;
            if (value instanceof SCO)
            {
                sco = (SCO) value;
            }
            else
            {
                // Replace with a SCO wrapper so that we can load it
                sco = (SCO) sm.replaceSCOField(fieldNumber, value);
            }
            if (sco instanceof SCOContainer)
            {
                // Container of objects so recurse through them to make sure they are loaded
                ((SCOContainer)sco).loadFieldsInFetchPlan(state);
            }
        }
        else if (value.getClass().isArray() &&
                 PersistenceCapable.class.isAssignableFrom(value.getClass().getComponentType()))
        {
            // Array of PersistenceCapable objects
            // Return a new array with all elements transient (use the existing element if already transient, otherwise
            // make it transient).
            Object[] arrValue = (Object[])value;
            for (int j=0;j<arrValue.length;j++)
            {
                PersistenceCapable pc = (PersistenceCapable)arrValue[j];
                ((PersistenceManager)pc.jdoGetPersistenceManager()).findStateManager(pc).loadFieldsInFetchPlan(state);
            }
        }
        else if (value instanceof PersistenceCapable)
        {
            PersistenceCapable pc = (PersistenceCapable) value;
            ((PersistenceManager)pc.jdoGetPersistenceManager()).findStateManager(pc).loadFieldsInFetchPlan(state);
        }
        // TODO We have omitted interface fields, interface array fields, object reference fields, object reference array fields

        return value;
    }

    /**
     * Method called when were arrive at the end of a branch
     * @param fieldNumber Number of the field
     * @return Object to return
     */
    protected Object endOfGraphOperation(int fieldNumber)
    {
        SingleValueFieldManager sfv = new SingleValueFieldManager();
        sm.provideFields(new int[]{fieldNumber}, sfv);
        Object value = sfv.fetchObjectField(fieldNumber);

        return value;
    }
}