/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
    ...
**********************************************************************/
package org.jpox.state;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.FetchPlanImpl.FetchPlanForClass;
import org.jpox.sco.SCO;
import org.jpox.sco.SCOContainer;
import org.jpox.store.fieldmanager.SingleValueFieldManager;

/**
 * Utility class to handle the making transient of fields.
 *
 * @version $Revision: 1.8 $
 **/
public class MakeTransientFieldManager extends AbstractFetchFieldManager
{
    /**
     * Constructor for a field manager for make transient process.
     * @param sm the StateManager of the instance being detached. An instance in Persistent or Transactional state
     * @param secondClassMutableFields The second class mutable fields for the class of this object
     * @param fpClass Fetch Plan for the class of this instance
     * @param state State object to hold any pertinent controls for the fetchplan process
     */
    public MakeTransientFieldManager(StateManager sm, boolean[] secondClassMutableFields, FetchPlanForClass fpClass, FetchPlanState state)
    {
        super(sm, secondClassMutableFields, fpClass, state);
    }

    /**
     * Method to fetch an object field whether it is SCO collection, PC, or whatever for the fetchplan process.
     * @param fieldNumber Number of the field
     * @return The object
     */
    protected Object internalFetchObjectField(int fieldNumber)
    {
        SingleValueFieldManager sfv = new SingleValueFieldManager();
        sm.provideFields(new int[]{fieldNumber}, sfv);
        Object value = sfv.fetchObjectField(fieldNumber);

        if (value == null)
        {
            // Do nothing
        }
        else if (secondClassMutableFields[fieldNumber])
        {
            SCO sco;
            if (value instanceof SCO)
            {
                sco = (SCO) value;
            }
            else
            {
                // Replace with a SCO wrapper so that we can make it transient
                sco = (SCO) sm.replaceSCOField(fieldNumber, value);
            }

            if (sco instanceof SCOContainer)
            {
                ((SCOContainer)sco).makeTransient(state);
            }
        }
        else if (value.getClass().isArray() &&
                 PersistenceCapable.class.isAssignableFrom(value.getClass().getComponentType()))
        {
            // Array of PersistenceCapable objects
            // Return a new array with all elements transient (use the existing element if already transient, otherwise
            // make it transient).
            Object[] arrValue = (Object[])value;
            for (int j=0;j<arrValue.length;j++)
            {
                PersistenceCapable pc = (PersistenceCapable)arrValue[j];
                if (pc.jdoIsPersistent())
                {
                    ((PersistenceManager)pc.jdoGetPersistenceManager()).internalMakeTransient(pc, state);
                }
            }
        }
        else if (value instanceof PersistenceCapable)
        {
            PersistenceCapable pc = (PersistenceCapable) value;
            if (pc.jdoIsPersistent())
            {
                ((PersistenceManager)pc.jdoGetPersistenceManager()).internalMakeTransient(pc, state);
            }
        }
        // TODO We have omitted interface fields, interface array fields, object reference fields, object reference array fields

        return value;
    }

    /**
     * Method to throw and EndOfFetchPlanGraphException since we're at the end of a branch in the tree.
     * @param fieldNumber Number of the field
     * @return Object to return
     */
    protected Object endOfGraphOperation(int fieldNumber)
    {
        SingleValueFieldManager sfv = new SingleValueFieldManager();
        sm.provideFields(new int[]{fieldNumber}, sfv);
        Object value = sfv.fetchObjectField(fieldNumber);

        if (value != null && secondClassMutableFields[fieldNumber])
        {
            SCO sco;
            if (value instanceof SCO)
            {
                // SCO field so unset its owner
                sco = (SCO) value;
                sco.unsetOwner();
            }
        }

        return value;
    }
}