/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Erik Bengtson - removed exist() operation
2003 Andy Jefferson - added localiser
2003 Erik Bengtson - added new constructor for App ID
2003 Erik Bengtson - fixed loadDefaultFetchGroup to call jdoPostLoad
2003 Erik Bengtson - fixed evict to call jdoPreClear
2004 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - reordered methods to put in categories, split String utilities across to StringUtils.
2004 Andy Jefferson - added Lifecycle Listener callbacks
2004 Andy Jefferson - removed JDK 1.4 methods so that we support 1.3 also
2005 Martin Taal - Contrib of detach() method for "detachOnClose" functionality.
 	...
 **********************************************************************/
package org.jpox.state;

import java.io.PrintWriter;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.FetchPlan;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.jdo.Transaction;
import javax.jdo.spi.Detachable;
import javax.jdo.spi.JDOImplHelper;
import javax.jdo.spi.PersistenceCapable;
import javax.jdo.spi.StateManager;

import org.jpox.AbstractPersistenceManager;
import org.jpox.ClassLoaderResolver;
import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.FetchPlanImpl.FetchPlanForClass;
import org.jpox.api.ApiAdapter;
import org.jpox.cache.CachedPC;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.MetaDataManager;
import org.jpox.metadata.MetaDataUtils;
import org.jpox.sco.SCO;
import org.jpox.sco.SCOCollection;
import org.jpox.sco.SCOContainer;
import org.jpox.sco.SCOMap;
import org.jpox.sco.SCOUtils;
import org.jpox.sco.UnsetOwners;
import org.jpox.state.AbstractFetchFieldManager.EndOfFetchPlanGraphException;
import org.jpox.store.DatastoreClass;
import org.jpox.store.FieldValues;
import org.jpox.store.OID;
import org.jpox.store.OIDFactory;
import org.jpox.store.StoreManager;
import org.jpox.store.fieldmanager.FieldManager;
import org.jpox.store.fieldmanager.SingleTypeFieldManager;
import org.jpox.store.fieldmanager.SingleValueFieldManager;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.AIDUtils;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;
import org.jpox.util.TypeConversionHelper;

/**
 * Implementation of the StateManager.
 * Implemented here as one StateManager per Object so adds on functionality particular 
 * to each object. All PersistenceCapable objects will have a StateManager when they 
 * have had communication with the PersistenceManager. They will typically always have
 * an identity also. The exception to that is for embedded/serialised objects.
 * 
 * <H3>Embedded/Serialised Objects</H3>
 * An object that is being embedded/serialised in an owning object will NOT have an identity
 * unless the object is subject to a makePersistent() call also. When an object
 * is embedded/serialised and a field is changed, the field will NOT be marked as dirty (unless
 * it is also an object in its own right with an identity). When a field is changed
 * any owning objects are updated so that they can update their tables accordingly.
 * 
 * @version $Revision: 1.213.2.16 $
 */
final class StateManagerImpl implements org.jpox.StateManager
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    private static final JDOImplHelper HELPER;

    private static final SingleTypeFieldManager HOLLOWFIELDMANAGER = new SingleTypeFieldManager();

    /** the PersistenceManager for this StateManager */
    private PersistenceManager myPM;

    /** The PersistenceCapable instance managed by this StateManager */
    protected PersistenceCapable myPC;

    /** The object identity in the datastore */
    private Object myID;

    /** The object identity in the JVM */
    private Object myInternalID;

    /** The actual LifeCycleState for the PersistenceCapable instance */
    private LifeCycleState myLC;

    private byte jdoFlags;

    /**
     * an image of the PersistenceCapable instance at the moment the instance
     * is enlisted in the transaction 
     */
    private PersistenceCapable savedImage = null;

    /**
     * flags of the PersistenceCapable instance at the moment the instance
     * is enlisted in the transaction 
     */
    private byte savedFlags;

    /**
     * The loaded fields of the PersistenceCapable instance at the moment the instance
     * is enlisted in the transaction 
     */
    private boolean[] savedLoadedFields = null;

    /** the metadata for the class. */
    private AbstractClassMetaData cmd;

    /** the StoreManager. */
    private StoreManager srm;

    /** number of managed fields from this class plus inherited classes. */
    private int fieldCount;

    /** if the PersistenceCapable instance is dirty. */
    private boolean dirty = false;

    /** the dirty fields of the PersistenceCapable instance. */
    private boolean[] dirtyFields;

    /** the loaded fields of the PersistenceCapable instance. */
    private boolean[] loadedFields;
    
    /** the explicitly unloaded fields of the PersistenceCapable instance, via unloadField(). */
    private boolean[] unloadedFields;

    private boolean[] defaultFetchGroupFields;

    private boolean[] secondClassMutableFields;
    
    /** Field numbers of the non-primary key fields (inc superclasses). */
    private boolean[] nonPrimaryKeyFields;    

    /** Field numbers of all fields */
    private int[] allFieldNumbers;

    /** Field numbers of all non-primary-key fields */
    private int[] allNonPrimaryKeyFieldNumbers;

    /** Field numbers of all DFG fields. */
    private int[] defaultFetchGroupFieldNumbers;

    /** Field numbers of all second class mutable fields. */
    private int[] secondClassMutableFieldNumbers;

    /** if the PersistenceCapable instance is new and was flushed to the datastore. */
    private boolean flushedNew = false;

    /** version field for optimistic transactions */
    private Object myVersion;

    /** version field for optimistic transactions, after a insert/update but not yet committed. */
    private Object transactionalVersion;

    /** Current FieldManager. */
    private FieldManager currFM = null;

    /** Whether to restore values at StateManager. If true, overwrites the restore values at tx level. */
    private boolean restoreValues = false;

    /** Fetch plan for the class of the managed object. */
    private FetchPlanImpl.FetchPlanForClass myFP;

    /** loaded fields in detached instances. */
    private BitSet jdoLoadedFields;

    /** modified fields in detached instances. */
    private BitSet jdoModifiedFields;

    /** state for transitions of activities. */
    private ActivityState inserting = ActivityState.NONE;

    private boolean flushing = false;

    private boolean changingState = false;

    private boolean postLoadPending = false;

    private boolean disconnecting = false;

    private boolean deleting = false;

    /** Flag to indicate if the object is in the process of being detached. */
    private boolean detaching = false;

    private PersistenceCapable detachingPC = null;

    /** Whether the object is being attached. */
    private boolean attaching = false;

    /** Attached PC object, when this SM refers to a detached PC (and the SM is temporarily connected to get the values) */
    private PersistenceCapable attachedPC = null;

    /** List of StateManagers that we must notify when we have completed inserting our record. */
    private List insertionNotifyList = new ArrayList();

    /** List of fields that we must update when notified of the insertion of other records. */
    private Map insertionUpdateFieldList = new HashMap();

    /** The type of the managed object (0 = PC, 1 = embedded PC, 2 = embedded element, 3 = embedded key, 4 = embedded value. */
    private int pcObjectType = 0;

    /** List of owners when embedded. */
    private List embeddedOwners = null;

    /**
     * Map of external field values (added to our object table where we dont have relations to them - unidirectional).
     * This will include FK and order mappings for 1-N uni.
     */
    private HashMap externalFieldValuesByMapping = new HashMap();

    static
    {
        HELPER = (JDOImplHelper) AccessController.doPrivileged(new PrivilegedAction()
        {
            public Object run()
            {
                try
                {
                    return JDOImplHelper.getInstance();
                }
                catch (SecurityException e)
                {
                    throw new JDOFatalUserException(LOCALISER.msg("StateManager.SecurityProblem"), e);
                }
            }
        });
    }

    /** Handler for callback events. */
    CallbackHandler callback;

    // ------------------------------ Constructors -----------------------------

    /**
     * Constructs a state manager to manage a hollow instance having the
     * given object ID.  This constructor is used for creating new instances of
     * existing persistent objects (i.e. via {@link PersistenceManager#getObjectById(Object, boolean, boolean)}).
     * @param pm the persistence manager controlling this state manager.
     * @param pcClass the class of the new instance to be created.
     * @param id the JDO identity of the object.
     */
    public StateManagerImpl(PersistenceManager pm, Class pcClass, Object id)
    {
        myPM = pm;
        srm = myPM.getStoreManager();
        myID = id;
        myLC = myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.HOLLOW);
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;

        // TODO interfaces
        // in this case JPOX supports only one level for interfaces
        // in case of many levels, JPOX supports only one branch
        if (pm.getPMFContext().getTypeManager().isReferenceType(pcClass))
        {
            cmd = pm.getMetaDataManager().getMetaDataForImplementationOfReference(pcClass, null, pm.getClassLoaderResolver());
            // calling the classloaderresolver will make sure the class is initialized
            pcClass = pm.getClassLoaderResolver().classForName(cmd.getFullClassName(), pcClass.getClassLoader(), true);
        }
        else
        {
            try
            {
                // calling the classloaderresolver will make sure the class is initialized
                pcClass = pm.getClassLoaderResolver().classForName(pcClass.getName(), pcClass.getClassLoader(), true);
                cmd = pm.getMetaDataManager().getMetaDataForClass(pcClass, getPersistenceManager().getClassLoaderResolver());
            }
            catch (ClassNotResolvedException e)
            {
                throw new JDOFatalUserException(LOCALISER.msg("StateManager.ClassNotFoundError", pcClass.getName()));
            }
        }

        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass));
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

        initialize();

        // Create the PC object
        if (id instanceof OID || id == null)
        {
            // Create a new pc class
            myPC = HELPER.newInstance(pcClass, this);
        }
        else
        {
            // Create a new pc class, and copy the key class to fields
            myPC = HELPER.newInstance(pcClass, this, myID);
        }
    }
    
    /**
     * Constructs a state manager to manage a PersistenceCapable instance that will
     * be EMBEDDED/SERIALISED into another PersistenceCapable object. The instance will not be
     * assigned an identity in the process since it is a SCO.
     * @param pc The PersistenceCapable to manage (see copyPc also)
     * @param pm The persistence manager controlling this state manager.
     * @param copyPc Whether the SM should manage a copy of the passed PC or that one
     */
    public StateManagerImpl(PersistenceCapable pc, PersistenceManager pm, boolean copyPc)
    {
        pcObjectType = EMBEDDED_PC; // Default to an embedded PC object
        myPM = pm;
        srm = myPM.getStoreManager();
        myID = null; // It is embedded at this point so dont need an ID since we're not persisting it
        myLC = myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.P_NEW);
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;

        myPC = pc;
        replaceStateManager(this); // Set SM for myPC to be this

        cmd = pm.getMetaDataManager().getMetaDataForClass(pc.getClass().getName(), pm.getClassLoaderResolver());
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pc.getClass().getName()));
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

        initialize();

        if (copyPc)
        {
            // Create a new PC with the same field values
            PersistenceCapable pcCopy = myPC.jdoNewInstance(this);
            pcCopy.jdoCopyFields(myPC, allFieldNumbers);

            // Swap the managed PC to be the copy and not the input
            pcCopy.jdoReplaceStateManager(this);
            myPC = pcCopy;
            disconnectClone(pc);
        }

        // Mark all fields as loaded since we are using the passed PersistenceCapable
        for (int i=0;i<loadedFields.length;i++)
        {
            loadedFields[i] = true;
        }
    }

    /**
     * Constructs a state manager to manage a hollow or pclean instance having the given
     * FieldValues. This constructor is used for creating new instances of
     * existing persistent objects using application identity.
     * {@link PersistenceManager#getObjectByAID}
     * @param pm the persistence manager controlling this state manager.
     * @param pcClass the class of the new instance to be created.
     * @param fv the initial field values of the object.
     */
    public StateManagerImpl(PersistenceManager pm, Class pcClass, FieldValues fv)
    {
        myPM = pm;
        srm = myPM.getStoreManager();
        myLC = myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.HOLLOW);
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;

        //calling the classloaderresolver will make sure the class is initialized
        pcClass = pm.getClassLoaderResolver().classForName(pcClass.getName(), pcClass.getClassLoader(), true);
        cmd = pm.getMetaDataManager().getMetaDataForClass(pcClass, getPersistenceManager().getClassLoaderResolver());
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass));
        }
        if (cmd.getIdentityType() != IdentityType.APPLICATION)
        {
            throw new JDOFatalUserException("This constructor is only for objects using application identity.");
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);
        initialize();
        myPC = HELPER.newInstance(pcClass, this);
        if (myPC == null)
        {
            if (!HELPER.getRegisteredClasses().contains(pcClass))
            {
                // probably never will get here, as JDOImplHelper.newInstance() internally already throws
                // JDOFatalUserException when class is not registered 
                throw new JDOFatalUserException(LOCALISER.msg("StateManager.ClassNotRegistered", pcClass.getName()));
            }
            else
            {
                // Provide advisory information since we can't create an instance of this class, so maybe they
                // have an error in their data ?
                throw new JDOFatalUserException(LOCALISER.msg("StateManager.ClassNotConstructable", pcClass.getName()));
            }
        }

        loadFieldValues(fv); // as a minimum the PK fields are loaded here

        // Create the ID now that we have the PK fields loaded
        myID = myPC.jdoNewObjectIdInstance();
        if (!cmd.usesSingleFieldIdentityClass())
        {
            myPC.jdoCopyKeyFieldsToObjectId(myID);
        }
    }

    /**
     * Look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     * @param fv the initial field values of the object.
     */
    public void checkInheritance(FieldValues fv)
    {
        // Inheritance case, check the level of the instance
        ClassLoaderResolver clr = myPM.getClassLoaderResolver();
        String className = srm.getClassNameForObjectID(myID, clr, myPM);
        if (className == null)
        {
            // className is null when id class exists, and object has been validated and doesn't exist.
            throw new JDOObjectNotFoundException(LOCALISER.msg("StateManager.ObjectDoesntExist"), myID);
        }
        else if (!cmd.getFullClassName().equals(className))
        {
            Class pcClass; 
            try
            {
                //load the class and make sure the class is initialized
                pcClass = getPersistenceManager().getClassLoaderResolver().classForName(className, myID.getClass().getClassLoader(), true);
                cmd = myPM.getMetaDataManager().getMetaDataForClass(pcClass, getPersistenceManager().getClassLoaderResolver());
            }
            catch (ClassNotResolvedException e)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("StateManager.GetObjectByIdClassNotFound", myID));
                throw new JDOUserException(LOCALISER.msg("StateManager.GetObjectByIdClassNotFound", myID), e);
            }
            if (cmd == null)
            {
                throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass));
            }
            if (cmd.getIdentityType() != IdentityType.APPLICATION)
            {
                throw new JDOFatalUserException("This constructor is only for objects using application identity.");
            }
            
            myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);
            
            initialize();
            
            myPC = HELPER.newInstance(pcClass, this);
            if (myPC == null)
            {
                throw new JDOFatalUserException(LOCALISER.msg("StateManager.ClassNotRegistered", cmd.getFullClassName()));
            }
            
            // Note that this will mean the fields are loaded twice (loaded earlier in this method)
            // and also that postLoad will be called twice
            loadFieldValues(fv);
            
            myID = myPC.jdoNewObjectIdInstance();
            if (!cmd.usesSingleFieldIdentityClass())
            {
                myPC.jdoCopyKeyFieldsToObjectId(myID);
            }                
        }
    }

    /**
     * Constructs a state manager to manage a recently populated hollow instance having the
     * given object ID and the given field values.  This constructor is used for
     * creating new instances of persistent objects obtained e.g. via a Query or
     * backed by a view (i.e. via {@link PersistenceManager#getObjectById(Object,FieldValues)}).
     * @param pm the persistence manager controlling this state manager.
     * @param pcClass the class of the new instance to be created.
     * @param id the JDO identity of the object.
     * @param fv the initial field values of the object.
     */
    public StateManagerImpl(PersistenceManager pm, Class pcClass, Object id, FieldValues fv)
    {
        this(pm, pcClass, id);

        // Fetch the required fields using any defined fetch plan
        loadFieldValues(fv);
    }

    /**
     * Constructs a state manager to manage a transient instance that is 
     * becoming newly persistent.  A new object ID for the
     * instance is obtained from the store manager and the object is inserted
     * in the data store.
     * <p>This constructor is used for assigning state managers to existing
     * instances that are transitioning to a persistent state (i.e. via
     * {@link PersistenceManager#makePersistent(Object)}).
     *
     * @param pm the persistence manager controlling this state manager.
     * @param pc the instance being make persistent.
     * @param preInsertChanges Any changes to make before inserting
     */
    public StateManagerImpl(PersistenceManager pm, PersistenceCapable pc, FieldValues preInsertChanges)
    {
        Class pcClass = pc.getClass();

        myPM = pm;
        srm = myPM.getStoreManager();
        myPC = pc;
        myLC = myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.P_NEW);
        jdoFlags = PersistenceCapable.READ_OK;

        if (pm.getPMFContext().getTypeManager().isReferenceType(pcClass))
        {
            cmd = pm.getMetaDataManager().getMetaDataForImplementationOfReference(pcClass, null, pm.getClassLoaderResolver());
        }
        else
        {
            //calling the classloaderresolver will make sure the class is initialized
            cmd = pm.getMetaDataManager().getMetaDataForClass(pm.getClassLoaderResolver().classForName(pcClass.getName(), pcClass.getClassLoader(), true), getPersistenceManager().getClassLoaderResolver());
        }
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass.getName()));
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

        initialize();

        for (int i = 0; i < fieldCount; ++i)
        {
            loadedFields[i] = true;
        }

        replaceStateManager(this);
        myPC.jdoReplaceFlags();

        saveFields();

        // Populate all fields that have "value-strategy" and are not datastore populated
        populateStrategyFields();

        if (preInsertChanges != null)
        {
            // Apply any pre-insert field updates
            preInsertChanges.fetchFields(this);
        }

        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            //load key fields from Application Id instance to PC instance

            //if a primary key field is of type PersistenceCapable, it must first be persistent
            for (int fieldNumber = 0; fieldNumber < allFieldNumbers.length; fieldNumber++)
            {
                AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
                if (fmd.isPrimaryKey())
                {
                    if (pm.getMetaDataManager().getMetaDataForClass(fmd.getType(), getPersistenceManager().getClassLoaderResolver()) != null)
                    {
                        FieldManager oldFM = currFM;
                        currFM = new SingleValueFieldManager();
                        myPC.jdoProvideField(fieldNumber);
                        PersistenceCapable pkFieldPC = (PersistenceCapable) ((SingleValueFieldManager) currFM).fetchObjectField(fieldNumber);
                        if (pkFieldPC == null)
                        {
                            throw new JDOUserException(LOCALISER.msg("StateManager.PrimaryKeyFieldIsNull",fmd.getFullFieldName()));
                        }
                        if (!pkFieldPC.jdoIsPersistent())
                        {
                            myPM.makePersistentInternal(pkFieldPC, null);
                        }
                        currFM = oldFM;
                    }
                }
            }
        }

        /* Set the identity
         * This must come after the above block, because in identifying relationships
         * the PK FK associations must be persisted before, otherwise we
         * don't have an id assigned to the PK FK associations
         */        
        setIdentity();

        myPM.enlistInTransaction(this);
    }

    /**
     * Constructs a state manager to manage a Transactional Transient instance.
     * A new object ID for the instance is obtained from the store manager and
     * the object is inserted in the data store.
     * <p>
     * This constructor is used for assigning state managers to Transient
     * instances that are transitioning to a transient clean state (i.e. via
     * {@link PersistenceManager#makeTransactional(Object)}).
     * @param pm the persistence manager controlling this state manager.
     * @param pc the instance being make persistent.
     */
    public StateManagerImpl(PersistenceManager pm, PersistenceCapable pc)
    {
        Class pcClass = pc.getClass();

        myPM = pm;
        srm = myPM.getStoreManager();
        myPC = pc;
        myLC = null;
        jdoFlags = PersistenceCapable.READ_OK;

        // Retrieve the MetaData for the class
        if (pm.getPMFContext().getTypeManager().isReferenceType(pcClass))
        {
            cmd = pm.getMetaDataManager().getMetaDataForImplementationOfReference(pcClass, pc, pm.getClassLoaderResolver());
        }
        else
        {
            //calling the classloaderresolver will make sure the class is initialized
            cmd = pm.getMetaDataManager().getMetaDataForClass(pm.getClassLoaderResolver().classForName(pcClass.getName(), pcClass.getClassLoader(), true), getPersistenceManager().getClassLoaderResolver());
        }
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass.getName()));
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

        initialize();

        for (int i = 0; i < fieldCount; ++i)
        {
            loadedFields[i] = true;
        }

        myPC.jdoReplaceFlags();

        // Populate all fields that have "value-strategy" and are not datastore populated
        populateStrategyFields();

        // Set the identity
        setIdentity();

        // for non transactional read, tx might be not active
        // TODO add verification if is non transactional read = true
        if (myPM.currentTransaction().isActive())
        {
            myPM.enlistInTransaction(this);
        }
    }

    /**
     * Constructor for creating SM instances to manage PersistenceCapable objects in detached state
     * @param pm PersistenceManager
     * @param apiAdapter the ApiAdapter
     * @param pc the detach object.
     * @param cmd the AbstractClassMetaData
     * @param id the JDO identity of the object.
     * @param version the detached version
     * @since 1.1
     */
    public StateManagerImpl(PersistenceManager pm, ApiAdapter apiAdapter, PersistenceCapable pc, 
    		AbstractClassMetaData cmd, Object id, Object version)
    {
        this.cmd = cmd;
        this.myID = id;
        this.myPC = pc;
        this.myPM = pm;
        this.myFP = null;
        this.srm = null;
        setVersion(version);

        // This lifecycle state is not always correct. It is certainly "detached"
        // but we dont know if it is CLEAN or DIRTY. We need this setting here since all objects
        // have a lifecycle state and other methods e.g isPersistent() depend on it.
        this.myLC = apiAdapter.getLifeCycleState(LifeCycleState.DETACHED_CLEAN);

        this.myPC.jdoReplaceFlags();
        initialize();
    }

    /**
     * Constructor to create a StateManager for a PersistenceCapable object, assigning the
     * specified id to the object. This is used when getting objects out of the L2 Cache,
     * where they have no StateManager assigned, and returning them as associated with a
     * particular PM.
     * @param pm PersistenceManager managing this object
     * @param id Id to assign to the PersistenceCapable object
     * @param pc The PersistenceCapable object
     * @param loaded The list of loaded fields (when put in the cache)
     */
    public StateManagerImpl(PersistenceManager pm, Object id, PersistenceCapable pc, boolean loaded[])
    {
        // Create a new copy of the input object type
        this(pm, pc.getClass(), id);

        myLC = myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.P_CLEAN);
        jdoFlags = PersistenceCapable.READ_OK;

        synchronized (pc)
        {
        	// Synchronise the L2 cached object while we connect a StateManager to get its fields
            // Copy the fields from the input object onto our copy that will be returned to the user
            pc.jdoReplaceStateManager(this);
            myPC.jdoCopyFields(pc, allFieldNumbers);

            // Reinstate the original loadedFields list
            for (int i=0; i<loadedFields.length; i++)
            {
                loadedFields[i] = loaded[i];
            }

            // Mark all relationships as not yet loaded since we can't maintain those whilst cached.
            for (int i=0; i<cmd.getPersistenceCapableFieldNumbers().length; i++)
            {
                loadedFields[cmd.getPersistenceCapableFieldNumbers()[i]] = false;
            }
            for (int i=0; i<secondClassMutableFieldNumbers.length; i++)
            {
                loadedFields[secondClassMutableFieldNumbers[i]] = false;
            }

            // Disconnect the input object
            disconnectClone(pc);
		}

        if (isFetchPlanLoaded())
        {
            // Should we call postLoad when getting the object out of the L2 cache ? Seems incorrect IMHO
            postLoad();
        }
    }

    /**
     * Convenience method to populate all fields in the PC object that have "value-strategy" specified
     * and that aren't datastore attributed. This applies not just to PK fields (where it is most
     * useful to use value-strategy) but also to any other field. Fields are populated only if they are null
     * This is called once on a PC object, when makePersistent is called.
     */
    private void populateStrategyFields()
    {
        int totalFieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();
        DatastoreClass table = null;
        if (!cmd.isEmbeddedOnly())
        {
            table = srm.getDatastoreClass(cmd.getFullClassName(), myPM.getClassLoaderResolver());
        }

        for (int fieldNumber =0; fieldNumber < totalFieldCount; fieldNumber++)
        {
            AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
            IdentityStrategy strategy = fmd.getValueStrategy();

            // Check for the strategy, and if it is a datastore attributed strategy
            if (strategy != null && !srm.isStrategyDatastoreAttributed(strategy))
            {
                // Assign the strategy value where required.
                // Default JDO2 behaviour is to always provide a strategy value when it is marked as using a strategy
                boolean applyStrategy = true;
                if (!fmd.getType().isPrimitive() && strategy != null &&
                    fmd.hasExtension("strategy-when-notnull") &&
                    fmd.getValueForExtension("strategy-when-notnull").equalsIgnoreCase("false") &&
                    this.provideField(fieldNumber) != null)
                {
                    // JPOX allows an extension to only provide a value-strategy value where the field is null at persistence.
                    applyStrategy = false;
                }

                if (applyStrategy)
                {
                    // Apply a strategy value for this field
	                DatastoreClass fieldTable = table.getBaseDatastoreClassWithField(fmd);
	                Object obj = srm.getStrategyValue(myPM, fieldTable, cmd, fieldNumber, strategy, fmd.getSequence(), fmd.getExtensions());
	                this.replaceField(fieldNumber, obj);
                }
            }
        }
    }

    /**
     * Convenience method to load the passed field values.
     * Loads the fields using any required fetch plan and calls jdoPostLoad() as appropriate.
     * @param fv Field Values to load (including any fetch plan to use when loading)
     */
    public void loadFieldValues(FieldValues fv)
    {
        // Fetch the required fields using any defined fetch plan
        FetchPlanImpl.FetchPlanForClass origFetchPlan = myFP;
        FetchPlanImpl loadFetchPlan = fv.getFetchPlanForLoading();
        if (loadFetchPlan != null)
        {
            myFP = loadFetchPlan.manageFetchPlanForClass(cmd);
        }

        boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);

        fv.fetchFields(this);

        if (callPostLoad && isFetchPlanLoaded())
        {
            postLoad();
        }

        // Reinstate the original (PM) fetch plan
        myFP = origFetchPlan;
    }

    /**
     * Utility to set the identity for the PersistenceCapable object.
     * Will only create the id instance if it is not attributed (in some way) by the datastore.
     * So, for example, autoassign cases will not gain their identity here. In such cases the identity
     * is only created when they are inserted into the datastore.
     */
    private void setIdentity()
    {
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            boolean isObjectIDDatastoreAttributed = false;
            int totalFieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();
            for (int fieldNumber =0; fieldNumber < totalFieldCount; fieldNumber++)
            {
                AbstractPropertyMetaData fmd=cmd.getManagedFieldAbsolute(fieldNumber);
                if (fmd.isPrimaryKey())
                {
                    if (srm.isStrategyDatastoreAttributed(fmd.getValueStrategy()))
                    {
                        isObjectIDDatastoreAttributed = true;
                        break;
                    }
                    else if (cmd.usesSingleFieldIdentityClass())
                    {
                        if (this.provideField(fieldNumber) == null)
                        {
                            // SingleFieldIdentity field has not had its value set (by user, or by value-strategy)
                            throw new JDOFatalUserException(LOCALISER.msg("StateManager.SingleFieldIdentityPKFieldIsNull", 
                                cmd.getFullClassName(), fmd.getName()));
                        }
                    }
                }
            }

            if (!isObjectIDDatastoreAttributed)
            {
                // Not using the datastore to attribute the id, so set it now
                myID = srm.newObjectID(myPM, cmd.getFullClassName(), myPC);
            }
        }
        else
        {
            if (cmd.getIdentityMetaData() == null || !srm.isStrategyDatastoreAttributed(cmd.getIdentityMetaData().getStrategyValue()))
            {
                myID = srm.newObjectID(myPM, cmd.getFullClassName(), myPC);
            }
        }

        if (myInternalID != myID && myID != null)
        {
            // Update the id with the PM if it is changing
            myPM.replaceObjectId(myPC, myInternalID, myID);
        }
    }

    /**
     * Accessor for a L2-Cacheable form of this PersistenceCapable object.
     * @return The L2 cacheable object
     */
    public CachedPC getL2CacheableObject()
    {
        PersistenceCapable pcCopy = myPC.jdoNewInstance(this, myPC.jdoGetObjectId());

        // Make a copy of the field values from the original object - basic fields + PC fields + SCO fields (omit SCO containers)
        int[] containerFieldNumbers = cmd.getSecondClassContainerFieldNumbers();
        int[] noncontainerFieldNumbers = new int[allFieldNumbers.length - containerFieldNumbers.length];
        boolean[] l2loadedFields = new boolean[allFieldNumbers.length];
        int nonNum = 0;
        for (int i=0;i<allFieldNumbers.length;i++)
        {
            boolean noncontainer = true;
            for (int j=0;j<containerFieldNumbers.length;j++)
            {
                if (containerFieldNumbers[j] == allFieldNumbers[i])
                {
                    noncontainer = false;
                    break;
                }
            }
            if (noncontainer)
            {
                noncontainerFieldNumbers[nonNum++] = allFieldNumbers[i];
                l2loadedFields[i] = (noncontainer && loadedFields[i]);
            }
            else
            {
                l2loadedFields[i] = false;
            }
        }
        pcCopy.jdoCopyFields(myPC, noncontainerFieldNumbers);

        // Reset jdoFlags in the copy to PersistenceCapable.READ_WRITE_OK and clear its state manager.
        pcCopy.jdoReplaceFlags();
        pcCopy.jdoReplaceStateManager(null);

        return new CachedPC(pcCopy, l2loadedFields);
    }

    /**
     * Utility method to initialise the state information.
     **/
    private void initialize()
    {
        fieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();

        dirtyFields = new boolean[fieldCount];
        loadedFields = new boolean[fieldCount];
        unloadedFields = new boolean[fieldCount];
        defaultFetchGroupFields = cmd.getDefaultFetchGroupFieldFlags();
        secondClassMutableFields = cmd.getSecondClassMutableFieldFlags();
        nonPrimaryKeyFields = cmd.getNonPrimaryKeyFieldFlags();

        allFieldNumbers = cmd.getAllFieldNumbers();
        allNonPrimaryKeyFieldNumbers = cmd.getNonPrimaryKeyFieldNumbers();

        defaultFetchGroupFieldNumbers = cmd.getDefaultFetchGroupFieldNumbers();
        secondClassMutableFieldNumbers = cmd.getSecondClassMutableFieldNumbers();
        
        if (myPM != null)
        {
            callback = myPM.getCallbackHandler();
        }
    }

    /**
     * Utility to update our object to use a different state manager.
     * @param sm The new state manager.
     **/
    private void replaceStateManager(StateManager sm)
    {
        try
        {
            myPC.jdoReplaceStateManager(sm);
        }
        catch (SecurityException e)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.SecurityProblem"), e);
        }
    }

    public void enlistInTransaction()
    {
        myPM.enlistInTransaction(this);

        if (jdoFlags == PersistenceCapable.LOAD_REQUIRED && isDFGLoaded())
        {
            /*
             * A transactional object whose DFG fields are loaded does not need
             * to contact us in order to read those fields.
             */
            jdoFlags = PersistenceCapable.READ_OK;
            myPC.jdoReplaceFlags();
        }
    }

    public void evictFromTransaction()
    {
        myPM.evictFromTransaction(this);

        /*
         * A non-transactional object needs to contact us on any field read no
         * matter what fields are loaded.
         */
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;
        myPC.jdoReplaceFlags();
    }

    /**
     * Method to save all fields of the object.
     **/
    public void saveFields()
    {
        savedImage = myPC.jdoNewInstance(this);
        savedImage.jdoCopyFields(myPC, allFieldNumbers);
        savedFlags = jdoFlags;
        savedLoadedFields = (boolean[])loadedFields.clone();
    }

    /**
     * Method to restore all fields of the object.
     **/
    public void restoreFields()
    {
        if (savedImage != null)
        {
            loadedFields = savedLoadedFields;
            jdoFlags = savedFlags;
            myPC.jdoReplaceFlags();
            myPC.jdoCopyFields(savedImage, allFieldNumbers);

            clearDirtyFlags();
            clearSavedFields();
        }
    }

    /**
     * Method to clear all fields of the object.
     **/
    public void clearFields()
    {
        clearFieldsByNumbers(allFieldNumbers);
    }

    /**
     * Method to clear all fields that are not part of the primary key of the object.
     **/
    public void clearNonPrimaryKeyFields()
    {
        clearFieldsByNumbers(allNonPrimaryKeyFieldNumbers);
    }
    
    private void clearFieldsByNumbers(int[] fieldNumbers)
    {
        try
        {
            callback.preClear(myPC);
        }
        finally
        {
            replaceFields(fieldNumbers, HOLLOWFIELDMANAGER);
            clearLoadedFlags();
            clearDirtyFlags();
            callback.postClear(myPC);
        }
    }

    /**
     * Method to clear all saved fields on the object.
     **/
    public void clearSavedFields()
    {
        savedImage = null;
        savedFlags = 0;
        savedLoadedFields = null;
    }

    /**
     * Method to clear all loaded flags on the object.
     **/
    public void clearLoadedFlags()
    {
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;
        myPC.jdoReplaceFlags();
        clearFlags(loadedFields);
    }

    /**
     * Method to clear all dirty flags on the object.
     **/
    private void clearDirtyFlags()
    {
        dirty = false;
        clearFlags(dirtyFields);
    }
    
    /**
     * Method to clear all dirty flags on the object.
     * @param fields the fields to clear
     **/
    private void clearDirtyFlags(int[] fields)
    {
        dirty = false;
        clearFlags(dirtyFields,fields);
    }

    /**
     * Accessor for the names of the fields that are dirty.
     * @return Names of the dirty fields
     */
    public String[] getDirtyFieldNames()
    {
        int[] dirtyFieldNumbers = getFlagsSetTo(dirtyFields, true);
        if (dirtyFieldNumbers != null && dirtyFieldNumbers.length > 0)
        {
            String[] dirtyFieldNames = new String[dirtyFieldNumbers.length];
            for (int i=0;i<dirtyFieldNumbers.length;i++)
            {
                dirtyFieldNames[i] = cmd.getManagedFieldAbsolute(dirtyFieldNumbers[i]).getName();
            }
            return dirtyFieldNames;
        }
        return null;
    }

    /**
     * Accessor for the names of the fields that are loaded.
     * @return Names of the loaded fields
     */
    public String[] getLoadedFieldNames()
    {
        int[] loadedFieldNumbers = getFlagsSetTo(loadedFields, true);
        if (loadedFieldNumbers != null && loadedFieldNumbers.length > 0)
        {
            String[] loadedFieldNames = new String[loadedFieldNumbers.length];
            for (int i=0;i<loadedFieldNumbers.length;i++)
            {
                loadedFieldNames[i] = cmd.getManagedFieldAbsolute(loadedFieldNumbers[i]).getName();
            }
            return loadedFieldNames;
        }
        return null;
    }

    /**
     * Marks the given field dirty.
     * @param field The no of field to mark as dirty. 
     */
    public void makeDirty(int field)
    {
        boolean wasDirty = preWriteField(field);
        postWriteField(wasDirty);
    }

    /**
     * Mark the associated PersistenceCapable field dirty.
     *
     * @param pc the calling PersistenceCapable instance
     * @param fieldName the name of the field
     */
    public void makeDirty(PersistenceCapable pc, String fieldName)
    {
        if (!disconnectClone(pc))
        {
            int fieldNumber = cmd.getFieldNumberAbsolute(fieldName);

            if (fieldNumber == -1)
            {
                throw new JDOUserException(LOCALISER.msg("StateManager.InvalidFieldForClass", fieldName, cmd.getFullClassName()));
            }
            
            makeDirty(fieldNumber);
        }
    }

    // -------------------------- Accessor Methods -----------------------------

    /**
     * Accessor for the StoreManager used for this object.
     * @return The StoreManager.
     **/
    public StoreManager getStoreManager()
    {
        return srm;
    }

    /**
     * Accessor for the ClassMetaData for this object.
     * @return The ClassMetaData.
     **/
    public AbstractClassMetaData getClassMetaData()
    {
        return cmd;
    }

    /**
     * Accessor for the MetaDataManager to use for this object.
     * Simply a wrapper accessor method. 
     * @return The MetaDataManager.
     **/
    public MetaDataManager getMetaDataManager()
    {
        return myPM.getMetaDataManager();
    }

    /**
     * Accessor for the PersistenceManager that owns this instance.
     * @param pc The PersistenceCapable instance
     * @return The PersistenceManager that owns this instance
     */
    public javax.jdo.PersistenceManager getPersistenceManager(PersistenceCapable pc)
    {
        //in identifying relationships, jdoCopyKeyFieldsFromId will call
        //this method, and at this moment, myPC in statemanager is null
        // Currently AbstractPersistenceManager.java putObjectInCache prevents any identifying relation object being put in L2

        //if not identifying relationship, do the default check of disconnectClone:
        //"this.disconnectClone(pc)"
        if (myPC != null && this.disconnectClone(pc))
        {
            return null;
        }
        else if (myPM == null)
        {
            return null;
        }
        else
        {
            myPM.hereIsStateManager(this, myPC);
            return myPM.getPMHandle();
        }
    }

    /**
     * Accessor for the PersistenceManager for this object.
     * @return The PersistenceManager.
     **/
    public PersistenceManager getPersistenceManager()
    {
        if (myPM == null)
        {
            return null;
        }
        else
        {
            return myPM.getPMHandle();
        }
    }

    /**
     * Accessor for the Persistent Capable object.
     * @return The PersistentCapable object
     **/
    public PersistenceCapable getObject()
    {
        return myPC;
    }

    /**
     * Accessor for the Restore Values flag 
     * @return Whether to restore values
     */
    public boolean isRestoreValues()
    {
        return restoreValues;
    }

    /**
     * Mutator for the Restore Values flag 
     * @param restore_values Whether to restore values
     */
    private void setRestoreValues(boolean restore_values)
    {
        restoreValues = restore_values;
    }

    /**
     * Return the object representing the JDO identity of the calling instance.
     *
     * According to the JDO specification, if the JDO identity is being changed
     * in the current transaction, this method returns the JDO identify as of
     * the beginning of the transaction.
     *
     * @param pc the calling PersistenceCapable instance
     * @return the object representing the JDO identity of the calling instance
     */
    public Object getObjectId(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return null;
        }
        else
        {
            return getExternalObjectId(pc);
        }
    }

    /**
     * Accessor for the internal object id of the object we are managing. 
     * @return The internal object id 
     */
    public Object getInternalObjectId()
    {
        if( myID != null)
        {
            return myID;
        }
        else if (myInternalID == null)
        {
            myInternalID = super.toString();
            return myInternalID;
        }
        else
        {
            return myInternalID;
        }
    }

	/**
     * If the id is obtained after inserting the object into the database, set
     * new a new id for persistent classes (for example, increment).
     * @param id the id received from the datastore
     */
    public void setPostStoreNewObjectId(Object id)
    {
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            try
            {
                myID = null;

                int fieldNumber = 0;
                for (fieldNumber =0; fieldNumber < fieldCount; fieldNumber++)
                {
                    AbstractPropertyMetaData fmd=cmd.getManagedFieldAbsolute(fieldNumber);
                    if (fmd.isPrimaryKey() &&
                        srm.isStrategyDatastoreAttributed(fmd.getValueStrategy()))
                    {
                        //replace the value of the id, but before convert the value to the field type if needed
                        replaceField(fieldNumber,TypeConversionHelper.convertTo(id,fmd.getType()));
                    }
                }
            }
            catch (Exception e)
            {
                JPOXLogger.RDBMS.error(e);
            }
            finally
            {
                myID = AIDUtils.getNewObjectId(getObject(),cmd);
            }
        }
        else
        {
            myID = OIDFactory.getInstance(cmd.getFullClassName(), id);
        }

        if (myInternalID != myID && myID != null)
        {
            // Update the id with the PM if it is set and is changing
            myPM.replaceObjectId(myPC, myInternalID, myID);
        }
    }

    /**
     * Return an object id that the user can use.
     * @param pc the PersistenceCapable object
     * @return the object id
     */
    public Object getExternalObjectId(PersistenceCapable pc)
    {
        // Note that we always create a new application identity since it is mutable and we can't allow
        // the user to change it. The only drawback of this is that we *must* have the relevant fields
        // set when this method is called, so that the identity can be generated.
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            if (!flushing)
            {
                // Flush any datastore changes so that we have all necessary fields populated
                // only if the datastore generates the field numbers
                if (!flushedNew && inserting == ActivityState.NONE && myLC.stateType() == LifeCycleState.P_NEW )
                {
                    int[] pkFieldNumbers = cmd.getPrimaryKeyFieldNumbers();
                    for (int i = 0; i < pkFieldNumbers.length; i++)
                    {
                        AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(i);
                        if (srm.isStrategyDatastoreAttributed(fmd.getValueStrategy()))
                        {
                            flush();
                            break;
                        }
                    }
                }
            }

            if (cmd.usesSingleFieldIdentityClass())
            {
                //SingleFieldIdentity classes are immutable.
                //Note, the instances of SingleFieldIdentity can be changed by the user using reflection,
                //but this is not allowed by the JDO spec
                return myID;
            }
            return AIDUtils.getNewObjectId(myPC, cmd);
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            if (!flushing)
            {
                // Flush any datastore changes so that myID is set by the time we return
                if (!flushedNew && inserting == ActivityState.NONE && myLC.stateType() == LifeCycleState.P_NEW )
                {
                    if (srm.isStrategyDatastoreAttributed(cmd.getIdentityMetaData().getStrategyValue()))
                    {
                        flush();
                    }
                }
            }
        }

        return myID;
    }

    /**
     * Replace the current value of jdoStateManager.
     *
     * <P>This method is called by the PersistenceCapable whenever
     * jdoReplaceStateManager is called and there is already
     * an owning StateManager.  This is a security precaution
     * to ensure that the owning StateManager is the only
     * source of any change to its reference in the PersistenceCapable.</p>
     *
     * @return the new value for the jdoStateManager
     * @param pc the calling PersistenceCapable instance
     * @param sm the proposed new value for the jdoStateManager
     */
    public StateManager replacingStateManager(PersistenceCapable pc, StateManager sm)
    {
        if (myLC == null)
        {
            throw new JDOFatalInternalException("Null LifeCycleState");
        }

        if (myLC.stateType() == LifeCycleState.DETACHED_CLEAN)
        {
            return sm;
        }

        if (pc == myPC)
        {
            //TODO check if we are really in transition to a transient instance
            if (sm == null)
            {
                return null;
            }
            if (sm == this)
            {
                return this;
            }

            if (this.myPM.getPMHandle() == ((StateManagerImpl) sm).myPM.getPMHandle())
            {
                // This is a race condition when makePersistent or
                // makeTransactional is called on the same PC instance for the
                // same PM. It has been already set to this SM - just 
                // disconnect the other one. Return this SM so it won't be
                // replaced.
                ((StateManagerImpl) sm).disconnect();
                return this;
            }

            if (sm != null)
            {
                throw new JDOUserException(LOCALISER.msg("StateManager.StateManagerChangeError"));
            }
            if (!disconnecting)
            {
                throw new JDOUserException(LOCALISER.msg("StateManager.StateManagerDisconnectError"));
            }

            if (JPOXLogger.JDO.isDebugEnabled())
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("StateManager.ClearingStateManager", StringUtils.toJVMIDString(pc)));
            }

            return null;
        }
        else if (pc == savedImage)
        {
            return null;
        }
        else
        {
            return sm;
        }
    }

    /**
     * Return the object representing the JDO identity
     * of the calling instance.  If the JDO identity is being changed in
     * the current transaction, this method returns the current identity as
     * changed in the transaction.
     *
     * @param pc the calling PersistenceCapable instance
     * @return the object representing the JDO identity of the calling instance
     */
    public Object getTransactionalObjectId(PersistenceCapable pc)
    {
        return getObjectId(pc);
    }

    /**
     * Tests whether this object is dirty.
     *
     * Instances that have been modified, deleted, or newly
     * made persistent in the current transaction return true.
     *
     * <P>Transient nontransactional instances return false (JDO spec), but the
     * JPOX implementation does not currently support the transient
     * transactional state.
     *
     * @see PersistenceCapable#jdoMakeDirty(String fieldName)
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance has been modified in current transaction.
     */
    public boolean isDirty(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isDirty();
        }
    }

    /**
     * Tests whether this object is transactional.
     *
     * Instances that respect transaction boundaries return true.  These
     * instances include transient instances made transactional as a result of
     * being the target of a makeTransactional method call; newly made
     * persistent or deleted persistent instances; persistent instances read
     * in data store transactions; and persistent instances modified in
     * optimistic transactions.
     * <P>
     * Transient nontransactional instances return false.
     *
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance is transactional.
     */
    public boolean isTransactional(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isTransactional();
        }
    }

    /**
     * Tests whether this object is persistent.
     * Instances whose state is stored in the data store return true.
     * <P>
     * Transient instances return false.
     *
     * @see PersistenceManager#makePersistent(Object pc)
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance is persistent.
     */
    public boolean isPersistent(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isPersistent();
        }
    }

    /**
     * Tests whether this object has been newly made persistent.
     * Instances that have been made persistent in the current transaction
     * return true.
     * <P>
     * Transient instances return false.
     *
     * @see PersistenceManager#makePersistent(Object pc)
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance was made persistent
     * in the current transaction.
     */
    public boolean isNew(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isNew();
        }
    }

    /**
     * Tests whether this object has been deleted.
     * Instances that have been deleted in the current transaction return true.
     *
     * <P>Transient instances return false.
     *
     * @see PersistenceManager#deletePersistent(Object pc)
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance was deleted
     * in the current transaction.
     */
    public boolean isDeleted(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isDeleted();
        }
    }

    // --------------------------- Load Field Methods --------------------------

    /**
     * Accessor for whether the current fetch plan fields are loaded.
     * @return Whether the fetch plan fields are all loaded.
     */
    private boolean isFetchPlanLoaded()
    {
        int[] fpFields = myFP.getFieldsInActualFetchPlan();
        for (int i=0; i<fpFields.length; ++i)
        {
            if (!loadedFields[fpFields[i]])
            {
                return false;
            }
        }
        return true;
    }

    /**
     * Accessor for whether the default fetch group fields are loaded.
     * @return Whether the default fetch group are loaded.
     **/
    private boolean isDFGLoaded()
    {
        for (int i=0; i<defaultFetchGroupFieldNumbers.length; ++i)
        {
            if (!loadedFields[defaultFetchGroupFieldNumbers[i]])
            {
                return false;
            }
        }
        return true;
    }

    /**
     * Fetchs from the database all fields in the default fetch group not
     * already loaded. Assumes myLC is already set to the state we're destined
     * for.
     */
    private void loadDFGFields()
    {
        int[] fieldNumbers = getFlagsSetTo(loadedFields, defaultFetchGroupFieldNumbers, false);
        if (fieldNumbers != null)
        {
            boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);
            srm.fetch(this, fieldNumbers);
            if (callPostLoad)
            {
                postLoad();
            }

            // Update the L2 cache to have this object (more loaded fields)
            myPM.putObjectIntoCache(this, false, true);
        }
    }

    /**
     * Fetchs from the database all SCO fields that are not containers that aren't already loaded.
     */
    private void loadSCONonContainerFields()
    {
        int[] noncontainerFieldNumbers = cmd.getSecondClassNonContainerFieldNumbers();
        int[] fieldNumbers = getFlagsSetTo(loadedFields, noncontainerFieldNumbers, false);
        if (fieldNumbers != null)
        {
            srm.fetch(this, fieldNumbers);

            // We currently dont call postLoad here since this is only called as part of attaching an object
            // and consequently we just read to get the current (attached) values. Could add a flag on input to allow postLoad

            // Update the L2 cache to have this object (more loaded fields)
            myPM.putObjectIntoCache(this, false, true);
        }
    }

    /**
     * Fetch the specified fields from the database.
     * Do NOT call with a default fetch group field.
     * @param fieldNumbers the numbers of the field(s) to fetch.
     */
    private void loadNonDFGFields(int[] fieldNumbers)
    {
        if (srm == null)
        {
            // Nothing to do since we're detached
            return;
        }

        if (isEmbedded())
        {
            // Should never happen since embedded will always retrieve all fields in one go.
        }
        else
        {
            srm.fetch(this, fieldNumbers);
        }

        // Update the L2 cache to have this object (more loaded fields)
        myPM.putObjectIntoCache(this, false, true);
    }

    /**
     * Fetch from the database all fields that are not currently loaded regardless of whether
     * they are in the current fetch group or not. Called by lifecycle transitions.
     * @since 1.1
     */
    public void loadUnloadedFields()
    {
        int[] fieldNumbers = getFlagsSetTo(loadedFields, allFieldNumbers, false);
        if (fieldNumbers != null)
        {
            boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);
            srm.fetch(this, fieldNumbers);

            // Make sure all SCO lazy-loaded fields have been loaded
            for (int i=0;i<secondClassMutableFieldNumbers.length;i++)
            {
                SingleValueFieldManager sfv = new SingleValueFieldManager();
                provideFields(new int[]{secondClassMutableFieldNumbers[i]}, sfv);
                Object value = sfv.fetchObjectField(i);
                if (value instanceof SCOContainer)
                {
                    ((SCOContainer)value).load();
                }
            }

            if (callPostLoad)
            {
                postLoad();
            }

            // Update the L2 cache to have this object (more loaded fields)
            myPM.putObjectIntoCache(this, false, true);
        }
    }

    /** State indicator for whether this instance is in the process of loading FP fields. */
    boolean loadingFetchPlanFields = false;

    /**
     * Method to load all unloaded fields in the FetchPlan.
     * Recurses through the FetchPlan objects and loads fields of sub-objects where needed.
     * Used as a precursor to detaching objects at commit since fields cant be loaded during
     * the postCommit phase when the detach actuall happens.
     * @param state The FetchPlan state
     */
    public void loadFieldsInFetchPlan(FetchPlanState state)
    {
        if (loadingFetchPlanFields)
        {
            // Already in the process so just return
            return;
        }

        // Load unloaded fields of this object
        loadingFetchPlanFields = true;
        loadUnloadedFieldsInFetchPlan();

        // Recurse through all fields
        int[] fieldNumbers = getFlagsSetTo(loadedFields, allFieldNumbers, true);
        if (fieldNumbers != null && fieldNumbers.length > 0)
        {
            // TODO Fix this to just access the fields of the FieldManager yet this actually does a replaceField
            replaceFields(fieldNumbers, new LoadFieldManager(this, secondClassMutableFields, myFP, state));
        }

        loadingFetchPlanFields = false;
    }

    /**
     * Fetchs from the database all fields in the actual fetch plan.
     * Called by life-cycle transitions.
     * @since 1.1
     */
    private void loadFieldsInFetchPlan()
    {
        int[] fieldNumbers = myFP.getFieldsInActualFetchPlan();
        if (fieldNumbers != null)
        {
            boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);
            srm.fetch(this, fieldNumbers);
            if (callPostLoad)
            {
                postLoad();
            }

            // Update the L2 cache to have this object (more loaded fields)
            myPM.putObjectIntoCache(this, false, true);
        }
    }

    /**
     * Fetchs from the database all fields that are not currently loaded and that are in the current
     * fetch group. Called by lifecycle transitions.
     * @since 1.1
     */
    public void loadUnloadedFieldsInFetchPlan()
    {
        int[] fieldNumbers = getFlagsSetTo(loadedFields, myFP.getFieldsInActualFetchPlan(), false);
        if (fieldNumbers != null)
        {
            boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);
            srm.fetch(this, fieldNumbers);
            if (callPostLoad)
            {
                postLoad();
            }

            // Update the L2 cache to have this object (more loaded fields)
            myPM.putObjectIntoCache(this, false, true);
        }
    }

    /**
     * Fetchs from the database all fields in the actual fetch plan.
     * Called by life-cycle transitions.
     * @since 1.1
     */
    public void loadUnloadedFieldsOfClassInFetchPlan(FetchPlan fetchPlan)
    {
        FetchPlanForClass fpc = ((FetchPlanImpl)fetchPlan).manageFetchPlanForClass(this.cmd);
        int[] fieldNumbers = getFlagsSetTo(loadedFields, fpc.getFieldsInActualFetchPlan(), false);
        if (fieldNumbers != null)
        {
            boolean callPostLoad = fpc.isToCallPostLoadFetchPlan(this.loadedFields);
            srm.fetch(this, fieldNumbers);
            if (callPostLoad)
            {
                postLoad();
            }

            // Update the L2 cache to have this object (more loaded fields)
            myPM.putObjectIntoCache(this, false, true);
        }
    }

    /**
     * Refreshes from the database all fields in fetch plan.
     * Called by life-cycle transitions.
     */
    public void refreshFieldsInFetchPlan()
    {
        int[] fieldNumbers = myFP.getFieldsInActualFetchPlan();

        if (fieldNumbers != null)
        {
            clearDirtyFlags(fieldNumbers);
            clearFlags(loadedFields,fieldNumbers);

            boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);
            srm.fetch(this, fieldNumbers);
            if (callPostLoad)
            {
                postLoad();
            }

            callback.postRefresh(myPC);

            // Update the L2 cache to have this object (more loaded fields)
            myPM.putObjectIntoCache(this, false, true);
        }
    }
    
    /**
     * Refreshes from the database all fields currently loaded.
     * Called by life-cycle transitions.
     */
    public void refreshLoadedFields()
    {
        int[] fieldNumbers = getFlagsSetTo(loadedFields, myFP.getFieldsInActualFetchPlan(), true);

        if (fieldNumbers != null)
        {
            clearDirtyFlags();
            clearFlags(loadedFields);

            boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);
            srm.fetch(this, fieldNumbers);
            if (callPostLoad)
            {
                postLoad();
            }

            // Update the L2 cache to have this object (more loaded fields)
            myPM.putObjectIntoCache(this, false, true);
        }
    }    

    public void unloadField(String fieldName)
    {
        FieldMetaData fmd = (FieldMetaData) getClassMetaData().getField(fieldName);
        unloadField(fmd.getAbsoluteFieldNumber());
    }
    
    /**
     * @param fieldNumber
     */
    void unloadField(int fieldNumber)
    {
        if (pcObjectType == PC)
        {
            loadedFields[fieldNumber] = false;
            unloadedFields[fieldNumber] = true;
        }
        else
        {
            throw new JDOUserException("Cannot unload field of embedded object");
        }
    }    
    
    /**
     * Convenience method to load a field from the datastore.
     * Used in attaching fields and checking their old values (so we dont
     * want any postLoad method being called).
     * TODO Merge this with one of the loadXXXFields methods.
     * @param fieldNumber The field number.
     */
    void loadFieldFromDatastore(int fieldNumber)
    {
        srm.fetch(this, new int[]{fieldNumber});
    }

    /**
     * Return true if the field is cached in the calling instance.
     * <P>
     * In the JPOX implementation of this method, isLoaded() will always
     * return true. If the field is not loaded, it will be loaded as a side
     * effect of the call to this method. If it is in the default fetch group,
     * the default fetch group, including this field, will be loaded.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the absolute field number
     * @return always returns true (this implementation)
     */
    public boolean isLoaded(PersistenceCapable pc, int field)
    {
        if (disconnectClone(pc))
        {
            return true;
        }
        else
        {
            transitionReadField(loadedFields[field]);
            if (!loadedFields[field])
            {
                if (!unloadedFields[field])
                {
                    Object value = provideField(field);
                    if (value != null)
                    {
                        if (secondClassMutableFields[field])
                        {
                            // When we (L2) cache objects we retain the values but don't restore the SCO fields directly
                            // Try to get SCO field from its current value
                            if (value instanceof SCO)
                            {
                                if (!(value instanceof SCOCollection) && !(value instanceof SCOMap))
                                {
                                    SCO newValue = (SCO) newSCOEmptyInstance(field);
                                    newValue.setValueFrom(value, true);
                                    loadedFields[field] = true;
                                    return true;
                                }
                            }
                        }
                        else if (isPCField(field))
                        {
                            // When we (L2) cache objects we retain the values but don't restore the SCO fields directly
                            // Try to load PC field from the cache (if cached)
                            if (value instanceof PersistenceCapable && loadPCFieldFromCache(field, (PersistenceCapable) value))
                            {
                                return true;
                            }
                        }
                    }
                }
                else 
                {
                    // ignore what our PC has as value for this field, (re-)load it anyway
                    // because it has explicitly been unloaded
                }

                if (pcObjectType != PC)
                {
                    // Embedded object so we assume that all was loaded before (when it was read)
                    return true;
                }
                
                if (defaultFetchGroupFields[field])
                {
                    loadDFGFields();
                }
                else
                {
                    loadNonDFGFields(new int[] {field});
                }
                unloadedFields[field] = false;
            }

            return true;
        }
    }   

    /**
     * Convenience method to return if a field number is a PC field
     * @param fieldNumber Number of field
     * @return Whether it is PC
     */
    private boolean isPCField(int fieldNumber)
    {
        int[] pcFields = cmd.getPersistenceCapableFieldNumbers();
        for (int i = 0; i < pcFields.length; i++)
        {
            if (pcFields[i] == fieldNumber)
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Method to try to retrieve a PC field from the L1/L2 caches.
     * Note that this only works for application identity PC objects.
     * @param fieldNumber Number of field to load
     * @param pc The PC object to load
     * @return Whether it was loaded from cache.
     */
    private boolean loadPCFieldFromCache(int fieldNumber, PersistenceCapable pc)
    {
        AbstractClassMetaData fieldCmd = myPM.getMetaDataManager().getMetaDataForClass(pc.getClass(), myPM.getClassLoaderResolver());
        if (fieldCmd.getIdentityType() == IdentityType.APPLICATION)
        {
            Object externalID = AIDUtils.getNewObjectId(pc, fieldCmd);
            AbstractPersistenceManager pm = (AbstractPersistenceManager) myPM;
            Object cachedPC = pm.getObjectFromCache(externalID);
            if (cachedPC != null)
            {
                replaceField(fieldNumber, cachedPC);
                loadedFields[fieldNumber] = true;
                pm.putObjectIntoCache(myPM.findStateManager((PersistenceCapable) cachedPC), true, true);
                return true;
            }
        }
        return false;
    }

    // ---------------------- Field Accessor/Mutator Methods -------------------

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public boolean getBooleanField(PersistenceCapable pc, int field, boolean currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public byte getByteField(PersistenceCapable pc, int field, byte currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public char getCharField(PersistenceCapable pc, int field, char currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public double getDoubleField(PersistenceCapable pc, int field, double currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public float getFloatField(PersistenceCapable pc, int field, float currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public int getIntField(PersistenceCapable pc, int field, int currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public long getLongField(PersistenceCapable pc, int field, long currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public short getShortField(PersistenceCapable pc, int field, short currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public String getStringField(PersistenceCapable pc, int field, String currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @param field the field number 
     * @param currentValue the current value of the field
     * @return the new value for the field
     */
    public Object getObjectField(PersistenceCapable pc, int field, Object currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * Called by the various setXXXField() methods once it's established that
     * the field really deserves to be written.
     * Makes the state transition and replaces the field value in the PC
     * instance.
     */
    private void writeField(int field, Object newValue)
    {
        replaceField(field, newValue);
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setBooleanField(PersistenceCapable pc, int field, boolean currentValue, boolean newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, newValue ? Boolean.TRUE : Boolean.FALSE);
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            if (!loadedFields[field] || currentValue != newValue)
            {
                boolean wasDirty = preWriteField(field);
                writeField(field, newValue ? Boolean.TRUE : Boolean.FALSE);
                postWriteField(wasDirty);
            }
        }
        else
        {
            replaceField(field, newValue ? Boolean.TRUE : Boolean.FALSE);
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setByteField(PersistenceCapable pc, int field, byte currentValue, byte newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Byte(newValue));
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            if (!loadedFields[field] || currentValue != newValue)
            {
                boolean wasDirty = preWriteField(field);
                writeField(field, new Byte(newValue));
                postWriteField(wasDirty);
            }
        }
        else
        {
            replaceField(field, new Byte(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setCharField(PersistenceCapable pc, int field, char currentValue, char newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Character(newValue));
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            if (!loadedFields[field] || currentValue != newValue)
            {
                boolean wasDirty = preWriteField(field);
                writeField(field, new Character(newValue));
                postWriteField(wasDirty);
            }
        }
        else
        {
            replaceField(field, new Character(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setDoubleField(PersistenceCapable pc, int field, double currentValue, double newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Double(newValue));
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            if (!loadedFields[field] || currentValue != newValue)
            {
                boolean wasDirty = preWriteField(field);
                writeField(field, new Double(newValue));
                postWriteField(wasDirty);
            }
        }
        else
        {
            replaceField(field, new Double(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setFloatField(PersistenceCapable pc, int field, float currentValue, float newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Float(newValue));
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            if (!loadedFields[field] || currentValue != newValue)
            {
                boolean wasDirty = preWriteField(field);
                writeField(field, new Float(newValue));
                postWriteField(wasDirty);
            }
        }
        else
        {
            replaceField(field, new Float(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setIntField(PersistenceCapable pc, int field, int currentValue, int newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Integer(newValue));
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            if (!loadedFields[field] || currentValue != newValue)
            {
                boolean wasDirty = preWriteField(field);
                writeField(field, new Integer(newValue));
                postWriteField(wasDirty);
            }
        }
        else
        {
            replaceField(field, new Integer(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setLongField(PersistenceCapable pc, int field, long currentValue, long newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Long(newValue));
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            if (!loadedFields[field] || currentValue != newValue)
            {
                boolean wasDirty = preWriteField(field);
                writeField(field, new Long(newValue));
                postWriteField(wasDirty);
            }
        }
        else
        {
            replaceField(field, new Long(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setShortField(PersistenceCapable pc, int field, short currentValue, short newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Short(newValue));
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            if (!loadedFields[field] || currentValue != newValue)
            {
                boolean wasDirty = preWriteField(field);
                writeField(field, new Short(newValue));
                postWriteField(wasDirty);
            }
        }
        else
        {
            replaceField(field, new Short(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setStringField(PersistenceCapable pc, int field, String currentValue, String newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, newValue);
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            if (!loadedFields[field] || !equals(currentValue, newValue))
            {
                boolean wasDirty = preWriteField(field);
                writeField(field, newValue);
                postWriteField(wasDirty);
            }
        }
        else
        {
            replaceField(field, newValue);
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setObjectField(PersistenceCapable pc, int field, Object currentValue, Object newValue)
    {
        if (currentValue != null && currentValue != newValue && currentValue instanceof PersistenceCapable)
        {
            // Where the object is embedded, remove the owner from its old value since it is no longer managed
            // by this StateManager
            StateManagerImpl currentSM = (StateManagerImpl)myPM.findStateManager((PersistenceCapable)currentValue);
            if (currentSM != null && currentSM.isEmbedded())
            {
                currentSM.removeEmbeddedOwner(this, field);
            }
        }

        if (pc != myPC)
        {
            replaceField(pc, field, newValue);
            disconnectClone(pc);
        }
        else if (myLC != null)
        {
            boolean loadedOldValue = false;
            Object oldValue = currentValue;
            AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(field);
            if (fmd.isDependent() && currentValue == null && newValue == null)
            {
                // Retrieve the old value (when not loaded) for any PC fields that are being nulled
                if (!loadedFields[field])
                {
                    loadNonDFGFields(new int[] {field});
                    loadedOldValue = true;
                    oldValue = provideField(field);
                }
            }

            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                if (currentValue instanceof SCO)
                {
                    ((SCO) currentValue).unsetOwner();
                }

                if (newValue instanceof SCO)
                {
                    SCO sco = (SCO) newValue;
                    Object owner = sco.getOwner();

                    if (owner != null)
                    {
                        throw new JDOUserException(LOCALISER.msg("StateManager.SecondClassObjectShareError", sco.getFieldName(), owner));
                    }
                }

                writeField(field, newValue);
            }
            else if (loadedOldValue)
            {
                // We've updated the value with the old value (when retrieving it above), so put the new value back again
                writeField(field, newValue);
            }
            postWriteField(wasDirty);

            if (fmd.isDependent() && oldValue != null && newValue == null && oldValue instanceof PersistenceCapable)
            {
                // Delete any dependent PC fields that have been nulled
                flush(); // Make sure that any null reference is set first (to avoid any FK constraint failures)
                JPOXLogger.JDO.debug(LOCALISER.msg("StateManager.DeleteDependentNulledField", oldValue, fmd.getFullFieldName()));
                myPM.deletePersistent(oldValue);
            }
        }
        else
        {
            replaceField(field, newValue);
        }
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedBooleanField(PersistenceCapable pc, int field, boolean currentValue)
    {
        currFM.storeBooleanField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedByteField(PersistenceCapable pc, int field, byte currentValue)
    {
        currFM.storeByteField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedCharField(PersistenceCapable pc, int field, char currentValue)
    {
        currFM.storeCharField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedDoubleField(PersistenceCapable pc, int field, double currentValue)
    {
        currFM.storeDoubleField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedFloatField(PersistenceCapable pc, int field, float currentValue)
    {
        currFM.storeFloatField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedIntField(PersistenceCapable pc, int field, int currentValue)
    {
        currFM.storeIntField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedLongField(PersistenceCapable pc, int field, long currentValue)
    {
        currFM.storeLongField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedShortField(PersistenceCapable pc, int field, short currentValue)
    {
        currFM.storeShortField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedStringField(PersistenceCapable pc, int field, String currentValue)
    {
        currFM.storeStringField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc the calling PersistenceCapable instance
     * @param fieldNumber the field number
     * @param currentValue the current value of the field
     */
    public void providedObjectField(PersistenceCapable pc, int fieldNumber, Object currentValue)
    {
        currFM.storeObjectField(fieldNumber, currentValue);
    }

    /**
     * The StateManager uses this method to supply the value of jdoFlags to the
     * associated PersistenceCapable instance.
     *
     * @param pc
     *   the calling PersistenceCapable instance
     * @return
     *   the value of jdoFlags to be stored in the PersistenceCapable instance
     */
    public byte replacingFlags(PersistenceCapable pc)
    {
        // If this is a clone, return READ_WRITE_OK.
        if (pc != myPC)
        {
            return PersistenceCapable.READ_WRITE_OK;
        }
        else
        {
            return jdoFlags;
        }
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a boolean field.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public boolean replacingBooleanField(PersistenceCapable pc, int field)
    {
        boolean value = currFM.fetchBooleanField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a byte field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public byte replacingByteField(PersistenceCapable obj, int field)
    {
        byte value = currFM.fetchByteField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a char field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public char replacingCharField(PersistenceCapable obj, int field)
    {
        char value = currFM.fetchCharField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a double field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public double replacingDoubleField(PersistenceCapable obj, int field)
    {
        double value = currFM.fetchDoubleField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a float field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public float replacingFloatField(PersistenceCapable obj, int field)
    {
        float value = currFM.fetchFloatField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a int field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public int replacingIntField(PersistenceCapable obj, int field)
    {
        int value = currFM.fetchIntField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a long field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public long replacingLongField(PersistenceCapable obj, int field)
    {
        long value = currFM.fetchLongField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a short field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public short replacingShortField(PersistenceCapable obj, int field)
    {
        short value = currFM.fetchShortField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a String field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public String replacingStringField(PersistenceCapable obj, int field)
    {
        String value = currFM.fetchStringField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of an Object field.
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public Object replacingObjectField(PersistenceCapable obj, int field)
    {
        try
        {
            Object value = currFM.fetchObjectField(field);
            loadedFields[field] = true;
            return value;
        }
        catch (EndOfFetchPlanGraphException eodge)
        {
            // Beyond the scope of the fetch-depth when detaching
            return null;
        }
    }

    /**
     * Method to return the current value of a particular field.
     * @param fieldNumber Number of field
     * @return The value of the field
     */
    public synchronized Object provideField(int fieldNumber)
    {
        if (!unloadedFields[fieldNumber])
        {
            return provideField(myPC, fieldNumber);
        }
        else
        {
            return null;
        }
    }

    /**
     * Method to change the value of a particular field.
     * @param fieldNumber Number of field
     * @param value New value
     */
    public synchronized void replaceField(int fieldNumber, Object value)
    {
        replaceField(myPC, fieldNumber, value);
    }

    /**
     * Method to retrieve the value of a field from the PC object.
     * @param pc The PC object
     * @param fieldNumber Number of field
     * @return The value of the field
     */
    private synchronized Object provideField(PersistenceCapable pc, int fieldNumber)
    {
        Object obj;

        FieldManager prevFM = currFM;
        currFM = new SingleValueFieldManager();
        try
        {
            // This assumes that the field is loaded!
            pc.jdoProvideField(fieldNumber);
            obj = currFM.fetchObjectField(fieldNumber);
        }
        finally
        {
            currFM = prevFM;
        }

        return obj;
    }

    /**
     * Method to change the value of a field in the PC object.
     * @param pc The PC object
     * @param fieldNumber Number of field
     * @param value The new value of the field
     */
    private synchronized void replaceField(PersistenceCapable pc, int fieldNumber, Object value)
    {
        if (embeddedOwners != null)
        {
            // Notify any owners that embed this object that it has just changed
            // We do this before we actually change the object so we can compare with the old value
            Iterator ownerIter = embeddedOwners.iterator();
            while (ownerIter.hasNext())
            {
                EmbeddedOwnerRelation owner = (EmbeddedOwnerRelation)ownerIter.next();
                StateManagerImpl ownerSM = (StateManagerImpl)owner.sm;

                AbstractPropertyMetaData ownerFmd = ownerSM.cmd.getManagedFieldAbsolute(owner.fieldNumber);
                if (ownerFmd.getCollection() != null)
                {
                    // PC Object embedded in collection
                    Object ownerField = ownerSM.provideField(owner.fieldNumber);
                    if (ownerField instanceof SCOCollection)
                    {
                        ((SCOCollection)ownerField).updateEmbeddedElement(myPC, fieldNumber, value);
                    }
                }
                else if (ownerFmd.getMap() != null)
                {
                    // PC Object embedded in map
                    Object ownerField = ownerSM.provideField(owner.fieldNumber);
                    if (ownerField instanceof SCOMap)
                    {
                        if (pcObjectType == EMBEDDED_MAP_KEY_PC)
                        {
                            ((SCOMap)ownerField).updateEmbeddedKey(myPC, fieldNumber, value);
                        }
                        if (pcObjectType == EMBEDDED_MAP_VALUE_PC)
                        {
                            ((SCOMap)ownerField).updateEmbeddedValue(myPC, fieldNumber, value);
                        }
                    }
                }
                else
                {
                    // PC Object embedded in PC object
                    if (!ownerSM.updatingEmbeddedFieldsWithOwner)
                    {
                        // Update the owner when one of our fields have changed, EXCEPT when they have just
                        // notified us of our owner field!
                        if (ownerSM.isEmbedded())
                        {
                            // Owner is embedded so just update its field
                            ownerSM.replaceField(owner.fieldNumber, pc);
                        }
                        else
                        {
                            // Owner is not embedded so mark the field as dirty too
                            boolean wasDirty = ownerSM.preWriteField(owner.fieldNumber);
                            ownerSM.replaceField(owner.fieldNumber, pc);
                            ownerSM.postWriteField(wasDirty);
                        }
                    }
                }
            }
        }

        // Update the field in our PC object
        FieldManager prevFM = currFM;
        currFM = new SingleValueFieldManager();

        try
        {
            currFM.storeObjectField(fieldNumber, value);
            pc.jdoReplaceField(fieldNumber);
        }
        finally
        {
            currFM = prevFM;
        }
    }

    /**
     * Called from the StoreManager after StoreManager.update() is called to
     * obtain updated values from the PersistenceCapable associated with this
     * StateManager.
     *
     * @param fieldNumbers An array of field numbers to be updated by the Store
     * @param fm The updated values are stored in this object. This object is only valid
     *   for the duration of this call.
     */
    public synchronized void provideFields(int fieldNumbers[], FieldManager fm)
    {
        FieldManager prevFM = currFM;
        currFM = fm;

        try
        {
            // This will respond by calling this.providedXXXFields() with the value of the field
            myPC.jdoProvideFields(fieldNumbers);
        }
        finally
        {
            currFM = prevFM;
        }
    }

    /**
     * Called from the StoreManager to refresh data in the PersistenceCapable
     * object associated with this StateManager.
     * @param fieldNumbers An array of field numbers to be refreshed by the Store
     * @param fm The updated values are stored in this object. This object is only valid
     *   for the duration of this call.
     * @param replaceWhenDirty Whether to replace the fields when they are dirty here
     */
    public synchronized void replaceFields(int fieldNumbers[], FieldManager fm, boolean replaceWhenDirty)
    {
        FieldManager prevFM = currFM;
        currFM = fm;

        try
        {
            int[] fieldsToReplace = fieldNumbers;
            if (!replaceWhenDirty)
            {
                // Dont replace any fields that are dirty
                int numberToReplace = fieldNumbers.length;
                for (int i=0;i<fieldNumbers.length;i++)
                {
                    if (dirtyFields[fieldNumbers[i]])
                    {
                        numberToReplace--;
                    }
                }
                if (numberToReplace > 0 && numberToReplace != fieldNumbers.length)
                {
                    fieldsToReplace = new int[numberToReplace];
                    int n = 0;
                    for (int i=0;i<fieldNumbers.length;i++)
                    {
                        if (!dirtyFields[fieldNumbers[i]])
                        {
                            fieldsToReplace[n++] = fieldNumbers[i];
                        }
                    }
                }
                else if (numberToReplace == 0)
                {
                    fieldsToReplace = null;
                }
            }

            if (fieldsToReplace != null)
            {
                myPC.jdoReplaceFields(fieldsToReplace);
            }
        }
        finally
        {
            currFM = prevFM;
        }
    }

    /**
     * Called from the StoreManager to refresh data in the PersistenceCapable
     * object associated with this StateManager.
     * @param fieldNumbers An array of field numbers to be refreshed by the Store
     * @param fm The updated values are stored in this object. This object is only valid
     *   for the duration of this call.
     */
    public synchronized void replaceFields(int fieldNumbers[], FieldManager fm)
    {
        replaceFields(fieldNumbers, fm, true);
    }

    /**
     * Called from the StoreManager to refresh data in the PersistenceCapable
     * object associated with this StateManager. Only not loaded fields are refreshed
     *
     * @param fieldNumbers An array of field numbers to be refreshed by the Store
     * @param fm The updated values are stored in this object. This object is only valid
     *   for the duration of this call.
     */
    public synchronized void replaceNonLoadedFields(int fieldNumbers[], FieldManager fm)
    {
        FieldManager prevFM = currFM;
        currFM = fm;

        boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);
        try
        {
            int[] fieldsToReplace = getFlagsSetTo(loadedFields, fieldNumbers, false);
            if (fieldsToReplace != null)
            {
                myPC.jdoReplaceFields(fieldsToReplace);
            }
        }
        finally
        {
            currFM = prevFM;
        }

        if (callPostLoad && isFetchPlanLoaded())
        {
            // The fetch plan is now loaded so fire off any necessary post load
            postLoad();
        }
    }

    /**
     * Convenience accessor for whether this StateManager manages an embedded/serialised object.
     * @return Whether the managed object is embedded/serialised.
     */
    public boolean isEmbedded()
    {
        return pcObjectType > 0;
    }

    /**
     * Method to set this StateManager as managing an embedded/serialised object.
     * @param embeddedType The type of object being managed
     */
    public void setPcObjectType(int embeddedType)
    {
        this.pcObjectType = embeddedType;
    }

    /**
     * Method to register an owner StateManager with this embedded/serialised object.
     * @param ownerSM The owning State Manager.
     * @param ownerFieldNumber The field number in the owner that the embedded/serialised object is stored as
     */
    public void addEmbeddedOwner(org.jpox.StateManager ownerSM, int ownerFieldNumber)
    {
        if (ownerSM == null)
        {
            return;
        }

        if (embeddedOwners == null)
        {
            embeddedOwners = new ArrayList();
        }
        embeddedOwners.add(new EmbeddedOwnerRelation(ownerSM, ownerFieldNumber));
    }

    /**
     * Method to remove an owner StateManager from this embedded/serialised objects owners list.
     * @param ownerSM The owner to remove
     * @param ownerFieldNumber The field in the owner where this object is stored
     */
    public void removeEmbeddedOwner(StateManager ownerSM, int ownerFieldNumber)
    {
        if (embeddedOwners != null)
        {
            Iterator iter = embeddedOwners.iterator();
            while (iter.hasNext())
            {
                EmbeddedOwnerRelation relation = (EmbeddedOwnerRelation)iter.next();
                if (relation.sm == ownerSM && relation.fieldNumber == ownerFieldNumber)
                {
                    iter.remove();
                    break;
                }
            }
        }
    }

    /**
     * Wrapper class storing the owning state manager, and the field of the
     * PC managed by the owning state manager where this object is embedded/serialised.
     */
    private class EmbeddedOwnerRelation
    {
        StateManager sm;
        int fieldNumber;

        /**
         * 
         * @param ownerSM the owner StateManager
         * @param ownerFieldNumber the absolute owner field number
         */
        public EmbeddedOwnerRelation(StateManager ownerSM, int ownerFieldNumber)
        {
            this.sm = ownerSM;
            this.fieldNumber = ownerFieldNumber;
        }
    }

    /** Flag whether this SM is updating the ownership of its embedded/serialised field(s). */
    private boolean updatingEmbeddedFieldsWithOwner = false;

    /**
     * Method to store a new SCO field at the specified field number.
     * @param fieldNumber The field number
     * @param value The value
     * @param forUpdate Whether this SCO needs updating in the datastore
     * @return the new SCO instance
     */
    public Object newSCOInstance(int fieldNumber, Object value, boolean forUpdate)
    {
        if (value instanceof PersistenceCapable)
        {
            AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
            if (fmd.getEmbeddedMetaData() != null &&
                fmd.getEmbeddedMetaData().getOwnerField() != null)
            {
                // Assign the embedded/serialised object "owner-field" if specified
                StateManagerImpl subSM = (StateManagerImpl)myPM.findStateManager((PersistenceCapable)value);
                int ownerAbsFieldNum = subSM.cmd.getFieldNumberAbsolute(fmd.getEmbeddedMetaData().getOwnerField());
                if (ownerAbsFieldNum >= 0)
                {
                    updatingEmbeddedFieldsWithOwner = true;
                    subSM.replaceField(ownerAbsFieldNum, myPC);
                    updatingEmbeddedFieldsWithOwner = false;
                }
            }
        }
        if (secondClassMutableFields[fieldNumber] && value != null)
        {
            if (!(value instanceof SCO) || myPC != ((SCO)value).getOwner())
            {
                AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
                SCO newValue = SCOUtils.newSCOInstance(fmd.getType(), this, fmd, value, forUpdate);

                value = newValue;
            }
        }

        return value;
    }

    /**
     * Method to replace a second-class object field value.
     * @param fieldNumber Number of field
     * @param value New value of the field
     * @return The new SCO object representing the new value
     */
    public Object replaceSCOField(int fieldNumber, Object value)
    {
        Object newValue = newSCOInstance(fieldNumber, value, true);
        if (newValue != value)
        {
            replaceField(fieldNumber, newValue);
        }
        return newValue;
    }

    /**
     * Method to create a new SCO proxy wraper object for a field.
     * The wrapper will be empty. e.g this could be a SCO collection
     * and will be a org.jpox.sco.{class} type with no elements initially.
     * @param fieldNumber Number of field
     * @return The new (empty) SCO object
     */
    public Object newSCOEmptyInstance(int fieldNumber)
    {
        AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
        SCO newValue = SCOUtils.newSCOInstance(fmd.getType(), this, fmd, null, false);

        return newValue;
    }

    /**
     * Method to change the object state to read-field.
     * @param isLoaded if the field was previously loaded
     **/
    private void transitionReadField(boolean isLoaded)
    {
        if (myLC == null)
        {
            return;
        }
        synchronized(myPC)
        {
            preStateChange();
            try
            {
                myLC = myLC.transitionReadField(this, isLoaded);
            }
            finally
            {
                postStateChange();
            }
        }
    }

    /**
     * Method to change the object state to read-field.
     **/
    private void transitionReadField()
    {
        transitionReadField(false);
    }

    /**
     * Method to change the object state to write-field.
     **/
    private void transitionWriteField()
    {
        synchronized(myPC)
        {
            preStateChange();
            try
            {
                myLC = myLC.transitionWriteField(this);
            }
            finally
            {
                postStateChange();
            }
        }
    }

    // -------------------------- Version handling ----------------------------

    /** 
     * Return the object representing the version 
     * of the calling instance.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @return the object representing the version of the calling instance
     * @since JDO 2.0
     */    
    public Object getVersion(PersistenceCapable pc)
    {
        if (pc == myPC)
        {
            return myVersion;
        }
        else
        {
            return null;
        }
    }

    /**
     * Sets the value for the version column in a transaction not yet committed
     * @param version The version
     */
    public void setTransactionalVersion(Object version)
    {
        this.transactionalVersion = version;
    }

    /** Return the object representing the transactional version 
     * of the calling instance.
     * @param pc the calling <code>PersistenceCapable</code> instance
     * @return the object representing the version of the calling instance
     * @since 1.1.1
     */    
    public Object getTransactionalVersion (PersistenceCapable pc)
    {
        return this.transactionalVersion;
    }

    /**
     * Sets the value for the version column in the datastore
     * @param version The version
     */
    public void setVersion(Object version)
    {
        this.myVersion = version;
        this.transactionalVersion = version;
    }

    /**
     * Accessor for the highest field number in this class
     * @return The highest field number
     */
    public int getHighestFieldNumber()
    {
        return cmd.getNoOfManagedFields() + cmd.getNoOfInheritedManagedFields();
    }

    // ------------------------- Lifecycle Methods -----------------------------

    /**
     * Accessor for the LifeCycleState
     * @return the LifeCycleState
     */
    public LifeCycleState getLifecycleState()
    {
        return myLC;
    }

    /**
     * Method to mark an object for reachability.
     * Provides the basis for "persistence-by-reachability", but run at commit time only.
     * The reachability algorithm is also run at makePersistent, but directly via InsertRequest.
     * The "reachables" list could be changed to store id's instead of StateManagers if preferred.
     * @param reachables List of StateManagers currently logged as reachable
     */
    public void runReachability(Set reachables)
    {
        if (reachables == null)
        {
            return;
        }
        if (!reachables.contains(getInternalObjectId()))
        {
            // Make sure all changes are persisted first
            flush();

            if (isDeleted(myPC))
            {
                // The object is deleted so nothing further will be reachable
                return;
            }

            if (getPersistenceManager().isEnlistedInTransaction(getInternalObjectId()))
            {
                // This object is enlisted in the transaction so make sure all fields are loaded
                loadUnloadedFields();
            }

            if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
            {
                JPOXLogger.JDO_REACHABILITY.debug("Adding instance of "+cmd.getFullClassName()+" with id \""+getObjectId(myPC)+"\" and lifecycle state \""+myLC+"\" to the list of reachables on commit.");
            }            
            // Add this StateManager since not yet reached
            reachables.add(getInternalObjectId());

            // Go through all (FetchPlan) fields for reachability
            for (int i = 0; i < loadedFields.length; i++)
            {
                if (loadedFields[i])
                {
                    Object value = provideField(i);
                    if (value == null)
                    {
                        if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
                        {
                            JPOXLogger.JDO_REACHABILITY.debug("Performing reachability on field "+cmd.getManagedFieldAbsolute(i).getFullFieldName()+" which is null");
                        }
                        // Do nothing
                    }
                    else if (secondClassMutableFields[i] && value instanceof SCOContainer)
                    {
                        if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
                        {
                            JPOXLogger.JDO_REACHABILITY.debug("Performing reachability on SCO field "+cmd.getManagedFieldAbsolute(i).getFullFieldName());
                        }
                        // SCO which maybe has PC components
                        SCOContainer sco = (SCOContainer) value;
                        sco.runReachability(reachables);
                    }
                    else if (value.getClass().isArray())
                    {
                        if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
                        {
                            JPOXLogger.JDO_REACHABILITY.debug("Performing reachability on Array field "+cmd.getManagedFieldAbsolute(i).getFullFieldName());
                        }
                        // Array of PC/Reference elements have PC objects
                        if (PersistenceCapable.class.isAssignableFrom(value.getClass().getComponentType()) ||
                            value.getClass().getComponentType().isInterface() ||
                            value.getClass().getComponentType() == java.lang.Object.class)
                        {
                            Object[] arrValue = (Object[])value;
                            for (int j=0;j<arrValue.length;j++)
                            {
                                PersistenceCapable pc = (PersistenceCapable)arrValue[j];
                                org.jpox.StateManager elementSM = myPM.findStateManager(pc);
                                if (elementSM != null)
                                {
                                    elementSM.runReachability(reachables);
                                }
                            }
                        }
                    }
                    else if (value instanceof PersistenceCapable)
                    {
                        if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
                        {
                            JPOXLogger.JDO_REACHABILITY.debug("Performing reachability on PersistenceCapable field "+cmd.getManagedFieldAbsolute(i).getFullFieldName());
                        }
                        // 1-1 relation with PC
                        PersistenceCapable pc = (PersistenceCapable) value;
                        org.jpox.StateManager valueSM = myPM.findStateManager(pc);
                        if (valueSM != null)
                        {
                            valueSM.runReachability(reachables);
                        }
                        else
                        {
                            if (JPOXLogger.JDO_REACHABILITY.isDebugEnabled())
                            {
                                JPOXLogger.JDO_REACHABILITY.debug("Cannot retrieve a StateManger for PersistenceCapable field "+cmd.getManagedFieldAbsolute(i).getFullFieldName()+" with object id "+pc.jdoGetObjectId());
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Method to make the object persistent.
     **/
    public void makePersistent()
    {
        if (flushedNew)
        {
            // With CompoundIdentity cases it is possible that the persist of the PK PC fields performed during
            // SM construction can cause flush() which persists our object before sm.makePersistent() is called.
            registerTransactional();
            return;
        }

        if (cmd.isEmbeddedOnly())
        {
            return;
        }

        // Call any lifecycle listeners waiting for this event.
        callback.postCreate(myPC);

        // If this is an embedded/serialised object becoming persistent in its own right, assign an identity.
        if (myID == null)
        {
            setIdentity();
        }

        dirty = true;
        
        if (myPM.isDelayDatastoreOperationsEnabled())
        {
            myPM.markDirty(this);
        }
        else
        {
            internalMakePersistent();
        }

        registerTransactional();
    }

    /**
     * Method to make the object persistent.
     **/
    private void internalMakePersistent()
    {
        inserting = ActivityState.INSERTING;

        try
        {
            callback.preStore(myPC); // This comes after setting the INSERTING flag so we know we are inserting it now

            //in InstanceLifecycleEvents this object could get dirty if a field is changed in preStore or
            //postCreate, we clear dirty flags to make sure this object will not be flushed again
            clearDirtyFlags();

            srm.insert(this);
            inserting = ActivityState.NONE;
            flushedNew = true;
        }
        finally
        {
            inserting = ActivityState.NONE;
            flushedNew = true;

            callback.postStore(myPC);
        }
    }

    /**
     * Tests whether this object is being inserted.
     * @return true if this instance is inserting.
     */
    public boolean isInserting()
    {
        return (inserting == ActivityState.INSERTING);
    }

    /**
     * Returns whether the field of this object is inserted in the datastore.
     * Only applies during the makePersistent process.
     * @param fieldNumber Number of the field
     * @return Whether it is inserted to the level of this field
     */
    public boolean isInserted(int fieldNumber)
    {
        if (!isInserting())
        {
            return true;
        }
        if (latestInsertedDatastoreClass == null)
        {
            // Not yet inserted anything
            return false;
        }
        AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
        if (fmd == null)
        {
            return false;
        }
        if (latestInsertedDatastoreClass instanceof ClassTable)
        {
            return latestInsertedDatastoreClass.managesClass(fmd.getClassName());
        }

        return false;
    }

    /**
     * Latest datastore class used during insertion. Only used during makePersistent.
     * Only the most recent datastore class is stored since with inheritance the tables are inserted
     * in order, starting at the root.
     */
    private DatastoreClass latestInsertedDatastoreClass = null;

    /**
     * Change the activity state.
     * @param activityState the new state
     * @param table Datastore class that has just been processed
     */
    public void changeActivityState(ActivityState activityState, DatastoreClass table)
    {
        if (activityState != ActivityState.INSERTING_CALLBACKS)
        {
            inserting = activityState;
        }
        if (activityState == ActivityState.INSERTING_CALLBACKS)
        {
            latestInsertedDatastoreClass = table;
            if (table.managesClass(cmd.getFullClassName()))
            {
                // Full insertion has just completed so notify all interested StateManagers
                inserting = activityState;
                synchronized(insertionNotifyList)
                {
                    for (int i=0; i<insertionNotifyList.size(); i++)
                    {
                        StateManagerImpl notifySM = (StateManagerImpl)insertionNotifyList.get(i);
                        notifySM.insertionCompleted(this);
                    }
                    insertionNotifyList.clear();
                }
            }
            else
            {
                // Not yet inserted fully so keep going
            }
        }
    }

    /**
     * Method to add a notifier that we must contact when we have finished our insertion.
     * @param sm the state manager
     * @param activityState the ActivityState (unused)
     */
    public void addInsertionNotifier(StateManager sm, ActivityState activityState)
    {
        // TODO Use the second param to add the StateManager to other lists for other events
        insertionNotifyList.add(sm);
    }

    /**
     * Marks the given field as being required to be updated when the specified object
     * has been inserted.
     * @param pc The PersistenceCapable
     * @param fieldNumber Number of the field.
     */
    public void updateFieldAfterInsert(PersistenceCapable pc, int fieldNumber)
    {
        StateManagerImpl otherSM = (StateManagerImpl) myPM.findStateManager(pc);

        // Register the other SM to update us when it is inserted
        otherSM.addInsertionNotifier(this, ActivityState.INSERTING_CALLBACKS);

        // Register that we should update this field when the other SM informs us
        FieldContainer cont = (FieldContainer)insertionUpdateFieldList.get(otherSM);
        if (cont == null)
        {
            cont = new FieldContainer(fieldNumber);
        }
        else
        {
            cont.set(fieldNumber);
        }
        insertionUpdateFieldList.put(otherSM, cont);

        if (JPOXLogger.JDO.isDebugEnabled())
        {
            JPOXLogger.JDO.debug(LOCALISER.msg("StateManager.FieldRegisteredForUpdateAfterInsertion", 
                cmd.getManagedFieldAbsolute(fieldNumber).getName(), myID));
        }
    }

    /**
     * Method called by another StateManager when this object has registered that it needs to know
     * when the other object has been inserted.
     * @param sm State Manager of the other object that has just been inserted
     */
    void insertionCompleted(StateManagerImpl sm)
    {
        // Go through our insertion update list and mark all required fields as dirty
        FieldContainer fldCont = (FieldContainer)insertionUpdateFieldList.get(sm);
        if (fldCont != null)
        {
            dirty = true;
            int[] fieldsToUpdate = fldCont.getFields();
            for (int i=0;i<fieldsToUpdate.length;i++)
            {
                if (JPOXLogger.JDO.isDebugEnabled())
                {
                    JPOXLogger.JDO.debug(LOCALISER.msg("StateManager.FieldBeingUpdatedNowValueInserted", 
                        cmd.getManagedFieldAbsolute(fieldsToUpdate[i]).getName(), myID));
                }
                dirtyFields[fieldsToUpdate[i]] = true;
            }
            insertionUpdateFieldList.remove(sm);

            // Perform our update
            flush();
        }
    }

    /**
     * Method to set the value for an external field stored against this object
     * @param mapping The mapping for the (external) field
     * @param value The value that this field has
     */
    public void setExternalFieldValueForMapping(JavaTypeMapping mapping, Object value)
    {
        externalFieldValuesByMapping.put(mapping, value);
    }

    /**
     * Accessor for the value of an external field.
     * This is used when inserting this object so that we can insert the external field values too.
     * @param mapping The external field mapping
     * @return The value for this mapping
     */
    public Object getValueForExternalField(JavaTypeMapping mapping)
    {
        return externalFieldValuesByMapping.get(mapping);
    }

    /** Private class storing the fields to be updated for a StateManager, when it is inserted */
    private class FieldContainer
    {
        boolean[] fieldsToUpdate = new boolean[allFieldNumbers.length];
        /**
         * Constructor
         * @param fieldNumber the absolute field number to flag true
         */
        public FieldContainer(int fieldNumber)
        {
            fieldsToUpdate[fieldNumber] = true;
        }
        /**
         * Flag to true the <code>fieldNumber</code>
         * @param fieldNumber the absolute field number to flag true
         */
        public void set(int fieldNumber)
        {
            fieldsToUpdate[fieldNumber] = true;
        }
        /**
         * Array with absolute field numbers with true flag
         * @return array of absolute field numbers
         */
        public int[] getFields()
        {
            return getFlagsSetTo(fieldsToUpdate,true);
        }
    }

    /**
     * Makes Transactional Transient instances persistent
     */
    public void makePersistentTransactionalTransient()
    {
        preStateChange();
        try
        {
            if (myLC.isTransactional && !myLC.isPersistent)
            {
                // make the transient instance persistent in the datastore, if is transactional and !persistent 
                makePersistent();
                myLC = myLC.transitionMakePersistent(this);
            }
        }
        finally
        {
            postStateChange();
        }

    }

    /**
     * Method to change the object state to transactional.
     **/
    public void makeTransactional()
    {
        preStateChange();
        try
        {
            if (myLC == null)
            {
                initializeSM(LifeCycleState.T_CLEAN);
                setRestoreValues(true);
            }
            else
            {
                myLC = myLC.transitionMakeTransactional(this);
            }
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to change the object state to nontransactional.
     **/
    public void makeNontransactional()
    {
        preStateChange();
        try
        {
            myLC = myLC.transitionMakeNontransactional(this);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to change the object state to transient.
     * @param state Object containing the state of any fetchplan processing
     **/
    public void makeTransient(FetchPlanState state)
    {
        if (state == null)
        {
            // No FetchPlan in use so just unset the owner of all loaded SCO fields
            int[] fieldNumbers = getFlagsSetTo(loadedFields, secondClassMutableFieldNumbers, true);
            if (fieldNumbers != null)
            {
                provideFields(fieldNumbers, new UnsetOwners());
            }
        }
        else
        {
            // Load any unloaded fetch-plan fields
            loadUnloadedFieldsInFetchPlan();

            // Make all loaded SCO fields transient appropriate to this fetch plan
            int[] fieldNumbers = getFlagsSetTo(loadedFields, secondClassMutableFieldNumbers, true);
            if (fieldNumbers != null && fieldNumbers.length > 0)
            {
                // TODO Fix this to just access the fields of the FieldManager yet this actually does a replaceField
                replaceFields(fieldNumbers, new MakeTransientFieldManager(this, secondClassMutableFields, myFP, state));
            }
        }

        preStateChange();
        try
        {
            myLC = myLC.transitionMakeTransient(this,state!=null);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to detach this object.
     * This is used to detach an object at the "detachOnClose" of a PersistenceManager.
     * @param state State for the detachment process
     */
    public void detach(FetchPlanState state)
    {
        if (!cmd.isDetachable())
        {
            return;
        }
        if (myLC.isDeleted())
        {
            // Ignore any objects that have already been deleted.
            return;
        }
        if (myPC.jdoIsDetached())
        {
            return;
        }
        if (detaching)
        {
            // In the process of detaching this so do nothing on this call.
            return;
        }

        if (JPOXLogger.JDO.isDebugEnabled())
        {
            try
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("PM.Detach", myPC, "" + state.getCurrentFetchDepth()));
            }
            catch (Exception e)
            {
                // we catch any exception that can be raised inside the PC.toString() method
                JPOXLogger.JDO.debug(LOCALISER.msg("PM.Detach", myPC.getClass().getName(), 
                		"" + state.getCurrentFetchDepth()));
            }
        }

        // Call any "pre-detach" listeners
        callback.preDetach(myPC);

        try
        {
            detaching = true;
    
            // Handle any field loading/unloading before the detach
            if ((myPM.getFetchPlan().getDetachmentOptions() & FetchPlan.DETACH_LOAD_FIELDS) != 0)
            {
                // Load any unloaded fetch-plan fields
                loadUnloadedFieldsInFetchPlan();
            }
    
            // Detach any fields in the current FetchPlan
            FieldManager detachFieldManager = new DetachFieldManager(this, secondClassMutableFields, myFP, state, false);
            for (int i = 0; i < loadedFields.length; i++)
            {
                if (loadedFields[i])
                {
                    AbstractPropertyMetaData fmd = myFP.getAbstractClassMetaData().getManagedFieldAbsolute(i);
                    final boolean requiresFCOFetching = MetaDataUtils.getInstance().storesFCO(fmd, getPersistenceManager().getClassLoaderResolver());
                    if (requiresFCOFetching)
                    {
                        try
                        {
                            detachFieldManager.fetchObjectField(i);
                        }
                        catch (EndOfFetchPlanGraphException eofpge)
                        {
                            // Do nothing
                        }
                    }
                }
            }

            // Migrate the lifecycle state to DETACHED_CLEAN
            myLC = myLC.transitionDetach(this);

            // Update the object with its detached state
            myPC.jdoReplaceFlags();
            ((Detachable)myPC).jdoReplaceDetachedState();
            // Call any "post-detach" listeners
            callback.postDetach(myPC, myPC); // there is no copy, so give the same object

            PersistenceCapable toCheckPC = myPC;
            Object toCheckID = myID;
            disconnect();
            // Sanity check on the objects detached state
            if (!toCheckPC.jdoIsDetached())
            {
                throw new JDOUserException(LOCALISER.msg("StateManager.DetachOfObjectFailed", toCheckPC.getClass().getName(), toCheckID));
            }
        }
        finally
        {
            detaching = false;
        }
    }

    /**
     * Return an array of field numbers that must be included in the detached object
     * @return the field numbers array
     */
    private int[] getFieldsNumbersToDetach()
    {
        // This will cause the detach of any other fields in the FetchPlan.
        int[] fieldsToDetach = myFP.getFieldsInActualFetchPlan();
        if ((myPM.getFetchPlan().getDetachmentOptions() & FetchPlan.DETACH_UNLOAD_FIELDS) == 0)
        {
            // Detach fetch-plan fields plus any other loaded fields
            int[] loadedFieldNumbers = getFlagsSetTo(loadedFields, allFieldNumbers, true);
            if (loadedFieldNumbers != null)
            {
                boolean[] flds = new boolean[allFieldNumbers.length];
                for (int i=0;i<fieldsToDetach.length;i++)
                {
                    flds[fieldsToDetach[i]] = true;
                }
                for (int i=0;i<loadedFieldNumbers.length;i++)
                {
                    flds[loadedFieldNumbers[i]] = true;
                }
                fieldsToDetach = getFlagsSetTo(flds,true);
            }
        }
        return fieldsToDetach;
    }
    
    /**
     * Method to make detached copy of this instance
     * @param state State for the detachment process
     * @return the detached PersistenceCapable instance
     * @since 1.1
     */
    public Object detachCopy(FetchPlanState state)
    {
        if (myLC.isDeleted())
        {
            throw new JDOUserException(LOCALISER.msg("StateManager.DetachOfDeletedObjectNotPossible", myPC.getClass().getName(), myID));
        }
        if (myPC.jdoIsDetached())
        {
            throw new JDOUserException(LOCALISER.msg("StateManager.DetachOfAlreadyDetachedObjectNotPossible", myPC.getClass().getName(), myID));
        }
        if (dirty)
        {
            myPM.flush();
        }

        // Object in the process of detaching (recursive) so return the object which will be the detached object
        if (detaching)
        {
            return detachingPC;
        }
        PersistenceCapable detachedPC = myPC.jdoNewInstance(this);
        detachingPC = detachedPC;

        // Check if detachable ... if so then we detach a copy, otherwise we return a transient copy
        boolean detachable = ClassUtils.isDetachable(myPC);

        // make sure a detaching PC is not read by another thread while we are detaching
        synchronized (detachingPC)
        {
            if (detachable)
            {
                if (JPOXLogger.JDO.isDebugEnabled())
                {
                    try
                    {
                        JPOXLogger.JDO.debug(LOCALISER.msg("PM.DetachCopy", myPC, "" + state.getCurrentFetchDepth()));
                    }
                    catch (Exception e)
                    {
                        // we catch any exception that can be raised inside the PC.toString() method
                        JPOXLogger.JDO.debug(LOCALISER.msg("PM.DetachCopy", myPC.getClass().getName(), 
                        		"" + state.getCurrentFetchDepth()));
                    }
                }

                // Call any "pre-detach" listeners
                callback.preDetach(myPC);
            }

            try
            {
                detaching = true;

                // Handle any field loading/unloading before the detach
                if ((myPM.getFetchPlan().getDetachmentOptions() & FetchPlan.DETACH_LOAD_FIELDS) != 0)
                {
                    // Load any unloaded fetch-plan fields
                    loadUnloadedFieldsInFetchPlan();
                }

                if (myLC == myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.HOLLOW) ||
                    myLC == myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.P_NONTRANS))
                {
                    // Migrate any HOLLOW/P_NONTRANS to P_CLEAN etc
                    myLC = myLC.transitionReadField(this, true);
                }

                // Create a "detached" SM
                StateManagerImpl smDetachedPC = new StateManagerImpl(myPM, myPM.getPMFContext().getApiAdapter(), detachedPC, 
                    this.cmd, getExternalObjectId(myPC), myVersion);
                detachedPC.jdoReplaceStateManager(smDetachedPC);

                smDetachedPC.replaceFields(getFieldsNumbersToDetach(), new DetachFieldManager(this, secondClassMutableFields, 
                		myFP, state, true));

                if (detachable)
                {
                   // Update the object with its detached state - not to be confused with the "state" object above
                    detachedPC.jdoReplaceFlags();
                    ((Detachable)detachedPC).jdoReplaceDetachedState();
                }
                else
                {
                    smDetachedPC.makeTransient(null);
                }

                // Remove its StateManager since now detached or transient
                detachedPC.jdoReplaceStateManager(null);
            }
            finally
            {
                detaching = false;
                detachingPC = null;
            }
            
            // Sanity check on the objects detached state
            if (detachable && !detachedPC.jdoIsDetached())
            {
                throw new JDOUserException(LOCALISER.msg("StateManager.DetachOfObjectFailed", detachedPC.getClass().getName(), myID));
            }

            if (detachable)
            {
                // Call any "post-detach" listeners
                callback.postDetach(myPC, detachedPC);
            }
        }
        return detachedPC;
    }

    /**
     * Accessor for the PC object that this detached object is being attached as.
     * @return The attached version of this object
     */
    public Object getAttachedPC()
    {
        return attachedPC;
    }

    /**
     * Method to attach the detached PersistenceCapable instance
     * @param detachedPC the detached PersistenceCapable instance to be attached
     * @param embedded Whether the object is stored embedded/serialised in another object
     * @return The attached copy
     * @since 1.1
     */
    public Object attachCopy(PersistenceCapable detachedPC, boolean embedded)
    {
        if (attaching)
        {
            return this;
        }
        attaching = true;

        try
        {
            // Check if the object is already persisted
            boolean persistent = false;
            if (embedded)
            {
                persistent = true;
            }
            else
            {
                if (!myPM.getPMFContext().getPmfConfiguration().getAttachSameDatastore())
                {
                    // We cant assume that this was detached from the same datastore, so do a check for existence
                    try
                    {
                        locate();
                        persistent = true;
                    }
                    catch (JDOObjectNotFoundException onfe)
                    {
                        // Not currently present!
                    }
                }
                else
                {
                    persistent = true;
                }
            }

            // Call any "pre-attach" listeners
            callback.preAttach(detachedPC);

            if (detachedPC.jdoIsDeleted())
            {
                // The detached object has been deleted
                myLC = myLC.transitionDeletePersistent(this);
            }

            if (myLC == myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.HOLLOW) ||
                myLC == myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.P_NONTRANS))
            {
                // Make the attached object transactional
                // There may be other states that we come in here with that need making transactional
                myLC = myLC.transitionMakeTransactional(this);
            }

            if (persistent)
            {
                // Make sure that all non-container SCO fields are loaded so we can make valid dirty checks
                // for whether these fields have been updated whilst detached. The detached object doesnt know if the contents
                // have been changed.
                loadSCONonContainerFields();
            }

            // Add a state manager to the detached PC so that we can retrieve its detached state
            StateManagerImpl smDetachedPC = new StateManagerImpl(myPM, myPM.getPMFContext().getApiAdapter(), detachedPC,
                this.cmd, getExternalObjectId(detachedPC), null);
            detachedPC.jdoReplaceStateManager(smDetachedPC);

            // Mark the temporary StateManager with the attached PC object in case it is accessed during attach
            smDetachedPC.attachedPC = myPC;

            // Retrieve the updated values from the detached object
            retrieveDetachState(smDetachedPC);

            if (!persistent)
            {
                // Object is not yet persisted! so make it persistent

                // Make sure all field values in the attach object are ready for inserts (but dont trigger any cascade attaches)
                internalAttachCopy(this, smDetachedPC, smDetachedPC.loadedFields, smDetachedPC.dirtyFields, persistent, smDetachedPC.myVersion, false);

                makePersistent();
            }

            // Go through all related fields and attach them (including relationships)
            internalAttachCopy(this, smDetachedPC, smDetachedPC.loadedFields, smDetachedPC.dirtyFields, persistent, smDetachedPC.myVersion, true);

            // Remove the state manager from the detached PC
            smDetachedPC.attachedPC = null;
            detachedPC.jdoReplaceStateManager(null);

            // Call any "post-attach" listeners
            callback.postAttach(myPC,detachedPC);
        }
        finally
        {
            attaching = false;
        }
        return myPC;
    }

    /**
     * Attach the fields of a persistent object.
     * @param sm StateManager for the attached object.
     * @param smDetached StateManager for the detached object.
     * @param loadedFields Fields that were detached with the object
     * @param dirtyFields Fields that have been modified while detached
     * @param persistent whether the object is already persistent
     * @param version the version
     * @param cascade Whether to cascade the attach to related fields
     * @since 1.1
     */
    private void internalAttachCopy(org.jpox.StateManager sm,
                                   org.jpox.StateManager smDetached,
                                   boolean[] loadedFields,
                                   boolean[] dirtyFields,
                                   boolean persistent,
                                   Object version,
                                   boolean cascade)
    {
        // Need to take all loaded fields plus all modified fields (maybe some werent detached but have been modified) and attach them
        int[] attachFieldNumbers = null;
        int numberOfAttachFields = 0;
        for (int i=0;i<loadedFields.length;i++)
        {
            if (loadedFields[i] || dirtyFields[i])
            {
                numberOfAttachFields++;
            }
        }
        attachFieldNumbers = new int[numberOfAttachFields];
        int n=0;
        for (int i=0;i<loadedFields.length;i++)
        {
            if (loadedFields[i] || dirtyFields[i])
            {
                attachFieldNumbers[n++] = allFieldNumbers[i];
            }
        }
        sm.setVersion(version);
        if (attachFieldNumbers != null)
        {
            // Only update the fields that were detached, and only update them if there are any to update
            smDetached.provideFields(attachFieldNumbers, 
                new AttachFieldManager(sm, secondClassMutableFields, dirtyFields, persistent, cascade));
        }
    }

    /**
     * Method to delete the object from persistence.
     */
    public void deletePersistent()
    {
        if (!myLC.isDeleted())
        {
            // Call any lifecycle listeners waiting for this event.
            callback.preDelete(myPC);

            // Perform any cascade on dependent fields
            srm.deleteDependent(this, allFieldNumbers);

            dirty = true;
            preStateChange();
            try
            {
                myLC = myLC.transitionDeletePersistent(this);
            }
            finally
            {
                postStateChange();
            }
            if (myPM.isDelayDatastoreOperationsEnabled())
            {
                myPM.markDirty(this);
            }
            else
            {
                internalDeletePersistent();
            }

            // Call any lifecycle listeners waiting for this event.
            callback.postDelete(myPC);
        }
    }

    /**
     * Nullify fields with reference to PersistenceCapable or SCO instances 
     */
    public void nullifyFields()
    {
        if (!myLC.isDeleted() && !myPC.jdoIsDetached())
        {
            // Update any relationships for fields of this object that aren't dependent
            replaceFields(allNonPrimaryKeyFieldNumbers, new NullifyRelationFieldManager(this));
            flush();
        }
    }

    /**
     * Method to delete the object from persistence.
     */
    private void internalDeletePersistent()
    {
        if (deleting)
        {
            throw new JDOUserException(LOCALISER.msg("StateManager.DeleteRecursiveError"));
        }

        deleting = true;

        try
        {
            if (dirty)
            {
                clearDirtyFlags();
                /*
                 * Clear the PM's knowledge of our being dirty. This will
                 * call our flush() method, which will do nothing.
                 */
                myPM.flush();
            }

            srm.delete(this);
        }
        finally
        {
            deleting = false;
        }
    }

    /**
     * Validates whether the persistence capable instance exists in the
     * datastore. If the instance does not exist in the datastore, this method
     * will fail raising a javax.jdo.JDOObjectNotFoundException.
     */
    public void validate()
    {
        if (!myLC.isTransactional)
        {
	        int[] fieldNumbers = getFlagsSetTo(loadedFields, defaultFetchGroupFieldNumbers, false);
	        if (fieldNumbers != null)
	        {
	            fieldNumbers = getFlagsSetTo(nonPrimaryKeyFields, fieldNumbers, true);
	        }
	        if (fieldNumbers != null)
	        {
	            fieldNumbers = getFlagsSetTo(secondClassMutableFields, fieldNumbers, false);
	        }
	        if (fieldNumbers != null && fieldNumbers.length > 0)
	        {
                // Load the fields in the current FetchPlan (JDO2 spec 12.6.5)
                transitionReadField();

                loadFieldsInFetchPlan();

                // Note that postLoad is not called here since this is only called as a secondary action for getObjectById()
                // and the object will have been loaded before this
	        }
	        else
	        {
	            locate();
	            transitionReadField();
	        }
        }
    }

    /**
     * Locates this object in the datastore. If the instance does not exist in
     * the datastore, this method will fail raising a
     * javax.jdo.JDOObjectNotFoundException.
     */
    private void locate()
    {
        srm.locate(this);
    }

    // ------------------------- Object State Methods --------------------------

    /**
     * Method to change the object state to evicted.
     **/
    public void evict()
    {
        if (myLC != myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.P_CLEAN) &&
            myLC != myPM.getPMFContext().getApiAdapter().getLifeCycleState(LifeCycleState.P_NONTRANS))
        {
            return;
        }
            
        preStateChange();
        try
        {
            try
            {
                callback.preClear(myPC);

                callback.postClear(myPC);
            }
            finally
            {
                myLC = myLC.transitionEvict(this);
            }
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to refresh the object.
     **/
    public void refresh()
    {
        preStateChange();
        try
        {
            myLC = myLC.transitionRefresh(this);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to retrieve the object.
     * @param fgOnly Only load the current fetch group fields
     **/
    public void retrieve(boolean fgOnly)
    {
        preStateChange();
        try
        {
            myLC = myLC.transitionRetrieve(this, fgOnly);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to retrieve the object.
     * @param fetchPlan the fetch plan to load fields
     **/
    public void retrieve(FetchPlan fetchPlan)
    {
        preStateChange();
        try
        {
            myLC = myLC.transitionRetrieve(this, fetchPlan);
        }
        finally
        {
            postStateChange();
        }
    }
    
    // --------------------------- Process Methods -----------------------------

    /**
     * Method called before a change in state.
     **/
    private void preStateChange()
    {
        changingState = true;
    }

    /**
     * Method called after a change in state.
     **/
    private void postStateChange()
    {
        changingState = false;

        if (postLoadPending && isFetchPlanLoaded())
        {
            // Only call postLoad when all FetchPlan fields are loaded
            postLoadPending = false;
            postLoad();
        }
    }

    /**
     * Method called before a write of the specified field.
     * @param field The field to write
     * @return true if the field was already dirty before
     **/
    private boolean preWriteField(int field)
    {
        boolean wasDirty = dirty;
        /*
         * If we're writing a field in the process of inserting it must be due 
         * to jdoPreStore().  We haven't actually done the INSERT yet so we 
         * don't want to mark anything as dirty, which would make us want to do 
         * an UPDATE later. 
         */
        if (inserting == ActivityState.NONE)
        {
            dirty = true;
            if (!wasDirty) // (only do it for first dirty event).
            {
                // Call any lifecycle listeners waiting for this event
                callback.preDirty(myPC);
            }

            transitionWriteField();

            dirty = true;
            dirtyFields[field] = true;
            loadedFields[field] = true;
        }
        return wasDirty;
    }

    /**
     * Method called after the write of a field.
     * @param wasDirty whether before writing this field the pc was dirty
     */
    private void postWriteField(boolean wasDirty)
    {
        if (dirty && !wasDirty) // (only do it for first dirty event).
        {
            // Call any lifecycle listeners waiting for this event
            callback.postDirty(myPC);
        }
        
        /*
         * If we're writing a field in the middle of flushing or inserting it
         * must be due to jdoPreStore(). To avoid an infinite recursion we don't
         * notify the PM or call flush(). if not transactional transient
         */
        if (inserting == ActivityState.NONE && !flushing && !(myLC.isTransactional && !myLC.isPersistent))
        {
            if (myPM.currentTransaction().isActive())
            {
                myPM.markDirty(this);
            }
            else
            {
                flush();
            }
        }
    }

    /**
     * Called whenever the default fetch group fields have all been loaded.
     * Updates jdoFlags and calls jdoPostLoad() as appropriate.
     * <p>
     * If it's called in the midst of a life-cycle transition both actions will
     * be deferred until the transition is complete.
     * <em>This deferral is important</em>. Without it, we could enter user
     * code (jdoPostLoad()) while still making a state transition, and that way
     * lies madness.
     * <p>
     * As an example, consider a jdoPostLoad() that calls other enhanced methods
     * that read fields (jdoPostLoad() itself is not enhanced). A P_NONTRANS
     * object accessed within a transaction would produce the following infinite
     * loop:
     * <p>
     * <blockquote>
     * 
     * <pre>
     * 
     *  isLoaded()
     *  transitionReadField()
     *  refreshLoadedFields()
     *  jdoPostLoad()
     *  isLoaded()
     *  ...
     *  
     * </pre>
     * 
     * </blockquote>
     * <p>
     * because the transition from P_NONTRANS to P_CLEAN can never be completed.
     */
    private void postLoad()
    {
        if (changingState)
        {
            postLoadPending = true;
        }
        else
        {
            /*
             * A transactional object whose DFG fields are loaded does not need
             * to contact us in order to read those fields, so we can safely set
             * READ_OK.
             *
             * A non-transactional object needs to notify us on all field reads
             * so that we can decide whether or not any transition should occur,
             * so we leave the flags at LOAD_REQUIRED.
             */

            if (jdoFlags == PersistenceCapable.LOAD_REQUIRED && myLC.isTransactional())
            {
                jdoFlags = PersistenceCapable.READ_OK;
                myPC.jdoReplaceFlags();
            }

            callback.postLoad(myPC);
        }
    }

    /**
     * This method is invoked when a commit is performed in a Transaction
     * involving the PersistenceCapable managed by this StateManager
     * @param tx The Transaction
     */
    public void postCommit(Transaction tx)
    {
        preStateChange();
        try
        {
            myLC = myLC.transitionCommit(this, tx);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * This method is invoked when a rollback is performed in a Transaction
     * involving the PersistenceCapable managed by this StateManager.
     * @param tx The Transaction been rolled back
     */
    public void preRollback(javax.jdo.Transaction tx)
    {
        preStateChange();
        try
        {
            myPM.clearDirty(this);
            myLC = myLC.transitionRollback(this,tx);
        }
        finally
        {
            postStateChange();
        }
    }

    /** Flag to signify that we are currently storing the PC object, so we dont detach it on any serialisation. */
    private boolean storingPC = false;

    /**
     * Method to set the storing PC flag.
     */
    public void setStoringPC()
    {
        storingPC = true;
    }

    /**
     * Method to unset the storing PC flag.
     */
    public void unsetStoringPC()
    {
        storingPC = false;
    }

    /**
     * Guarantee that the serializable transactional and persistent fields
     * are loaded into the instance.  This method is called by the generated
     * jdoPreSerialize method prior to serialization of the instance.
     *
     * @param pc the calling PersistenceCapable instance
     */
    public void preSerialize(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return;
        }

        // Retrieve all fields prior to serialisation
        retrieve(false);

        if (myPM.currentTransaction().isActive())
        {
            myLC = myLC.transitionSerialize(this);
        }

        if (!storingPC && pc instanceof Detachable)
        {
            if (!myLC.isDeleted && myLC.isPersistent)
            {
                if (myLC.isDirty)
                {
                    flush();
                }

                // Normal PC Detachable object being serialised so load up the detached state into the instance
                // JDO2 spec "For Detachable classes, the jdoPreSerialize method must also initialize the jdoDetachedState
                // instance so that the detached state is serialized along with the instance."
                ((Detachable)pc).jdoReplaceDetachedState();
            }
        }
    }

    /**
     * Flushes any dirty fields to the data store.
     *
     * <p>Note that "dirty" in this case is not equated to being in the P_DIRTY
     * state.  The P_DIRTY state means that at least one field in the object has
     * been written by the user during the current transaction, whereas for the
     * purposes of this method, a field is "dirty" if it's been written by the
     * user but not yet updated in the data store.  The difference is, it's
     * possible for an object's state to be P_DIRTY, yet have no "dirty" fields
     * because flush() has been called at least once during the transaction.
     */
    public void flush()
    {
        if (dirty)
        {
            if (flushing)
            {
                // In the case of persisting a new object using autoincrement id within an optimistic
                // transaction, flush() will initially be called at the point of recognising that the
                // id is generated in the datastore, and will then be called again at the point of doing
                // the InsertRequest for the object itself. Just return since we are flushing right now
                return;
            }

            flushing = true;
            try
            {
                if (myLC.stateType() == LifeCycleState.P_NEW && !flushedNew)
                {
                    if (!isEmbedded())
                    {
                        // internalMakePersistent does preStore, postStore
                        internalMakePersistent();
                    }
                    else
                    {
                        callback.preStore(myPC);

                        callback.postStore(myPC);
                    }
                    dirty = false;
                }
                else if (myLC.stateType() == LifeCycleState.P_DELETED)
                {
                    callback.preDelete(myPC);
                    if (!isEmbedded())
                    {
                        internalDeletePersistent();
                    }
                    callback.postDelete(myPC);
                }
                else if (myLC.stateType() == LifeCycleState.P_NEW_DELETED)
                {
                    if (flushedNew)
                    {
                        callback.preDelete(myPC);
                        if (!isEmbedded())
                        {
                            internalDeletePersistent();
                        }
                        callback.postDelete(myPC);
                    }
                    else
                    {
                        dirty = false;
                    }
                }
                else
                {
                    if (!deleting)
                    {
                        callback.preStore(myPC);
                    }

                    int[] dirtyFieldNumbers = getFlagsSetTo(dirtyFields, true);
                    if (dirtyFieldNumbers == null)
                    {
                        throw new JDOFatalInternalException(LOCALISER.msg("StateManager.InconsistentDirtyStateError"));
                    }
                    if (!isEmbedded())
                    {
                        srm.update(this, dirtyFieldNumbers);

                        // Update the object in the cache(s)
                        myPM.putObjectIntoCache(this, true, true);
                    }

                    clearDirtyFlags();

                    callback.postStore(myPC);
                }
            }
            finally
            {
                flushing = false;
            }
        }
    }

    /**
     * Initialize SM reference in PC and Oid
     * @param newState The new StateManager state 
     **/
    private void initializeSM(int newState)
    {
        final StateManagerImpl thisSM = this;
        myLC = myPM.getPMFContext().getApiAdapter().getLifeCycleState(newState);

        try
        {
            if (myLC.isPersistent())
            {
                myPM.addStateManager(this);
            }

            // Everything OK so far. Now we can set SM reference in PC 
            // It can be done only after myLC is set to deligate validation
            // to the LC and objectId verified for uniqueness
            AccessController.doPrivileged(
            // Need to have privileges to perform jdoReplaceStateManager.
                    new PrivilegedAction()
                    {
                        public Object run()
                        {
                            try
                            {
                                myPC.jdoReplaceStateManager(thisSM);
                                return null;
                            }
                            catch (SecurityException e)
                            {
                                throw new JDOFatalUserException("EXC_CannotSetStateManager", e);
                            }
                        }
                    });

        }
        catch (SecurityException e)
        {
            throw new JDOUserException(e.getMessage());

        }
        catch (JDOException jdoException)
        {
            if (myPM.getStateManagerById(myID) == this)
            {
                myPM.removeStateManager(this);
            }
            throw jdoException;
        }
    }

    /**
     * Method to disconnect any cloned persistence capable objects from their
     * StateManager.
     * @param pc The PersistenceCapable object
     * @return Whether the object was disconnected.
     **/
    private boolean disconnectClone(PersistenceCapable pc)
    {
        if (detaching)
        {
            return false;
        }
        if (pc != myPC)
        {
            if (JPOXLogger.JDO.isDebugEnabled())
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("StateManager.DisconnectClone", StringUtils.toJVMIDString(pc), this));
            }

            // Reset jdoFlags in the clone to PersistenceCapable.READ_WRITE_OK 
            // and clear its state manager.
            pc.jdoReplaceFlags();
            pc.jdoReplaceStateManager(null);
            return true;
        }
        else
        {
            return false;
        }
    }

    /**
     * Convenience method to unset the owners of all SCO fields in the PC object.
     */
    private void unsetOwnerInSCOFields()
    {
        // Call unsetOwner() on all loaded SCO fields.
        int[] fieldNumbers = getFlagsSetTo(loadedFields, secondClassMutableFieldNumbers, true);
        if (fieldNumbers != null)
        {
            provideFields(fieldNumbers, new UnsetOwners());
        }        
    }

    /**
     * Disconnect the StateManager from the PersistenceManager and PC object.
     */
    public void disconnect()
    {
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            JPOXLogger.JDO.debug(LOCALISER.msg("StateManager.Disconnect", StringUtils.toJVMIDString(myPC), this));
        }

        //we are transitioning to TRANSIENT state, so if any postLoad
        //action is pending we do it before. This usually happens when
        //we make transient instances using the fetch plan and some
        //fields were loaded during this action which triggered a
        //jdoPostLoad event
        if (postLoadPending)
        {
            changingState = false; //hack to make sure postLoad does not return without processing
            postLoadPending = false;
            postLoad();
        }
        
        unsetOwnerInSCOFields();

        myPM.removeStateManager(this);
        jdoFlags = PersistenceCapable.READ_WRITE_OK;
        myPC.jdoReplaceFlags();

        disconnecting = true;
        try
        {
            replaceStateManager(null);
        }
        finally
        {
            disconnecting = false;
        }

        clearSavedFields();
        myPM = null;
        myFP = null;
        myPC = null;
        myID = null;
        myLC = null;
        cmd = null;
        srm = null;
    }

    /**
     * Registers the pc class in the cache
     */
    public void registerTransactional()
    {
        myPM.addStateManager(this);
    }

    // ------------------------------ Detach Methods ---------------------------

    /** State variable as to whether we are retrieving detached state from the detached object. */
    private boolean retrievingDetachedState = false;

    /** State variable as to whether we are resetting the detached state. */
    private boolean resettingDetachedState = false;

    /**
     * Convenience method to retrieve the detach state from the passed State Manager's object
     * @param sm The State Manager
     */
    public void retrieveDetachState(org.jpox.StateManager sm)
    {
        if (sm.getObject() instanceof Detachable)
        {
            ((StateManagerImpl)sm).retrievingDetachedState = true;
            ((Detachable)sm.getObject()).jdoReplaceDetachedState();
            ((StateManagerImpl)sm).retrievingDetachedState = false;
        }
    }

    /**
     * Convenience method to reset the detached state in the current object.
     */
    public void resetDetachState()
    {
        if (getObject() instanceof Detachable)
        {
            resettingDetachedState = true;
            try
            {
                ((Detachable)getObject()).jdoReplaceDetachedState();
            }
            finally
            {
                resettingDetachedState = false;
            }
        }
    }

    /**
     * Method to update the "detached state" in the detached object
     * to obtain the "detached state" from the detached object, or
     * to reset it (to null).
     * @param pc The PersistenceCapable beind updated
     * @param currentState The current state values
     * @return The detached state to assign to the object
     */
    public Object[] replacingDetachedState(Detachable pc, Object[] currentState)
    {
        if (resettingDetachedState)
        {
            return null;
        }
        else if (retrievingDetachedState)
        {
            // Retrieving the detached state from the detached object
            // Don't need the id or version since they can't change
            jdoLoadedFields = (BitSet)currentState[2];
            for (int i = 0; i < this.loadedFields.length; i++)
            {
                this.loadedFields[i] = jdoLoadedFields.get(i);
            }

            jdoModifiedFields = (BitSet)currentState[3];
            for (int i = 0; i < dirtyFields.length; i++)
            {
                dirtyFields[i] = jdoModifiedFields.get(i);
            }
            setVersion(currentState[1]);
            return currentState;
        }
        else
        {
            // Updating the detached state in the detached object with our state
            Object[] state = new Object[4];
            state[0] = myID;
            state[1] = myVersion;

            // Loaded fields
            if (currentState == null || currentState[2] == null)
            {
                BitSet loadedState = new BitSet();
                for (int i = 0; i < loadedFields.length; i++)
                {
                    if (loadedFields[i])
                    {
                        loadedState.set(i);
                    }
                    else
                    {
                        loadedState.clear(i);
                    }
                }
                state[2] = loadedState;
            }
            else
            {
                // TODO Why are we just passing the same state back ?
                state[2] = currentState[2];
            }

            // Modified fields
            if (currentState == null || currentState[3] == null)
            {
                BitSet modifiedState = new BitSet();
                for (int i = 0; i < dirtyFields.length; i++)
                {
                    if (dirtyFields[i])
                    {
                        modifiedState.set(i);
                    }
                    else
                    {
                        modifiedState.clear(i);
                    }
                }
                state[3] = modifiedState;
            }
            else
            {
                // TODO Why are we just passing the same state back ?
                state[3] = currentState[3];
            }
            return state;
        }
    }

    // ------------------------------ Helper Methods ---------------------------

    /**
     * Compares two objects for equality, where one or both of the object
     * references may be null.
     *
     * @return  <code>true</code> if the objects are both <code>null</code> or
     *          compare equal according to their equals() method,
     *          <code>false</code> otherwise.
     */
    private static boolean equals(Object o1, Object o2)
    {
        return o1 == null ? (o2 == null) : o1.equals(o2);
    }

    /**
     * Utility to clear the supplied flags.
     * @param flags
     **/
    private static void clearFlags(boolean[] flags)
    {
        for (int i = 0; i < flags.length; i++)
        {
            flags[i] = false;
        }
    }

    /**
     * Utility to clear the supplied flags.
     * @param flags
     * @param fields fields numbers where the flags will be cleared
     **/
    private static void clearFlags(boolean[] flags, int[] fields)
    {
        for (int i = 0; i < fields.length; i++)
        {
            flags[fields[i]] = false;
        }
    }
    
    /**
     * Returns an array of integers containing the indices of all elements in
     * <tt>flags</tt> that are in the <tt>state</tt> passed as argument.
     * @param flags Array of flags (true or false)
     * @param state The state to search (true or false)
     * @return The settings of the flags
     */
    private static int[] getFlagsSetTo(boolean[] flags, boolean state)
    {
        int[] temp = new int[flags.length];
        int j = 0;

        for (int i = 0; i < flags.length; i++)
        {
            if (flags[i] == state)
            {
                temp[j++] = i;
            }
        }

        if (j != 0)
        {
            int[] fieldNumbers = new int[j];
            System.arraycopy(temp, 0, fieldNumbers, 0, j);

            return fieldNumbers;
        }
        else
        {
            return null;
        }
    }

    /**
     * Returns an array of integers containing the indices of all elements in
     * <tt>flags</tt> whose index occurs in <tt>indices</tt> and whose value is
     * <tt>state</tt>.
     */
    private static int[] getFlagsSetTo(boolean[] flags, int[] indices, boolean state)
    {
        int[] temp = new int[indices.length];
        int j = 0;

        for (int i = 0; i < indices.length; i++)
        {
            if (flags[indices[i]] == state)
            {
                temp[j++] = indices[i];
            }
        }

        if (j != 0)
        {
            int[] fieldNumbers = new int[j];
            System.arraycopy(temp, 0, fieldNumbers, 0, j);

            return fieldNumbers;
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Utility to take a peek at a field in the PersistenceCapable object.
     * @param obj The PersistenceCapable object
     * @param fieldName The field to peek at
     * @return The value of the field.
     **/
    private static Object peekField(Object obj, String fieldName)
    {
        try
        {
            /*
             * This doesn't work due to security problems but you get the idea.
             * I'm trying to get field values directly without going through
             * the provideField machinery.
             */
            Object value = obj.getClass().getDeclaredField(fieldName).get(obj);
            if (value instanceof PersistenceCapable)
            {
                return StringUtils.toJVMIDString(value);
            }
            else
            {
                return value;
            }
        }
        catch (Exception e)
        {
            return e.toString();
        }
    }

    /**
     * Method to dump a PersistenceCapable object to the specified
     * PrintWriter.
     * @param pc The PersistenceCapable object
     * @param out The PrintWriter
     **/
    private static void dumpPC(PersistenceCapable pc, PrintWriter out)
    {
        out.println(StringUtils.toJVMIDString(pc));

        if (pc == null)
        {
            return;
        }

        out.print("jdoStateManager = " + peekField(pc, "jdoStateManager"));
        out.print("jdoFlags = ");
        Object flagsObj = peekField(pc, "jdoFlags");
        if (flagsObj instanceof Byte)
        {
            out.println(StringUtils.jdoFlagsToString(((Byte) flagsObj).byteValue()));
        }
        else
        {
            out.println(flagsObj);
        }

        Class c = pc.getClass();

        do
        {
            String[] fieldNames = HELPER.getFieldNames(c);

            for (int i = 0; i < fieldNames.length; ++i)
            {
                out.print(fieldNames[i]);
                out.print(" = ");
                out.println(peekField(pc, fieldNames[i]));
            }

            c = c.getSuperclass();
        }
        while (c != null && PersistenceCapable.class.isAssignableFrom(c));
    }

    /**
     * Utility to dump the contents of the StateManager.
     *
     * @param out PrintWriter to dump to
     **/
    public void dump(PrintWriter out)
    {
        out.println("myPM = " + myPM);
        out.println("myID = " + myID);
        out.println("myLC = " + myLC);
        out.println("cmd = " + cmd);
        out.println("srm = " + srm);
        out.println("fieldCount = " + fieldCount);
        out.println("dirty = " + dirty);
        out.println("flushing = " + flushing);
        out.println("changingState = " + changingState);
        out.println("postLoadPending = " + postLoadPending);
        out.println("disconnecting = " + disconnecting);
        out.println("dirtyFields = " + StringUtils.booleanArrayToString(dirtyFields));
        out.println("defaultFetchGroupFields = " + StringUtils.booleanArrayToString(defaultFetchGroupFields));
        out.println("secondClassMutableFields = " + StringUtils.booleanArrayToString(secondClassMutableFields));
        out.println("allFieldNumbers = " + StringUtils.intArrayToString(allFieldNumbers));
        out.println("defaultFetchGroupFieldNumbers = " + StringUtils.intArrayToString(defaultFetchGroupFieldNumbers));
        out.println("secondClassMutableFieldNumbers = " + StringUtils.intArrayToString(secondClassMutableFieldNumbers));

        out.println();
        out.println("jdoFlags = " + StringUtils.jdoFlagsToString(jdoFlags));
        out.println("loadedFields = " + StringUtils.booleanArrayToString(loadedFields));
        out.print("myPC = ");
        dumpPC(myPC, out);

        out.println();
        out.println("savedFlags = " + StringUtils.jdoFlagsToString(savedFlags));
        out.println("savedLoadedFields = " + StringUtils.booleanArrayToString(savedLoadedFields));

        out.print("savedImage = ");
        dumpPC(savedImage, out);
    }
}
