/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2002 Mike Martin - unknown changes
2003 Andy Jefferson - commented
2003 Erik Bengtson - optimistic transaction
    ...
**********************************************************************/
package org.jpox.state.jpa;

import javax.jdo.FetchPlan;
import javax.jdo.JDOUserException;
import javax.jdo.Transaction;

import org.jpox.state.IllegalStateTransitionException;
import org.jpox.state.LifeCycleState;
import org.jpox.StateManager;


/**
 * Class representing the life cycle state of PersistentNontransactional
 *
 * @version $Revision: 1.2.2.1 $
 **/
class PersistentNontransactional extends LifeCycleState
{
    /** Protected Constructor to prevent external instantiation. */
    protected PersistentNontransactional()
    {
		isPersistent = true;    	
        isDirty = false;
        isNew = false;
        isDeleted = false;
        isTransactional = false;

        stateType = P_NONTRANS;
    }

    /**
     * Method to transition to delete persistent.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionDeletePersistent(StateManager sm)
    {
        return changeState(sm, P_DELETED);
    }

    /**
     * Method to transition to transactional.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakeTransactional(StateManager sm)
    {
        sm.refreshLoadedFields();
        return changeState(sm, P_CLEAN);
    }

    /**
     * Method to transition to transient.
     * @param sm StateManager.
     * @param useFetchPlan to make transient the fields in the fetch plan
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakeTransient(StateManager sm, boolean useFetchPlan)
    {
        if (useFetchPlan)
        {
            sm.loadUnloadedFieldsInFetchPlan();
        }
        return changeState(sm, TRANSIENT);
    }

    /**
     * Method to transition to commit state.
     * @param sm StateManager.
     * @param tx the Transaction been committed.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionCommit(StateManager sm, Transaction tx)
    {
        throw new IllegalStateTransitionException(this, "commit", sm);
    }

    /**
     * Method to transition to rollback state.
     * @param sm StateManager.
     * @param tx The transaction
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRollback(StateManager sm,Transaction tx)
    {
        throw new IllegalStateTransitionException(this, "rollback", sm);
    }

    /**
     * Method to transition to refresh state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRefresh(StateManager sm)
    {
        sm.refreshFieldsInFetchPlan();
        return this;
    }
    
    /**
     * Method to transition to evict state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionEvict(StateManager sm)
    {
        sm.clearNonPrimaryKeyFields();
        sm.clearSavedFields();
        return changeState(sm, HOLLOW);
    }

    /**
     * Method to transition to read-field state.
     * @param sm StateManager.
     * @param isLoaded if the field was previously loaded.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionReadField(StateManager sm, boolean isLoaded)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
		if (!tx.isActive() && !tx.getNontransactionalRead())
		{
	        throw new JDOUserException(LOCALISER.msg("LifeCycle.PersistentNontransactionalReadFieldNotAllowedNotActiveTx"),sm.getInternalObjectId());
		}
        if (tx.isActive() && ! tx.getOptimistic())
        {
            // Save the fields for rollback.
            sm.saveFields();
            sm.refreshLoadedFields();
            return changeState(sm, P_CLEAN);
        }
        else
        {
            return this;
        }
    }

    /**
     * Method to transition to write-field state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionWriteField(StateManager sm)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
        if (tx.isActive())
        {
            // Save the fields for rollback.
            sm.saveFields();

            // This line was present previously (from 1.1.0-beta-4 approx until 1.1.4) but resulted in optimistic tests 
            // failing due to refreshing all fields (including version) from the datastore when only one was updated
            // No idea why it was here since it had no comment to explain its purpose but is now commented out
            // If it is required please provide a comment what is its role relative to the spec
//            sm.refreshLoadedFields();

            return changeState(sm, P_DIRTY);
        }
        else
        {
            return this;
        }
    }

    /**
     * Method to transition to retrieve state.
     * @param sm StateManager.
	 * @param fgOnly only the current fetch group fields
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRetrieve(StateManager sm, boolean fgOnly)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
        if (tx.isActive() && !tx.getOptimistic())
        {
            // Save the fields for rollback.
            sm.saveFields();
    		if (fgOnly)
            {
                sm.loadUnloadedFieldsInFetchPlan();
            }
    		else
            {
    			sm.loadUnloadedFields();
            }             
            return changeState(sm, P_CLEAN);
        }
        else if (tx.isActive() && tx.getOptimistic())
        {
            // Save the fields for rollback.
            sm.saveFields();
    		if (fgOnly)
            {
                sm.loadUnloadedFieldsInFetchPlan();
            }
    		else
            {
    			sm.loadUnloadedFields();
            }
    		return this;
        }
        else
        {
    		if (fgOnly)
            {
                sm.loadUnloadedFieldsInFetchPlan();
            }
    		else
            {
    			sm.loadUnloadedFields();
            }
    		return this;
        }
    }

    /**
     * Method to transition to retrieve state.
     * @param sm StateManager.
     * @param fetchPlan the fetch plan to load fields
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRetrieve(StateManager sm, FetchPlan fetchPlan)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
        if (tx.isActive() && !tx.getOptimistic())
        {
            // Save the fields for rollback.
            sm.saveFields();
            sm.loadUnloadedFieldsOfClassInFetchPlan(fetchPlan);
            return changeState(sm, P_CLEAN);
        }
        else if (tx.isActive() && tx.getOptimistic())
        {
            // Save the fields for rollback.
            sm.saveFields();
            sm.loadUnloadedFieldsOfClassInFetchPlan(fetchPlan);
            return this;
        }
        else
        {
            sm.loadUnloadedFieldsOfClassInFetchPlan(fetchPlan);
            return this;
        }
    }

    /**
     * Method to transition when serialised.
     * @param sm State Manager
     * @return The new LifeCycle state
     */
    public LifeCycleState transitionSerialize(StateManager sm)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
        if (tx.isActive() && !tx.getOptimistic())
        {
            return changeState(sm, P_CLEAN);
        }
        return this;
    }

    /**
     * Method to return a string version of this object.
     * @return The string "P_NONTRANS".
     **/
    public String toString()
    {
        return "P_NONTRANS";
    }
}