/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.store;

import javax.jdo.JDOFatalUserException;

import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ExtendableMetaData;
import org.jpox.util.Localiser;

/**
 * Representation of a class/field managed by the StoreManager.
 * <H3>Type of data</H3>
 * There are 2 constructors, one for specifying a class (FCO) being managed, and one
 * for specifying a field (SCO) being managed. The "type" is set accordingly to represent
 * which one is being used. 
 * <H3>Ownership of Tables</H3>
 * In JPOX, a class can have a table assigned where the class is not the "owner" of the table.
 * This happens when the class has its fields persisted in the table of another class
 * - so they share the table.
 *
 * @version $Revision: 1.10 $
 */
public class StoreData
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.Localisation");

    /** First class object (FCO) type */
    public static final int FCO_TYPE = 1;

    /** Second class object (SCO) type */
    public static final int SCO_TYPE = 2;

    /** Name of the class/field. */
    protected final String name;

    /** If this class is an implementation of a persistent interface, this attribute will hold the name of the interface. */
    protected final String interfaceName;
    
    /** Name of the datastore object where it is stored. */
    protected String tableName;

    /** Whether this class is the owner of the table. */
    protected final boolean isTableOwner;

    /** Type of data being stored. */
    protected final int type;

    /** MetaData for this object. */
    protected final ExtendableMetaData metaData;

    /** The datastore container (table) where this class/field is persisted. */
    protected DatastoreContainerObject table;

    /**
     * Constructor. To be used when creating for the start mechanism.
     * @param name Name of the class/field
     * @param tableName Name of the table associated
     * @param tableOwner Whether this is the owner
     * @param type The type (FCO/SCO)
     * @param interfaceName if this class is an implementation of a persistent interface (multiple persistent interface names are comma separated), otherwise is null.
     */
    public StoreData(String name, String tableName, boolean tableOwner, int type, String interfaceName)
    {
        this.metaData = null;
        this.name = name;
        this.tableName = tableName;
        this.isTableOwner = tableOwner;
        this.type = type;
        this.interfaceName = interfaceName;
    }

    /**
     * Constructor for FCO data.
     * @param cmd MetaData for the class.
     * @param table Table where the class is stored.
     * @param tableOwner Whether the class is the owner of the table.
     */
    public StoreData(ClassMetaData cmd, DatastoreContainerObject table, boolean tableOwner)
    {
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StoreData.MetaDataNullError", tableName));
        }
        this.name = cmd.getFullClassName();
        this.table = table;
        this.tableName = (table != null ? table.toString() : null);
        this.isTableOwner = tableOwner;
        this.metaData = cmd;
        this.type = FCO_TYPE;
        String interfaces = null;
        for( int i=0; i<cmd.getImplementsMetaData().length; i++)
        {
        	if( interfaces==null )
        	{
        		interfaces = "";
        	}
        	else
        	{
        		interfaces += ",";
        	}
        	interfaces += cmd.getImplementsMetaData()[i].getName();
        }
        this.interfaceName = interfaces;
        
    }

    /**
     * Constructor for SCO data.
     * @param fmd MetaData for the field.
     * @param table Table where the field is stored.
     */
    public StoreData(AbstractPropertyMetaData fmd, DatastoreContainerObject table)
    {
        if (fmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StoreData.MetaDataNullError", tableName));
        }
        this.name = fmd.getFullFieldName();
        this.table = table;
        this.tableName = (table != null ? table.toString() : null);
        this.isTableOwner = true; // Always the owner of the table
        this.metaData = fmd;
        this.type = SCO_TYPE;
        this.interfaceName = table.getStoreManager().getMetaDataManager().isPersistentInterface(fmd.getType().getName()) ? fmd.getType().getName() : null;
    }

    /**
     * Accessor for class/field name.
     * @return Returns the class/field name.
     */
    public String getName()
    {
        return name;
    }


    /**
     * Accessor for the persistent interface name
     * @return Returns the persistent interface name
     */
    public String getInterfaceName()
    {
        return interfaceName;
    }

    /**
     * Accessor for tableName.
     * @return Returns the tableName.
     */
    public String getTableName()
    {
        return tableName;
    }

    /**
     * Accessor for whether this class is the owner of the table.
     * @return Whether it owns the table
     */
    public boolean isTableOwner()
    {
        return isTableOwner;
    }

    /**
     * Accessor for whether this has a table representation.
     * @return Whether it has a table
     */
    public boolean hasTable()
    {
        return (tableName != null);
    }

    /**
     * Accessor for whether this represents FCO data
     * @return Whether it is FCO
     */
    public boolean isFCO()
    {
        return type == FCO_TYPE;
    }

    /**
     * Accessor for whether this represents SCO data.
     * @return Whether it is SCO.
     */
    public boolean isSCO()
    {
        return type == SCO_TYPE;
    }

    /**
     * Accessor for type.
     * @return Returns the type.
     */
    public int getType()
    {
        return type;
    }
    
    /**
     * Accessor for any superclass for this class, if this is a class.
     * @return The superclass (if exists). If this is a field, just return null.
     **/
    public String getSuperclass()
    {
        if (metaData instanceof ClassMetaData)
        {
            return ((ClassMetaData)metaData).getPersistenceCapableSuperclass();
        }
        else
        {
            return null;
        }
    }

    /**
     * Accessor for class/field meta data.
     * @return Returns the class/field meta data.
     */
    public ExtendableMetaData getMetaData()
    {
        return metaData;
    }

    /**
     * Accessor for the Table details.
     * @return The Table
     **/
    public DatastoreContainerObject getDatastoreContainerObject()
    {
        return table;
    }

    /**
     * Accessor for the identifier for the table.
     * @return The table identifier
     **/
    public DatastoreIdentifier getDatastoreIdentifier()
    {
        if (table != null)
        {
            return table.getIdentifier();
        }

        // If the class doesn't map to a table return null
        return null;
    }

    /**
     * Convenience to set the table. To be used in cases where the table isn't known
     * until after the initial create
     * @param table The table
     */
    public void setDatastoreContainerObject(DatastoreClass table)
    {
        if (table != null)
        {
            tableName = table.toString();
            this.table = table;
        }
    }

    /**
     * Method to return this class/field managed object as a string.
     * @return String version of this class/field managed object.
     **/
    public String toString()
    {
        if (metaData instanceof ClassMetaData)
        {
            ClassMetaData cmd = (ClassMetaData)metaData;
            if (tableName != null)
            {
                return LOCALISER.msg("StoreData.DataForClass", name, tableName,
                    cmd.getInheritanceMetaData().getStrategyValue().toString());
            }
            else
            {
                return LOCALISER.msg("StoreData.DataForClass", name, "(none)",
                    cmd.getInheritanceMetaData().getStrategyValue().toString());
            }
        }
        else if (metaData instanceof AbstractPropertyMetaData)
        {
            return LOCALISER.msg("StoreData.DataForField", name, tableName);
        }
        else
        {
            return LOCALISER.msg("StoreData.DataForClass", name, tableName);
        }
    }
}
