/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store;

import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.Extent;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;
import javax.jdo.Query;
import javax.jdo.datastore.Sequence;
import javax.jdo.identity.SingleFieldIdentity;
import javax.jdo.spi.PersistenceCapable;
import javax.resource.spi.security.PasswordCredential;
import javax.sql.DataSource;

import org.jpox.AbstractPersistenceManagerFactory;
import org.jpox.ClassLoaderResolver;
import org.jpox.PMFContext;
import org.jpox.PersistenceManager;
import org.jpox.PersistenceManagerImpl;
import org.jpox.StateManager;
import org.jpox.Transaction;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ExtensionMetaData;
import org.jpox.metadata.IdentityStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.InheritanceStrategy;
import org.jpox.metadata.MetaDataManager;
import org.jpox.metadata.SequenceMetaData;
import org.jpox.resource.ManagedConnectionImpl;
import org.jpox.store.exceptions.DatastoreInitialisationException;
import org.jpox.store.exceptions.DatastorePermissionException;
import org.jpox.store.exceptions.NoTableManagedException;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.poid.PoidManager;
import org.jpox.store.scostore.ArrayStore;
import org.jpox.store.scostore.CollectionStore;
import org.jpox.store.scostore.MapStore;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.MacroString;
import org.jpox.util.MultiMap;
import org.jpox.util.StringUtils;

/**
 * An abstract representation of a Store Manager.
 * Manages the persistence of objects to the store.
 * Will be implemented for the type of datastore (RDBMS, ODBMS, OLAP) in question. 
 * The store manager's responsibilities include:
 * <ul>
 * <li>Creating and/or validating datastore tables according to the persistent classes being 
 * accessed by the application.</li>
 * <li>Serving as the primary intermediary between StateManagers and the database.</li>
 * <li>Serving as the base Extent and Query factory.</li>
 * </ul>
 * <p>
 * A store manager's knowledge of its contents is typically not complete. It knows about
 * the classes that it has encountered in its lifetime. The PersistenceManager can make the
 * StoreManager aware of a class, and can check if the StoreManager knows about a particular class.
 * The Auto-Start mechanism provides a way of inheriting knowledge from the last time the store
 * was used.
 *
 * @version $Revision: 1.84.2.4 $
 */
public abstract class StoreManager
{
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.Localisation");
    
    /** Datasource */
    protected final DataSource[] ds;

    /** Non transactional Datasource */
    protected final DataSource[] dsNonTx;
    
    /** Adapter for the datastore being used. */
    protected DatastoreAdapter dba;

    /** Username for the datastore being used. */
    protected final String userName;

    /** Password for the datastore being used. */
    protected final String password;

    /** Whether this datastore is read only. */
    protected final boolean readOnlyDatastore;

    /** What should happen when readonly and an update occurs. */
    protected final String readOnlyDatastoreAction;

    /** Whether this datastore is fixed (no mods to datastore classes allowed). */
    protected final boolean fixedDatastore;
    
    /** Auto-Start mechanism to use. */
    protected AutoStartMechanism starter = null;

    /** Whether the AutoStart mechanism is initialised */
    protected boolean starterInitialised = false;

    /** PMF context. */
    protected final PMFContext pmfContext;

    /** Whether to auto create any tables. */
    protected final boolean autoCreateTables;

    /** Whether to auto create any columns that are missing. */
    protected final boolean autoCreateColumns;

    /** Whether to auto create any constraints */
    protected final boolean autoCreateConstraints;

    /** Whether to warn only when any errors occur on auto-create. */
    protected final boolean autoCreateWarnOnError;

    /** Whether to validate any tables */
    protected final boolean validateTables;

    /** Whether to validate any columns */
    protected final boolean validateColumns;

    /** Whether to validate any constraints */
    protected final boolean validateConstraints;
    
    /** Whether to check if table/view exists */
    protected final boolean checkExistTablesOrViews;    

    /** Manager for identity generation. */
    protected final PoidManager poidManager;

    /** Factory for identifiers for this datastore. */
    protected IdentifierFactory identifierFactory;

    /** Map of all managed store data, keyed by the class/field name. */
    protected Map storeDataByClass = new HashMap();

    /** Map of all managed store data using Application Identity, keyed by the app id PK class */
    protected MultiMap storeDataByAppIdClass = new MultiMap();

    /**
     * Map of all managed datastore containers (tables) keyed by the datastore identifier.
     * Only currently used for storing SequenceTable.
     */
    protected Map datastoreContainerByIdentifier = new HashMap();

    /**
     * Constructor for a new StoreManager. Stores the basic information
     * required for the datastore management.
     * @param clr the ClassLoaderResolver
     * @param pmf The corresponding PersistenceManagerFactory. This factory's
     *            non-transactional data source will be used to get database
     *            connections as needed to perform management functions.
     * @param userName The database user name.
     * @param password The database user's password.
     * @see StoreManagerFactory
     */
    protected StoreManager(ClassLoaderResolver clr, 
                           AbstractPersistenceManagerFactory pmf, 
                           String userName,
                           String password)
    {
        ds = pmf.getTransactionalDataSource();
        dsNonTx = pmf.getNontransactionalDataSource();
        this.userName = userName;
        this.password = password;
        this.pmfContext = pmf.getPMFContext();
        this.readOnlyDatastore = pmf.getReadOnlyDatastore();
        this.readOnlyDatastoreAction = pmf.getReadOnlyDatastoreAction();
        this.fixedDatastore = pmf.getFixedDatastore();

        this.poidManager = new PoidManager();

        autoCreateColumns = pmf.getPMFConfiguration().getAutoCreateColumns();
        autoCreateTables = pmf.getPMFConfiguration().getAutoCreateTables();
        autoCreateConstraints = pmf.getPMFConfiguration().getAutoCreateConstraints();
        autoCreateWarnOnError = pmf.getPMFConfiguration().getAutoCreateWarnOnError();

        validateTables = pmf.getPMFConfiguration().getValidateTables();
        validateColumns = pmf.getPMFConfiguration().getValidateColumns();
        validateConstraints = pmf.getPMFConfiguration().getValidateConstraints();

        checkExistTablesOrViews = pmf.getPMFConfiguration().getCheckExistTablesOrViews();
    }

	/**
	 * Release of resources
	 */
	public void close()
	{
        poidManager.clear();
		storeDataByClass.clear();
		storeDataByAppIdClass.clear();
        starterInitialised = false;
        starter = null;
        dba = null;
    }

    /**
     * Method to obtain a nonmanaged transaction for the specified persistence manager for this store.
     * @param pm The PersistenceManager
     * @param userName The username
     * @param password The password
     * @return The transaction
     */
    public abstract Transaction getNonmanagedTransaction(PersistenceManagerImpl pm, String userName, String password);

    /**
     * Method to obtain a managed transaction for the specified persistence manager for this store.
     * @param pm The PersistenceManager implementation for managed environment
     * @param credential Password credential
     * @return The transaction
     */
    public abstract Transaction getManagedTransaction(ManagedConnectionImpl pm, PasswordCredential credential);

    /**
     * Method to return a datastore sequence for this datastore matching the passed sequence MetaData.
     * @param pm The PersistenceManager
     * @param seqmd SequenceMetaData
     * @return The Sequence
     */
    public abstract Sequence getSequence(PersistenceManager pm, SequenceMetaData seqmd);

    // -------------------------------- Management of Classes --------------------------------

    /**
     * Method to register some data with the store.
     * This will also register the data with the starter process.
     * @param data The StoreData to add
     */
    protected void registerStoreData(StoreData data)
    {
        if (data.isFCO())
        {
            // Index any classes by the class name
            storeDataByClass.put(data.getName(), data);

            // If it's a class, using APPLICATION identity and is users own AID then store the PK class.
            // We don't need SingleFieldIdentity in here since they define the class being used
            ClassMetaData cmd = (ClassMetaData)data.getMetaData();
            if (cmd.getIdentityType() == IdentityType.APPLICATION &&
                !cmd.usesSingleFieldIdentityClass())
            {
                storeDataByAppIdClass.put(cmd.getObjectidClass(), data);
            }
        }
        else
        {
            // Index any fields by the class name of the field
            storeDataByClass.put(data.getMetaData(), data);
        }

        if (JPOXLogger.RDBMS_SCHEMA.isInfoEnabled())
        {
            JPOXLogger.RDBMS_SCHEMA.info(LOCALISER.msg("StoreManager.AddStoreData", data));
        }

        // Keep the AutoStarter in step with our managed classes/fields
        if (starter != null && starterInitialised)
        {
            starter.addClass(data);
        }
    }

    /**
     * Method to deregister all existing store data so that we are managing nothing.
     */
    protected void deregisterAllStoreData()
    {
        storeDataByClass.clear();
        storeDataByAppIdClass.clear();
        starterInitialised = false;
        if (JPOXLogger.RDBMS_SCHEMA.isInfoEnabled())
        {
            JPOXLogger.RDBMS_SCHEMA.info(LOCALISER.msg("StoreManager.RemoveStoreData"));
        }

        // Keep the AutoStarter in step with our managed classes/fields
        clearAutoStarter();
    }

    /**
     * Utility to return all StoreData for a Datastore Container identifier.
     * Returns StoreData with this table identifier and where the class is the owner of the table.
     * @param tableIdentifier Identifier for the table
     * @return The StoreData for this table (if managed).
     */
    public synchronized StoreData[] getStoreDataForDatastoreContainerObject(DatastoreIdentifier tableIdentifier)
    {
        Collection dataForTable = null;
        Iterator iterator = storeDataByClass.values().iterator();
        while (iterator.hasNext())
        {
            StoreData data = (StoreData) iterator.next();
            if (data.getDatastoreIdentifier() != null && data.getDatastoreIdentifier().equals(tableIdentifier) && data.isTableOwner)
            {
                if (dataForTable == null)
                {
                    dataForTable = new HashSet();
                }
                dataForTable.add(data);
            }
        }
        if (dataForTable != null)
        {
            return (StoreData[])dataForTable.toArray(new StoreData[dataForTable.size()]);
        }
        return null;
    }

    /**
     * Accessor for the datastore container (table) for a given datastore identifier
     * @param identifier The identifier
     * @return The datastore container (or null if not managed here)
     */
    public DatastoreClass getDatastoreContainerForIdentifier(DatastoreIdentifier identifier)
    {
        return (DatastoreClass)datastoreContainerByIdentifier.get(identifier);
    }

    /**
     * Method to add a datastore container to the managed datastore classes
     * @param table The datastore container
     */
    public void addDatastoreContainer(DatastoreContainerObject table)
    {
        if (table != null && datastoreContainerByIdentifier.get(table.getIdentifier()) == null)
        {
            datastoreContainerByIdentifier.put(table, table.getIdentifier());
        }
    }

    // -------------------------------- Auto Starter ------------------------------------

    /**
     * Method to initialise the auto-start mechanism, loading up the classes
     * from its store into memory so that we start from where we got to last
     * time.
     * @param mechanism AutoStart mechanism
     * @param mode Mode for autostart mechanism
     * @param clr The ClassLoaderResolver
     * @throws DatastoreInitialisationException
     */
    protected void initialiseAutoStart(String mechanism, String mode, ClassLoaderResolver clr)
    throws DatastoreInitialisationException
    {
        if (starterInitialised)
        {
            return;
        }

        String autoStarterClassName = getPMFContext().getPluginManager().getAttributeValueForExtension("org.jpox.autostart",
            "name", mechanism, "class-name");
        if (autoStarterClassName != null)
        {
            starter = (AutoStartMechanism) ClassUtils.newInstance(clr.classForName(autoStarterClassName),
                new Class[] {org.jpox.store.StoreManager.class, org.jpox.ClassLoaderResolver.class}, 
                new Object[] {this, clr});
            starter.setMode(mode);
        }
        if (starter == null)
        {
            starterInitialised = true;
            return;
        }

        boolean illegalState = false;
        try
        {
            if (!starter.isOpen())
            {
                starter.open();
            }
            Collection existingData = starter.getAllClassData();
            if (existingData != null && existingData.size() > 0)
            {
                List classesNeedingAdding = new ArrayList();
                Iterator existingDataIter = existingData.iterator();
                while (existingDataIter.hasNext())
                {
                    StoreData data = (StoreData) existingDataIter.next();
                    if (data.isFCO())
                    {
                        // Catch classes that don't exist (e.g in use by a different app)
                        // TODO Catch classes with no MetaData
                        boolean classFound = false;
                        try
                        {
                            if (clr.classForName(data.getName()) != null)
                            {
                                classFound = true;
                            }
                        }
                        catch (ClassNotResolvedException cnre)
                        {
                            if (data.getInterfaceName() != null)
                            {
                                try
                                {
                                    getPMFContext().getImplementationCreator().newInstance(clr.classForName(data.getInterfaceName()), 
                                        getMetaDataManager(), clr);
                                    if (clr.classForName(data.getName()) != null)
                                    {
                                        classFound = true;
                                    }
                                }
                                catch (ClassNotResolvedException cnre2)
                                {
                                    // Do nothing
                                }
                            }
                            // Thrown if class not found
                        }

                        
                        if (classFound)
                        {
                            JPOXLogger.JDO.info(LOCALISER.msg("StoreManager.AddPreviousStoreData", data.getName()));
                            classesNeedingAdding.add(data.getName());
                        }
                        else
                        {
                            String msg = LOCALISER.msg("AutoStarter.ClassNotFoundError", data.getName());
                            if (starter.getMode().equals(AutoStartMechanism.MODE_CHECKED))
                            {
                                JPOXLogger.JDO.error(msg);
                                throw new DatastoreInitialisationException(msg);
                            }
                            else if (starter.getMode().equals(AutoStartMechanism.MODE_IGNORED))
                            {
                                JPOXLogger.JDO.warn(msg);
                            }
                            else if (starter.getMode().equals(AutoStartMechanism.MODE_QUIET))
                            {
                                JPOXLogger.JDO.warn(msg);
                                JPOXLogger.JDO.warn(LOCALISER.msg("AutoStarter.DeleteClass", data.getName()));
                                starter.deleteClass(data.getName());
                            }
                        }
                    }
                }
                String[] classesToLoad = new String[classesNeedingAdding.size()];
                Iterator classesNeedingAddingIter = classesNeedingAdding.iterator();
                int n = 0;
                while (classesNeedingAddingIter.hasNext())
                {
                    classesToLoad[n++] = (String)classesNeedingAddingIter.next();
                }

                // Load the classes
                try
                {
                    addClasses(classesToLoad, clr, null);
                }
                catch (Exception e)
                {
                    // Exception while adding so some of the (referenced) classes dont exist
                    JPOXLogger.JDO.warn(LOCALISER.msg("AutoStarter.UnknownError", e));
                    //if an exception happens while loading AutoStart data, them we disable it, since
                    //it was unable to load the data from AutoStart. The memory state of AutoStart does
                    //not represent the database, and if we don't disable it, we could
                    //think that the autostart store is empty, and we would try to insert new entries in
                    //the autostart store that are already in there
                    illegalState = true;
                    
                    // TODO
                    // TODO Go back and add classes one-by-one to eliminate the class(es) with the problem
                }
            }
        }
        finally
        {
            if (starter.isOpen())
            {
                starter.close();
            }
            if( illegalState )
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("AutoStarter.IllegalStateDisablingAutoStart"));
                starter = null;
            }
        }

        starterInitialised = true;
    }

    /**
     * Method to clear the Auto-Starter status.
     */
    protected void clearAutoStarter()
    {
        // AutoStarter - Delete all supported classes
        if (starter != null)
        {
            try
            {
                if (!starter.isOpen())
                {
                    starter.open();
                }
                starter.deleteAllClasses();
            }
            finally
            {
                if (starter.isOpen())
                {
                    starter.close();
                }
            }
        }
    }

    /**
     * Accessor for the {@link AutoStartMechanism}
     * @return may return null if auto start mechanism is not initialized
     */
    public AutoStartMechanism getAutoStartMechanism()
    {
        return starter;
    }
    
    // ------------------------------- Class Management -----------------------------------

    /**
     * Accessor for the names of the classes currently managed by this store.
     * @return Collection of the names of the classes being managed
     */
    public Collection getManagedClasses()
    {
        return Collections.unmodifiableCollection(storeDataByClass.keySet());
    }

    /**
     * Accessor for whether the specified class is managed currently
     * @param className The name of the class
     * @return Whether it is managed
     */
    public boolean managesClass(String className)
    {
        return storeDataByClass.containsKey(className);
    }
    
    /**
     * Method to add a class to the managed list for this datastore manager.
     * @param className Name of the class
     * @param clr The ClassLoaderResolver
     */
    public abstract void addClass(String className, ClassLoaderResolver clr);
    
    /**
     * Add classes to the persistence model for the datastore.
     * In the case of an RDBMS, this will map these classes to RDBMS tables.
     * In the case of XML, this will map these classes to XML datafiles. 
     * <p>
     * This method is primarily useful for applications that wish to perform all
     * of their datastore initialization up front, rather than wait for the JPOX
     * runtime to do it on-demand.
     * @param classes The class(es) to be added.
     * @param clr The ClassLoaderResolver
     * @param writer Optional writer when you just want the DDL for persisting the specified classes
     * @exception org.jpox.store.exceptions.DatastoreValidationException
     *      If there is some mismatch between the current datastore contents and
     *      those necessary to enable persistence of the given classes.
     */
    public abstract void addClasses(String[] classes, ClassLoaderResolver clr, FileWriter writer);

    /**
     * Drops all tables in the datastore.
     * This empties the datastore of all datastore objects managed by JPOX.
     * All objects not managed by JPOX are left untouched.
     * @param clr The ClassLoaderResolver
     */
    public abstract void removeAllClasses(ClassLoaderResolver clr);

    /**
     * Method to return the class(es) that has a table managing the persistence of
     * the fields of the supplied class. For the 3 inheritance strategies, the following
     * occurs :-
     * <UL>
     * <LI>new-table : will return the same ClassMetaData</LI>
     * <LI>subclass-table : will return all subclasses that have a table managing its fields</LI>
     * <LI>superclass-table : will return the next superclass that has a table</LI>
     * </UL> 
     * @param cmd The supplied class.
     * @param clr ClassLoader resolver
     * @return The ClassMetaData's managing the fields of the supplied class
     */
    public AbstractClassMetaData[] getClassesManagingTableForClass(AbstractClassMetaData cmd, ClassLoaderResolver clr)
    {
        // Null input, so just return null;
        if (cmd == null)
        {
            return null;
        }

        // This class manages a table so return the classes metadata.
        if (cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NEW_TABLE)
        {
            return new AbstractClassMetaData[] {cmd};
        }
        else if (cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
        {
            // Check the subclasses that we have metadata for and make sure they are managed before proceeding
            String[] subclasses = getMetaDataManager().getSubclassesForClass(cmd.getFullClassName(), true);
            if (subclasses != null)
            {
                for (int i=0;i<subclasses.length;i++)
                {
                    if (storeDataByClass.get(subclasses[i]) == null)
                    {
                        addClass(subclasses[i], clr);
                    }
                }
            }

            // Find subclasses who manage the tables winto which our class is persisted
            HashSet managingClasses=new HashSet();
            Set managedClasses=storeDataByClass.keySet();
            Iterator managedClassesIter = managedClasses.iterator();
            while (managedClassesIter.hasNext())
            {
                StoreData data = (StoreData)storeDataByClass.get(managedClassesIter.next());
                if (data.isFCO() && ((AbstractClassMetaData)data.getMetaData()).getSuperAbstractClassMetaData() != null &&
                    ((AbstractClassMetaData)data.getMetaData()).getSuperAbstractClassMetaData().getFullClassName() == cmd.getFullClassName())
                {
                    AbstractClassMetaData[] superCmds = getClassesManagingTableForClass((AbstractClassMetaData)data.getMetaData(), clr);
                    if (superCmds != null)
                    {
                        for (int i=0;i<superCmds.length;i++)
                        {
                            managingClasses.add(superCmds[i]);
                        }
                    }
                }
            }

            Iterator managingClassesIter = managingClasses.iterator();
            AbstractClassMetaData managingCmds[] = new AbstractClassMetaData[managingClasses.size()];
            int i=0;
            while (managingClassesIter.hasNext())
            {
                managingCmds[i++] = (AbstractClassMetaData)(managingClassesIter.next());
            }
            return managingCmds;
        }
        else if (cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
        {
            // Fields managed by superclass, so recurse to that
            return getClassesManagingTableForClass(cmd.getSuperAbstractClassMetaData(), clr);
        }
        return null;
    }

    /**
     * Convenience method to ensure that the class defined by the passed OID
     * is managed by this store.
     * @param id OID
     * @param clr ClassLoader resolver
     * @return The class name of the class associated to this identity
     */
    public String manageClassForIdentity(OID id, ClassLoaderResolver clr)
    {
        String className = null;

        // Check that the implied class is managed
        className = id.getPcClass();
        AbstractClassMetaData cmd = getMetaDataManager().getMetaDataForClass(className, clr);
        if (cmd.getIdentityType() != IdentityType.DATASTORE)
        {
            throw new JDOUserException(LOCALISER.msg("Identity.AssignedToIncorrectClass", id, cmd.getFullClassName()));
        }

        // If the class is not yet managed, manage it
        if (!managesClass(className))
        {
            addClass(className, clr);
        }

        return className;
    }

    /**
     * Convenience method to ensure that the class defined by the passed SingleFieldIdentity
     * is managed by this store.
     * @param id Single Field Identity
     * @param clr ClassLoader resolver
     * @return The class name of the class associated to this identity
     */
    public String manageClassForIdentity(SingleFieldIdentity id, ClassLoaderResolver clr)
    {
        String className = null;

        // Check that the implied class is managed
        className = id.getTargetClassName();
        AbstractClassMetaData cmd = getMetaDataManager().getMetaDataForClass(className, clr);
        if (cmd.getIdentityType() != IdentityType.APPLICATION || !cmd.getObjectidClass().equals(id.getClass().getName()))
        {
            throw new JDOUserException(LOCALISER.msg("Identity.AssignedToIncorrectClass", id, cmd.getFullClassName()));
        }

        // If the class is not yet managed, manage it
        if (!managesClass(className))
        {
            addClass(className, clr);
        }

        return className;
    }

    // ------------------------------ PersistenceManager interface -----------------------------------

    /**
     * Interface to getting an Extent for a class.
     * @param pm The Persistence Manager
     * @param c  The class requiring the Extent
     * @param subclasses Whether to include subclasses of 'c'
     * @return The Extent.
     */
    public abstract Extent getExtent(org.jpox.PersistenceManager pm,Class c,boolean subclasses);

    /**
     * Accessor for whether this query language is supported.
     * @param language The language
     * @return Whether it is supported.
     * @see org.jpox.store.StoreManager#supportsQueryLanguage(java.lang.String)
     */
    public abstract boolean supportsQueryLanguage(String language);

    /**
     * Method to generate a new query using the passed query as basis.
     * @param language The query language
     * @param pm The PM
     * @param query The query filter (String) or a previous Query
     * @return The Query
     */
    public Query newQuery(String language, PersistenceManager pm, Object query)
    {
        if (language == null)
        {
            return null;
        }

        String queryClassName = getPMFContext().getPluginManager().getAttributeValueForExtension("org.jpox.store_query_query",
            "name", language, "class-name");
        if (queryClassName != null)
        {
            if (query == null)
            {
                return (Query) ClassUtils.newInstance(pm.getClassLoaderResolver().classForName(queryClassName,
                    AbstractPersistenceManagerFactory.class.getClassLoader()), 
                    new Class[] {org.jpox.PersistenceManager.class}, 
                    new Object[] {pm});
            }
            else
            {
                return (Query) ClassUtils.newInstance(pm.getClassLoaderResolver().classForName(queryClassName,
                    AbstractPersistenceManagerFactory.class.getClassLoader()), 
                    new Class[]{org.jpox.PersistenceManager.class, query.getClass()},
                    new Object[]{pm, query});
            }
        }
        return null;
    }

    /**
     * Returns the class corresponding to the given object JDO ID. If the object
     * is an OID, return the PersistenceCapable class. If the object is a SCOID,
     * return the SCO class. If the object is a PersistenceCapable class, return
     * the class. If the object is an AppID PK, return the PersistenceCapable
     * class.
     * @param id The JDO identity of some object.
     * @param clr ClassLoader resolver
     * @param pm Persistence Manager
     * @return For datastore identity, return the class of the corresponding
     * object. For application identity, return the class of the corresponding
     * object or null if object does not exist.
     * @exception ClassCastException If the type of ID is not recognized ({@link OID}
     * or {@link SCOID}).
     */
    public abstract String getClassNameForObjectID(Object id, ClassLoaderResolver clr, org.jpox.PersistenceManager pm);

    /**
     * Check if the strategy is attributed by the database when the
     * PersistenceCapable object is inserted into the database
     * @param identityStrategy the identityStrategy
     * @return if the object for the strategy is attributed by the database
     */
    public boolean isStrategyDatastoreAttributed(IdentityStrategy identityStrategy)
    {
        if (identityStrategy == null)
        {
            return false;
        }

        /*if (identityStrategy == IdentityStrategy.NATIVE && dba.supportsAutoIncrementFields())
        {
            // If the user leaves it to us and the DBA supports autoincrement then we use it
            // which means attributing the id in the datastore.
            return true;
        }*/

        // "identity" is processed in the datastore
        if (identityStrategy == IdentityStrategy.IDENTITY)
        {
            // TODO "identity" could imply using a SEQUENCE, so we need to allow for this here
            return true;
        }

        return false;
    }

    /**
     * Method to retrieve the value for a strategy for a particular field.
     * @param pm The Persistence Manager
     * @param table The datastore table
     * @param cmd AbstractClassMetaData for the class
     * @param absoluteFieldNumber The field number
     * @param strategy The strategy
     * @param sequence The sequence name if strategy is sequence-based
     * @param extensions Any extensions to control the value generation process
     * @return The value
     */
    public abstract Object getStrategyValue(PersistenceManager pm,
                                            DatastoreClass table,
                                            AbstractClassMetaData cmd,
                                            int absoluteFieldNumber,
                                            IdentityStrategy strategy,
                                            String sequence,
                                            ExtensionMetaData[] extensions);

    /**
     * Returns a new, unique ID for an object of the given class.
     * @param pm The PersistenceManager
     * @param className Name of the class of the object.
     * @param pc The persistenceCapable object. Useful for AppID
     * @return A new object ID.
     */
    public Object newObjectID(PersistenceManager pm, String className, PersistenceCapable pc)
    {
        AbstractClassMetaData cmd = getMetaDataManager().getMetaDataForClass(className, pm.getClassLoaderResolver()); 
        if (cmd.getIdentityType() == IdentityType.DATASTORE || cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            DatastoreClass t = getDatastoreClass(className, pm.getClassLoaderResolver());
            if (t == null && cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
            {
                // TODO Move this resource to store level
                throw new JDOUserException(LOCALISER.msg("StoreManager.NoTableForClass", className));
            }

            // Go up to overall superclass to find id for that class.
            boolean has_superclass = true;
            while (has_superclass)
            {
                DatastoreClass supert = t.getSuperDatastoreClass();
                if (supert != null)
                {
                    t = supert;
                }
                else
                {
                    has_superclass = false;
                }
            }
            Object oid = t.newOID(pm, pc);
            if (oid instanceof OID)
            {
                oid = OIDFactory.getInstance(className,((OID) oid).keyValue());
            }
            return oid;
        }
        else
        {
            return new SCOID(className);
        }
    }

    /**
     * Inserts a persistent object into the database.
     * @param sm The state manager of the object to be inserted.
     */
    public void insert(StateManager sm)
    {
        if (readOnlyDatastore)
        {
        	if (readOnlyDatastoreAction.equalsIgnoreCase("EXCEPTION"))
        	{
                throw new DatastorePermissionException(LOCALISER.msg("StoreManager.InsertObjectWhileReadOnlyError",
                		StringUtils.toJVMIDString(sm.getObject())));
        	}
        	else
        	{
        		if (JPOXLogger.JDO.isDebugEnabled())
                {
                    JPOXLogger.JDO.debug(LOCALISER.msg("StoreManager.InsertObjectWhileReadOnlyIgnore",
                        StringUtils.toJVMIDString(sm.getObject())));
                }
        		return;
        	}
        }

        DatastoreClass dc = getDatastoreClass(sm.getObject().getClass().getName(), sm.getPersistenceManager().getClassLoaderResolver());
        if (dc == null)
        {
            AbstractClassMetaData cmd = getMetaDataManager().getMetaDataForClass(sm.getObject().getClass(), sm.getPersistenceManager().getClassLoaderResolver()); 
            if (cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
            {
                throw new JDOUserException(LOCALISER.msg("StoreManager.NoTableForClass", sm.getObject().getClass().getName()));
            }
            throw new JDOFatalInternalException(LOCALISER.msg("StoreManager.CantFindTableForClass", sm.getObject().getClass().getName(), cmd.getInheritanceMetaData().getStrategyValue()));
        }
        dc.insert(sm);
    }

    /**
     * Fetches a persistent object from the database.
     * @param sm The state manager of the object to be fetched.
     * @param fieldNumbers The numbers of the fields to be fetched.
     */
    public void fetch(StateManager sm, int fieldNumbers[])
    {
        AbstractPropertyMetaData[] fmds = null;
        if (fieldNumbers != null)
        {
            // Convert the field numbers for this class into their metadata for the table
            fmds = new AbstractPropertyMetaData[fieldNumbers.length];
            for (int i=0;i<fmds.length;i++)
            {
                fmds[i] = sm.getClassMetaData().getManagedFieldAbsolute(fieldNumbers[i]);
            }
        }

        getDatastoreClass(sm.getObject().getClass().getName(), sm.getPersistenceManager().getClassLoaderResolver()).fetch(sm, fmds);
    }

    /**
     * Updates a persistent object in the database.
     * @param sm The state manager of the object to be updated.
     * @param fieldNumbers The numbers of the fields to be updated.
     */
    public void update(StateManager sm, int fieldNumbers[])
    {
        if (readOnlyDatastore)
        {
        	if (readOnlyDatastoreAction.equalsIgnoreCase("EXCEPTION"))
        	{
                throw new DatastorePermissionException(LOCALISER.msg("StoreManager.UpdateObjectWhileReadOnlyError",
                		StringUtils.toJVMIDString(sm.getObject())));
        	}
        	else
        	{
        		if (JPOXLogger.JDO.isDebugEnabled())
                {
                    JPOXLogger.JDO.debug(LOCALISER.msg("StoreManager.UpdateObjectWhileReadOnlyIgnore",
                        StringUtils.toJVMIDString(sm.getObject())));
                }
        		return;
        	}
        }

        AbstractPropertyMetaData[] fmds = null;
        if (fieldNumbers != null)
        {
            // Convert the field numbers for this class into their metadata for the table
            fmds = new AbstractPropertyMetaData[fieldNumbers.length];
            for (int i=0;i<fmds.length;i++)
            {
                fmds[i] = sm.getClassMetaData().getManagedFieldAbsolute(fieldNumbers[i]);
            }
        }

        getDatastoreClass(sm.getObject().getClass().getName(), sm.getPersistenceManager().getClassLoaderResolver()).update(sm, fmds);
    }

    /**
     * Deletes a persistent object from the database.
     * @param sm The state manager of the object to be deleted.
     */
    public void delete(StateManager sm)
    {
        if (readOnlyDatastore)
        {
        	if (readOnlyDatastoreAction.equalsIgnoreCase("EXCEPTION"))
        	{
                throw new DatastorePermissionException(LOCALISER.msg("StoreManager.DeleteObjectWhileReadOnlyError",
                		StringUtils.toJVMIDString(sm.getObject())));
        	}
        	else
        	{
        		if (JPOXLogger.JDO.isDebugEnabled())
                {
                    JPOXLogger.JDO.debug(LOCALISER.msg("StoreManager.DeleteObjectWhileReadOnlyIgnore",
                        StringUtils.toJVMIDString(sm.getObject())));
                }
        		return;
        	}
        }

        getDatastoreClass(sm.getObject().getClass().getName(), sm.getPersistenceManager().getClassLoaderResolver()).delete(sm);
    }

    /**
     * Locates this object in the datastore.
     * @param sm The StateManager for the object to be found 
     * @throws javax.jdo.JDOObjectNotFoundException If the instance does not exist in
     * the datastore
     */
    public void locate(StateManager sm)
    {
        getDatastoreClass(sm.getObject().getClass().getName(), sm.getPersistenceManager().getClassLoaderResolver()).locate(sm);
    }

    /**
     * Delete dependent objects of a persistent object.
     * @since 1.1
     * @param sm StateManager for the object being deleted.
     * @param fieldNumbers the fields to delete.
     */
    public void deleteDependent(StateManager sm, int[] fieldNumbers)
    {
        if (readOnlyDatastore)
        {
        	if (readOnlyDatastoreAction.equalsIgnoreCase("EXCEPTION"))
        	{
                throw new DatastorePermissionException(LOCALISER.msg("StoreManager.DeleteDependentObjectWhileReadOnlyError",
                		StringUtils.toJVMIDString(sm.getObject())));
        	}
        	else
        	{
        		if (JPOXLogger.JDO.isDebugEnabled())
                {
                    JPOXLogger.JDO.debug(LOCALISER.msg("StoreManager.DeleteDependentObjectWhileReadOnlyIgnore",
                        StringUtils.toJVMIDString(sm.getObject())));
                }
        		return;
        	}
        }

        AbstractPropertyMetaData[] fmds = null;
        if (fieldNumbers != null)
        {
            // Convert the field numbers for this class into their metadata for the table
            fmds = new AbstractPropertyMetaData[fieldNumbers.length];
            for (int i=0;i<fmds.length;i++)
            {
                fmds[i] = sm.getClassMetaData().getManagedFieldAbsolute(fieldNumbers[i]);
            }
        }

        getDatastoreClass(sm.getObject().getClass().getName(), sm.getPersistenceManager().getClassLoaderResolver()).deleteDependent(sm, fmds);
    }

    /**
     * Called by Mapping objects to request the creation of a DatastoreObject (table).
     * @param fmd The field metadata describing the field.
     * @param clr The ClassLoaderResolver
     * @return The DatastoreContainerObject
     */
    public abstract DatastoreContainerObject newJoinDatastoreContainerObject(AbstractPropertyMetaData fmd, ClassLoaderResolver clr);

    /**
     * Utility to return the names of the classes that are known subclasses of the provided
     * class. Actually uses the MetaDataManager for determining what is a subclass
     * since the MetaData is often registered before being needed by the Store.
     * @param className Class for which we search for subclasses.
     * @param includeDescendents Whether to include subclasses of subclasses etc
     * @param clr The ClassLoaderResolver
     * @return Set of classes that are subclasses of the input class.
     */
    public HashSet getSubClassesForClass(String className, boolean includeDescendents, ClassLoaderResolver clr)
    {
        HashSet subclasses = new HashSet();

        String[] subclassNames = getMetaDataManager().getSubclassesForClass(className, includeDescendents);
        if (subclassNames != null)
        {
            // Load up the table for any classes that are not yet loaded
            for (int i=0;i<subclassNames.length;i++)
            {
                if (storeDataByClass.get(subclassNames[i]) == null)
                {
                    // We have no knowledge of this class so load it now
                    addClass(subclassNames[i], clr);
                }
                subclasses.add(subclassNames[i]);
            }
        }

        return subclasses;
    }

    // ------------------------------- Accessors ------------------------------------

    /**
     * Accessor for the identifier factory.
     * @return Returns the identifier factory.
     */
    public IdentifierFactory getIdentifierFactory()
    {
        return identifierFactory;
    }

    /**
     * Accessor for the context in which this RDBMSManager is running
     * @return Returns the pmfContext.
     */   
    public PMFContext getPMFContext()
    {
        return pmfContext;
    }
 
    /**
     * Gets the MetaDataManager to use for this store.
     * @return Returns the MetaDataManager.
     */   
    public MetaDataManager getMetaDataManager()
    {
        return pmfContext.getMetaDataManager();
    }

    /**
     * Gets the MappingManager to use for this store.
     * @return Returns the MappingManager.
     */
    public MappingManager getMappingManager()
    {
        return dba.getMappingManager();
    }

    /**
     * Gets the DatastoreAdapter to use for this store.
     * @return Returns the DatastoreAdapter
     */   
    public DatastoreAdapter getDatastoreAdapter()
    {
        return dba;
    }

    /**
     * Returns the primary datastore container serving as backing for the given class. 
     * If the class is not yet known to the store manager, {@link #addClass}is called
     * to add it. Classes which have inheritance strategy of "new-table" and
     * "superclass-table" will return a table here, whereas "subclass-table" will
     * return null since it doesn't have a table as such.
     * <p>
     * @param className Name of the class whose table is be returned.
     * @param clr The ClassLoaderResolver
     * @return The corresponding class table.
     * @exception NoTableManagedException If the given class has no table managed in the database.
     */
    public DatastoreClass getDatastoreClass(String className, ClassLoaderResolver clr)
    {
        DatastoreClass ct = null;
        if (className == null)
        {
            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("StoreManager.TableRetrievalError"));
            return null;
        }

        boolean classKnown = false;
        synchronized (this)
        {
            // Check if we know about the class
            StoreData sd = (StoreData) storeDataByClass.get(className);
            if (sd != null)
            {
                classKnown = true;
                ct = (DatastoreClass) sd.getDatastoreContainerObject();
            }
        }

        if (!classKnown)
        {
            // Consider adding it to our list of supported classes. Currently we only consider PC classes
            boolean toBeAdded = false;
            if (clr != null)
            {
                Class cls = clr.classForName(className);
                if (cls != null && !cls.isInterface() && PersistenceCapable.class.isAssignableFrom(cls))
                {
                    toBeAdded = true;
                }
            }
            else
            {
                toBeAdded = true;
            }

            if (toBeAdded)
            {
                // Add the class to our supported list
                addClass(className, clr);

                // Retry
                synchronized (this)
                {
                    StoreData sd = (StoreData) storeDataByClass.get(className);
                    if (sd != null)
                    {
                        classKnown = true;
                        ct = (DatastoreClass) sd.getDatastoreContainerObject();
                    }
                }
            }
        }

        // Throw an exception if class still not known and no table
        // Note : "subclass-table" inheritance strategies will return null from this method
        if (!classKnown && ct == null)
        {
            throw new NoTableManagedException(className);
        }

        return ct;
    }

    /**
     * Returns the datastore container (table) for the specified field. 
     * Returns 'null' if the field is not (yet) known to the store manager.
     * @param fmd The metadata for the field.
     * @return The corresponding datastore container, or 'null'.
     */
    public synchronized DatastoreContainerObject getDatastoreContainerObject(AbstractPropertyMetaData fmd)
    {
        StoreData sd = (StoreData) storeDataByClass.get(fmd);
        if (sd != null)
        {
            return sd.getDatastoreContainerObject();
        }
        else
        {
            return null;
        }
    }

    /**
     * Returns the JDO table having the given SQL identifier.
     * Returns 'null' if no such table is (yet) known to the store manager.
     * @param name The identifier name of the table.
     * @return The corresponding JDO table, or 'null'
     */
    public synchronized DatastoreClass getDatastoreClass(DatastoreIdentifier name)
    {
        Iterator iterator = storeDataByClass.values().iterator();
        while (iterator.hasNext())
        {
            StoreData sd = (StoreData) iterator.next();
            if (sd.hasTable() && sd.getDatastoreIdentifier().equals(name))
            {
                return (DatastoreClass) sd.getDatastoreContainerObject();
            }
        }
        return null;
    }

    /**
     * Utility to navigate the inheritance hierarchy to find the base class that defines the primary keys
     * for this tree. This will either go up to the next class in the hierarchy that has a table
     * OR go up to the base class, whichever is first.
     * @param cmd AbstractClassMetaData for this class
     * @param clr The ClassLoaderResolver
     * @return The AbstractClassMetaData for the class defining the primary keys
     */
    public AbstractClassMetaData getClassWithPrimaryKeyForClass(AbstractClassMetaData cmd, ClassLoaderResolver clr)
    {
        if (cmd == null)
        {
            return null;
        }

        // Base class will have primary key fields
        if (cmd.getSuperAbstractClassMetaData() == null)
        {
            return cmd;
        }
        // Class has its own table so has the PK fields already
        else if (getDatastoreClass(cmd.getFullClassName(), clr) != null)
        {
            return cmd;
        }

        return getClassWithPrimaryKeyForClass(cmd.getSuperAbstractClassMetaData(), clr);
    }

    /**
     * Resolves an identifier macro. The public fields <var>clazz </var>,
     * <var>fieldName </var>, and <var>subfieldName </var> of the given macro
     * are taken as inputs, and the public <var>value </var> field is set to the
     * SQL identifier of the corresponding database table or column.
     * @param im The macro to resolve.
     * @param clr The ClassLoaderResolver
     */
    public abstract void resolveIdentifierMacro(MacroString.IdentifierMacro im, ClassLoaderResolver clr);

    /**
     * Method to create a new fetch statement for the passed table.
     * @param table The table to fetch from
     * @return The fetch statement
     */
    public abstract FetchStatement getFetchStatement(DatastoreContainerObject table);

    /**
     * Method to return a backing store for a Collection, consistent with this store and the instantiated type.
     * @param fmd MetaData for the field that has this collection
     * @param datastoreTable The datastore object
     * @param clr ClassLoader resolver
     * @param instantiated Whether it is already instantiated
     * @param listBased whether the collection is list based
     * @return The Backing store for this collection in this store
     */
    public abstract CollectionStore getBackingStoreForCollection(AbstractPropertyMetaData fmd, 
            DatastoreObject datastoreTable, ClassLoaderResolver clr, boolean instantiated, boolean listBased);

    /**
     * Method to return a backing store for a Map, consistent with this store and the instantiated type.
     * @param fmd MetaData for the field that has this map
     * @param datastoreTable The datastore object
     * @param clr ClassLoader resolver
     * @return The Backing store for this map in this store
     */
    public abstract MapStore getBackingStoreForMap(AbstractPropertyMetaData fmd, DatastoreObject datastoreTable, ClassLoaderResolver clr);

    /**
     * Method to return a backing store for an array, consistent with this store and the instantiated type.
     * @param fmd MetaData for the field that has this array
     * @param datastoreTable The datastore object
     * @param clr ClassLoader resolver
     * @return The Backing store for this array in this store
     */
    public abstract ArrayStore getBackingStoreForArray(AbstractPropertyMetaData fmd, DatastoreObject datastoreTable, ClassLoaderResolver clr);
}