/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - renamed from CandidateSetExpression
    ...
**********************************************************************/
package org.jpox.store.expression;

import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.query.StatementText;
import org.jpox.store.scostore.CollectionStore;

/**
 * An expression that represents some collection field in a query candidate
 * class, or a collection field in an object linked from the candidate class
 * by navigation.
 * <p>
 * When navigated through using contains(expr), the elements of the collection
 * are relationally joined onto the query statement.
 * </p>
 * <p>
 * As this is a Collection Expression it works equally for Sets and Lists,
 * and we use CollectionStore as the backing store interface, so that
 * ListStore and SetStore are equally applicable.
 *
 * @version $Revision: 1.31 $
 */
public class CollectionExpression extends ScalarExpression
{
    private final CollectionStore collStore;
    private final String fieldName;

    /**
     * Constructor.
     * @param qs The Query Statement
     * @param ownerMapping The mapping to the owner of this collection
     * @param te The Table Expression
     * @param collStore the backing store.
     * @param fieldName Name of the field for the collection.
     **/
    public CollectionExpression(QueryExpression qs, 
                                JavaTypeMapping ownerMapping,
                                LogicSetExpression te,
                                CollectionStore collStore,
                                String fieldName)
    {
        super(qs);

        this.mapping = ownerMapping;
        this.collStore = collStore;
        this.fieldName = fieldName;
        this.te = te;
    }

    /**
     * Executed when the size() method is found in a query filter.
     * @return  The NumericExpression resulting from the size() method.
     */
    public NumericExpression sizeMethod()
    {
        IdentifierFactory idFactory = qs.getStoreManager().getIdentifierFactory();
        String ctIdentifier = idFactory.newIdentifier(te.getRangeVariable(), fieldName).getIdentifier();
        DatastoreIdentifier ctRangeVar = idFactory.newIdentifier(IdentifierFactory.TABLE, ctIdentifier);

        return new ContainerSizeExpression(qs, collStore.getSizeSubquery(qs, mapping, te, ctRangeVar));
    }

    /**
     * Executed when the contains() method is found in a query filter.
     * @param expr The ScalarExpression passed as a parameter to contains().
     * @return The BooleanExpression resulting from the contains() method.
     */
    public BooleanExpression containsMethod(ScalarExpression expr)
    {
        // ct... = "collection table"
        // et... = "element table"

        IdentifierFactory idFactory = qs.getStoreManager().getIdentifierFactory();
        if (expr instanceof NullLiteral)
        {
            // JPOX doesn't currently support storing nulls in Collections so just return "1 = 0"
            // TODO Add support for storing nulls in collections and querying for this situation
            return new BooleanLiteral(qs, mapping, false).eq(new BooleanLiteral(qs, mapping, true));
        }
        else if (expr instanceof UnboundVariable)
        {
            UnboundVariable var = (UnboundVariable)expr;
            if (var.getVariableType() == null)
            {
                //Set the variable type to be the element type for this collection
                //implicit variable type. We now set the type to the collection type 
                var.setVariableType(qs.getClassLoaderResolver().classForName(collStore.getElementType()));
            }
            //ctJavaName += + '.' + var.getVariableName();
            //String etJavaName = qs.getDefaultTableExpression().getRangeVariable().getJavaName() + '.' + var.getVariableName();

            String etIdentifier = "UNBOUND" + '.' + var.getVariableName();
            String ctIdentifier = idFactory.newIdentifier(idFactory.newIdentifier(te.getRangeVariable(), fieldName), var.getVariableName()).getIdentifier();

            DatastoreIdentifier ctRangeVar = idFactory.newIdentifier(IdentifierFactory.TABLE, ctIdentifier);
            DatastoreIdentifier etRangeVar = idFactory.newIdentifier(IdentifierFactory.TABLE, etIdentifier);
            QueryExpression qexpr = collStore.getExistsSubquery(qs, mapping, te, ctRangeVar);
            qexpr.setParent(qs);
            ScalarExpression bindTo = collStore.joinElementsTo(qexpr,
                qs,
                mapping,
                te,
                ctRangeVar,
                var.getVariableType(),
                expr,
                expr.te == null ? etRangeVar : expr.te.getRangeVariable());
            var.bindTo(bindTo);
            //see JDOQLContainerTest.testContainsResultVariable
            if( qs.hasCrossJoin(qs.getTableExpression(etRangeVar)))
            {
                qexpr.andCondition(bindTo.eq(expr.mapping.newScalarExpression(qs, qs.getTableExpression(etRangeVar))));
            }
            //END see JDOQLContainerTest.

            return new ExistsExpression(qs, qexpr, true);

        }
        else if (expr instanceof Literal)
        {
            String ctIdentifier = idFactory.newIdentifier(te.getRangeVariable(), fieldName).getIdentifier();
            DatastoreIdentifier ctRangeVar = qs.getStoreManager().getIdentifierFactory().newIdentifier(IdentifierFactory.TABLE, ctIdentifier);
            DatastoreIdentifier etRangeVar;             
            if (expr.te == null) // literals
            {
                int n = 0;

                do
                {
                    String etIdentifier = ctIdentifier + '.' + (++n);
                    etRangeVar = idFactory.newIdentifier(IdentifierFactory.TABLE, etIdentifier);
                } while (qs.getTableExpression(etRangeVar) != null);
            }
            else
            {
                etRangeVar = expr.te.getRangeVariable();
            }

            ClassLoaderResolver clr = qs.getClassLoaderResolver();
            QueryExpression qexpr = collStore.getExistsSubquery(qs, mapping, te, ctRangeVar);
            ScalarExpression expr1 = collStore.joinElementsTo(qexpr, qs, mapping, te,
                ctRangeVar,
                clr.classForName(expr.getMapping().getType()),
                expr,
                expr.te == null ? etRangeVar : expr.te.getRangeVariable());
            qexpr.andCondition(expr.eq(expr1));
            return new ExistsExpression(qs, qexpr, true);
        }        
        else
        {
            String ctIdentifier = idFactory.newIdentifier(te.getRangeVariable(), fieldName).getIdentifier();
            DatastoreIdentifier ctRangeVar = idFactory.newIdentifier(IdentifierFactory.TABLE, ctIdentifier);
            DatastoreIdentifier etRangeVar;             
            if (expr.te == null) // literals
            {
                int n = 0;

                do
                {
                    String etIdentifier = ctIdentifier + '.' + (++n);
                    etRangeVar = idFactory.newIdentifier(IdentifierFactory.TABLE, etIdentifier);
                } while (qs.getTableExpression(etRangeVar) != null);
            }
            else
            {
                etRangeVar = expr.te.getRangeVariable();
            }

            ClassLoaderResolver clr = qs.getClassLoaderResolver();
            ScalarExpression joinExpr =
               collStore.joinElementsTo(expr.getQueryExpression(),
                                        qs,
                                        mapping,
                                        te,
                                        ctRangeVar,
                                        clr.classForName(collStore.getElementType()),
                                        expr,
                                        etRangeVar);

            return joinExpr.eq(expr);
        }
    }

    /**
     * Return the BooleanExpression for a query filter in the form "collection.isEmpty()".
     * @return The BooleanExpression for a query filter in the form 
     * "collection.isEmpty()".
     */
    public BooleanExpression isEmptyMethod()
    {
        // ct... = "collection table"
        IdentifierFactory idFactory = qs.getStoreManager().getIdentifierFactory();
        String ctIdentifier = idFactory.newIdentifier(te.getRangeVariable(), fieldName).getIdentifier();
        DatastoreIdentifier ctRangeVar = idFactory.newIdentifier(IdentifierFactory.TABLE, ctIdentifier);
        QueryExpression qexpr = collStore.getExistsSubquery(qs, mapping, te, ctRangeVar);
        qexpr.setParent(qs);
        return new ExistsExpression(qs, qexpr, false);
    }

    /**
     * Method to return the expression for comparing a collection with a value.
     * JPOX only supports comparisons with null currently.
     * @param expr The value to compare with.
     * @return The expression of equality
     */
    public BooleanExpression eq(ScalarExpression expr)
    {
        if (expr instanceof NullLiteral)
        {
            return isEmptyMethod();
        }
        else
        {
            throw new JDOUserException(LOCALISER.msg("CollectionExpression.EqualsNotNullInvalid"));
        }
    }

    /**
     * Method to return the statement text.
     * @param mode (0=PROJECTION;1=FILTER)
     * @return The statement
     * @throws JDOUserException since this object is inaccessible directly.
     **/
    public StatementText toStatementText(int mode)
    {
        throw new JDOUserException("Cannot reference collection object directly: field name = " + fieldName);
    }
}