/**********************************************************************
Copyright (c) 2005 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.store.expression;

/**
 * Interface that allows one datastore adapt operations
 * @version $Revision: 1.10 $
 */
public interface ExpressionMethodAdapter
{
    /**
     * Returns the absolute expression for the JDOQL Math.abs(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * ABS(str)
     * </pre></blockquote>
     *
     * @param expr   The argument to the abs() method.
     * @return The expression.
     */
    NumericExpression absMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.sqrt(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * SQRT(str)
     * </pre></blockquote>
     *
     * @param expr   The argument to the sqrt() method.
     * @return  The expression.
     */
    NumericExpression sqrtMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.cos(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * COS(str)
     * </pre></blockquote>
     *
     * @param expr   The argument to the cos() method.
     * @return  The expression.
     */
    NumericExpression cosMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.sin(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * SIN(str)
     * </pre></blockquote>
     *
     * @param expr The argument to the sin() method.
     * @return  The expression.
     */
    NumericExpression sinMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.tan(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * TAN(str)
     * </pre></blockquote>
     *
     * @param expr The argument to the tan() method.
     * @return  The expression.
     */
    NumericExpression tanMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.acos(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * ACOS(str)
     * </pre></blockquote>
     *
     * @param expr   The argument to the cos() method.
     * @return  The expression.
     */
    NumericExpression acosMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.asin(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * ASIN(str)
     * </pre></blockquote>
     *
     * @param expr The argument to the asin() method.
     * @return  The expression.
     */
    NumericExpression asinMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.tan(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * ATAN(str)
     * </pre></blockquote>
     *
     * @param expr The argument to the atan() method.
     * @return  The expression.
     */
    NumericExpression atanMethod(ScalarExpression expr);

    /**
     * Returns whether this string ends with the specified string.
     * @param leftOperand the source string
     * @param rightOperand The string to compare against.
     * @return Whether it ends with the string.
     **/
    BooleanExpression endsWithMethod(ScalarExpression leftOperand, ScalarExpression rightOperand);
    
    /**
     * Returns the expression for the JDOQL Math.exp(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * EXP(str)
     * </pre></blockquote>
     *
     * @param expr The argument to the exp() method.
     * @return  The expression.
     */
    NumericExpression expMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.log(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * LOG(str)
     * </pre></blockquote>
     *
     * @param expr The argument to the log() method.
     * @return  The expression.
     */
    NumericExpression logMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.floor(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * FLOOR(str)
     * </pre></blockquote>
     *
     * @param expr The argument to the floor() method.
     * @return  The expression.
     */
    NumericExpression floorMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL Math.ceil(EXPRESSION)
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * CEIL(str)
     * </pre></blockquote>
     *
     * @param expr The argument to the ceil() method.
     * @return  The expression.
     */
    NumericExpression ceilMethod(ScalarExpression expr);

    /**
     * Returns the expression for the JDOQL String.length()
     * method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * CHAR_LENGTH(str)
     * </pre></blockquote>
     *
     * @param str   The argument to the length() method.
     * @return  The expression.
     */
    NumericExpression lengthMethod(StringExpression str);

    /**
     * Returns the expression for the JDOQL
     * String.substring(str,begin) method.
     * In SQL, it may compile to:
     * <p>
     * <pre>
     * SUBSTRING(str FROM begin)
     * </pre>
     * Note that the value of <var>begin</var> is base 0(Java-style), while most
     * SQL string functions use base 1.
     *
     * @param str The first argument to the substring() method.
     * @param begin The second argument to the substring() method.
     * @return  The expression.
     */
    StringExpression substringMethod(StringExpression str, NumericExpression begin);

    /**
     * Returns the expression for the JDOQL
     * String.substring(str,begin,end) method.
     * In SQL, it may compile to:
     * <p>
     * <blockquote><pre>
     * SUBSTRING(str FROM begin FOR len)
     * </pre></blockquote>
     * Note that the value of <var>begin</var> is base 0 (Java-style), while most
     * SQL string functions use base 1.
     * Note also that an end position is given, while most SQL substring
     * functions take a length.
     *
     * @param str   The first argument to the substring() method.
     * @param begin The second argument to the substring() method.
     * @param end   The third argument to the substring() method.
     * @return  The expression.
     */
    StringExpression substringMethod(StringExpression str,
                                     NumericExpression begin,
                                     NumericExpression end);

    /**
     * Method to handle the starts with operation.
     * @param source The expression with the searched string
     * @param str The expression for the search string 
     * @return The expression.
     **/
    BooleanExpression startsWithMethod(ScalarExpression source, ScalarExpression str);

    /**
     * Method to handle the indexOf operation.
     * @param source The expression with the searched string
     * @param str The expression for the search string
     * @param from Position to start searching from
     * @return The expression.
     **/
    NumericExpression indexOfMethod(ScalarExpression source, ScalarExpression str, NumericExpression from);

    /**
     * Returns the expression for the JDOQL
     * String.trim(str) method.
     * In SQL, it may compile to:
     * <p>
     * <pre>
     * TRIM(str)
     * </pre>
     * </p>
     * @param str The first argument to the trim() method.
     * @return  The expression.
     */
    StringExpression trimMethod(StringExpression str);

    /**
     * Method to handle the Date.getDay() method in JDOQL.
     * @param source The date expression to find the day of month for
     * @return The expression for the day of the month.
     **/
    NumericExpression getDayMethod(SqlTemporalExpression source);

    /**
     * Method to handle the Date.getMonth() method in JDOQL.
     * @param source The date expression to find the month for
     * @return The expression for the month.
     **/
    NumericExpression getMonthMethod(SqlTemporalExpression source);

    /**
     * Method to handle the Date.getYear() method in JDOQL.
     * @param source The date expression to find the year for
     * @return The expression for the year.
     **/
    NumericExpression getYearMethod(SqlTemporalExpression source);

    /**
     * Method to handle the Time.getHour() method in JDOQL.
     * @param source The time expression to find the hour for
     * @return The expression for the hour.
     **/
    NumericExpression getHourMethod(SqlTemporalExpression source);

    /**
     * Method to handle the Time.getMinute() method in JDOQL.
     * @param source The time expression to find the minute for
     * @return The expression for the minute.
     **/
    NumericExpression getMinuteMethod(SqlTemporalExpression source);

    /**
     * Method to handle the Time.getSecond() method in JDOQL.
     * @param source The time expression to find the second for
     * @return The expression for the second.
     **/
    NumericExpression getSecondMethod(SqlTemporalExpression source);
    
    /**
     * Matches this to the argument expression pattern. Use "." to find any
     * character and ".*" for wildcard matches. A global case-insensitive flag
     * "(?i)" can be set for the pattern. If used, the global case-insensitive
     * flag must prefix the pattern. The pattern passed to matches must be a
     * literal or parameter.
     * @param text   The argument to the length() method.
     * @param pattern The literal expression with the pattern.
     * @return the match expression.
     */
    BooleanExpression matchesMethod(StringExpression text, StringExpression pattern);
    
}