/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOHelper;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.ClassNameConstants;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ContainerMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreObject;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NoDatastoreMappingException;
import org.jpox.store.scostore.Store;
import org.jpox.util.ClassUtils;

/**
 * Mapping for a field that represents a container of objects, such as a List,
 * a Set, a Collection, a Map, or an array. Has an owner table. 
 * Can be represented in the following ways :-
 * <ul>
 * <li>Using a Join-Table, where the linkage between owner and elements/keys/values is stored in this table</li>
 * <li>Using a Foreign-Key in the element/key/value</li>
 * <li>Serialised into a single column in the owner table.</li>
 * <li>Embedded into the Join-Table</li>
 * <li>Serialised into a single-column in the Join-Table</li>
 * </ul>
 * The contents of the container are backed by a store, handling interfacing with the datastore.
 *
 * @version $Revision: 1.30.2.1 $
 */
public abstract class AbstractContainerMapping extends SingleFieldMapping
{
    /** Name of the field that is the container. */
    protected String fieldName;

    /** Store Manager being used. */
    protected StoreManager storeMgr;

    /** Datastore class of the contents. */
    protected DatastoreObject contentsTable;

    /** Store for the contents. */
    protected Store contentsStore;

    /** The type of the instantiated object in this field. */
    protected Class instantiatedType;

    /**
     * Initialize this JavaTypeMapping with the given DatastoreAdapter for
     * the given FieldMetaData.
     *  
     * @param dba The Datastore Adapter that this Mapping should use.
     * @param fmd FieldMetaData for the field to be mapped (if any)
     * @param container The datastore container storing this mapping (if any)
     * @param clr the ClassLoaderResolver
     */
    public void initialize(DatastoreAdapter dba, AbstractPropertyMetaData fmd, DatastoreContainerObject container, ClassLoaderResolver clr)
    {
		super.initialize(dba, fmd, container, clr);
        this.fieldName = fmd.getName();
        this.storeMgr = datastoreContainer.getStoreManager();

        ContainerMetaData conmd = fmd.getContainer();
        if (conmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("ContainerMapping.ContainerMetaDataError", fmd.getFullFieldName()));
        }

        if (!containerIsStoredInSingleColumn())
        {
            // Not serialised so we use JoinTable or ForeignKey
            contentsTable = storeMgr.newJoinDatastoreContainerObject(fmd, clr);
        }
        contentsStore = null;
    }

    /**
     * Method to prepare a field mapping for use in the datastore.
     * This creates the column in the table.
     */
    protected void prepareDatastoreMapping()
    {
        if (containerIsStoredInSingleColumn())
        {
            // Serialised collections/maps/arrays should just create a (typically BLOB) column as normal in the owning table
            super.prepareDatastoreMapping();
        }
    }

    /**
     * Accessor for the name of the java-type actually used when mapping the particular datastore
     * field. This java-type must have an entry in the datastore mappings.
     * @param index requested datastore field index.
     * @return the name of java-type for the requested datastore field.
     */
    public String getJavaTypeForDatastoreMapping(int index)
    {
        if (containerIsStoredInSingleColumn())
        {
            // Serialised container so just return serialised
            return ClassNameConstants.JAVA_IO_SERIALIZABLE;
        }
        return super.getJavaTypeForDatastoreMapping(index);
    }

    /**
     * Method to set a field in the passed JDBC PreparedStatement using this mapping.
     * Only valid when the collection is serialised.
     * @param pm The Persistence Manager
     * @param preparedStatement The JDBC Prepared Statement to be populated
     * @param exprIndex The parameter positions in the JDBC statement to populate.
     * @param value The value to populate into it
     */
    public void setObject(PersistenceManager pm, Object preparedStatement, int[] exprIndex, Object value)
    {
        if (fmd == null || !containerIsStoredInSingleColumn())
        {
            throw new JDOFatalInternalException(failureMessage("setObject"));
        }

        StateManager[] sms = null;
        if (value != null)
        {
            Collection smsColl = null;
            if (value instanceof java.util.Collection)
            {
                Iterator elementsIter = ((java.util.Collection)value).iterator();
                while (elementsIter.hasNext())
                {
                    Object elem = elementsIter.next();
                    if (elem instanceof PersistenceCapable)
                    {
                        StateManager sm = pm.findStateManager((PersistenceCapable)elem);
                        if (sm != null)
                        {
                            if (smsColl == null)
                            {
                                smsColl = new HashSet();
                            }
                            smsColl.add(sm);
                        }
                    }
                }
            }
            else if (value instanceof java.util.Map)
            {
                Iterator entriesIter = ((java.util.Map)value).entrySet().iterator();
                while (entriesIter.hasNext())
                {
                    Map.Entry entry = (Map.Entry)entriesIter.next();
                    Object key = entry.getKey();
                    Object val = entry.getValue();
                    if (key instanceof PersistenceCapable)
                    {
                        StateManager sm = pm.findStateManager((PersistenceCapable)key);
                        if (sm != null)
                        {
                            if (smsColl == null)
                            {
                                smsColl = new HashSet();
                            }
                            smsColl.add(sm);
                        }
                    }
                    if (val instanceof PersistenceCapable)
                    {
                        StateManager sm = pm.findStateManager((PersistenceCapable)val);
                        if (sm != null)
                        {
                            if (smsColl == null)
                            {
                                smsColl = new HashSet();
                            }
                            smsColl.add(sm);
                        }
                    }
                }
            }
            if (smsColl != null)
            {
                sms = (StateManager[])smsColl.toArray(new StateManager[smsColl.size()]);
            }
        }

        if (sms != null)
        {
            // Set all PC objects as being stored (so we dont detach them in any serialisation process)
            for (int i=0;i<sms.length;i++)
            {
                sms[i].setStoringPC();
            }
        }
        getDataStoreMapping(0).setObject(preparedStatement, exprIndex[0], value);
        if (sms != null)
        {
            // Unset all PC objects now they are stored
            for (int i=0;i<sms.length;i++)
            {
                sms[i].unsetStoringPC();
            }
        }
    }

    /**
     * Method to retrieve an object from the passed JDBC ResultSet.
     * Only valid when the collection is serialised.
     * @param pm Persistence Manager
     * @param resultSet The JDBC ResultSet
     * @param exprIndex The JDBC parameter position(s) to extract the object from
     * @return The collection object
     */
    public Object getObject(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        if (fmd == null || !containerIsStoredInSingleColumn())
        {
            throw new JDOFatalInternalException(failureMessage("getObject"));
        }
        return getDataStoreMapping(0).getObject(resultSet, exprIndex[0]);
    }

    /**
     * Accessor for the datastore class.
     * @return The datastore class
     */
    public DatastoreContainerObject getDatastoreContainer()
    {
        if (containerIsStoredInSingleColumn())
        {
            // Serialised into owner table
            return datastoreContainer;
        }
        else
        {
            return null;
        }
    }

    /**
     * Accessor for the number of datastore fields
     * @return The number of datastore fields
     */
    public int getNumberOfDatastoreFields()
    {
        if (containerIsStoredInSingleColumn())
        {
            // Serialised into owner table
            return super.getNumberOfDatastoreFields();
        }
        else
        {
            // By default, we have no columns as such for the container
            return 0;
        }
    }

    /**
     * Accessor for a datastore mapping
     * @param index The id of the mapping
     * @return The datastore mapping
     */
    public DatastoreMapping getDataStoreMapping(int index)
    {
        if (containerIsStoredInSingleColumn())
        {
            // Serialised into owner table
            return super.getDataStoreMapping(index);
        }
        else
        {
            throw new NoDatastoreMappingException(fieldName);
        }
    }

    /**
     * Accessor for the datastore mappings for this java type
     * @return The datastore mapping(s)
     */
    public DatastoreMapping[] getDataStoreMappings()
    {
        if (containerIsStoredInSingleColumn())
        {
            // Serialised into owner table
            return super.getDataStoreMappings();
        }
        else
        {
            throw new NoDatastoreMappingException(fieldName);
        }
    }

    /**
     * Convenience method to return if the container (collection or map) is stored
     * in the owning table as a column. The container is stored in a single column in the following situations :-
     * <UL>
     * <LI>The FieldMetaData has 'serialized="true"'</LI>
     * <LI>The collection has embedded-element="true" but no join table (and so serialised)</LI>
     * <LI>The map has embedded-key/value="true" but no join table (and so serialised)</LI>
     * <LI>The array has embedded-element="true" but no join table (and so serialised)</LI>
     * <LI>The array has no join table and non-PC elements (and so serialised)</LI>
     * </UL>
     * @return Whether it is stored in a single column in the main table.
     */
    protected boolean containerIsStoredInSingleColumn()
    {
        return (fmd != null && fmd.isSerialized());
    }

    /**
     * This mapping is included in the select statement.
     * @return Whether to include in select statement
     */
    public boolean includeInFetchStatement()
    {
        // Only include in a fetch when it is serialised into 1 column
        return containerIsStoredInSingleColumn();
    }

    /**
     * This mapping is included in the update statement.
     * @return Whether to include in update statement
     */
    public boolean includeInUpdateStatement()
    {
        // Only include in an update when it is serialised into 1 column
        return containerIsStoredInSingleColumn();
    }

    /**
     * This mapping is included in the insert statement.
     * @return Whether to include in insert statement
     */
    public boolean includeInInsertStatement()
    {
        // Only include in an insert when it is serialised into 1 column
        return containerIsStoredInSingleColumn();
    }

    /**
     * Hash code function.
     * @return The hash code for this object
     */
    public int hashCode()
    {
        return fmd == null || storeMgr == null ? super.hashCode() : (fmd.hashCode() ^ storeMgr.hashCode());
    }

    /**
     * Accessor for a sample value for this mapping.
     * @return The sample value
     */
    public Object getSampleValue(ClassLoaderResolver clr)
    {
        return null;
    }

    /**
     * Method to return a SCO wrapper for this container.
     * This will return a wrapper suitable for the passed instantiated type
     * and if it is null will return a wrapper for the declared type of the field.
     * @param ownerSM State Manager for the owning bject
     * @param fieldName The field name
     * @return The SCO wrapper object
     */
    protected Object newWrapper(StateManager ownerSM, String fieldName)
    {
        String typeName = getType();
        if (instantiatedType != null)
        {
            // If we have a current value, use the instantiated type to define the wrapper type
            typeName = instantiatedType.getName();
        }
        Class newType = storeMgr.getPMFContext().getTypeManager().getSecondClassWrapper(typeName);
        if (newType == null)
        {
            // Users implementation not supported directly so try the defined type
            typeName = fmd.getTypeName();
            newType = storeMgr.getPMFContext().getTypeManager().getSecondClassWrapper(typeName);
            if (newType == null)
            {
                throw new JDOUserException(LOCALISER.msg("ContainerMapping.NewWrapperTypeNotValid", typeName));
            }
        }
        return ClassUtils.newInstance(newType,
            new Class[]{org.jpox.StateManager.class, String.class}, new Object[]{ownerSM, fmd.getName()});
    }

    // ---------------- Implementation of MappingCallbacks --------------------

    /**
     * Method to be called after any fetch of the owner class element.
     * @param sm StateManager of the owner
     */
    public void postFetch(StateManager sm)
    {
        if (containerIsStoredInSingleColumn())
        {
            // Do nothing when serialised since we are handled in the main request
            return;
        }

        // Retrieve the type of the class so we can instantiate the correct wrapper type
        Object value = sm.provideField(fmd.getAbsoluteFieldNumber());
        if (instantiatedType == null)
        {
            // instantiated type is not yet set, so set it based on what has been provided
            if (value == null)
            {
                if (fmd.getOrderMetaData() != null && fmd.getType() == java.util.Collection.class)
                {
                    // We have a collection and the user has defined an ordering, so give them a List-type as a minimum
                    instantiatedType = java.util.ArrayList.class;
                }
            }
            else
            {
                // Just use the provided type
                instantiatedType = value.getClass();
            }
        }

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newWrapper(sm, fieldName));
    }

    /**
     * Method to check if an element is already persistent, or is managed by a different
     * Persistencemanager. If not persistent, this will persist it.
     * @param sm The state manager of this owner
     * @param element The element
     */
    protected void validateElementForWriting(StateManager sm, Object element)
    {
        if (element != null)
        {
            if (element instanceof PersistenceCapable)
            {
                // Element = PC stored in own table
                PersistenceManager pm = sm.getPersistenceManager();

                if (!JDOHelper.isPersistent(element))
                {
                    boolean exists = false;
                    if (JDOHelper.isDetached(element))
                    {
                        try
                        {
                            // Check if the (attached) object exists in this datastore
                            Object obj = pm.getObjectById(((PersistenceCapable)element).jdoGetObjectId(), true, false, element.getClass().getName());
                            if (obj != null)
                            {
                                // PM.getObjectById creates a dummy object to represent this object and automatically
                                // enlists it in the txn. Evict it to avoid issues with reachability
                                StateManager objSM = pm.findStateManager((PersistenceCapable)obj);
                                if (objSM != null)
                                {
                                    pm.evictFromTransaction(objSM);
                                }
                            }
                            exists = true;
                        }
                        catch (JDOObjectNotFoundException onfe)
                        {
                            exists = false;
                        }
                    }
                    if (!exists)
                    {
                        pm.makePersistentInternal(element, null);
                    }
                }
                else if (pm != JDOHelper.getPersistenceManager(element))
                {
                    //TODO fix this message
                    throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Collection.WriteInvalidWithDifferentPM"),JDOHelper.getObjectId(element));
                }
            }
        }
    }
    
    /**
     * Method to check if an element is already persistent, or is managed by a different
     * Persistencemanager. If not persistent, this will persist it.
     * @param sm The state manager of this owner
     * @param elements The elements
     */
    protected void validateElementsForWriting(StateManager sm, Object elements)
    {
        if (elements != null)
        {
            if (elements.getClass().isArray())
            {
                if (!elements.getClass().getComponentType().isPrimitive())
                {
                    Object[] obj = ((Object[]) elements);
                    for (int i = 0; i < obj.length; i++)
                    {
                        validateElementForWriting(sm, obj[i]);
                    }
                }
            }
            else if (elements instanceof Collection)
            {
                Collection col = (Collection) elements;
                Iterator it = col.iterator();
                while (it.hasNext())
                {
                    validateElementForWriting(sm, it.next());
                }
            }
        }
    }
}
