/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.lang.reflect.Array;
import java.util.List;

import javax.jdo.JDOFatalUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.StateManager;
import org.jpox.metadata.MetaDataUtils;
import org.jpox.store.expression.ArrayExpression;
import org.jpox.store.expression.ArrayLiteral;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.scostore.ArrayStore;

/**
 * Mapping for an array.
 *
 * @version $Revision: 1.14.2.1 $
 */
public class ArrayMapping extends AbstractContainerMapping implements MappingCallbacks
{
    private static Byte mappingSampleValue = new Byte("0");

    /**
     * Equality operator.
     * @param obj Object to compare against
     * @return Whether they are equal
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!obj.getClass().equals(getClass()))
        {
            return false;
        }

        ArrayMapping sm = (ArrayMapping)obj;

        return fmd.equals(sm.fmd) && storeMgr.equals(sm.storeMgr);
    }

    /**
     * Accessor for the Java type represented here.
     * @return The java type
     */
    public Class getJavaType()
    {
        if (fmd != null)
        {
            return fmd.getType();
        }
        return null;
    }

    /**
     * Accessor for a sample value for this mapping.
     * @return The sample value
     */
    public Object getSampleValue(ClassLoaderResolver clr)
    {
        return mappingSampleValue;
    }

    /**
     * Accessor for the Store of the contents.
     * @param clr The ClassLoaderResolver
     * @return The backing store
     **/
    public synchronized ArrayStore getBackingStore(ClassLoaderResolver clr)
    {
        if (containerIsStoredInSingleColumn())
        {
            // No backing store needed since stored in a single column
            return null;
        }

        if (contentsStore == null)
        {
            contentsStore = storeMgr.getBackingStoreForArray(fmd, contentsTable, clr);
        }

        return (ArrayStore)contentsStore;
    }

    /**
     * Convenience method to return if the array is stored in the owning table as a column.
     * Overrides the superclass since arrays can be stored in a single column also when the no join is
     * specified and the array is of a primitive/wrapper type.
     * @return Whether it is stored in a single column in the main table.
     */
    protected boolean containerIsStoredInSingleColumn()
    {
        if (super.containerIsStoredInSingleColumn())
        {
            return true;
        }
        if (fmd != null && fmd.hasArray() && fmd.getJoinMetaData() == null && MetaDataUtils.arrayStorableAsByteArrayInSingleColumn(fmd))
        {
            return true;
        }
        return false;
    }

    // ---------------------- Implementation of MappingCallbacks ----------------------------------

    /**
     * Method to be called after the insert of the owner class element.
     * @param sm StateManager of the owner
     **/
    public void postInsert(StateManager sm)
    {
        Object value = sm.provideField(fmd.getAbsoluteFieldNumber());
        if (containerIsStoredInSingleColumn())
        {
            // Check that the elements are not managed by other PM, or are not yet persisted
            validateElementsForWriting(sm, value);
            return;
        }

        if (value == null)
        {
            return;
        }

        // Insert the array
        getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).set(sm, value);
    }

    /**
     * Method to be called after any fetch of the owner class element.
     * @param sm StateManager of the owner
     */
    public void postFetch(StateManager sm)
    {
        if (containerIsStoredInSingleColumn())
        {
            // Do nothing when stored in a single column since we are handled in the main request
            return;
        }

        List elements = getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).getArray(sm);
        if (elements != null)
        {
            boolean primitiveArray = fmd.getType().getComponentType().isPrimitive();
            Object array = Array.newInstance(fmd.getType().getComponentType(), elements.size());
            for (int i=0;i<elements.size();i++)
            {
                Object element = elements.get(i);
                if (primitiveArray)
                {
                    // Handle the conversion back to the primitive
                    if (element instanceof Boolean)
                    {
                        Array.setBoolean(array, i, ((Boolean)element).booleanValue());
                    }
                    else if (element instanceof Byte)
                    {
                        Array.setByte(array, i, ((Byte)element).byteValue());
                    }
                    else if (element instanceof Character)
                    {
                        Array.setChar(array, i, ((Character)element).charValue());
                    }
                    else if (element instanceof Double)
                    {
                        Array.setDouble(array, i, ((Double)element).doubleValue());
                    }
                    else if (element instanceof Float)
                    {
                        Array.setFloat(array, i, ((Float)element).floatValue());
                    }
                    else if (element instanceof Integer)
                    {
                        Array.setInt(array, i, ((Integer)element).intValue());
                    }
                    else if (element instanceof Long)
                    {
                        Array.setLong(array, i, ((Long)element).longValue());
                    }
                    else if (element instanceof Short)
                    {
                        Array.setShort(array, i, ((Short)element).shortValue());
                    }
                }
                else
                {
                    Array.set(array, i, element);
                }
            }
            if (elements.size() == 0)
            {
                sm.replaceField(fmd.getAbsoluteFieldNumber(), null);
            }
            else
            {
                sm.replaceField(fmd.getAbsoluteFieldNumber(), array);
            }
        }
        else
        {
            sm.replaceField(fmd.getAbsoluteFieldNumber(), null);
        }
    }

    /**
     * Method to be called after any update of the owner class element.
     * @param sm StateManager of the owner
     */
    public void postUpdate(StateManager sm)
    {
        Object value = sm.provideField(fmd.getAbsoluteFieldNumber());
        if (containerIsStoredInSingleColumn())
        {
            // Check that the elements are not managed by other PM, or are not yet persisted
            validateElementsForWriting(sm, value);
            return;
        }

        if (value == null)
        {
            // array is now null so remove any elements in the array
            getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).clear(sm);
            return;
        }

        // Update the datastore
        getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).clear(sm);
        getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).set(sm, value);
    }

    /**
     * Method to be called before any delete of the owner class element,
     * if the field in the owner is dependent
     * @param sm StateManager of the owner
     **/
    public void preDelete(StateManager sm)
    {
        if (containerIsStoredInSingleColumn())
        {
            // Do nothing when stored in a single column since we are handled in the main request
            return;
        }

        // Clear the array
        getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).clear(sm);
    }

    /**
     * Method to delete all dependent objects (the contents of the array).
     * @param sm State Manager of the owner
     */
    public void deleteDependent(StateManager sm)
    {
        if (containerIsStoredInSingleColumn())
        {
            // Do nothing when stored in a single column since we are handled in the main request
            return;
        }

        // Delete dependent elements in the array
        // TODO Implement delete-dependent for arrays
    }

    // --------------------------------------- JDOQL Query Methods ---------------------------------

    /**
     * Accessor for a literal representing this type.
     * @param qs The Query
     * @param value the value of this object in the literal
     * @return The literal
     */
    public ScalarExpression newLiteral(QueryExpression qs, 
                                       Object value)
    {
        if (containerIsStoredInSingleColumn())
        {
            throw new JDOFatalUserException(LOCALISER.msg("ArrayMapping.QueryInvalidIfSingleColumn", fmd.getFullFieldName()));
        }
        return new ArrayLiteral(qs, this, value);
    }

    /**
     * Accessor for a scalar expression involving this object.
     * @param qs The Query
     * @param te The table holding this object.
     * @return The expression
     */
    public ScalarExpression newScalarExpression(QueryExpression qs,
                                                LogicSetExpression te)
    {
        if (containerIsStoredInSingleColumn())
        {
            throw new JDOFatalUserException(LOCALISER.msg("ArrayMapping.QueryInvalidIfSingleColumn", fmd.getFullFieldName()));
        }
        return new ArrayExpression(qs, datastoreContainer.getIDMapping(), te, getBackingStore(qs.getClassLoaderResolver()), fieldName);
    }
}
