/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Andy Jefferson - coding standards
2005 Andy Jefferson - basic serialisation support
2005 Andy Jefferson - updated serialisation using SCOUtils methods
2005 Andy Jefferson - changed to allow the store to be set or list
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.util.Collection;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOHelper;

import org.jpox.ClassLoaderResolver;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.sco.SCO;
import org.jpox.store.expression.CollectionExpression;
import org.jpox.store.expression.CollectionLiteral;
import org.jpox.store.expression.CollectionSubqueryExpression;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.query.Queryable;
import org.jpox.store.scostore.CollectionStore;
import org.jpox.store.scostore.ListStore;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.JavaUtils;

/**
 * Mapping for Collection (Set/List) types.
 *
 * @version $Revision: 1.41 $
 **/
public class CollectionMapping extends AbstractContainerMapping implements MappingCallbacks
{
    /**
     * Return whether the collection is list based.
     * @return Whether it needs list ordering
     */
    protected boolean isListBased()
    {
        if (instantiatedType == null)
        {
            return false;
        }
        else if (java.util.List.class.isAssignableFrom(instantiatedType))
        {
            return true;
        }
        else if (JavaUtils.isJRE1_5OrAbove())
        {
            if (java.util.Queue.class.isAssignableFrom(instantiatedType))
            {
                // Queue needs ordering
                return true;
            }
        }
        return false;
    }

    /**
     * Accessor for the Store of the contents.
     * @param clr The ClassLoaderResolver
     * @return The backing store
     **/
    public synchronized CollectionStore getBackingStore(ClassLoaderResolver clr)
    {
        if (containerIsStoredInSingleColumn())
        {
            return null;
        }

        if (contentsStore == null)
        {
            contentsStore = storeMgr.getBackingStoreForCollection(fmd, contentsTable, clr, 
                instantiatedType != null, isListBased());
        }
        else if (instantiatedType != null)
        {
            // Check if the existing contentsStore needs changing because the instantiatedType is inconsistent
            // This happens if we have a declaredType of java.util.Collection and the user then
            // assigns a List-based type to the field
            boolean recreateBackingStore = false;
            if (contentsStore instanceof ListStore && java.util.Set.class.isAssignableFrom(instantiatedType))
            {
                // Currently using List but need Set
                recreateBackingStore = true;
            }
            else if (contentsStore instanceof SetStore  && java.util.List.class.isAssignableFrom(instantiatedType))
            {
                // Currently using Set but need List
                recreateBackingStore = true;
            }
            if (recreateBackingStore)
            {
                contentsStore = storeMgr.getBackingStoreForCollection(fmd, contentsTable, clr,
                    instantiatedType != null, isListBased());
            }
        }
        return (CollectionStore)contentsStore;
    }

    /**
     * Equality operator.
     * @param obj Object to compare against
     * @return Whether they are equal
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!obj.getClass().equals(getClass()))
        {
            return false;
        }

        CollectionMapping sm = (CollectionMapping)obj;

        return fmd.equals(sm.fmd) && storeMgr.equals(sm.storeMgr);
    }

    /**
     * Accessor for the Java type represented here.
     * @return The java type
     */
    public Class getJavaType()
    {
        return Collection.class;
    }

    // --------------- Implementation of MappingCallbacks -------------------

    /**
     * Method to be called after the insert of the owner class element.
     * @param sm StateManager of the owner
     **/
    public void postInsert(StateManager sm)
    {
        Collection value = (Collection) sm.provideField(fmd.getAbsoluteFieldNumber());
        if (containerIsStoredInSingleColumn())
        {
            // Check that the elements are not managed by other PM, or are not yet persisted
            validateElementsForWriting(sm,value); // Why do this ? We are storing them serialised/embedded in a single col!
            return;
        }

        if (value == null)
        {
            // replace null collections with an empty SCO wrapper
            sm.replaceField(fmd.getAbsoluteFieldNumber(), newWrapper(sm, fieldName));
            return;
        }

        if (instantiatedType == null)
        {
            // Set the instantiated type since not yet set
            instantiatedType = value.getClass();
        }

        // Check if some elements need attaching
        // TODO Investigate if we can just use the attachCopy route below and skip off this check
        boolean needsAttaching = false;
        Object[] collElements = value.toArray();
        for (int i=0;i<collElements.length;i++)
        {
            if (JDOHelper.isDetached(collElements[i]))
            {
                needsAttaching = true;
                break;
            }
        }

        if (needsAttaching)
        {
            // Create a wrapper and attach the elements (and add the others)
            SCO collWrapper = (SCO)newWrapper(sm, fieldName);
            collWrapper.attachCopy(value);

            // Replace the field with the SCO wrapper
            sm.replaceField(fmd.getAbsoluteFieldNumber(), collWrapper);
        }
        else
        {
            // Add the elements direct to the backing store
            CollectionStore backingStore = getBackingStore(sm.getPersistenceManager().getClassLoaderResolver());
            if (backingStore instanceof ListStore)
            {
                ((ListStore)backingStore).addAll(sm, value, true);
            }
            else
            {
                getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).addAll(sm, value);
            }

            // Replace the field with a SCO wrapper (which will pick up the elements)
            sm.replaceField(fmd.getAbsoluteFieldNumber(), newWrapper(sm, fieldName));
        }
    }

    /**
     * Method to be called after any update of the owner class element.
     * @param sm StateManager of the owner
     */
    public void postUpdate(StateManager sm)
    {
        Collection value = (Collection) sm.provideField(fmd.getAbsoluteFieldNumber());
        if (containerIsStoredInSingleColumn())
        {
            // Check that the elements are not managed by other PM, or are not yet persisted
            validateElementsForWriting(sm,value); // Why do this ? We are storing them serialised/embedded in a single col!
            return;
        }

        if (value == null)
        {
            // remove any elements in the collection and replace it with an empty SCO wrapper
            getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).clear(sm);
            sm.replaceField(fmd.getAbsoluteFieldNumber(), newWrapper(sm, fieldName));
            return;
        }

        if (value instanceof SCO)
        {
            SCO sco = (SCO) value;
            if (sm.getObject() == sco.getOwner() && fieldName.equals(sco.getFieldName()))
            {
                return;
            }

            if (sco.getOwner() != null)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("CollectionMapping.WrongOwnerError"));
            }
        }

        if (instantiatedType == null)
        {
            // Set the instantiated type since not yet set
            instantiatedType = value.getClass();
        }

        // TODO Change this. To remove all elements and then add a new collection is very inefficient!
        CollectionStore backingStore = getBackingStore(sm.getPersistenceManager().getClassLoaderResolver());
        backingStore.clear(sm);
        if (backingStore instanceof ListStore)
        {
            ((ListStore)backingStore).addAll(sm, value, true);
        }
        else
        {
            getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).addAll(sm, value);
        }

        // This is one alternative but involves using remove()/contains() and this brings in issues when handling serialised elements
//        SCOUtils.updateStoreWithCollection(getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()), sm, value);

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newWrapper(sm, fieldName));
    }

    /**
     * Method to be called before any delete of the owner class element.
     * @param sm StateManager of the owner
     **/
    public void preDelete(StateManager sm)
    {
        // Do nothing - dependent deletion is performed by deleteDependent()
    }

    /**
     * Method to delete all dependent objects (the contents of the Collection).
     * @param sm State Manager of the owner
     */
    public void deleteDependent(StateManager sm)
    {
        if (containerIsStoredInSingleColumn())
        {
            // Do nothing when serialised since we are handled in the main request
            return;
        }

        boolean isDependentElement = fmd.getCollection().isDependentElement();
        boolean hasJoin = (fmd.getJoinMetaData() != null);
        boolean hasFK = false;
        if (!hasJoin)
        {
            if (fmd.getElementMetaData() != null && fmd.getElementMetaData().getForeignKeyMetaData() != null)
            {
                // FK collection, using <element> FK spec
                hasFK = true;
            }
            else if (fmd.getForeignKeyMetaData() != null)
            {
                // FK collection, using <field> FK spec
                hasFK = true;
            }
            AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(sm.getPersistenceManager().getClassLoaderResolver());
            if (relatedFmd != null && relatedFmd.getForeignKeyMetaData() != null)
            {
                // FK collection (bidir), using <field> FK spec at other end
                hasFK = true;
            }
        }
        if (sm.getPersistenceManager().getPMFContext().getPmfConfiguration().getDeletionPolicy().equals("JDO2"))
        {
            // JDO2 doesnt currently (2.0 spec) take note of foreign-key
            hasFK = false;
        }

        if (isDependentElement || hasJoin || !hasFK)
        {
            // If we have a FK with a FK relation we do nothing here and just let the delete be handled by the datastore
            // otherwise we need to clear things out for the user
            getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).clear(sm);
        }
    }

    // ------------------------------- JDOQL Query Methods --------------------------------------

    /**
     * Accessor for a literal representing this type.
     * @param qs The Query
     * @param value the value of this object in the literal
     * @return The literal
     */
    public ScalarExpression newLiteral(QueryExpression qs, Object value)
    {
        if (containerIsStoredInSingleColumn())
        {
            throw new JDOFatalUserException(LOCALISER.msg("ContainerMapping.QueryInvalidIfSerialised", fmd.getFullFieldName()));
        }
        if (value instanceof Queryable)
        {
            return new CollectionSubqueryExpression(qs, ((Queryable)value).newQueryStatement());
        }
        else
        {
            return new CollectionLiteral(qs, this, (Collection)value);
        }
    }

    /**
     * Accessor for a scalar expression involving this object.
     * @param qs The Query
     * @param te The table holding this object.
     * @return The expression
     */
    public ScalarExpression newScalarExpression(QueryExpression qs, LogicSetExpression te)
    {
        if (containerIsStoredInSingleColumn())
        {
            throw new JDOFatalUserException(LOCALISER.msg("ContainerMapping.QueryInvalidIfSerialised", fmd.getFullFieldName()));
        }
        return new CollectionExpression(qs, datastoreContainer.getIDMapping(), te, getBackingStore(qs.getClassLoaderResolver()), fieldName);
    }
}