/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.EmbeddedMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.state.StateManagerFactory;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.expression.ObjectExpression;
import org.jpox.store.expression.ObjectLiteral;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;

/**
 * Mapping for an embedded PC object. 
 * The PC object can be embedded directly (1-1 relation) or be the element of 
 * a collection, or be the key or value of a map.
 *
 * @version $Revision: 1.16 $
 **/
public abstract class EmbeddedMapping extends SingleFieldMapping implements SimpleDatastoreRepresentation
{
    /** Mappings of the fields of the embedded PC. */
    protected List javaTypeMappings;

    /** ClassLoader resolver */
    protected ClassLoaderResolver clr;

    /** EmbeddedMetaData for the object being embedded. */
    protected EmbeddedMetaData emd;

    /** Type name for the object being embedded. */
    protected String typeName;

    /** Type of PC object. Corresponds to the values in StateManagerImpl. */
    protected int objectType = -1;

    /**
     * Initialize this JavaTypeMapping with the given DatastoreAdapter for
     * the given FieldMetaData.
     * <p>
     * Subclass must override this method to call
     * initialize(DatastoreAdapter, AbstractPropertyMetaData, DatastoreContainerObject, ClassLoaderResolver, EmbeddedMetaData, String, int)
     *  
     * @param dba The Datastore Adapter that this Mapping should use.
     * @param fmd FieldMetaData for the field to be mapped (if any)
     * @param container The datastore container storing this mapping (if any)
     * @param clr the ClassLoaderResolver
     * @throws JDOFatalInternalException
     */
    public void initialize(DatastoreAdapter dba, AbstractPropertyMetaData fmd, DatastoreContainerObject container, ClassLoaderResolver clr)
    {
        throw new JDOFatalInternalException("subclass must override this method");
    }
    
    /**
     * Initialize this JavaTypeMapping with the given DatastoreAdapter for
     * the given FieldMetaData.
     * 
     * @param dba Datastore Adapter
     * @param fmd AbstractPropertyMetaData for the field
     * @param datastoreContainer Table for persisting this field
     * @param clr The ClassLoaderResolver
     * @param emd Embedded MetaData for the object being embedded
     * @param typeName type of the embedded PC object
     * @param objectType Type of the PC object being embedded (see StateManagerImpl object types)
     */
    public void initialize(DatastoreAdapter dba, 
                           AbstractPropertyMetaData fmd,
                           DatastoreContainerObject datastoreContainer,
                           ClassLoaderResolver clr,
                           EmbeddedMetaData emd,
                           String typeName,
                           int objectType)
    {
    	super.initialize(dba, fmd, datastoreContainer, clr);
        this.clr = clr;
        this.emd = emd;
        this.typeName = typeName;
        this.objectType = objectType;

        // Find the MetaData for the embedded PC class
        AbstractClassMetaData pcCmd = fmd.getAbstractClassMetaData().getMetaDataManager().getMetaDataForClass(typeName, clr);
        if (pcCmd == null)
        {
            // Not found so maybe its a persistent interface
            pcCmd = fmd.getAbstractClassMetaData().getMetaDataManager().getMetaDataForInterface(clr.classForName(typeName), clr);
        }

        AbstractPropertyMetaData[] embFmds;
        if (emd == null && pcCmd.isEmbeddedOnly())
        {
            // No <embedded> block yet the class is defined as embedded-only so just use its own definition of fields
            embFmds = pcCmd.getManagedFields();
        }
        else
        {
            // <embedded> block so use those field definitions
            embFmds = emd.getFieldMetaData();
        }

        // Add all fields of the embedded class (that are persistent)
        int[] pcFieldNumbers = pcCmd.getAllFieldNumbers();
        for (int i=0;i<pcFieldNumbers.length;i++)
        {
            AbstractPropertyMetaData pcFmd = pcCmd.getManagedFieldAbsolute(pcFieldNumbers[i]);
            if (pcFmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
            {
                if (emd != null && emd.getOwnerField() != null && emd.getOwnerField().equals(pcFmd.getName()))
                {
                    // Do nothing since we dont map owner fields (since the owner is the containing object)
                }
                else
                {
                    AbstractPropertyMetaData embeddedFmd = null;
                    for (int j=0;j<embFmds.length;j++)
                    {
                        // Why are these even possible ? Why are they here ? Why dont they use localised messages ?
                        if (embFmds[j] == null)
                        {
                            throw new RuntimeException("embFmds[j] is null"+pcCmd.toString()+" type "+ typeName);
                        }
                        if (pcCmd.getField(embFmds[j].getName()) == null)
                        {
                            throw new RuntimeException("pcCmd.getField(embFmds[j].getName()) is null"+pcCmd.toString()+" type "+ typeName+" embFmds[j].getName() "+embFmds[j].getName());
                        }

                        // Check if this field has a mapping in the embedded specification
                        if (pcCmd.getField(embFmds[j].getName()).getAbsoluteFieldNumber() == pcFieldNumbers[i])
                        {
                            embeddedFmd = embFmds[j];
                            break;
                        }
                    }

                    JavaTypeMapping embFmdMapping;
                    if (embeddedFmd != null)
                    {
                        // User has provided a field definition so map with that
                        embFmdMapping = dba.getMappingManager().getMapping(datastoreContainer, embeddedFmd, dba, clr, MappingManager.MAPPING_FIELD);
                    }
                    else
                    {
                        // User hasn't provided a field definition so map with the classes own definition
                        embFmdMapping = dba.getMappingManager().getMapping(datastoreContainer, pcFmd, dba, clr, MappingManager.MAPPING_FIELD);
                    }

                    this.addJavaTypeMapping(embFmdMapping);
                    for (int j=0; j<embFmdMapping.getNumberOfDatastoreFields(); j++)
                    {
                        DatastoreMapping datastoreMapping = embFmdMapping.getDataStoreMapping(j);
                        this.addDataStoreMapping(datastoreMapping);
                    }
                }
            }
        }
    }

    /**
     * Method to prepare a field mapping for use in the datastore.
     * Overridden so it does nothing
     */
    protected void prepareDatastoreMapping()
    {
    }

    /**
     * Add a new JavaTypeMapping to manage.
     * @param mapping the JavaTypeMapping
     */
    public void addJavaTypeMapping(JavaTypeMapping mapping)
    {
        if (javaTypeMappings == null)
        {
            javaTypeMappings = new ArrayList();
        }
        if (mapping == null)
        {
            throw new JDOFatalInternalException("mapping argument in EmbeddedMapping.addJavaTypeMapping is null");
        }
        javaTypeMappings.add(mapping);
    }

    /**
     * Accessor for the number of java type mappings
     * @return Number of java type mappings of the fields of the embedded PC element
     */
    public int getNumberOfJavaTypeMappings()
    {
        return javaTypeMappings != null ? javaTypeMappings.size() : 0;
    }

    /**
     * Accessor for the java type mappings
     * @param i the index position of the java type mapping
     * @return the java type mapping
     */
    public JavaTypeMapping getJavaTypeMapping(int i)
    {
        if (javaTypeMappings == null)
        {
            return null;
        }
        return (JavaTypeMapping)javaTypeMappings.get(i);
    }

    /**
     * Accessor for the sub type mapping for a particular field name
     * @param fieldName The field name
     * @return The type mapping for that field in the embedded object
     */
    public JavaTypeMapping getJavaTypeMapping(String fieldName)
    {
        if (javaTypeMappings == null)
        {
            return null;
        }
        Iterator iter = javaTypeMappings.iterator();
        while (iter.hasNext())
        {
            JavaTypeMapping m = (JavaTypeMapping)iter.next();
            if (m.getFieldMetaData().getName().equals(fieldName))
            {
                return m;
            }
        }
        return null;
    }

    /**
     * Mutator for the embedded object in the datastore.
     * @param pm The Persistence Manager managing this object
     * @param ps The Prepared Statement
     * @param param Param numbers in the PreparedStatement for the fields of this object
     * @param value The embedded object to use
     */
    public void setObject(PersistenceManager pm, Object ps, int[] param, Object value)
    {
        setObject(pm, ps, param, value, null, -1);
    }

    /**
     * Mutator for the embedded object in the datastore.
     * @param pm The Persistence Manager managing this object
     * @param ps The Prepared Statement
     * @param param Param numbers in the PreparedStatement for the fields of this object
     * @param value The embedded object to use
     * @param ownerSM StateManager of the owning object containing this embedded object
     * @param ownerFieldNumber Field number in the owning object where this is stored
     */
    public void setObject(PersistenceManager pm, Object ps, int[] param, Object value, StateManager ownerSM, int ownerFieldNumber)
    {
        if (value == null)
        {
            int n = 0;
            String nullColumn = null;
            String nullValue = null;
            if (emd != null)
            {
                nullColumn = emd.getNullIndicatorColumn();
                nullValue = emd.getNullIndicatorValue();
            }
            for (int i=0; i<javaTypeMappings.size(); i++)
            {
                JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
                int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
                for (int j=0; j<posMapping.length; j++)
                {
                    posMapping[j] = param[n++];
                }

                // Null out this field unless it is the null-indicator column and has a value
                // in which case apply the required value
                if (nullColumn != null && nullValue != null &&
                    mapping.getFieldMetaData().getColumnMetaData().length > 0 &&
                    mapping.getFieldMetaData().getColumnMetaData()[0].getName().equals(nullColumn))
                {
                    // Try to cater for user having an integer based column and value
                    if (mapping instanceof IntegerMapping ||
                        mapping instanceof BigIntegerMapping ||
                        mapping instanceof LongMapping ||
                        mapping instanceof ShortMapping)
                    {
                        Object convertedValue = null;
                        try
                        {
                            if (mapping instanceof IntegerMapping ||
                                mapping instanceof ShortMapping)
                            {
                                convertedValue = new Integer(nullValue);
                            }
                            else if (mapping instanceof LongMapping ||
                                mapping instanceof BigIntegerMapping)
                            {
                                convertedValue = new Long(nullValue);
                            }
                        }
                        catch (Exception e)
                        {
                        }
                        mapping.setObject(pm, ps, posMapping, convertedValue);
                    }
                    else
                    {
                        mapping.setObject(pm, ps, posMapping, nullValue);
                    }
                }
                else
                {
                    if (mapping.getNumberOfDatastoreFields() > 0)
                    {
                        mapping.setObject(pm, ps, posMapping, null);
                    }
                }
            }
        }
        else
        {
            if (!(value instanceof PersistenceCapable))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("Mapping.ObjectIsNotPersistenceCapable",value.getClass(),value));
            }

            PersistenceCapable pc = (PersistenceCapable) value;
            AbstractClassMetaData embCmd = pm.getMetaDataManager().getMetaDataForClass(pc.getClass().getName(), pm.getClassLoaderResolver());
            StateManager embSM = pm.findStateManager(pc);
            if (embSM == null || pc.jdoGetPersistenceManager() == null)
            {
                // Assign a StateManager to manage our embedded object
                embSM = StateManagerFactory.newStateManager(pc, pm, false);
                embSM.addEmbeddedOwner(ownerSM, ownerFieldNumber);
                embSM.setPcObjectType(objectType);
            }

            int n = 0;
            for (int i=0; i<javaTypeMappings.size(); i++)
            {
                JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
                int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
                for (int j=0; j<posMapping.length; j++)
                {
                    posMapping[j] = param[n++];
                }

                // Retrieve value of field from Embedded StateManager
                int embAbsFieldNum = embCmd.getFieldNumberAbsolute(mapping.getFieldMetaData().getName());
                Object fieldValue = embSM.provideField(embAbsFieldNum);

                if (mapping instanceof EmbeddedPCMapping)
                {
                    mapping.setObject(pm, ps, posMapping, fieldValue, embSM, embAbsFieldNum);
                }
                else
                {
                    if (mapping.getNumberOfDatastoreFields() > 0)
                    {
                        mapping.setObject(pm, ps, posMapping, fieldValue);
                    }
                }
            }
        }
    }

    /**
     * Accessor for the embedded object from the result set
     * @param pm Persistence Manager managing this object
     * @param rs The ResultSet
     * @param param Array of param numbers in the ResultSet for the fields of this object
     * @return The embedded object
     */
    public Object getObject(PersistenceManager pm, Object rs, int[] param)
    {
        return getObject(pm, rs, param, null, -1);
    }

    /**
     * Accessor for the embedded object from the result set
     * @param pm Persistence Manager managing this object
     * @param rs The ResultSet
     * @param param Array of param numbers in the ResultSet for the fields of this object
     * @param ownerSM StateManager of the owning object containing this embedded object
     * @param ownerFieldNumber Field number in the owning object where this is stored
     * @return The embedded object
     */
    public Object getObject(PersistenceManager pm, Object rs, int[] param, StateManager ownerSM, int ownerFieldNumber)
    {
        Object value = null;

        // Create a PersistenceCapable to put the values into
        StateManager embSM = StateManagerFactory.newStateManager(pm, getJavaType(), (Object)null);
        embSM.setPcObjectType(objectType);
        value = embSM.getObject();

        String nullColumn = null;
        String nullValue = null;
        if (emd != null)
        {
            nullColumn = emd.getNullIndicatorColumn();
            nullValue = emd.getNullIndicatorValue();
        }
        AbstractClassMetaData embCmd = pm.getMetaDataManager().getMetaDataForClass(typeName, clr);
        int n = 0;
        for (int i=0; i<javaTypeMappings.size(); i++)
        {
            JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
            int embAbsFieldNum = embCmd.getFieldNumberAbsolute(mapping.getFieldMetaData().getName());
            if (mapping instanceof EmbeddedPCMapping)
            {
                // We have a nested embedded
                int numSubParams = mapping.getNumberOfDatastoreFields();
                int[] subParam = new int[numSubParams];
                int k = 0;
                for (int j=n;j<n+numSubParams;j++)
                {
                    subParam[k++] = param[j];
                }
                n += numSubParams;

                // Use the sub-object mapping to extract the value for that object
                Object subValue = mapping.getObject(pm, rs, subParam, embSM, embAbsFieldNum);
                if (subValue != null)
                {
                    embSM.replaceField(embAbsFieldNum, subValue);
                }

                // TODO Check the null column and its value in the sub-embedded ?
            }
            else
            {
                // Extract the value(s) for this field and update the PC if it is not null
                int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
                for (int j=0; j<posMapping.length; j++)
                {
                    posMapping[j] = param[n++];
                }
                Object fieldValue = mapping.getObject(pm, rs, posMapping);
                if (fieldValue != null)
                {
                    embSM.replaceField(embAbsFieldNum, fieldValue);
                }

                // Check for the null column and its value
                if (nullColumn != null &&
                    mapping.getFieldMetaData().getColumnMetaData()[0].getName().equals(nullColumn))
                {
                    if ((nullValue == null && fieldValue == null) ||
                        (nullValue != null && fieldValue.toString().equals(nullValue)))
                    {
                        value = null;
                        break;
                    }
                }
            }
        }

        // Update owner field in the element (if present)
        if (emd != null)
        {
            String ownerField = emd.getOwnerField();
            if (ownerField != null)
            {
                int ownerFieldNumberInElement = embCmd.getFieldNumberAbsolute(ownerField);
                if (ownerFieldNumberInElement >= 0)
                {
                    embSM.replaceField(ownerFieldNumberInElement, ownerSM.getObject());
                }
            }
        }
        
        // Register our owner now that we have our values set
        if (value != null && ownerSM != null)
        {
            embSM.addEmbeddedOwner(ownerSM, ownerFieldNumber);
        }

        return value;
    }

    /**
     * Accessor for the sample value for this type.
     * @return Sample value
     */
    public Object getSampleValue(ClassLoaderResolver clr)
    {
        return null;
    }

    /**
     * Accessor for the Java type being represented here.
     * @return The Java type
     */
    public Class getJavaType()
    {
        return clr.classForName(typeName);
    }

    // --------------------------------------- JDOQL Query Methods ------------------------------------------

    public ScalarExpression newLiteral(QueryExpression qs, Object value)
    {
       ScalarExpression expr = new ObjectLiteral(qs, this, value, getType());
       return expr;        
    }

    public ScalarExpression newScalarExpression(QueryExpression qs, LogicSetExpression te)
    {
        ScalarExpression expr = new ObjectExpression(qs, this, te);
        return expr;        
    }
}