/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
    ...
**********************************************************************/
package org.jpox.store.mapping;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.Relation;
import org.jpox.store.OID;
import org.jpox.store.OIDFactory;
import org.jpox.store.exceptions.NotYetFlushedException;
import org.jpox.store.exceptions.NullValueException;
import org.jpox.util.JPOXLogger;

/**
 * Mapping to represent multiple mappings.
 * This mapping can be used to represent, for example,
 * <UL>
 * <LI>a reference field (where there are multiple implementations)</LI>
 * <LI>an element in a collection and the element has no table of its own, but multiple subclasses</LI>
 * <LI>a FK of a PC field (where there may be multiple fields in the PK of the PC object)</LI>
 * </UL>
 * @version $Revision: 1.16 $
 **/
public abstract class MultiMapping extends JavaTypeMapping
{
    /** The Java mappings represented by this mapping. */
    protected JavaTypeMapping[] javaTypeMappings = new JavaTypeMapping[0];

    /** Number of DatastoreField - convenience field to improve performance **/
    private int numberOfDatastoreFields = 0;

    /**
     * Method to add a Java type mapping for a field
     * @param mapping The mapping to add
     */
    public void addJavaTypeMapping(JavaTypeMapping mapping)
    {
        JavaTypeMapping[] jtm = javaTypeMappings; 
        javaTypeMappings = new JavaTypeMapping[jtm.length+1]; 
        System.arraycopy(jtm, 0, javaTypeMappings, 0, jtm.length);
        javaTypeMappings[jtm.length] = mapping;
    }

    /**
     * Accessor for the Java type mappings
     * @return The Java type mappings
     */
    public JavaTypeMapping[] getJavaTypeMapping()
    {
        return javaTypeMappings;
    }

    /**
     * Accessor for the number of datastore fields.
     * @return The number of datastore fields.
     */
    public int getNumberOfDatastoreFields()
    {
        if (numberOfDatastoreFields == 0)
        {
            for (int i=0; i<javaTypeMappings.length; i++)
            {
                numberOfDatastoreFields += javaTypeMappings[i].getNumberOfDatastoreFields();
            }
        }
        return numberOfDatastoreFields;
    }

    /**
     * Accessor for a datastore mapping.
     * @param index The position of the mapping to return
     * @return The datastore mapping
     */
    public DatastoreMapping getDataStoreMapping(int index)
    {
        int currentIndex = 0;
        int numberJavaMappings = javaTypeMappings.length;
        for (int i=0; i<numberJavaMappings; i++)
        {
            int numberDatastoreMappings = javaTypeMappings[i].getNumberOfDatastoreFields();
            for (int j=0; j<numberDatastoreMappings; j++)
            {
                if (currentIndex == index)
                {
                    return javaTypeMappings[i].getDataStoreMapping(j);
                }
                currentIndex++;
            }
        }

        // TODO Localise this message
        throw new JDOFatalInternalException("Invalid index " + index + " for DataStoreMapping.");
    }

    /**
     * Method to set the parameters in the PreparedStatement with the fields of
     * this object.
     * @param pm Persistence Manager
     * @param ps The PreparedStatement
     * @param pos The parameter positions
     * @param value The object to populate the statement with
     * @throws NotYetFlushedException Thrown if the object is not yet flushed to the datastore
     */
    public void setObject(PersistenceManager pm, Object ps, int[] pos, Object value)
    {
    	setObject(pm, ps, pos, value, null, -1);
    }
    
    /**
     * Sets a <code>value</code> into <code>preparedStatement</code> 
     * at position specified by <code>exprIndex</code>.
     * @param pm the PersistenceManager 
     * @param ps a datastore object that executes statements in the database 
     * @param pos the position of the value in the statement
     * @param value the value
     * @param ownerSM the owner StateManager
     * @param ownerFieldNumber the owner absolute field number
     */    
    public void setObject(PersistenceManager pm, Object ps, int[] pos, Object value, StateManager ownerSM, int ownerFieldNumber)
    {
        ClassLoaderResolver clr = pm.getClassLoaderResolver();
        int n = 0;
        boolean foundClassAssignableFromValue = false;
        NotYetFlushedException notYetFlushed = null;
        Class type = null;
        
        PersistenceCapable pc = (PersistenceCapable)value;
        if (pc != null)
        {
	        if (!pm.isInserting(pc))
	        {
	            // Object either already exists, or is not yet being inserted.
	            Object id = pc.jdoGetObjectId();
	            
	            // Check if the PersistenceCapable exists in this datastore
	            boolean requiresPersisting = false;
	            if (pc.jdoIsDetached() && ownerSM != null)
	            {
	                // Detached object that needs attaching (or persisting if detached from a different datastore)
	                requiresPersisting = true;
	            }
	            else if (id == null)
	            {
	                // Transient object, so we need to persist it
	                requiresPersisting = true;
	            }
	            else
	            {
	                if (((PersistenceManager)pc.jdoGetPersistenceManager()) != null &&
	                    pm.getPMHandle() != ((PersistenceManager)pc.jdoGetPersistenceManager()).getPMHandle())
	                {
	                    throw new JDOUserException(LOCALISER.msg("Mapping.FieldNotAssignableIsFromAnotherPM"), id);
	                }
	            }
	            
	            if (requiresPersisting)
	            {
	                // The object is either not yet persistent or is detached and so needs attaching
	                PersistenceCapable pcNew = (PersistenceCapable)pm.makePersistentInternal(value, null);
	                pm.flush();
	                id = pcNew.jdoGetObjectId();
	                if (pc.jdoIsDetached() && ownerSM != null)
	                {
	                    // Update any detached reference to refer to the attached variant
	                    ownerSM.replaceField(ownerFieldNumber, pcNew);
                        int relationType = fmd.getRelationType(clr);
                        if (relationType == Relation.ONE_TO_ONE_BI)
                        {
                            StateManager relatedSM = pm.findStateManager(pcNew);
                            relatedSM.replaceField(fmd.getRelatedFieldMetaData(clr).getAbsoluteFieldNumber(), ownerSM.getObject());
                        }
                        else if (relationType == Relation.MANY_TO_ONE_BI)
                        {
                            // TODO Update the container element with the attached variant
                            if (JPOXLogger.JDO.isInfoEnabled())
                            {
                                JPOXLogger.JDO.info("PCMapping.setObject : object " + ownerSM.getInternalObjectId() + 
                                    " has field " + ownerFieldNumber + 
                                    " that is 1-N bidirectional - should really update the reference in the relation. Not yet supported");
                            }
                        }
	                }
	            }        
		        if (getNumberOfDatastoreFields() <= 0)
		        {
		            // If the field doesnt map to any datastore fields, omit the set process
		        	return;
		        }	        
	        }
        }
        StateManager sm = null;
        if (value != null)
        {
            sm = pm.findStateManager(pc);
        } 
        try
        {
        	if (pos == null)
        	{
        		return;
        	}
        	if (sm != null)
	        {
	            sm.setStoringPC();
	        }        
	        for (int i=0; i<javaTypeMappings.length; i++)
	        {
	            if (n >=pos.length)
	            {
	            	//this means we store all implementations to the same columns, so we reset the index
	            	n = 0;
                }
	            int[] posMapping;
	            if (javaTypeMappings[i].getReferenceMapping() != null)
	            {           
	                posMapping = new int[javaTypeMappings[i].getReferenceMapping().getNumberOfDatastoreFields()];            	
	            }
	            else
	            {
	            	posMapping = new int[javaTypeMappings[i].getNumberOfDatastoreFields()];
	            }
	            for (int j=0; j<posMapping.length; j++)
	            {
	                posMapping[j] = pos[n++];
	            }
	            try
	            {
	            	if (pm.getStoreManager().getMetaDataManager().getMetaDataForInterface(clr.classForName(getType()),clr) != null)
	            	{
	            		type = clr.classForName(getType());
	            	}
	            	else
	            	{
	            		type = clr.classForName(javaTypeMappings[i].getType());
	            	}
	                if (value != null && type.isAssignableFrom(value.getClass()))
	                {
	                    foundClassAssignableFromValue = true;
                        javaTypeMappings[i].setObject(pm, ps, posMapping, value);
	                }
	                else
	                {
                        javaTypeMappings[i].setObject(pm, ps, posMapping, null);
	                }
	            }
	            catch (NotYetFlushedException e)
	            {
	                notYetFlushed = e;
	            }
	        }
	        if (notYetFlushed != null)
	        {
	            throw notYetFlushed;
	        }
        }
        finally
        {
            if (sm != null)
            {
                sm.unsetStoringPC();
            }        	
        }
        if (value != null && !foundClassAssignableFromValue)
        {
            // as required by the JDO spec, a ClassCastException is thrown
            // TODO Change this to a multiple field mapping localised message
            throw new ClassCastException(LOCALISER.msg("ReferenceMapping.NotValidImplementation", 
                fmd != null ? fmd.getFullFieldName() : "", type.getName(), value.getClass().getName()));
        }
    }

    /**
     * Method to retrieve an object of this type from the ResultSet.
     * @param pm Persistence Manager
     * @param rs The ResultSet
     * @param pos The parameter positions
     * @return The object
     */
    public Object getObject(PersistenceManager pm, final Object rs, int[] pos)
    {
        // Go through the possible types for this field and find a non-null value (if there is one)
        int n = 0;
        for (int i=0; i<javaTypeMappings.length; i++)
        {
            int[] posMapping;
            if( n >=pos.length)
            {
            	//this means we store all implementations to the same columns, so we reset the index
            	n = 0;
            }

            if (javaTypeMappings[i].getReferenceMapping() != null)
            {
                posMapping = new int[javaTypeMappings[i].getReferenceMapping().getNumberOfDatastoreFields()];
            }
            else
            {
                posMapping = new int[javaTypeMappings[i].getNumberOfDatastoreFields()];
            }
            for (int j=0; j<posMapping.length; j++)
            {
                posMapping[j] = pos[n++];
            }

            Object value = null;
            try
            {
                // Retrieve the value (PC object) for this mappings' object
                value = javaTypeMappings[i].getObject(pm, rs, posMapping);
            }
            catch (NullValueException e)
            {
                // expected if implementation object is null and has primitive
                // fields in the primary key
            }
            catch (JDOObjectNotFoundException e)
            {
                // expected, will try next implementation
            }
            if (value != null)
            {
                if (value instanceof OID)
                {
                    // What situation is this catering for exactly ?
                    String className;
                    if (javaTypeMappings[i].getReferenceMapping() != null)
                    {
                        className = javaTypeMappings[i].getReferenceMapping().getDataStoreMapping(0).getDatastoreField().getStoredJavaType();
                    }
                    else
                    {
                        className = javaTypeMappings[i].getDataStoreMapping(0).getDatastoreField().getStoredJavaType();
                    }
                    value = OIDFactory.getInstance(className, ((OID)value).keyValue());
                    return pm.getObjectById(value, false);
                }
                else if (pm.getClassLoaderResolver().classForName(getType()).isAssignableFrom(value.getClass()))               	
                {
                	return value;
                }
            }
        }
        return null;
    }
}