/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Erik Bengtson - Important changes regarding the application identity
                     support when fetching PreparedStatements. Added an
                     inner class that should be moved away.
2004 Erik Bengtson - added commentary and Localisation
2004 Andy Jefferson - added capability to handle Abstract PC fields for
                      both SingleFieldIdentity and AID
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.lang.reflect.Field;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.ResultSet;
import java.sql.SQLException;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;
import javax.jdo.identity.IntIdentity;
import javax.jdo.identity.LongIdentity;
import javax.jdo.identity.ShortIdentity;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.IdentityStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.InheritanceStrategy;
import org.jpox.metadata.Relation;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.FieldValues;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.OID;
import org.jpox.store.OIDFactory;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NotYetFlushedException;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.expression.ObjectExpression;
import org.jpox.store.expression.ObjectLiteral;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.rdbms.fieldmanager.ResultSetGetter;
import org.jpox.store.rdbms.table.CorrespondentColumnsMapping;
import org.jpox.store.rdbms.table.ElementContainerTable;
import org.jpox.util.AIDUtils;
import org.jpox.util.JPOXLogger;

/**
 * Maps a java field to a PersistenceCapable class. 
 * For PersistenceCapable classes using datastore identity most of the necessary behaviour
 * is coded in the OIDMapping super class.
 * TODO Split this from OIDMapping since a PCMapping may represent an application
 * identity object instead of a datastore identity object
 *
 * @version $Revision: 1.89.2.4 $ 
 **/
public class PersistenceCapableMapping extends OIDMapping
{
    /** Mappings for all fields necessary to represent the id of the PC object. */
    private JavaTypeMapping[] javaTypeMappings = new JavaTypeMapping[0];

    //----------------------- convenience fields to improve performance ----------------------// 
    /** ClassMetaData for the represented class. Create a new one on each getObject invoke is expensive **/
    private AbstractClassMetaData cmd;

    /** Whether this mapping represents the identity only, or the object. */
    private boolean representsIdentity = false;

    /** number of DatastoreField - convenience field to improve performance **/
    private int numberOfDatastoreFields = 0;

    /**
     * Create a new empty PersistenceCapableMapping.  The caller must call one
     * of the initialize methods to initialize the instance with the
     * DatastoreAdapter and its type.
     */
    public PersistenceCapableMapping() {
    }
    
    /**
     * Constructor used to generate a PCMapping representing only the identity of the object.
     * This is typically used where the user has selected the id in a JDOQL query as a result field.
     * @param pcMapping The mapping to base it on
     */
    protected PersistenceCapableMapping(PersistenceCapableMapping pcMapping)
    {
        initialize(pcMapping.dba, pcMapping.type);
        datastoreContainer = pcMapping.datastoreContainer;

        // Add the same field mappings to the identity
        javaTypeMappings = new JavaTypeMapping[pcMapping.javaTypeMappings.length]; 
        System.arraycopy(pcMapping.javaTypeMappings, 0, javaTypeMappings, 0, javaTypeMappings.length);

        representsIdentity = true;
    }

    /**
     * Initialize this JavaTypeMapping with the given DatastoreAdapter for
     * the given FieldMetaData.
     *  
     * @param dba The Datastore Adapter that this Mapping should use.
     * @param fmd FieldMetaData for the field to be mapped (if any)
     * @param container The datastore container storing this mapping (if any)
     * @param clr the ClassLoaderResolver
     */
    public void initialize(DatastoreAdapter dba, AbstractPropertyMetaData fmd, DatastoreContainerObject container, ClassLoaderResolver clr)
    {
    	super.initialize(dba, fmd, container, clr);
        prepareDatastoreMapping(clr);
    }
    
    /**
     * Utility to return a PC mapping representing only the identity and not the object.
     * @return The id mapping
     */
    public JavaTypeMapping getIdMapping()
    {
        return new PersistenceCapableMapping(this);
    }

    /**
     * Add a new JavaTypeMapping
     * @param mapping the JavaTypeMapping
     */
	public void addJavaTypeMapping(JavaTypeMapping mapping)
	{
	    if (mapping == null)
	    {
	        throw new JDOFatalInternalException("mapping argument in PersistenceCapableMapping.addJavaTypeMapping is null");
	    }
        JavaTypeMapping[] jtm = javaTypeMappings; 
        javaTypeMappings = new JavaTypeMapping[jtm.length+1]; 
        System.arraycopy(jtm, 0, javaTypeMappings, 0, jtm.length);
        javaTypeMappings[jtm.length] = mapping;
	}
    
    /**
     * Method to prepare the PC mapping and add its associated datastore mappings.
     */
    protected void prepareDatastoreMapping()
    {
        // Does nothing - added to prevent column creation by the method in OIDMapping
        // If we change the inheritance so this doesnt extend OIDMapping, then this can be deleted
    }

	/**
     * Method to prepare the PC mapping and add its associated datastore mappings.
     * @param clr The ClassLoaderResolver
     */
    protected void prepareDatastoreMapping(ClassLoaderResolver clr)
    {
        // Either one end of a 1-1 relation, or the N end of a N-1
        StoreManager storeMgr = datastoreContainer.getStoreManager();
        AbstractClassMetaData refCmd = storeMgr.getMetaDataManager().getMetaDataForClass(fmd.getType(), clr);
        JavaTypeMapping referenceMapping = null;
        if (refCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
        {
            // Find the actual tables storing the other end (can be multiple subclasses)
            AbstractClassMetaData[] cmds = storeMgr.getClassesManagingTableForClass(refCmd, clr);
            if (cmds != null && cmds.length > 0)
            {
                if (cmds.length > 1)
                {
                    JPOXLogger.JDO.warn("Field " + fmd.getFullFieldName() + " represents either a 1-1 relation, or a N-1 relation " + 
                        "where the other end uses \"subclass-table\" inheritance strategy and more than 1 subclasses with a table. " +
                        "This is not fully supported by JPOX");
                }
            }
            else
            {
                // No subclasses of the class using "subclasses-table" so no mapping!
                // TODO Throw an exception ?
                return;
            }
            // TODO We need a mapping for each of the possible subclass tables
            referenceMapping = storeMgr.getDatastoreClass(cmds[0].getFullClassName(), clr).getIDMapping();
        }
        else
        {
            referenceMapping = storeMgr.getDatastoreClass(fmd.getType().getName(), clr).getIDMapping();
        }

        // Generate a mapping from the columns of the referenced object to this mapping's ColumnMetaData
        CorrespondentColumnsMapping correspondentColumnsMapping = new CorrespondentColumnsMapping(fmd, referenceMapping, true);

        // Find any related field where this is part of a bidirectional relation
        int relationType = fmd.getRelationType(clr);
        boolean createDatastoreMappings = true;
        if (relationType == Relation.MANY_TO_ONE_BI)
        {
            AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
            createDatastoreMappings = (relatedFmd.getJoinMetaData() == null);
        }
        else if (relationType == Relation.ONE_TO_ONE_BI)
        {
            // Put the FK at the end without "mapped-by"
            createDatastoreMappings = (fmd.getMappedBy() == null);
        }

        // Loop through the datastore fields in the referenced class and create a datastore field for each
        for (int i=0; i<referenceMapping.getNumberOfDatastoreFields(); i++)
        {
            DatastoreMapping refDatastoreMapping = referenceMapping.getDataStoreMapping(i);
            JavaTypeMapping mapping = dba.getMapping(refDatastoreMapping.getJavaTypeMapping().getJavaType(), storeMgr);
            this.addJavaTypeMapping(mapping);

            // Create physical datastore columns where we require a FK link to the related table.
            if (createDatastoreMappings)
            {
                // Find the Column MetaData that maps to the referenced datastore field
                ColumnMetaData colmd = correspondentColumnsMapping.getColumnMetaDataByIdentifier(
                    refDatastoreMapping.getDatastoreField().getIdentifier());
                if (colmd == null)
                {
                    throw new JDOFatalUserException(LOCALISER.msg("PersistenceCapableMapping.PrimaryKeyColumnNotMapped",
                        refDatastoreMapping.getDatastoreField().getIdentifier(), toString()));
                }

                // Create a Datastore field to equate to the referenced classes datastore field
                MappingManager mmgr = datastoreContainer.getStoreManager().getMappingManager();
                DatastoreField col = mmgr.createDatastoreField(fmd, datastoreContainer, dba, mapping, 
                    colmd, refDatastoreMapping.getDatastoreField(), clr);

                // Add its datastore mapping
                DatastoreMapping datastoreMapping = mmgr.createDatastoreMapping(mapping, storeMgr, col,
                    refDatastoreMapping.getJavaTypeMapping().getJavaTypeForDatastoreMapping(i));
                this.addDataStoreMapping(datastoreMapping);
            }
            else
            {
                mapping.setReferenceMapping(referenceMapping);
            }
        }
    }

    /**
     * Accessor for the Java type mappings
     * @return The Java type mappings
     */
    public JavaTypeMapping[] getJavaTypeMapping()
    {
        return javaTypeMappings;
    }

    /**
     * Accessor for the number of datastore fields.
     * Zero datastore fields implies that the mapping uses a FK in the associated class
     * to reference back.
     * @return Number of datastore fields for this PC mapping.
     */
    public int getNumberOfDatastoreFields()
    {
        if (numberOfDatastoreFields == 0)
        {
            for (int i=0; i<javaTypeMappings.length; i++)
            {
                numberOfDatastoreFields += javaTypeMappings[i].getNumberOfDatastoreFields();
            }
        }
        return numberOfDatastoreFields;
    }

	/**
     * Accessor for a datastore mapping.
     * This method works through the java type mappings, and for each mapping through its datastore mappings
     * incrementing the index with each datastore mapping.
     * @param index The position of the mapping.
     * @return The datastore mapping.
	 */
	public DatastoreMapping getDataStoreMapping(int index)
	{
        int currentIndex = 0;
        int numberJavaMappings = javaTypeMappings.length;
        for (int i=0; i<numberJavaMappings; i++)
        {
            int numberDatastoreMappings = javaTypeMappings[i].getNumberOfDatastoreFields();
            for (int j=0; j<numberDatastoreMappings; j++)
            {
                if (currentIndex == index)
                {
                    return javaTypeMappings[i].getDataStoreMapping(j);
                }
                currentIndex++;
            }
        }
        // TODO Localise this message
        throw new JDOFatalInternalException("Invalid index " + index + " for DataStoreMapping.");
	}

	/**
	 * Method to set an object in the datastore.
	 * @param pm The Persistence Manager
	 * @param ps The Prepared Statement
	 * @param param The parameter ids in the statement
	 * @param value The value to put in the statement at these ids
	 * @throws NotYetFlushedException
	 */
    public void setObject(PersistenceManager pm, Object ps, int[] param, Object value)
    {
        setObject(pm, ps, param, value, null, -1);
    }

    /**
     * Method to set an object reference (FK) in the datastore.
     * @param pm The Persistence Manager
     * @param ps The Prepared Statement
     * @param param The parameter ids in the statement
     * @param value The value to put in the statement at these ids
     * @param ownerSM StateManager for the owner object
     * @param ownerFieldNumber Field number of this PC object in the owner
     * @throws NotYetFlushedException
     */
    public void setObject(PersistenceManager pm, Object ps, int[] param, Object value, StateManager ownerSM, int ownerFieldNumber)
    {
        Object id;

        if (value == null)
        {
            // Null out the PC object
            int n=0;
            for (int i=0; i<javaTypeMappings.length; i++)
            {
                JavaTypeMapping mapping = javaTypeMappings[i];
                if (mapping.getNumberOfDatastoreFields() > 0)
                {
                    // Only populate the PreparedStatement for the object if it has any datastore mappings
                    int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
                    for (int j=0; j<posMapping.length; j++)
                    {
                        posMapping[j] = param[n++];
                    }
                    mapping.setObject(pm, ps, posMapping, null);
                }
            }
        }
        else
        {
            if (!(value instanceof PersistenceCapable))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("Mapping.ObjectIsNotPersistenceCapable",value.getClass(), value));
            }

            StateManager sm = null;
            if (value != null)
            {
                sm = pm.findStateManager((PersistenceCapable)value);
            }

            try
            {
                PersistenceCapable pc = (PersistenceCapable)value;

                // Check if the field is attributed in the datastore
                boolean isStrategyDatastoreAttributed = false;
                if (this.fmd != null)
                {
                    // Object is associated to a field (i.e not a join table)
                    AbstractClassMetaData acmd = pm.getMetaDataManager().getMetaDataForClass(this.fmd.getType(),pm.getClassLoaderResolver());
                    if (acmd.getIdentityType() == IdentityType.APPLICATION)
                    {
                        for (int i=0; i<acmd.getPrimaryKeyFieldNumbers().length; i++)
                        {
                            IdentityStrategy strategy = acmd.getManagedFieldAbsolute(acmd.getPrimaryKeyFieldNumbers()[i]).getValueStrategy();
                            if (strategy != null)
                            {
                                //if strategy is null, then it's user attributed value
                                isStrategyDatastoreAttributed |= pm.getStoreManager().isStrategyDatastoreAttributed(strategy);
                            }
                        }
                    }
                }

                boolean inserted = false;
                if (ownerFieldNumber >= 0)
                {
                    inserted = pm.isInserted(pc, ownerFieldNumber);
                }

                if (sm != null && sm.getAttachedPC() != null && ownerSM != null)
                {
                    // The value is really still detached but has a temporary StateManager whilst attaching so
                    // replace this field reference to be for the newly attached object
                    ownerSM.replaceField(ownerFieldNumber, sm.getAttachedPC());
                }

                // TODO Why exactly is there the final part of this "if" ? If something is datastore-attributed AND inserted then it is
                // ok to set the FK to that object since it is inserted!!! What other situation is there ? If there is one then
                // why not document it here?
                if (!pm.isInserting(pc) || inserted || (!isStrategyDatastoreAttributed && (this.fmd != null && this.fmd.isPrimaryKey())))
                {
                    // The PC is either already inserted, or inserted down to the level we need, or not inserted at all,
                    // or the field is a PK and identity not attributed by the datastore

                    // Object either already exists, or is not yet being inserted.
                    id = pc.jdoGetObjectId();

                    // Check if the PersistenceCapable exists in this datastore
                    boolean requiresPersisting = false;
                    if (pc.jdoIsDetached() && ownerSM != null)
                    {
                        // Detached object that needs attaching (or persisting if detached from a different datastore)
                        requiresPersisting = true;
                    }
                    else if (id == null)
                    {
                        // Transient object, so we need to persist it
                        requiresPersisting = true;
                    }
                    else
                    {
                        if (((PersistenceManager)pc.jdoGetPersistenceManager()) != null &&
                                pm.getPMHandle() != ((PersistenceManager)pc.jdoGetPersistenceManager()).getPMHandle())
                        {
                            throw new JDOUserException(LOCALISER.msg("Mapping.FieldNotAssignableIsFromAnotherPM"), id);
                        }
                    }

                    if (requiresPersisting)
                    {
                        // The object is either not yet persistent or is detached and so needs attaching
                        PersistenceCapable pcNew = (PersistenceCapable)pm.makePersistentInternal(value, null);
                        if(isStrategyDatastoreAttributed)
                        {
                            pm.flush();
                        }
                        id = pcNew.jdoGetObjectId();
                        if (pc.jdoIsDetached() && ownerSM != null)
                        {
                            // Update any detached reference to refer to the attached variant
                            ownerSM.replaceField(ownerFieldNumber, pcNew);
                            ClassLoaderResolver clr = pm.getClassLoaderResolver();
                            int relationType = fmd.getRelationType(clr);
                            if (relationType == Relation.MANY_TO_ONE_BI)
                            {
                                // TODO Update the container to refer to the attached object
                                if (JPOXLogger.JDO.isInfoEnabled())
                                {
                                    JPOXLogger.JDO.info("PCMapping.setObject : object " + ownerSM.getInternalObjectId() + 
                                        " has field " + ownerFieldNumber + " that is 1-N bidirectional." + 
                                        " Have just attached the N side so should really update the reference in the 1 side collection" +
                                        " to refer to this attached object. Not yet implemented");
                                }
                            }
                            else if (relationType == Relation.ONE_TO_ONE_BI)
                            {
                                AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
                                StateManager relatedSM = pm.findStateManager(pcNew);
                                relatedSM.replaceField(relatedFmd.getAbsoluteFieldNumber(), ownerSM.getObject());
                            }
                        }
                    }

                    if (sm != null)
                    {
                        sm.setStoringPC();
                    }

                    // If the field doesnt map to any datastore fields, omit the set process
                    if (getNumberOfDatastoreFields() > 0)
                    {
                        if (id instanceof OID)
                        {
                            super.setObject(pm, ps, param, id);
                        }
                        else
                        {
                            pc.jdoCopyKeyFieldsFromObjectId(new AppIDObjectIdFieldConsumer(param, pm, ps, javaTypeMappings), id);
                        }
                    }
                }
                else
                {
                    if (sm != null)
                    {
                        sm.setStoringPC();
                    }

                    if (getNumberOfDatastoreFields() > 0)
                    {
                        // Object is in the process of being inserted so we cant use its id currently and we need to store
                        // a foreign key to it (which we cant yet do). Just put "null" in and throw "NotYetFlushedException",
                        // to be caught by ParameterSetter and will signal to the PC object being inserted that it needs 
                        // to inform this object when it is inserted.
                        int n=0;
                        for (int i=0; i<javaTypeMappings.length; i++)
                        {
                            JavaTypeMapping mapping = javaTypeMappings[i];
                            if (mapping.getNumberOfDatastoreFields() > 0)
                            {
                                // Only update the object if it has any datastore mappings
                                int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
                                for (int j=0; j<posMapping.length; j++)
                                {
                                    posMapping[j] = param[n++];
                                }
                                mapping.setObject(pm, ps, posMapping, null);
                            }
                        }
                        throw new NotYetFlushedException(pc);
                    }
                }
            }
            finally
            {
                if (sm != null)
                {
                    sm.unsetStoringPC();
                }
            }
        }
    }

    /**
     * Returns a instance of a PersistenceCapable class.
     * Processes a FK field and converts the id stored firstly into an OID/AID
     * and then into the object that the FK id relates to.
     * @param pm The Persistence Manager
     * @param rs The ResultSet
     * @param param Array of parameter ids in the ResultSet to retrieve
     * @return The Persistence Capable object
     */
    public Object getObject(PersistenceManager pm, final Object rs, int[] param)
    {
        // Check for null FK
        try
        {
            //if the first param is null, then the field is null 
            if (((ResultSet) rs).getObject(param[0]) == null)
            {
                return null;
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(e.toString());
        }
       
        if (cmd == null)
        {
            cmd = pm.getMetaDataManager().getMetaDataForClass(getType(),pm.getClassLoaderResolver());
        }

        if (representsIdentity)
        {
            // Return the identity represented by this mapping
            if (cmd.getIdentityType() == IdentityType.DATASTORE)
            {
                OID oid = (OID)super.getObject(pm, rs, param);

                // Check that we have the correct inheritance level for the object
                // TODO Use any available discriminator to avoid this extra datastore call
                String className = pm.getStoreManager().getClassNameForObjectID(oid, pm.getClassLoaderResolver(), pm);
                if (!oid.getPcClass().equals(className))
                {
                    return OIDFactory.getInstance(className, oid.keyValue());
                }
                return oid;
            }
            else if (cmd.getIdentityType() == IdentityType.APPLICATION)
            {
                ClassLoaderResolver clr = pm.getClassLoaderResolver();
                Class objectIdClass = clr.classForName(cmd.getObjectidClass());
                Class pcClass = clr.classForName(cmd.getFullClassName());
                Object id;
                // TODO Will this give the object at the right inheritance level ? not tested!
                if (cmd.usesSingleFieldIdentityClass())
                {
                    id = createSingleFieldIdentity(pm, rs, param, cmd, objectIdClass, pcClass); 
                }
                else
                {
                    id = createObjectIdInstanceReflection(pm, rs, param, cmd, objectIdClass); 
                }
                return id;
            }
            else
            {
                // NONDURABLE has no identity
                return null;
            }
        }
        else
        {
            // Return the object represented by this mapping
            if (cmd.getIdentityType() == IdentityType.DATASTORE)
            {
                return getObjectForDatastoreIdentity(pm,rs,param,cmd);
            }
            else if (cmd.getIdentityType() == IdentityType.APPLICATION)
            {
                return getObjectForApplicationIdentity(pm,rs,param,cmd);
            }
            else
            {
                return null;
            }
        }
    }

    // ---------------------------- JDOQL Query Methods --------------------------------------

    public ScalarExpression newLiteral(QueryExpression qs, Object value)
    {
        ScalarExpression expr = new ObjectLiteral(qs, this, value,getType());
        return expr;        
    }

    public ScalarExpression newScalarExpression(QueryExpression qs, LogicSetExpression te)
    {
        if (getNumberOfDatastoreFields() > 0)
        {
            return new ObjectExpression(qs, this, te);
        }
        else
        {
            ClassLoaderResolver clr = qs.getClassLoaderResolver();
            int relationType = fmd.getRelationType(clr);
            if (relationType == Relation.ONE_TO_ONE_BI)
            {
                // Create an expression joining to the related field in the related table
                DatastoreClass targetTable = qs.getStoreManager().getDatastoreClass(fmd.getTypeName(), clr);
                JavaTypeMapping refMapping = targetTable.getFieldMapping(fmd.getRelatedFieldMetaData(clr));
                JavaTypeMapping selectMapping = targetTable.getIDMapping();
                DatastoreIdentifier targetTableIdentifier =
                    qs.getStoreManager().getIdentifierFactory().newIdentifier(IdentifierFactory.TABLE, "RELATED" + fmd.getAbsoluteFieldNumber());
                LogicSetExpression targetTe = qs.newTableExpression(targetTable, targetTableIdentifier);
                return new ObjectExpression(qs, this, te, refMapping, targetTe, selectMapping);
            }
            else if (relationType == Relation.MANY_TO_ONE_BI)
            {
                AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
                if (fmd.getJoinMetaData() != null || relatedFmd.getJoinMetaData() != null)
                {
                    // Join table relation
                    // Create an expression this table to the join table on the element id, selecting the owner id
                    DatastoreContainerObject targetTable = qs.getStoreManager().getDatastoreContainerObject(relatedFmd);
                    JavaTypeMapping refMapping = null;
                    JavaTypeMapping selectMapping = null;
                    ElementContainerTable elementTable = (ElementContainerTable)targetTable;
                    refMapping = elementTable.getElementMapping();
                    selectMapping = elementTable.getOwnerMapping();
                    DatastoreIdentifier targetTableIdentifier =
                        qs.getStoreManager().getIdentifierFactory().newIdentifier(IdentifierFactory.TABLE, "JOINTABLE" + fmd.getAbsoluteFieldNumber());
                    LogicSetExpression targetTe = qs.newTableExpression(targetTable, targetTableIdentifier);
                    return new ObjectExpression(qs, this, te, refMapping, targetTe, selectMapping);
                }
            }
        }

        // TODO Throw an exception since should be impossible
        return null;
    }

    /**
     * Method to handle deletion of any dependent fields (before deleting this object)
     * @param sm StateManager of the object
     */
    public void deleteDependent(StateManager sm)
    {
        // makes sure field is loaded
        int fieldNumber = fmd.getAbsoluteFieldNumber();
        sm.isLoaded(sm.getObject(), fieldNumber);

        // Check if the field has a FK defined
        ClassLoaderResolver clr = sm.getPersistenceManager().getClassLoaderResolver();
        AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
        boolean hasFK = false;
        if (fmd.getForeignKeyMetaData() != null)
        {
            hasFK = true;
        }
        if (relatedFmd != null && relatedFmd.getForeignKeyMetaData() != null)
        {
            hasFK = true;
        }
        if (sm.getPersistenceManager().getPMFContext().getPmfConfiguration().getDeletionPolicy().equals("JDO2"))
        {
            // JDO2 doesnt currently (2.0 spec) take note of foreign-key
            hasFK = false;
        }

        PersistenceCapable pc = (PersistenceCapable) sm.provideField(fieldNumber);
        if (pc != null)
        {
            if (!hasFK || fmd.getRelationType(clr) == Relation.MANY_TO_ONE_BI)
            {
                // No FK specified so null the related field where possible
                boolean isNullable = isNullable();
                if (getNumberOfDatastoreFields() == 0)
                {
                    // Other end of a bidir so get the end with the column(s)
                    if (relatedFmd != null)
                    {
                        DatastoreClass relatedTable = sm.getStoreManager().getDatastoreClass(relatedFmd.getClassName(), sm.getPersistenceManager().getClassLoaderResolver());
                        JavaTypeMapping relatedMapping = relatedTable.getFieldMapping(relatedFmd);
                        isNullable = relatedMapping.isNullable();
                    }
                }
                if (isNullable)
                {
                    // Null out the reference to this object since the reference column is nullable
                    sm.replaceField(fieldNumber, null);
                    sm.makeDirty(fieldNumber);
                }
            }

            // delete related object
            sm.getPersistenceManager().deletePersistent(pc);
        }
    }

    // ------------------------------------- Utility Methods ------------------------------------------

    /**
     * Get the object instance for a class using datastore identity
     * @param pm the PersistenceManager
     * @param rs the ResultSet
     * @param param the parameters
     * @param cmd the AbstractClassMetaData
     * @return the id
     */
    private Object getObjectForDatastoreIdentity(PersistenceManager pm, final Object rs, int[] param, AbstractClassMetaData cmd)
    {
        // Datastore Identity - retrieve the OID for the class.
        // Note that this is a temporary OID that is simply formed from the type of base class in the relationship
        // and the id stored in the FK. The real OID for the object may be of a different class.
        // For that reason we get the object by checking the inheritance (final param in getObjectById())
        Object oid = super.getObject(pm, rs, param);
        if (oid instanceof PersistenceCapable) //why check this?
        {
        	return oid;
        }
        return oid == null ? null : pm.getObjectById(oid, false, true, null);
    }

    /**
     * Create a SingleFieldIdentity instance
     * @param pm the PersistenceManager
     * @param rs the ResultSet
     * @param param the parameters
     * @param cmd the AbstractClassMetaData
     * @param objectIdClass the object id class
     * @param pcClass the PersistenceCapable class
     * @return the id
     */
    private Object createSingleFieldIdentity(PersistenceManager pm, final Object rs, int[] param, AbstractClassMetaData cmd, Class objectIdClass, Class pcClass)
    {
        // SingleFieldIdentity
        int paramNumber = param[0];
        try
        {
            Object idObj = ((ResultSet)rs).getObject(paramNumber);
            if (idObj == null)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("PersistenceCapableMapping.CantCreateSingleFieldIdentityWithNull"));
            }
            else
            {
                idObj = convertIfNecessary(objectIdClass, idObj);
            }
            return AIDUtils.getNewSingleFieldIdentity(objectIdClass, pcClass, idObj);
        }
        catch (Exception e)
        {
            JPOXLogger.JDO.error(LOCALISER.msg("PersistenceCapableMapping.CantCreateAID", cmd.getObjectidClass(), e));
            return null;
        }
    }

    private Object convertIfNecessary(Class objectIdClass, Object idObj)
    {
        // Check for any needed type conversion between the JDBC returned type and the object id type
        if (idObj instanceof BigInteger && LongIdentity.class.isAssignableFrom(objectIdClass))
        {
            idObj = new Long(((BigInteger)idObj).longValue());
        }
        else if (idObj instanceof BigInteger && IntIdentity.class.isAssignableFrom(objectIdClass))
        {
            idObj = new Long(((BigInteger)idObj).intValue());
        }
        else if (idObj instanceof BigDecimal && LongIdentity.class.isAssignableFrom(objectIdClass))
        {
            idObj = new Long(((BigDecimal)idObj).longValue());
        }
        else if (idObj instanceof BigDecimal && IntIdentity.class.isAssignableFrom(objectIdClass))
        {
            idObj = new Integer(((BigDecimal)idObj).intValue());
        }
        else if (idObj instanceof Long && IntIdentity.class.isAssignableFrom(objectIdClass))
        {
            idObj = new Integer(((Long)idObj).intValue());
        }
        else if (idObj instanceof Long && ShortIdentity.class.isAssignableFrom(objectIdClass))
        {
            idObj = new Short(((Long)idObj).shortValue());
        }
        else if (idObj instanceof Integer && ShortIdentity.class.isAssignableFrom(objectIdClass))
        {
            idObj = new Short(((Integer)idObj).shortValue());
        }
        return idObj;
    }

    /**
     * Create an object id instance and fill the fields using reflection
     * @param pm the PersistenceManager
     * @param rs the ResultSet
     * @param param the parameters
     * @param cmd the AbstractClassMetaData
     * @param objectIdClass the object id class
     * @return the id
     */
    private Object createObjectIdInstanceReflection(PersistenceManager pm, final Object rs, int[] param, AbstractClassMetaData cmd, Class objectIdClass)
    {
        // Users own AID
        Object fieldValue = null;
        try
        {
            // Create an AID
            Object id = objectIdClass.newInstance();

            // Set the fields of the AID
            int paramIndex = 0;
            for (int i=0; i<cmd.getPrimaryKeyFieldNumbers().length; ++i)
            {
                AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(cmd.getPrimaryKeyFieldNumbers()[i]);
                Field field = objectIdClass.getField(fmd.getName());
                
                JavaTypeMapping m = pm.getStoreManager().getDatastoreClass(cmd.getFullClassName(), pm.getClassLoaderResolver()).getFieldMapping(fmd);
                // NOTE This assumes that each field has one datastore column.
                for (int j = 0; j < m.getNumberOfDatastoreFields(); j++)
                {
                    Object obj = ((ResultSet)rs).getObject(param[paramIndex++]);
                    if ((obj instanceof BigDecimal))
                    {
                        BigDecimal bigDecimal = (BigDecimal) obj;
                        // Oracle 10g returns BigDecimal for NUMBER columns,
                        // resulting in IllegalArgumentException when reflective 
                        // setter is invoked for incompatible field type
                        // (see http://www.jpox.org/servlet/jira/browse/CORE-2624)
                        obj = convertIfNecessary(field.getType(), bigDecimal);
                        if (!bigDecimal.subtract(new BigDecimal("" + obj)).equals(new BigDecimal("0")))
                        {
                            throw new JDOFatalInternalException("Cannot convert retrieved BigInteger value to field of object id class!");
                        }
                    }
                    // field with multiple columns should have values returned from db merged here
                    fieldValue = obj;
                }
                field.set(id, fieldValue);
            }
            return id;
        }
        catch (Exception e)
        {
            JPOXLogger.JDO.error(LOCALISER.msg("PersistenceCapableMapping.CantCreateAIDWithReflection",
                cmd.getObjectidClass(), fmd == null ? null : fmd.getName(), fieldValue, e));
            return null;
        }
    }

    /**
     * Create an object id instance and fill the fields using reflection
     * @param pm the PersistenceManager
     * @param rs the ResultSet
     * @param param the parameters
     * @param cmd the AbstractClassMetaData
     * @return the id
     */
    private Object getObjectForAbstractClass(PersistenceManager pm, final Object rs, int[] param, AbstractClassMetaData cmd)
    {
        ClassLoaderResolver clr = pm.getClassLoaderResolver();

        // Abstract class, so we need to generate an AID before proceeding
        Class objectIdClass = clr.classForName(cmd.getObjectidClass());
        Class pcClass = clr.classForName(cmd.getFullClassName());
        Object id;
        if (cmd.usesSingleFieldIdentityClass())
        {
            id = createSingleFieldIdentity(pm, rs, param, cmd, objectIdClass, pcClass); 
        }
        else
        {
            id = createObjectIdInstanceReflection(pm, rs, param, cmd, objectIdClass); 
        }
        return pm.getObjectById(id, false, true, null);
    }

    /**
     * Get the object instance for a class using application identity
     * @param pm the PersistenceManager
     * @param rs the ResultSet
     * @param param the parameters
     * @param cmd the AbstractClassMetaData
     * @return the id
     */
    private Object getObjectForApplicationIdentity(PersistenceManager pm, final Object rs, int[] param, AbstractClassMetaData cmd)
    {
        ClassLoaderResolver clr = pm.getClassLoaderResolver();

        // Abstract class
        if (((ClassMetaData)cmd).isAbstractPersistenceCapable() && cmd.getObjectidClass() != null)
        {
            return getObjectForAbstractClass(pm,rs,param,cmd);
        }
        
        
        // Concrete class
        int totalFieldCount = cmd.getNoOfManagedFields() + cmd.getNoOfInheritedManagedFields();
        final StatementExpressionIndex[] statementExpressionIndex = new StatementExpressionIndex[totalFieldCount];
        int paramIndex = 0;
        
        DatastoreClass datastoreClass = pm.getStoreManager().getDatastoreClass(cmd.getFullClassName(), clr);
        final int[] pkFieldNumbers = cmd.getPrimaryKeyFieldNumbers();

        for (int i=0; i<pkFieldNumbers.length; ++i)
        {
            AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(pkFieldNumbers[i]);
            JavaTypeMapping m = datastoreClass.getFieldMapping(fmd);
            statementExpressionIndex[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
            statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setMapping(m);
            int expressionsIndex[] = new int[m.getNumberOfDatastoreFields()];
            for (int j = 0; j < expressionsIndex.length; j++)
            {
                expressionsIndex[j] = param[paramIndex++];
            }
            statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setExpressionIndex(expressionsIndex);
        }        
        
        return pm.getObjectByAID(clr.classForName(cmd.getFullClassName()),
            new FieldValues()
            {
            // StateManager calls the fetchFields method
            public void fetchFields(StateManager sm)
            {
                sm.replaceFields(pkFieldNumbers, new ResultSetGetter(sm, (ResultSet) rs, statementExpressionIndex));
            }
            public void fetchNonLoadedFields(StateManager sm)
            {
                sm.replaceNonLoadedFields(pkFieldNumbers, new ResultSetGetter(sm, (ResultSet) rs, statementExpressionIndex));
            }
            public FetchPlanImpl getFetchPlanForLoading()
            {
                return null;
            }
            }, false, true);
    }
}