/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.store.mapping;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.state.StateManagerFactory;

/**
 * Mapping for a serialised reference (Interface/Object) field.
 * Extends SerialisedMapping since that provides the basic serialisation mechanism,
 * adding on the addition of StateManagers to the serialised object whenever it is required
 * (since the object is a PersistenceCapable).
 *
 * @version $Revision: 1.7 $
 */
public class SerialisedReferenceMapping extends SerialisedMapping
{
    /**
     * Method to populate parameter positions in a PreparedStatement with this object
     * @param pm The Persistence Manager
     * @param preparedStatement The Prepared Statement
     * @param exprIndex The parameter positions to populate
     * @param value The value of the PC to use in populating the parameter positions
     */
    public void setObject(PersistenceManager pm, Object preparedStatement, int[] exprIndex, Object value)
    {
        if (fmd != null)
        {
            setObject(pm, preparedStatement, exprIndex, value, null, fmd.getAbsoluteFieldNumber());
        }
        else
        {
            super.setObject(pm, preparedStatement, exprIndex, value);
        }
    }

    /**
     * Method to populate parameter positions in a PreparedStatement with this object
     * @param pm The Persistence Manager
     * @param preparedStatement The Prepared Statement
     * @param exprIndex The parameter positions to populate
     * @param value The value of the PC to use in populating the parameter positions
     * @param ownerSM State Manager for the owning object
     * @param fieldNumber field number of this object in the owning object
     */
    public void setObject(PersistenceManager pm, Object preparedStatement, int[] exprIndex, Object value, 
            StateManager ownerSM, int fieldNumber)
    {
        if (value != null && value instanceof PersistenceCapable)
        {
            // Assign a StateManager to the serialised object if none present
            PersistenceCapable pc = (PersistenceCapable)value;
            StateManager embSM = pm.findStateManager(pc);
            if (embSM == null || pc.jdoGetPersistenceManager() == null)
            {
                embSM = StateManagerFactory.newStateManager(pc, pm, false);
                embSM.addEmbeddedOwner(ownerSM, fieldNumber);
            }
        }

        StateManager sm = null;
        if (value != null && value instanceof PersistenceCapable)
        {
            // Find SM for serialised PC object
            sm = pm.findStateManager((PersistenceCapable)value);
        }

        if (sm != null)
        {
            sm.setStoringPC();
        }
        getDataStoreMapping(0).setObject(preparedStatement, exprIndex[0], value);
        if (sm != null)
        {
            sm.unsetStoringPC();
        }
    }

    /**
     * Method to extract the value of the PersistenceCapable from a ResultSet.
     * @param pm The Persistence Manager
     * @param resultSet The ResultSet
     * @param exprIndex The parameter positions in the result set to use.
     * @return The (deserialised) PersistenceCapable object
     */
    public Object getObject(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        if (fmd != null)
        {
            return getObject(pm, resultSet, exprIndex, null, fmd.getAbsoluteFieldNumber());
        }
        else
        {
            return super.getObject(pm, resultSet, exprIndex);
        }
    }

    /**
     * Method to extract the value of the PersistenceCapable from a ResultSet.
     * @param pm The Persistence Manager
     * @param resultSet The ResultSet
     * @param exprIndex The parameter positions in the result set to use.
     * @param ownerSM The owning object
     * @param fieldNumber Absolute number of field in owner object
     * @return The (deserialised) PersistenceCapable object
     */
    public Object getObject(PersistenceManager pm, Object resultSet, int[] exprIndex, StateManager ownerSM, int fieldNumber)
    {
        Object obj = getDataStoreMapping(0).getObject(resultSet, exprIndex[0]);
        if (obj != null && obj instanceof PersistenceCapable)
        {
            // Assign a StateManager to the serialised object if none present
            PersistenceCapable pc = (PersistenceCapable)obj;
            StateManager embSM = pm.findStateManager(pc);
            if (embSM == null || pc.jdoGetPersistenceManager() == null)
            {
                embSM = StateManagerFactory.newStateManager(pc, pm, false);
                embSM.addEmbeddedOwner(ownerSM, fieldNumber); // Feed dirty flags to the owning object
            }
        }
        return obj;
    }
}