/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import javax.jdo.JDOFatalInternalException;

import org.jpox.util.JPOXLogger;
import org.jpox.util.JavaUtils;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Controller for execution of SQL statements to the underlying RDBMS datastore.
 * CURRENTLY THIS CLASS IS NOT UTILISED BY JPOX. IT IS PLANNED TO BE ENABLED IN THE FUTURE.
 * <p>
 * Separates statements into 2 categories.
 * <ul>
 * <li><b>Updates</b> : INSERT/UPDATE/DELETE statements that can potentially (depending on the JDBC driver capabilities) 
 * be batched and hence give much speed benefit.</li>
 * <li><b>Queries</b> : SELECT statements that can provide flexibility in the capabilities of ResultSet.</li>
 * </ul>
 * </p>
 * <p>
 * Maintains statement status for a Connection in a particular thread. This means that each Connection in each thread can be
 * controlled here providing batching etc individually.
 * </p>
 *
 * @version $Revision: 1.2 $
 */
public class SQLExecutionController
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    /** Whether batching is supported for this controller (datastore). */
    protected boolean supportsBatching = false;

    /** Maximum batch size (-1 implies no limit). */
    protected int maxBatchSize = -1;

    /** Timeout to apply to queries (where required). */
    protected int queryTimeout = 0;

    /**
     * State of a connection used by a thread.
     * Maintains the current (active) update and query statements (if any) and this can
     * then be used for providing batching etc.
     * 
     * @version $Revision: 1.2 $
     */
    class ConnectionStatementState
    {
        /** The current (active) PreparedStatement for updates. */
        PreparedStatement updateStmt = null;

        /** The text for the active update statement. */
        String updateStmtText = null;

        /** Number of statements currently batched in the update statement. */
        int updateStmtBatchSize = 0;

        /** The current (active) PreparedStatement for queries. */
        PreparedStatement queryStmt = null;

        /** The text for the active query statement. */
        String queryStmtText = null;
    }

    /** Map of the ConnectionStatementState keyed by the Connection+Thread */
    Map connectionStatements = new HashMap();

    /**
     * Constructor.
     * @param supportsBatching Whether batching is to be supported.
     * @param maxBatchSize The maximum batch size
     * @param queryTimeout Timeout for queries (ms)
     */
    public SQLExecutionController(boolean supportsBatching, int maxBatchSize, int queryTimeout)
    {
        this.supportsBatching = supportsBatching;
        this.maxBatchSize = maxBatchSize;
        this.queryTimeout = queryTimeout;
    }

    /**
     * Method to return a PreparedStatement for updating objects.
     * @param conn The Connection to use
     * @param stmtText The statement to execute
     * @param batchable Whether the statement is considered batchable
     * @return The PreparedStatement
     */
    public PreparedStatement getStatementForUpdate(Connection conn, String stmtText, boolean batchable)
    throws SQLException
    {
        ConnectionStatementState state = getConnectionStatementState(conn);
        if (state == null)
        {
            state = new ConnectionStatementState();
            setConnectionStatementState(conn, state);
        }

        // TODO What if we already have an updateStmt set ? Can we have 2 at once and execute out of order ?
        if (supportsBatching && batchable && 
            state.updateStmtText != null && state.updateStmtText.equals(stmtText) && maxBatchSize > state.updateStmtBatchSize)
        {
            // Same as the current update statement and we are allowed to batch this so batch it
            state.updateStmt.addBatch();
            state.updateStmtBatchSize++;
        }
        else
        {
            // Not batchable so we need to process the existing statement and provide a new one
            if (state.updateStmt != null)
            {
                processStatementForUpdate(conn);
            }

            state.updateStmt = conn.prepareStatement(stmtText);
            state.updateStmtText = stmtText;

            if (JPOXLogger.RDBMS_SQL.isDebugEnabled() && JavaUtils.isJRE1_4OrAbove())
            {
                // Wrap the statement with our JDK1.4 parameter wrapper
                state.updateStmt = new ParamLoggingPreparedStatement(state.updateStmt, stmtText);
            }
        }

        return state.updateStmt;
    }

    /**
     * Method to return a PreparedStatement for querying objects.
     * @param conn The Connection to use
     * @param stmtText The statement to execute
     * @param applyTimeout Whether this query should apply any timeout (where specified)
     * @param resultSetType Type of result set
     * @param resultSetConcurrency Concurrency for the result set
     * @return The PreparedStatement
     */
    public PreparedStatement getStatementForQuery(Connection conn, String stmtText, boolean applyTimeout, String resultSetType, String resultSetConcurrency)
    throws SQLException
    {
        ConnectionStatementState state = getConnectionStatementState(conn);
        if (state == null)
        {
            state = new ConnectionStatementState();
            setConnectionStatementState(conn, state);
        }
        // TODO What if we already have a queryStmt set ?
        // TODO What if we already have an updateStmt set ? Can we have 2 at once and execute out of order ?

        if (resultSetType != null || resultSetConcurrency != null)
        {
            int rsTypeValue = ResultSet.TYPE_FORWARD_ONLY;
            if (resultSetType.equals("scroll-sensitive"))
            {
                rsTypeValue = ResultSet.TYPE_SCROLL_SENSITIVE;
            }
            else if (resultSetType.equals("scroll-insensitive"))
            {
                rsTypeValue = ResultSet.TYPE_SCROLL_INSENSITIVE;
            }
            
            int rsConcurrencyValue = ResultSet.CONCUR_READ_ONLY;
            if (resultSetConcurrency.equals("updateable"))
            {
                rsConcurrencyValue = ResultSet.CONCUR_UPDATABLE;
            }

            state.queryStmt = conn.prepareStatement(stmtText, rsTypeValue, rsConcurrencyValue);
        }
        else
        {
            state.queryStmt = conn.prepareStatement(stmtText);
        }
        state.queryStmtText = stmtText;

        if (applyTimeout)
        {
            if (queryTimeout > 0)
            {
                // Apply any query timeout
                state.queryStmt.setQueryTimeout(queryTimeout);
            }
        }
        
        if (JPOXLogger.RDBMS_SQL.isDebugEnabled() && JavaUtils.isJRE1_4OrAbove())
        {
            state.queryStmt = new ParamLoggingPreparedStatement(state.queryStmt, stmtText);
        }

        return state.queryStmt;
    }

    /**
     * Method to process the current update statement.
     * @return The return from the statement
     */
    public int[] processStatementForUpdate(Connection conn)
    throws SQLException
    {
        ConnectionStatementState state = getConnectionStatementState(conn);
        if (state == null)
        {
            throw new JDOFatalInternalException("No Connection state exists for Connection " + 
                StringUtils.toJVMIDString(conn) + " and thread " + Thread.currentThread().getName());
        }

        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            if (state.updateStmt instanceof ParamLoggingPreparedStatement)
            {
                JPOXLogger.RDBMS_SQL.debug(((ParamLoggingPreparedStatement)state.updateStmt).getStatementWithParamsReplaced());
            }
            else
            {
                JPOXLogger.RDBMS_SQL.debug(state.updateStmtText);
            }
        }

        long startTime = System.currentTimeMillis();
        int[] inds = null;
        if (state.updateStmtBatchSize > 0)
        {
            state.updateStmt.addBatch(); // Most recent statement needs adding to the batch before processing

            inds = state.updateStmt.executeBatch();
        }
        else
        {
            inds = new int[1];
            inds[0] = state.updateStmt.executeUpdate();
        }

        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
        }
        
        // TODO Close the statement (or make it optional on the method)

        state.updateStmt = null;
        state.updateStmtText = null;
        state.updateStmtBatchSize = 0;

        return inds;
    }

    /**
     * Method to process the current query statement.
     * @return The return from the statement
     */
    public ResultSet processStatementForQuery(Connection conn)
    throws SQLException
    {
        ConnectionStatementState state = getConnectionStatementState(conn);
        if (state == null)
        {
            throw new JDOFatalInternalException("No Connection state exists for Connection " + 
                StringUtils.toJVMIDString(conn) + " and thread " + Thread.currentThread().getName());
        }

        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            if (state.queryStmt instanceof ParamLoggingPreparedStatement)
            {
                JPOXLogger.RDBMS_SQL.debug(((ParamLoggingPreparedStatement)state.queryStmt).getStatementWithParamsReplaced());
            }
            else
            {
                JPOXLogger.RDBMS_SQL.debug(state.queryStmtText);
            }
        }

        long startTime = System.currentTimeMillis();
        ResultSet rs = state.queryStmt.executeQuery();

        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime", (System.currentTimeMillis() - startTime)));
        }

        state.queryStmt = null;
        state.queryStmtText = null;

        return rs;
    }

    /**
     * Convenience method to remove the state for this connection/thread.
     * This is typically called when a Connection is closed.
     * @param conn The Connection
     */
    protected void removeConnectionStatementState(Connection conn)
    {
        connectionStatements.remove(getStateKey(conn));
    }

    /**
     * Convenience method to get the state for this connection/thread.
     * @param conn The Connection
     * @return The state (if any)
     */
    protected ConnectionStatementState getConnectionStatementState(Connection conn)
    {
        return (ConnectionStatementState)connectionStatements.get(getStateKey(conn));
    }

    /**
     * Convenience method to set the state for this connection/thread.
     * @param conn The Connection
     * @param state The state
     */
    protected void setConnectionStatementState(Connection conn, ConnectionStatementState state)
    {
        connectionStatements.put(getStateKey(conn), state);
    }

    /**
     * Method to set the key for a ConnectionStatementState (concatenation of the connection object and the thread).
     * @param conn The Connection
     * @return The key
     */
    protected String getStateKey(Connection conn)
    {
        return StringUtils.toJVMIDString(conn) + "_" + Thread.currentThread().getName();
    }
}