/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.sql.DatabaseMetaData;
import java.sql.ResultSet;

import org.jpox.store.rdbms.typeinfo.InformixTypeInfo;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Provides methods for adapting SQL language elements to the Informix
 * database. Overrides some methods in DatabaseAdapter where Informix behaviour
 * differs.
 * 
 * Informix databases must be created WITH LOG MODE ANSI, otherwise
 * errors like "Transaction Not Supported", "Not in transaction" will appear.
 * See the informix info
 *
 * @version $Revision: 1.8 $ 
 */
class InformixAdapter extends DatabaseAdapter
{
    /**
     * Constructor.
     * @param metadata MetaData for the DB
     **/
    public InformixAdapter(DatabaseMetaData metadata)
    {
        super(metadata);
    }

    public String getVendorID()
    {
        return "informix";
    }

    /**
     * Factory for TypeInfo objects.
     * @param rs The ResultSet from DatabaseMetaData.getTypeInfo(). 
     * @return A TypeInfo object.
     **/
    public TypeInfo newTypeInfo(ResultSet rs)
    {
        return new InformixTypeInfo(rs);
    }

    /**
     * Accessor for whether to use order by column number in selects.
     * @return Whether to use order by column number
     */
    public boolean includeOrderByColumnsInSelect()
    {
        // Informix uses this
        return true;
    }

    /**
     * Accessor for an identifier quote string.
     * @return Identifier quote string.
     **/
    public String getIdentifierQuoteString()
    {
    	/*
    	 * the documentation says a double quote, but it seems to not work
    	 */
        return "";
    }   

    // ---------------------------- AutoIncrement Support ---------------------------

    /**
     * Whether we support autoincrementing fields.
     * @return whether we support autoincrementing fields.
     **/
    public boolean supportsAutoIncrementFields()
    {
        return true; // SERIAL columns
    }

    /**
     * Accessor for the autoincrement sql access statement for this datastore.
     * @param tableName Name of the table that the autoincrement is for
     * @param columnName Name of the column that the autoincrement is for
     * @return The statement for getting the latest auto-increment key
     **/
    public String getAutoIncrementStmt(String tableName, String columnName)
    {
        // Which one of these works ? Not got Informix so cant check
        return "SELECT first 1 dbinfo('sqlca.sqlerrd1') from systables";
        // return "SELECT unique dbinfo('sqlca.sqlerrd1') s from " + JDBCUtils.getIdentifierNameStripped(tableName, this);
    }

    /**
     * Accessor for the auto-increment keyword for generating DDLs (CREATE TABLEs...).
     * @return The keyword for a column using auto-increment
     **/
    public String getAutoIncrementKeyword()
    {
        return "SERIAL";
    }

    /**
     * Whether we support auto-increment/identity keys with nullability specification.
     * @return whether we support auto-increment keys with nullability spec.
     **/
    public boolean supportsAutoIncrementKeysNullSpecification()
    {
        return false;
    }

    /**
     * Whether we support auto-increment/identity keys with column type specification.
     * @return whether we support auto-increment keys with column type spec.
     **/
    public boolean supportsAutoIncrementColumnTypeSpecification()
    {
        return false;
    }
}