/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.jdo.JDODataStoreException;

import org.jpox.ClassLoaderResolver;
import org.jpox.plugin.ConfigurationElement;
import org.jpox.plugin.Extension;
import org.jpox.plugin.PluginManager;
import org.jpox.store.DatastoreAdapter;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Factory of RDBMS datastore adapters for JPOX. 
 * Acts as a registry of adapters to RDBMS that can be selected for use in JPOX. 
 * Uses a singleton pattern, and the factory can be accessed using the getInstance() method.
 * <H3>Autodetection</H3>
 * JPOX provides a level of autodetection of the adapter to use for a particular
 * RDBMS. It uses the DatabaseMetaData to extract the "product name" and matches
 * this against a series of "adapter aliases" that JPOX has defined internally
 * that it will find as part of the JDBC "product name". These "adapter aliases"
 * can be primary or secondary. The primary is the main alias to match on. Typically
 * if the primary is matched then that will define the DatabaseAdapter to use.
 * In some situations (e.g DB2 AS/400) a primary alias matches ("db2"), yet
 * we need to use an alternative adapter so we have a secondary alias ("as/400").
 * If a secondary alias matches then that takes priority in the autodetection process.
 * <H3>Specification of adapter class</H3>
 * You can call getDatastoreAdapter() passing the adapter class name directly if you know
 * which you want to use. This allows for user-defined database adapters.
 *
 * @version $Revision: 1.11 $ 
 */
public class RDBMSAdapterFactory
{
    /** Localisation of messages */
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    /** The adapter classes, mapped by naming string. */
    HashMap adapters = new HashMap();

    /** The secondary level aliases. */
    HashSet secondaryAliases = new HashSet();

    /** Singleton instance */
    static RDBMSAdapterFactory adapterMgr = new RDBMSAdapterFactory();
    
    /**
     * Accessor for the RDBMSAdapterFactory for JPOX (singleton)
     * @return The manager of type information
     */
    public static RDBMSAdapterFactory getInstance()
    {
        return adapterMgr;
    }

    /**
     * Protected constructor to prevent outside instantiation
     */
    protected RDBMSAdapterFactory()
    {
    }

    /**
     * Method to add support for a RDBMS product via an adapter.
     * @param alias Naming string found as part of the "product name" of the JDBC DatabaseMetaData
     * @param adapterClassName Name of the adapter class to use
     * @param secondary Whether this naming alias is secondary level
     */
    public void addAdapter(String alias, String adapterClassName, boolean secondary)
    {
        if (alias == null || adapterClassName == null)
        {
            return;
        }

        adapters.put(alias, adapterClassName);
        if (secondary)
        {
            secondaryAliases.add(alias);
        }
    }

    /**
     * Accessor for an adapter, given a Connection to the datastore.
     * @param clr ClassLoader resolver for resolving the adapter class
     * @param conn The Connection
     * @param adapterClassName Name of the class of the database adapter to use
     * @param pluginMgr the Plug-in manager
     * @return The database adapter for this connection.
     * @throws SQLException Thrown if a DB error occurs.
     **/
    public DatastoreAdapter getDatastoreAdapter(ClassLoaderResolver clr, Connection conn, String adapterClassName, PluginManager pluginMgr)
    throws SQLException
    {
        loadDatastoreAdapterExtension(pluginMgr);
        DatastoreAdapter adapter = null;
        DatabaseMetaData metadata = conn.getMetaData();

        // Get a new adapter
        adapter = getNewDatastoreAdapter(clr, metadata, adapterClassName);
        if (adapter == null)
        {
            // Nothing suitable found so warn the user and continue with the generic adapter
            JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMS.Adapter.NoAdapterFound"));
            adapter = new DatabaseAdapter(metadata);
        }

        return adapter;
    }

    /**
     * Method to "autodetect" the adapter to be used, based on the supplied
     * "product name". Adaopts a simple procedure of finding one of our registered
     * adapters that is part of the product name.
     * @param productName The product name
     * @return The class name of the adapter to use
     */
    protected String getAdapterClassNameForProductName(String productName)
    {
        String adapterClassName = null;

        Set adapter_keys = adapters.keySet();
        if (adapter_keys != null)
        {
            Iterator keys_iter = adapter_keys.iterator();
            while (keys_iter.hasNext())
            {
                String alias = (String)keys_iter.next();
                if (productName.toLowerCase().indexOf(alias.toLowerCase()) >= 0)
                {
                    if (adapterClassName == null)
                    {
                        adapterClassName = (String)adapters.get(alias);
                        if (secondaryAliases.contains(alias))
                        {
                            // This is a secondary alias so proceed
                            break;
                        }
                    }
                    else
                    {
                        if (secondaryAliases.contains(alias))
                        {
                            // This is a secondary alias so takes priority
                            adapterClassName = (String)adapters.get(alias);
                            break;
                        }
                    }
                }
            }
        }
        return adapterClassName;
    }

    /**
     * Accessor for the adapter for a specified datastore product.
     * @param clr ClassLoader resolver for resolving the adapter class
     * @param metadata Database MetaData for the RDBMS
     * @param adapterClassName Name of the class of the database adapter (null implies use autodetect)
     * @return Instance of the database adapter
     */
    protected DatastoreAdapter getNewDatastoreAdapter(ClassLoaderResolver clr,
                                                      DatabaseMetaData metadata,
                                                      String adapterClassName)
    {
        if (metadata == null)
        {
            return null;
        }

        if (adapterClassName == null)
        {
            // No adapter specified, so use "autodetection" based on the metadata to find the most suitable
            String productName = null;
            try
            {
                productName = metadata.getDatabaseProductName();
                if (productName == null)
                {
                    JPOXLogger.RDBMS.error(LOCALISER.msg("RDBMS.Adapter.JDBC.NoProductNameError"));
                    return null;
                }
            }
            catch (SQLException sqe)
            {
                JPOXLogger.RDBMS.error(LOCALISER.msg("RDBMS.Adapter.JDBC.ProductNameAccessError",sqe));
                return null;
            }

            adapterClassName = getAdapterClassNameForProductName(productName);
        }

        // Instantiate the adapter class
        if (adapterClassName != null)
        {
            final Object adapter_obj;
            final Object[] ctr_args = new Object[] {metadata};
            final Class[] ctr_args_classes = new Class[] {DatabaseMetaData.class};
            try
            {
                Class adapterClass;
                try
                {
                    // try first with this classloader, since we are dealing with jpox internal classes and not user classes
                    adapterClass = Class.forName(adapterClassName, true, RDBMSAdapterFactory.class.getClassLoader());
                }
                catch (ClassNotFoundException e1)
                {
                    // try with jdo classloader
                    adapterClass = clr.classForName(adapterClassName, true);
                }

                // Create an instance of the datastore adapter
                final Constructor ctr = adapterClass.getConstructor(ctr_args_classes);
                try
                {
                    ctr.setAccessible(true);
                    adapter_obj = ctr.newInstance(ctr_args);
                }
                catch (InvocationTargetException ite)
                {
                    if (ite.getTargetException()!= null && ite.getTargetException() instanceof JDODataStoreException)
                    {
                        throw (JDODataStoreException) ite.getTargetException();
                    }
                    return null;
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                    JPOXLogger.RDBMS.error(LOCALISER.msg("RDBMS.Adapter.InstantiationError", adapterClassName, e));
                    return null;
                }
            }
            catch (NoSuchMethodException nsme)
            {
                return null;
            }

            return (DatastoreAdapter)adapter_obj;
        }

        return null;
    }
    
    /**
     * Load "org.jpox.store.datastoreadapter" extensions 
     * @param pluginMgr the Plug-in Manager
     */
    private void loadDatastoreAdapterExtension(PluginManager pluginMgr)
    {
        if (adapters.size() > 0)
        {
            //already loaded
            return;
        }
        Extension[] ex = pluginMgr.getExtensionPoint("org.jpox.store_datastoreadapter").getExtensions();
        for (int i=0; i<ex.length; i++)
        {
            ConfigurationElement[] confElm = ex[i].getConfigurationElements();
            for( int c=0; c<confElm.length; c++)
            {
                adapterMgr.addAdapter(confElm[c].getAttribute("vendor-id"),
                                      confElm[c].getAttribute("class-name"),
                                      Boolean.valueOf(confElm[c].getAttribute("secondary")).booleanValue());
            }
        }
    }
}