/**********************************************************************
Copyright (c) 2005 Andrew Hoffman and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.columninfo;

import java.sql.ResultSet;
import java.sql.Types;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Represents the metadata of a specific table column in Oracle.
 * 
 * @version $Revision: 1.3 $
 */
public class OracleColumnInfo extends ColumnInfo
{
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    /**
     * Constructs a column information object from the current row of the
     * given result set. The {@link ResultSet} object passed must
     * have been obtained from a call to DatabaseMetaData.getColumns().
     * <p>
     * This method only retrieves the values from the current row; the
     * caller is required to advance to the next row with {@link ResultSet#next}.
     * @param rs The result set returned from DatabaseMetaData.getColumns().   
     */
    public OracleColumnInfo(ResultSet rs)
    {
        super(rs);

        // set dataType value based on typeName column
        this.dataType = new Integer(OracleColumnInfo.getColumnType(this.typeName)).shortValue();
    }

    /**
     * Internal method used to translate String typeName into java.sql.Types identifier.
     * @param oracleDataType String value for Oracle data type, as given in 
     * DATA_TYPE column of ALL_TAB_COLUMNS dictionary view
     * @return java.sql.Types identifier appropriate to provided oracleDataType
     */
    private static int getColumnType(String oracleDataType)
    {
        if (oracleDataType == null)
        {
            return Types.NULL;
        }
        else if (oracleDataType.equals("ROWID"))
        {
            return Types.INTEGER;
        }
        else if (oracleDataType.equals("NUMBER") || oracleDataType.equals("VARNUM"))
        {
            return Types.NUMERIC;
        }
        else if (oracleDataType.equals("VARCHAR2"))
        {
            return Types.VARCHAR;
        }
        else if (oracleDataType.equals("CHAR"))
        {
            return Types.CHAR;
        }
        else if (oracleDataType.equals("DATE"))
        {
            // should this be TIMESTAMP instead of DATE ??
            return Types.DATE;
        }
        else if (oracleDataType.equals("CLOB") || oracleDataType.equals("NCLOB"))
        {
            return Types.CLOB;
        }
        else if (oracleDataType.equals("BLOB"))
        {
            return Types.BLOB; 
        }
        else if (oracleDataType.equals("LONG"))
        {
            return Types.LONGVARCHAR;
        }
        else if (oracleDataType.equals("LONG RAW"))
        {
            return Types.LONGVARBINARY;
        }
        else if (oracleDataType.equals("RAW"))
        {
            return Types.VARBINARY;
        }
        else if (oracleDataType.startsWith("TIMESTAMP"))
        {
            return Types.TIMESTAMP;
        }
        else if (oracleDataType.equals("FLOAT"))
        {
            return Types.FLOAT;
        }
        else
        {
            JPOXLogger.RDBMS.warn(LOCALISER.msg("TypeInfo.UnrecognizedDatatype", oracleDataType));
            return Types.OTHER;
        }
    }
}