/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.fieldmanager;

import java.sql.PreparedStatement;

import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.NullValue;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.exceptions.NotYetFlushedException;
import org.jpox.store.fieldmanager.AbstractFieldManager;
import org.jpox.store.mapping.EmbeddedPCMapping;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.mapping.SerialisedReferenceMapping;
import org.jpox.store.mapping.SerialisedPCMapping;
import org.jpox.util.Localiser;

/**
 * Parameter setter implementation of a field manager.
 * 
 * @version $Revision: 1.12 $
 **/
public class ParameterSetter extends AbstractFieldManager
{
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.rdbms.Localisation"); 

    private final StateManager sm;
    private final PersistenceManager pm;
    private final PreparedStatement ps;
    private final StatementExpressionIndex[] statementExpressionIndex;
    private final boolean checkNonNullable;

    /**
     * Constructor.
     * @param sm The state manager for the object.
     * @param ps The Prepared Statement to set values on.
     * @param statementExpressionIndex the index of parameters/mappings
     * @param checkNonNullable Whether to check for nullability
     **/
    public ParameterSetter(StateManager sm,
                           PreparedStatement ps,
                           StatementExpressionIndex[] statementExpressionIndex,
                           boolean checkNonNullable)
    {
        this.sm = sm;
        this.pm = sm.getPersistenceManager();
        this.ps = ps;
        this.statementExpressionIndex = statementExpressionIndex;
        this.checkNonNullable = checkNonNullable;
    }

    public void storeBooleanField(int fieldNumber, boolean value)
    {
        statementExpressionIndex[fieldNumber].getMapping().setBoolean(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
    }

    public void storeCharField(int fieldNumber, char value)
    {
        statementExpressionIndex[fieldNumber].getMapping().setChar(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
    }

    public void storeByteField(int fieldNumber, byte value)
    {
        statementExpressionIndex[fieldNumber].getMapping().setByte(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
    }

    public void storeShortField(int fieldNumber, short value)
    {
        statementExpressionIndex[fieldNumber].getMapping().setShort(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
    }

    public void storeIntField(int fieldNumber, int value)
    {
        statementExpressionIndex[fieldNumber].getMapping().setInt(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
    }

    public void storeLongField(int fieldNumber, long value)
    {
        statementExpressionIndex[fieldNumber].getMapping().setLong(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
    }

    public void storeFloatField(int fieldNumber, float value)
    {
        statementExpressionIndex[fieldNumber].getMapping().setFloat(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
    }

    public void storeDoubleField(int fieldNumber, double value)
    {
        statementExpressionIndex[fieldNumber].getMapping().setDouble(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
    }

    public void storeStringField(int fieldNumber, String value)
    {
        if (checkNonNullable &&
            value == null &&
            statementExpressionIndex[fieldNumber].getMapping().getFieldMetaData().getNullValue() == NullValue.EXCEPTION)
        {
            throw new JDOUserException(LOCALISER.msg("ParameterSetter.FieldCannotBeNull",statementExpressionIndex[fieldNumber].getMapping().getFieldMetaData().getFullFieldName()));
        }
        statementExpressionIndex[fieldNumber].getMapping().setString(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
    }

    public void storeObjectField(int fieldNumber, Object value)
    {
        if (checkNonNullable &&
            value == null &&
            statementExpressionIndex[fieldNumber].getMapping().getFieldMetaData().getNullValue() == NullValue.EXCEPTION)
        {
            throw new JDOUserException(LOCALISER.msg("ParameterSetter.FieldCannotBeNull",statementExpressionIndex[fieldNumber].getMapping().getFieldMetaData().getFullFieldName()));
        }

        try
        {
            JavaTypeMapping mapping = statementExpressionIndex[fieldNumber].getMapping();
            if (mapping instanceof EmbeddedPCMapping ||
                mapping instanceof SerialisedPCMapping ||
                mapping instanceof SerialisedReferenceMapping ||
                mapping instanceof PersistenceCapableMapping ||
                mapping instanceof InterfaceMapping)
            {
                // Pass in the owner StateManager/field for any mappings that need to perform management of the relation
                mapping.setObject(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value, sm, fieldNumber);
            }
            else
            {
                mapping.setObject(pm, ps, statementExpressionIndex[fieldNumber].getParameterIndex(), value);
            }
	        sm.replaceSCOField(fieldNumber, value);
        }
        catch (NotYetFlushedException e)
        {
            sm.updateFieldAfterInsert(e.getPersistenceCapable(),fieldNumber);
        }
    }
}
