/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2004 Andy Jefferson - added getTypeInfo method
2005 Andy Jefferson - changed to use StoreManager instead of DatabaseAdapter
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import javax.jdo.JDOFatalInternalException;

import org.jpox.store.StoreManager;
import org.jpox.store.mapping.DatastoreMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.adapter.RDBMSAdapter;
import org.jpox.store.rdbms.typeinfo.TypeInfo;
import org.jpox.util.Localiser;

/**
 * Implementation of the mapping of an RDBMS type.
 *
 * @version $Revision: 1.12 $
 **/
public abstract class RDBMSMapping implements DatastoreMapping
{
    protected static final Localiser LOCALISER_STORE = Localiser.getInstance("org.jpox.store.Localisation");
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    /** Store Manager to use for mapping. */
    protected final StoreManager storeMgr;

    /** Mapping of the Java type. */
    protected final JavaTypeMapping mapping;

    /**
     * Create a new Mapping with the given DatabaseAdapter for the given type.
     * 
     * @param storeMgr The Store Manager that this Mapping should use.
     * @param mapping The Class that this mapping maps to the database.
     */
    protected RDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        this.storeMgr = storeMgr;
        this.mapping = mapping;
        mapping.addDataStoreMapping(this);
    }

    /**
     * Convenience to access the Datastore adapter as a DatabaseAdapter.
     * @return The adapter in use
     */
    protected RDBMSAdapter getDatabaseAdapter()
    {
        return (RDBMSAdapter)storeMgr.getDatastoreAdapter();
    }

    /**
     * Accessor for the java type mapping
     * @return The java type mapping used
     */
    public JavaTypeMapping getJavaTypeMapping()
    {
        return mapping;
    }

    /**
     * Convenience method to apply the valid values to the Column.
     * Adds a constraint "CHECK(col IN (x,y,z)).
     * @param column The column
     * @param validValues valid values
     */
    protected static void applyValidValuesToColumn(Column column, Object[] validValues)
    {
        StringBuffer constraints = new StringBuffer("CHECK (");
        constraints.append(column.getIdentifier());
        constraints.append(" IN (");
        for (int i=0;i<validValues.length;i++)
        {
            if (i > 0)
            {
                constraints.append(",");
            }
            constraints.append(validValues[i]);
        }
        constraints.append(")");
        if (column.isNullable())
        {
            constraints.append(" OR " + column.getIdentifier() + " IS NULL");
        }
        constraints.append(")");
        column.setConstraints(constraints.toString());
    }

    /**
     * Accessor for the type info for this datastore type.
     * TODO Move this to DatastoreMapping when we have a generic
     * "TypeInfo" that could apply to other datastores.
     * @return The type info
     */
    public abstract TypeInfo getTypeInfo();

    /**
     * Accessor for whether the mapping is nullable.
     * @return Whether it is nullable
     */
    public abstract boolean isNullable();

    /**
     * Whether this mapping is included in the fetch statement.
     * @return Whether to include in fetch statement
     */
    public abstract boolean includeInFetchStatement();

    /**
     * Accessor for the string to put in any retrieval datastore statement for this field.
     * In RDBMS, this is typically a ? to be used in JDBC statements.
     * @return The input parameter
     */
    public abstract String getInsertionInputParameter();

    /**
     * Accessor for the string to put in any update datastore statements for this field.
     * In RDBMS, this is typically a ? to be used in JDBC statements.
     * @return The input parameter.
     */
    public abstract String getUpdateInputParameter();
 
    /**
     * Utility to output any error message.
     * @param method The method that failed.
     * @return The localised failure message
     **/
    protected String failureMessage(String method)
    {
        return LOCALISER_STORE.msg("Mapping.FailureMessage",getClass().getName(),method);
    }

    // ------------------- Accessors & Mutators for JDBC -----------------------
 
    public void setBoolean(Object preparedStatement, int exprIndex, boolean value)
    {
        throw new JDOFatalInternalException(failureMessage("setBoolean"));
    }

    public boolean getBoolean(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getBoolean"));
    }

    public void setChar(Object preparedStatement, int exprIndex, char value)
    {
        throw new JDOFatalInternalException(failureMessage("setChar"));
    }

    public char getChar(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getChar"));
    }

    public void setByte(Object preparedStatement, int exprIndex, byte value)
    {
        throw new JDOFatalInternalException(failureMessage("setByte"));
    }

    public byte getByte(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getByte"));
    }

    public void setShort(Object preparedStatement, int exprIndex, short value)
    {
        throw new JDOFatalInternalException(failureMessage("setShort"));
    }

    public short getShort(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getShort"));
    }

    public void setInt(Object preparedStatement, int exprIndex, int value)
    {
        throw new JDOFatalInternalException(failureMessage("setInt"));
    }

    public int getInt(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getInt"));
    }

    public void setLong(Object preparedStatement, int exprIndex, long value)
    {
        throw new JDOFatalInternalException(failureMessage("setLong"));
    }

    public long getLong(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getLong"));
    }

    public void setFloat(Object preparedStatement, int exprIndex, float value)
    {
        throw new JDOFatalInternalException(failureMessage("setFloat"));
    }

    public float getFloat(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getFloat"));
    }

    public void setDouble(Object preparedStatement, int exprIndex, double value)
    {
        throw new JDOFatalInternalException(failureMessage("setDouble"));
    }

    public double getDouble(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getDouble"));
    }

    public void setString(Object preparedStatement, int exprIndex, String value)
    {
        throw new JDOFatalInternalException(failureMessage("setString"+mapping.getFieldMetaData().getFullFieldName()));
    }

    public String getString(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getString"));
    }

    public void setObject(Object preparedStatement, int exprIndex, Object value)
    {
        throw new JDOFatalInternalException(failureMessage("setObject"));
    }

    public Object getObject(Object resultSet, int exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getObject"));
    }
}