/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2004 Andy Jefferson - localised messages
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import javax.jdo.JDODataStoreException;

import org.jpox.store.DatastoreField;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NullValueException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Mapping of a REAL RDBMS type.
 * 
 * @version $Revision: 1.8 $
 */
public class RealRDBMSMapping extends ColumnMapping
{
    /**
     * @param storeMgr Store Manager
     * @param mapping java type mapping
     */
    protected RealRDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }

    /**
     * Constructor.
     * @param mapping Java type mapping
     * @param storeMgr Store Manager
     * @param field Field to be mapped
     */
    public RealRDBMSMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField field)
    {
		super(storeMgr, mapping);
		column = (Column) field;
		initialize();
	}
    
    private void initialize()
    {
        if (column != null)
        {
            column.checkPrimitive();
        }
		initTypeInfo();
    }
    
    public TypeInfo getTypeInfo()
    {
        return getDatabaseAdapter().getTypeInfo(Types.REAL);
    }

    public void setFloat(Object ps, int param, float value)
    {
        try
        {
            ((PreparedStatement) ps).setFloat(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","float","" + value, column, e.getMessage()), e);
        }
    }

    public float getFloat(Object rs, int param)
    {
        float value;

        try
        {
            value = ((ResultSet) rs).getFloat(param);
            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMS.Mapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {           
            try
            {
                //when value is real in database, cause cause a parse error when calling getFloat
                //JDBC error:Value can not be converted to requested type.

                value = Float.parseFloat(((ResultSet) rs).getString(param));
                if (((ResultSet) rs).wasNull())
                {
                    throw new NullValueException(LOCALISER.msg("RDBMS.Mapping.IllegalNullValueInColumn",column));
                }
            }
            catch (SQLException e1)
            {
                try
                {
                    throw new JDODataStoreException("Can't get float result: param = " + param + " - " +((ResultSet) rs).getString(param), e);
                }
                catch (SQLException e2)
                {
                    throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","float","" + param, column, e.getMessage()), e);
                }
            }
        }

        return value;
    }

    public void setObject(Object ps, int param, Object value)
    {
        try
        {
            if (value == null)
            {
                ((PreparedStatement) ps).setNull(param, getTypeInfo().dataType);
            }
            else
            {
                ((PreparedStatement) ps).setFloat(param, ((Float)value).floatValue());
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","Object","" + value, column, e.getMessage()), e);
        }
    }

    public Object getObject(Object rs, int param)
    {
        Object value;

        try
        {
            float f = ((ResultSet) rs).getFloat(param);
            value = ((ResultSet) rs).wasNull() ? null : new Float(f);
        }
		catch (SQLException e)
		{           
			try
			{
				//when value is real in database, cause cause a parse error when calling getFloat
				//JDBC error:Value can not be converted to requested type.

				value = new Float(Float.parseFloat(((ResultSet) rs).getString(param)));
				if (((ResultSet) rs).wasNull())
				{
					throw new NullValueException(LOCALISER.msg("RDBMS.Mapping.IllegalNullValueInColumn",column));
				}
			}
			catch (SQLException e1)
			{
				try
				{
					throw new JDODataStoreException("Can't get float result: param = " + param + " - " +((ResultSet) rs).getString(param), e);
				}
				catch (SQLException e2)
				{
					throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","Object","" + param, column, e.getMessage()), e);
				}
			}
		}        

        return value;
    }
}