/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.AbstractList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryResult;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Abstract representation of a QueryResult for RDBMS queries.
 *
 * @version $Revision: 1.3 $ 
 */
public abstract class AbstractQueryResult extends AbstractList
    implements QueryResult, java.io.Serializable
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    /** Number of these results in the PM */
    protected int pmResultsNumber = -1;

    /** The Query object. */
    protected final Query query;

    /** Result Object Factory used for accessing the elements. */
    protected final ResultObjectFactory rof;

    /** The ResultSet containing the results. */
    protected ResultSet rs;

    /** The Query statement used to generate these results. */
    protected QueryExpression qs;

    /** Whether the results are close. */
    protected boolean closed = false;

    /**
     * Constructor of the result from a Query.
     * @param qs The Query Statement
     * @param query The Query
     * @param rof The factory to retrieve results from
     * @param rs The ResultSet from the Query Statement
     */
    public AbstractQueryResult(QueryExpression qs, Query query, ResultObjectFactory rof, ResultSet rs)
    {
        this.query = query;
        this.rof = rof;
        this.rs = rs;
        this.qs = qs;

        // Register this result with the controlling PM.
        pmResultsNumber = ((PersistenceManager)query.getPersistenceManager()).addQueryResult(this);
    }

    /**
     * Accessor whether the results are open.
     * @return Whether it is open.
     */
    protected boolean isOpen()
    {
        return closed == false;
    }

    /**
     * Internal method to throw a JDOUserException if the ResultSet is open.
     */
    protected void assertIsOpen()
    {
        if (!isOpen())
        {
            throw new JDOUserException(LOCALISER.msg("QueryResult.Closed"));
        }
    }

    /**
     * Internal method to close the ResultSet.
     */
    protected void closeResultSet()
    {
        if (rs != null)
        {
            try
            {
                Statement stmt = null;
                try
                {
                    stmt = rs.getStatement();

                    // Close the result set
                    rs.close();
                }
                catch (SQLException e)
                {
                    JPOXLogger.RDBMS.error(LOCALISER.msg("QueryResult.CloseError",e));
                }
                finally
                {
                    try
                    {
                        if (stmt != null)
                        {
                            // Close the original statement
                            stmt.close();
                        }
                    }
                    catch (SQLException e)
                    {
                        // Do nothing
                    }
                }
            }
            finally
            {
                rs = null;
            }
        }
    }

    /**
     * Method to close the results.
     */
    public synchronized void close()
    {
        closed = true;

        // Deregister this result with the managing PM
        ((PersistenceManager)query.getPersistenceManager()).removeQueryResult(pmResultsNumber);

        // Close the result set
        closeResultSet();
    }

    /**
     * Finalise method to make sure that we close correctly.
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable
    {
        try
        {
            if (!closed)
            {
                close();
            }
        }
        finally
        {
            super.finalize();
        }
    }

    // ------------------------- Implementation of the List -------------------------

    /**
     * Method to add a result. Unsupported.
     * @param index The position to add
     * @param element The results to add
     */
    public void add(int index, Object element)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
    }

    /**
     * Method to add results. Unsupported.
     * @param o The result to add
     * @return true if added successfully
     */
    public boolean add(Object o)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
    }

    /**
     * Method to add results. Unsupported.
     * @param index The position to add
     * @param c The results to add
     * @return true if added successfully
     */
    public boolean addAll(int index, Collection c)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
    }

    /**
     * Method to clear the results.
     */
    public void clear()
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
    }

    /**
     * Method to check if the specified object is contained in this result.
     * @param o The object
     * @return Whether it is contained here.
     */
    public boolean contains(Object o)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotSupported"));
    }

    /**
     * Method to check if all of the specified objects are contained here.
     * @param c The collection of objects
     * @return Whether they are all contained here.
     */
    public boolean containsAll(Collection c)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotSupported"));
    }

    /**
     * Equality operator for QueryResults.
     * Overrides the AbstractList implementation since that uses 
     * size() and iterator() and that would cause problems when closed.
     * @param o The object to compare against
     * @return Whether they are equal
     */
    public abstract boolean equals(Object o);

    /**
     * Method to retrieve a particular element from the list.
     * @param index The index of the element
     * @return The element at index
     */
    public abstract Object get(int index);

    /**
     * Accessor for the hashcode of this object
     * @return The hash code
     */
    public int hashCode()
    {
        // Just base equality on the QueryStatement for now
        // NB For SQL queries the QueryStatement will be null so use Query
        if (qs != null)
        {
            return qs.hashCode();
        }
        else
        {
            return query.hashCode();
        }
    }

    /**
     * Method to check the index of a result. Not supported.
     * @param o The result
     * @return The position
     */
    public int indexOf(Object o)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotSupported"));
    }

    /**
     * Returns <tt>true</tt> if this collection contains no elements.<p>
     *
     * This implementation check if the result is empty.
     *
     * @return <tt>true</tt> if this collection contains no elements.
     */
    public abstract boolean isEmpty();

    /**
     * Accessor for an iterator for the results.
     * @return The iterator
     */
    public abstract Iterator iterator();

    /**
     * Method to check the last index of a result. Not supported.
     * @param o The result
     * @return The last index
     */
    public int lastIndexOf(Object o)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotSupported"));
    }

    /**
     * Accessor for a list iterator for the results.
     * @return a ListIterator with the query results
     */
    public abstract ListIterator listIterator();

    /**
     * Method to remove a result. Not supported.
     * @param index The position of the result.
     * @return The removed object.
     */
    public Object remove(int index)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
    }

    /**
     * Method to set the position of a result. Not supported.
     * @param index Position of the result
     * @param element The result
     * @return The element
     */
    public Object set(int index, Object element)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
    }

    /**
     * Method to return the size of the result.
     * @return The size of the result.
     */
    public abstract int size();

    /**
     * Method return a sub list of results. Not supported.
     * @param fromIndex start position
     * @param toIndex end position (exclusive)
     * @return The list of results
     */
    public List subList(int fromIndex, int toIndex)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotSupported"));
    }

    /**
     * Method to return the results as an array.
     * @return The array.
     */
    public Object[] toArray()
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotSupported"));
    }

    /**
     * Method to return the results as an array.
     * @param a The array to copy into. 
     * @return The array.
     */
    public Object[] toArray(Object[] a)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotSupported"));
    }

    // -------------------------------------- JDOQL Query Methods --------------------------------------

    /* (non-Javadoc)
     * @see org.jpox.store.query.Queryable#newQueryStatement()
     */
    public QueryExpression newQueryStatement()
    {
        return qs;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.query.Queryable#newQueryStatement(java.lang.Class)
     */
    public QueryExpression newQueryStatement(Class candidateClass)
    {
        qs.reset(); //reset this query, so it can be used as candidate for another query
        return qs;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.query.Queryable#newResultObjectFactory(org.jpox.store.QueryStatement)
     */
    public ResultObjectFactory newResultObjectFactory(QueryExpression stmt,boolean ignoreCache,Class resultClass, boolean useFetchPlan)
    {
        return rof;
    }
}