/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Map;

import javax.jdo.Extent;
import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryResult;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Base definition of a query using SQL.
 *
 * @see Query
 * @version $Revision: 1.4 $
 */
public abstract class BaseSQLQuery extends RDBMSQuery
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER_RDBMS=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    protected transient final String sqlText;

    protected transient String jdbcSqlText = null;

    protected transient StatementExpressionIndex[] statementExpressionIndex;

    /**
     * Constructs a new query instance from the existing query.
     * @param pm Persistence Manager
     * @param query Existing query
     */
    public BaseSQLQuery(PersistenceManager pm, BaseSQLQuery query)
    {
        this(pm, query.sqlText);
    }

    /**
     * Constructs a new query instance having the same criteria as the given
     * query.
     * @param pm The PersistenceManager
     * @param sql_text The SQL query string
     */
    public BaseSQLQuery(PersistenceManager pm, String sql_text)
    {
        super(pm);

        candidateClass = null;
        filter = null;
        imports = null;
        variables = null;
        parameters = null;
        ordering = null;

        if (sql_text == null)
        {
            throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.StatementNullError"));
        }
        this.sqlText = sql_text.trim();
    }

    /**
     * Utility to discard any compiled query.
     * @see org.jpox.store.query.Query#discardCompiled()
     */
    protected void discardCompiled()
    {
        super.discardCompiled();

        jdbcSqlText = null;
        statementExpressionIndex = null;
    }

    /**
     * Convenience method to return whether the query should return a single row.
     * @return Whether a single row should result
     */
    protected boolean shouldReturnSingleRow()
    {
        if (unique)
        {
            // UNIQUE implies a single row
            return true;
        }
        return false;
    }

    /**
     * Set the candidate Extent to query.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     *
     * @param pcs   the Candidate Extent.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#setCandidates(javax.jdo.Extent)
     */
    public void setCandidates(Extent pcs)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.CandidateExtentsInapplicableError"));
    }

    /**
     * Set the candidate Collection to query.
     *
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     *
     * @param pcs the Candidate collection.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#setCandidates(java.util.Collection)
     */
    public void setCandidates(Collection pcs)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.CandidateCollectionsInapplicableError"));
    }

    /**
     * Set the result for the results. The application might want to get results
     * from a query that are not instances of the candidate class. The results
     * might be fields of persistent instances, instances of classes other than
     * the candidate class, or aggregates of fields.
     * 
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * 
     * @param result The result parameter consists of the optional keyword
     * distinct followed by a commaseparated list of named result expressions or
     * a result class specification.
     * @see javax.jdo.Query#setResult
     * @since 1.1
     */
    public void setResult(String result)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.ResultInapplicableError"));
    }

    /**
     * Set the range of the results. Not applicable to SQL/JPOXSQL queries.
     * @param fromIncl From element no (inclusive) to return
     * @param toExcl To element no (exclusive) to return
     * @see javax.jdo.Query#setRange(long, long)
     * @since 1.1
     */
    public void setRange(int fromIncl, int toExcl)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.RangeInapplicableError"));
    }

    /**
     * Method to set whether to use subclasses. This is not used with JPOXSQL.
     * @param subclasses Whether to use subclasses
     */
    public void setSubclasses(boolean subclasses)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.CandidateExtentsInapplicableError"));
    }

    /**
     * Set the filter for the query.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * @param filter the query filter.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#setFilter
     */
    public void setFilter(String filter)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.FilterStringsInapplicableError"));
    }

    /**
     * Declare the unbound variables to be used in the query.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * @param variables the variables separated by semicolons.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#declareVariables
     */
    public void declareVariables(String variables)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.VariablesInapplicableError"));
    }

    /**
     * Set the grouping specification for the result Collection.
     * @param grouping the grouping specification.
     * @see javax.jdo.Query#setGrouping
     */
    public void setGrouping(String grouping)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.GroupingNotExplicitError"));
    }

    /**
     * Set the ordering specification for the result Collection.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * @param ordering  the ordering specification.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#setOrdering
     */
    public void setOrdering(String ordering)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.OrderingNotExplicitError"));
    }

    /**
     * Verify the elements of the query and provide a hint to the query to
     * prepare and optimize an execution plan.
     * @see javax.jdo.Query#compile
     */
    public void compile()
    {
        if (!isCompiled)
        {
            super.compile();
            generateQueryStatement();
            if (JPOXLogger.JDO_QUERY.isDebugEnabled())
            {
                JPOXLogger.JDO_QUERY.debug(LOCALISER_RDBMS.msg("SQL.Statement.Summary", jdbcSqlText));
            }

            isCompiled = true;
        }
    }

    /**
     * Convenience method to take the user specified SQL statement, and align it with
     * field names for the candidate class etc.
     */
    protected abstract void generateQueryStatement();

    /**
     * Execute the query and return the filtered QueryResult.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult.
     */
    public abstract Collection performExecute(Map parameters);

    /**
     * Utility to take a ResultSet and return a QueryResult for those results,
     * assuming that no candidate class is supplied. The QueryResult will return
     * either a result class type, or Object/Object[] depending on whether a
     * ResultClass has been defined.
     * @param rs The ResultSet
     * @return The QueryResult
     * @throws SQLException Thrown when an error occurs creating the QueryResult
     */
    protected QueryResult getQueryResultForNoCandidateClass(ResultSet rs)
    throws SQLException
    {
        // No candidate class, so use resultClass or Object/Object[]
        Class requiredResultClass = resultClass;
        int numberOfColumns = 0;
        String[] resultFieldNames = null;
        try
        {
            ResultSetMetaData rsmd = rs.getMetaData();
            numberOfColumns = rsmd.getColumnCount();
            if (requiredResultClass == null)
            {
                if (numberOfColumns == 1)
                {
                    requiredResultClass = Object.class;
                }
                else
                {
                    requiredResultClass = Object[].class;
                }
            }

            // Generate names to use for the fields based on the column names
            resultFieldNames = new String[numberOfColumns];
            for (int i=0;i<numberOfColumns;i++)
            {
                resultFieldNames[i] = rsmd.getColumnName(i+1);
            }
        }
        catch (SQLException sqe)
        {
            // Do nothing
        }

        QueryResult qr = null;
        if (getResultSetType().equals("scroll-insensitive"))
        {
            qr = new InsensitiveQueryResult(null, this, new ResultClassROF(requiredResultClass, resultFieldNames), rs, null);
        }
        else
        {
            qr = new ForwardQueryResult(null, this, new ResultClassROF(requiredResultClass, resultFieldNames), rs, null);
        }
        return qr;
    }

    /**
     * Execute the query to delete persistent objects.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult of the deleted objects.
     */
    protected long performDeletePersistentAll(Map parameters)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.DeletePersistent.NotSupported"));
    }
}