/**********************************************************************
Copyright (c) 2003 Mike Martin (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2003 Erik Bengtson - removed unused import
2003 Andy Jefferson - commented and localised
2004 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - added support for uniqueness of query
2004 Andy Jefferson - added Single-String query support
2005 Andy Jefferson - added timeout support
2005 Andy Jefferson - added implicit variable/parameter support
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Map.Entry;

import javax.jdo.Extent;
import javax.jdo.JDODataStoreException;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.plugin.ConfigurationElement;
import org.jpox.plugin.Extension;
import org.jpox.plugin.PluginManager;
import org.jpox.store.exceptions.NoSuchPersistentFieldException;
import org.jpox.store.expression.AggregateExpression;
import org.jpox.store.expression.ArrayExpression;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.ClassExpression;
import org.jpox.store.expression.CollectionExpression;
import org.jpox.store.expression.Literal;
import org.jpox.store.expression.MapExpression;
import org.jpox.store.expression.NewObjectExpression;
import org.jpox.store.expression.NullLiteral;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.UnboundVariable;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.JDOQLResultSetMetaData;
import org.jpox.store.query.Parser;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryResult;
import org.jpox.store.query.Queryable;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.RDBMSTransaction;
import org.jpox.util.ClassUtils;
import org.jpox.util.Imports;
import org.jpox.util.JPOXLogger;

/**
 * A JDO query that uses the JQOQL language.
 * The query can be specified via method calls, or via a single-string form.
 *
 * @see Query
 * @version $Revision: 1.3.2.1 $
 */
public class JDOQLQuery extends RDBMSQuery
{
    /** Keywords used in single-string JDOQL. Uppercase variants specified here, but we allow the lowercase form. */
    public static final String[] SINGLE_STRING_KEYWORDS = {
        "SELECT", "UNIQUE", "INTO", "FROM", "EXCLUDE", "SUBCLASSES", "WHERE",
        "VARIABLES", "PARAMETERS", "GROUP", "ORDER", "BY", "RANGE"
        };

    /** Keywords in lowercase (we avoid calling toLowerCase() multiple times, which is expensive operation) **/
    public static final String[] SINGLE_STRING_KEYWORDS_LOWERCASE = {
        "select", "unique", "into", "from", "exclude", "subclasses", "where",
        "variables", "parameters", "group", "order", "by", "range"
        };

    protected transient Extent candidateExtent = null;
    protected transient Collection candidateCollection = null;
    protected transient Queryable candidates = null;

    private transient AbstractClassMetaData candidateCmd = null;
    private transient QueryExpression queryStmt = null;
    private transient ResultObjectFactory rof = null;
    /** distinct results **/
    private transient boolean distinct = false;
    
    /** Result metadata **/
    private transient JDOQLResultSetMetaData resultMetaData;

    /**
     * registered user defined ScalarExpressions. keyed by fully qualified class name
     * and values are the classes extending ScalarExpression class
     * A user ScalarExpression must have the original class + the expression class
     * e.g. the java.lang.Math (original class) + org.jpox.store.expression.MathExpression (expression class)
     */
    private static transient Map userDefinedScalarExpressions = new Hashtable();
    
    /**
     * Constructs a new query instance that uses the given persistence manager.
     * @param pm the associated persistence manager for this query.
     */
    public JDOQLQuery(PersistenceManager pm)
    {
        super(pm);
        registerScalarExpressions(pm.getPMFContext().getPluginManager(), pm.getClassLoaderResolver());

        candidateClass = null;
        candidateClassName = null;
        filter = null;
        imports = null;
        variables = null;
        parameters = null;
        grouping = null;
        ordering = null;
        result = null;
        resultClass = null;
        resultClassName = null;
        range = null;
    }

    /**
     * Constructs a new query instance having the same criteria as the given query.
     * @param pm The PersistenceManager
     * @param q The query from which to copy criteria.
     */
    public JDOQLQuery(PersistenceManager pm, JDOQLQuery q)
    {
        super(pm);
        registerScalarExpressions(pm.getPMFContext().getPluginManager(), pm.getClassLoaderResolver());

        if (q == null)
        {
            candidateClass = null;
            candidateClassName = null;
            filter = null;
            imports = null;
            variables = null;
            parameters = null;
            grouping = null;
            ordering = null;
            result = null;
            resultClass = null;
            resultClassName = null;
            range = null;
            fromInclNo = 0;
            toExclNo = Long.MAX_VALUE;
        }
        else
        {
            candidateClass = q.candidateClass;
            candidateClassName = q.candidateClassName;
            filter = q.filter;
            imports = q.imports;
            variables = q.variables;
            parameters = q.parameters;
            grouping = q.grouping;
            ordering = q.ordering;
            result = q.result;
            resultClass = q.resultClass;
            resultClassName = q.resultClassName;
            range = q.range;
            fromInclNo = q.fromInclNo;
            toExclNo = q.toExclNo;

            // JDO 1.0.1, Sect 12.6.3 : IgnoreCache is preserved when using other query baseline
            ignoreCache = q.ignoreCache;
        }
    }

    /**
     * Constructor for a JDOQL query where the query is specified using the "Single-String" format.
     * @param pm The persistence manager
     * @param query The query string
     */
    public JDOQLQuery(PersistenceManager pm, String query)
    {
        super(pm);
        registerScalarExpressions(pm.getPMFContext().getPluginManager(), pm.getClassLoaderResolver());
        
        new JDOQLSingleStringParser(this, query).parse();
    }

    /**
     * Register ScalarExpression classes delcared as plug-ins extensions
     * TODO currently register only the first time this class is instantiated. Should be registered per PMF?
     * @param pluginMgr The PluginManager
     * @param clr The ClassLoaderResolver to load the literal and ScalarExpression classes
     */
    private void registerScalarExpressions(PluginManager pluginMgr, ClassLoaderResolver clr)
    {
        if (userDefinedScalarExpressions.isEmpty())
        {
            Extension[] ex = pluginMgr.getExtensionPoint("org.jpox.store_expression_scalarexpression").getExtensions();
            for( int i=0; i<ex.length; i++)
            {
                ConfigurationElement[] confElm = ex[i].getConfigurationElements();
                for (int c=0; c<confElm.length; c++)
                {       
                    Class literalClass = null;
                    if( confElm[c].getAttribute("literal-class") != null )
                    {
                        literalClass = pluginMgr.loadClass(confElm[c].getExtension().getPlugin().getId(),confElm[c].getAttribute("literal-class"));
                    }
                    Class scalarExpression = null;
                    if( confElm[c].getAttribute("scalar-expression-class") != null )
                    {
                        scalarExpression = pluginMgr.loadClass(confElm[c].getExtension().getPlugin().getId(),confElm[c].getAttribute("scalar-expression-class"));
                    }
                    registerScalarExpression(literalClass, scalarExpression, confElm[c].getAttribute("name"));
                }
            }
        }
    }

    /**
     * Register ScalarExpressions for the given <code>cls</code>. It allows
     * to perform operations in the query on <i>cls.method([arglist])</i>.
     * @param literal the class providing the operations; e.g. java.lang.Math.class
     * @param scalarExpressionClass the class with the corresponding ScalarExpression. eg. org.jpox.store.expression.MathExpression.class
     * @param name alternative name of the given literal class
     */
    public static void registerScalarExpression(Class literal, Class scalarExpressionClass, String name)
    {
        userDefinedScalarExpressions.put(name == null ? literal.getName() : name, scalarExpressionClass);
    }
    
    /**
     * Equality operator for JDOQL.
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof JDOQLQuery) || !super.equals(obj))
        {
            return false;
        }

        return true;
    }

    /**
     * Set the candidate Extent to query.
     *
     * @param pcs the Candidate Extent.
     * @see javax.jdo.Query#setCandidates(javax.jdo.Extent)
     */
    public void setCandidates(Extent pcs)
    {
        discardCompiled();
        assertIsModifiable();

        if (pcs == null)
        {
            JPOXLogger.JDO_QUERY.warn(LOCALISER.msg("Candidates.ExtentCantBeNull"));
            return;
        }
        if (!(pcs instanceof Queryable))
        {
            throw new JDOUnsupportedOptionException(LOCALISER.msg("JDOQL.ExtentNotQueryableError",pcs.getClass().getName()));
        }

        setSubclasses(pcs.hasSubclasses());
        setClass(pcs.getCandidateClass());

        candidateExtent = pcs;
        candidateCollection = null; // We have an Extent, so remove any collection
    }

    /**
     * Set the candidate Collection to query.
     *
     * @param pcs the Candidate collection.
     * @see javax.jdo.Query#setCandidates(java.util.Collection)
     */
    public void setCandidates(Collection pcs)
    {
        discardCompiled();
        assertIsModifiable();
        if (pcs == null)
        {
            JPOXLogger.JDO_QUERY.warn(LOCALISER.msg("Candidates.CollectionCantBeNull"));
            return;
        }

        candidateExtent = null;
        candidateCollection = pcs; // We have a Collection, so remove any Extent
    }

    /** Cached form of the single string form of the query. */
    String singleString = null;

    /**
     * Method to take the defined parameters for the query and form a single string.
     * This is used to print out the query for logging.
     * @return The single string
     */
    private String getSingleStringQuery()
    {
        if (singleString != null)
        {
            return singleString;
        }

        StringBuffer str = new StringBuffer("SELECT ");        
        if (unique)
        {
            str.append("UNIQUE ");
        }
        if (result != null)
        {
            str.append(result + " ");
        }
        if (resultClassName != null)
        {
            str.append("INTO " + resultClassName + " ");
        }
        if (candidateClassName != null)
        {
            str.append("FROM " + candidateClassName + " ");
        }
        if (!subclasses)
        {
            str.append("EXCLUDE SUBCLASSES ");
        }
        if (filter != null)
        {
            str.append("WHERE " + filter + " ");
        }
        if (variables != null)
        {
            str.append("VARIABLES " + variables + " ");
        }
        if (parameters != null)
        {
            str.append("PARAMETERS " + parameters + " ");
        }
        if (imports != null)
        {
            str.append(imports + " ");
        }
        if (grouping != null)
        {
            str.append("GROUP BY " + grouping + " ");
        }
        if (ordering != null)
        {
            str.append("ORDER BY " + ordering + " ");
        }

        if (range != null)
        {
            str.append("RANGE " + range);
        }
        else if (fromInclNo > 0 || toExclNo != Long.MAX_VALUE)
        {
            str.append("RANGE " + fromInclNo + "," + toExclNo);
        }

        singleString = str.toString().trim();
        return singleString;
    }

    /**
     * Verify the elements of the query and provide a hint to the query to
     * prepare and optimize an execution plan.
     * @see javax.jdo.Query#compile
     */
    public void compile()
    {
        if (JPOXLogger.JDO_QUERY.isDebugEnabled())
        {
            JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("JDOQL.Statement.Precompile", getSingleStringQuery()));
        }

        // Perform basic checks on the parameters and variables
        super.compile();
        Iterator it = userDefinedScalarExpressions.keySet().iterator();
        while( it.hasNext() )
        {
            parsedImports.importPackage((String)it.next());
        }
        if (isPreCompile)
        {
            // Query.compile() called, so use the compiler to perform a "pre-compile" check to validate the expressions
            new Compiler(null).preCompile();
        }

        isCompiled = true;
    }

    /**
     * Method to discard our current compiled query due to changes.
     * @see org.jpox.store.query.Query#discardCompiled()
     */
    protected void discardCompiled()
    {
        super.discardCompiled();
        singleString = null;
        queryStmt = null;
        rof = null;
    }

    /**
     * Convenience method to return whether the query should return a single row.
     * @return Whether a single row should result
     */
    protected boolean shouldReturnSingleRow()
    {
        if (unique)
        {
            // UNIQUE implies a single row
            return true;
        }
        else if (grouping != null)
        {
            // GROUP BY implies multiple possible
            return false;
        }
        else if (candidates instanceof ResultExpressionsQueryable)
        {
            // All result expressions being aggregate implies a single row
            return ((ResultExpressionsQueryable)candidates).hasAggregatedExpressionsOnly();
        }
        return false;
    }


    /**
     * Execute the query and return the filtered QueryResult.
     * @param parameters Map containing all of the parameters.
     * @return the filtered QueryResult.
     */
    protected Collection performExecute(Map parameters)
    {
        // Throw away any existing compilation remnants
        isPreCompile = false;
        discardCompiled();

        try
        {
            // execution-compile the query statement
            if (JPOXLogger.JDO_QUERY.isDebugEnabled())
            {
                JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("JDOQL.Statement.Summary", getSingleStringQuery()));
            }

            long startTime = System.currentTimeMillis();
            ExecutedCompileCache cache = (ExecutedCompileCache) compiledCache.get(getSingleStringQuery());
            if (cache == null || candidateCollection != null)
            {
                // Compile imports, explicit parameters and explicit variables
                super.compile();
                Iterator it = userDefinedScalarExpressions.keySet().iterator();
                while (it.hasNext())
                {
                    parsedImports.importPackage((String)it.next());
                }
                Compiler c = new Compiler(parameters);
                queryStmt = c.executionCompile();
                if (candidateCollection == null)
                {
                    //only cache if use dont use candidates, otherwise the query using candidates may be different
                    //TODO temporarily disabled
                    //compiledCache.put(getSingleStringQuery(),new ExecutedCompileCache(queryStmt,getSingleStringQuery(),candidates,parameterNames));
                }
            }
            else
            {
                candidates = cache.getCandidates();
                queryStmt = cache.getQueryExpression();
                parameterNames = cache.getParameterNames();
                queryStmt.reset();
            }

            // Apply the Query FetchPlan to the query
            // Note : we could have added getFetchPlan() to Queryable, but this would affect *many* classes
            // so leaving like this for now
            if (candidates instanceof CollectionCandidates)
            {
                ((CollectionCandidates)candidates).getFetchPlan().setGroups(fetchPlan.getGroups());
            }
            else if (candidates instanceof Extent)
            {
                ((Extent)candidates).getFetchPlan().setGroups(fetchPlan.getGroups());
            }

            boolean useFetchPlan = pm.getPMFContext().getPmfConfiguration().getQueryUseFetchPlan();
            if (extensions != null && extensions.containsKey("org.jpox.query.useFetchPlan"))
            {
                useFetchPlan = Boolean.valueOf((String)extensions.get("org.jpox.query.useFetchPlan")).booleanValue();
            }

            if (pm.hasQueryRun(toString()))
            {
                rof = candidates.newResultObjectFactory(queryStmt, getIgnoreCache(), resultClass, useFetchPlan);
            }
            else
            {
                rof = candidates.newResultObjectFactory(queryStmt, getIgnoreCache(), resultClass, useFetchPlan);
            }

            if (JPOXLogger.JDO_QUERY.isDebugEnabled())
            {
                JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("JDOQL.Statement.CompileTime",(System.currentTimeMillis() - startTime)));
            }

            isCompiled = true;
        }
        finally
        {
            if (!isCompiled)
            {
                discardCompiled();
            }
        }

        if (candidates.isEmpty())
        {
            // if the candidates are empty, we don't have to go to the database
            // to find out that we have no elements to return
            return new ArrayList();
        }

        isPreCompile = true;

        // Execute the query
        QueryResult qr = null;

        // Find the locking strategy (use the txn value, and override with any local query setting
        boolean useUpdateLock = ((RDBMSTransaction)pm.currentTransaction()).useUpdateLockOnFetch();
        if (extensions != null && extensions.containsKey("org.jpox.query.useUpdateLock"))
        {
            useUpdateLock = Boolean.valueOf((String)extensions.get("org.jpox.query.useUpdateLock")).booleanValue();
        }

        try
        {
            RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
            Connection conn = storeMgr.getConnection(pm, false, true);

            try
            {
                StatementText stmtText = queryStmt.toStatementText(useUpdateLock);
                PreparedStatement ps = getStatement(conn, stmtText);
                try
                {
                    // Apply timeouts, result set constraints etc
                    prepareStatementForExecution(ps);

                    // Add a limit on the number of rows to include the maximum we may need
                    if (toExclNo != 0 && toExclNo != Long.MAX_VALUE)
                    {
                        ps.setMaxRows((int)toExclNo);
                    }

                    ResultSet rs = ((RDBMSManager)pm.getStoreManager()).executeStatementQuery(stmtText.toString(), ps);
                    try
                    {
                        // Check the type of result set needed
                        boolean forwardResultSet = true;
                        if (getResultSetType().equals("scroll-insensitive"))
                        {
                            forwardResultSet = false;
                        }

                        if (forwardResultSet)
                        {
                            qr = new ForwardQueryResult(queryStmt, this, rof, rs, distinct ? null : candidateCollection);
                        }
                        else
                        {
                            qr = new InsensitiveQueryResult(queryStmt, this, rof, rs, distinct ? null : candidateCollection);
                        }
                    }
                    finally
                    {
                        if (qr == null)
                        {
                            rs.close();
                        }
                    }
                }
                finally
                {
                    if (qr == null)
                    {
                        ps.close();
                    }
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("JDOQL.ExecutionError", queryStmt.toStatementText(useUpdateLock), e.getMessage()), e);
        }

        pm.addQueryRun(toString());
        queryResults.add(qr);

        return qr;
    }

    /**
     * Execute the query to delete persistent objects.
     * @param parameters the Map containing all of the parameters.
     * @return the number of deleted objects.
     */
    protected long performDeletePersistentAll(Map parameters)
    {
        // Save the required value for unique (find the instances without it)
        boolean requiresUnique = unique;
        if (unique)
        {
            unique = false;
        }

        Collection results = performExecute(parameters);
        int number = 0;
        if (results == null)
        {
            number = 0;
        }
        else
        {
            // TODO : To make this most efficient we shouldn't instantiate things into memory
            // but instead check if the object to be deleted implements DeleteCallback, or there are
            // any lifecycle listeners for this object type, or if there are any dependent fields
            // and only then do we instantiate it.
            number = results.size();
            if (requiresUnique && number > 1)
            {
                throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.UniqueSpecifiedButMultipleResults"));
            }

            // Instances to be deleted are flushed first (JDO2 [14.8-4])
            Iterator resultsIter = results.iterator();
            while (resultsIter.hasNext())
            {
                PersistenceCapable pc = (PersistenceCapable)resultsIter.next();
                StateManager sm = pm.findStateManager(pc);
                sm.flush();
            }

            // Perform the deletion
            pm.deletePersistentAll(results);

            // TODO Do we need to remove these from the PM.queryRun store and this.queryResults ?
            ((QueryResult)results).close();
        }
        return number;
    }
    
    private static Map compiledCache = new HashMap();
    
    /**
     * Cache a compiled query
     */
    public static class ExecutedCompileCache
    {
        private final QueryExpression queryStmt;
        private final String singleStringQuery;
        private final Queryable candidates;
        private final List parameterNames;
        public ExecutedCompileCache(QueryExpression queryStmt, String singleStringQuery, Queryable candidates, List parameterNames)
        {
            this.queryStmt = queryStmt;
            this.singleStringQuery = singleStringQuery;
            this.candidates = candidates;
            this.parameterNames = parameterNames;
        }
        
        public QueryExpression getQueryExpression()
        {
            return this.queryStmt;
        }
        
        public Queryable getCandidates()
        {
            return candidates;
        }
        
        public List getParameterNames()
        {
            return parameterNames;
        }
        
        public int hashCode()
        {
            return singleStringQuery.hashCode();
        }
        
        public boolean equals(Object obj)
        {
            if (obj == null)
            {
                return false;
            }
            if (!(obj instanceof ExecutedCompileCache))
            {
                return false;
            }
            ExecutedCompileCache cached = (ExecutedCompileCache) obj;
            return cached.queryStmt.equals(queryStmt) && cached.singleStringQuery.equals(singleStringQuery);
        }
    }

    /**
     * Retrieve the metadata for the results 
     * @return the ResultSetMetaData
     */
    public JDOQLResultSetMetaData getResultSetMetaData()
    {
        if( resultMetaData == null )
        {
            throw new JDOUserException("You must compile the query before calling this method.");
        }
        return resultMetaData;
    }
    
    /**
     * Inner class to compile the query for execution.
     **/
    public class Compiler implements UnboundVariable.VariableBinder
    {
        private final Map parameters;
        private final Map expressionsByVariableName = new HashMap();
        private QueryExpression qs = null;
        private Parser p = null;
        private boolean explicitParameters = false;
        private int implicitParamNo = 0;
        private boolean explicitVariables = false;

        /**
         * Constructor
         * @param parameters map of declared parameters in the query
         */
        public Compiler(Map parameters)
        {
            if (JDOQLQuery.this.variables != null && JDOQLQuery.this.variables.length() > 0)
            {
                // Using explicit variables
                explicitVariables = true;
            }

            if (JDOQLQuery.this.parameters != null && JDOQLQuery.this.parameters.length() > 0)
            {
                // Using explicit parameters
                explicitParameters = true;
                if (parameters != null)
                {
                    if (parameters.size() != parameterNames.size())
                    {
                        // Number of explicit parameters doesn't match the number of parameter values
                        throw new JDOUserException(LOCALISER.msg("JDOQL.IncorrectNumberOfParametersError",
                            "" + parameterNames.size(), "" + parameters.size()));
                    }

                    for (Iterator it = parameterTypesByName.entrySet().iterator(); it.hasNext(); )
                    {
                        Map.Entry entry = (Entry) it.next();
                        Object key = entry.getKey();
                        Object value = parameters.get(key);
                        if (value == null)
                        {
                            // primitive parameter can't be null
                            if (((Class)entry.getValue()).isPrimitive())
                            {
                                throw new JDOUserException(LOCALISER.msg("Query.InvalidNullArgumentForPrimitiveParameter",
                                    entry.getKey(), ((Class)entry.getValue()).getName()));
                            }
                        }
                        else
                        {
                            if (value instanceof PersistenceCapable)
                            {
                                // parameters bound to a different PM can't be used with this PM
                                PersistenceCapable pc = (PersistenceCapable)value;
                                if (pc.jdoGetPersistenceManager() != null && pc.jdoGetPersistenceManager() != pm)
                                {
                                    throw new JDOUserException(LOCALISER.msg("JDOQL.ParameterBoundToDifferentPM", key));
                                }
                            }
                        }
                    }
                }
            }
            else
            {
                // Using implicit parameters
                // TODO Need to check the number of implicit params in the query (only do-able after compiling)
                if (parameters != null)
                {
                    Iterator parameterEntryIter = parameters.entrySet().iterator();
                    while (parameterEntryIter.hasNext())
                    {
                        Map.Entry entry = (Map.Entry)parameterEntryIter.next();
                        Object value = entry.getValue();
                        if (value != null && value instanceof PersistenceCapable)
                        {
                            // parameters bound to a different PM can't be used with this PM
                            PersistenceCapable pc = (PersistenceCapable)value;
                            if (pc.jdoGetPersistenceManager() != null && pc.jdoGetPersistenceManager() != pm)
                            {
                                // TODO Really should put the param name in here, but we don't know it
                                throw new JDOUserException(LOCALISER.msg("JDOQL.ParameterBoundToDifferentPM", pc));
                            }
                        }
                    }
                }
            }

            this.parameters = parameters;
        }

        /**
         * Method to provide pre-compilation checks to catch errors.
         * This is performed when calling the JDO Query.compile() method.
         */
        public void preCompile()
        {
            // Compile the candidates
            compileCandidates();

            // Create a QueryStatement so that we can generate the query expressions for the parts of the query
            // The statement is thrown away after the precompile currently
            qs = candidates.newQueryStatement(candidateClass);

            performCompile(qs);
        }

        /**
         * Method to execution-compile the query.
         * Generates the query and returns it.
         * @return the execution compiled query
         */
        public QueryExpression executionCompile()
        {
            // Compile the candidates
            compileCandidates();

            // Create the QueryStatement so we can generate the executable query
            qs = candidates.newQueryStatement(candidateClass);

            // Apply any restrictions on joining before generating the statement content
            boolean useInner = pm.getPMFContext().getPmfConfiguration().getQueryUseInnerJoinsOnly();
            if (extensions != null && extensions.containsKey("org.jpox.jdoql.useInnerJoinsOnly"))
            {
                useInner = Boolean.valueOf((String)extensions.get("org.jpox.jdoql.useInnerJoinsOnly")).booleanValue();
            }
            if (useInner)
            {
                qs.setUseInnerJoinsOnly(true);
            }

            performCompile(qs);

            return qs;
        }
        
        /**
         * Perform the actual compilation of the query.
         * @param qs The QueryExpression to use during compilation (if required)
         */
        private void performCompile(QueryExpression qs)
        {
            // Compile and apply the result/result-class to the query
            fieldExpressions.clear();
            compileResult(qs);
            ScalarExpression[] resultFieldExprs = (ScalarExpression[]) fieldExpressions.toArray(new ScalarExpression[fieldExpressions.size()]);
            for (int i=0; i<resultFieldExprs.length; i++)
            {
                if( resultFieldExprs[i].getLogicSetExpression()==null)
                {
                    if (resultFieldExprs[i] instanceof UnboundVariable)
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.SpecifiedFieldNotFound", ((UnboundVariable)resultFieldExprs[i]).getVariableName()));
                    }
                }
                qs.crossJoin(resultFieldExprs[i].getLogicSetExpression(),true);
            }
            // Compile and apply the filter to the query
            compileFilter(qs);
            // Compile and apply the grouping to the query
            ScalarExpression[] groupingFieldExprs = null;
            ScalarExpression[] havingFieldExprs = null;
            if (grouping != null && grouping.length() > 0)
            {
                // The "grouping" string will be of the form "...., ...., ... HAVING ...."
                // so we parse it into the former part as a grouping clause, and the latter part as a having clause
                String groupingClause = null;
                String havingClause = null;
                if (grouping.indexOf("HAVING") >= 0)
                {
                    havingClause = grouping.substring(grouping.indexOf("HAVING") + 7);
                    groupingClause = grouping.substring(0, grouping.indexOf("HAVING")-1);
                }
                else if (grouping.indexOf("having") >= 0)
                {
                    havingClause = grouping.substring(grouping.indexOf("having") + 7);
                    groupingClause = grouping.substring(0, grouping.indexOf("having")-1);
                }
                else
                {
                    groupingClause = grouping;
                }

                // Compile the "grouping"
                fieldExpressions.clear();
                compileGrouping(qs, groupingClause);
                groupingFieldExprs = (ScalarExpression[]) fieldExpressions.toArray(new ScalarExpression[fieldExpressions.size()]);

                // Compile the "having"
                fieldExpressions.clear();
                compileHaving(qs, havingClause);
                havingFieldExprs = (ScalarExpression[]) fieldExpressions.toArray(new ScalarExpression[fieldExpressions.size()]);
            }

            // Compile and apply the ordering to the query
            fieldExpressions.clear();
            compileOrdering(qs);
            ScalarExpression[] orderingFieldExprs = (ScalarExpression[]) fieldExpressions.toArray(new ScalarExpression[fieldExpressions.size()]);

            // Check that all result expression fields, having fields and ordering fields are defined in any grouping specification
            if (groupingFieldExprs != null)
            {
                // Check that all ordering expression fields are in the grouping clause
                checkExpressionsAgainstGrouping(orderingFieldExprs, groupingFieldExprs, "JDOQL.FieldInOrderingNotSpecifiedInGrouping");

                // Check having fields against what is in the grouping spec
                checkExpressionsAgainstGrouping(havingFieldExprs, groupingFieldExprs, "JDOQL.FieldInHavingNotSpecifiedInGrouping");

                // Check that all result clause fields are in the grouping clause
                checkExpressionsAgainstGrouping(resultFieldExprs, groupingFieldExprs, "JDOQL.FieldInResultNotSpecifiedInGrouping");
            }

            // Compile and apply the range to the query
            compileRange(qs);

            // Check that all variables have been bound to the query
            checkVariableBinding();
        }

        /**
         * Convenience method to check the expressions against those specified in the grouping.
         * Throws a JDOUserException if one of the expressions is not present in the grouping expressions.
         * @param exprs The expressions to check
         * @param groupExprs The grouping expressions
         * @param localiserErrorString Name of a localiser error message to throw as the JDOUserException message.
         */
        private void checkExpressionsAgainstGrouping(ScalarExpression[] exprs, ScalarExpression[] groupExprs, String localiserErrorString)
        {
            if (exprs == null)
            {
                return;
            }
            for (int i=0; i<exprs.length; i++)
            {
                boolean exists = false;
                for (int j=0; j<groupExprs.length; j++)
                {
                    if (exprs[i].equals(groupExprs[j]))
                    {
                        exists = true;
                        break;
                    }
                }
                if (!(exprs[i] instanceof AggregateExpression) && !exists)
                {
                    throw new JDOUserException(LOCALISER.msg(localiserErrorString, exprs[i]));
                }
            }
        }

        /** Internal list of the field expressions that are just being parsed. Used for checking validity of the query components. */
        private List fieldExpressions = new ArrayList();

        /**
         * Bind a variable to the query.
         * @param name Name of the variable
         * @param expr The expression
         */
        public void bindVariable(String name, ScalarExpression expr)
        {
            ScalarExpression previousExpr = (ScalarExpression)expressionsByVariableName.put(name, expr);
            if (previousExpr != null)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("JDOQL.VariableBindError",name,expr,previousExpr));
            }
        }

        /**
         * Convenience method to check that all variables have been bound to the query.
         * @throws JDOUserException Thrown if a variable is found that is not bound.
         */
        private void checkVariableBinding()
        {
            for (int i=0;i<variableNames.size();i++)
            {
                String variableName = (String)variableNames.get(i);
                if (expressionsByVariableName.get(variableName) == null)
                {
                    boolean foundInResult = false;
                    //we allow unbounded variables to be projected in result
                    if( candidates instanceof ResultExpressionsQueryable)
                    {
                        ScalarExpression[] exprs = ((ResultExpressionsQueryable)candidates).getResultExpressions();
                        for( int j=0; j<exprs.length; j++ )
                        {
                            if( exprs[j] instanceof UnboundVariable )
                            {
                                if( ((UnboundVariable)exprs[j]).getVariableName().equals(variableName) )
                                {
                                    foundInResult = true;
                                }
                            }
                                    
                        }
                    }
                    //only throw exception if unbounded variable is not in result
                    if( !foundInResult )
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.VariableNotBoundError", variableName));
                    }
                }
            }
        }

        /**
         * Convenience method to process the candidates for this query.
         * Processes the "candidateClassName" and "candidateClass" and sets up "candidates".
         */
        private void compileCandidates()
        {
            // Check the candidate class existence
            if (candidateClass == null && candidateClassName != null)
            {
                try
                {
                    candidateClass = pm.getClassLoaderResolver().classForName(candidateClassName, true);
                }
                catch (JDOException jdoe)
                {
                    candidateClass = resolveClassDeclaration(candidateClassName);
                }
            }

            // Set the "candidates"
            if (candidateExtent != null)
            {
                candidates = (Queryable)candidateExtent;
            }
            else if (candidateCollection != null)
            {
                candidates = new CollectionCandidates(pm, candidateClass, candidateCollection);
                // Using this code below means that if the passed in is a Queryable (e.g org.jpox.sco.ArrayList)
                // and the candidate table uses a discriminator then it wont be selected, so commented out for now
                /*if (!(candidateCollection instanceof Queryable))
                {
                    candidates = new CollectionCandidates(pm, candidateClass, candidateCollection);
                }
                else
                {
                    candidates = (Queryable)candidateCollection;
                }*/
            }
            else
            {
                if (candidateClass == null)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.NoCandidateClassError"));
                }
                candidates = (Queryable)pm.getExtent(candidateClass, subclasses);
            }

            distinct = false;
            if (result != null)
            {
                // User has specified a result so set the candidates based on the result definition
                if (candidateCollection != null)
                {
                    candidates = new ResultExpressionsQueryable(pm, candidateClass, 
                        ((CollectionCandidates)candidates).getUserCandidates(), subclasses);
                }
                else
                {
                    candidates = new ResultExpressionsQueryable(pm, candidateClass, subclasses);
                }

                if (result != null && result.toLowerCase().startsWith("distinct"))
                {
                    distinct = true;
                }
            }
        }

        /**
         * Compile the result expressions and class.
         * @param qs Query Expression to apply the result to (if required)
         */
        private void compileResult(QueryExpression qs)
        {
            // Compile any result expressions
            ScalarExpression[] resultExprs = null;
            distinct = false;
            if (result != null)
            {
                String resultDefinition = result;
                if (result != null && result.toLowerCase().startsWith("distinct"))
                {
                    // DISTINCT specified so add direct to the query statement
                    if (qs != null)
                    {
                        qs.setDistinctResults(true);
                    }
                    resultDefinition = result.substring(8);
                    distinct = true;
                }
                resultExprs = compileExpressionsFromString(resultDefinition, parsedImports);
                Class[] resultTypes = new Class[resultExprs.length];
                
                for (int i=0; i<resultExprs.length; i++)
                {
                    //TODO localise these messages
                    if (resultExprs[i] instanceof CollectionExpression)
                    {
                        throw new JDOUserException(resultExprs[i].toStatementText(ScalarExpression.PROJECTION) +" is of type java.util.Collection and cannot be in the result.");
                    }
                    else if (resultExprs[i] instanceof MapExpression)
                    {
                        throw new JDOUserException(resultExprs[i].toStatementText(ScalarExpression.PROJECTION) +" is of type java.util.Map and cannot be in the result.");
                    }
                    //e.g.UnboundVariables have no mapping, so we ignore it
                    if( resultExprs[i].getMapping() != null )
                    {
                        if( resultExprs[i].getMapping().getType() != null )
                        {
                            resultTypes[i] = resolveClassDeclaration(resultExprs[i].getMapping().getType());
                        }
                        else
                        {
                            resultTypes[i] = resultExprs[i].getMapping().getJavaType();
                        }
                    }
                }
                ((ResultExpressionsQueryable)candidates).setResultExpressions(resultExprs);
                resultMetaData = new JDOQLResultSetMetaData(resultTypes);
            }
            else
            {
                resultMetaData = new JDOQLResultSetMetaData(new Class[] {candidateClass} );
            }

            if (resultClass == null && resultClassName != null)
            {
                // Resolve any result class
                ScalarExpression expr = compileExpressionFromString(resultClassName,parsedImports);
                resultClass = ((ClassExpression)expr).getCls();
            }

            if (resultClass != null && resultExprs != null)
            {
                if (QueryUtils.resultClassIsSimple(resultClass.getName()))
                {
                    if (resultExprs.length > 1)
                    {
                        // Invalid number of result expressions
                        throw new JDOUserException(LOCALISER.msg("Query.ResultClassSimpleButMultipleFields", resultClass.getName()));
                    }

                    Class exprType = resultExprs[0].getMapping().getJavaType();
                    boolean typeConsistent = false;
                    if (exprType == resultClass)
                    {
                        typeConsistent = true;
                    }
                    else if (exprType.isPrimitive())
                    {
                        Class resultClassPrimitive = ClassUtils.getPrimitiveTypeForType(resultClass);
                        if (resultClassPrimitive == exprType)
                        {
                            typeConsistent = true;
                        }
                    }
                    if (!typeConsistent)
                    {
                        // Inconsistent expression type not matching the result class type
                        throw new JDOUserException(LOCALISER.msg("Query.ResultClassSimpleButWrongType", resultClass.getName(), exprType));
                    }
                }
                else if (QueryUtils.resultClassIsUserType(resultClass.getName()))
                {
                    // Check for valid constructor (either using param types, or using default ctr)
                    Class[] ctrTypes = new Class[resultExprs.length];
                    for (int i=0;i<ctrTypes.length;i++)
                    {
                        ctrTypes[i] = resultExprs[i].getMapping().getJavaType();
                    }
                    Constructor ctr = ClassUtils.getConstructorWithArguments(resultClass, ctrTypes);
                    if (ctr == null && !ClassUtils.hasDefaultConstructor(resultClass))
                    {
                        // No valid constructor found!
                        throw new JDOUserException(LOCALISER.msg("Query.ResultClassNotCreatable", resultClass.getName()));
                    }
                    else if (ctr == null)
                    {
                        // We are using default constructor, so check the types of the result expressions for means of input
                        for (int i=0;i<resultExprs.length;i++)
                        {
                            String fieldName = resultExprs[i].getAlias();
                            Class fieldType = resultExprs[i].getMapping().getJavaType();
                            if (fieldName == null && resultExprs[i].getMapping().getFieldMetaData() != null)
                            {
                                fieldName = resultExprs[i].getMapping().getFieldMetaData().getName();
                            }

                            if (fieldName != null)
                            {
                                // Check for the field of that name in the result class
                                Class resultFieldType = null;
                                boolean publicField = true;
                                try
                                {
                                    Field fld = resultClass.getDeclaredField(fieldName);
                                    resultFieldType = fld.getType();

                                    // Check the type of the field
                                    if (!ClassUtils.typesAreCompatible(fieldType, resultFieldType) && 
                                        !ClassUtils.typesAreCompatible(resultFieldType, fieldType))
                                    {
                                        throw new JDOUserException(LOCALISER.msg("JDOQL.ResultClassFieldTypeMismatch", fieldName, fieldType.getName(), resultFieldType.getName()));
                                    }
                                    if (!Modifier.isPublic(fld.getModifiers()))
                                    {
                                        publicField = false;
                                    }
                                }
                                catch (NoSuchFieldException nsfe)
                                {
                                    publicField = false;
                                }

                                // Check for a public set method
                                if (!publicField)
                                {
                                    Method setMethod = QueryUtils.getPublicSetMethodForFieldOfResultClass(resultClass, fieldName, resultFieldType);
                                    if (setMethod == null)
                                    {
                                        // No setter, so check for a public put(Object, Object) method
                                        Method putMethod = QueryUtils.getPublicPutMethodForResultClass(resultClass);
                                        if (putMethod == null)
                                        {
                                            throw new JDOUserException(LOCALISER.msg("JDOQL.ResultClassFieldSetterNotFound", resultClass.getName(), fieldName));
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        /**
         * Convenience method to compile the filter.
         * Processes the filter and updates the QueryExpression accordingly.
         * @param qs The Query Expression to apply the filter to (if specified)
         */
        private void compileFilter(QueryExpression qs)
        {
            if (filter != null && filter.length() > 0)
            {
                ScalarExpression expr = compileExpressionFromString(filter, parsedImports);
                if (!(expr instanceof BooleanExpression))
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.FilterExpressionNotBooleanError", filter));               
                }
                if (qs != null)
                {
                    qs.andCondition((BooleanExpression)expr, true);
                }
            }
        }

        /**
         * Convenience method to compile the ordering defintion.
         * Processes any ordering definition and updates the QueryExpression accordingly.
         * @param qs The Query Expression to apply the ordering to (if specified)
         */
        private void compileOrdering(QueryExpression qs)
        {
            if (ordering != null && ordering.length() > 0)
            {
                // Compile the ordering
                StringTokenizer t1 = new StringTokenizer(ordering, ",");

                int n = t1.countTokens();
                ScalarExpression[] orderExprs = new ScalarExpression[n];
                boolean[] descending = new boolean[n];
                for (n = 0; t1.hasMoreTokens(); ++n)
                {
                    String orderExpression = t1.nextToken().trim(); 

                    // Allow all sensible forms of direction specification (not just JDOQL standard)
                    if (orderExpression.endsWith("ascending") || orderExpression.endsWith("ASCENDING"))
                    {
                        descending[n] = false;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"ascending".length());
                    }
                    else if (orderExpression.endsWith("asc") || orderExpression.endsWith("ASC"))
                    {
                        descending[n] = false;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"asc".length());
                    }
                    else if (orderExpression.endsWith("descending") || orderExpression.endsWith("DESCENDING"))
                    {
                        descending[n] = true;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"descending".length());
                    }
                    else if (orderExpression.endsWith("desc") || orderExpression.endsWith("DESC"))
                    {
                        descending[n] = true;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"desc".length());
                    }
                    else
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.InvalidOrderDirectionError",ordering));
                    }

                    orderExprs[n] = compileExpressionFromString(orderExpression, parsedImports);
                }

                // Update the query statement
                if (qs != null)
                {
                    qs.setOrdering(orderExprs, descending);
                }
            }
        }

        /**
         * Convenience method to compile the grouping.
         * Compiles the grouping definition and applies it to the passed Query Expression as appropriate.
         * @param qs The QueryExpression to update (if specified)
         * @param groupingClause The grouping clause string
         */
        private void compileGrouping(QueryExpression qs, String groupingClause)
        {
            if (groupingClause != null)
            {
                ScalarExpression[] groupExprs = compileExpressionsFromString(groupingClause, parsedImports);
                if (groupExprs != null && qs != null)
                {
                    for (int i=0;i<groupExprs.length;i++)
                    {
                        qs.addGroupingExpression(groupExprs[i]);
                    }
                }
            }
        }

        /**
         * Convenience method to compile the having clause
         * @param qs The QueryExpression to update (if specified)
         * @param havingClause The having clause string
         */
        private void compileHaving(QueryExpression qs, String havingClause)
        {
            if (havingClause != null)
            {
                ScalarExpression havingExpr = compileExpressionFromString(havingClause, parsedImports);
                if (qs != null)
                {
                    if (!(havingExpr instanceof BooleanExpression))
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.HavingExpressionNotBoolean", havingExpr));
                    }
                    qs.setHaving((BooleanExpression)havingExpr);
                }
            }
        }

        /**
         * Convenience method to compile the range.
         * Compiles any range string and extracts the fromInclNo, toExclNo as appropriate.
         * @param qs QueryExpression to apply the range to (if specified)
         */
        private void compileRange(QueryExpression qs)
        {
            if (range != null)
            {
                ScalarExpression[] exprs = compileExpressionsFromString(range, parsedImports);
                if (exprs.length > 0)
                {
                    if (!(exprs[0] instanceof Literal))
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.RangeNeedsLiteralExpression", "FROM", exprs[0]));
                    }
                    if (!(((Literal)exprs[0]).getValue() instanceof Number))
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.RangeLiteralExpressionNumberOnly", "FROM", exprs[0]));
                    }
                    fromInclNo = ((Number)((Literal)exprs[0]).getValue()).longValue();
                }
                if (exprs.length > 1)
                {
                    if (!(exprs[1] instanceof Literal))
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.RangeNeedsLiteralExpression", "TO", exprs[1]));
                    }
                    if (!(((Literal)exprs[1]).getValue() instanceof Number))
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.RangeLiteralExpressionNumberOnly", "TO", exprs[1]));
                    }
                    toExclNo = ((Number)((Literal)exprs[1]).getValue()).longValue();
                }
            }

            if (qs != null && fromInclNo > 0 || toExclNo != Long.MAX_VALUE)
            {
                qs.setRangeConstraint(fromInclNo, toExclNo != Long.MAX_VALUE ? toExclNo-fromInclNo : -1);
            }
        }

        /**
         * Convenience method to parse an expression string into its query expression.
         * @param str The string
         * @param imports The imports to utilise
         * @return The query expression for the passed string
         */
        private ScalarExpression compileExpressionFromString(String str, Imports imports)
        {
            p = new Parser(str, imports);
            ScalarExpression expr = compileExpression();
            if (!p.parseEOS())
            {
                throw new ExpressionSyntaxException(LOCALISER.msg("JDOQL.InvalidExpressionError"));
            }
            return expr;
        }

        /**
         * Convenience method to parse an expression string into its component query expressions.
         * This splits expressions at comma boundaries, whilst respecting that the string could include
         * expressions like "new MyClass(a, b, c)" and so keeping braced arguments together. If it wasn't
         * for this requirement we would have been able to just use a StringTokenizer.
         * @param str The string
         * @param imports The imports to utilise
         * @return The query expressions for the passed string
         */
        private ScalarExpression[] compileExpressionsFromString(String str, Imports imports)
        {
            String[] exprList = QueryUtils.getExpressionsFromString(str);
            if (exprList != null && exprList.length > 0)
            {
                ScalarExpression[] exprs = new ScalarExpression[exprList.length];
                for (int i=0;i<exprs.length;i++)
                {
                    exprs[i] = compileExpressionFromString(exprList[i], imports);
                }
                return exprs;
            }
            return null;
        }

        /**
         * Principal method for compiling an expression.
         * An expression could be the filter, the range, the result, etc.
         * @return The compiled expression
         */
        private ScalarExpression compileExpression()
        {
            return compileConditionalOrExpression();
            /*
             * OR ("||")
             * AND ("&&")
             * Bitwise OR ("|")
             * Bitwise XOR ("^")
             * Bitwise AND ("&")
             * Equality ("==") ("!=")
             * Relational (">=") (">") ("<=") ("<")
             * Additive ("+") ("-")
             * Multiplicative ("*") ("/") ("%")
             * Unary ("+") ("-")
             * Unary ("~") ("!")
             * Cast
             */
        }

        /**
         * This method deals with the OR condition
         * A condition specifies a combination of one or more expressions and logical (Boolean) operators and returns a value of TRUE, FALSE, or unknown
         */
        private ScalarExpression compileConditionalOrExpression()
        {
            ScalarExpression expr = compileConditionalAndExpression();

            while (p.parseString("||"))
            {
                expr = expr.ior(compileConditionalAndExpression());
            }

            return expr;
        }

        /**
         * This method deals with the AND condition
         * A condition specifies a combination of one or more expressions and
         * logical (Boolean) operators and returns a value of TRUE, FALSE, or 
         * unknown
         */
        private ScalarExpression compileConditionalAndExpression()
        {
            ScalarExpression expr = compileInclusiveOrExpression();

            while (p.parseString("&&"))
            {
                expr = expr.and(compileInclusiveOrExpression());
            }

            return expr;
        }

        private ScalarExpression compileInclusiveOrExpression()
        {
            ScalarExpression expr = compileExclusiveOrExpression();

            while (p.parseChar('|', '|'))
            {
                expr = expr.ior(compileExclusiveOrExpression());
            }

            return expr;
        }

        private ScalarExpression compileExclusiveOrExpression()
        {
            ScalarExpression expr = compileAndExpression();

            while (p.parseChar('^'))
            {
                expr = expr.eor(compileExclusiveOrExpression());
            }

            return expr;
        }

        private ScalarExpression compileAndExpression()
        {
            ScalarExpression expr = compileEqualityExpression();

            while (p.parseChar('&', '&'))
            {
                expr = expr.and(compileEqualityExpression());
            }

            return expr;
        }

        private ScalarExpression compileEqualityExpression()
        {
            ScalarExpression expr = compileRelationalExpression();

            for (;;)
            {
                if (p.parseString("=="))
                {
                    expr = expr.eq(compileRelationalExpression());
                }
                else if (p.parseString("!="))
                {
                    expr = expr.noteq(compileRelationalExpression());
                }
                else if (p.parseString("="))
                {
                    // Assignment operator is invalid (user probably meant to specify "==")
                    throw new JDOUserException(LOCALISER.msg("JDOQL.InvalidAssignmentOperator", p.getInput()));
                }
                else
                {
                    break;
                }
            }

            return expr;
        }

        private ScalarExpression compileRelationalExpression()
        {
            ScalarExpression expr = compileAdditiveExpression();

            for (;;)
            {
                if (p.parseString("<="))
                {
                    expr = expr.lteq(compileAdditiveExpression());
                }
                else if (p.parseString(">="))
                {
                    expr = expr.gteq(compileAdditiveExpression());
                }
                else if (p.parseChar('<'))
                {
                    expr = expr.lt(compileAdditiveExpression());
                }
                else if (p.parseChar('>'))
                {
                    expr = expr.gt(compileAdditiveExpression());
                }
                else if (p.parseString("instanceof"))
                {
                    expr = expr.instanceOf(compileAdditiveExpression());
                }
                else if (p.parseString("AS") || p.parseString("as"))
                {
                    String asName = p.parseName();
                    expr = expr.as(asName);
                }
                else
                {
                    break;
                }
            }

            return expr;
        }

        private ScalarExpression compileAdditiveExpression()
        {
            ScalarExpression expr = compileMultiplicativeExpression();

            for (;;)
            {
                if (p.parseChar('+'))
                {
                    expr = expr.add(compileMultiplicativeExpression());
                }
                else if (p.parseChar('-'))
                {
                    expr = expr.sub(compileMultiplicativeExpression());
                }
                else
                {
                    break;
                }
            }

            return expr;
        }

        private ScalarExpression compileMultiplicativeExpression()
        {
            ScalarExpression expr = compileUnaryExpression();

            for (;;)
            {
                if (p.parseChar('*'))
                {
                    expr = expr.mul(compileUnaryExpression());
                }
                else if (p.parseChar('/'))
                {
                    expr = expr.div(compileUnaryExpression());
                }
                else if (p.parseChar('%'))
                {
                    expr = expr.mod(compileUnaryExpression());
                }
                else
                {
                    break;
                }
            }

            return expr;
        }

        private ScalarExpression compileUnaryExpression()
        {
            ScalarExpression expr;
            if (p.parseString("++"))
            {
                throw new JDOUserException("Unsupported operator '++'");
            }
            else if (p.parseString("--"))
            {
                throw new JDOUserException("Unsupported operator '--'");
            }               

            if (p.parseChar('+'))
            {
                expr = compileUnaryExpression();
            }
            else if (p.parseChar('-'))
            {
                expr = compileUnaryExpression().neg();
            }
            else
            {
                expr = compileUnaryExpressionNotPlusMinus();
            }
            return expr;
        }

        private ScalarExpression compileUnaryExpressionNotPlusMinus()
        {
            ScalarExpression expr;

            if (p.parseChar('~'))
            {
                expr = compileUnaryExpression().com();
            }
            else if (p.parseChar('!'))
            {
                expr = compileUnaryExpression().not();
            }
            else if ((expr = compileCastExpression()) == null)
            {
                expr = compilePrimary();
            }

            return expr;
        }

        private ScalarExpression compileCastExpression()
        {
            Class type;
            if ((type = p.parseCast(qs.getClassLoaderResolver(), getCandidateLoader())) == null)
            {
                return null;
            }

            return compileUnaryExpression().cast(type);
        }

        /**
         * this compiles a primary. First look for a literal (e.g. "text"), then
         * an identifier(e.g. variable) In the next step, call a function, if
         * executing a function, on the literal or the identifier found.
         * 
         * @return Scalar Expression for the primary compiled expression
         */
        private ScalarExpression compilePrimary()
        {
            // try to find a literal
            ScalarExpression expr = compileLiteral();

            // if expr == null, literal not found...
            if (expr == null)
            {
                //try to find an expression
                if (p.parseChar('('))
                {
                    expr = compileExpression();

                    if (!p.parseChar(')'))
                    {
                        throw new ExpressionSyntaxException("')' expected");
                    }
                    expr.encloseWithInParentheses();
                }
                //try to find an array
                else if (p.parseChar('{'))
                {
                    List exprs = new ArrayList();
                    while(!p.parseChar('}'))
                    {
                        exprs.add(compileExpression());
                        if(p.parseChar('}'))
                        {
                            break;
                        }
                        else if (!p.parseChar(','))
                        {
                            throw new ExpressionSyntaxException("',' or '}' expected");
                        }
                    }
                    expr = new ArrayExpression(qs,(ScalarExpression[])exprs.toArray(new ScalarExpression[exprs.size()]));
                }
                else
                {
                    String methodId = p.parseMethod();
                    if (methodId == null)
                    {
	                    /*
	                     * We will have an identifier
	                     * an identifier can be an variable, parameter or a field in the candidate class
	                     */
	                    expr = compileIdentifier();
                    }
                    else
                    {
                        // we are running arbitrary methods. the namespace here is "this",
                        // because "this" has no methods we assume we are dealing with
                        // aggregate methods
                        if (p.parseChar('('))
                        {
                            // Aggregate function expression
                            List args = new ArrayList();
                            boolean isDistinct = false;
                            if (!p.parseChar(')'))
                            {
                                // Check for use of DISTINCT/distinct
                                isDistinct = p.parseString("DISTINCT");
                                if (!isDistinct)
                                {
                                    isDistinct = p.parseString("distinct");
                                }

                                do
                                {
                                    ScalarExpression argExpr = compileExpression();
                                    args.add(argExpr);
                                    fieldExpressions.remove(argExpr); // Remove from field expressions list since we will include aggregates
                                } while (p.parseChar(','));

                                if (!p.parseChar(')'))
                                {
                                    throw new ExpressionSyntaxException("')' expected");
                                }
                            }
                            expr = new AggregateExpression(qs);
                            if (isDistinct)
                            {
                                ((AggregateExpression)expr).setDistinct();
                            }

                            // Allow for aggregate functions specified in uppercase [JDO2 spec 14.4]
                            expr = expr.callMethod(methodId.toLowerCase(), args);
                            fieldExpressions.add(expr); // Remove from field expressions list since we dont include aggregates
                        }
                    }
                }
            }

            /*
             * run function on literals or identifiers
             * e.g. "primary.runMethod(arg)"
             */
            while (p.parseChar('.'))
            {
                String id = p.parseIdentifier();
                if (id == null)
                {
                    throw new ExpressionSyntaxException("Identifier expected");
                }
                if (p.parseChar('('))
                {
                    // Found syntax for a method, so invoke the method
                    ArrayList args = new ArrayList();

                    if (!p.parseChar(')'))
                    {
                        do
                        {
                            args.add(compileExpression());
                        } while (p.parseChar(','));
 
                        if (!p.parseChar(')'))
                        {
                            throw new ExpressionSyntaxException("')' expected");
                        }
                    }

                    expr = expr.callMethod(id, args);
                }
                else
                {
                    // access the field from an identifier
                    expr = expr.accessField(id, false);
                }
            }
            return expr;
        }

        /**
         * A literal is one value of any type.
         * Supported literals are of types String, Floating Point, Integer,
         * Character, Boolean and null e.g. 'J', "String", 1, 1.8, true, false, null.
         * @return The compiled literal
         */
        private ScalarExpression compileLiteral()
        {
            Class  litType;
            Object litValue;

            String     sLiteral;
            BigDecimal fLiteral;
            BigInteger iLiteral;
            Boolean    bLiteral;

            boolean single_quote_next = p.nextIsSingleQuote();
            if ((sLiteral = p.parseStringLiteral()) != null)
            {
                // Both String and Character are allowed to use single-quotes
                // so we need to check if it was single-quoted and
                // use CharacterLiteral if length is 1.
                if (sLiteral.length() == 1 && single_quote_next)
                {
                    litType = Character.class;
                    litValue = new Character(sLiteral.charAt(0));
                }
                else
                {
                    litType  = String.class;
                    litValue = sLiteral;
                }
            }
            else if ((fLiteral = p.parseFloatingPointLiteral()) != null)
            {
                litType  = BigDecimal.class;
                litValue = fLiteral;
            }
            else if ((iLiteral = p.parseIntegerLiteral()) != null)
            {
                litType  = Long.class;
                litValue = new Long(iLiteral.longValue());
            }
            else if ((bLiteral = p.parseBooleanLiteral()) != null)
            {
                litType  = Boolean.class;
                litValue = bLiteral;
            }
            else if (p.parseNullLiteral())
            {
                return new NullLiteral(qs);
            }
            else
            {
                return null;
            }

            JavaTypeMapping m = dba.getMapping(litType, qs.getStoreManager(), qs.getClassLoaderResolver());
            return m.newLiteral(qs, litValue);
        }

        /**
         * An identifier always designates a reference to a single value.
         * A single value can be one collection, one field.
         * @return The compiled identifier
         */
        private ScalarExpression compileIdentifier()
        {
            ScalarExpression expr;

            String id = p.parseIdentifier();
	        if (id == null)
	        {
	            throw new ExpressionSyntaxException("Identifier expected");
	        }

            // Check if the identifier is a JDOQL keyword
            for (int i=0;i<SINGLE_STRING_KEYWORDS.length;i++)
            {
                if (id.equals(SINGLE_STRING_KEYWORDS[i]) || id.equals(SINGLE_STRING_KEYWORDS_LOWERCASE[i]))
                {
                    throw new ExpressionSyntaxException(LOCALISER.msg("JDOQL.IdentifierIsJDOQLKeyword", id));
                }
            }

            if (id.startsWith(":"))
            {
                id = id.substring(1); // Omit the ":"
                if (explicitParameters)
                {
                    // JDO2 14.6.3 Parameters should be either explicit, or implicit, but not both.
                    // The user has declared at least 1 parameter yet this new parameter is implicit, so throw an exception
                    throw new JDOUserException(LOCALISER.msg("JDOQL.MixedParametersInvalid", parameters, id));
                }

                if (parameters != null && parameters.size() > 0)
                {
                    if (parameters.containsKey(id))
                    {
                        // Implicit parameter already found, so reuse the value
                        Object paramValue = parameters.get(id);
                        if (paramValue != null)
                        {
                            JavaTypeMapping m = qs.getStoreManager().getDatastoreAdapter().getMapping(paramValue.getClass(), 
                                qs.getStoreManager(), qs.getClassLoaderResolver());
                            return m.newLiteral(qs, paramValue);
                        }
                        else
                        {
                            return new NullLiteral(qs);
                        }
                    }

                    if (parameters.size() < implicitParamNo+1)
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.ImplicitParameterWithNoValue", "" + id, "" + implicitParamNo));
                    }

                    // Replace the value key with the true name
                    if (!parameters.containsKey("JPOX_" + implicitParamNo))
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.ImplicitParameterWithNoValue", "" + id, "" + implicitParamNo));
                    }
                    Object paramValue = parameters.get("JPOX_" + implicitParamNo);
                    parameters.put(id, paramValue);
                    parameters.remove("JPOX_" + implicitParamNo);

                    implicitParamNo++;
                    if (paramValue != null)
                    {
                        JavaTypeMapping m = qs.getStoreManager().getDatastoreAdapter().getMapping(paramValue.getClass(), 
                            qs.getStoreManager(), qs.getClassLoaderResolver());
                        return m.newLiteral(qs, paramValue);
                    }
                    else
                    {
                        return new NullLiteral(qs);
                    }
                }
                else
                {
                    if (isPreCompile)
                    {
                        // We have no value available at precompile so just give it a null for now
                        return new NullLiteral(qs);
                    }
                    else
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.ImplicitParameterWithNoValue", "" + id, "" + implicitParamNo));
                    }
                }
            }

            if (id.equals("new"))
            {
                // Found syntax for "new MyObject(param1, param2)" - result expression
                String newClsName = p.parseName();
                Class newCls = null;
                try
                {
                    newCls = resolveClassDeclaration(newClsName);
                }
                catch (JDOUserException jdoe)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.NewObject.ClassNotFound", newClsName));
                }
                ArrayList args = new ArrayList();
                if (p.parseChar('('))
                {
                    if (!p.parseChar(')'))
                    {
                        do
                        {
                            ScalarExpression argExpr = compileExpression();
                            args.add(argExpr);
                            fieldExpressions.add(argExpr); // Add to field expressions list
                        } while (p.parseChar(','));
                        if (!p.parseChar(')'))
                        {
                            throw new ExpressionSyntaxException("')' expected, but was "+p.remaining());
                        }
                    }
                }
                else
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.NewObject.InvalidSyntax", getSingleStringQuery()));
                }
                expr = new NewObjectExpression(qs, newCls, args);
            }
            else if (parameterNames.contains(id))
            {
                // Identifier is an explicit parameter
                Class parameterClass = (Class)parameterTypesByName.get(id);
                boolean serialised = false;
                if (parameterClass == java.lang.Object.class)
                {
                    serialised = true; // Don't yet support Object as non-serialised. TODO Remove this when available
                }

                JavaTypeMapping m = dba.getMapping(parameterClass, serialised, false, qs.getStoreManager(), qs.getClassLoaderResolver());
                if (isPreCompile)
                {
                    expr = m.newLiteral(qs, m.getSampleValue(this.qs.getClassLoaderResolver()));
                }
                else if (parameters == null || !parameters.containsKey(id))
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.RequiredParameterNotProvidedError",id));
                }
                else
                {
                    Object parameterValue = parameters.get(id);
                    if (parameterValue == null)
                    {
                        expr = new NullLiteral(qs);
                    }
                    else
                    {
                        expr = m.newLiteral(qs, parameterValue);
                    }
                }
            }
            else if (variableNames.contains(id))
            {
                // Identifier is an explicit variable
                expr = (ScalarExpression)expressionsByVariableName.get(id);
                if (expr == null)
                {
                    expr = new UnboundVariable(qs, id, (Class)variableTypesByName.get(id), this);
                }
                fieldExpressions.add(expr); // Add to the field expressions list
            }
            else
            {
                try
                {
                    // Check if the identifier is a field of the candidate class
                    expr = qs.getDefaultTableExpression().newFieldExpression(id);

                    if (!id.equals("this"))
                    {
                        // Make a check on whether the field exists in the candidate class (TableExpression only checks the same table).
                        if (candidateCmd == null)
                        {
                            candidateCmd = pm.getMetaDataManager().getMetaDataForClass(candidateClass,pm.getClassLoaderResolver());
                        }
                        if (candidateCmd.getField(id) == null)
                        {
                            // The field doesnt exist in the candidate class, so no point proceeding since prohibited by the JDO spec
                            throw new JDOUserException(LOCALISER.msg("JDOQL.SpecifiedFieldNotFound", id));
                        }
                    }

                    fieldExpressions.add(expr); // Add to the field expressions list
                }
                catch (NoSuchPersistentFieldException nspfe)
                {
                    // Not a field so retrieve the full name and it is either a class, or an implicit variable
                    String name = id;
                    if (p.nextIsDot())
                    {
                        p.parseChar('.');
                        name += ".";
                        name += p.parseName();
                    }

                    Class cls = null;
                    try
                    {
                        cls = resolveClassDeclaration(name);
                        expr = new ClassExpression(qs, cls);
                    }
                    catch (JDOUserException jdoe)
                    {
                        if (name.indexOf('.') > 0)
                        {
                            // Try without the last part of the name (in case its a class name plus field or method)
                            String partialName = name.substring(0, name.lastIndexOf('.'));
                            String finalNamePart = name.substring(name.lastIndexOf('.')+1);
                            try
                            {
                                //try method invocation
                                expr = callUserDefinedScalarExpression(name);
                                if( expr == null )
                                {
                                    cls = resolveClassDeclaration(partialName);
                                    //try field access
                                    expr = new ClassExpression(qs, cls);
                                    expr = expr.accessField(finalNamePart, true);
                                }
                            }
                            catch (JDOUserException jdoe2)
                            {
                                throw new JDOUserException(LOCALISER.msg("JDOQL.SpecifiedClassNotFound", partialName),jdoe2);
                            }
                        }
                        else
                        {
                            try
                            {
                                //try with candidate class
                                //try field access
                                expr = new ClassExpression(qs, candidateClass);
                                expr = expr.accessField(name, true);
                            }
                            catch (JDOUserException jdoe2)
                            {
                                // Implicit variable
                                if (explicitVariables)
                                {
                                    throw new JDOUserException(LOCALISER.msg("JDOQL.MixedVariablesInvalid", JDOQLQuery.this.variables, name));
                                }
                               // if (!implicitParameter)
                               //// {
                                  //  throw new JDOUserException(LOCALISER.msg("JDOQL.SpecifiedFieldNotFound", id));
                                //}
                                expr = (ScalarExpression)expressionsByVariableName.get(name);
                                if (expr == null)
                                {
                                    // Type will be null here since we have no types specified for implicit variables
                                    // The type will be set later when we know its context
                                    expr = new UnboundVariable(qs, name, (Class)variableTypesByName.get(name), this);
                                    variableNames.add(name);
                                    // We should really add this to variableTypesByName but we don't know the type here
                                    fieldExpressions.add(expr); // Add to the field expressions list
                                }
                            }
                        }
                    }
                }
            }

            return expr;
        }

        /**
         * Instanciate a ScalarExpression and invoke a method
         * @param method the method name prefixed by the class name (fully qualified or not) 
         * @return the ScalarExpression instance
         */
        private ScalarExpression callUserDefinedScalarExpression(String method)
        {
            String className = method.substring(0, method.lastIndexOf('.'));
            String methodName = method.substring(method.lastIndexOf('.')+1);
            if( !userDefinedScalarExpressions.containsKey(className) )
            {
                Class cls = resolveClassDeclaration(className);
                className = cls.getName();
            }
            if( userDefinedScalarExpressions.containsKey(className) )
            {
                ScalarExpression expr = newScalarExpression( (Class) userDefinedScalarExpressions.get(className) );
                if (p.parseChar('('))
                {
                    // Found syntax for a method, so invoke the method
                    ArrayList args = new ArrayList();

                    if (!p.parseChar(')'))
                    {
                        do
                        {
                            args.add(compileExpression());
                        } while (p.parseChar(','));
 
                        if (!p.parseChar(')'))
                        {
                            throw new ExpressionSyntaxException("')' expected");
                        }
                    }

                    return expr.callMethod(methodName, args);
                }
            }
            return null;
        }
        
        /**
         * Create a ScalarExpression instance for the given <code>cls</code>
         * @param cls the class that extends ScalarExpression 
         * @return the ScalarExpression instance
         */
        private ScalarExpression newScalarExpression(Class cls)
        {
            try
            {
                return (ScalarExpression) cls.getConstructor(new Class[]{QueryExpression.class}).newInstance(new Object[] {qs});
            }
            catch (IllegalArgumentException e)
            {
                throw new JDOFatalInternalException("Cannot create ScalarExpression for class "+cls.getName()+" due to "+e.getMessage(),e);
            }
            catch (SecurityException e)
            {
                throw new JDOFatalInternalException("Cannot create ScalarExpression for class "+cls.getName()+" due to "+e.getMessage(),e);
            }
            catch (InstantiationException e)
            {
                throw new JDOFatalInternalException("Cannot create ScalarExpression for class "+cls.getName()+" due to "+e.getMessage(),e);
            }
            catch (IllegalAccessException e)
            {
                throw new JDOFatalInternalException("Cannot create ScalarExpression for class "+cls.getName()+" due to "+e.getMessage(),e);
            }
            catch (InvocationTargetException e)
            {
                throw new JDOFatalInternalException("Cannot create ScalarExpression for class "+cls.getName()+" due to "+e.getMessage(),e);
            }
            catch (NoSuchMethodException e)
            {
                throw new JDOFatalInternalException("Cannot create ScalarExpression for class "+cls.getName()+" due to "+e.getMessage(),e);
            }
        }
        
        /**
         * Inner class representing an exception in the syntax handling.
         **/
        private class ExpressionSyntaxException extends JDOUserException
        {
            /**
             * Constructor
             * @param msg error message
             */
            public ExpressionSyntaxException(String msg)
            {
                super(msg + " at character " + (p.getIndex() + 1) + " in \"" + p.getInput() + '"');
            }
        }
    }

    /**
     * Stringifier method
     * @return Single-string form of this JDOQL query.
     */
    public String toString()
    {
        return getSingleStringQuery();
    }
}