/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Andy Jefferson - commented and localised
2005 Andy Jefferson - added timeout support
2005 Andy Jefferson - added support for use without candidate class
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.exceptions.ClassNotPersistenceCapableException;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.metadata.IdentityType;
import org.jpox.store.DatastoreField;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryResult;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.exceptions.PersistentSuperclassNotAllowedException;
import org.jpox.util.MacroString;

/**
 * A Query using SQL, and allowing specification of parameter names and imports,
 * known internally as "JPOXSQL", an extension to JDO.
 * @see Query
 * @version $Revision: 1.10 $
 */
public class JPOXSQLQuery extends BaseSQLQuery
{
    protected transient List parameterOccurrences = null;

    protected transient List fieldColumnNames = null;

    protected transient int[] fieldNumbers = null;

    /**
     * Constructor for a new query using the existing query.
     * @param pm Persistence Manager
     * @param query The existing query
     */
    public JPOXSQLQuery(PersistenceManager pm, JPOXSQLQuery query)
    {
        super(pm, query);
    }

    /**
     * Constructs a new query instance having the same criteria as the given
     * query.
     * @param pm The PersistenceManager
     */
    public JPOXSQLQuery(PersistenceManager pm)
    {
        super(pm, (String)null);
    }

    /**
     * Constructs a new query instance having the same criteria as the given
     * query.
     * @param pm The PersistenceManager
     * @param sql_text The JPOX SQL query string
     */
    public JPOXSQLQuery(PersistenceManager pm, String sql_text)
    {
        super(pm, sql_text);
    }

    protected void discardCompiled()
    {
        super.discardCompiled();
        parameterOccurrences = null;
        fieldColumnNames = null;
        fieldNumbers = null;
    }

    /**
     * Equality operator
     * @param obj The object to compare against
     * @return Whether they are equal
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof JPOXSQLQuery) || !super.equals(obj))
        {
            return false;
        }

        return sqlText.equals(((JPOXSQLQuery)obj).sqlText);
    }

    /**
     * Method to perform any necessary pre-processing on the users query statement
     * before we execute it.
     */
    protected void generateQueryStatement()
    {
        final AbstractClassMetaData candidateCmd;
        if (candidateClass == null)
        {
            // We have no candidate so don't need to bother about aligning the return columns to our mappings.
            fieldColumnNames = new ArrayList(); // Empty
            candidateCmd = null;
        }
        else
        {
            candidateCmd = pm.getMetaDataManager().getMetaDataForClass(candidateClass,pm.getClassLoaderResolver());
            if (candidateCmd == null)
            {
                throw new ClassNotPersistenceCapableException(candidateClass.getName());
            }
            if (candidateCmd.getPersistenceCapableSuperclass() != null)
            {
                throw new PersistentSuperclassNotAllowedException(candidateClass.getName());
            }
            if (candidateCmd.isRequiresExtent())
            {
                throw new JDOUserException(LOCALISER_RDBMS.msg("JPOXSQL.InvalidCandidateClassExtentError",candidateClass.getName()));
            }
            if (candidateCmd.getIdentityType() != IdentityType.NONDURABLE)
            {
                throw new JDOUserException(LOCALISER_RDBMS.msg("JPOXSQL.InvalidCandidateClassNondurableError",candidateClass.getName()));
            }

            int fieldCount = candidateCmd.getNoOfManagedFields();
            int[] fn = new int[fieldCount];
            statementExpressionIndex = new StatementExpressionIndex[fieldCount];
            fieldColumnNames = new ArrayList(fieldCount);
            int n = 0;

            for (int fieldNumber = 0; fieldNumber < fieldCount; ++fieldNumber)
            {
                statementExpressionIndex[fieldNumber] = new StatementExpressionIndex();
                AbstractPropertyMetaData fmd = candidateCmd.getManagedField(fieldNumber);
                String fieldName = fmd.getName();
                Class fieldType = fmd.getType();

                if (fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
                {
                    JavaTypeMapping m = dba.getMapping(fieldType, getStoreManager(), pm.getClassLoaderResolver());
                    if (m.includeInFetchStatement())
                    {
                        statementExpressionIndex[fieldNumber].setMapping(m);
                        fn[n++] = fieldNumber;

                        String columnName = null;
                        if (fmd.getColumnMetaData() != null && fmd.getColumnMetaData().length > 0)
                        {
                            columnName = fmd.getColumnMetaData()[0].getName();
                        }
                        else
                        {
                            columnName = pm.getStoreManager().getIdentifierFactory().newDatastoreFieldIdentifier(fieldName, 
                                pm.getPMFContext().getTypeManager().isDefaultEmbeddedType(fieldType), DatastoreField.ROLE_NONE).getIdentifier();
                        }
                        fieldColumnNames.add(columnName);
                    }

                    if (m instanceof MappingCallbacks)
                    {
                        // TODO Remove this. Why is it needed ?
                        throw new JDOFatalUserException("Mapping " + m +
                            " not suitable for a JPOXSQL result column, field = " + fieldName);
                    }
                }
                else if (fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL)
                {
                    throw new JDOFatalInternalException("Invalid persistence modifier on field " + fieldName);
                }
            }

            if (n == 0)
            {
                throw new JDOFatalUserException(
                    "View class has no persistent fields: " + candidateClass.getName());
            }

            fieldNumbers = new int[n];
            System.arraycopy(fn, 0, fieldNumbers, 0, n);
        }

        // Generate the actual JDBC SQL text by processing the embedded parameter/field macros
        // in the provided JPOXSQL text.
        parameterOccurrences = new ArrayList();
        MacroString ms = new MacroString(candidateClass != null ? candidateClass.getName() : null, 
                                         candidateClass != null ? imports : null, sqlText);
        jdbcSqlText = ms.substituteMacros(new MacroString.MacroHandler()
            {
            public void onIdentifierMacro(MacroString.IdentifierMacro im)
            {
                // Only process identifier macros when the candidate class is known
                if (candidateClass != null)
                {
                    if (im.className.equals(candidateClass.getName()))
                    {
                        if (im.fieldName == null)
                        {
                            throw new JDOUserException(LOCALISER_RDBMS.msg("JPOXSQL.ResultClassesHaveNoTableError",im));
                        }
                        if (im.subfieldName != null)
                        {
                            throw new JDOUserException(LOCALISER_RDBMS.msg("JPOXSQL.NoSuchFieldInResultClassError",im.className,im));
                        }
                        int fieldNumber = candidateCmd.getFieldNumber(im.fieldName);
                        
                        if (fieldNumber < 0)
                        {
                            throw new JDOUserException(LOCALISER_RDBMS.msg("JPOXSQL.NoSuchFieldInResultClassError",im.className,im));
                        }
                        im.value = (String) fieldColumnNames.get(fieldNumber);
                    }
                    else
                    {
                        getStoreManager().resolveIdentifierMacro(im, pm.getClassLoaderResolver());
                    }
                }
            }
            
            public void onParameterMacro(MacroString.ParameterMacro pm)
            {
                parameterOccurrences.add(pm.parameterName);
            }
            }, pm.getClassLoaderResolver());
    }

    /**
     * Execute the query and return the filtered QueryResult.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult
     */
    public Collection performExecute(Map parameters)
    {
        compile();

        if (parameters.size() != parameterNames.size())
        {
            throw new JDOUserException(LOCALISER_RDBMS.msg("JPOXSQL.IncorrectNumberOfParametersError","" + parameterNames.size(),"" + parameters.size()));
        }

        QueryResult qr = null;
        try
        {
            RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
            Connection conn = storeMgr.getConnection(pm, false, true);

            try
            {
                PreparedStatement ps = conn.prepareStatement(jdbcSqlText);
                try
                {
                    Iterator iter = parameterOccurrences.iterator();
                    int stmtParamNum = 1;

                    while (iter.hasNext())
                    {
                        String paramName = (String) iter.next();
                        Class paramType = (Class) parameterTypesByName.get(paramName);

                        if (!parameters.containsKey(paramName))
                        {
                            throw new JDOUserException(LOCALISER_RDBMS.msg("JPOXSQL.ParameterNotProvidedError",paramName));
                        }
                        if (paramType == null)
                        {
                            throw new JDOUserException(LOCALISER_RDBMS.msg("JPOXSQL.UndeclaredParameterError",paramName));
                        }

                        JavaTypeMapping mapping = dba.getMapping(paramType, getStoreManager(),pm.getClassLoaderResolver() );
                        Object paramValue = parameters.get(paramName);

                        mapping.setObject(pm, ps, Mappings.getParametersIndex(stmtParamNum,mapping), paramValue);
                        if (mapping.getNumberOfDatastoreFields() == 0)
                        {
                            stmtParamNum++;
                        }
                        else
                        {
                            stmtParamNum += mapping.getNumberOfDatastoreFields();
                        }
                    }

                    // Apply any user-specified constraints over timeouts and ResultSet
                    prepareStatementForExecution(ps);

                    // Execute the query
                    ResultSet rs = storeMgr.executeStatementQuery(jdbcSqlText, ps);
                    try
                    {
                        if (candidateClass != null)
                        {
                            ResultSetMetaData rsmd = rs.getMetaData();
                            HashSet remainingColumnNames = new HashSet(fieldColumnNames);

                            int colCount = rsmd.getColumnCount();
                            for (int colNum = 1; colNum <= colCount; ++colNum)
                            {
                                String colName = rsmd.getColumnName(colNum);
                                int fieldNumber = fieldColumnNames.indexOf(colName);
                                if (fieldNumber >= 0)
                                {
                                    statementExpressionIndex[fieldNumber].setExpressionIndex(new int[]{colNum});
                                    remainingColumnNames.remove(colName);
                                }
                            }
                            
                            if (!remainingColumnNames.isEmpty())
                            {
                                throw new JDOUserException(LOCALISER_RDBMS.msg("JPOXSQL.ResultSetMissingColumnsError",remainingColumnNames));
                            }

                            if (resultClass != null)
                            {
                                if (getResultSetType().equals("scroll-insensitive"))
                                {
                                    qr = new InsensitiveQueryResult(null, this, new ResultClassROF(resultClass, statementExpressionIndex), rs, null);
                                }
                                else
                                {
                                    qr = new ForwardQueryResult(null, this, new ResultClassROF(resultClass, statementExpressionIndex), rs, null);
                                }
                            }
                            else
                            {
                                if (getResultSetType().equals("scroll-insensitive"))
                                {
                                    qr = new InsensitiveQueryResult(null, this, new ResultClassROF(resultClass, statementExpressionIndex), rs, null);
                                }
                                else
                                {
                                    qr = new ForwardQueryResult(null, this, new TransientIDROF(candidateClass, fieldNumbers, statementExpressionIndex), rs, null);
                                }
                            }
                        }
                        else
                        {
                            qr = getQueryResultForNoCandidateClass(rs);
                        }
                    }
                    finally
                    {
                        if (qr == null)
                        {
                            rs.close();
                        }
                    }
                }
                finally
                {
                    if (qr == null)
                    {
                        ps.close();
                    }
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER_RDBMS.msg("JPOXSQL.ExecutionError", jdbcSqlText), e);
        }

        queryResults.add(qr);

        return qr;
    }
}