/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Erik Bengtson - removed subclasses operation
2003 Andy Jefferson - comments, and update to returned class name
2003 Erik Bengtson - added getObjectByAid
2004 Erik Bengtson - throws an JDOObjectNotFoundException
2004 Erik Bengtson - removed unused variable and import
2004 Erik Bengtson - added support for ignoreCache
2004 Andy Jefferson - coding standards
2005 Andy Jefferson - added support for using discriminator to distinguish objects 
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.sql.ResultSet;
import java.util.Map;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.InterfaceMetaData;
import org.jpox.store.DatastoreClass;
import org.jpox.store.FieldValues;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.OIDMapping;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.store.rdbms.fieldmanager.ResultSetGetter;
import org.jpox.util.SoftValueMap;

/**
 * Class to take a JDBC ResultSet and create a PersistenceCapable object instance
 * for each row in the ResultSet. Each row in the ResultSet represents an instance of
 * a particular object type. We can determine the object type in one of two ways :
 * <UL>
 * <LI><B>discriminator</B> - when the candidate class and all possible subclasses in the
 * query are stored in the same table they are distinguished using a "discriminator" column
 * that will either store the class name, or a value representing that type.</LI>
 * <LI><B>JPOXMETADATA</B> - in all other cases JPOX adds an extra column to the SELECT
 * that contains the class name of the object.</LI>
 * </UL>
 *
 * @version $Revision: 1.2.2.1 $
 */
public final class PersistentIDROF implements ResultObjectFactory
{
    /**
     * Table where the objects of the candidate class are stored.
     * If the candidate uses "subclass-table" inheritance then this will be one or the other and may cause issues!
     */
    private final DatastoreClass table;

    /** Whether to ignore the cache */
    private final boolean ignoreCache;

    /** FIeld numbers to populate. */
    protected int[] fieldNumbers;

    /** Mapping from ResultSet to field numbers and mappings. */
    protected StatementExpressionIndex[] statementExpressionIndex;

    /** Indices of datastore identity columns */
    protected final int[] datastoreIdentityExpressionIndex;

    /** Index of version column(s). */
    protected final int[] versionIndex;

    /** Flag whether we should use a discriminator column to distinguish object types (otherwise use JPOXMETADATA). */
    protected boolean discriminator;
    
    /** if the results have included a meta data column with the class name **/
    protected final boolean hasMetaDataInResults;

    /** Fetch Plan to use when loading fields (if any). */
    protected final FetchPlanImpl fetchPlan;
    
    /** fieldnumbers taken from this MetaData **/
    protected final AbstractClassMetaData acmd;

    /** Resolved classes for metadata / discriminator keyed by class names **/
    private Map resolvedClasses = new SoftValueMap();

    /** StatementExpressionIndex keyed by full class name **/
    private Map stmtExprIndexes = new SoftValueMap();
    
    /**
     * Constructor.
     * @param table Table being selected from
     * @param fieldNumbers Numbers of the fields in the select. May be null if no fields are retrieved
     * @param statementExpressionIndex Index of columns to statement columns
     * @param datastoreIdentityExpressionIndex index for datastore identity
     * @param versionIndex Indices of the version column(s)
     * @param ignoreCache Whether to ignore the cache
     * @param discriminator Whether we use a discriminator column to distinguish object 
     *                      types, or whether we use a JPOXMETADATA column
     * @param hasMetaDataInResults whether we use JPOXMETADATA column    
     * @param fetchPlan the Fetch Plan
     */
    public PersistentIDROF(DatastoreClass table,
                           int[] fieldNumbers,
                           AbstractClassMetaData acmd,
                           StatementExpressionIndex[] statementExpressionIndex,
                           int[] datastoreIdentityExpressionIndex,
                           int[] versionIndex,
                           boolean ignoreCache,
                           boolean discriminator,
                           boolean hasMetaDataInResults,
                           FetchPlanImpl fetchPlan)
    {
        this.statementExpressionIndex = statementExpressionIndex;
        if( fieldNumbers == null )
        {
            this.fieldNumbers = new int[0];
        }
        else
        {
            this.fieldNumbers = fieldNumbers;
        }
        this.acmd = acmd;
        this.table = table;
        this.datastoreIdentityExpressionIndex = datastoreIdentityExpressionIndex;
        this.versionIndex = versionIndex;
        this.ignoreCache = ignoreCache;
        this.discriminator = discriminator;
        this.hasMetaDataInResults = hasMetaDataInResults;
        this.fetchPlan = fetchPlan;
    }

    /**
     * Method to convert the current ResultSet row into an Object.
     * @param pm The PersistenceManager
     * @param rs The ResultSet from the Query.
     * @param pcClass The candidate class that we are looking for (may be base class).
     * @return The (persisted) object.
     **/
    public Object getObject(final PersistenceManager pm, final Object rs, Class pcClass)
    {
        Class pcClassForObject = pcClass;
        boolean requiresInheritanceCheck = true;
        // PC, stored in their own table (or multiple tables)
        String className = null;
        if (discriminator)
        {
            className = QueryUtils.getClassNameFromDiscriminatorResultSetRow(table, (ResultSet)rs, pm);
        }
        else if (hasMetaDataInResults)
        {
            // Extract the object type using the JPOXMETADATA column (if available)
            try
            {
                className = QueryUtils.getClassNameFromJPOXMetaDataResultSetRow((ResultSet)rs);
            }
            catch (Exception e)
            {
                // Do nothing
            }
        }
        ClassLoaderResolver clr = pm.getClassLoaderResolver();
        if (className != null)
        {
            Class cls = (Class) resolvedClasses.get(className);

            if( cls != null )
            {
                pcClassForObject = cls;                    
            }
            else
            {
                if( pcClass.getName().equals(className) )
                {
                    pcClassForObject = pcClass;                    
                }
                else
                {
                    pcClassForObject = clr.classForName(className, pcClass.getClassLoader());
                }
                resolvedClasses.put(className, pcClassForObject);
            }
            requiresInheritanceCheck = false;
        }
        
        int[] fieldNumbers = new int[this.fieldNumbers.length];
        AbstractClassMetaData cmd = pm.getMetaDataManager().getMetaDataForClass(pcClassForObject, clr);
        for (int i=0; i<this.fieldNumbers.length; i++)
        {
            if( acmd instanceof InterfaceMetaData )
            {
                fieldNumbers[i] = cmd.getFieldNumberAbsolute(acmd.getManagedFieldAbsolute(this.fieldNumbers[i]).getName());
            }
            else
            {
                 fieldNumbers[i] = cmd.getFieldNumberAbsolute(acmd.getManagedFieldAbsolute(this.fieldNumbers[i]).getClassName(),acmd.getManagedFieldAbsolute(this.fieldNumbers[i]).getName());
                 if( fieldNumbers[i]==-1 )
                 {
                     fieldNumbers[i] = cmd.getFieldNumberAbsolute(acmd.getManagedFieldAbsolute(this.fieldNumbers[i]).getName());
                 }
            }
        }
        Object obj = null;
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            obj = getObjectByAID(pm, (ResultSet)rs, pcClassForObject, fieldNumbers, cmd, requiresInheritanceCheck);
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            JavaTypeMapping mapping = new OIDMapping();
            mapping.initialize(table.getStoreManager().getDatastoreAdapter(), pcClassForObject.getName());
            mapping.addDataStoreMapping(table.getDataStoreObjectIdMapping().getDataStoreMapping(0));
            Object id = mapping.getObject(pm, rs, datastoreIdentityExpressionIndex);
            if (fieldNumbers == null)
            {
                obj = pm.getObjectById(id, false, requiresInheritanceCheck, null);
            }
            else
            {
            	if (requiresInheritanceCheck)
            	{
                    obj = getObjectById(pm, id, (ResultSet)rs, fieldNumbers, cmd);
            	}
            	else
            	{
            		obj = getObjectById(pm, id, (ResultSet)rs, pcClassForObject, fieldNumbers, cmd);
            	}
            }
        }
        else if (cmd.getIdentityType() == IdentityType.NONDURABLE)
        {
            Object id = null;
            if (fieldNumbers == null)
            {
                obj = pm.getObjectById(id, false, requiresInheritanceCheck, null);
            }
            else
            {
                obj = getObjectById(pm, id, (ResultSet)rs, pcClassForObject, fieldNumbers, cmd);
            }
        }

        if (versionIndex != null)
        {
            // Apply the version if required
            JavaTypeMapping mapping = table.getVersionMapping();
            Object version = mapping.getObject(pm, rs, versionIndex);
            StateManager objSM = pm.findStateManager((PersistenceCapable)obj);
            objSM.setVersion(version);
        }

        return obj;
    }

    /**
     * Returns a PC instance from an Application ID 
     * @param rs JDBC ResultSet
     * @param pcClass PersistenceCapable class
     * @return The object with this Application ID
     */
    private Object getObjectByAID(final PersistenceManager pm, final ResultSet rs, Class pcClass, final int[] fieldNumbers, AbstractClassMetaData cmd, boolean requiresInheritanceCheck)
    {
        StatementExpressionIndex[] indexes = (StatementExpressionIndex[]) stmtExprIndexes.get(cmd.getFullClassName());
        if( indexes == null )
        {
        	indexes = new StatementExpressionIndex[cmd.getNoOfManagedFields()+cmd.getNoOfInheritedManagedFields()];
            for (int i=0; i<fieldNumbers.length; i++)
            {
            	indexes[fieldNumbers[i]] = statementExpressionIndex[this.fieldNumbers[i]]; 
            }
            stmtExprIndexes.put(cmd.getFullClassName(), indexes);
        }
        final StatementExpressionIndex[] stmtExprIndex = indexes;            
        return pm.getObjectByAID(pcClass, new FieldValues()
        {
            // StateManager calls the fetchFields method
            public void fetchFields(StateManager sm)
            {
                sm.replaceFields(fieldNumbers, new ResultSetGetter(sm, rs, stmtExprIndex), false);
            }
            public void fetchNonLoadedFields(StateManager sm)
            {
                sm.replaceNonLoadedFields(fieldNumbers, new ResultSetGetter(sm, rs, stmtExprIndex));
            }
            public FetchPlanImpl getFetchPlanForLoading()
            {
                return fetchPlan;
            }
        }, ignoreCache, requiresInheritanceCheck);
    }

    /**
     * Returns a PC instance from an Datastore ID 
     * @param oid The object id
     * @param rs The ResultSet
     * @param pcClass The PersistenceCapable class
     * @return The Object
     */
    private Object getObjectById(final PersistenceManager pm, Object oid, final ResultSet rs, Class pcClass, final int[] fieldNumbers, AbstractClassMetaData cmd)
    {
        StatementExpressionIndex[] indexes = (StatementExpressionIndex[]) stmtExprIndexes.get(cmd.getFullClassName());
        if( indexes == null )
        {
            indexes = new StatementExpressionIndex[cmd.getNoOfManagedFields()+cmd.getNoOfInheritedManagedFields()];
            for (int i=0; i<fieldNumbers.length; i++)
            {
                indexes[fieldNumbers[i]] = statementExpressionIndex[this.fieldNumbers[i]]; 
            }
            stmtExprIndexes.put(cmd.getFullClassName(), indexes);
        }
        final StatementExpressionIndex[] stmtExprIndex = indexes;            
        return pm.getObjectById(oid, new FieldValues()
        {
            // StateManager calls the fetchFields method
            public void fetchFields(StateManager sm)
            {
                sm.replaceFields(fieldNumbers, new ResultSetGetter(sm, rs, stmtExprIndex), false);
            }
            public void fetchNonLoadedFields(StateManager sm)
            {
                sm.replaceNonLoadedFields(fieldNumbers, new ResultSetGetter(sm, rs, stmtExprIndex));
            }
            public FetchPlanImpl getFetchPlanForLoading()
            {
                return fetchPlan;
            }
        }, pcClass, ignoreCache);
    }
    
    /**
     * Returns a PC instance from an Datastore ID 
     * @param oid The object id
     * @param rs The ResultSet
     * @param pcClass The PersistenceCapable class
     * @return The Object
     */
    private Object getObjectById(final PersistenceManager pm, Object oid, final ResultSet rs, final int[] fieldNumbers, AbstractClassMetaData cmd)
    {
        StatementExpressionIndex[] indexes = (StatementExpressionIndex[]) stmtExprIndexes.get(cmd.getFullClassName());
        if( indexes == null )
        {
            indexes = new StatementExpressionIndex[cmd.getNoOfManagedFields()+cmd.getNoOfInheritedManagedFields()];
            for (int i=0; i<fieldNumbers.length; i++)
            {
                indexes[fieldNumbers[i]] = statementExpressionIndex[this.fieldNumbers[i]]; 
            }
            stmtExprIndexes.put(cmd.getFullClassName(), indexes);
        }
        final StatementExpressionIndex[] stmtExprIndex = indexes;            
        return pm.getObjectById(oid, new FieldValues()
        {
            // StateManager calls the fetchFields method
            public void fetchFields(StateManager sm)
            {
                sm.replaceFields(fieldNumbers, new ResultSetGetter(sm, rs, stmtExprIndex), false);
            }
            public void fetchNonLoadedFields(StateManager sm)
            {
                sm.replaceNonLoadedFields(fieldNumbers, new ResultSetGetter(sm, rs, stmtExprIndex));
            }
            public FetchPlanImpl getFetchPlanForLoading()
            {
                return fetchPlan;
            }
        });
    }    
}