/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.store.query.Query;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.util.TypeConversionHelper;

/**
 * Extension of Query providing for queries for RDBMS datastores.
 * @version $Revision: 1.2 $
 */
public abstract class RDBMSQuery extends Query
{
    /**
     * Constructs a new query instance that uses the given persistence manager.
     * @param pm The Persistence Manager for this query.
     */
    public RDBMSQuery(PersistenceManager pm)
    {
        super(pm);
    }

    /**
     * Accessor for the result set type.
     * Checks both the PMF, and also the query extensions.
     * @return The result set type string
     */
    protected String getResultSetType()
    {
        String rsTypeString = pm.getPMFContext().getPmfConfiguration().getQueryResultSetType();
        if (extensions != null && extensions.containsKey("org.jpox.query.resultSetType"))
        {
            rsTypeString = (String)extensions.get("org.jpox.query.resultSetType");
        }
        return rsTypeString;
    }

    /**
     * Accessor for the result set concurrency.
     * Checks both the PMF, and also the query extensions.
     * @return The result set concurrency string
     */
    protected String getResultSetConcurrency()
    {
        String rsConcurrencyString = pm.getPMFContext().getPmfConfiguration().getQueryResultSetConcurrency();
        if (extensions != null && extensions.containsKey("org.jpox.query.resultSetConcurrency"))
        {
            rsConcurrencyString = (String)extensions.get("org.jpox.query.resultSetConcurrency");
        }
        return rsConcurrencyString;
    }

    /**
     * Method to create a PreparedStatement for use with the query.
     * @param conn the Connection
     * @param queryStmt The statement text for the query
     * @return the PreparedStatement
     * @throws SQLException Thrown if an error occurs creating the statement
     */
    protected PreparedStatement getStatement(Connection conn, String queryStmt)
    throws SQLException
    {
        // Apply any non-standard result set definition if required (either from the PMF, or via query extensions)
        String rsTypeString = getResultSetType();
        if (rsTypeString != null &&
            (!rsTypeString.equals("scroll-sensitive") && !rsTypeString.equals("forward-only") && !rsTypeString.equals("scroll-insensitive")))
        {
            throw new JDOUserException("JDOQL extension for JPOX org.jpox.query.resultSetType has valid values of scroll-sensitive,scroll-insensitive,forward-only only.");
        }
        String rsConcurrencyString = getResultSetConcurrency();
        if (rsConcurrencyString != null &&
            (!rsConcurrencyString.equals("read-only") && !rsConcurrencyString.equals("updateable")))
        {
            throw new JDOUserException("JDOQL extension for JPOX org.jpox.query.resultSetConcurrency has valid values of read-only,updateable only.");
        }

        RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
        PreparedStatement ps = storeMgr.getStatement(conn, queryStmt, true, rsTypeString, rsConcurrencyString);

        return ps;
    }

    /**
     * Method to create a PreparedStatement for use with the query.
     * @param conn the Connection
     * @param stmtText The statement text
     * @return the PreparedStatement
     * @throws SQLException Thrown if an error occurs creating the statement
     */
    protected PreparedStatement getStatement(Connection conn, StatementText stmtText)
    throws SQLException
    {
        // Apply any non-standard result set definition if required (either from the PMF, or via query extensions)
        String rsTypeString = getResultSetType();
        if (rsTypeString != null &&
            (!rsTypeString.equals("scroll-sensitive") && !rsTypeString.equals("forward-only") && !rsTypeString.equals("scroll-insensitive")))
        {
            throw new JDOUserException("JDOQL extension for JPOX org.jpox.query.resultSetType has valid values of scroll-sensitive,scroll-insensitive,forward-only only.");
        }
        String rsConcurrencyString = getResultSetConcurrency();
        if (rsConcurrencyString != null &&
            (!rsConcurrencyString.equals("read-only") && !rsConcurrencyString.equals("updateable")))
        {
            throw new JDOUserException("JDOQL extension for JPOX org.jpox.query.resultSetConcurrency has valid values of read-only,updateable only.");
        }

        PreparedStatement ps = stmtText.prepareStatement(pm, conn, rsTypeString, rsConcurrencyString);

        return ps;
    }

    /**
     * Method to apply any query timeouts, and to add any restrictions to the created ResultSet.
     * @param ps The PreparedStatement
     * @throws SQLException Thrown when an error occurs applying the constraints
     */
    protected void prepareStatementForExecution(PreparedStatement ps)
    throws SQLException
    {
        // Apply any user-specified timeout
        int timeout = pm.getPMFContext().getPmfConfiguration().getQueryTimeout();
        if (extensions != null && extensions.containsKey("org.jpox.query.timeout"))
        {
            // Accept timeout as an Integer or String
            Object val = extensions.get("org.jpox.query.timeout");
            if (val instanceof Integer)
            {
                timeout = ((Integer)val).intValue();
            }
            else if (val instanceof String)
            {
                timeout = TypeConversionHelper.intFromString((String)val, 0);
            }
        }
        if (timeout > 0)
        {
            ps.setQueryTimeout(timeout);
        }

        // Apply any fetch size
        int fetchSize = pm.getPMFContext().getPmfConfiguration().getQueryFetchSize();
        if (extensions != null && extensions.containsKey("org.jpox.query.fetchSize"))
        {
            Object val = extensions.get("org.jpox.query.fetchSize");
            if (val instanceof Integer)
            {
                fetchSize = ((Integer)val).intValue();
            }
            else if (val instanceof String)
            {
                fetchSize = TypeConversionHelper.intFromString((String)val, 0);
            }
        }
        if ((dba.supportsQueryFetchSizeLowerThanOne() && fetchSize < 1) || (fetchSize > 0))
        {
            ps.setFetchSize(fetchSize);
        }

        // Apply any fetch direction
        String fetchDir = pm.getPMFContext().getPmfConfiguration().getQueryFetchDirection();
        if (extensions != null && extensions.containsKey("org.jpox.query.fetchDirection"))
        {
            fetchDir = (String)extensions.get("org.jpox.query.fetchDirection");
            if (!fetchDir.equals("forward") && !fetchDir.equals("reverse") && !fetchDir.equals("unknown"))
            {
                throw new JDOUserException("Query extension for JPOX org.jpox.query.fetchDirection has valid values of forward,reverse,unknown only");
            }
        }
        if (fetchDir.equals("reverse"))
        {
            ps.setFetchDirection(ResultSet.FETCH_REVERSE);
        }
        else if (fetchDir.equals("unknown"))
        {
            ps.setFetchDirection(ResultSet.FETCH_UNKNOWN);
        }
    }
}