/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2004 Andy Jefferson - commented and localised
2004 Andy Jefferson - changed to allow for null candidate class
2005 Andy Jefferson - changed to use JDBC parameters
2005 Andy Jefferson - added timeout support
2005 Andy Jefferson - changed to use PersistenceCapable candidate class
2005 Andy Jefferson - added checks on missing columns to compile step
2006 Andy Jefferson - implemented executeWithMap taking Integer args
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;

import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.exceptions.ClassNotPersistenceCapableException;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.metadata.IdentityType;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.DatastoreMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryResult;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.util.StringUtils;

/**
 * A Query using SQL, and keeping to the JDO2 definition of a SQL query.
 * The majority of this has to be specified in the query filter itself.
 * Parameters are embodied in the query and are unnamed/untyped. There
 * are no variables/imports. Ordering/grouping is explicit in the query.
 *
 * @see Query
 * @version $Revision: 1.13.2.2 $
 */
public final class SQLQuery extends BaseSQLQuery
{
    /**
     * Constructor for a new query using the existing query.
     * @param pm Persistence Manager
     * @param query The existing query
     */
    public SQLQuery(PersistenceManager pm, SQLQuery query)
    {
        super(pm, query);
    }

    /**
     * Constructs a new query instance having the same criteria as the given query.
     * @param pm The PersistenceManager
     */
    public SQLQuery(PersistenceManager pm)
    {
        super(pm, (String)null);
    }

    /**
     * Constructs a new query instance having the same criteria as the given query.
     * @param pm The PersistenceManager
     * @param sql_text The SQL query string
     */
    public SQLQuery(PersistenceManager pm, String sql_text)
    {
        super(pm, sql_text);

        if (!pm.getPMFContext().getPmfConfiguration().isQueryAllowAllSQLStatements())
        {
            // JDO spec section 14.7. SQL queries must start with SELECT
            if (!sqlText.startsWith("SELECT") && !sqlText.startsWith("select"))
            {
                throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.StatementNotSelectError", sqlText));
            }
        }
    }

    /**
     * Equality operator
     * @param obj The object to compare against
     * @return Whether they are equal
     **/
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof SQLQuery) || !super.equals(obj))
        {
            return false;
        }

        return sqlText.equals(((SQLQuery)obj).sqlText);
    }

    /**
     * Declare the parameters to be used in the query.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     *
     * @param parameters the parameters separated by semicolons.
     * @exception JDOUserException Always thrown.
     * @see javax.jdo.Query#declareVariables
     */
    public void declareParameters(String parameters)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.ParametersInapplicableError"));
    }

    /**
     * Set the import statements to be used to identify the fully qualified
     * name of variables or parameters.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * 
     * @param imports   import statements separated by semicolons.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#declareImports
     */
    public void declareImports(String imports)
    {
        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.ImportInapplicableError"));
    }

    /**
     * Method to perform any necessary pre-processing on the users query statement
     * before we execute it. SQL queries are not modified in any way, as per JDO2 spec
     * section 14.7.
     */
    protected void generateQueryStatement()
    {
        jdbcSqlText = sqlText; // Just take the users SQL query direct - no substitution of params etc.

        // Perform any sanity checking on the users input
        if (candidateClass != null)
        {
            AbstractClassMetaData cmd = pm.getMetaDataManager().getMetaDataForClass(candidateClass,pm.getClassLoaderResolver());
            if (cmd == null)
            {
                throw new ClassNotPersistenceCapableException(candidateClass.getName());
            }
            if (cmd.getPersistenceCapableSuperclass() != null)
            {
               // throw new PersistentSuperclassNotAllowedException(candidateClass.getName());
            }

            if (resultClass == null)
            {
                // Check the presence of the required columns (id, version, discriminator) in the candidate class
                String selections = sqlText.trim().substring(7); // Skip "SELECT "
                int fromStart = selections.indexOf("FROM");
                if (fromStart == -1)
                {
                    fromStart = selections.indexOf("from");
                }
                selections = selections.substring(0, fromStart).trim();
                String[] selectedColumns = StringUtils.split(selections, ",");

                if (selectedColumns == null || selectedColumns.length == 0)
                {
                    throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.NoSelectionsError", sqlText));
                }
                if (selectedColumns.length == 1 && selectedColumns[0].trim().equals("*"))
                {
                    // SQL Query using * so just return since all possible columns will be selected
                    return;
                }

                // Generate id column field information for later checking the id is present
                DatastoreClass table = pm.getStoreManager().getDatastoreClass(candidateClass.getName(), pm.getClassLoaderResolver());
                PersistenceCapableMapping idMapping = (PersistenceCapableMapping)table.getIDMapping();
                String[] idColNames = new String[idMapping.getNumberOfDatastoreFields()];
                boolean[] idColMissing = new boolean[idMapping.getNumberOfDatastoreFields()];
                for (int i=0;i<idMapping.getNumberOfDatastoreFields();i++)
                {
                    DatastoreMapping m = idMapping.getDataStoreMapping(i);
                    idColNames[i] = m.getDatastoreField().getIdentifier().toString();
                    idColMissing[i] = true;
                }

                // Generate discriminator/version information for later checking they are present
                String discriminatorColName = table.getDiscriminatorMapping() != null ? 
                        table.getDiscriminatorMapping().getDataStoreMapping(0).getDatastoreField().getIdentifier().toString() : null;
                String versionColName = table.getVersionMapping() != null ? 
                        table.getVersionMapping().getDataStoreMapping(0).getDatastoreField().getIdentifier().toString() : null;
                boolean discrimMissing = (discriminatorColName != null);
                boolean versionMissing = true;
                if (versionColName == null)
                {
                    versionMissing = false;
                }

                // Go through the selected fields and check the existence of id,
                // version, discriminator cols
                final AbstractClassMetaData candidateCmd = pm.getMetaDataManager().getMetaDataForClass(candidateClass, pm.getClassLoaderResolver());
                for (int i = 0; i < selectedColumns.length; i++)
                {
                    String colName = selectedColumns[i].trim();
                    if (colName.indexOf(" AS ") > 0)
                    {
                        // Allow for user specification of "XX.YY AS ZZ"
                        colName = colName.substring(colName.indexOf(" AS ")+4).trim();
                    }
                    else if (colName.indexOf(" as ") > 0)
                    {
                        // Allow for user specification of "XX.YY as ZZ"
                        colName = colName.substring(colName.indexOf(" as ")+4).trim();
                    }

                    if (candidateCmd.getIdentityType() == IdentityType.DATASTORE)
                    {
                        // Check for existence of id column, allowing for any
                        // RDBMS using quoted identifiers
                        if (columnNamesAreTheSame(idColNames[0], colName))
                        {
                            idColMissing[0] = false;
                        }
                    }
                    else if (candidateCmd.getIdentityType() == IdentityType.APPLICATION)
                    {
                        for (int j=0; j<idColNames.length; j++)
                        {
                            // Check for existence of id column, allowing for
                            // any RDBMS using quoted identifiers
                            if (columnNamesAreTheSame(idColNames[j], colName))
                            {
                                idColMissing[j] = false;
                            }
                        }
                    }

                    if (discrimMissing && columnNamesAreTheSame(discriminatorColName, colName))
                    {
                        discrimMissing = false;
                    }
                    else if (versionMissing && columnNamesAreTheSame(versionColName, colName))
                    {
                        versionMissing = false;
                    }
                }

                if (discrimMissing)
                {
                    throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.Candidate.DiscriminatorColumnNotSelected", sqlText, candidateClass.getName(),
                        discriminatorColName));
                }
                if (versionMissing)
                {
                    throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.Candidate.VersionColumnNotSelected", sqlText, candidateClass.getName(), versionColName));
                }
                for (int i = 0; i < idColMissing.length; i++)
                {
                    if (idColMissing[i])
                    {
                        throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.Candidate.IdColumnNotSelected", sqlText, candidateClass.getName(), 
                            idColNames[i]));
                    }
                }
            }
        }
    }

    /**
     * Convenience method to compare two column names.
     * Allows for case sensitivity issues, and for database added quoting.
     * @param name1 The first name (from the datastore)
     * @param name2 The second name (from the user SQL statement)
     * @return Whether they are the same
     */
    private boolean columnNamesAreTheSame(String name1, String name2)
    {
        if (name1.equalsIgnoreCase(name2) ||
            name1.equalsIgnoreCase(dba.getIdentifierQuoteString() + name2 + dba.getIdentifierQuoteString()))
        {
            return true;
        }
        return false;
    }

    /**
     * Execute the query and return the filtered List. Overrides the version in
     * Query so that we can allow for not having parameter names in SQL.
     * @param parameters the Object array with all of the parameters.
     * @return the filtered List.
     * @see javax.jdo.Query#executeWithArray(Object[])
     */
    public Object executeWithArray(Object[] parameters)
    {
        isPreCompile = false;
        compile();

        // Clear the parameterNames that are set in compile since we assign ours using the parameterMap passed in
        parameterNames.clear();

        // Insert dummy parameter names that are just "1", "2", etc
        HashMap parameterMap = new HashMap();
        if (parameters != null)
        {
            for (int i = 0; i < parameters.length; ++i)
            {
                String name = "" + (i+1);
                parameterNames.add(name);
                parameterMap.put(new Integer(i+1), parameters[i]);
            }
        }

        Object theReturn = super.executeWithMap(parameterMap);
        isPreCompile = true;
        return theReturn; 
    }

    /**
     * Execute the query and return the filtered List. Overrides the version in
     * Query so that we can allow for not having parameter names in SQL and impose
     * the restriction that all map keys are integers starting at 1.
     * @param parameterMap the Map with all of the parameters.
     * @return the filtered List.
     * @see javax.jdo.Query#executeWithArray(Object[])
     */
    public Object executeWithMap(Map parameterMap)
    {
        isPreCompile = false;
        compile();

        // Clear the parameterNames that are set in compile since we assign ours using the parameterMap passed in
        parameterNames.clear();

        // Check number of parameters required for the SQL statement
        int numParamsExpected = 0;
        boolean complete = false;
        int position = 0;
        while (!complete)
        {
            int paramPos = jdbcSqlText.indexOf('?', position);
            if (paramPos < 0)
            {
                complete = true;
            }
            else
            {
                position = paramPos+1;
                numParamsExpected++;
            }
        }

        if (numParamsExpected > 0 && parameterMap == null)
        {
            // We need "numParamsExpected" yet the user gives us none!
            throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.ParametersRequiredButNoneSupplied", 
                jdbcSqlText, "" + numParamsExpected));
        }

        Map executeMap = new HashMap();
        if (parameterMap != null)
        {
            // Just check for parameters of type Integer at the parameter positions expected
            for (int i=1;i<numParamsExpected+1;i++)
            {
                Integer key = new Integer(i);
                if (!parameterMap.containsKey(key))
                {
                    // No parameter provided for position "i"
                    throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.ParameterAtPositionNotSupplied", 
                        jdbcSqlText, "" + i));
                }
                else
                {
                    executeMap.put(key, parameterMap.get(key));
                    parameterNames.add("" + i);
                }
            }
        }

        Object theReturn = super.executeWithMap(executeMap);
        isPreCompile = true;
        return theReturn; 
    }

    /**
     * Execute the query and return the filtered QueryResult.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult.
     */
    public Collection performExecute(Map parameters)
    {
        compile();

        if (parameters.size() != parameterNames.size())
        {
            throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.IncorrectNumberOfParametersError","" + parameterNames.size(),"" + parameters.size()));
        }

        QueryResult qr = null;
        try
        {
            RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
            Connection conn = storeMgr.getConnection(pm, false, true);

            try
            {
                PreparedStatement ps = getStatement(conn, jdbcSqlText);
                try
                {
                    // Set the values of any parameters
                    if (parameters != null)
                    {
                        for (int i=0;i<parameters.size();i++)
                        {
                            Object obj = parameters.get(new Integer(i+1));
                            ps.setObject((i+1), obj);
                        }
                    }

                    // Apply any user-specified constraints over timeouts and ResultSet
                    prepareStatementForExecution(ps);

                    // Execute the query
                    ResultSet rs = storeMgr.executeStatementQuery(jdbcSqlText, ps);
                    try
                    {
                        if (resultClass != null || candidateClass == null)
                        {
                            // Where a result class is provided, or where we have no candidate class then we
                            // massage the results into a class
                            qr = getQueryResultForNoCandidateClass(rs);
                        }
                        else
                        {
                            // Create an index listing for ALL (fetchable) fields in the result class.
                            final AbstractClassMetaData candidateCmd = pm.getMetaDataManager().getMetaDataForClass(candidateClass, pm.getClassLoaderResolver());
                            int fieldCount = candidateCmd.getNoOfManagedFields() + candidateCmd.getNoOfInheritedManagedFields();
                            Map columnFieldNumberMap = new HashMap(); // Map of field numbers keyed by the column name
                            statementExpressionIndex = new StatementExpressionIndex[fieldCount];
                            for (int fieldNumber = 0; fieldNumber < fieldCount; ++fieldNumber)
                            {
                                statementExpressionIndex[fieldNumber] = new StatementExpressionIndex();
                                AbstractPropertyMetaData fmd = candidateCmd.getManagedFieldAbsolute(fieldNumber);
                                String fieldName = fmd.getName();
                                Class fieldType = fmd.getType();

                                if (fmd.getPersistenceModifier() != FieldPersistenceModifier.NONE)
                                {
                                    JavaTypeMapping m = dba.getMapping(fieldType, getStoreManager(), pm.getClassLoaderResolver());
                                    if (m.includeInFetchStatement())
                                    {
                                        // Set mapping for this field since it can potentially be returned from a fetch
                                        statementExpressionIndex[fieldNumber].setMapping(m);

                                        String columnName = null;
                                        if (fmd.getColumnMetaData() != null && fmd.getColumnMetaData().length > 0)
                                        {
                                            for (int colNum = 0;colNum<fmd.getColumnMetaData().length;colNum++)
                                            {
                                                columnName = fmd.getColumnMetaData()[colNum].getName();
                                                columnFieldNumberMap.put(columnName, new Integer(fieldNumber));
                                            }
                                        }
                                        else
                                        {
                                            columnName = pm.getStoreManager().getIdentifierFactory().newDatastoreFieldIdentifier(fieldName, 
                                                pm.getPMFContext().getTypeManager().isDefaultEmbeddedType(fieldType), DatastoreField.ROLE_NONE).getIdentifier();
                                            columnFieldNumberMap.put(columnName, new Integer(fieldNumber));
                                        }
                                    }
                                    else
                                    {
                                        // Dont put anything in this position (field has no col in the result set)
                                    }
                                }
                                else
                                {
                                    // Dont put anything in this position (field has no col in the result set)
                                }
                            }
                            if (columnFieldNumberMap.size() == 0)
                            {
                            	// Check if we have no fields in the class that have a column in the datastore (so nothing to select)
                                // TODO Localise the message
                                throw new JDOFatalUserException("SQL query class has no persistent fields to SELECT : " + candidateClass.getName());
                            }

                            // Generate id column field information for later checking the id is present
                            DatastoreClass table = pm.getStoreManager().getDatastoreClass(candidateClass.getName(), pm.getClassLoaderResolver());
                            PersistenceCapableMapping idMapping = (PersistenceCapableMapping)table.getIDMapping();
                            String[] idColNames = new String[idMapping.getNumberOfDatastoreFields()];
                            boolean[] idColMissing = new boolean[idMapping.getNumberOfDatastoreFields()];
                            for (int i=0;i<idMapping.getNumberOfDatastoreFields();i++)
                            {
                                DatastoreMapping m = idMapping.getDataStoreMapping(i);
                                idColNames[i] = m.getDatastoreField().getIdentifier().toString();
                                idColMissing[i] = true;
                            }

                            // Generate discriminator/version information for later checking they are present
                            String discriminatorColName = table.getDiscriminatorMapping() != null ?
                                    table.getDiscriminatorMapping().getDataStoreMapping(0).getDatastoreField().getIdentifier().toString() : null;
                            String versionColName = table.getVersionMapping() != null ?
                                    table.getVersionMapping().getDataStoreMapping(0).getDatastoreField().getIdentifier().toString() : null;
                            boolean discrimMissing = (discriminatorColName != null);
                            boolean versionMissing = true;
                            if (versionColName == null)
                            {
                                versionMissing = false;
                            }

                            // Go through the fields of the ResultSet and map to the required fields in the candidate
                            // Note that we check the existence of the columns again here even though they were checked at compilation
                            // TODO This could be removed from here since its now done at compile time
                            ResultSetMetaData rsmd = rs.getMetaData();
                            HashSet remainingColumnNames = new HashSet(columnFieldNumberMap.size());
                            int colCount = rsmd.getColumnCount();
                            int[] datastoreIndex = null;
                            int[] versionIndex = null;

                            int[] matchedFieldNumbers = new int[colCount];
                            int fieldNumberPosition = 0;
                            for (int colNum=1; colNum<=colCount; ++colNum)
                            {
                                String colName = rsmd.getColumnName(colNum);
                                int fieldNumber = -1;
                                Integer fieldNum = (Integer)columnFieldNumberMap.get(colName);
                                if (fieldNum != null)
                                {
                                    fieldNumber = fieldNum.intValue();
                                }
                                if (fieldNumber >= 0)
                                {
                                    int[] exprIndices = null;
                                    if (statementExpressionIndex[fieldNumber].getExpressionIndex() != null)
                                    {
                                        exprIndices = new int[statementExpressionIndex[fieldNumber].getExpressionIndex().length+1];
                                        for (int i=0;i<statementExpressionIndex[fieldNumber].getExpressionIndex().length;i++)
                                        {
                                            exprIndices[i] = statementExpressionIndex[fieldNumber].getExpressionIndex()[i];
                                        }
                                        exprIndices[exprIndices.length-1] = colNum;
                                    }
                                    else
                                    {
                                        exprIndices = new int[] {colNum};
                                    }
                                    statementExpressionIndex[fieldNumber].setExpressionIndex(exprIndices);
                                    remainingColumnNames.remove(colName);
                                    matchedFieldNumbers[fieldNumberPosition++] = fieldNumber;
                                }

                                if (versionColName != null && colName.equals(versionColName))
                                {
                                    // Identify the location of the version column
                                    versionIndex = new int[1];
                                    versionIndex[0] = colNum;
                                    versionMissing = false;
                                }

                                if (candidateCmd.getIdentityType() == IdentityType.DATASTORE)
                                {
                                    // Check for existence of id column, allowing for any RDBMS using quoted identifiers
                                    if (columnNamesAreTheSame(idColNames[0], colName))
                                    {
                                        datastoreIndex = new int[1];
                                        datastoreIndex[0] = colNum;
                                        idColMissing[0] = false;
                                    }
                                }
                                else if (candidateCmd.getIdentityType() == IdentityType.APPLICATION)
                                {
                                    for (int j=0;j<idColNames.length;j++)
                                    {
                                        // Check for existence of id column, allowing for any RDBMS using quoted identifiers
                                        if (columnNamesAreTheSame(idColNames[j], colName))
                                        {
                                            idColMissing[j] = false;
                                        }
                                    }
                                }
                                if (discrimMissing && columnNamesAreTheSame(discriminatorColName, colName))
                                {
                                    discrimMissing = false;
                                }
                                else if (versionMissing && columnNamesAreTheSame(versionColName, colName))
                                {
                                    versionMissing = false;
                                }
                            }

                            // Set the field numbers found to match what we really have
                            int[] fieldNumbers = new int[fieldNumberPosition];
                            for (int i=0;i<fieldNumberPosition;i++)
                            {
                                fieldNumbers[i] = matchedFieldNumbers[i];
                            }

                            if (discrimMissing)
                            {
                                throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.Candidate.DiscriminatorColumnNotSelected", 
                                    sqlText, candidateClass.getName(), discriminatorColName));
                            }
                            if (versionMissing)
                            {
                                throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.Candidate.VersionColumnNotSelected", 
                                    sqlText, candidateClass.getName(), versionColName));
                            }
                            for (int i=0;i<idColMissing.length;i++)
                            {
                                if (idColMissing[i])
                                {
                                    throw new JDOUserException(LOCALISER_RDBMS.msg("SQL.Candidate.IdColumnNotSelected", 
                                        sqlText, candidateClass.getName(), idColNames[i]));
                                }
                            }

                            PersistentIDROF rof = new PersistentIDROF(table, fieldNumbers, candidateCmd, statementExpressionIndex, datastoreIndex,
                                versionIndex, ignoreCache, false, false, (FetchPlanImpl)fetchPlan);
                            if (getResultSetType().equals("scroll-insensitive"))
                            {
                                qr = new InsensitiveQueryResult(null, this, rof, rs, null);
                            }
                            else
                            {
                                qr = new ForwardQueryResult(null, this, rof, rs, null);
                            }
                        }
                    }
                    finally
                    {
                        if (qr == null)
                        {
                            rs.close();
                        }
                    }
                }
                finally
                {
                    if (qr == null)
                    {
                        ps.close();
                    }
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER_RDBMS.msg("SQL.ExecutionError", jdbcSqlText), e);
        }

        queryResults.add(qr);

        return qr;
    }
}