/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Erik Bengtson - the fields to fetch are better managed for application
                     identity. Their selection were moved from execute
                     method to the constructor
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - conversion to use Logger
2004 Erik Bengtson - changed to use mapping consumer
2004 Andy Jefferson - added discriminator support
2005 Andy Jefferson - fixed 1-1 bidir order of insertion
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.Relation;
import org.jpox.state.ActivityState;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.exceptions.NotYetFlushedException;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.mapping.VersionMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.adapter.RDBMSAdapter;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.SecondaryTable;
import org.jpox.store.rdbms.table.Table;
import org.jpox.util.JPOXLogger;
import org.jpox.util.JavaUtils;
import org.jpox.util.StringUtils;

/**
 * Class to provide a means of insertion of records to a data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * insert operation.
 * <p>
 * When inserting an object with inheritance this will involve 1 InsertRequest
 * for each table involved. So if we have a class B that extends class A and they
 * both use "new-table" inheritance strategy, we will have 2 InsertRequests, one for
 * table A, and one for table B. When the InsertRequest starts to populate its statement
 * and it has a PC field, this calls PersistenceCapableMapping.setObject(). This then
 * checks if the other PC object is yet persistent and, if not, will persist it
 * before processing this objects INSERT. This forms the key to
 * "persistence-by-reachability".
 * </p>
 *
 * @version $Revision: 1.58.2.1 $ 
 **/
public class InsertRequest extends Request
{
    private static final int IDPARAMNUMBER = 1;

    private final MappingCallbacks[] callbacks;

    /** Fields to be provided in the INSERT statement. */
    private final int[] fieldsToBeProvided;

    /** Primary key fields to be provided on INSERT to define the record. */
    private final int[] pkfieldsToBeProvided;

    /**
     * Fields to be retrieved that this record references but that it has no FK to.
     * These fields are 1-1 bidir/1-N join table bidir using mapped by where there are no datastore columns
     * representing the field in this table.
     */
    private final int[] fieldsToBeRetrieved;

    /** SQL statement for the INSERT. */
    private final String insertStmt;

    /** Whether the class has an auto-increment column */
    private boolean hasAutoIncrementColumn = false;

    /** one StatementExpressionIndex for each field **/
    private StatementExpressionIndex[] statementExpressionIndex;

    /** StatementExpressionIndex for fields to be "retrieved" **/
    private StatementExpressionIndex[] retrievedStatementExpressionIndex;

    /** StatementExpressionIndex for version **/
    private StatementExpressionIndex versionStatementExpressionIndex;

    /** StatementExpressionIndex for discriminator **/
    private StatementExpressionIndex discriminatorStatementExpressionIndex;

    /** StatementExpressionIndex for external FKs */
    private StatementExpressionIndex[] externalFKStatementExpressionIndex;

    /** StatementExpressionIndex for external FK discriminators (shared FKs) */
    private StatementExpressionIndex[] externalFKDiscrimStatementExpressionIndex;

    /** StatementExpressionIndex for external indices */
    private StatementExpressionIndex[] externalOrderStatementExpressionIndex;

    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query.
     * @param table The Class Table representing the datastore table to insert.
     * @param cls Class of objects being updated
     * @param clr ClassLoader resolver
     **/
    public InsertRequest(DatastoreClass table, Class cls, ClassLoaderResolver clr)
    {
        super(table);

        AbstractClassMetaData cmd = table.getStoreManager().getMetaDataManager().getMetaDataForClass(cls, clr);

        InsertMappingConsumer consumer = new InsertMappingConsumer(clr, cmd, IDPARAMNUMBER);
        table.provideDatastoreIdMappings(consumer);
        table.provideNonPrimaryKeyMappings(consumer);
        table.providePrimaryKeyMappings(consumer);
        table.provideVersionMappings(consumer);
        table.provideDiscriminatorMappings(consumer);
        table.provideExternalMappings(consumer, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK);
        table.provideExternalMappings(consumer, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK_DISCRIM);
        table.provideExternalMappings(consumer, MappingConsumer.MAPPING_TYPE_EXTERNAL_INDEX);
        callbacks = (MappingCallbacks[])consumer.getMappingCallbacks().toArray(new MappingCallbacks[consumer.getMappingCallbacks().size()]);

        statementExpressionIndex = consumer.getStatementExpressionIndex();
        versionStatementExpressionIndex = consumer.getVersionStatementExpressionIndex();
        discriminatorStatementExpressionIndex = consumer.getDiscriminatorStatementExpressionIndex();
        externalFKStatementExpressionIndex = consumer.getExternalFKStatementExpressionIndex();
        externalFKDiscrimStatementExpressionIndex = consumer.getExternalFKDiscrimStatementExpressionIndex();
        externalOrderStatementExpressionIndex = consumer.getExternalOrderStatementExpressionIndex();
        pkfieldsToBeProvided = consumer.getPrimaryKeyFieldsToBeProvided();
        if (table.getIdentityType() == IdentityType.APPLICATION &&
            pkfieldsToBeProvided.length < 1 &&
            !hasAutoIncrementColumn)
        {
	        throw new JDOFatalException(LOCALISER.msg("RDBMS.Request.InsertWithNoPrimaryKeyError", cmd.getFullClassName()));
        }
        fieldsToBeProvided = consumer.getFieldsToBeProvided();
        retrievedStatementExpressionIndex = consumer.getRetrievedStatementExpressionIndex();
        fieldsToBeRetrieved = consumer.getFieldsToBeRetrieved();

        insertStmt = consumer.getInsertStmt();
    }

    /**
     * Method performing the insertion of the record from the datastore. Takes
     * the constructed insert query and populates with the specific record
     * information.
     * @param sm The state manager for the record to be inserted
     **/
    public void execute(StateManager sm)
    {
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            // Debug information about what we are inserting
            StringBuffer str = new StringBuffer("Inserting object " + StringUtils.toJVMIDString(sm.getObject()) +
                " of type " + sm.getObject().getClass().getName() + " into table " + table);
            JPOXLogger.JDO.debug(str);
        }

        try
        {
            // Set the state to "inserting" (may already be at this state if multiple inheritance level INSERT)
            sm.changeActivityState(ActivityState.INSERTING, table);

            PersistenceManager pm = sm.getPersistenceManager();
            RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            try
            {
                PreparedStatement ps = storeMgr.getStatement(conn, insertStmt, false);

                try
                {
                    // Provide the primary key field(s)
                    if (table.getIdentityType() == IdentityType.DATASTORE)
                    {
                        if (!table.isObjectIDDatastoreAttributed() || !table.isBaseDatastoreClass())
                        {
                            int[] paramNumber = {IDPARAMNUMBER};
                            table.getDataStoreObjectIdMapping().setObject(pm, ps, paramNumber, sm.getInternalObjectId());
                        }
                    }
                    else if (table.getIdentityType() == IdentityType.APPLICATION)
                    {
                        sm.provideFields(pkfieldsToBeProvided, new ParameterSetter(sm, ps, statementExpressionIndex, true));
                    }

                    // Provide all non-key fields needed for the insert.
                    // This provides "persistence-by-reachability" for these fields
                    int numberOfFieldsToProvide = 0;
                    for (int i = 0; i < fieldsToBeProvided.length; i++)
                    {
                        if (fieldsToBeProvided[i] < sm.getHighestFieldNumber())
                        {
                            numberOfFieldsToProvide++;
                        }
                    }
                    int j = 0;
                    int[] classFieldNumbers = new int[numberOfFieldsToProvide];
                    for (int i = 0; i < fieldsToBeProvided.length; i++)
                    {
                        if (fieldsToBeProvided[i] < sm.getHighestFieldNumber())
                        {
                            classFieldNumbers[j++] = fieldsToBeProvided[i];
                        }
                        else
                        {
                            // Any fields out of range for the class, we default or null.
                            // This happens when we insert a base class object
                            // and the sub-class is persisted to the
                            // same table ("superclass-table").
                            StatementExpressionIndex stmtExprIndex = statementExpressionIndex[fieldsToBeProvided[i]];
                            stmtExprIndex.getMapping().setDefault(pm, ps, stmtExprIndex.getParameterIndex());
                        }
                    }
                    sm.provideFields(classFieldNumbers, new ParameterSetter(sm, ps, statementExpressionIndex, true));

                    // Version mapping (optional)
                    if (table.getVersionMapping() != null)
                    {
                        Object nextOptimisticVersion = ((VersionMapping) table.getVersionMapping()).getNextVersion(sm.getVersion(sm.getObject()));
                        table.getVersionMapping().setObject(pm, ps, versionStatementExpressionIndex.getParameterIndex(), nextOptimisticVersion);
                        sm.setTransactionalVersion(nextOptimisticVersion);
                    }

                    // Discriminator mapping (optional)
                    if (table.getDiscriminatorMapping() != null)
                    {
                        DiscriminatorMetaData dismd = table.getDiscriminatorMetaData();
                        if (dismd.getStrategy() == DiscriminatorStrategy.CLASS_NAME)
                        {
                            table.getDiscriminatorMapping().setObject(pm, ps, discriminatorStatementExpressionIndex.getParameterIndex(),
                                sm.getObject().getClass().getName());
                        }
                        else if (dismd.getStrategy() == DiscriminatorStrategy.VALUE_MAP)
                        {
                            // Use Discriminator info for the actual class
                            dismd = sm.getClassMetaData().getInheritanceMetaData().getDiscriminatorMetaData();
                            table.getDiscriminatorMapping().setObject(pm, ps, discriminatorStatementExpressionIndex.getParameterIndex(), dismd.getValue());
                        }
                    }

                    // External FK columns (optional)
                    if (externalFKStatementExpressionIndex != null)
                    {
                        for (int i=0;i<externalFKStatementExpressionIndex.length;i++)
                        {
                            Object fkValue = sm.getValueForExternalField(externalFKStatementExpressionIndex[i].getMapping());
                            if (fkValue != null)
                            {
                                // Need to provide the owner field number so PCMapping can work out if it is inserted yet
                                AbstractPropertyMetaData ownerFmd = 
                                    table.getMetaDataForExternalMapping(externalFKStatementExpressionIndex[i].getMapping(), 
                                        MappingConsumer.MAPPING_TYPE_EXTERNAL_FK);
                                externalFKStatementExpressionIndex[i].getMapping().setObject(pm, ps,
                                    externalFKStatementExpressionIndex[i].getParameterIndex(), fkValue, null, ownerFmd.getAbsoluteFieldNumber());
                            }
                            else
                            {
                                // We're inserting a null so dont need the owner field
                                externalFKStatementExpressionIndex[i].getMapping().setObject(pm, ps,
                                    externalFKStatementExpressionIndex[i].getParameterIndex(), null);
                            }
                        }
                    }

                    // External FK discriminator columns (optional)
                    if (externalFKDiscrimStatementExpressionIndex != null)
                    {
                        for (int i=0;i<externalFKDiscrimStatementExpressionIndex.length;i++)
                        {
                            Object discrimValue = sm.getValueForExternalField(externalFKDiscrimStatementExpressionIndex[i].getMapping());
                            externalFKDiscrimStatementExpressionIndex[i].getMapping().setObject(pm, ps, externalFKDiscrimStatementExpressionIndex[i].getParameterIndex(), discrimValue);
                        }
                    }

                    // External order columns (optional)
                    if (externalOrderStatementExpressionIndex != null)
                    {
                        for (int i=0;i<externalOrderStatementExpressionIndex.length;i++)
                        {
                            Object orderValue = sm.getValueForExternalField(externalOrderStatementExpressionIndex[i].getMapping());
                            if (orderValue == null)
                            {
                                // No order value so use -1
                                orderValue = new Integer(-1);
                            }
                            externalOrderStatementExpressionIndex[i].getMapping().setObject(pm, ps, externalOrderStatementExpressionIndex[i].getParameterIndex(), orderValue);
                        }
                    }

                    storeMgr.executeStatementUpdate(insertStmt, ps);

                    // Retrieve the key for auto-increment/identity keys
                    if (hasAutoIncrementColumn)
                    {
                        Object newId = null;

                        if (((RDBMSAdapter) pm.getStoreManager().getDatastoreAdapter()).supportsStatementGetGeneratedKeys())
                        {
                            if (JavaUtils.isJRE1_4OrAbove())
                            {
                                // JDK 1.4+ so try the getGeneratedKeys() method for more efficiency
                                ResultSet rs = null;
                                try
                                {
                                    rs = ps.getGeneratedKeys();
                                    if (rs != null && rs.next())
                                    {
                                        newId = rs.getObject(1);
                                    }
                                }
                                catch (Throwable e)
                                {
                                    // Not supported maybe (e.g HSQL), or the driver is too old
                                }
                                finally
                                {
                                    if (rs != null)
                                    {
                                        rs.close();
                                    }
                                }
                            }
                        }
                        if (newId == null)
                        {
                            // No identity found, so try the native method for retrieving it
                            String tableName = table.getIdentifier().toString();
                            String columnName = null;
                            JavaTypeMapping idMapping = table.getIDMapping();
                            if (idMapping != null)
                            {
                                for (int i=0;i<idMapping.getNumberOfDatastoreFields();i++)
                                {
                                    Column col = (Column)idMapping.getDataStoreMapping(i).getDatastoreField();
                                    if (col.isAutoIncrement())
                                    {
                                        columnName = col.getIdentifier().toString();
                                        break;
                                    }
                                }
                            }
                            String autoIncStmt = ((RDBMSAdapter) pm.getStoreManager().getDatastoreAdapter()).getAutoIncrementStmt(tableName, columnName);
                            PreparedStatement psAutoIncrement = storeMgr.getStatement(conn, autoIncStmt, false);
                            ResultSet rs = null;
                            try
                            {
                                rs = storeMgr.executeStatementQuery(autoIncStmt, psAutoIncrement);
                                if (rs.next())
                                {
                                    newId = rs.getObject(1);
                                }
                            }
                            finally
                            {
                                if (rs != null)
                                {
                                    rs.close();
                                }
                                if (psAutoIncrement != null)
                                {
                                    psAutoIncrement.close();
                                }                            
                            }
                        }

                        if (newId == null)
                        {
                            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Request.InsertOfAutoincrementReturnedNullIdentity",this.table));
                        }
                        sm.setPostStoreNewObjectId(newId);
                    }

                    // Update the insert status for this table
                    sm.changeActivityState(ActivityState.INSERTING_CALLBACKS, table);

                    // Retrieve all other fields that relate to this object
                    // This provides "persistence-by-reachability" for these fields
                    int numberOfFieldsToRetrieve = 0;
                    for (int i = 0; i < fieldsToBeRetrieved.length; i++)
                    {
                        if (fieldsToBeRetrieved[i] < sm.getHighestFieldNumber())
                        {
                            numberOfFieldsToRetrieve++;
                        }
                    }
                    classFieldNumbers = new int[numberOfFieldsToRetrieve];
                    j = 0;
                    for (int i = 0; i < fieldsToBeRetrieved.length; i++)
                    {
                        if (fieldsToBeRetrieved[i] < sm.getHighestFieldNumber())
                        {
                            classFieldNumbers[j++] = fieldsToBeRetrieved[i];
                        }
                        else
                        {
                            // Any fields out of range for the class, we default or null.
                            // This happens when we insert a base class object
                            // and the sub-class is persisted to the same table ("superclass-table").
                            StatementExpressionIndex stmtExprIndex = retrievedStatementExpressionIndex[fieldsToBeRetrieved[i]];
                            stmtExprIndex.getMapping().setDefault(pm, ps, stmtExprIndex.getParameterIndex());
                        }
                    }
                    sm.provideFields(classFieldNumbers, new ParameterSetter(sm, ps, retrievedStatementExpressionIndex, true));
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("Insert request failed: " + insertStmt, e);
        }

        // Execute any mapping actions now that we have inserted the element
        // (things like inserting any association parent-child).
        for (int i = 0; i < callbacks.length; ++i)
        {
            try
            {
                callbacks[i].postInsert(sm);
            }
            catch (NotYetFlushedException e)
            {
                sm.updateFieldAfterInsert(e.getPersistenceCapable(), ((JavaTypeMapping) callbacks[i]).getFieldMetaData().getAbsoluteFieldNumber());
            }
        }
    }

    /**
     * Internal class to provide mapping consumption.
     * TODO Think about making this a InsertStatement (copying FetchStatement)
     */
    private class InsertMappingConsumer implements MappingConsumer
    {
        /** fields to be provided */
        List ftbp = new ArrayList();

        /** primary key fields to be provided */
        List pkbp = new ArrayList();

        /** fields to be retrieved */
        List ftbr = new ArrayList();

        StringBuffer columnNames = new StringBuffer();
        StringBuffer columnValues = new StringBuffer();

        Map assignedColumns = new HashMap();

        List mc = new ArrayList();

        boolean initialized = false;

        int paramIndex;

        /** one StatementExpressionIndex for each field **/
        private StatementExpressionIndex[] statementExpressionIndex;

        /** statement indexes for fields to be "retrieved". */
        private StatementExpressionIndex[] retrievedStatementExpressionIndex;

        /** StatementExpressionIndex for version **/
        private StatementExpressionIndex versionStatementExpressionIndex;

        /** StatementExpressionIndex for discriminator **/
        private StatementExpressionIndex discriminatorStatementExpressionIndex;

        private StatementExpressionIndex[] externalFKStmtExprIndex;

        private StatementExpressionIndex[] externalFKDiscrimStmtExprIndex;

        private StatementExpressionIndex[] externalOrderStmtExprIndex;

        private final ClassLoaderResolver clr;
        private final AbstractClassMetaData cmd;

        /**
         * Constructor
         * @param clr the ClassLoaderResolver
         * @param cmd ClassMetaData
         * @param initialParamIndex the initial index to use for the JDBC Parameter
         */
        public InsertMappingConsumer(ClassLoaderResolver clr, AbstractClassMetaData cmd, int initialParamIndex)
        {
            super();
            this.clr = clr;
            this.cmd = cmd;
            this.paramIndex = initialParamIndex;
        }

        public void preConsumeMapping(int highestFieldNumber)
        {
            if (!initialized)
            {
                statementExpressionIndex = new StatementExpressionIndex[highestFieldNumber];
                retrievedStatementExpressionIndex = new StatementExpressionIndex[highestFieldNumber];
                initialized = true;
            }
        }

        public void consumeMapping(JavaTypeMapping m, AbstractPropertyMetaData fmd)
        {
            if (!fmd.getAbstractClassMetaData().isSameOrAncestorOf(cmd))
            {
                // Make sure we only accept mappings from the correct part of any inheritance tree
                return;
            }
            if (m.includeInInsertStatement())
            {
                if (m.getNumberOfDatastoreFields() == 0 && (m instanceof PersistenceCapableMapping || m instanceof InterfaceMapping))
                {
                    // Fields to be "retrieved" (that relate to this object but have no FK in the table)
                    retrievedStatementExpressionIndex[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
                    retrievedStatementExpressionIndex[fmd.getAbsoluteFieldNumber()].setMapping(m);
                    int relationType = fmd.getRelationType(clr);
                    if (relationType == Relation.ONE_TO_ONE_BI)
                    {
                        if (fmd.getMappedBy() != null)
                        {
                            // 1-1 bidirectional field without datastore column(s) (with single FK)
                            ftbr.add(new Integer(fmd.getAbsoluteFieldNumber()));
                        }
                    }
                    else if (relationType == Relation.MANY_TO_ONE_BI)
                    {
                        AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
                        if (fmd.getJoinMetaData() != null || relatedFmd.getJoinMetaData() != null)
                        {
                            // N-1 bidirectional using join table for relation
                            ftbr.add(new Integer(fmd.getAbsoluteFieldNumber()));
                        }
                    }
                }
                else
                {
                    // Fields to be "provided" (that have a datastore column)
                    statementExpressionIndex[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
                    statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setMapping(m);

                    // create the expressions index (columns index)
                    int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];
                    for (int j = 0; j < parametersIndex.length; j++)
                    {
                        // check if the column was not already assigned
                        Column c = (Column)m.getDataStoreMapping(j).getDatastoreField();
                        DatastoreIdentifier columnId = c.getIdentifier();
                        if (!assignedColumns.containsKey(columnId.toString()))
                        {
                            // Either we are a field in a secondary table.
                            // Or we are a subclass table.
                            // Or we are not datastore attributed.
                            if (table instanceof SecondaryTable ||
                                (!table.getStoreManager().isStrategyDatastoreAttributed(fmd.getValueStrategy()) && !c.isAutoIncrement()) ||
                                !table.isBaseDatastoreClass())
                            {
                                if (columnNames.length() > 0)
                                {
                                    columnNames.append(',');
                                    columnValues.append(',');
                                }
                                String param = ((RDBMSMapping)m.getDataStoreMapping(j)).getInsertionInputParameter();

                                columnNames.append(columnId);
                                columnValues.append(param);

                                if (param.indexOf("?") > -1) // only add fields to be replaced by the real values only if the param value has ?
                                {
                                    Integer abs_field_num = new Integer(fmd.getAbsoluteFieldNumber());
                                    if (fmd.isPrimaryKey())
                                    {
                                        if (!pkbp.contains(abs_field_num))
                                        {
                                            pkbp.add(abs_field_num);
                                        }
                                    }
                                    else
                                    {
                                        if (!ftbp.contains(abs_field_num))
                                        {
                                            ftbp.add(abs_field_num);
                                        }
                                    }
                                    parametersIndex[j] = paramIndex++;
                                }
                                assignedColumns.put(c.getIdentifier().toString(),new Integer(fmd.getAbsoluteFieldNumber()));
                            }
                            else
                            {
                                hasAutoIncrementColumn = true;
                            }
                        }
                        else
                        {
                            parametersIndex[j] = ((Integer)assignedColumns.get(c.getIdentifier().toString())).intValue();
                        }
                        statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setParameterIndex(parametersIndex);
                    }
                }
            }
            if (m instanceof MappingCallbacks)
            {
                mc.add(m);
            }
        }

        /**
         * Consumes a mapping not associated to a field.
         * @param m The mapping.
         * @param mappingType the Mapping type
         */
        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
            if (mappingType == MappingConsumer.MAPPING_TYPE_VERSION)
            {
                // version column
                if (table.getVersionMapping() != null)
                {
                    String val = ((RDBMSMapping)table.getVersionMapping().getDataStoreMapping(0)).getUpdateInputParameter();
                    if (columnNames.length() > 0)
                    {
                        columnNames.append(',');
                        columnValues.append(',');
                    }
                    columnNames.append(((Column)table.getVersionMapping().getDataStoreMapping(0).getDatastoreField()).getIdentifier());
                    columnValues.append(val);

                    versionStatementExpressionIndex = new StatementExpressionIndex();
                    versionStatementExpressionIndex.setMapping(table.getVersionMapping());
                    int[] param = { paramIndex++ };
                    versionStatementExpressionIndex.setParameterIndex(param);
                }
                else
                {
                    versionStatementExpressionIndex = null;
                }
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_DISCRIMINATOR)
            {
                // discriminator column
                if (table.getDiscriminatorMapping() != null)
                {
                    String val = ((RDBMSMapping)table.getDiscriminatorMapping().getDataStoreMapping(0)).getUpdateInputParameter();

                    if (columnNames.length() > 0)
                    {
                        columnNames.append(',');
                        columnValues.append(',');
                    }
                    columnNames.append(((Column)table.getDiscriminatorMapping().getDataStoreMapping(0).getDatastoreField()).getIdentifier());
                    columnValues.append(val);
                    discriminatorStatementExpressionIndex = new StatementExpressionIndex();
                    discriminatorStatementExpressionIndex.setMapping(table.getDiscriminatorMapping());
                    int[] param = { paramIndex++ };
                    discriminatorStatementExpressionIndex.setParameterIndex(param);
                }
                else
                {
                    discriminatorStatementExpressionIndex = null;
                }
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
            {
                if (table.getIdentityType() == IdentityType.DATASTORE)
                {
                    if (!table.isObjectIDDatastoreAttributed() || !table.isBaseDatastoreClass())
                    {
                        Iterator iterator = key.getColumns().iterator();
                        if (columnNames.length() > 0)
                        {
                            columnNames.append(',');
                            columnValues.append(',');
                        }
                        columnNames.append(((Column) iterator.next()).getIdentifier().toString());
                        columnValues.append("?");
                        paramIndex++;
                    }
                    else
                    {
                        hasAutoIncrementColumn = true;
                    }
                }
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_FK)
            {
                // External FK mapping (1-N uni)
                externalFKStmtExprIndex = processExternalMapping(m, statementExpressionIndex, externalFKStmtExprIndex);
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_FK_DISCRIM)
            {
                // External FK discriminator mapping (1-N uni with shared FK)
                externalFKDiscrimStmtExprIndex = processExternalMapping(m, statementExpressionIndex, externalFKDiscrimStmtExprIndex);
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_INDEX)
            {
                // External FK order mapping (1-N uni List)
                externalOrderStmtExprIndex = processExternalMapping(m, statementExpressionIndex, externalOrderStmtExprIndex);
            }
        }

        /**
         * Convenience method to process an external mapping.
         * @param mapping The external mapping
         * @param fieldStmtExprIndex The indices for the fields
         * @param stmtExprIndex The current external mapping indices
         * @return The updated external mapping indices
         */
        private StatementExpressionIndex[] processExternalMapping(JavaTypeMapping mapping, 
                StatementExpressionIndex[] fieldStmtExprIndex, StatementExpressionIndex[] stmtExprIndex)
        {
            // Check that we dont already have this as a field
            for (int i=0;i<fieldStmtExprIndex.length;i++)
            {
                if (fieldStmtExprIndex[i] != null && fieldStmtExprIndex[i].getMapping() == mapping)
                {
                    // Already present so ignore it
                    return stmtExprIndex;
                }
            }

            int pos = 0;
            if (stmtExprIndex == null)
            {
                stmtExprIndex = new StatementExpressionIndex[1];
                pos = 0;
            }
            else
            {
                // Check that we dont already have this external order mapping (shared?)
                for (int i=0;i<stmtExprIndex.length;i++)
                {
                    if (stmtExprIndex[i].getMapping() == mapping)
                    {
                        // Shared order mapping so ignore it
                        return stmtExprIndex;
                    }
                }

                StatementExpressionIndex[] tmpStmtExprIndex = stmtExprIndex;
                stmtExprIndex = new StatementExpressionIndex[tmpStmtExprIndex.length+1];
                for (int i=0;i<tmpStmtExprIndex.length;i++)
                {
                    stmtExprIndex[i] = tmpStmtExprIndex[i];
                }
                pos = tmpStmtExprIndex.length;
            }
            stmtExprIndex[pos] = new StatementExpressionIndex();
            stmtExprIndex[pos].setMapping(mapping);
            int[] param = new int[mapping.getNumberOfDatastoreFields()];
            for (int i=0;i<mapping.getNumberOfDatastoreFields();i++)
            {
                if (columnNames.length() > 0)
                {
                    columnNames.append(',');
                    columnValues.append(',');
                }
                columnNames.append(((Column)mapping.getDataStoreMapping(i).getDatastoreField()).getIdentifier());
                columnValues.append(((RDBMSMapping)mapping.getDataStoreMapping(i)).getUpdateInputParameter());
                param[i] = paramIndex++;
            }
            stmtExprIndex[pos].setParameterIndex(param);
            
            return stmtExprIndex;
        }

        /**
         * @return Returns the mappingCallbacks.
         */
        public List getMappingCallbacks()
        {
            return mc;
        }

        /**
         * Obtain the field numbers which values are to be provided by the StateManager
         * @return the array of field numbers 
         */        
        public int[] getFieldsToBeProvided()
        {
            int[] fieldsToBeProvided = new int[ftbp.size()];
            for (int i = 0; i < ftbp.size(); ++i)
            {
                fieldsToBeProvided[i] = ((Integer) ftbp.get(i)).intValue();
            }
            return fieldsToBeProvided;
        }

        /**
         * Obtain the primary key field numbers which values are to be provided by the StateManager
         * @return the array of primary key field numbers 
         */        
        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }

        /**
         * Obtain the field numbers which values are to be retrieved
         * @return the array of field numbers 
         */        
        public int[] getFieldsToBeRetrieved()
        {
            int[] fieldsToBeRetrieved = new int[ftbr.size()];
            for (int i = 0; i < ftbr.size(); ++i)
            {
                fieldsToBeRetrieved[i] = ((Integer) ftbr.get(i)).intValue();
            }
            return fieldsToBeRetrieved;
        }

        /** 
         * Obtain the StatementExpressionIndex
         * @return the array of StatementExpressionIndex indexed by absolute field numbers
         */
        public StatementExpressionIndex[] getStatementExpressionIndex()
        {
            return statementExpressionIndex;
        }

        /** 
         * Obtain the StatementExpressionIndex for the "retrieved" fields.
         * @return the array of StatementExpressionIndex indexed by absolute field numbers
         */
        public StatementExpressionIndex[] getRetrievedStatementExpressionIndex()
        {
            return retrievedStatementExpressionIndex;
        }

        /** 
         * Obtain the StatementExpressionIndex for the version
         * @return the StatementExpressionIndex
         */
        public StatementExpressionIndex getVersionStatementExpressionIndex()
        {
            return versionStatementExpressionIndex;
        }

        /** 
         * Obtain the StatementExpressionIndex for the discriminator
         * @return the StatementExpressionIndex
         */
        public StatementExpressionIndex getDiscriminatorStatementExpressionIndex()
        {
            return discriminatorStatementExpressionIndex;
        }

        /** 
         * Obtain the StatementExpressionIndex for any external FKs
         * @return the StatementExpressionIndex
         */
        public StatementExpressionIndex[] getExternalFKStatementExpressionIndex()
        {
            return externalFKStmtExprIndex;
        }

        /** 
         * Obtain the StatementExpressionIndex for any external FK disciminators
         * @return the StatementExpressionIndex
         */
        public StatementExpressionIndex[] getExternalFKDiscrimStatementExpressionIndex()
        {
            return externalFKDiscrimStmtExprIndex;
        }

        /** 
         * Obtain the StatementExpressionIndex for any external indexes
         * @return the StatementExpressionIndex
         */
        public StatementExpressionIndex[] getExternalOrderStatementExpressionIndex()
        {
            return externalOrderStmtExprIndex;
        }

        /**
         * Obtain the insert statement
         * @return the SQL statement
         */
        public String getInsertStmt()
        {
            // Construct the statement for the INSERT
            if (columnNames.length() > 0 && columnValues.length() > 0)
            {
                return "INSERT INTO " + table.toString() + " (" + columnNames + ") VALUES (" + columnValues + ")";
            }
            else
            {
                // No columns in the INSERT statement
                return ((RDBMSAdapter)table.getStoreManager().getDatastoreAdapter()).getInsertStatementForNoColumns((Table)table);
            }
        }
    }
}
