/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOObjectNotFoundException;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.FetchStatement;
import org.jpox.store.OID;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.rdbms.RDBMSFetchStatement;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.util.JPOXLogger;

/**
 * Class to locate a record in the data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * select operation.
 *
 * @version $Revision: 1.20 $ 
 **/
public class LocateRequest extends Request
{
    /** Index of the statements. */
    private final MappingStatementIndex mappingStatementIndex;

    /** SQL statement for the locate. */
    private final String locateStmt;

    /** Field numbers of the PK fields to be provided. */
    private final int[] pkFieldsToBeProvided;

    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query.
     * @param table The Class Table representing the datastore table to retrieve
     **/
    public LocateRequest(DatastoreClass table)
    {
        super(table);

        mappingStatementIndex = new MappingStatementIndex();
        FetchStatement fetchStmt = new RDBMSFetchStatement(table);
        FetchMappingConsumer consumer = new FetchMappingConsumer(fetchStmt);
        fetchStmt.select("1");
        table.provideDatastoreIdMappings(consumer);
        table.providePrimaryKeyMappings(consumer);

        locateStmt = fetchStmt.toString();
        pkFieldsToBeProvided = consumer.getPrimaryKeyFieldsToBeProvided();
    }

    /**
     * Method performing the retrieval of the record from the datastore. Takes
     * the constructed retrieval query and populates with the specific record
     * information.
     *
     * @param sm The state manager for the record to be retrieved
     **/
    public void execute(StateManager sm)
    {
        if (locateStmt != null)
        {
            PersistenceManager pm = sm.getPersistenceManager();
            RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
            try
            {
                Connection conn = storeMgr.getConnection(pm, false, false);

                try
                {
                    PreparedStatement ps = storeMgr.getStatement(conn, locateStmt, true);

                    try
                    {
                        if (sm.getInternalObjectId() instanceof OID)
                        {
                            // datastore identity
                            table.getDataStoreObjectIdMapping().setObject(pm, ps, mappingStatementIndex.getDatastoreId().getParameterIndex(), 
                                sm.getInternalObjectId());
                        }
                        else
                        {
                            // application identity - provide primary key fields
                            sm.provideFields(pkFieldsToBeProvided, new ParameterSetter(sm, ps, mappingStatementIndex.getFields(), false));
                        }

                        ResultSet rs = storeMgr.executeStatementQuery(locateStmt, ps);
                        try
                        {
                            if (!rs.next())
                            {
                                JPOXLogger.RDBMS_SQL.warn(LOCALISER.msg("RDBMS.Manager.ObjectWithIdNotFound",sm.getInternalObjectId()));
                                throw new JDOObjectNotFoundException("No such database row", sm.getInternalObjectId());
                            }
                        }
                        finally
                        {
                            rs.close();
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    storeMgr.releaseConnection(pm, conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException("Locate request failed: " + locateStmt, e);
            }
        }
    }
    
    private class FetchMappingConsumer implements MappingConsumer
    {
        private int paramIndex = 1;

        final FetchStatement fetchStmt;

        /* pk fields to be provided */
        List pkbp = new ArrayList();
        
        boolean initialized = false;
        
        /**
         * @param fetchStmt
         */
        public FetchMappingConsumer(final FetchStatement fetchStmt)
        {
            super();
            this.fetchStmt = fetchStmt;
        }
        
        public void preConsumeMapping(int highestFieldNumber)
        {
            if (!initialized)
            {
                mappingStatementIndex.setPrimaryKeys(new StatementExpressionIndex[highestFieldNumber]);
                mappingStatementIndex.setFields(new StatementExpressionIndex[highestFieldNumber]);
                initialized = true;
            }
        }

        public void consumeMapping(JavaTypeMapping m, AbstractPropertyMetaData fmd)
        {
            if (m.includeInFetchStatement())
            {
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setMapping(m);

                /*
                 * check if the our column is primary key. if the
                 * first column for the mapping is primary key,
                 * all columns for the same mapping must be primary key
                 */
                if (fmd.isPrimaryKey())
                {
                    int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];
                    for (int j = 0; j < parametersIndex.length; j++)
                    {
                        fetchStmt.andCondition(fetchStmt.referenceDatastoreField(m.getDataStoreMapping(j).getDatastoreField()) + " = ?");
                        
                        pkbp.add(new Integer(fmd.getAbsoluteFieldNumber()));
                        
                        parametersIndex[j] = paramIndex++;
                    }
                    mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setParameterIndex(parametersIndex);
                }
            } 
          
        }
        
        /**
         * Consumes a mapping not associated to a field
         * @param m
         * @param mappingType the Mapping type
         */
        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
	        if (mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
	        {            
	            fetchStmt.andCondition(fetchStmt.referenceDatastoreField((DatastoreField) key.getColumns().get(0)) + " = ?");
	            mappingStatementIndex.getDatastoreId().setParameterIndex(new int[] {paramIndex} );   
	            paramIndex++;
	        }
        }
        
        /**
         * Obtain the primary key field numbers which values are to be provided by the StateManager
         * @return the array of primary key field numbers 
         */
        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }       
    }
}