/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - updated to support inherited objects
2003 Andy Jefferson - revised logging
2004 Andy Jefferson - merged IteratorStmt and GetStmt into GetRangeStmt
2005 Andy Jefferson - added embedded PC element capability
2005 Andy Jefferson - added dependent-element when removed from collection
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.ListIterator;

import javax.jdo.JDODataStoreException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.mapping.EmbeddedElementPCMapping;
import org.jpox.store.mapping.ReferenceMapping;
import org.jpox.store.mapping.SerialisedPCMapping;
import org.jpox.store.mapping.SerialisedReferenceMapping;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.store.rdbms.RDBMSTransaction;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.JoinTable;
import org.jpox.store.scostore.ListStore;
import org.jpox.util.SQLWarnings;

/**
 * Abstract representation of a backing store for a List.
 *
 * @version $Revision: 1.63.2.3 $
 **/
abstract class AbstractListStore extends AbstractCollectionStore implements ListStore
{
    protected String listName;

    protected String addStmt;
    protected String removeAtStmt;
    protected String setStmt;
    protected String shiftStmt;
    protected String indexOfStmt;
    protected String lastIndexOfStmt;

    /**
     * Constructor. Protected to prevent instantiation.
     * @param storeMgr Manager for the store
     * @param clr ClassLoader resolver
     **/
    protected AbstractListStore(StoreManager storeMgr, ClassLoaderResolver clr)
    {
        super(storeMgr, clr);
    }

    /**
     * Method to initialise the statements being used.
     **/
    protected void initialiseStatements()
    {
        super.initialiseStatements();

        addStmt = getAddStmt();

        removeAtStmt = getRemoveAtStmt();
        indexOfStmt = getIndexOfStmt();
        lastIndexOfStmt = getLastIndexOfStmt();
        setStmt = getSetStmt();
        shiftStmt = getShiftStmt();
    }

    // -------------------------- List Method implementations ------------------

    /**
     * Accessor for an iterator through the list elements.
     * @param sm State Manager for the container.
     * @return The Iterator
     **/
    public Iterator iterator(StateManager sm)
    {
        return listIterator(sm);
    }

    /**
     * Accessor for an iterator through the list elements.
     * @param sm State Manager for the container.
     * @return The List Iterator
     **/
    public ListIterator listIterator(StateManager sm)
    {
        return listIterator(sm, 0);
    }

    /**
     * Accessor for an iterator through the list elements.
     * @param ownerSM State Manager for the container.
     * @param start The start point in the list.
     * @return The List Iterator
     **/
    public ListIterator listIterator(StateManager ownerSM, int start)
    {
        // Create the basic statement (without any selected columns)
        QueryExpression stmt = null;
        if (start >= 0)
        {
            stmt = getIteratorStatement(ownerSM, start, -1);
        }
        else
        {
            stmt = getIteratorStatement(ownerSM, -1, -1);
        }

        // Add the required field selections to the statement
        ResultObjectFactory rof = newResultObjectFactory(ownerSM, stmt, false, true);

        PersistenceManager pm = ownerSM.getPersistenceManager();
        RDBMSTransaction tx = (RDBMSTransaction) pm.currentTransaction();
        String statement = stmt.toStatementText(tx.useUpdateLockOnFetch()).toString();

        ListIterator iter;
        try
        {
            Connection conn = storeMgr.getConnection(pm, false, false);

            try
            {
                PreparedStatement ps = stmt.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm, conn);
                try
                {
                    ResultSet rs = storeMgr.executeStatementQuery(statement, ps);
                    try
                    {
                        iter = new ListStoreIterator(ownerSM, rs, rof);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.IteratorRequestFailed",statement),e);
        }

        return iter;
    }

    /**
     * Inner class providing iterator.
     **/
    private class ListStoreIterator implements ListIterator
    {
        private final StateManager sm;
        private final PersistenceManager pm;
        private final ListIterator delegate;
        private Object lastElement = null;
        private int currentIndex = -1;

        /**
         * Constructor
         * @param sm the StateManager
         * @param rs the ResultSet
         * @param rof the Query.ResultObjectFactory
         * @throws SQLException
         */
        public ListStoreIterator(StateManager sm, ResultSet rs, ResultObjectFactory rof) 
        throws SQLException
        {
            this.sm = sm;
            this.pm = sm.getPersistenceManager();

            ArrayList results = new ArrayList();
            if (rs != null)
            {
                while (rs.next())
                {
                    Object nextElement;
                    if (elementsAreEmbedded || elementsAreSerialised)
                    {
                        int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                        for (int i = 0; i < param.length; ++i)
                        {
                            param[i] = i + 1;
                        }

                        if (elementMapping instanceof SerialisedPCMapping ||
                            elementMapping instanceof SerialisedReferenceMapping ||
                            elementMapping instanceof EmbeddedElementPCMapping)
                        {
                            // Element = Serialised
                            int ownerFieldNumber = -1;
                            if (containerTable != null)
                            {
                                ownerFieldNumber = ((JoinTable)containerTable).getOwnerFieldMetaData().getAbsoluteFieldNumber();
                            }
                            nextElement = elementMapping.getObject(pm, rs, param, sm, ownerFieldNumber);
                        }
                        else
                        {
                            // Element = Non-PC
                            nextElement = elementMapping.getObject(pm, rs, param);
                        }
                    }
                    else if (elementMapping instanceof ReferenceMapping)
                    {
                        // Element = Reference (Interface/Object)
                        int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                        for (int i = 0; i < param.length; ++i)
                        {
                            param[i] = i + 1;
                        }
                        nextElement = elementMapping.getObject(pm, rs, param);
                    }
                    else
                    {
                        nextElement = rof.getObject(pm, rs, pm.getClassLoaderResolver().classForName(elementType));
                    }

                    results.add(nextElement);
                }
            }
            delegate = results.listIterator();
        }

        public void add(Object o)
        {
            currentIndex = delegate.nextIndex();
            AbstractListStore.this.add(sm, currentIndex, o);
            delegate.add(o);
            lastElement = null;
        }

        public boolean hasNext()
        {
            return delegate.hasNext();
        }

        public boolean hasPrevious()
        {
            return delegate.hasPrevious();
        }

        public Object next()
        {
            currentIndex = delegate.nextIndex();
            lastElement = delegate.next();

            return lastElement;
        }

        public int nextIndex()
        {
            return delegate.nextIndex();
        }

        public Object previous()
        {
            currentIndex = delegate.previousIndex();
            lastElement = delegate.previous();

            return lastElement;
        }

        public int previousIndex()
        {
            return delegate.previousIndex();
        }

        public synchronized void remove()
        {
            if (lastElement == null)
            {
                throw new IllegalStateException("No entry to remove");
            }

            AbstractListStore.this.remove(sm, currentIndex);
            delegate.remove();

            lastElement = null;
            currentIndex = -1;
        }

        public synchronized void set(Object o)
        {
            if (lastElement == null)
            {
                throw new IllegalStateException("No entry to replace");
            }

            AbstractListStore.this.set(sm, currentIndex, o, true);
            delegate.set(o);

            lastElement = o;
        }
    }

    /**
     * Method to add an element to the List.
     * @param sm The state manager
     * @param element The element to remove
     * @return Whether it was added successfully.
     **/
    public boolean add(StateManager sm, Object element)
    {
        return internalAdd(sm, 0, true, Collections.singleton(element), false);
    }

    /**
     * Method to add an element to the List.
     * @param sm The state manager.
     * @param index The location to add at
     * @param element The element to add.
     **/
    public void add(StateManager sm, int index, Object element)
    {
        internalAdd(sm, index, false, Collections.singleton(element), false);
    }

    /**
     * Method to add a collection of elements to the List.
     * @param sm The state manager
     * @param elements The elements to remove
     * @return Whether they were added successfully.
     **/
    public boolean addAll(StateManager sm, Collection elements)
    {
        return internalAdd(sm, 0, true, elements, false);
    }

    /**
     * Method to add all elements from a Collection to the List.
     * @param sm The state manager
     * @param index The location to add at
     * @param c The collection
     * @return Whether it was successful
     **/
    public boolean addAll(StateManager sm, int index, Collection c)
    {
        return internalAdd(sm, index, false, c, false);
    }

    /**
     * Method to add all elements from a Collection to the List.
     * @param sm The state manager
     * @param c The collection
     * @param initialising Whether we are initialising the list with these elements
     * @return Whether it was successful
     **/
    public boolean addAll(StateManager sm, Collection c, boolean initialising)
    {
        return internalAdd(sm, 0, false, c, initialising);
    }

    /**
     * Method to retrieve an element from the List.
     * @param sm The state manager
     * @param index The location of the element.
     * @return The object
     **/
    public Object get(StateManager sm, int index)
    {
        java.util.List list = internalGetRange(sm, index, index);
        if (list == null || list.size() == 0)
        {
            return null;
        }

        return list.get(0);
    }

    /**
     * Accessor for the indexOf an object in the List.
     * @param sm The state manager
     * @param element The element.
     * @return The index
     **/
    public int indexOf(StateManager sm, Object element)
    {
        return internalIndexOf(sm, element, indexOfStmt);
    }

    /**
     * Method to retrieve the last index of an object in the list.
     * @param sm The state manager.
     * @param element The object
     * @return The last index
     **/
    public int lastIndexOf(StateManager sm, Object element)
    {
        return internalIndexOf(sm, element, lastIndexOfStmt);
    }

    /**
     * Method to remove an element from the List.
     * @param sm The state manager
     * @param element The element to remove.
     * @return Whether it was removed successfully.
     **/
    public boolean remove(StateManager sm, Object element)
    {
        if (!validateElementForReading(sm, element))
        {
            return false;
        }

        // Get the indices of the elements to remove in reverse order (highest first)
        // This is done because the element could be duplicated in the list.
        Collection elements = new ArrayList();
        elements.add(element);
        int[] indices = getIndicesOf(sm, elements);

        boolean modified = false;

        // Remove each element in turn, doing the shifting of indexes each time
        // TODO : Change this to remove all in one go and then shift once
        for (int i=0;i<indices.length;i++)
        {
            removeAt(sm, indices[i]);
            modified = true;
        }

        CollectionMetaData collmd = ownerFieldMetaData.getCollection();
        if (collmd.isDependentElement() && !collmd.isEmbeddedElement())
        {
            // Delete the element if it is dependent
            // TODO What if this element never existed in the List?
            sm.getPersistenceManager().deletePersistent(element);
        }

        return modified;
    }

    /**
     * Remove all elements from a collection from the association owner vs
     * elements.
     * TODO : Change the query to do all in one go for efficiency. Currently
     * removes an element and shuffles the indexes, then removes an element
     * and shuffles the indexes, then removes an element and shuffles the
     * indexes etc ... a bit inefficient !!!
     * @param sm State Manager for the container
     * @param elements Collection of elements to remove 
     * @return Whether the database was updated 
     */
    public boolean removeAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        boolean modified = false;

        // Get the indices of the elements to remove in reverse order (highest first)
        int[] indices = getIndicesOf(sm,elements);

        // Remove each element in turn, doing the shifting of indexes each time
        // TODO : Change this to remove all in one go and then shift once
        for (int i=0;i<indices.length;i++)
        {
            removeAt(sm, indices[i]);
            modified = true;
        }

        if (ownerFieldMetaData.getCollection().isDependentElement())
        {
            // "delete-dependent" : delete elements if the collection is marked as dependent
            // TODO What if the collection contains elements that are not in the List ? should not delete them
            sm.getPersistenceManager().deletePersistentAll(elements);
        }

        return modified;
    }

    /**
     * Method to remove an object at a location in the List.
     * @param sm The state manager.
     * @param index The location
     * @return The object that was removed. 
     **/
    public Object remove(StateManager sm, int index)
    {
        Object element = get(sm, index);
        removeAt(sm, index);

        CollectionMetaData collmd = ownerFieldMetaData.getCollection();
        if (collmd.isDependentElement() && !collmd.isEmbeddedElement())
        {
            if (!contains(sm, element))
            {
                // Delete the element if it is dependent and doesnt have a duplicate entry in the list
                sm.getPersistenceManager().deletePersistent(element);
            }
        }

        return element;
    }

    /**
     * Method to set an object in the List.
     * @param sm The state manager
     * @param index The item index
     * @param element What to set it to.
     * @param allowDependentField Whether to allow dependent field deletes
     * @return The value before setting.
     **/
    public Object set(StateManager sm, int index, Object element, boolean allowDependentField)
    {
        validateElementForWriting(sm, element);
        Object o = get(sm, index);

        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            try
            {
                PreparedStatement ps = storeMgr.getStatement(conn, setStmt, false);
                try
                {
                    int jdbcPosition = 1;
                    jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);
                    jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
                    jdbcPosition = populateOrderInStatement(pm, ps, index, jdbcPosition);
                    if (relationDiscriminatorMapping != null)
                    {
                        jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                    }

                    storeMgr.executeStatementUpdate(setStmt, ps);
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.SetRequestFailed",setStmt),e);
        }

        CollectionMetaData collmd = ownerFieldMetaData.getCollection();
        if (collmd.isDependentElement() && !collmd.isEmbeddedElement() && allowDependentField)
        {
            if (o != null && !contains(sm, o))
            {
                // Delete the element if it is dependent and doesnt have a duplicate entry in the list
                sm.getPersistenceManager().deletePersistent(o);
            }
        }

        return o;
    }


    /**
     * Method to retrieve a list of elements in a range.
     * @param sm The state manager.
     * @param start From index (inclusive)
     * @param end To index (exclusive)
     * @return Sub List of elements in this range.
     **/
    public java.util.List subList(StateManager sm, int start, int end)
    {
        return internalGetRange(sm,start,end);
    }

    // -------------------------- Common Statements ----------------------------

    /**
     * Generate statement for getting the index of an item.
     * <PRE>
     * SELECT INDEXCOL FROM LISTTABLE
     * WHERE OWNERCOL=? 
     * AND ELEMENTCOL=? 
     * [AND EMBEDDEDFIELD1=? AND EMBEDDEDFIELD2=? AND ...]
     * [AND DISTINGUISHER=?]
     * ORDER BY INDEXCOL
     * </PRE>
     * @return The Statement for getting the index of an item 
     */
    protected String getIndexOfStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
        }
        stmt.append(" FROM ");
        stmt.append(containerTable.toString());
        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }

        stmt.append(" ORDER BY ");
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
        }
        return stmt.toString();
    }

    /**
     * Generates the statement for getting the index of the last item.
     * <PRE> 
     * SELECT INDEXCOL FROM LISTTABLE 
     * WHERE OWNERCOL=? 
     * AND ELEMENTCOL=?
     * [AND EMBEDDEDFIELD1=? AND EMBEDDEDFIELD2=? AND ...]
     * [AND DISTINGUISHER=?]
     * ORDER BY INDEXCOL DESC
     * </PRE>
     * @return The Statement for getting the last item 
     */
    protected String getLastIndexOfStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
        }
        stmt.append(" FROM ");
        stmt.append(containerTable.toString());
        stmt.append(" WHERE ");

        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }

        stmt.append(" ORDER BY ");
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" DESC ");
        }
        return stmt.toString();
    }

    /**
     * Generates the statement for removing an item.
     * <PRE> 
     * DELETE FROM LISTTABLE
     * WHERE OWNERCOL = ? 
     * AND INDEXCOL = ?
     * [AND DISTINGUISHER=?]
     * </PRE>
     * @return The Statement for removing an item from a position
     */
    protected String getRemoveAtStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("DELETE FROM ");
        stmt.append(containerTable.toString());
        stmt.append(" WHERE ");

        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)orderMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }
        return stmt.toString();
    }

    /**
     * Generates the statement for setting an item.
     * <PRE>
     * UPDATE LISTTABLE SET [ELEMENTCOL = ?]
     * [EMBEDDEDFIELD1=?, EMBEDDEDFIELD2=?, ...]
     * WHERE OWNERCOL = ? 
     * AND INDEXCOL = ?
     * [AND DISTINGUISHER=?]
     * </PRE>
     * @return The Statement for setting an item
     */
    protected String getSetStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(containerTable.toString());
        stmt.append(" SET ");

        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)orderMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }

        return stmt.toString();
    }

    /**
     * Generates the statement for shifting items.
     * <PRE> 
     * UPDATE LISTTABLE SET INDEXCOL = ?
     * WHERE OWNERCOL = ?
     * AND INDEXCOL = ?
     * [AND DISTINGUISHER=?]
     * </PRE>
     * @return The Statement for shifting elements
     */
    protected String getShiftStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(containerTable.toString());
        stmt.append(" SET ");

        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)orderMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            stmt.append(" + ");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
        }

        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)orderMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }
        return stmt.toString();
    }

    /**
     * Generates the statement for getting the indices of a collection of
     * element. Order into descending index order (highest first) so they will
     * NOT be in the same order as they appear in the input collection "elements".
     * <PRE>
     * SELECT INDEXCOL FROM LISTTABLE 
     * WHERE (OWNERCOL=? AND ELEMENT_COL=? [AND DISTINGUISHER=?]) OR
     *       (OWNERCOL=? AND ELEMENT_COL=? [AND DISTINGUISHER=?]) OR
     *       (OWNERCOL=? AND ELEMENT_COL=? [AND DISTINGUISHER=?]) 
     * ORDER BY INDEXCOL DESC
     * </PRE>
     * @param elements The elements to retrieve the indices for. 
     * @return The Statement for getting the indices of the collection.
     */
    protected String getIndicesOfStmt(Collection elements)
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
        }
        stmt.append(" FROM ");
        stmt.append(containerTable.toString());
        stmt.append(" WHERE ");
        Iterator iter=elements.iterator();
        boolean first_element=true;
        while (iter.hasNext())
        {
            iter.next(); // Move to next element

            if (!first_element)
            {
                stmt.append(" OR (");
            }
            else
            {
                stmt.append("(");
            }

            for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
            {
                if (i > 0)
                {
                    stmt.append(" AND ");
                }
                stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }

            for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
            if (relationDiscriminatorMapping != null)
            {
                for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
                {
                    stmt.append(" AND ");
                    stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                    stmt.append(" = ");
                    stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
                }
            }

            stmt.append(")");
            first_element = false;
        }

        stmt.append(" ORDER BY ");
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" DESC");
        }

        return stmt.toString();
    }

    /**
     * Method to prepare the indicesOf statement for use. Populates the various
     * parameters. This is required because the query is built dynamically
     * depending on the number of elements to retrieve the indices for.
     * @param sm State Manager of the container.
     * @param ps The Prepared Statement
     * @param elements Collection of elements
     **/
    protected void prepareIndicesOfStmt(StateManager sm,
                                        PreparedStatement ps,
                                        Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return;
        }

        PersistenceManager pm = sm.getPersistenceManager();
        Iterator iter=elements.iterator();
        int jdbcPosition = 1;
        while (iter.hasNext())
        {
            Object element = iter.next();

            jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
            jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);
            if (relationDiscriminatorMapping != null)
            {
                jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
            }
        }
    }

    /**
     * Method to remove an element from the specified position
     * @param sm The State Manager for the list
     * @param index The index of the element
     */
    protected void removeAt(StateManager sm, int index)
    {
        internalRemoveAt(sm, index, removeAtStmt);
    }

    // -------------------------- Internal Methods -----------------------------

    /**
     * Internal method to remove an object at a location in the List.
     * @param sm The state manager.
     * @param index The location
     * @param stmt The statement to remove the element from the List
     **/
    protected void internalRemoveAt(StateManager sm, int index, String stmt)
    {
        int size = size(sm);

        PersistenceManager pm = sm.getPersistenceManager();
        try
        {
            Connection conn = storeMgr.getConnection(pm, true, false);
            try
            {
                PreparedStatement ps = storeMgr.getStatement(conn, stmt, false);
                try
                {
                    int jdbcPosition = 1;
                    jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
                    jdbcPosition = populateOrderInStatement(pm, ps, index, jdbcPosition);
                    if (relationDiscriminatorMapping != null)
                    {
                        jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                    }

                    int rowsDeleted = storeMgr.executeStatementUpdate(stmt, ps);
                    if (rowsDeleted == 0)
                    {
                        //?? throw exception??
                    }
                }
                finally
                {
                    ps.close();
                }

                // shift down
                if (index != size - 1)
                {
                    ps = storeMgr.getStatement(conn, shiftStmt, false);

                    try
                    {
                        for (int i=index + 1; i<size; i++)
                        {
                            // Shift this index down 1
                            processShift(sm, ps, false, i, -1);
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.RemoveRequestFailed",stmt),e);
        }
    }

    /**
     * Internal method to add element(s) to the List.
     * Performs the add in 2 steps.
     * <ol>
     * <li>Shift all existing elements into their new positions so we can insert.</li>
     * <li>Insert all new elements directly at their desired positions>/li>
     * </ol>
     * Both steps can be batched (separately).
     * @param sm The state manager
     * @param start The start location (if required)
     * @param atEnd Whether to add the element at the end
     * @param c The collection of objects to add.
     * @param initialising Whether we are initialising the list with these elements
     * @return Whether it was successful
     **/ 
    protected boolean internalAdd(StateManager sm, int start, boolean atEnd, Collection c, boolean initialising)
    {
        if (c == null || c.size() == 0)
        {
            return true;
        }

        // Calculate the amount we need to shift any existing elements by
        // This is used where inserting between existing elements and have to shift down all elements after the start point
        int shift = c.size();

        // check all elements are valid for persisting and exist (persistence-by-reachability)
        Iterator iter = c.iterator();
        while (iter.hasNext())
        {
            validateElementForWriting(sm, iter.next());
        }

        int currentListSize = 0;
        if (!initialising)
        {
            // Not initialising so retrieve the current size of the list
            currentListSize = size(sm);
        }

        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            try
            {
                // Shift any existing elements so that we can insert the new element(s) at their position
                if (!atEnd && start != currentListSize)
                {
                    PreparedStatement ps = storeMgr.getStatement(conn, shiftStmt, false);
                    try
                    {
                        boolean batched = allowsBatching() && currentListSize-start > 0;

                        for (int i = currentListSize - 1; i >= start; i--)
                        {
                            // Shift the index for this row by "shift"
                            processShift(sm, ps, batched, i, shift);
                        }
                        if (batched)
                        {
                            storeMgr.executeStatementBatch(shiftStmt, ps);
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                else
                {
                    start = currentListSize;
                }

                // Insert the elements at their required location
                PreparedStatement ps = storeMgr.getStatement(conn, addStmt, false);
                int jdbcPosition = 1;
                try
                {
                    boolean batched = allowsBatching() && c.size() > 1;

                    iter = c.iterator();
                    while (iter.hasNext())
                    {
                        Object element = iter.next();

                        jdbcPosition = 1;
                        jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
                        jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);
                        jdbcPosition = populateOrderInStatement(pm, ps, start, jdbcPosition);
                        if (relationDiscriminatorMapping != null)
                        {
                            jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                        }
                        start++;

                        if (batched)
                        {
                            ps.addBatch();
                        }
                        else
                        {
                            storeMgr.executeStatementUpdate(addStmt, ps);
                        }
                    }
                    if (batched)
                    {
                        storeMgr.executeStatementBatch(addStmt, ps);
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.AddRequestFailed",addStmt),e);
        }
        return true;
    }

    /**
     * Method to process a "shift" statement, updating the index in the list of the specified index.
     * @param ownerSM StateManager of the owner
     * @param ps The PreparedStatement
     * @param batched Whether the statement is batched
     * @param oldIndex The old index
     * @param amount Amount to shift by (negative means shift down)
     * @throws SQLException Thrown if an error occurs
     */
    protected void processShift(StateManager ownerSM, PreparedStatement ps, boolean batched, int oldIndex, int amount)
    throws SQLException
    {
        PersistenceManager pm = ownerSM.getPersistenceManager();
        int jdbcPosition = 1;
        jdbcPosition = populateOrderInStatement(pm, ps, amount, jdbcPosition);
        jdbcPosition = populateOwnerInStatement(ownerSM, pm, ps, jdbcPosition);
        jdbcPosition = populateOrderInStatement(pm, ps, oldIndex, jdbcPosition);
        if (relationDiscriminatorMapping != null)
        {
            jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
        }

        if (batched)
        {
            ps.addBatch();
        }
        else
        {
            storeMgr.executeStatementUpdate(shiftStmt, ps);
        }
    }

    /**
     * Internal method to find the index of an element.
     * @param sm The state manager.
     * @param element The element
     * @param stmt The statement to find the element.
     * @return The index of the element in the List.
     **/
    private int internalIndexOf(StateManager sm, Object element, String stmt)
    {
        validateElementForReading(sm, element);

        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, false, false);
            try
            {
                PreparedStatement ps = storeMgr.getStatement(conn, stmt, false);
                try
                {
                    int jdbcPosition = 1;
                    jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
                    jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);
                    if (relationDiscriminatorMapping != null)
                    {
                        jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                    }

                    ResultSet rs = storeMgr.executeStatementQuery(stmt, ps);
                    try
                    {
                        boolean found = rs.next();
                        if (!found)
                        {
                            SQLWarnings.log(rs);
                            return -1;
                        }
                        int index = rs.getInt(1);
                        SQLWarnings.log(rs);
                        return index;
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.IndexofRequestFailed",stmt),e);
        }
    }

    /**
     * Utility to find the indices of a collection of elements.
     * The returned list are in reverse order (highest index first).
     * @param sm The state manager.
     * @param elements The elements
     * @return The indices of the elements in the List.
     **/
    protected int[] getIndicesOf(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return null;
        }

        Iterator iter=elements.iterator();
        while (iter.hasNext())
        {
            Object element=iter.next();
            validateElementForReading(sm, element);
        }

        String stmt = getIndicesOfStmt(elements);
        int[] indices = new int[elements.size()];
        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, false, false);
            try
            {
                PreparedStatement ps = storeMgr.getStatement(conn, stmt, false);
                try
                {
                    prepareIndicesOfStmt(sm,ps,elements);

                    ResultSet rs = storeMgr.executeStatementQuery(stmt, ps);
                    try
                    {
                        int i=0;
                        while (rs.next())
                        {
                            indices[i++] = rs.getInt(1);
                        }

                        if (i < elements.size())
                        {
                            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.GetIndicesFailed.InvalidElements",stmt));
                        }
                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.IndicesofRequestFailed",stmt),e);
        }

        return indices;
    }

    /**
     * Accessor for the iterator statement. This has 5 modes
     * <UL>
     * <LI>start and end have the same value and &ge;0 then get that element</LI>
     * <LI>start and end have values &ge;0 and are different then get a range</LI>
     * <LI>start is &ge;0 and end is -1 the get all from start (inclusive).</LI>
     * <LI>start is -1 and end is &ge;0 the get all up to end(exclusive).</LI> 
     * <LI>start and end values are both -1 then get all elements</LI> 
     * </UL> 
     * @param ownerSM The StateManager
     * @param startIndex The start position in the List. 
     * @param endIndex The end position in the List. 
     * @return The QueryStatement to iterate through the required elements.
     **/
    protected abstract QueryExpression getIteratorStatement(StateManager ownerSM, int startIndex, int endIndex);

    /**
     * Utility to retrieve (as a List) the elements for the List in a range
     * depending on the specified start/end positions.
     * @param sm State Manager for the container
     * @param start Start position in the List.
     * @param end End position in the List.
     * @return The List of elements matching the range specification.
     **/
    protected java.util.List internalGetRange(StateManager sm,int start,int end)
    {
        java.util.List list = new java.util.ArrayList();

        PersistenceManager pm = sm.getPersistenceManager();
        QueryExpression stmt = getIteratorStatement(sm, start, end);
        ResultObjectFactory getROF = newResultObjectFactory(sm,stmt,false,true);
        RDBMSTransaction tx = (RDBMSTransaction)pm.currentTransaction();
        String statement = stmt.toStatementText(tx.useUpdateLockOnFetch()).toString();

        try
        {
            Connection conn = storeMgr.getConnection(pm, false, false);
            try
            {
                PreparedStatement ps = stmt.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm,conn);
                try
                {
                    ResultSet rs = storeMgr.executeStatementQuery(statement, ps);
                    try
                    {
                        while (rs.next())
                        {
                            Object obj = null;

                            if (elementsAreEmbedded || elementsAreSerialised)
                            {
                                int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                                for (int i = 0; i < param.length; ++i)
                                {
                                    param[i] = i + 1;
                                }
                                obj = elementMapping.getObject(pm, rs, param);
                            }
                            else if (elementMapping instanceof ReferenceMapping)
                            {
                                // Element = Reference (Interface/Object)
                                int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                                for (int i = 0; i < param.length; ++i)
                                {
                                    param[i] = i + 1;
                                }
                                obj = elementMapping.getObject(pm, rs, param);
                            }
                            else
                            {
                                obj = getROF.getObject(pm, rs, pm.getClassLoaderResolver().classForName(elementType));
                            }

                            list.add(obj);
                        }

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.GetRequestFailed",statement),e);
        }

        return list;
    }
}