/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2002 Mike Martin (TJDO)
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - changed to use Logger
2004 Andy Jefferson - moved statements from subclasses to this class.
2005 Andy Jefferson - added embedded PC element capability
2005 Andy Jefferson - added dependent-element when removed from collection
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.mapping.EmbeddedElementPCMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.ReferenceMapping;
import org.jpox.store.mapping.SerialisedReferenceMapping;
import org.jpox.store.mapping.SerialisedPCMapping;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.store.rdbms.RDBMSTransaction;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.JoinTable;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.JPOXLogger;

/**
 * Abstract representation of the backing store for a Set.
 *
 * @version $Revision: 1.70.2.1 $
 **/
abstract class AbstractSetStore extends AbstractCollectionStore implements SetStore
{
    protected String setName;

    protected String addStmt;
    protected String removeStmt;

    /**
     * Constructor.
     * @param storeMgr Manager for the store
     */
    protected AbstractSetStore(StoreManager storeMgr, ClassLoaderResolver clr)
    {
        super(storeMgr, clr);
    }

    /**
     * Method to initialise the statements being used.
     **/
    protected void initialiseStatements()
    {
        super.initialiseStatements();

        removeStmt = getRemoveStmt();
    }

    /**
     * Accessor for the statement for the iterator.
     * @param ownerSM the owner StateManager
     * @return The iterator Query Statement.
     **/
    protected abstract QueryExpression getIteratorStatement(StateManager ownerSM);

    /**
     * Accessor for an iterator for the set.
     * @param ownerSM State Manager for the set. 
     * @return Iterator for the set.
     **/
    public Iterator iterator(StateManager ownerSM)
    {
        // Create the basic statement (without any selected columns)
        QueryExpression stmt = getIteratorStatement(ownerSM);
        if (stmt == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.SCO.IteratorStatementIsNull"));
        }

        // Add the required field selections to the statement and get a factory for extracting them
        ResultObjectFactory rof = newResultObjectFactory(ownerSM, stmt, false, true);

        PersistenceManager pm = ownerSM.getPersistenceManager();
        RDBMSTransaction tx = (RDBMSTransaction) pm.currentTransaction();
        String statement = stmt.toStatementText(tx.useUpdateLockOnFetch()).toString();

        if (statement == null)
        {
            // this happens when collection is null
            throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.SCO.IteratorStatementIsNull"));
        }

        Iterator iter;
        try
        {
            Connection conn = storeMgr.getConnection(pm, false, false);
            try
            {
                PreparedStatement ps = stmt.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm, conn);
                try
                {
                    ResultSet rs = storeMgr.executeStatementQuery(statement, ps);
                    try
                    {
                        iter = new SetStoreIterator(ownerSM, rs, rof);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.IteratorRequestFailed",statement),e);
        }

        return iter;
    }

    /**
     * Inner class representing an iterator of the Set.
     **/
    private class SetStoreIterator implements Iterator
    {
        private final StateManager sm;
        private final PersistenceManager pm;
        private final Iterator delegate;
        private Object lastElement = null;

        /**
         * Constructor
         * @param sm the StateManager
         * @param rs the ResultSet
         * @param rof the Query.ResultObjectFactory
         * @throws SQLException
         */
        public SetStoreIterator(StateManager sm, ResultSet rs, ResultObjectFactory rof)
        throws SQLException
        {
            this.sm = sm;
            this.pm = sm.getPersistenceManager();

            ArrayList results = new ArrayList();
            if (rs != null)
            {
                while (rs.next())
                {
                    Object nextElement;
                    if (elementsAreEmbedded || elementsAreSerialised)
                    {
                        int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                        for (int i = 0; i < param.length; ++i)
                        {
                            param[i] = i + 1;
                        }

                        if (elementMapping instanceof SerialisedPCMapping ||
                            elementMapping instanceof SerialisedReferenceMapping ||
                            elementMapping instanceof EmbeddedElementPCMapping)
                        {
                            // Element = Serialised
                            int ownerFieldNumber = -1;
                            if (containerTable != null)
                            {
                                ownerFieldNumber = ((JoinTable)containerTable).getOwnerFieldMetaData().getAbsoluteFieldNumber();
                            }
                            nextElement = elementMapping.getObject(pm, rs, param, sm, ownerFieldNumber);
                        }
                        else
                        {
                            // Element = Non-PC
                            nextElement = elementMapping.getObject(pm, rs, param);
                        }
                    }
                    else if (elementMapping instanceof ReferenceMapping)
                    {
                        // Element = Reference (Interface/Object)
                        int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                        for (int i = 0; i < param.length; ++i)
                        {
                            param[i] = i + 1;
                        }
                        nextElement = elementMapping.getObject(pm, rs, param);
                    }
                    else
                    {
                        // Element = PC
                        nextElement = rof.getObject(pm, rs, pm.getClassLoaderResolver().classForName(elementType));
                    }
 
                    results.add(nextElement);
                }
            }
            delegate = results.iterator();
        }

        public boolean hasNext()
        {
            return delegate.hasNext();
        }

        public Object next()
        {
            lastElement = delegate.next();

            return lastElement;
        }

        public synchronized void remove()
        {
            if (lastElement == null)
            {
                throw new IllegalStateException("No entry to remove");
            }

            AbstractSetStore.this.remove(sm, lastElement);
            delegate.remove();

            lastElement = null;
        }
    }

    /**
     * Adds one element to the association owner vs elements
     * @param sm State Manager for the container
     * @param element The element to add
     * @return Whether it was successful 
     */
    public boolean add(StateManager sm, Object element)
    {
        validateElementForWriting(sm, element);

        boolean modified = false;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = storeMgr.getConnection(pm, true, false);
            PreparedStatement ps = storeMgr.getStatement(conn, addStmt, false);
            try
            {
                int num = processAddStmt(sm, ps, false, element);
                if (num > 0)
                {
                    modified = true;
                }
            }
            finally
            {
                ps.close();
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            JPOXLogger.RDBMS.error(e);
            String msg = LOCALISER.msg("RDBMS.SCO.AddRequestFailed", addStmt);
            JPOXLogger.RDBMS.error(msg);
            throw new JDODataStoreException(msg, e);
        }

        return modified;
    }
    
    /**
     * Adds all elements from a collection to the association owner vs elements
     * @param sm State Manager for the container
     * @param elements The elements to add
     * @return Whether it was successful 
     **/
    public boolean addAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        boolean modified = false;
        boolean batched = allowsBatching() && elements.size() > 1;

        // Validate all elements for writing
        Iterator iter = elements.iterator();
        while (iter.hasNext())
        {
            validateElementForWriting(sm, iter.next());
        }

        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            PreparedStatement ps = storeMgr.getStatement(conn, addStmt, false);
            try
            {
                // Loop through all elements to be added
                iter = elements.iterator();
                Object element = null;
                while (iter.hasNext())
                {
                    element = iter.next();

                    int num = processAddStmt(sm, ps, false, element);
                    if (num > 0)
                    {
                        modified = true;
                    }
                }

                if (batched)
                {
                    int[] num = storeMgr.executeStatementBatch(addStmt, ps);
                    if (num == null)
                    {
                        modified = false;
                    }
                    else
                    {
                        for (int i=0;i<num.length;i++)
                        {
                            if (num[i] > 0)
                            {
                                modified = true;
                            }
                        }
                    }
                }
            }
            finally
            {
                ps.close();
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            JPOXLogger.RDBMS.error(e);
            String msg = LOCALISER.msg("RDBMS.SCO.AddRequestFailed", addStmt);
            JPOXLogger.RDBMS.error(msg);
            throw new JDODataStoreException(msg, e);
        }

        return modified;
    }

    /**
     * Method to process an "add" statement.
     * @param ownerSM StateManager for the owner
     * @param ps The PreparedStatement
     * @param batched Whether we are batching it
     * @param element The element
     * @return Number of datastore records changed (always 0 if batch since nothing yet changed)
     * @throws SQLException Thrown if an error occurs
     */
    protected int processAddStmt(StateManager ownerSM, PreparedStatement ps, boolean batched, Object element)
    throws SQLException
    {
        int number = 0;
        PersistenceManager pm = ownerSM.getPersistenceManager();

        int jdbcPosition = 1;
        jdbcPosition = populateOwnerInStatement(ownerSM, pm, ps, jdbcPosition);
        jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);
        if (relationDiscriminatorMapping != null)
        {
            jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
        }

        if (batched)
        {
            // Batch the statement
            ps.addBatch();
            number = 1;
        }
        else
        {
            // Execute the statement
            number = storeMgr.executeStatementUpdate(addStmt, ps);
        }

        return number;
    }

    /**
     * Removes the association to one element
     * @param sm State Manager for the container
     * @param element Element to remove
     * @return Whether it was successful 
     */
    public boolean remove(StateManager sm, Object element)
    {
        if (!validateElementForReading(sm, element))
        {
            JPOXLogger.RDBMS.debug("AbstractSetStore::remove element=" + element + " doesn't exist in this Set.");
            return false;
        }

        boolean modified = false;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = storeMgr.getConnection(pm, true, false);
            try
            {
                PreparedStatement ps = storeMgr.getStatement(conn, removeStmt, false);
                try
                {
                    int jdbcPosition = 1;
                    jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
                    jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);
                    if (relationDiscriminatorMapping != null)
                    {
                        jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                    }

                    int rowsDeleted = storeMgr.executeStatementUpdate(removeStmt, ps);
                    modified = (rowsDeleted == 1);
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            JPOXLogger.RDBMS.error(e);
            String msg = LOCALISER.msg("RDBMS.SCO.RemoveRequestFailed",removeStmt);
            JPOXLogger.RDBMS.error(msg);
            throw new JDODataStoreException(msg, e);
        }

        CollectionMetaData collmd = ownerFieldMetaData.getCollection();
        if (collmd.isDependentElement() && !collmd.isEmbeddedElement())
        {
            // TODO Could this element be stored as a duplicate in this set ?
            // Delete the element if it is dependent
            pm.deletePersistent(element);
        }

        return modified;
    }

    /**
     * Remove all elements from a collection from the association owner vs
     * elements. This implementation iterates around the remove() method doing
     * each element 1 at a time. Please refer to the JoinSetStore and
     * FKSetStore for the variations used there. This is used for Map key and value
     * stores.
     * @param sm State Manager for the container
     * @param elements Collection of elements to remove 
     * @return Whether the database was updated
     */
    public boolean removeAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        boolean batched = allowsBatching() && elements.size() > 1;

        // Validate all elements exist
        Iterator iter = elements.iterator();
        while (iter.hasNext())
        {
            Object element = iter.next();
            if (!validateElementForReading(sm, element))
            {
                JPOXLogger.RDBMS.debug("AbstractSetStore::removeAll element=" + element + " doesn't exist in this Set.");
                return false;
            }
        }

        boolean modified = false;
        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            PreparedStatement ps = storeMgr.getStatement(conn, removeStmt, false);
            try
            {
                Object element = null;
                iter = elements.iterator();
                while (iter.hasNext())
                {
                    element = iter.next();
 
                    // Populate statement for this element
                    int rowsDeleted = processRemoveStmt(sm, ps, batched, element);
                    if (rowsDeleted > 0)
                    {
                        modified = true;
                    }
                }
            }
            finally
            {
                ps.close();
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            JPOXLogger.RDBMS.error(e);
            String msg = LOCALISER.msg("RDBMS.SCO.RemoveRequestFailed", removeStmt);
            JPOXLogger.RDBMS.error(msg);
            throw new JDODataStoreException(msg, e);
        }

        return modified;
    }

    /**
     * Method to process a "remove" statement.
     * @param ownerSM StateManager for the owner
     * @param ps The PreparedStatement
     * @param batched Whether we are batching it
     * @param element The element
     * @return Number of records changed (0 if batch since nothing yet changed)
     * @throws SQLException Thrown if an error occurs
     */
    protected int processRemoveStmt(StateManager ownerSM, PreparedStatement ps, boolean batched, Object element)
    throws SQLException
    {
        int number = 0;
        PersistenceManager pm = ownerSM.getPersistenceManager();

        int jdbcPosition = 1;
        jdbcPosition = populateOwnerInStatement(ownerSM, pm, ps, jdbcPosition);
        jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);
        if (relationDiscriminatorMapping != null)
        {
            jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
        }

        if (batched)
        {
            // Batch the statement
            ps.addBatch();
        }
        else
        {
            // Execute the statement
            number = storeMgr.executeStatementUpdate(removeStmt, ps);
        }

        return number;
    }

    // ---------------------------- Internal Methods ---------------------------

    /**
     * Generate statement for deleting an item from the Set.
     * <PRE>
     * DELETE FROM SETTABLE 
     * WHERE OWNERCOL=?
     * AND ELEMENTCOL = ?
     * [AND DISCRIM = ?]
     * </PRE>
     * @return Statement for deleting an item from the Set.
     */
    protected String getRemoveStmt()
    {
        StringBuffer stmt = new StringBuffer();
        String containerAlias = "THIS";
        String joinedElementAlias = "ELEM";
        stmt.append("DELETE FROM ");
        stmt.append(containerTable.toString()).append(" ").append(containerAlias);

        // Add join to element table if required (only allows for 1 element table currently)
        boolean joinedDiscrim = false;
        if (elementInfo != null && elementInfo[0].getDatastoreClass() != containerTable && elementInfo[0].getDiscriminatorMapping() != null)
        {
            joinedDiscrim = true;
            JavaTypeMapping elemIdMapping = elementInfo[0].getDatastoreClass().getIDMapping();
            stmt.append(" INNER JOIN ");
            stmt.append(elementInfo[0].getDatastoreClass().toString()).append(" ").append(joinedElementAlias).append(" ON ");
            for (int i=0;i<elementMapping.getNumberOfDatastoreFields();i++)
            {
                if (i > 0)
                {
                    stmt.append(" AND ");
                }
                stmt.append(containerAlias).append(".");
                stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier());
                stmt.append("=");
                stmt.append(joinedElementAlias).append(".").append(elemIdMapping.getDataStoreMapping(0).getDatastoreField().getIdentifier());
            }
        }

        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(containerAlias).append(".");
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(containerAlias).append(".");
            stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            if (elementsAreSerialised)
            {
                // Can't directly compare serialised element fields
                stmt.append(" LIKE ");
            }
            else
            {
                stmt.append(" = ");
            }
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        if (elementInfo != null && elementInfo[0].getDiscriminatorMapping() != null)
        {
            JavaTypeMapping discrimMapping = elementInfo[0].getDiscriminatorMapping();
            for (int i=0;i<discrimMapping.getNumberOfDatastoreFields();i++)
            {
                stmt.append(" AND ");
                if (joinedDiscrim)
                {
                    stmt.append(joinedElementAlias);
                }
                else
                {
                    stmt.append(containerAlias);
                }
                stmt.append(".");
                stmt.append(discrimMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)discrimMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }
        return stmt.toString();
    }
}