/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Andy Jefferson - updated retrieval of ColumnList's
2003 Andy Jefferson - updated to support inherited elements
2003 Andy Jefferson - fixed sizeStmt to ignore records with no index value
2004 Andy Jefferson - merged IteratorStmt and GetStmt into GetRangeStmt.
2004 Andy Jefferson - moved majority statement construction to AbstractListStore.
2004 Andy Jefferson - fixed use of shift statement
2004 Andy Jefferson - added support for inverse 1-N unidirectional
2005 Andy Jefferson - added dependent-element when removed from collection
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOHelper;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.ClassNameConstants;
import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.FieldValues;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.DiscriminatorIteratorStatement;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.query.UnionIteratorStatement;
import org.jpox.util.JPOXLogger;

/**
 * Representation of an Inverse List as part of a relationship. This class is
 * used where you have a 1-N and the tables are not joined via a join table.
 * There is an owner table and an element table, and the element table has a
 * column being the id of the owner table. This is in contrast to
 * NormalListStore which represents 1-N relationships using a join table.
 * There are 2 possible uses here
 * <UL>
 * <LI><B>bidirectional</B> - where the owner has a List of elements, and
 * the element has an owner. In this case the element class will have an owner field
 * and index which can be updated directly</LI>
 * <LI><B>unidirectional</B> - where the owner has a List of elements, but
 * the element knows nothing about the owner. In this case the element class has no
 * owner field. In this case the owner/index columns in the element table have to be
 * updated directly.</LI>
 * </UL>
 *
 * @version $Revision: 1.21.2.5 $
 **/
public class FKListStore extends AbstractListStore
{
    /** Field number of owner link in element class. */
    private final int ownerFieldNumber;

    private final String unsetStmt;
    private final String removeAtNullifyStmt;
    private final String clearNullifyStmt;

    /** Statement for updating a foreign key in a 1-N unidirectional */
    private final String updateFkStmt;

    /**
     * Constructor.
     * @param fmd Field MetaData for the field that this represents
     * @param storeMgr The Store Manager in use
     * @param clr The ClassLoaderResolver
     */
    public FKListStore(AbstractPropertyMetaData fmd, RDBMSManager storeMgr, ClassLoaderResolver clr)
    {
        super(storeMgr, clr);

        ownerFieldMetaData = fmd;
        CollectionMetaData colmd = fmd.getCollection();
        if (colmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.CollectionMetaDataMissing", fmd.getFullFieldName()));
        }

        // Load the element class
        elementType = colmd.getElementType();
        Class element_class = clr.classForName(elementType);

        if (storeMgr.getPMFContext().getTypeManager().isReferenceType(element_class))
        {
            // Take the metadata for the first implementation of the reference type
            emd = storeMgr.getMetaDataManager().getMetaDataForImplementationOfReference(element_class,null,clr);
            if (emd != null)
            {
                // Pretend we have a relationship with this one implementation
                elementType = emd.getFullClassName();
            }
        }
        else
        {
            // Check that the element class has MetaData
            emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class, clr);
        }
        if (emd == null)
        {
            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.ElementMetaDataNotFound", element_class.getName(), fmd.getFullFieldName()));
        }

        elementInfo = getElementInformationForClass();
        if (elementInfo != null && elementInfo.length > 1)
        {
            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Collection.ElementWithSubclassTableWithMultipleTablesNotSupported", 
                ownerFieldMetaData.getFullFieldName()));
        }

        elementMapping = elementInfo[0].getDatastoreClass().getIDMapping(); // Just use the first element type as the guide for the element mapping
        elementsAreEmbedded = false; // Can't embed element when using FK relation
        elementsAreSerialised = false; // Can't serialise element when using FK relation

        // Get the field in the element table (if any)
        String mappedByFieldName = fmd.getMappedBy();
        if (mappedByFieldName != null)
        {
            // 1-N FK bidirectional
            // The element class has a field for the owner.
            AbstractPropertyMetaData eofmd = emd.getField(mappedByFieldName);
            if (eofmd == null)
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.MappedByNonExistent", fmd.getFullFieldName(), 
                    mappedByFieldName, element_class.getName()));
            }

            // Check that the type of the element "mapped-by" field is consistent with the owner type
            if (!clr.isAssignableFrom(eofmd.getType(), fmd.getAbstractClassMetaData().getFullClassName()))
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.MappedByTypeInvalid", fmd.getFullFieldName(), 
                    eofmd.getFullFieldName(), eofmd.getTypeName(), fmd.getAbstractClassMetaData().getFullClassName()));
            }

            String ownerFieldName = eofmd.getName();
            ownerFieldNumber = emd.getFieldNumberAbsolute(ownerFieldName);
            ownerMapping = elementInfo[0].getDatastoreClass().getFieldMapping(eofmd);
            if (ownerMapping == null)
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Collection.InverseOwnerMappedByFieldNotPresent",
                    fmd.getAbstractClassMetaData().getFullClassName(), fmd.getName(), elementType, ownerFieldName));
            }
            if (isEmbeddedMapping(ownerMapping))
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.InverseOwnerMappedByFieldInvalidType",
                    ownerFieldName, elementType, eofmd.getTypeName(), fmd.getClassName()));
            }
        }
        else
        {
            // 1-N FK unidirectional
            // The element class knows nothing about the owner (but its table has external mappings)
            ownerFieldNumber = -1;
            ownerMapping = elementInfo[0].getDatastoreClass().getExternalMapping(fmd, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK);
            if (ownerMapping == null)
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Collection.InverseOwnerFieldNotPresent", 
                    fmd.getAbstractClassMetaData().getFullClassName(), fmd.getName(), elementType));
            }
        }

        orderMapping = elementInfo[0].getDatastoreClass().getExternalMapping(fmd, MappingConsumer.MAPPING_TYPE_EXTERNAL_INDEX);
        if (orderMapping == null)
        {
            // Check if the user has omitted the order mapping
            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.List.InverseIndexFieldNotPresent", 
                fmd.getAbstractClassMetaData().getFullClassName(), fmd.getName(), elementType));
        }

        relationDiscriminatorMapping = elementInfo[0].getDatastoreClass().getExternalMapping(fmd, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK_DISCRIM);
        if (relationDiscriminatorMapping != null)
        {
            relationDiscriminatorValue = fmd.getValueForExtension("relation-discriminator-value");
            if (relationDiscriminatorValue == null)
            {
                // No value defined so just use the field name
                relationDiscriminatorValue = fmd.getFullFieldName();
            }
        }

        // TODO Cater for multiple element tables
        containerTable = elementInfo[0].getDatastoreClass();
        listName = "inverseList";

        // Generate the statements
        initialiseStatements();
        addStmt = null;
        removeAtNullifyStmt = getRemoveAtNullifyStmt();
        clearNullifyStmt = getClearNullifyStmt();
        updateFkStmt = getUpdateFkStmt();
        unsetStmt = getUnsetStmt();
    }

    /**
     * Generates the statement for setting an item to be at a position.
     * <PRE>
     * UPDATE LISTTABLE SET OWNERCOL=?, INDEXCOL = ? [,DISTINGUISHER=?]
     * WHERE ELEMENTCOL = ?
     * </PRE>
     * @return The Statement for setting an item
     */
    protected String getSetStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        // TODO Allow for multiple element tables
        stmt.append(containerTable.toString());
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)orderMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(",");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }

        stmt.append(" WHERE ");

        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        return stmt.toString();
    }

    /**
     * Generates the statement for unsetting an item from a list position.
     * <PRE> 
     * UPDATE LISTTABLE SET OWNERCOL=NULL, INDEXCOL=-1 [, DISTINGUISHER = NULL]
     * WHERE OWNERCOL = ? AND INDEXCOL = ? [AND DISTINGUISHER = ?]
     * </PRE>
     * @return The Statement for unsetting an item
     */
    protected String getUnsetStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        // TODO Allow for multiple element tables
        stmt.append(containerTable.toString());
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = NULL");
        }

        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append("=-1");
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(",");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = NULL");
            }
        }

        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)orderMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }
        return stmt.toString();
    }

    /**
     * Method to set an object in the List at a position.
     * @param sm The state manager
     * @param index The item index
     * @param element What to set it to.
     * @param allowDependentField Whether to enable dependent-field deletes during the set
     * @return The value before setting.
     **/
    public Object set(StateManager sm, int index, Object element, boolean allowDependentField)
    {
        validateElementForWriting(sm, element, -1); // Last argument means dont set the position on any INSERT
        Object o = get(sm, index);

        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            try
            {
                // Unset the existing object from this position
                PreparedStatement ps = storeMgr.getStatement(conn, unsetStmt, false);
                try
                {
                    int jdbcPosition = 1;
                    jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
                    jdbcPosition = populateOrderInStatement(pm, ps, index, jdbcPosition);
                    if (relationDiscriminatorMapping != null)
                    {
                        jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                    }

                    storeMgr.executeStatementUpdate(unsetStmt, ps);
                }
                finally
                {
                    ps.close();
                }

                // Set the new object at this position
                PreparedStatement ps2 = storeMgr.getStatement(conn, setStmt, false);
                try
                {
                    int jdbcPosition = 1;
                    jdbcPosition = populateOwnerInStatement(sm, pm, ps2, jdbcPosition);
                    jdbcPosition = populateOrderInStatement(pm, ps2, index, jdbcPosition);
                    if (relationDiscriminatorMapping != null)
                    {
                        jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps2, jdbcPosition);
                    }
                    jdbcPosition = populateElementInStatement(pm, ps2, element, jdbcPosition);

                    storeMgr.executeStatementUpdate(setStmt, ps2);
                }
                finally
                {
                    ps2.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.SetRequestFailed",setStmt),e);
        }

        if (ownerFieldMetaData.getCollection().isDependentElement() && allowDependentField)
        {
            if (o != null)
            {
                // Delete the element if it is dependent and doesnt have a duplicate entry in the list
                sm.getPersistenceManager().deletePersistent(o);
            }
        }

        return o;
    }

    /**
     * Utility to update a foreign-key in the element in the case of
     * a unidirectional 1-N relationship.
     * @param sm StateManager for the owner
     * @param element The element to update
     * @param owner The owner object to set in the FK
     * @param index The index position (or -1 if not known)
     * @return Whether it was performed successfully
     */
    private boolean updateElementFk(StateManager sm, Object element, Object owner, int index)
    {
        if (element == null)
        {
            return false;
        }

        boolean retval;
        PersistenceManager pm = sm.getPersistenceManager();
        try
        {
            Connection conn = storeMgr.getConnection(pm, false, false);
            try
            {
                PreparedStatement ps = storeMgr.getStatement(conn, updateFkStmt, false);
                try
                {
                    int jdbcPosition = 1;
                    if (elementInfo.length > 1)
                    {
                        storeMgr.getDatastoreClass(element.getClass().getName(), clr);
                    }
                    if (owner == null)
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbcPosition, ownerMapping), null);
                        jdbcPosition += ownerMapping.getNumberOfDatastoreFields();
                    }
                    else
                    {
                        jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
                    }
                    jdbcPosition = populateOrderInStatement(pm, ps, index, jdbcPosition);
                    if (relationDiscriminatorMapping != null)
                    {
                        jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                    }
                    jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);

                    storeMgr.executeStatementUpdate(updateFkStmt, ps);
                    retval = true;
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.UpdateFkRequestFailed", updateFkStmt),e);
        }

        return retval;
    }

    /**
     * Generate statement for updating the owner, index columns in an inverse 1-N.
     * Will result in the statement
     * <PRE>
     * UPDATE ELEMENTTABLE SET FK_COL_1 = ?, FK_COL_2 = ?, FK_IDX = ? [,DISTINGUISHER=?]
     * WHERE ELEMENT_ID = ?
     * </PRE>
     * when we have a single element table, and
     * <PRE>
     * UPDATE ? SET FK_COL_1=?, FK_COL_2=?, FK_IDX=? [,DISTINGUISHER=?]
     * WHERE ELEMENT_ID=?
     * </PRE>
     * when we have multiple element tables possible
     * @return Statement for updating the owner/index of an element in an inverse 1-N
     */
    private String getUpdateFkStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        if (elementInfo.length > 1)
        {
            stmt.append("?");
        }
        else
        {
            stmt.append(elementInfo[0].getDatastoreClass().toString());
        }
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)orderMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(",");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }

        stmt.append(" WHERE ");

        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        return stmt.toString();
    }

    /**
     * Accessor for the iterator statement to retrieve element(s) in a range 
     * from the List. This has 5 modes
     * <UL>
     * <LI>start and end have the same value and &ge;0 then get that
     * element</LI>
     * <LI>start and end have values &ge;0 and are different then get a
     * range</LI>
     * <LI>start is &ge;0 and end is -1 the get all from start (inclusive).</LI>
     * <LI>start is -1 and end is &ge;0 the get all up to end(exclusive).</LI>
     * <LI>start and end values are both -1 then get all elements</LI> 
     * </UL>
     * @param ownerSM The StateManager
     * @param start_index The start position in the List.
     * @param end_index The end position in the List.
     * @return The QueryStatement.
     **/
    protected QueryExpression getIteratorStatement(StateManager ownerSM, int start_index, int end_index)
    {
        if (elementInfo == null)
        {
            return null;
        }

        final ClassLoaderResolver clr = ownerSM.getPersistenceManager().getClassLoaderResolver();
        QueryExpression stmt = null;
        if (elementInfo.length == 1 && 
            elementInfo[0].getDatastoreClass().getDiscriminatorMetaData() != null && 
            elementInfo[0].getDatastoreClass().getDiscriminatorMetaData().getStrategy() != DiscriminatorStrategy.NONE)
        {
            if (storeMgr.getPMFContext().getTypeManager().isReferenceType(clr.classForName(ownerFieldMetaData.getCollection().getElementType())))
            {                
                // Take the metadata for the first implementation of the reference type
                String[] clsNames = storeMgr.getMetaDataManager().getClassesImplementingInterface(ownerFieldMetaData.getCollection().getElementType(), clr);
                Class[] cls = new Class[clsNames.length];
                for( int i=0; i<clsNames.length; i++)
                {
                    cls[i] = clr.classForName(clsNames[i]);
                }
                stmt = new DiscriminatorIteratorStatement(clr, cls, true, this.storeMgr, true).getQueryStatement();
            }
            else
            {
                stmt = new DiscriminatorIteratorStatement(clr, new Class[] {clr.classForName(elementInfo[0].getClassName())},
                    true, this.storeMgr, true).getQueryStatement();
            }
            iterateUsingDiscriminator = true;
        }
        else
        {
            for (int i=0;i<elementInfo.length;i++)
            {
                final int elementNo = i;
                QueryExpression subStmt = new UnionIteratorStatement(
                    clr,
                    clr.classForName(this.elementType),
                    true,
                    this.storeMgr,
                    new UnionIteratorStatement.AssociationEnd() 
                    {
                        public JavaTypeMapping getMapping()
                        {
                            return elementMapping;
                        }
                        public Class getType()
                        {
                            return clr.classForName(elementType);
                        }
                        public DatastoreContainerObject getDatastoreContainerObject()
                        {
                            return elementInfo[elementNo].getDatastoreClass();
                        }
                        public boolean useJoin()
                        {
                            return false;
                        }                    
                    }).getQueryStatement();
                if (stmt == null)
                {
                    stmt = subStmt;
                }
                else
                {
                    stmt.union(subStmt);
                }
            }
        }

        // Apply condition on join-table owner field to filter by owner
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, ownerSM.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        // Apply condition on distinguisher field to filter by distinguisher (when present)
        if (relationDiscriminatorMapping != null)
        {
            ScalarExpression distinguisherExpr = relationDiscriminatorMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
            ScalarExpression distinguisherVal = relationDiscriminatorMapping.newLiteral(stmt, relationDiscriminatorValue);
            stmt.andCondition(distinguisherExpr.eq(distinguisherVal), true);
        }

        // Add restriction(s) on index values (where required)
        boolean returning_range=false;
        if (start_index == -1 && end_index == -1)
        {
            // Add no more constraints. We want all elements.
            returning_range = true; 
        }
        else if (start_index == end_index && start_index >= 0)
        {
            // Apply condition on join-table owner field to filter by index
            ScalarExpression indexExpr = orderMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
            ScalarExpression indexVal = orderMapping.newLiteral(stmt, new Integer(start_index));
            
            stmt.andCondition(indexExpr.eq(indexVal), true);     
        }
        else
        {
            returning_range = true; 

            // Apply filter to retrieve indexes in the range
            // a). start_index (inclusive) constraint
            if (start_index >= 0)
            { 
                // Apply condition on join-table index field to filter by index
                ScalarExpression indexExpr = orderMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
                ScalarExpression indexVal = orderMapping.newLiteral(stmt, new Integer(start_index));
                stmt.andCondition(indexExpr.gteq(indexVal), true);   
            }
            // b). end_index (exclusive) constraint
            if (end_index >= 0)
            {
                // Apply condition on join-table index field to filter by index
                ScalarExpression indexExpr = orderMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
                ScalarExpression indexVal = orderMapping.newLiteral(stmt, new Integer(end_index));
                stmt.andCondition(indexExpr.lt(indexVal), true); 
            }
        }

        if (returning_range)
        {
            // If returning a range, add ordering by index
		    ScalarExpression exprIndex[] = new ScalarExpression[orderMapping.getNumberOfDatastoreFields()];
		    boolean descendingOrder[] = new boolean[orderMapping.getNumberOfDatastoreFields()];
            exprIndex = orderMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression()).getExpressionList().toArray();
		    stmt.setOrdering(exprIndex, descendingOrder);
        }

        return stmt;
    }

    /**
     * Internal method for adding an item to the List.
     * @param sm The state manager
     * @param startAt The start position
     * @param atEnd Whether to add at the end
     * @param c The Collection of elements to add.
     * @param initialising Whether we are initialising the list with these elements (so we know the size is 0 currently)
     * @return Whether it was successful
     **/
    protected boolean internalAdd(StateManager sm, int startAt, boolean atEnd, Collection c, boolean initialising)
    {
        if (c == null || c.size() == 0)
        {
            return true;
        }

        // Check what we have persistent already
        int currentListSize = 0;
        if (!initialising)
        {
            // Not initialising so retrieve the current size of the list
            currentListSize = size(sm);
        }
        boolean shiftingElements = true;
        if (atEnd || startAt == currentListSize)
        {
            shiftingElements = false;
            startAt = currentListSize; // Not shifting so we insert from the end
        }

        boolean elementsNeedPositioning = false;
        int position = startAt;
        Iterator elementIter = c.iterator();
        while (elementIter.hasNext())
        {
            // Persist any non-persistent objects optionally at their final list position (persistence-by-reachability)
            if (shiftingElements)
            {
                // We have to shift things so dont bother with positioning
                position = -1;
            }
            boolean inserted = validateElementForWriting(sm, elementIter.next(), position);
            if (!inserted || shiftingElements)
            {
                // This element wasnt positioned in the validate so we need to set the positions later
                elementsNeedPositioning = true;
            }
            if (!shiftingElements)
            {
                position++;
            }
        }

        if (shiftingElements)
        {
            // We need to shift existing elements before positioning the new ones
            try
            {
                PersistenceManager pm = sm.getPersistenceManager();
                Connection conn = storeMgr.getConnection(pm, true, false);

                // Calculate the amount we need to shift any existing elements by
                // This is used where inserting between existing elements and have to shift down all elements after the start point
                int shift = c.size();

                try
                {
                    // shift up existing elements after start position by "shift"
                    PreparedStatement ps = storeMgr.getStatement(conn, shiftStmt, false);

                    try
                    {
                        for (int i=currentListSize-1; i>=startAt; i--)
                        {
                            // Shift the index of this row by "shift"
                            processShift(sm, ps, false, i, shift);
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    storeMgr.releaseConnection(pm, conn);
                }
            }
            catch (SQLException e)
            {
                // An error was encountered during the shift process so abort here
                throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.AddRequestFailed", shiftStmt), e);
            }
        }

        if (shiftingElements || elementsNeedPositioning)
        {
            // Some elements have been shifted so the new elements need positioning now, or we already had some
            // of the new elements persistent and so they need their positions setting now
            elementIter = c.iterator();
            while (elementIter.hasNext())
            {
                Object element = elementIter.next();
                updateElementFk(sm, element, sm.getObject(), startAt);
                startAt++;
            }
        }

        return true;
    }

    /**
     * Generates the statement for removing an item by nulling it out.
     * When there is only a single element table the statement will be
     * <PRE>
     * UPDATE LISTTABLE SET OWNERCOL=NULL, INDEXCOL=-1
     * WHERE OWNERCOL = ?
     * AND INDEXCOL = ?
     * [AND DISTINGUISHER = ?]
     * </PRE>
     * and when there are multiple element tables the statement will be
     * <PRE>
     * UPDATE ? SET OWNERCOL=NULL, INDEXCOL=-1
     * WHERE OWNERCOL=?
     * AND INDEXCOL=?
     * [AND DISTINGUISHER = ?]
     * </PRE>
     * @return The Statement for removing an item from a position
     */
    protected String getRemoveAtNullifyStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        if (elementInfo.length > 1)
        {
            stmt.append("?");
        }
        else
        {
            stmt.append(elementInfo[0].getDatastoreClass().toString());
        }
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(", ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString() + "=NULL");
        }
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(", ");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString() + "=-1");
        }
        
        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)orderMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }
        return stmt.toString();
    }

    /**
     * Method to remove an element from the List.
     * Differs from the JoinTable list in that duplicate positions are not possible
     * for an element so it has a simpler job to remove it.
     * @param sm The state manager
     * @param element The element to remove.
     * @return Whether it was removed successfully.
     **/
    public boolean remove(StateManager sm, Object element)
    {
        if (!validateElementForReading(sm, element))
        {
            return false;
        }

        // The element can be at one position only (no duplicates allowed in FK list)
        int index = indexOf(sm, element);
        if (index == -1)
        {
            return false;
        }
        removeAt(sm, index);

        if (ownerFieldMetaData.getCollection().isDependentElement())
        {
            // Delete the element if it is dependent
            sm.getPersistenceManager().deletePersistent(element);
        }

        return true;
    }

    /**
     * Internal method to remove an object at a location in the List.
     * Differs from the Normal List in that it nulls out the owner FK. 
     * @param sm The state manager.
     * @param index The location
     **/
    protected void removeAt(StateManager sm, int index)
    {
        String stmt = removeAtStmt;
        if (ownerMapping.isNullable() && orderMapping.isNullable())
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("RDBMS.SCO.List.RemoveElementFromInverseNull"));
            // TODO When using this statement we need to plug in the element table name, and do it for 
            // all possible element tables
            stmt = removeAtNullifyStmt;
        }
        else
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("RDBMS.SCO.List.RemoveElementFromInverseDelete"));
        }
        internalRemoveAt(sm, index, stmt);
    }

    /**
     * Generates the statement for clearing items by nulling the owner link out.
     * The statement will be
     * <PRE>
     * UPDATE LISTTABLE SET OWNERCOL=NULL, INDEXCOL=-1 [,DISTINGUISHER=NULL]
     * WHERE OWNERCOL=? [AND DISTINGUISHER=?]
     * </PRE>
     * when there is only one element table, and will be
     * <PRE>
     * UPDATE ? SET OWNERCOL=NULL, INDEXCOL=-1 [,DISTINGUISHER=NULL]
     * WHERE OWNERCOL=? [AND DISTINGUISHER=?]
     * </PRE>
     * when there is more than 1 element table.
     * @return The Statement for clearing items for the owner.
     */
    protected String getClearNullifyStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        if (elementInfo.length > 1)
        {
            stmt.append("?");
        }
        else
        {
            stmt.append(elementInfo[0].getDatastoreClass().toString());
        }
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(", ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString() + "=NULL");
        }
        for (int i=0; i<orderMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(", ");
            stmt.append(orderMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString() + "=-1");
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(", ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append("=NULL");
            }
        }

        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }
        return stmt.toString();
    }

    /**
     * Generate statement for getting the size of the container.
     * The order part is only present when an order mapping is used.
     * The discriminator part is only present when the element type has
     * inheritance strategy of "superclass-table" and is Inverse.
     * <PRE>
     * SELECT COUNT(*) FROM CONTAINERTABLE
     * WHERE OWNERCOL=?
     * [AND ORDERCOL IS NOT NULL]
     * [AND (DISCRIMINATOR=? OR DISCRMINATOR=? OR DISCRIMINATOR=?)]
     * </PRE>
     * The discriminator part includes all subclasses of the element type
     * @return The Statement returning the size of the container.
     */
    protected String getSizeStmt()
    {
        // TODO Allow for multiple element tables
        return super.getSizeStmt();
    }

    /**
     * Generate statement for retrieving the contents of the Collection.
     * The discriminator part is only present when the element type has
     * inheritance strategy of "superclass-table" and is Inverse.
     * <PRE> 
     * SELECT OWNERCOL FROM COLLECTIONTABLE 
     * WHERE OWNERCOL=?
     * AND ELEMENTCOL = ?
     * [AND DISCRIMINATOR = ?]
     * </PRE>
     * @return Statement for retrieving the contents of the Collection. 
     */
    protected String getContainsStmt()
    {
        // TODO Allow for multiple element tables
        return super.getContainsStmt();
    }

    /**
     * Generate statement for clearing the container.
     * <PRE>
     * DELETE FROM CONTAINERTABLE
     * WHERE OWNERCOL = ?
     * </PRE>
     * @return Statement for clearing the container.
     */
    protected String getClearStmt()
    {
        // TODO Allow for multiple element tables
        return super.getClearStmt();
    }

    /**
     * Method to clear the List.
     * This is called by the List.clear() method and also when the container object is being deleted
     * and the elements are to be removed (maybe for dependent field).
     * @param ownerSM The state manager
     **/
    public void clear(StateManager ownerSM)
    {
        boolean deleteElements = false;
        if (ownerFieldMetaData.getCollection().isDependentElement())
        {
            // Elements are dependent and can't exist on their own, so delete them all
            JPOXLogger.RDBMS.debug(LOCALISER.msg("RDBMS.SCO.Collection.ClearElementsFromInverseDeleteDependent"));
            deleteElements = true;
        }
        else
        {
            if (ownerMapping.isNullable() && orderMapping.isNullable())
            {
                // Field is not dependent, and nullable so we null the FK
                JPOXLogger.RDBMS.debug(LOCALISER.msg("RDBMS.SCO.Collection.ClearElementsFromInverseNull"));
                deleteElements = false;
            }
            else
            {
                // Field is not dependent, and not nullable so we just delete the elements
                JPOXLogger.RDBMS.debug(LOCALISER.msg("RDBMS.SCO.Collection.ClearElementsFromInverseDelete"));
                deleteElements = true;
            }
        }

        if (deleteElements)
        {
            ownerSM.isLoaded(ownerSM.getObject(), ownerFieldMetaData.getAbsoluteFieldNumber()); // Make sure the field is loaded
            Collection value = (Collection) ownerSM.provideField(ownerFieldMetaData.getAbsoluteFieldNumber());
            if (value != null && !value.isEmpty())
            {
                ownerSM.getPersistenceManager().deletePersistentAll(value.toArray());
            }
        }
        else
        {
            // TODO If the relation is bidirectional we need to clear the owner in the element
            try
            {
                PersistenceManager pm = ownerSM.getPersistenceManager();
                Connection conn = storeMgr.getConnection(pm, true, false);
                try
                {
                    PreparedStatement ps = storeMgr.getStatement(conn, clearNullifyStmt, false);
                    try
                    {
                        int jdbcPosition = 1;
                        jdbcPosition = populateOwnerInStatement(ownerSM, pm, ps, jdbcPosition);
                        if (relationDiscriminatorMapping != null)
                        {
                            jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                        }
                        storeMgr.executeStatementUpdate(clearNullifyStmt, ps);
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    storeMgr.releaseConnection(pm, conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.ClearRequestFailed",clearNullifyStmt),e);
            }
        }
    }

    /**
     * Method to validate that an element is valid for writing to the datastore.
     * TODO This is totally different to FKSetStore.validateElementForWriting. They should be very similar!
     * @param sm StateManager for the List
     * @param element The element to validate
     * @param index The position that the element is being stored at in the list
     * @return Whether the element was inserted
     */
    protected boolean validateElementForWriting(StateManager sm, Object element, int index)
    {
        if (!validateElementType(sm.getPersistenceManager().getClassLoaderResolver(), element))
        {
            throw new ClassCastException(LOCALISER.msg("RDBMS.SCO.Collection.ElementIsInvalid",element.getClass().getName(),elementType));
        }

        boolean inserted = false;
        if (element != null && !elementsAreEmbedded && !elementsAreSerialised)
        {
            PersistenceManager pm = sm.getPersistenceManager();

            final Object newOwner = sm.getObject();
            final StateManager ownerSM = sm;
            final int elementPosition = index;
            if (!JDOHelper.isPersistent(element))
            {
                boolean exists = false;
                if (JDOHelper.isDetached(element))
                {
                    try
                    {
                        // Check if the (attached) object exists in this datastore
                        Object obj = pm.getObjectById(((PersistenceCapable)element).jdoGetObjectId(), true, false, element.getClass().getName());
                        if (obj != null)
                        {
                            // PM.getObjectById creates a dummy object to represent this object and automatically
                            // enlists it in the txn. Evict it to avoid issues with reachability
                            StateManager objSM = pm.findStateManager((PersistenceCapable)obj);
                            if (objSM != null)
                            {
                                pm.evictFromTransaction(objSM);
                            }
                        }
                        exists = true;
                    }
                    catch (JDOObjectNotFoundException onfe)
                    {
                        exists = false;
                    }
                }
                if (!exists)
                {
                    // The new element is not yet persistent. Update the owner field and make it persistent
                    // Making the change before DB insertion avoids an unnecessary UPDATE and allows the 
                    // owner field to be non-nullable.
                    pm.makePersistentInternal(element, new FieldValues()
                        {
                        public void fetchFields(StateManager esm)
                        {
                            boolean isPersistentInterface = storeMgr.getMetaDataManager().isPersistentInterface(elementType);
                            DatastoreClass elementTable = null;
                            if (!isPersistentInterface)
                            {
                                elementTable = storeMgr.getDatastoreClass(elementType, clr);
                            }
                            else
                            {
                                elementTable = storeMgr.getDatastoreClass(storeMgr.getMetaDataManager().getImplementationNameForPersistentInterface(elementType), clr);
                            }
                            if (elementTable != null)
                            {
                                JavaTypeMapping externalFKMapping = elementTable.getExternalMapping(ownerFieldMetaData, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK);
                                if (externalFKMapping != null)
                                {
                                    // The element has an external FK mapping so set the value it needs to use in the INSERT
                                    esm.setExternalFieldValueForMapping(externalFKMapping, ownerSM.getObject());
                                }
                                if (relationDiscriminatorMapping != null)
                                {
                                    esm.setExternalFieldValueForMapping(relationDiscriminatorMapping, relationDiscriminatorValue);
                                }
                                if (orderMapping != null && elementPosition >= 0)
                                {
                                    if (ownerFieldMetaData.getOrderMetaData() != null && ownerFieldMetaData.getOrderMetaData().getMappedBy() != null)
                                    {
                                        // Order is stored in a field of the element so update it with the position
                                        // We support mapped-by fields of types int/long/Integer/Long currently
                                        Object indexValue = null;
                                        if (orderMapping.getFieldMetaData().getTypeName().equals(ClassNameConstants.JAVA_LANG_LONG) ||
                                            orderMapping.getFieldMetaData().getTypeName().equals(ClassNameConstants.LONG))
                                        {
                                            indexValue = new Long(elementPosition);
                                        }
                                        else
                                        {
                                            indexValue = new Integer(elementPosition);
                                        }
                                        esm.replaceField(orderMapping.getFieldMetaData().getAbsoluteFieldNumber(), indexValue);
                                    }
                                    else
                                    {
                                        // Order is stored in surrogate column so save it with the element for later processing
                                        esm.setExternalFieldValueForMapping(orderMapping, new Integer(elementPosition));
                                    }
                                }
                            }
                            if (ownerFieldNumber >= 0)
                            {
                                esm.replaceField(ownerFieldNumber, newOwner);
                            }
                        }
                        public void fetchNonLoadedFields(StateManager sm)
                        {
                        }
                        public FetchPlanImpl getFetchPlanForLoading()
                        {
                            return null;
                        }
                        });
                    StateManager elementSM = pm.findStateManager((PersistenceCapable)element);
                    if (elementSM != null)
                    {
                        // Make sure that this is in the datastore before continuing
                        elementSM.flush();
                    }
                    inserted = true;
                }
            }
            else
            {
                PersistenceCapable pcElement = (PersistenceCapable) element;
                /*
                 * The new element is already persistent.
                 * "Add" the new value to the set by updating its owner field
                 * to the appropriate value. This is done with the same methods
                 * the PC itself would use if the application code modified the
                 * field. It should result in no actual database activity if 
                 * the field was already set to the right value.
                 */
                if (pm != pcElement.jdoGetPersistenceManager())
                {
                    throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.List.WriteInvalidWithDifferentPM"), pcElement.jdoGetObjectId());
                }

                StateManager esm = pm.findStateManager(pcElement);
                if (ownerFieldNumber >= 0)
                {
                    // 1-N bidir

                    // Ensure the current owner value is loaded.
                    esm.isLoaded(pcElement, ownerFieldNumber);

                    // Get the current value.
                    Object oldOwner = esm.provideField(ownerFieldNumber);

                    // Replace with new value.
                    esm.setObjectField(pcElement, ownerFieldNumber, oldOwner, newOwner);
                }
                esm.flush(); // Make sure it is in the datastore now since we will soon update its owner/position
            }
        }
        return inserted;
    }

    // ---------------------------- Query Methods ------------------------------

    /**
     * Utility method to return a new QueryStatement for retrieval of the
     * elements of this List.
     * @param sm StateManager for this object
     * @param candidateClass Class for the element end of the link.
     * @return The QueryStatement
     **/
    public QueryExpression newQueryStatement(StateManager sm, String candidateClass)
    {
        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(elementType,candidateClass))
        {
            throw new IncompatibleQueryElementTypeException(elementType, candidateClass);
        }

        DatastoreClass candidateTable = storeMgr.getDatastoreClass(candidateClass, sm.getPersistenceManager().getClassLoaderResolver());
        QueryExpression stmt = dba.newQueryStatement(candidateTable,sm.getPersistenceManager().getClassLoaderResolver());

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal));

        stmt.select(elementMapping);

        return stmt;
    }

    /**
     * Method used in queries when contains() has been invoked.
     * @param stmt The Query Statement 
     * @param parentStmt the parent Query Statement. If there is no parent, <code>parentStmt</code> must be equals to <code>stmt</code> 
     * @param ownerMapping the mapping for the owner. 
     * @param ownerTe Table Expression for the owner
     * @param listRangeVar   The range variable for the "List" table. 
     * @param filteredElementType The Class Type for the filtered element
     * @param elmExpr The Expression for the element
     * @param elementRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the element table.
     * @return expression to the join
     **/
    public ScalarExpression joinElementsTo(
                        QueryExpression stmt,
                        QueryExpression parentStmt,
    					JavaTypeMapping ownerMapping,
    					LogicSetExpression ownerTe,
    					DatastoreIdentifier listRangeVar,
                        Class filteredElementType,
                        ScalarExpression elmExpr,
                        DatastoreIdentifier elementRangeVar)
    {
        ClassLoaderResolver clr=stmt.getClassLoaderResolver();
        if (!clr.isAssignableFrom(elementType,filteredElementType) &&
            !clr.isAssignableFrom(filteredElementType,elementType))
        {
            throw new IncompatibleQueryElementTypeException(elementType, filteredElementType.getName());
        }

        // Join the element table on the owner ID column.
        DatastoreContainerObject filteredElementTable = storeMgr.getDatastoreClass(filteredElementType.getName(), stmt.getClassLoaderResolver());
        stmt.newTableExpression(filteredElementTable, elementRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,ownerTe);
        DatastoreIdentifier containerRangeVar = listRangeVar;
        if( stmt.getTableExpression(containerRangeVar) == null )
        {
            containerRangeVar = elementRangeVar;
        }
        ScalarExpression ownerSetExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(containerRangeVar));
        stmt.newTableExpression(containerTable, containerRangeVar);

        stmt.andCondition(ownerExpr.eq(ownerSetExpr),true);

        JavaTypeMapping elementTableID = filteredElementTable.getIDMapping();
        return elementTableID.newScalarExpression(stmt,stmt.getTableExpression(containerRangeVar));

    }
}