/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 

Contributors:
2003 Andy Jefferson - added comments
2003 Erik Bengtson - removed unused variable and import
2003 Erik Bengtson - fixed bug [832635] Imcompatiblities in casting
                    inherited classes in query
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - updated retrieval of ColumnLists
2004 Andy Jefferson - addition of removeAll()
2004 Andy Jefferson - moved statements to AbstractSetStore
2004 Andy Jefferson - added support for inverse 1-N unidirectional
2005 Andy Jefferson - added dependent-element when removed from collection
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.FieldValues;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.DiscriminatorIteratorStatement;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.query.UnionIteratorStatement;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * Representation of an Inverse Set as part of a relationship. This class is
 * used where you have a 1-N and the tables are not joined via a link table.
 * That is there is an owner table, and a collection table, and the collection
 * table has a column being the id of the owner table. This is in contrast to
 * NormalSetStore which represents 1-N relationships using a link table.
 * There are 2 possible uses here
 * <UL>
 * <LI><B>bidirectional</B> - where the owner has a Collection of elements, and
 * the element has an owner. In this case the element class will have an owner field
 * which can be updated directly</LI>
 * <LI><B>unidirectional</B> - where the owner has a Collection of elements, but
 * the element knows nothing about the owner. In this case the element class has no
 * owner field. In this case the column in the element table has to be updated
 * directly.</LI>
 * </UL>
 *
 * @version $Revision: 1.14.2.2 $
 **/
public class FKSetStore extends AbstractSetStore
{
    /** Field number of owner link in element class. */
    private final int ownerFieldNumber;

    /** Statement for updating a foreign key in a 1-N unidirectional */
    private final String updateFkStmt;

    /** Statement for nullifying a FK in the element. */
    private final String clearNullifyStmt;

    private final boolean isPersistentInterface;

    /**
     * Constructor for the relationship representation.
     * @param fmd The MetaData for the field that this represents
     * @param storeMgr The RDBMSManager managing the associated datastore.
     * @param clr The ClassLoaderResolver
     **/
    public FKSetStore(AbstractPropertyMetaData fmd, RDBMSManager storeMgr, ClassLoaderResolver clr)
    {
        super(storeMgr, clr);

        ownerFieldMetaData = fmd;
        CollectionMetaData colmd = fmd.getCollection();
        if (colmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.CollectionMetaDataMissing", fmd.getFullFieldName()));
        }

        // Load the element class
        elementType = colmd.getElementType();
        Class element_class = clr.classForName(elementType);

        if (storeMgr.getPMFContext().getTypeManager().isReferenceType(element_class))
        {
            isPersistentInterface = storeMgr.getMetaDataManager().isPersistentInterface(element_class.getName());
            if (isPersistentInterface)
            {
                emd = storeMgr.getMetaDataManager().getMetaDataForInterface(element_class,clr);
            }
            else
            {
                // Take the metadata for the first implementation of the reference type
                emd = storeMgr.getMetaDataManager().getMetaDataForImplementationOfReference(element_class,null,clr);
                if (emd != null)
                {
                    // Pretend we have a relationship with this one implementation
                    //elementType = emd.getFullClassName();
                }
            }
        }
        else
        {
            // Check that the element class has MetaData
            emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class, clr);
            isPersistentInterface = false;
        }
        if (emd == null)
        {
            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.ElementMetaDataNotFound", element_class.getName(), fmd.getFullFieldName()));
        }

        elementInfo = getElementInformationForClass();
        /*
        if (elementInfo != null && elementInfo.length > 1)
        {
            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Collection.ElementWithSubclassTableWithMultipleTablesNotSupported", 
                ownerFieldMetaData.getFullFieldName()));
        }
*/
        elementMapping = elementInfo[0].getDatastoreClass().getIDMapping(); // Just use the first element type as the guide for the element mapping
        elementsAreEmbedded = false; // Can't embed element when using FK relation
        elementsAreSerialised = false; // Can't serialise element when using FK relation

        // Get the field in the element table (if any)
        if (fmd.getMappedBy() != null)
        {
            // 1-N FK bidirectional
            // The element class has a field for the owner.
            AbstractPropertyMetaData eofmd = emd.getField(fmd.getMappedBy());
            if (eofmd == null)
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.MappedByNonExistent", fmd.getFullFieldName(),
                    fmd.getMappedBy(), element_class.getName()));
            }

            // Check that the type of the element "mapped-by" field is consistent with the owner type
            //TODO check does not work if mapped by has relation to super class. Enable this
            //and run the PersistentInterfacesTest to reproduce the issue
            //TODO there is equivalent code in FKList and FKMap that was
            //not commented out. When fixing, add tests for all types of Inverses
            /*
            if (!clr.isAssignableFrom(eofmd.getType(), fmd.getAbstractClassMetaData().getFullClassName()))
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.MappedByTypeInvalid", fmd.getFullFieldName(),
                    eofmd.getFullFieldName(), eofmd.getTypeName(), fmd.getAbstractClassMetaData().getFullClassName()));
            }
            */
            String ownerFieldName = eofmd.getName();
            ownerFieldNumber = emd.getFieldNumberAbsolute(ownerFieldName);
            ownerMapping = elementInfo[0].getDatastoreClass().getFieldMapping(eofmd);
            if (ownerMapping == null)
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Collection.InverseOwnerMappedByFieldNotPresent",
                    fmd.getAbstractClassMetaData().getFullClassName(), fmd.getName(), elementType, ownerFieldName));
            }
            if (isEmbeddedMapping(ownerMapping))
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.InverseOwnerMappedByFieldInvalidType",
                    ownerFieldName, elementType, eofmd.getTypeName(), fmd.getClassName()));
            }
        }
        else
        {
            // 1-N FK unidirectional
            // The element class knows nothing about the owner (but its table has external mappings)
            ownerFieldNumber = -1;
            ownerMapping = elementInfo[0].getDatastoreClass().getExternalMapping(fmd, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK);
            if (ownerMapping == null)
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Collection.InverseOwnerFieldNotPresent", 
                    fmd.getAbstractClassMetaData().getFullClassName(), fmd.getName(), elementType));
            }
        }

        relationDiscriminatorMapping = elementInfo[0].getDatastoreClass().getExternalMapping(fmd, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK_DISCRIM);
        if (relationDiscriminatorMapping != null)
        {
            relationDiscriminatorValue = fmd.getValueForExtension("relation-discriminator-value");
            if (relationDiscriminatorValue == null)
            {
                // No value defined so just use the field name
                relationDiscriminatorValue = fmd.getFullFieldName();
            }
        }

        // TODO Remove use of containerTable - just use elementTable[0] or equivalent
        containerTable = elementInfo[0].getDatastoreClass();
        setName = "inverseSet";

        // Generate the statements
        initialiseStatements();
        addStmt = null;
        removeStmt = null;
        clearStmt = null;
        clearNullifyStmt = getClearNullifyStmt();
        updateFkStmt = getUpdateFkStmt();
    }

    /**
     * Generate statement for getting the size of the container.
     * The order part is only present when an order mapping is used.
     * The discriminator part is only present when the element type has
     * inheritance strategy of "superclass-table" and is Inverse.
     * <PRE>
     * SELECT COUNT(*) FROM CONTAINERTABLE
     * WHERE OWNERCOL=?
     * [AND ORDERCOL IS NOT NULL]
     * [AND (DISCRIMINATOR=? OR DISCRMINATOR=? OR DISCRIMINATOR=?)]
     * </PRE>
     * The discriminator part includes all subclasses of the element type
     * @return The Statement returning the size of the container.
     */
    protected String getSizeStmt()
    {
        // TODO Allow for multiple element tables
        return super.getSizeStmt();
    }

    /**
     * Generate statement for retrieving the contents of the Collection.
     * The discriminator part is only present when the element type has
     * inheritance strategy of "superclass-table" and is Inverse.
     * <PRE> 
     * SELECT OWNERCOL FROM COLLECTIONTABLE 
     * WHERE OWNERCOL=?
     * AND ELEMENTCOL = ?
     * [AND DISCRIMINATOR = ?]
     * </PRE>
     * @return Statement for retrieving the contents of the Collection. 
     */
    protected String getContainsStmt()
    {
        // TODO Allow for multiple element tables
        return super.getContainsStmt();
    }
    /**
     * Generates the statement for clearing items by nulling the owner link out.
     * The statement will be
     * <PRE>
     * UPDATE LISTTABLE SET OWNERCOL=NULL [,DISTINGUISHER=NULL]
     * WHERE OWNERCOL=?
     * </PRE>
     * when there is only one element table, and will be
     * <PRE>
     * UPDATE ? SET OWNERCOL=NULL [,DISTINGUISHER=NULL]
     * WHERE OWNERCOL=?
     * </PRE>
     * when there is more than 1 element table.
     * @return The Statement for clearing items for the owner.
     */
    protected String getClearNullifyStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        if (elementInfo.length > 1)
        {
            //CANNOT USE ? to replace. JDBC drivers does not accept, so we replace 
            //the <TABLE NAME> right before firing the statement to the database
            stmt.append("<TABLE NAME>");
        }
        else
        {
            stmt.append(elementInfo[0].getDatastoreClass().toString());
        }
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(", ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString() + " = NULL");
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(", ");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString() + " = NULL");
            }
        }
        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Generate statement for updating a Foreign Key in an inverse 1-N.
     * The statement generated will be
     * <PRE>
     * UPDATE ELEMENTTABLE SET FK_COL_1=?, FK_COL_2=?, [DISTINGUISHER=?]
     * WHERE ELEMENT_ID = ?
     * </PRE>
     * where there is a single element table, and
     * <PRE>
     * UPDATE ? SET FK_COL_1=?, FK_COL_2=?, [DISTINGUISHER=?]
     * WHERE ELEMENT_ID=?
     * </PRE>
     * where there are more than 1 element tables
     * @return Statement for updating the FK in an inverse 1-N
     */
    private String getUpdateFkStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        if (elementInfo.length > 1)
        {
            //CANNOT USE ? to replace. JDBC drivers does not accept, so we replace 
            //the <TABLE NAME> right before firing the statement to the database
            stmt.append("<TABLE NAME>");
        }
        else
        {
            stmt.append(elementInfo[0].getDatastoreClass().toString());
        }
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        if (relationDiscriminatorMapping != null)
        {
            for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(",");
                stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }

        stmt.append(" WHERE ");
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        return stmt.toString();
    }

    /**
     * Utility to update a foreign-key (and distinguisher) in the element in the case of
     * a unidirectional 1-N relationship.
     * @param sm StateManager for the owner
     * @param element The element to update
     * @param owner The owner object to set in the FK
     * @return Whether it was performed successfully
     */
    private boolean updateElementFk(StateManager sm, Object element, Object owner)
    {
        if (element == null)
        {
            return false;
        }
        validateElementForWriting(sm, element);

        boolean retval;
        PersistenceManager pm = sm.getPersistenceManager();
        try
        {
            Connection conn = storeMgr.getConnection(pm, false, false);
            try
            {
                String stmt = updateFkStmt;
                int jdbcPosition = 1;
                if (elementInfo.length > 1)
                {
                    DatastoreClass table = storeMgr.getDatastoreClass(element.getClass().getName(), clr);
                    if (table != null)
                    {
                        stmt = StringUtils.replaceAll(stmt, "<TABLE NAME>", table.toString());
                    }
                    else
                    {
                        JPOXLogger.JDO.warn(">> InverseArrayStore.updateElementFK : need to set table in statement but dont know table where to store " + element);
                    }
                }                
                PreparedStatement ps = storeMgr.getStatement(conn, stmt, false);
                try
                {

                    if (owner == null)
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbcPosition, ownerMapping), null);
                    }
                    else
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbcPosition, ownerMapping), sm.getObject());
                    }
                    jdbcPosition += ownerMapping.getNumberOfDatastoreFields();
                    if (relationDiscriminatorMapping != null)
                    {
                        jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                    }
                    elementMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbcPosition, elementMapping), element);
                    jdbcPosition += elementMapping.getNumberOfDatastoreFields();

                    storeMgr.executeStatementUpdate(updateFkStmt, ps);
                    retval = true;
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.UpdateFkRequestFailed",updateFkStmt),e);
        }

        return retval;
    }

    /**
     * Method to check if an element is already persistent, or is managed by a different
     * Persistencemanager. If not persistent, this will persist it.
     * @param sm The state manager of this collection
     * @param element The element
     */
    protected void validateElementForWriting(StateManager sm, Object element)
    {
        super.validateElementForWriting(sm,element);
        if (element != null && !elementsAreEmbedded && !elementsAreSerialised)
        {
            if (element instanceof PersistenceCapable)
            {
                // make sure the element is inserted in the database before we write in it
                PersistenceManager pm = sm.getPersistenceManager();
                pm.findStateManager((PersistenceCapable)element).flush();
            }
        }
    }

    /**
     * This method was added sometime during getting "persistence-interfaces" to work.
     * What exactly it does that is not catered for by returning "ownerFieldNumber" is best
     * asked of the person who added it ... since it lacked any javadoc. If it is to represent
     * the "ownerFieldNumber" then why not just give that field the correct value in the first place?
     * @param sm StateManager of the owner
     * @return The "owner field owner" whatever that is
     */
    protected int getOwnerFieldOwner(StateManager sm)
    {
        if (ownerFieldNumber < 0)
        {
            return -1;
        }
        return sm.getClassMetaData().getFieldNumberAbsolute(ownerFieldMetaData.getMappedBy());
    }

    /**
     * Method to add an object to the relationship at the collection end.
     * @param sm StateManager of the owner of the Set
     * @param element Element to be added
     * @return Success indicator
     **/
    public boolean add(StateManager sm, Object element)
    {
        // TODO This doesnt use validateElementForWriting but should instead of copying code etc
        // Should be commonised with FKListStore
        if (element == null)
        {
            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Set.NullElementsNotAllowed"));
        }
        if (!isPersistentInterface && !sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(elementType,element.getClass()))
        {
            throw new ClassCastException(LOCALISER.msg("RDBMS.SCO.Collection.ElementIsInvalid",element.getClass().getName(), 
                ownerFieldMetaData.getFullFieldName(), elementType));
        }

        PersistenceManager pm = sm.getPersistenceManager();
        PersistenceCapable pcElement = (PersistenceCapable) element;
        final Object newOwner = sm.getObject();
        final StateManager ownerSM = sm;

        if (pcElement.jdoIsPersistent())
        {
            /*
             * The new element is already persistent.
             *
             * "Add" the new value to the set by updating its owner field to the
             * appropriate value.  This is done with the same methods the PC
             * itself would use if the application code modified the field.  It
             * should result in no actual database activity if the field was
             * already set to the right value.
             */
            if (pm != pcElement.jdoGetPersistenceManager())
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Set.WriteInvalidWithDifferentPM"),pcElement.jdoGetObjectId());
            }

            StateManager elementSM = pm.findStateManager(pcElement);
            if (getOwnerFieldOwner(elementSM) >= 0)
            {
                // Bidirectional 1-N, so update the owner as appropriate

                // Ensure the current owner value is loaded.
                elementSM.isLoaded(pcElement, getOwnerFieldOwner(elementSM));

                // Get the current value and replace with the new value if it is changed
                Object oldOwner = elementSM.provideField(getOwnerFieldOwner(elementSM));
                if (oldOwner != newOwner)
                {
                    elementSM.setObjectField(pcElement, getOwnerFieldOwner(elementSM), oldOwner, newOwner);
                }

                return oldOwner != newOwner;
            }
            else
            {
                // Unidirectional 1-N, so check if contained already
                boolean contained = contains(sm, element);
                if (!contained)
                {
                    // Not present so set the owner
                    return updateElementFk(sm, pcElement, newOwner);
                }
                else
                {
                    // Already contained so cant add again
                    return false;
                }
            }
        }
        else
        {
            /*
             * The new element is not yet persistent.
             *
             * Update its owner field to the appropriate value and *then* make
             * it persistent.  Making the change before DB insertion avoids an
             * unnecessary UPDATE and allows the owner field to be non-nullable.
             */
            pm.makePersistentInternal(element, new FieldValues()
            {
                public void fetchFields(StateManager esm)
                {
                    boolean isPersistentInterface = storeMgr.getMetaDataManager().isPersistentInterface(elementType);
                    DatastoreClass elementTable = null;
                    if (!isPersistentInterface)
                    {
                        elementTable = storeMgr.getDatastoreClass(elementType, clr);
                    }
                    else
                    {
                        elementTable = storeMgr.getDatastoreClass(storeMgr.getMetaDataManager().getImplementationNameForPersistentInterface(elementType), clr);
                    }
                    if (elementTable != null)
                    {
                        JavaTypeMapping externalFKMapping = elementTable.getExternalMapping(ownerFieldMetaData, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK);
                        if (externalFKMapping != null)
                        {
                            // The element has an external FK mapping so set the value it needs to use in the INSERT
                            esm.setExternalFieldValueForMapping(externalFKMapping, ownerSM.getObject());
                        }
                        if (relationDiscriminatorMapping != null)
                        {
                            esm.setExternalFieldValueForMapping(relationDiscriminatorMapping, relationDiscriminatorValue);
                        }
                    }
                    if (getOwnerFieldOwner(esm) >= 0)
                    {
                        esm.replaceField(getOwnerFieldOwner(esm), newOwner);
                    }
                }

                public void fetchNonLoadedFields(StateManager sm)
                {
                }
                public FetchPlanImpl getFetchPlanForLoading()
                {
                    return null;
                }
            });
            StateManager elementSM = pm.findStateManager((PersistenceCapable)element);
            if (elementSM != null)
            {
                // Make sure that this is in the datastore before continuing
                elementSM.flush();
            }

            return true;
        }
    }
 
    /**
     * Method to add a collection of object to the relationship at the
     * collection end.
     * @param sm StateManager of the Set
     * @param elements Elements to be added
     * @return Success indicator
     **/
    public boolean addAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        boolean success = false;

        Iterator iter = elements.iterator();
        while (iter.hasNext())
        {
            if (add(sm, iter.next()))
            {
                success = true;
            }
        }

        return success;
    }

    /**
     * Method to remove the link to the collection object specified.
     * Depending on the column characteristics in the collection table, the id
     * of the owner field may be NULLed, or the record may be deleted
     * completely (as per cascade-delete in EJB).
     *
     * @param sm The StateManager of the Set
     * @param element The element of the collection to be deleted.
     * @return A success indicator.
     **/
    public boolean remove(StateManager sm, Object element)
    {
        if (element == null)
        {
            return false;
        }
        if (!validateElementForReading(sm, element))
        {
            return false;
        }

        PersistenceManager pm = sm.getPersistenceManager();
        PersistenceCapable pcElement = (PersistenceCapable) element;
        StateManager esm = pm.findStateManager(pcElement);

        Object oldOwner = null;
        if (ownerFieldNumber >= 0)
        {
            if (!esm.isDeleted(pcElement))
            {
                // Find the existing owner if the record hasn't already been deleted
                esm.isLoaded(pcElement, ownerFieldNumber);
                oldOwner = esm.provideField(ownerFieldNumber);
            }
        }
        else
        {
            // TODO Check if the element is managed by a different owner now
        }

        // Owner of the element has been changed
        if (ownerFieldNumber >= 0 && oldOwner != sm.getObject())
        {
            return false;
        }

        if (ownerFieldMetaData.getCollection().isDependentElement())
        {
            // Delete the element since it is dependent
            pm.deletePersistent(element);
        }
        else if (ownerMapping.isNullable())
        {
            // Bidirectional 1-N, so nullify the owner
            if (ownerFieldNumber >= 0)
            {
                esm.setObjectField(pcElement, ownerFieldNumber, oldOwner, null);
            }
            // Unidirectional 1-N, so nullify the owner
            else
            {
                updateElementFk(sm, pcElement, null);
            }
        }
        else
        {
            // otherwise remove the element since we can't null its FK
            pm.deletePersistent(element);
        }

        return true;
    }

    /**
     * Method to remove the links to a collection of elements specified.
     * Depending on the column characteristics in the collection table, the id
     * of the owner fields may be NULLed, or the records may be deleted
     * completely.
     *
     * @param sm The StateManager of the Set
     * @param elements The elements of the collection to be deleted.
     * @return A success indicator.
     **/
    public boolean removeAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        // Check the first element for whether we can null the column or
        // whether we have to delete
        boolean success = true;

        Iterator iter=elements.iterator();
        while (iter.hasNext())
        {
            if (remove(sm, iter.next()))
            {
                success = false;
            }
        }

        return success;
    }

    /**
     * Method to allow the Set relationship to be cleared out.
     * This is called by the Collection.clear() method and also when the container object is being deleted
     * and the elements are to be removed (maybe for dependent field).
     * @param ownerSM StateManager of the Set
     **/
    public void clear(StateManager ownerSM)
    {
        boolean deleteElements = false;
        if (ownerFieldMetaData.getCollection().isDependentElement())
        {
            // Elements are dependent and can't exist on their own, so delete them all
            JPOXLogger.RDBMS.debug(LOCALISER.msg("RDBMS.SCO.Collection.ClearElementsFromInverseDeleteDependent"));
            deleteElements = true;
        }
        else
        {
            if (ownerMapping.isNullable())
            {
                // Field is not dependent, but is nullable so we null the FK
                JPOXLogger.RDBMS.debug(LOCALISER.msg("RDBMS.SCO.Collection.ClearElementsFromInverseNull"));
                deleteElements = false;
            }
            else
            {
                // Field is not dependent, and is not nullable so we just delete the elements
                JPOXLogger.RDBMS.debug(LOCALISER.msg("RDBMS.SCO.Collection.ClearElementsFromInverseDelete"));
                deleteElements = true;
            }
        }

        if (deleteElements)
        {
            ownerSM.isLoaded(ownerSM.getObject(), ownerFieldMetaData.getAbsoluteFieldNumber()); // Make sure the field is loaded
            Collection value = (Collection) ownerSM.provideField(ownerFieldMetaData.getAbsoluteFieldNumber());
            if (value != null && !value.isEmpty())
            {
                ownerSM.getPersistenceManager().deletePersistentAll(value.toArray());
            }
        }
        else
        {
            // TODO If the relation is bidirectional we need to clear the owner in the element
            
            String stmt = clearNullifyStmt;
            if (elementInfo.length > 1)
            {
                // TODO we should loop over all tables that has elements stored into and clear all them
                DatastoreClass table = storeMgr.getDatastoreClass(elementInfo[0].getClassName(), clr);
                if (table != null)
                {
                    stmt = StringUtils.replaceAll(clearNullifyStmt, "<TABLE NAME>", table.toString());
                }
                else
                {
                    JPOXLogger.JDO.warn(">> InverseArrayStore.updateElementFK : need to set table in statement but dont know table where to store " + elementInfo[0].getClassName());
                }
            }                
            try
            {
                PersistenceManager pm = ownerSM.getPersistenceManager();
                Connection conn = storeMgr.getConnection(pm, true, false);
                try
                {
                    PreparedStatement ps = storeMgr.getStatement(conn, stmt, false);
                    try
                    {
                        int jdbcPosition = 1;
                        jdbcPosition = populateOwnerInStatement(ownerSM, pm, ps, jdbcPosition);
                        storeMgr.executeStatementUpdate(stmt, ps);
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    storeMgr.releaseConnection(pm, conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.ClearRequestFailed",stmt),e);
            }
        }
    }

    /**
     * Accessor for a QueryStatement to retrieve the elements of the set.
     * @param ownerSM the owner StateManager
     * @return The QueryStatement to retrieve the elements
     **/    
    protected QueryExpression getIteratorStatement(StateManager ownerSM)
    {
        if (elementInfo == null || elementInfo.length == 0)
        {
            return null;
        }

        final ClassLoaderResolver clr = ownerSM.getPersistenceManager().getClassLoaderResolver();
        QueryExpression stmt = null;
        if (elementInfo[0].getDatastoreClass().getDiscriminatorMetaData() != null && 
            elementInfo[0].getDatastoreClass().getDiscriminatorMetaData().getStrategy() != DiscriminatorStrategy.NONE)
        {
            //when one uses discriminator, all must have
            if (storeMgr.getPMFContext().getTypeManager().isReferenceType(clr.classForName(ownerFieldMetaData.getCollection().getElementType())))
            {                
                // Take the metadata for the first implementation of the reference type
                String[] clsNames = storeMgr.getMetaDataManager().getClassesImplementingInterface(ownerFieldMetaData.getCollection().getElementType(), clr);
                Class[] cls = new Class[clsNames.length];
                for( int i=0; i<clsNames.length; i++)
                {
                    cls[i] = clr.classForName(clsNames[i]);
                }
                stmt = new DiscriminatorIteratorStatement(clr, cls, true, this.storeMgr, true).getQueryStatement();
            }
            else
            {
                stmt = new DiscriminatorIteratorStatement(clr, new Class[] {clr.classForName(elementInfo[0].getClassName())},
                    true, this.storeMgr, true).getQueryStatement();
            }
            iterateUsingDiscriminator = true;
        }
        else
        {
            for (int i=0;i<elementInfo.length;i++)
            {
                final Class elementCls = clr.classForName(this.elementInfo[i].getClassName());
                final int elementNo = i;
                QueryExpression subStmt = new UnionIteratorStatement(
                    clr,
                    elementCls,
                    true,
                    this.storeMgr,
                    new UnionIteratorStatement.AssociationEnd()
                    {
                        public JavaTypeMapping getMapping()
                        {
                            return elementMapping;
                        }
                        public Class getType()
                        {
                            return elementCls;
                        }
                        public DatastoreContainerObject getDatastoreContainerObject()
                        {
                            return elementInfo[elementNo].getDatastoreClass();
                        }
                        public boolean useJoin()
                        {
                            return false;
                        }
                    }).getQueryStatement();
                if (stmt == null)
                {
                    stmt = subStmt;
                }
                else
                {
                    stmt.union(subStmt);
                }
            }
        }

        // Apply condition on join-table owner field to filter by owner
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, ownerSM.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        // Apply condition on distinguisher field to filter by distinguisher (when present)
        if (relationDiscriminatorMapping != null)
        {
            ScalarExpression distinguisherExpr = relationDiscriminatorMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
            ScalarExpression distinguisherVal = relationDiscriminatorMapping.newLiteral(stmt, relationDiscriminatorValue);
            stmt.andCondition(distinguisherExpr.eq(distinguisherVal), true);
        }

        if (orderMapping != null)
        {
            // Order by the ordering column, when present
            ScalarExpression exprIndex[] = new ScalarExpression[orderMapping.getNumberOfDatastoreFields()];
            boolean descendingOrder[] = new boolean[orderMapping.getNumberOfDatastoreFields()];
            exprIndex = orderMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression()).getExpressionList().toArray();
            stmt.setOrdering(exprIndex, descendingOrder);
        }

        return stmt;
    }

    // --------------------------- JDOQL Query Methods -------------------------------

    /**
     * Utility method to return a new QueryStatement for retrieval of the
     * elements of this Set.
     *
     * @param sm StateManager for this object
     * @param candidateClass Class for the element end of the link.
     * @return The QueryStatement
     **/
    public QueryExpression newQueryStatement(StateManager sm, String candidateClass)
    {
        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(elementType,candidateClass))
        {
            throw new IncompatibleQueryElementTypeException(elementType, candidateClass);
        }

        DatastoreContainerObject candidateTable = storeMgr.getDatastoreClass(candidateClass, sm.getPersistenceManager().getClassLoaderResolver());
        QueryExpression stmt = dba.newQueryStatement(candidateTable,sm.getPersistenceManager().getClassLoaderResolver());

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal));

        stmt.select(elementMapping);

        return stmt;
    }

    /**
     * Utility for use in building a query, joining the element table and the
     * owner table.
     *
     * @param stmt The Query Statement 
     * @param parentStmt the parent Query Statement. If there is no parent, <code>parentStmt</code> must be equals to <code>stmt</code> 
     * @param ownerMapping the mapping for the owner. 
     * @param ownerTe Table Expression for the owner
     * @param setRangeVar   The range variable for the "Set" table. 
     * @param filteredElementType The Class Type for the filtered element
     * @param elmExpr The Expression for the element
     * @param elementRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the element table.
     * @return A subquery for the provided query that joins the element table
     * to the owner table.
     */
    public ScalarExpression joinElementsTo(
        QueryExpression stmt,
        QueryExpression parentStmt,
        JavaTypeMapping ownerMapping,
        LogicSetExpression ownerTe,
        DatastoreIdentifier setRangeVar,
        Class filteredElementType,
        ScalarExpression elmExpr,
        DatastoreIdentifier elementRangeVar)
    {
        ClassLoaderResolver clr=stmt.getClassLoaderResolver();
        if (!clr.isAssignableFrom(elementType,filteredElementType) &&
            !clr.isAssignableFrom(filteredElementType,elementType))
        {
            throw new IncompatibleQueryElementTypeException(elementType, filteredElementType.getName());
        }

        // Join the element table on the owner ID column.
        DatastoreContainerObject filteredElementTable = storeMgr.getDatastoreClass(filteredElementType.getName(), stmt.getClassLoaderResolver());
        stmt.newTableExpression(filteredElementTable, elementRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,ownerTe);
        DatastoreIdentifier containerRangeVar = setRangeVar;
        if( stmt.getTableExpression(containerRangeVar) == null )
        {
            containerRangeVar = elementRangeVar;
        }
        ScalarExpression ownerSetExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(containerRangeVar));
        stmt.newTableExpression(containerTable, containerRangeVar);

        stmt.andCondition(ownerExpr.eq(ownerSetExpr),true);

        JavaTypeMapping elementTableID = filteredElementTable.getIDMapping();
        return elementTableID.newScalarExpression(stmt,stmt.getTableExpression(containerRangeVar));
    }
}