/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.MetaDataUtils;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.ReferenceMapping;
import org.jpox.store.rdbms.query.UnionIteratorStatement;
import org.jpox.store.rdbms.table.ArrayTable;
import org.jpox.util.ClassUtils;

/**
 * Backing store for an array stored using a join table.
 * Can be used for all types of arrays :- PersistenceCapable arrays,
 * non-PersistenceCapable arrays, reference arrays.
 *
 * @version $Revision: 1.8 $
 */
public class JoinArrayStore extends AbstractArrayStore
{
    /**
     * Constructor.
     * @param joinTable Join table storing the relationship between owner and element
     * @param clr ClassLoader resolver
     */
    public JoinArrayStore(ArrayTable joinTable, ClassLoaderResolver clr)
    {
        super(joinTable.getStoreManager(), clr);

        this.containerTable = joinTable;
        this.ownerFieldMetaData = joinTable.getOwnerFieldMetaData();

        ownerMapping = joinTable.getOwnerMapping();
        elementMapping = joinTable.getElementMapping();
        orderMapping = joinTable.getOrderMapping();
        relationDiscriminatorMapping = joinTable.getRelationDiscriminatorMapping();
        relationDiscriminatorValue = joinTable.getRelationDiscriminatorValue();

        elementType = joinTable.getElementType();
        elementsAreEmbedded = joinTable.isEmbeddedElement();
        elementsAreSerialised = joinTable.isSerialisedElement();

        if (elementsAreSerialised)
        {
            elementInfo = null;
        }
        else
        {
            Class element_class=clr.classForName(elementType);
            if (storeMgr.getPMFContext().getTypeManager().isReferenceType(element_class))
            {
                // Array of reference types (interfaces/Objects)
                String[] implNames = MetaDataUtils.getInstance().getImplementationNamesForReferenceField(ownerFieldMetaData, 
                    DatastoreField.ROLE_ARRAY_ELEMENT, clr);
                elementInfo = new ElementInfo[implNames.length];
                for (int i=0;i<implNames.length;i++)
                {
                    DatastoreClass table = storeMgr.getDatastoreClass(implNames[i], clr);
                    AbstractClassMetaData cmd = storeMgr.getMetaDataManager().getMetaDataForClass(implNames[i], clr);
                    elementInfo[i] = new ElementInfo(cmd, table);
                }
            }
            else
            {
                // Collection of PC or non-PC
                emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class, clr);
                if (emd != null)
                {
                    elementType  = emd.getFullClassName();
                    if (!elementsAreEmbedded && !elementsAreSerialised)
                    {
                        elementInfo = getElementInformationForClass();
                        if (elementInfo != null && elementInfo.length > 1)
                        {
                            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Array.ElementWithSubclassTableWithMultipleTablesNotSupported", 
                                ownerFieldMetaData.getFullFieldName()));
                        }
                    }
                    else
                    {
                        elementInfo = null;
                    }
                }
                else
                {
                    elementInfo = null;
                }
            }
        }

        // Generate the statements
        initialiseStatements();
        addStmt = getAddStmt();
    }

    /**
     * Accessor for the iterator statement to retrieve the element(s) for the array.
     * @param ownerSM The StateManager
     * @return The Query Statement.
     **/
    protected QueryExpression getIteratorStatement(StateManager ownerSM)
    {
        QueryExpression stmt = null;
        final ClassLoaderResolver clr = ownerSM.getPersistenceManager().getClassLoaderResolver();
        if (elementsAreEmbedded || elementsAreSerialised)
        {
            // Element = Non-PC
            stmt = dba.newQueryStatement(containerTable, clr);
            stmt.select(elementMapping);
        }
        else if (elementMapping instanceof ReferenceMapping)
        {
            // Element = Reference type (interface/Object)
            // Just select the join table since we're going to return the implementation id columns only
            stmt = dba.newQueryStatement(containerTable, clr);
        }
        else
        {
            // Element = PC
            stmt = new UnionIteratorStatement(clr, clr.classForName(this.elementType), true, this.storeMgr,
                new UnionIteratorStatement.AssociationEnd()
                {
                    public JavaTypeMapping getMapping()
                    {
                        return elementMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(elementType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return containerTable;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }
                }).getQueryStatement();
        }

        // Apply condition on join-table owner field to filter by owner
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, ownerSM.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        // Apply condition on distinguisher field to filter by distinguisher (when present)
        if (relationDiscriminatorMapping != null)
        {
            ScalarExpression distinguisherExpr = relationDiscriminatorMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
            ScalarExpression distinguisherVal = relationDiscriminatorMapping.newLiteral(stmt, relationDiscriminatorValue);
            stmt.andCondition(distinguisherExpr.eq(distinguisherVal), true);
        }

        // Order by the position
        ScalarExpression exprIndex[] = new ScalarExpression[orderMapping.getNumberOfDatastoreFields()];
        boolean descendingOrder[] = new boolean[orderMapping.getNumberOfDatastoreFields()];
        exprIndex = orderMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression()).getExpressionList().toArray();
        stmt.setOrdering(exprIndex, descendingOrder);

        return stmt;
    }

    /**
     * Method used in queries when contains() has been invoked.
     * @param stmt The Query Statement 
     * @param parentStmt the parent Query Statement. If there is no parent, <code>parentStmt</code> must be equals to <code>stmt</code> 
     * @param ownerMapping the mapping for the owner. 
     * @param ownerTe Table Expression for the owner
     * @param listRangeVar   The range variable for the "List" table. 
     * @param filteredElementType The Class Type for the filtered element
     * @param elmExpr The Expression for the element
     * @param elementRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the element table.
     * @return expression to the join
     **/
    public ScalarExpression joinElementsTo(QueryExpression stmt, 
                                           QueryExpression parentStmt,
                                           JavaTypeMapping ownerMapping,
                                           LogicSetExpression ownerTe,
                                           DatastoreIdentifier listRangeVar,
                                           Class filteredElementType,
                                           ScalarExpression elmExpr,
                                           DatastoreIdentifier elementRangeVar)
    {
        ClassLoaderResolver clr=stmt.getClassLoaderResolver();
        Class primitiveType = ClassUtils.getWrapperTypeForPrimitiveType(clr.classForName(elementType));
        if (!clr.isAssignableFrom(elementType, filteredElementType) &&
            !clr.isAssignableFrom(filteredElementType, elementType) &&
            (primitiveType !=null && clr.isAssignableFrom(elementType,primitiveType.getName())))
        {
            throw new IncompatibleQueryElementTypeException(elementType, filteredElementType.getName());
        }

        /* TODO this comment is no longer valid, since we use sub queries
         * create a Join from the owner to the set table based on the owner ID column.
         * FROM OWNER THIS
         * INNER JOIN OWNER_ELEMENT THIS_SETTABLE 
         * ON THIS.OWNER_ID = THIS_SETTABLE.OWNER_OID */
        LogicSetExpression ownTblExpr = stmt.newTableExpression(containerTable, listRangeVar);
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, ownerTe);
        ScalarExpression ownerSetExpr = this.ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
        if (!parentStmt.hasCrossJoin(ownTblExpr))
        {
            stmt.crossJoin(ownTblExpr, true);
        }
        stmt.andCondition(ownerExpr.eq(ownerSetExpr),true);

        if (storeMgr.getPMFContext().getTypeManager().isSupportedType(filteredElementType.getName()))
        {
            // Element = Non-PC(embedded)
            return elementMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
        }
        else if (elementsAreEmbedded || elementsAreSerialised)
        {
            // Element = PC(embedded), PC(serialised)
            return elementMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
        }
        else
        {
            // Element = PC
            // Join to element table on id column(s) of element
            DatastoreClass elementTable = storeMgr.getDatastoreClass(filteredElementType.getName(), stmt.getClassLoaderResolver());
            DatastoreClass joiningClass = (elmExpr.getLogicSetExpression() == null ? elementTable : (DatastoreClass)elmExpr.getLogicSetExpression().getMainTable());
            JavaTypeMapping elementTableID = joiningClass.getIDMapping();

            /* TODO this comment is no longer valid, since we use sub queries
             * if elementType == filteredElementType
             *    INNER JOIN ELEMENT THIS_FILTER_ELEMENT_VARIABLE
             *    ON THIS_SETTABLE.ELEMENT_EID = THIS_FILTER_ELEMENT_VARIABLE.ELEMENT_ID
             * 
             * else if elementType != filteredElementType but filteredElementType is a superclass of elementType
             * 
             *    INNER JOIN FILTER_ELEMENT_TYPE THIS_FILTER_ELEMENT_VARIABLE
             *    ON THIS_SETTABLE.ELEMENT_EID = THIS_FILTER_ELEMENT_VARIABLE.FILTER_ELEMENT_TYPE_ID */
            LogicSetExpression elmTblExpr = stmt.getTableExpression(elementRangeVar);
            if (elmTblExpr==null)
            {
                elmTblExpr = stmt.newTableExpression(elementTable,elementRangeVar);
            }
            ScalarExpression elmSetExpr = elementMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
            if (!parentStmt.hasCrossJoin(elmTblExpr))
            {
                stmt.crossJoin(elmTblExpr, true);
            }
            if (elmExpr.getLogicSetExpression()!= null && !elementTable.equals(elmExpr.getLogicSetExpression().getMainTable()))
            {
                //elmExpr might express a FK in another to the ELEMENT table 
                stmt.andCondition(elmSetExpr.eq(elmExpr),true);
                return this.elementMapping.newScalarExpression(stmt,stmt.getTableExpression(listRangeVar));
            }
            else
            {
                //elmExpr might be a PK of the ELEMENT table 
                ScalarExpression elementExpr = elementTableID.newScalarExpression(stmt, stmt.getTableExpression(elementRangeVar));
                stmt.andCondition(elmSetExpr.eq(elementExpr),true);
                return elementExpr;
            }                    
        }
    }
}