/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import org.jpox.ClassLoaderResolver;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ArrayMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.PrimaryKeyMetaData;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.rdbms.RDBMSManager;

/**
 * Join table for storing an array of objects.
 * The join table is used in the following situations :-
 * <UL>
 * <LI><B>Simple[]</B> - a list of simple elements.
 * For example a array of ints. The elements are stored in their own
 * column in this join table. There is no separate element table</LI>
 * <LI><B>PersistenceCapable[]</B> - a list of
 * PersistenceCapable elements with the elements stored in their own table.
 * A FK is added to this table to get to the element table</LI>
 * <LI><B>PersistenceCapable[] (embedded)</B> - an array of
 * PersistenceCapable elements with the elements embedded in this table.</LI>
 * <LI><B>PersistenceCapable[], Simple[] (serialised)</B> - an array
 * with the elements serialised into a single column in this table.</LI>
 * </UL>
 * 
 * @version $Revision: 1.11 $
 */
public class ArrayTable extends ElementContainerTable
{
    /**
     * Constructor.
     * @param tableName Identifier name of the table
     * @param fmd MetaData for the field of the owner
     * @param storeMgr The Store Manager managing these tables.
     */
    public ArrayTable(DatastoreIdentifier tableName,
                      AbstractPropertyMetaData fmd,
                      RDBMSManager storeMgr)
    {
        super(tableName, fmd, storeMgr);
    }

    /**
     * Method to initialise the table definition.
     * @param clr The ClassLoaderResolver
     */
    public void initialize(ClassLoaderResolver clr)
    {
        super.initialize(clr);

        PrimaryKeyMetaData pkmd = (fmd.getJoinMetaData() != null ? fmd.getJoinMetaData().getPrimaryKeyMetaData() : null);
        boolean pkColsSpecified = (pkmd != null ? pkmd.getColumnMetaData() != null : false);
        boolean pkRequired = requiresPrimaryKey();

        if (isSerialisedElementPC() || isEmbeddedElementPC())
        {
            // Element = PC(embedded), PC(serialised)

            // Add field(s) for element
            elementMapping = dba.getMappingManager().getMapping(this, fmd, dba, clr, MappingManager.MAPPING_ARRAY_ELEMENT);
        }
        else
        {
            // Element = PC, Non-PC(embedded)

            // Create the element mapping, taking a serialised mapping if the element is embedded
            ArrayMetaData arrmd = fmd.getArray();
            elementMapping = ColumnCreator.createColumnsForJoinTables(fmd.getType().getComponentType(), fmd,
                fmd.getElementMetaData() == null ? null : fmd.getElementMetaData().getColumnMetaData(), storeMgr, this,
                false, true, arrmd.isSerializedElement(), false, DatastoreField.ROLE_ARRAY_ELEMENT, clr, null, false);
        }

        // Add the list ordering index - JDO2 spec [18.5] order column is assumed to be "int"
        ColumnMetaData colmd = null;
        if (fmd.getOrderMetaData() != null && fmd.getOrderMetaData().getColumnMetaData() != null &&
            fmd.getOrderMetaData().getColumnMetaData().length > 0)
        {
            colmd = fmd.getOrderMetaData().getColumnMetaData()[0];
        }
        else
        {
            colmd = new ColumnMetaData(fmd, "IDX"); // Give index a default name of IDX when not specified
        }
        orderMapping = dba.getMapping(int.class, storeMgr); // JDO2 spec [18.5] order column is assumed to be "int"
        ColumnCreator.createAdapterColumn(orderMapping, storeMgr, clr, this, colmd, pkRequired && !pkColsSpecified);

        if (pkRequired && pkColsSpecified)
        {
            // Apply the users PK specification
            applyUserPrimaryKeySpecification(pkmd);
        }

        state = TABLE_STATE_INITIALIZED;
    }

    /**
     * Accessor for the element type stored in this array.
     * @return Name of element type.
     */
    public String getElementType()
    {
        return fmd.getType().getComponentType().getName();
    }

    /**
     * Accessor for whether the element is serialised into this table.
     * This can be a serialised PersistenceCapable, or a serialised simple type
     * @return Whether the element is serialised.
     */
    public boolean isSerialisedElement()
    {
        if (fmd.getArray() != null && fmd.getArray().isSerializedElement())
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for whether the element is embedded into this table.
     * This can be an embedded PersistenceCapable, or an embedded simple type
     * @return Whether the element is embedded.
     */
    public boolean isEmbeddedElement()
    {
        if (fmd.getArray() != null && fmd.getArray().isSerializedElement())
        {
            // Serialised takes precedence
            return false;
        }
        else if (fmd.getArray() != null && fmd.getArray().isEmbeddedElement())
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for whether the element is a PersistenceCapable(serialised)
     * @return Whether the element is PC and is serialised
     */
    public boolean isSerialisedElementPC()
    {
        if (fmd.getArray() != null && fmd.getArray().isSerializedElement())
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for whether the element is a PersistenceCapable(embedded)
     * @return Whether the element is PC and is embedded
     */
    public boolean isEmbeddedElementPC()
    {
        if (fmd.getArray() != null && fmd.getArray().isSerializedElement())
        {
            // Serialisation takes precedence over embedding
            return false;
        }
        if (fmd.getElementMetaData() != null && fmd.getElementMetaData().getEmbeddedMetaData() != null)
        {
            return true;
        }
        return false;
    }
}