/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Andy Jefferson - added localiser
2003 Andy Jefferson - replaced TableMetadata with identifier.
2003 Erik Bengtson - refactored the datastore identity together with the SQLIdentifier class
2003 Erik Bengtson - added OptimisticMapping
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - merged with JDOBaseTable
2004 Andy Jefferson - changed to store map of fieldMappings keyed by absolute field num. Added start point for inheritance strategy handling
2004 Andy Jefferson - changed consumer to set highest field number based on actual highest (to allow for inheritance strategies)
2004 Andy Jefferson - added DiscriminatorMapping
2004 Andy Jefferson - enabled use of Long/String datastore identity column
2004 Andy Jefferson - added capability to handle 1-N inverse unidirectional FKs
2004 Andy Jefferson - removed the majority of the value-strategy code - done elsewhere
    ...
 **********************************************************************/
package org.jpox.store.rdbms.table;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFConfiguration;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.ColumnMetaDataContainer;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.ExtendableMetaData;
import org.jpox.metadata.ExtensionMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.metadata.ForeignKeyAction;
import org.jpox.metadata.ForeignKeyMetaData;
import org.jpox.metadata.IdentityMetaData;
import org.jpox.metadata.IdentityStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.IndexMetaData;
import org.jpox.metadata.InheritanceMetaData;
import org.jpox.metadata.InheritanceStrategy;
import org.jpox.metadata.JoinMetaData;
import org.jpox.metadata.OrderMetaData;
import org.jpox.metadata.PrimaryKeyMetaData;
import org.jpox.metadata.PropertyMetaData;
import org.jpox.metadata.Relation;
import org.jpox.metadata.UniqueMetaData;
import org.jpox.metadata.VersionMetaData;
import org.jpox.metadata.VersionStrategy;
import org.jpox.sco.SCOUtils;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.OIDFactory;
import org.jpox.store.exceptions.ClassDefinitionException;
import org.jpox.store.exceptions.NoSuchPersistentFieldException;
import org.jpox.store.exceptions.NoTableManagedException;
import org.jpox.store.mapping.DatastoreMapping;
import org.jpox.store.mapping.DiscriminatorMapping;
import org.jpox.store.mapping.EmbeddedPCMapping;
import org.jpox.store.mapping.IndexMapping;
import org.jpox.store.mapping.IntegerMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.LongMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.mapping.ReferenceMapping;
import org.jpox.store.mapping.VersionMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.exceptions.DuplicateColumnNameException;
import org.jpox.store.rdbms.key.CandidateKey;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.key.Index;
import org.jpox.store.rdbms.key.PrimaryKey;
import org.jpox.store.rdbms.request.DeleteRequest;
import org.jpox.store.rdbms.request.FetchRequest;
import org.jpox.store.rdbms.request.InsertRequest;
import org.jpox.store.rdbms.request.LocateRequest;
import org.jpox.store.rdbms.request.UpdateRequest;
import org.jpox.store.rdbms.sqlidentifier.RDBMSIdentifierFactory;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.util.AIDUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.MacroString;
import org.jpox.util.MultiMap;

/**
 * Table representing a Java class (or classes) as a first class object (FCO).
 * Uses the inheritance strategy to control whether this represents multiple classes
 * or just the one class.
 * <H3>Mappings</H3>
 * This class adds some additional mappings over what the superclass provides. Here we add
 * <UL>
 * <LI><B>externalFkMappings</B> - any mappings for Collections that have no associated
 * field in this class, providing the foreign key column(s)</LI>
 * <LI><B>externalOrderMappings</B> - mappings for any ordering column used by Lists for ordering
 * elements of this class</LI>
 * <LI><B>externalFkDiscriminatorMappings</B> - mappings for any discriminator column used when sharing
 * external foreign keys to distinguish the element owner field</LI>
 * </UL>
 * <H3>Classes</H3>
 * A table can represent multiple classes. It has a nominal owner which is the class
 * that has an inheritance strategy of "new-table". All classes that utilise this table
 * have their MetaData stored in this object.
 * <H3>Secondary Tables</H3>
 * This class represents a "primary" table. That is, the main table where objects of a
 * class are persisted. It can have several "secondary" tables where some of the classes
 * fields are stored at persistence.
 *
 * @version $Revision: 1.197.2.3 $
 **/
public class ClassTable extends AbstractClassTable implements DatastoreClass 
{
    /**
     * MetaData for the principal class being stored here.
     * In inheritance situations where multiple classes share the same table, this will
     * be the class which uses "new-table" strategy.
     */
    private final ClassMetaData cmd;

    /** MetaData for all classes being managed here. */
    private final Collection managedClassMetaData = new HashSet();

    /** Table above this table storing superclass information (if any). */
    private ClassTable supertable;

    /** Secondary tables for this table (if any). */
    private Map secondaryTables;

    /**
     * Mappings for (Inverse) Collections/Lists FKs not managed by this class (1-N unidirectional).
     * Keyed by the FieldMetaData of the owner field.
     */
    private HashMap externalFkMappings;

    /**
     * Mappings for (Inverse) Collections/Lists FK relation discriminators not managed by this class (1-N unidirectional).
     * Keyed by the FieldMetaData of the owner field.
     */
    private HashMap externalFkDiscriminatorMappings;

    /**
     * Mappings for (Inverse) Lists managed by this class (1-N unidirectional).
     * Keyed by the FieldMetaData of the owner field.
     */
    private HashMap externalOrderMappings;

    /** User defined table schema **/
    private MacroString tableDef;

    /** DDL statement for creating the table, if using user defined table schema **/
    private String createStatementDDL;
    
    /** All callbacks for class tables waiting to be performed. */
    private static MultiMap callbacks = new MultiMap();

    HashMap candidateKeysByMapField = new HashMap();
    
    /**
     * Constructor.
     * @param tableName Table name SQL identifier
     * @param storeMgr Store Manager to manage this table
     * @param cmd MetaData for the class.
     */
    public ClassTable(DatastoreIdentifier tableName, RDBMSManager storeMgr, ClassMetaData cmd)
    {
        super(tableName, storeMgr);

        this.cmd = cmd;

        // Check for valid class to map to a table
        if (cmd.getInheritanceMetaData().getStrategyValue() != InheritanceStrategy.NEW_TABLE)
        {
            throw new JDOFatalUserException(LOCALISER.msg("RDBMS.Table.ClassHasNoTableError", cmd.getFullClassName(),
                cmd.getInheritanceMetaData().getStrategyValue().toString()));
        }

        highestFieldNumber = cmd.getNoOfManagedFields() + cmd.getNoOfInheritedManagedFields();
        
        // Extract the table definition from MetaData, if exists
        String tableImpStr = cmd.getValueForExtension("ddl-imports");
        String tableDefStr = null;
        if (dba.getVendorID() != null)
        {
            tableDefStr = cmd.getValueForExtension("ddl-definition" + '-' + dba.getVendorID());
        }
        if (tableDefStr == null)
        {
            tableDefStr = cmd.getValueForExtension("ddl-definition");
        }
        if (tableDefStr != null)
        {
            tableDef = new MacroString(cmd.getFullClassName(), tableImpStr, tableDefStr);
        }
    }

    /**
     * Pre-initialize.
     * We require any supertable, and the PK to be ready before we start initialisation.
     * @param clr the ClassLoaderResolver
     */
    public void preInitialize(final ClassLoaderResolver clr)
    {
        assertIsPKUninitialized();

        // Find the super table in any inheritance hierarchy (navigates to find any table above this)
        supertable = getSupertable(cmd, clr);

        // Make sure that the supertable is preinitialised before we think about initialising here
        if (supertable != null && !supertable.isInitialized() && !supertable.isPKInitialized())
        {
            supertable.preInitialize(clr);
        }

        // Initialise the PK field(s)
        if (!isPKInitialized())
        {
            initializePK(clr);
        }
    }

    /**
     * Method to initialise the table. This adds the columns based on the
     * MetaData representation for the class being represented by this table.
     * @param clr The ClassLoaderResolver
     */
    public void initialize(ClassLoaderResolver clr)
    {
        assertIsUninitialized();

        // Add the fields for this class (and any other superclasses that we need to manage the
        // fields for (inheritance-strategy="subclass-table" in the superclass)
        initializeForClass(cmd, clr);

        // Add Version where specified in MetaData
        VersionMetaData vmd = cmd.getVersionMetaData();
        if (vmd != null)
        {
            if (vmd.getVersionStrategy() == VersionStrategy.NONE)
            {
                // No optimistic locking but the idiot wants a column for that :-)
                versionMapping = new VersionMapping(dba, this, dba.getMapping(Long.class, storeMgr));
            }
            else if (vmd.getVersionStrategy() == VersionStrategy.VERSION_NUMBER)
            {
                versionMapping = new VersionMapping(dba, this, dba.getMapping(Long.class, storeMgr));
            }
            else if (vmd.getVersionStrategy() == VersionStrategy.DATE_TIME)
            {
                versionMapping = new VersionMapping(dba, this, dba.getMapping(Timestamp.class, storeMgr));
            }
        }

        // Add Discriminator where specified in MetaData
        InheritanceMetaData imd = cmd.getInheritanceMetaData();
        if (imd != null && imd.getStrategyValue() == InheritanceStrategy.NEW_TABLE)
        {
            DiscriminatorMetaData dismd = imd.getDiscriminatorMetaData();
            if (dismd != null)
            {
                if (dismd.getStrategy() == DiscriminatorStrategy.CLASS_NAME)
                {
                    discriminatorMapping = new DiscriminatorMapping(dba, this, dba.getMapping(String.class, storeMgr));
                }
                else if (dismd.getStrategy() == DiscriminatorStrategy.VALUE_MAP)
                {
                    ColumnMetaData disColmd = dismd.getColumnMetaData();
                    if (disColmd != null && disColmd.getJdbcType() != null)
                    {
                        if (disColmd.getJdbcType().equalsIgnoreCase("INTEGER") ||
                            disColmd.getJdbcType().equalsIgnoreCase("BIGINT") ||
                            disColmd.getJdbcType().equalsIgnoreCase("NUMERIC"))
                        {
                            discriminatorMapping = new DiscriminatorMapping(dba, this, dba.getMapping(long.class, storeMgr));
                        }
                        else
                        {
                            discriminatorMapping = new DiscriminatorMapping(dba, this, dba.getMapping(String.class, storeMgr));
                        }
                    }
                    else
                    {
                        discriminatorMapping = new DiscriminatorMapping(dba, this, dba.getMapping(String.class, storeMgr));
                    }
                }
            }
        }

        // Initialise any SecondaryTables
        if (secondaryTables != null)
        {
            Set secondaryTableNames = secondaryTables.keySet();
            Iterator secondaryTableNamesIter = secondaryTableNames.iterator();
            while (secondaryTableNamesIter.hasNext())
            {
                String secondaryTableName = (String)secondaryTableNamesIter.next();
                SecondaryTable second = (SecondaryTable)secondaryTables.get(secondaryTableName);
                if (!second.isInitialized())
                {
                    second.initialize(clr);
                }
            }
        }

        state = TABLE_STATE_INITIALIZED;

    }
    
    /**
     * Post initilize. For things that must be set after all classes have been initialized before 
     * @param clr the ClassLoaderResolver
     */
    public void postInitialize(final ClassLoaderResolver clr)
    {
        assertIsInitialized();
        runCallBacks(clr);

        if (tableDef != null)
        {
            createStatementDDL = tableDef.substituteMacros(new MacroString.MacroHandler()
                {
                    public void onIdentifierMacro(MacroString.IdentifierMacro im)
                    {
                        storeMgr.resolveIdentifierMacro(im, clr);
                    }
    
                    public void onParameterMacro(MacroString.ParameterMacro pm)
                    {
                        throw new JDOUserException(LOCALISER.msg("RDBMS.Table.MacrosNotAllowedInTableDefinition", cmd.getFullClassName(), pm));
                    }
                }, clr
            );
        }
    }

    /** Name of class currently being processed in manageClass (if any). */
    protected transient String managingClassCurrent = null;

    /** Flag to run the callbacks after the current class is managed fully. */
    protected boolean runCallbacksAfterManageClass = false;

    /**
     * Method to give management of the persistence of the fields of the specified class
     * to this table. Goes through all (non-PK) fields in the specified class and
     * adds a mapping for each. Each mapping will then call <I>prepareDatastoreMapping()</I>
     * where the columns are defined for each mapping.
     * @param theCmd ClassMetaData for the class to be managed
     * @param clr The ClassLoaderResolver
     */
    public void manageClass(AbstractClassMetaData theCmd, ClassLoaderResolver clr)
    {
        if (JPOXLogger.RDBMS.isDebugEnabled())
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("RDBMS.Table.ManagingClassFields", toString(), theCmd.getFullClassName(), 
                theCmd.getInheritanceMetaData().getStrategyValue().toString()));
        }

        managingClassCurrent = theCmd.getFullClassName();
        managedClassMetaData.add(theCmd);

        addFieldMetaData(theCmd, clr, theCmd.getManagedFields());
        addFieldMetaData(theCmd, clr, theCmd.getOverriddenFields());

        managingClassCurrent = null;
        if (runCallbacksAfterManageClass)
        {
            // We need to run the callbacks now that this class is fully managed
            runCallBacks(clr);
            runCallbacksAfterManageClass = false;
        }
    }

    /**
     * Accessor for the names of all classes managed by this table.
     * @return Names of the classes managed (stored) here
     */
    public String[] getManagedClasses()
    {
        String[] classNames = new String[managedClassMetaData.size()];
        Iterator iter = managedClassMetaData.iterator();
        int i = 0;
        while (iter.hasNext())
        {
            classNames[i++] = ((ClassMetaData)iter.next()).getFullClassName();
        }
        return classNames;
    }

    /**
     * Goes through all (non-PK) fields in the specified class and
     * adds a mapping for each. Each mapping will then call <I>prepareDatastoreMapping()</I>
     * where the columns are defined for each mapping.
     * @param theCmd ClassMetaData for the class to be managed
     * @param clr The ClassLoaderResolver
     * @param fields the fields
     */
    private void addFieldMetaData(AbstractClassMetaData theCmd, ClassLoaderResolver clr, AbstractPropertyMetaData[] fields)
    {
        // Go through the fields for this class and add columns for them
        for (int fieldNumber=0; fieldNumber<fields.length; fieldNumber++)
        {
            // Primary key fields are added by the initialisePK method
            AbstractPropertyMetaData fmd = fields[fieldNumber];
            if (!managesField(fmd.getFullFieldName()))
            {
                // Manage the field if not already managed (may already exist if overriding a superclass field)
                if (!fmd.isPrimaryKey())
                {
                    if (fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
                    {
                        boolean isPrimary = true;
                        if (fmd.getTable() != null && fmd.getJoinMetaData() == null)
                        {
                            // Field has a table specified and is not a 1-N with join table
                            // so is mapped to a secondary table
                            isPrimary = false;
                        }
                        
                        if (isPrimary)
                        {
                            // Add the field to this table
                            addFieldMapping(dba.getMappingManager().getMapping(this, fmd, dba, clr, MappingManager.MAPPING_FIELD));
                        }
                        else
                        {
                            // Add the field to the appropriate secondary table
                            if (secondaryTables == null)
                            {
                                secondaryTables = new HashMap();
                            }
                            SecondaryTable secTable = (SecondaryTable)secondaryTables.get(fmd.getTable());
                            if (secTable == null)
                            {
                                // Secondary table doesnt exists yet so create it to users specifications.
                                JoinMetaData[] joinmds = theCmd.getJoinMetaData();
                                JoinMetaData joinmd = null;
                                if (joinmds != null)
                                {
                                    for (int j=0;j<joinmds.length;j++)
                                    {
                                        if (joinmds[j].getTable().equalsIgnoreCase(fmd.getTable()))
                                        {
                                            joinmd = joinmds[j];
                                            break;
                                        }
                                    }
                                }

                                SQLIdentifier secTableIdentifier = 
                                    (SQLIdentifier)storeMgr.getIdentifierFactory().newDatastoreContainerIdentifier(fmd.getTable());
                                secTableIdentifier.setCatalogName(getCatalogName());
                                secTableIdentifier.setSchemaName(getSchemaName());

                                secTable = new SecondaryTable(secTableIdentifier, storeMgr, this, joinmd, clr);
                                secTable.preInitialize(clr);
                                secTable.initialize(clr);
                                secTable.postInitialize(clr);
                                secondaryTables.put(fmd.getTable(), secTable);
                            }
                            secTable.addFieldMapping(dba.getMappingManager().getMapping(secTable, fmd, dba, clr, MappingManager.MAPPING_FIELD));
                        }
                    }
                    else if (fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL)
                    {
                        throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.Table.InvalidPersistenceModifierForField",fmd.getName()));
                    }

                    // Calculate if we need a FK adding due to a 1-N (FK) relationship
                    boolean needsFKToContainerOwner = false;
                    int relationType = fmd.getRelationType(clr);
                    if (relationType == Relation.ONE_TO_MANY_BI)
                    {
                        AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
                        if (fmd.getJoinMetaData() == null && relatedFmd.getJoinMetaData() == null)
                        {
                            needsFKToContainerOwner = true;
                        }
                    }
                    else if (relationType == Relation.ONE_TO_MANY_UNI)
                    {
                        if (fmd.getJoinMetaData() == null)
                        {
                            needsFKToContainerOwner = true;
                        }
                    }

                    if (needsFKToContainerOwner)
                    {
                        // 1-N uni/bidirectional using FK, so update the element side with a FK
                        if ((fmd.getCollection() != null && !SCOUtils.collectionHasSerialisedElements(fmd)) ||
                            (fmd.getArray() != null && !SCOUtils.arrayIsStoredInSingleColumn(fmd)))
                        {
                            // 1-N ForeignKey collection/array, so add FK to element table
                            AbstractClassMetaData elementCmd = null;
                            if (fmd.hasCollection())
                            {
                                // Collection
                                elementCmd = storeMgr.getMetaDataManager().getMetaDataForClass(fmd.getCollection().getElementType(), clr); 
                            }
                            else
                            {
                                // Array
                                elementCmd = storeMgr.getMetaDataManager().getMetaDataForClass(fmd.getType().getComponentType(), clr);
                            }
                            if (elementCmd == null)
                            {
                                // Elements that are reference types or non-PC will come through here
                                if (fmd.hasCollection())
                                {
                                    JPOXLogger.METADATA.warn(LOCALISER.msg("RDBMS.Table.Collection.ElementHasNoMetaData", 
                                        theCmd.getFullClassName(), fmd.getCollection().getElementType()));
                                }
                                else
                                {
                                    JPOXLogger.METADATA.warn(LOCALISER.msg("RDBMS.Table.Array.ElementHasNoMetaData", 
                                        theCmd.getFullClassName(), fmd.getType().getComponentType().getName()));
                                }
                            }
                            else
                            {
                                AbstractClassMetaData[] elementCmds = null;
                                // TODO : Cater for interface elements, and get the metadata for the implementation classes here
                                if (elementCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
                                {
                                    elementCmds = storeMgr.getClassesManagingTableForClass(elementCmd, clr);
                                }
                                else
                                {
                                    elementCmds = new ClassMetaData[1];
                                    elementCmds[0] = elementCmd;
                                }

                                // Run callbacks for each of the element classes.
                                for (int i=0;i<elementCmds.length;i++)
                                {
                                    callbacks.put(elementCmds[i].getFullClassName(), new CallBack(fmd));
                                    DatastoreClass dc = storeMgr.getDatastoreClass(elementCmds[i].getFullClassName(), clr);
                                    ClassTable ct = (ClassTable) dc;
                                    if (ct.isInitialized())
                                    {
                                        // if the target table is already initialized, run the callbacks
                                        ct.runCallBacks(clr);
                                    }
                                }
                            }
                        }
                        else if (fmd.getMap() != null && !SCOUtils.mapHasSerialisedKeysAndValues(fmd))
                        {
                            // 1-N ForeignKey map, so add FK to value table
                            if (fmd.getKeyMetaData() != null && fmd.getKeyMetaData().getMappedBy() != null)
                            {
                                // Key is stored in the value table so add the FK to the value table
                                AbstractClassMetaData valueCmd = storeMgr.getMetaDataManager().getMetaDataForClass(fmd.getMap().getValueType(), clr);
                                if (valueCmd == null)
                                {
                                    // Interface elements will come through here and java.lang.String and others as well
                                    JPOXLogger.METADATA.warn(LOCALISER.msg("RDBMS.Table.Map.ValueHasNoMetaData", 
                                        theCmd.getFullClassName(), fmd.getMap().getValueType()));
                                }
                                else
                                {
                                    AbstractClassMetaData[] valueCmds = null;
                                    // TODO : Cater for interface values, and get the metadata for the implementation classes here
                                    if (valueCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
                                    {
                                        valueCmds = storeMgr.getClassesManagingTableForClass(valueCmd, clr);
                                    }
                                    else
                                    {
                                        valueCmds = new ClassMetaData[1];
                                        valueCmds[0] = valueCmd;
                                    }
                                    
                                    // Run callbacks for each of the value classes.
                                    for (int i=0;i<valueCmds.length;i++)
                                    {
                                        callbacks.put(valueCmds[i].getFullClassName(), new CallBack(fmd));
                                        DatastoreClass dc = storeMgr.getDatastoreClass(valueCmds[i].getFullClassName(), clr);
                                        ClassTable ct = (ClassTable) dc;
                                        if (ct.isInitialized())
                                        {
                                            // if the target table is already initialized, run the callbacks
                                            ct.runCallBacks(clr);
                                        }
                                    }
                                }
                            }
                            else if (fmd.getValueMetaData() != null && fmd.getValueMetaData().getMappedBy() != null)
                            {
                                // Value is stored in the key table so add the FK to the key table
                                AbstractClassMetaData keyCmd = storeMgr.getMetaDataManager().getMetaDataForClass(fmd.getMap().getKeyType(), clr);
                                if (keyCmd == null)
                                {
                                    // Interface elements will come through here and java.lang.String and others as well
                                    JPOXLogger.METADATA.warn(LOCALISER.msg("RDBMS.Table.Map.KeyHasNoMetaData", 
                                        theCmd.getFullClassName(), fmd.getMap().getKeyType()));
                                }
                                else
                                {
                                    AbstractClassMetaData[] keyCmds = null;
                                    // TODO : Cater for interface values, and get the metadata for the implementation classes here
                                    if (keyCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
                                    {
                                        keyCmds = storeMgr.getClassesManagingTableForClass(keyCmd, clr);
                                    }
                                    else
                                    {
                                        keyCmds = new ClassMetaData[1];
                                        keyCmds[0] = keyCmd;
                                    }

                                    // Run callbacks for each of the key classes.
                                    for (int i=0;i<keyCmds.length;i++)
                                    {
                                        callbacks.put(keyCmds[i].getFullClassName(), new CallBack(fmd));
                                        DatastoreClass dc = storeMgr.getDatastoreClass(keyCmds[i].getFullClassName(), clr);
                                        ClassTable ct = (ClassTable) dc;
                                        if (ct.isInitialized())
                                        {
                                            // if the target table is already initialized, run the callbacks
                                            ct.runCallBacks(clr);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Accessor for whether this table manages the specified class
     * @param className Name of the class
     * @return Whether it is managed by this table
     */
    public boolean managesClass(String className)
    {
        if (className == null)
        {
            return false;
        }

        Iterator iter = managedClassMetaData.iterator();
        while (iter.hasNext())
        {
            ClassMetaData managedCmd = (ClassMetaData)iter.next();
            if (managedCmd.getFullClassName().equals(className))
            {
                return true;
            }
        }

        return false;
    }

    /**
     * Method to initialise the table primary key field(s).
     * @param clr The ClassLoaderResolver
     **/
    protected void initializePK(ClassLoaderResolver clr)
    {
        assertIsPKUninitialized();
        AbstractPropertyMetaData[] fieldsToAdd = new AbstractPropertyMetaData[cmd.getNoOfPrimaryKeyFields()];
        
        // Initialise Primary Key mappings for application id with PK fields in this class
        int pkFieldNum=0;
        int fieldCount = cmd.getNoOfManagedFields();
        boolean hasPrimaryKeyInThisClass = false; 
        if (cmd.getNoOfPrimaryKeyFields() > 0)
        {
            pkMappings = new JavaTypeMapping[cmd.getNoOfPrimaryKeyFields()];
            for (int relativeFieldNumber = 0; relativeFieldNumber < fieldCount; ++relativeFieldNumber)
            {
                AbstractPropertyMetaData fmd = cmd.getManagedField(relativeFieldNumber);
                if (fmd.isPrimaryKey())
                {
                    if (fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
                    {
                        fieldsToAdd[pkFieldNum++] = fmd;
                        hasPrimaryKeyInThisClass = true;
                    }
                    else if (fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL)
                    {
                        throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.Table.InvalidPersistenceModifierForField",fmd.getName()));
                    }

                    // Check if auto-increment and that it is supported by this RDBMS
                    if ((fmd.getValueStrategy() == IdentityStrategy.IDENTITY) && !dba.supportsAutoIncrementFields())
                    {
                        throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.Table.AutoIncrementNotSupportedByRDBMS", 
                            cmd.getFullClassName(), fmd.getName()));
                    }
                }
            }
        }

        // No Primary Key defined, so search for superclass or handle datastore id
        if (!hasPrimaryKeyInThisClass)
        {
            // application identity will inherit PK columns from super class
            if (cmd.getIdentityType() == IdentityType.APPLICATION)
            {
                // Check if the superclass has a table and, if so, copy its PK mappings
                DatastoreClass elementCT = storeMgr.getDatastoreClass(cmd.getPersistenceCapableSuperclass(), clr);
                if (elementCT != null)
                {
                    addApplicationIdUsingClassTableId(
                        cmd.getInheritanceMetaData() != null ? cmd.getInheritanceMetaData().getJoinMetaData() : null,
                        elementCT, clr, cmd);
                }
                else
                {
                    // Superclass has no table so create new mappings and columns
                    AbstractClassMetaData pkCmd = storeMgr.getClassWithPrimaryKeyForClass(cmd.getSuperAbstractClassMetaData(), clr);
                    if (pkCmd != null)
                    {
                        pkMappings = new JavaTypeMapping[pkCmd.getNoOfPrimaryKeyFields()];
                        pkFieldNum = 0;
                        fieldCount = pkCmd.getNoOfInheritedManagedFields() + pkCmd.getNoOfManagedFields();
                        for (int absFieldNum = 0; absFieldNum < fieldCount; ++absFieldNum)
                        {
                            AbstractPropertyMetaData fmd = pkCmd.getManagedFieldAbsolute(absFieldNum);
                            if (fmd.isPrimaryKey())
                            {
                                AbstractPropertyMetaData overriddenFmd = cmd.getOverriddenField(fmd.getName());
                                if (overriddenFmd != null)
                                {
                                    // PK field is overridden so use the overriding definition
                                    fmd = overriddenFmd;
                                }

                                if (fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
                                {
                                    fieldsToAdd[pkFieldNum++] = fmd;
                                }
                                else if (fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL)
                                {
                                    throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.Table.InvalidPersistenceModifierForField",fmd.getName()));
                                }
                            }
                        }
                    }
                }
            }
            // datastore identity
            else if (cmd.getIdentityType() == IdentityType.DATASTORE)
            {
                addDatastoreId(cmd.getIdentityMetaData(), null, cmd);
            }
            else if (cmd.getIdentityType() == IdentityType.NONDURABLE)
            {
                // Do nothing since no identity!
            }
        }

        //add field mappings in the end, so we compute all columns after the post initialize
        for( int i=0; i<fieldsToAdd.length; i++)
        {
            if (fieldsToAdd[i] != null)
            {
                try
                {
                    DatastoreClass datastoreClass = getStoreManager().getDatastoreClass(fieldsToAdd[i].getType().getName(),clr);
                    if (datastoreClass.getIDMapping() == null)
                    {
                        throw new JDOFatalInternalException("Unsupported relationship with field "+fieldsToAdd[i].getFullFieldName());
                    }
                }
                catch (NoTableManagedException ex)
                {
                    //do nothing
                }
                JavaTypeMapping fieldMapping = dba.getMappingManager().getMapping(this, fieldsToAdd[i], dba, clr, MappingManager.MAPPING_FIELD);
                addFieldMapping(fieldMapping);
                pkMappings[i] = fieldMapping;
            }
        }
        initializeIDMapping();
        
        state = TABLE_STATE_PK_INITIALIZED;
    }

    /**
     * Method to initialise this table to include all fields in the specified class.
     * This is used to recurse up the hierarchy so that we include all immediate superclasses
     * that have "subclass-table" specified as their inheritance strategy. If we encounter the parent of
     * this class with other than "subclass-table" we stop the process.
     * @param theCmd The ClassMetaData for the class
     */
    private void initializeForClass(AbstractClassMetaData theCmd, ClassLoaderResolver clr)
    {
        // Manage the fields for this class
        manageClass(theCmd, clr);

        // Now check the superclass of this class
        // This caters for superclasses which have their fields persisted with the subclass ("subclass-table" strategy)
        AbstractClassMetaData parentCmd = theCmd.getSuperAbstractClassMetaData();
        if (parentCmd != null && parentCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
        {
            initializeForClass(parentCmd, clr);
        }
    }

    /**
     * Create a newOID for a class.
     * Works for datastore identity, or application identity (inc composite keys)
     * @param pm The PersistenceManager
     * @param pcClass The persistence capable class
     * @return The new OID
     */
    public synchronized final Object newOID(PersistenceManager pm, PersistenceCapable pcClass)
    {
        Object id = null;

        // Obtain metadata specific to the passed-in class
        ClassMetaData cmd = ((ClassTable)this.getStoreManager().getDatastoreClass(pcClass.getClass().getName(), pm.getClassLoaderResolver())).getClassMetaData();

        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            // All values will have been populated by this point
            id = AIDUtils.getNewObjectId(pcClass, cmd);
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            // Populate any strategy value for the "datastore-identity" element
            IdentityMetaData imd = cmd.getIdentityMetaData();
            Object nextIdentifier = storeMgr.getStrategyValue(pm, this, cmd, -1, imd.getStrategyValue(), imd.getSequence(),
                imd.getExtensions());
            id = OIDFactory.getInstance(cmd.getFullClassName(), nextIdentifier);
        }

        return id;
    }

    /**
     * Execute the callbacks for the classes that this table maps to.
     * @param clr ClassLoader resolver
     */
    private void runCallBacks(ClassLoaderResolver clr)
    {
        // Run callbacks for all classes managed by this table
        Iterator cmdIter = managedClassMetaData.iterator();
        while (cmdIter.hasNext())
        {
            ClassMetaData managedCmd = (ClassMetaData)cmdIter.next();
            if (managingClassCurrent != null && managingClassCurrent.equals(managedCmd.getFullClassName()))
            {
                // We can't run callbacks for this class since it is stil being initialised. Mark callbacks to run after it completes
                runCallbacksAfterManageClass = true;
                break;
            }
            Collection c = (Collection)callbacks.remove(managedCmd.getFullClassName());
            runSpecifiedCallbacks(managedCmd, c, clr);
        }
    }

    /**
     * Convenience method to execute callbacks to add FK fields.
     * @param classCmd ClassMetaData for the class whose callbacks we're adding
     * @param callbacks Callbacks to be run
     * @param clr ClassLoader resolver
     */
    private void runSpecifiedCallbacks(ClassMetaData classCmd, Collection callbacks, ClassLoaderResolver clr)
    {
        if (callbacks != null)
        {
            for (Iterator it = callbacks.iterator(); it.hasNext();)
            {
                CallBack callback = (CallBack) it.next();

                if (callback.fmd.getJoinMetaData() == null)
                {
                    // 1-N FK relationship
                    AbstractPropertyMetaData ownerFmd = callback.fmd;
                    if (ownerFmd.getMappedBy() != null)
                    {
                        // Bidirectional (element has a PC mapping to the owner)
                        // Check that the "mapped-by" field in the other class actually exists
                        AbstractPropertyMetaData fmd = classCmd.getField(ownerFmd.getMappedBy());
                        if (fmd == null)
                        {
                            throw new JDOUserException(LOCALISER.msg("RDBMS.Table.RelationshipMappedByFieldNotFound",
                                ownerFmd.getMappedBy(), classCmd.getFullClassName(), ownerFmd.getFullFieldName()));
                        }

                        if (ownerFmd.getMap() != null && storeMgr.getPMFContext().getPmfConfiguration().getUniqueConstraintsMapInverse())
                        {
                            initializeFKMapUniqueConstraints(ownerFmd);
                        }

                        boolean duplicate = false;
                        JavaTypeMapping fkDiscrimMapping = null;
                        JavaTypeMapping orderMapping = null;
                        if (ownerFmd.hasExtension("relation-discriminator-column"))
                        {
                            // Collection has a relation discriminator so we need to share the FK. Check for the required discriminator
                            String colName = ownerFmd.getValueForExtension("relation-discriminator-column");
                            if (colName == null)
                            {
                                // No column defined so use a fallback name
                                colName = "RELATION_DISCRIM";
                            }
                            Set fkDiscrimEntries = getExternalFkDiscriminatorMappings().entrySet();
                            Iterator discrimMappingIter = fkDiscrimEntries.iterator();
                            while (discrimMappingIter.hasNext())
                            {
                                Map.Entry entry = (Map.Entry)discrimMappingIter.next();
                                JavaTypeMapping discrimMapping = (JavaTypeMapping)entry.getValue();
                                String discrimColName = ((ColumnMetaData)discrimMapping.getDataStoreMapping(0).getDatastoreField().getExtendableMetaData()).getName();
                                if (discrimColName.equalsIgnoreCase(colName))
                                {
                                    duplicate = true;
                                    fkDiscrimMapping = discrimMapping;
                                    orderMapping = (JavaTypeMapping)getExternalOrderMappings().get(entry.getKey());
                                    break;
                                }
                            }

                            if (!duplicate)
                            {
                                // Create the relation discriminator column since we dont have this discriminator
                                ColumnMetaData colmd = new ColumnMetaData(null, colName);
                                colmd.setAllowsNull(Boolean.TRUE); // Allow for elements not in any discriminated collection
                                fkDiscrimMapping = dba.getMapping(String.class, storeMgr); // Only support String discriminators currently
                                ColumnCreator.createAdapterColumn(fkDiscrimMapping, storeMgr, clr, this, colmd, false);
                            }

                            if (fkDiscrimMapping != null)
                            {
                                getExternalFkDiscriminatorMappings().put(ownerFmd, fkDiscrimMapping);
                            }
                        }

                        if (storeMgr.getPMFContext().getTypeManager().isSCOList(ownerFmd.getType()) ||
                            ownerFmd.hasArray() || ownerFmd.getOrderMetaData() != null)
                        {
                            // if the field is list or array type, add index column
                            state = TABLE_STATE_NEW;
                            if (orderMapping == null)
                            {
                                // Create new order mapping since we aren't using a shared FK
                                orderMapping = this.addOrderColumn(ownerFmd, clr);
                            }
                            getExternalOrderMappings().put(ownerFmd, orderMapping);
                            state = TABLE_STATE_INITIALIZED;
                        }
                    }
                    else
                    {
                        // Unidirectional (element knows nothing about the owner)
                        String ownerClassName = ownerFmd.getAbstractClassMetaData().getFullClassName();
                        JavaTypeMapping fkMapping = new PersistenceCapableMapping();
                        fkMapping.initialize(dba, ownerClassName);
                        JavaTypeMapping fkDiscrimMapping = null;
                        JavaTypeMapping orderMapping = null;
                        boolean duplicate = false;

                        try
                        {
                            // Get the owner id mapping of the "1" end
                            JavaTypeMapping ownerIdMapping = storeMgr.getDatastoreClass(ownerClassName, clr).getIDMapping();
                            ColumnMetaDataContainer colmdContainer = null;
                            if (ownerFmd.hasCollection() || ownerFmd.hasArray())
                            {
                                // 1-N Collection/array
                                colmdContainer = ownerFmd.getElementMetaData();
                            }
                            else if (ownerFmd.hasMap() && ownerFmd.getKeyMetaData() != null && ownerFmd.getKeyMetaData().getMappedBy() != null)
                            {
                                // 1-N Map with key stored in the value
                                colmdContainer = ownerFmd.getValueMetaData();
                            }
                            else if (ownerFmd.hasMap() && ownerFmd.getValueMetaData() != null && ownerFmd.getValueMetaData().getMappedBy() != null)
                            {
                                // 1-N Map with value stored in the key
                                colmdContainer = ownerFmd.getKeyMetaData();
                            }
                            CorrespondentColumnsMapping correspondentColumnsMapping = 
                                new CorrespondentColumnsMapping(colmdContainer, ownerIdMapping, true);
                            int countIdFields = ownerIdMapping.getNumberOfDatastoreFields();
                            for (int i=0; i<countIdFields; i++)
                            {
                                DatastoreMapping refDatastoreMapping = ownerIdMapping.getDataStoreMapping(i);
                                JavaTypeMapping mapping = dba.getMapping(refDatastoreMapping.getJavaTypeMapping().getJavaType(), storeMgr);
                                ColumnMetaData colmd = correspondentColumnsMapping.getColumnMetaDataByIdentifier(((Column)refDatastoreMapping.getDatastoreField()).getIdentifier());
                                if (colmd == null)
                                {
                                    throw new JDOFatalUserException(LOCALISER.msg("RDBMS.Table.PrimaryKeyColumnNotMapped",
                                        ((Column)refDatastoreMapping.getDatastoreField()).getIdentifier(), toString()));
                                }

                                DatastoreIdentifier identifier = null;
                                IdentifierFactory idFactory = storeMgr.getIdentifierFactory();
                                if (colmd.getName() == null || colmd.getName().length() < 1)
                                {
                                    // No user provided name so generate one
                                    identifier = ((RDBMSIdentifierFactory)idFactory).newForeignKeyFieldIdentifier(ownerFmd, 
                                        null, refDatastoreMapping.getDatastoreField().getIdentifier(), 
                                        storeMgr.getPMFContext().getTypeManager().isDefaultEmbeddedType(mapping.getJavaType()),
                                        DatastoreField.ROLE_OWNER);
                                }
                                else
                                {
                                    // User-defined name
                                    identifier = idFactory.newDatastoreFieldIdentifier(colmd.getName());
                                }
                                DatastoreField refColumn = addDatastoreField(mapping.getJavaType().getName(), identifier, mapping, colmd);
                                ((Column)refDatastoreMapping.getDatastoreField()).copyConfigurationTo(refColumn);

                                if (colmd == null || (colmd != null && !colmd.isAllowsNullSet()) ||
                                    (colmd != null && colmd.isAllowsNullSet() && colmd.isAllowsNull()))
                                {
                                    // User either wants it nullable, or havent specified anything, so make it nullable
                                    refColumn.setNullable();
                                }

                                fkMapping.addDataStoreMapping(getStoreManager().getMappingManager().createDatastoreMapping(mapping, storeMgr, refColumn, refDatastoreMapping.getJavaTypeMapping().getJavaType().getName()));
                                ((PersistenceCapableMapping)fkMapping).addJavaTypeMapping(mapping);
                            }
                        }
                        catch (DuplicateColumnNameException dcne)
                        {
                            // If the user hasnt specified "relation-discriminator-column" here we dont allow the sharing of columns
                            if (!ownerFmd.hasExtension("relation-discriminator-column"))
                            {
                                throw dcne;
                            }

                            // Find the FK using this column and use it instead of creating a new one since we're sharing
                            Iterator fkIter = getExternalFkMappings().entrySet().iterator();
                            fkMapping = null;
                            while (fkIter.hasNext())
                            {
                                Map.Entry entry = (Map.Entry)fkIter.next();
                                JavaTypeMapping existingFkMapping = (JavaTypeMapping)entry.getValue();
                                for (int j=0;j<existingFkMapping.getNumberOfDatastoreFields();j++)
                                {
                                    if (existingFkMapping.getDataStoreMapping(j).getDatastoreField().getIdentifier().toString().equals(dcne.getConflictingColumn().getIdentifier().toString()))
                                    {
                                        // The FK is shared (and so if it is a List we also share the index)
                                        fkMapping = existingFkMapping;
                                        fkDiscrimMapping = (JavaTypeMapping)externalFkDiscriminatorMappings.get(entry.getKey());
                                        orderMapping = (JavaTypeMapping)getExternalOrderMappings().get(entry.getKey());
                                        break;
                                    }
                                }
                            }
                            if (fkMapping == null)
                            {
                                // Should never happen since we know there is a col duplicating ours
                                throw dcne;
                            }
                            duplicate = true;
                        }

                        if (!duplicate && ownerFmd.hasExtension("relation-discriminator-column"))
                        {
                            // Create the relation discriminator column
                            String colName = ownerFmd.getValueForExtension("relation-discriminator-column");
                            if (colName == null)
                            {
                                // No column defined so use a fallback name
                                colName = "RELATION_DISCRIM";
                            }
                            ColumnMetaData colmd = new ColumnMetaData(null, colName);
                            colmd.setAllowsNull(Boolean.TRUE); // Allow for elements not in any discriminated collection
                            fkDiscrimMapping = dba.getMapping(String.class, storeMgr); // Only support String discriminators currently
                            ColumnCreator.createAdapterColumn(fkDiscrimMapping, storeMgr, clr, this, colmd, false);
                        }

                        // Save the external FK
                        getExternalFkMappings().put(ownerFmd, fkMapping);
                        if (fkDiscrimMapping != null)
                        {
                            getExternalFkDiscriminatorMappings().put(ownerFmd, fkDiscrimMapping);
                        }

                        if (storeMgr.getPMFContext().getTypeManager().isSCOList(ownerFmd.getType()) ||
                            ownerFmd.hasArray() || ownerFmd.getOrderMetaData() != null)
                        {
                            // if the field is list or array type, add index column
                            state = TABLE_STATE_NEW;
                            if (orderMapping == null)
                            {
                                // Create new order mapping since we need one and we aren't using a shared FK
                                orderMapping = this.addOrderColumn(ownerFmd, clr);
                            }
                            getExternalOrderMappings().put(ownerFmd, orderMapping);
                            state = TABLE_STATE_INITIALIZED;
                        }
                    }
                }
            }
        }
    }

    /**
     * Accessor for the main class represented.
     * @return The name of the class
     **/
    public String getType()
    {
        return cmd.getFullClassName();
    }

    /**
     * Accessor for the identity-type. 
     * @return identity-type tag value
     */    
    public IdentityType getIdentityType()
    {
        return cmd.getIdentityType();
    }

    /**
     * Accessor for versionMetaData
     * @return Returns the versionMetaData.
     */
    public final VersionMetaData getVersionMetaData()
    {
        return cmd.getVersionMetaData();
    }

    /**
     * Accessor for Discriminator MetaData
     * @return Returns the Discriminator MetaData.
     */
    public final DiscriminatorMetaData getDiscriminatorMetaData()
    {
        return cmd.getInheritanceMetaData().getDiscriminatorMetaData();
    }

    /**
     * Whether this table or super table has id (primary key) attributed by the datastore
     * @return true if the id attributed by the datastore
     */
    public boolean isObjectIDDatastoreAttributed()
    {
        boolean attributed = storeMgr.isStrategyDatastoreAttributed(cmd.getIdentityMetaData().getStrategyValue());
        if (attributed)
        {
            return true;
        }
        for (int i=0; i<columns.size(); i++)
        {
            Column col = (Column) columns.get(i);
            if (col.isPrimaryKey() && col.isAutoIncrement())
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Whether this table is the base table in the inheritance hierarchy.
     * @return true if this table is a root table
     */
    public boolean isBaseDatastoreClass()
    {
        return supertable == null ? true : false;
    }

    /**
     * Accessor for the supertable for this table.
     * @return The supertable
     **/
    public DatastoreClass getSuperDatastoreClass()
    {
        assertIsInitialized();

        return supertable;
    }

    /**
     * Utility to find the table above this one. Will recurse to cater for inheritance
     * strategies where fields are handed up to the super class, or down to this class.
     * @param theCmd ClassMetaData of the class to find the supertable for.
     * @return The table above this one in any inheritance hierarchy
     */
    private ClassTable getSupertable(AbstractClassMetaData theCmd, ClassLoaderResolver clr)
    {
        AbstractClassMetaData superCmd = theCmd.getSuperAbstractClassMetaData();
        if (superCmd != null)
        {
            if (superCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NEW_TABLE)
            {
                // This class has its own table, so return it.
                return (ClassTable) storeMgr.getDatastoreClass(superCmd.getFullClassName(), clr);
            }
            else if (superCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
            {
                // This class is mapped to the same table, so go up another level.
                return getSupertable(superCmd, clr);
            }
            else
            {
                // This class is mapped to its superclass table, so go up to that and try again.
                return getSupertable(superCmd, clr);
            }
        }
        return null;
    }

    /**
     * Convenience accessor for the base table for this table which has the specified field.
     * @param fmd Field MetaData for this field
     * @return The base table which has the field specified
     */
    public DatastoreClass getBaseDatastoreClassWithField(AbstractPropertyMetaData fmd)
    {
        if (fmd.isPrimaryKey())
        {
            // If this is a PK field then it will be in all tables up to the base so we need to continue navigating up
            if (getSuperDatastoreClass() != null)
            {
                return getSuperDatastoreClass().getBaseDatastoreClassWithField(fmd);
            }
        }
        if (fieldMappingsMap.get(fmd) != null)
        {
            // We have this field so return this table
            return this;
        }
        else if (getSuperDatastoreClass() == null)
        {
            // We don't have the field, but have no superclass, so return null
            return this;
        }
        else
        {
            // Return the superclass sicne we don't have it
            return getSuperDatastoreClass().getBaseDatastoreClassWithField(fmd);
        }
    }

    /**
     * Accessor for the (primary) class MetaData.
     * Package-level access to restrict to other table types only.
     * @return The (primary) class MetaData
     **/
    ClassMetaData getClassMetaData()
    {
        return cmd;
    }

    /**
     * Accessor for the indices for this table. This includes both the
     * user-defined indices (via MetaData), and the ones required by foreign
     * keys (required by relationships).
     * @param clr The ClassLoaderResolver
     * @return The indices
     */
    protected Set getExpectedIndices(ClassLoaderResolver clr)
    {
        // The indices required by foreign keys (BaseTable)
        Set indices = super.getExpectedIndices(clr);

        // Add on any user-required indices for the fields
        Set fieldNumbersSet = fieldMappingsMap.keySet();
        Iterator iter = fieldNumbersSet.iterator();
        while (iter.hasNext())
        {
            AbstractPropertyMetaData fmd = (AbstractPropertyMetaData) iter.next();
            JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmd);
            if (fieldMapping instanceof EmbeddedPCMapping)
            {
                // Add indexes for fields of this embedded PC object
                EmbeddedPCMapping embMapping = (EmbeddedPCMapping)fieldMapping;
                for (int i=0;i<embMapping.getNumberOfJavaTypeMappings();i++)
                {
                    JavaTypeMapping embFieldMapping = embMapping.getJavaTypeMapping(i);
                    IndexMetaData imd = embFieldMapping.getFieldMetaData().getIndexMetaData();
                    if (imd != null)
                    {
                        Index index = TableUtils.getIndexForField(this, imd, embFieldMapping);
                        if (index != null)
                        {
                            indices.add(index);
                        }
                    }
                }
            }
            else
            {
                // Add any required index for this field
                IndexMetaData imd = fmd.getIndexMetaData();
                if (imd != null)
                {
                    Index index = TableUtils.getIndexForField(this, imd, fieldMapping);
                    if (index != null)
                    {
                        indices.add(index);
                    }
                }
            }
        }

        // Check if any version column needs indexing
        if (versionMapping != null)
        {
            VersionMetaData vermd = getVersionMetaData();
            IndexMetaData idxmd = vermd.getIndexMetaData();
            if (idxmd != null)
            {
                Index index = new Index(this, idxmd.isUnique() != null ? idxmd.isUnique().booleanValue() : false, idxmd.getValueForExtension("extended-setting"));
                if (idxmd.getName() != null)
                {
                    index.setName(idxmd.getName());
                }
                int countVersionFields = versionMapping.getNumberOfDatastoreFields();
                for (int i=0; i<countVersionFields; i++)
                {
                    index.addDatastoreField(versionMapping.getDataStoreMapping(i).getDatastoreField());
                }
                indices.add(index);
            }
        }

        // Check if any discriminator column needs indexing
        if (discriminatorMapping != null)
        {
            DiscriminatorMetaData dismd = getDiscriminatorMetaData();
            IndexMetaData idxmd = dismd.getIndexMetaData();
            if (idxmd != null)
            {
                Index index = new Index(this, idxmd.isUnique() != null ? idxmd.isUnique().booleanValue() : false, idxmd.getValueForExtension("extended-setting"));
                if (idxmd.getName() != null)
                {
                    index.setName(idxmd.getName());
                }
                int countDiscrimFields = discriminatorMapping.getNumberOfDatastoreFields();
                for (int i=0; i<countDiscrimFields; i++)
                {
                    index.addDatastoreField(discriminatorMapping.getDataStoreMapping(i).getDatastoreField());
                }
                indices.add(index);
            }
        }

        // Add on any order fields (for lists, arrays, collections) that need indexing
        Set orderMappingsEntries = getExternalOrderMappings().entrySet();
        Iterator orderMappingsEntriesIter = orderMappingsEntries.iterator();
        while (orderMappingsEntriesIter.hasNext())
        {
            Map.Entry entry = (Map.Entry)orderMappingsEntriesIter.next();
            AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)entry.getKey();
            JavaTypeMapping mapping = (JavaTypeMapping)entry.getValue();
            OrderMetaData omd = fmd.getOrderMetaData();
            if (omd != null && omd.getIndexMetaData() != null)
            {
                Index index = getIndexForIndexMetaDataAndMapping(omd.getIndexMetaData(), mapping);
                if (index != null)
                {
                    indices.add(index);
                }
            }
        }

        // Add on any user-required indices for the class(es) as a whole (subelement of <class>)
        Iterator cmdIter = managedClassMetaData.iterator();
        while (cmdIter.hasNext())
        {
            ClassMetaData thisCmd = (ClassMetaData)cmdIter.next();
            IndexMetaData[] classIndices = thisCmd.getIndexMetaData();
            if (classIndices != null)
            {
                for (int i=0;i<classIndices.length;i++)
                {
                    Index index = getIndexForIndexMetaData(classIndices[i]);
                    if (index != null)
                    {
                        indices.add(index);
                    }
                }
            }
        }

        return indices;
    }

    /**
     * Convenience method to convert an IndexMetaData and a mapping into an Index.
     * @param imd The Index MetaData
     * @param mapping The mapping
     * @return The Index
     */
    private Index getIndexForIndexMetaDataAndMapping(IndexMetaData imd, JavaTypeMapping mapping)
    {
        // Verify if a unique index is needed
        boolean unique=false;
        if (imd.isUnique() != null && imd.isUnique().booleanValue())
        {
            unique = true;
        }

        Index index = new Index(this, unique, imd.getValueForExtension("extended-setting"));

        // Set the index name if required
        if (imd.getName() != null)
        {
            index.setName(imd.getName());
        }

        int numCols = mapping.getNumberOfDatastoreFields();
        for (int i=0;i<numCols;i++)
        {
            index.addDatastoreField(mapping.getDataStoreMapping(i).getDatastoreField());
        }

        return index;
    }

    /**
     * Convenience method to convert an IndexMetaData into an Index.
     * @param imd The Index MetaData
     * @return The Index
     */
    private Index getIndexForIndexMetaData(IndexMetaData imd)
    {
        // Verify if a unique index is needed
        boolean unique=false;
        if (imd.isUnique() != null && imd.isUnique().booleanValue())
        {
            unique = true;
        }

        Index index = new Index(this, unique, imd.getValueForExtension("extended-setting"));

        // Set the index name if required
        if (imd.getName() != null)
        {
            index.setName(imd.getName());
        }

        // Set the column(s) to index
        // Class-level index so use its column definition
        ColumnMetaData[] colmds = imd.getColumnMetaData();
        AbstractPropertyMetaData[] fmds = imd.getFieldMetaData();
        // a). Columns specified directly
        if (colmds != null && colmds.length > 0)
        {
            for (int i=0;i<colmds.length;i++)
            {
                DatastoreIdentifier colName = storeMgr.getIdentifierFactory().newDatastoreFieldIdentifier(colmds[i].getName());
                Column col = (Column)columnsByName.get(colName);
                if (col == null)
                {
                    JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMS.Table.Index.ColumnNotFound", 
                        toString(), index.getName(), colmds[i].getName()));
                    break;
                }
                else
                {
                    index.addDatastoreField(col);
                }
            }
        }
        // b). Columns specified using fields
        else if (fmds != null && fmds.length > 0)
        {
            for (int i=0;i<fmds.length;i++)
            {
                // Find the metadata for the actual field with the same name as this "index" field
                AbstractPropertyMetaData realFmd = getFieldMetaData(fmds[i].getName());
                JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(realFmd);
                int countFields = fieldMapping.getNumberOfDatastoreFields();
                for (int j=0; j<countFields; j++)
                {
                    index.addDatastoreField(fieldMapping.getDataStoreMapping(j).getDatastoreField());
                }
            }
        }
        else
        {
            // We can't have an index of no columns
            JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMS.Table.Index.NoColumnsSpecified", 
                toString(), index.getName()));
            return null;
        }

        return index;
    }

    /**
     * Convenience method to add the expected FKs for an embedded PC field.
     * @param foreignKeys The list of FKs to add the FKs to
     * @param autoMode Whether operating in "auto-mode" where JPOX can create its own FKs
     * @param clr ClassLoader resolver
     * @param embeddedMapping The embedded PC mapping
     */
    private void addExpectedForeignKeysForEmbeddedPCField(List foreignKeys, boolean autoMode, ClassLoaderResolver clr, 
            EmbeddedPCMapping embeddedMapping)
    {
        for (int i=0;i<embeddedMapping.getNumberOfJavaTypeMappings();i++)
        {
            JavaTypeMapping embFieldMapping = embeddedMapping.getJavaTypeMapping(i);
            if (embFieldMapping instanceof EmbeddedPCMapping)
            {
                // Nested embedded PC so add the FKs for that
                addExpectedForeignKeysForEmbeddedPCField(foreignKeys, autoMode, clr, (EmbeddedPCMapping)embFieldMapping);
            }
            else
            {
                AbstractPropertyMetaData embFmd = embFieldMapping.getFieldMetaData();
                if (storeMgr.getPMFContext().getTypeManager().isReferenceType(embFmd.getType()) && 
                    embFieldMapping instanceof ReferenceMapping)
                {
                    // Field is a reference type, so add a FK to the table of the PC for each PC implementation
                    Collection fks = TableUtils.getForeignKeysForReferenceField(embFieldMapping, embFmd, autoMode, storeMgr, clr);
                    foreignKeys.addAll(fks);
                }
                else if (storeMgr.getMetaDataManager().getMetaDataForClass(embFmd.getType(), clr) != null &&
                        embFieldMapping.getNumberOfDatastoreFields() > 0 &&
                        embFieldMapping instanceof PersistenceCapableMapping)
                {
                    // Field is for a PC class with the FK at this side, so add a FK to the table of this PC
                    ForeignKey fk = TableUtils.getForeignKeyForPCField(embFieldMapping, embFmd, autoMode, storeMgr, clr);
                    if (fk != null)
                    {
                        foreignKeys.add(fk);
                    }
                }
            }
        }
    }

    /**
     * Accessor for the expected foreign keys for this table.
     * @param clr The ClassLoaderResolver
     * @return The expected foreign keys.
     **/
    protected List getExpectedForeignKeys(ClassLoaderResolver clr)
    {
        assertIsInitialized();

        // Auto mode allows JPOX to decide which FKs are needed as well as using
        // what is in the users MetaData.
        boolean autoMode = false;
        if (storeMgr.getPMFContext().getPmfConfiguration().getForeignKeyCreateMode() == PMFConfiguration.FOREIGN_KEY_AUTO_MODE)
        {
            autoMode = true;
        }

        ArrayList foreignKeys = new ArrayList();

        // Check each field for FK requirements (user-defined, or required)
        // <field><foreign-key>...</foreign-key></field>
        Set fieldNumbersSet = fieldMappingsMap.keySet();
        Iterator iter = fieldNumbersSet.iterator();
        while (iter.hasNext())
        {
            AbstractPropertyMetaData fmd = (AbstractPropertyMetaData) iter.next();
            JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmd);

            if (fmd.getEmbeddedMetaData() != null && fieldMapping instanceof EmbeddedPCMapping)
            {
                EmbeddedPCMapping embMapping = (EmbeddedPCMapping)fieldMapping;
                addExpectedForeignKeysForEmbeddedPCField(foreignKeys, autoMode, clr, embMapping);
            }
            else
            {
                if (storeMgr.getPMFContext().getTypeManager().isReferenceType(fmd.getType()) && fieldMapping instanceof ReferenceMapping)
                {
                    // Field is a reference type, so add a FK to the table of the PC for each PC implementation
                    Collection fks = TableUtils.getForeignKeysForReferenceField(fieldMapping, fmd, autoMode, storeMgr, clr);
                    foreignKeys.addAll(fks);
                }
                else if (storeMgr.getMetaDataManager().getMetaDataForClass(fmd.getType(), clr) != null &&
                        fieldMapping.getNumberOfDatastoreFields() > 0 &&
                        fieldMapping instanceof PersistenceCapableMapping)
                {
                    // Field is for a PC class with the FK at this side, so add a FK to the table of this PC
                    ForeignKey fk = TableUtils.getForeignKeyForPCField(fieldMapping, fmd, autoMode, storeMgr, clr);
                    if (fk != null)
                    {
                        foreignKeys.add(fk);
                    }
                }
            }
        }

        // FK from id column(s) to id column(s) of superclass, as specified by
        // <inheritance><join><foreign-key ...></join></inheritance>
        ForeignKeyMetaData idFkmd = (cmd.getInheritanceMetaData().getJoinMetaData() != null) ? 
                cmd.getInheritanceMetaData().getJoinMetaData().getForeignKeyMetaData() : null;
        if (supertable != null && (autoMode || (idFkmd != null && idFkmd.getDeleteAction() != ForeignKeyAction.NONE)))
        {
            ForeignKey fk = new ForeignKey(getIDMapping(), dba, supertable, false);
            if (idFkmd != null && idFkmd.getName() != null)
            {
                fk.setName(idFkmd.getName());
            }
            foreignKeys.add(0, fk);
        }

        // Add any user-required FKs for the class as a whole
        // <class><foreign-key>...</foreign-key></field>
        Iterator cmdIter = managedClassMetaData.iterator();
        while (cmdIter.hasNext())
        {
            ClassMetaData thisCmd = (ClassMetaData)cmdIter.next();
            ForeignKeyMetaData[] fkmds = thisCmd.getForeignKeyMetaData();
            if (fkmds != null)
            {
                // TODO [CORE-2584] Enable this and fully implement the getForeignKeyForForeignKeyMetaData method
                /*for (int i=0;i<fkmds.length;i++)
                {
                    ForeignKey fk = getForeignKeyForForeignKeyMetaData(fkmds[i]);
                    if (fk != null)
                    {
                        foreignKeys.add(fk);
                    }
                }*/
            }
        }

        HashMap externalFks = getExternalFkMappings();
        if (!externalFks.isEmpty())
        {
            // 1-N FK relationships - FK to id column(s) of owner table where this is the element table and we have a FK
            Collection externalFkKeys = externalFks.keySet();
            Iterator externalFkKeysIter = externalFkKeys.iterator();
            while (externalFkKeysIter.hasNext())
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)externalFkKeysIter.next();
                DatastoreClass referencedTable = storeMgr.getDatastoreClass(fmd.getAbstractClassMetaData().getFullClassName(), clr);
                if (referencedTable != null)
                {
                    // Take <foreign-key> from either <field> or <element>
                    ForeignKeyMetaData fkmd = fmd.getForeignKeyMetaData();
                    if (fkmd == null && fmd.getElementMetaData() != null)
                    {
                        fkmd = fmd.getElementMetaData().getForeignKeyMetaData();
                    }
                    if ((fkmd != null && fkmd.getDeleteAction() != ForeignKeyAction.NONE) || autoMode)
                    {
                        // Either has been specified by user, or using autoMode, so add FK
                        JavaTypeMapping fkMapping = (JavaTypeMapping)externalFks.get(fmd);
                        ForeignKey fk = new ForeignKey(fkMapping, dba, referencedTable, true);
                        fk.setForMetaData(fkmd); // Does nothing when no FK MetaData
                        if (!foreignKeys.contains(fk))
                        {
                            // Only add when not already present (in the case of shared FKs there can be dups here)
                            foreignKeys.add(fk);
                        }
                    }
                }
            }
        }

        return foreignKeys;
    }

    /**
     * Convenience method to create a FK for the specified ForeignKeyMetaData.
     * Typically used for foreign-keys specified at &lt;class&gt; level.
     * @param fkmd ForeignKey MetaData
     * @return The ForeignKey
     */
    protected ForeignKey getForeignKeyForForeignKeyMetaData(ForeignKeyMetaData fkmd)
    {
        if (fkmd == null)
        {
            return null;
        }

        ForeignKey fk = new ForeignKey(fkmd.isDeferred());
        fk.setForMetaData(fkmd);

        // Set the column(s) to use in the FK
        // Class-level foreign-key so use its column definition
        ColumnMetaData[] colmds = fkmd.getColumnMetaData();
        AbstractPropertyMetaData[] fmds = fkmd.getFieldMetaData();

        // a). Columns specified directly
        if (colmds != null && colmds.length > 0)
        {
            // FK specified via <column>
            for (int i=0;i<colmds.length;i++)
            {
                // TODO [CORE-2584] Implement this - presumably we want to find the PK cols of "table"
            }
        }
        else if (fmds != null && fmds.length > 0)
        {
            // FK specified via <field>
            for (int i=0;i<fmds.length;i++)
            {
                // Find the metadata for the actual field with the same name as this "foreign-key" field
                AbstractPropertyMetaData realFmd = getFieldMetaData(fmds[i].getName());
                JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(realFmd);
                int countFields = fieldMapping.getNumberOfDatastoreFields();
                for (int j=0; j<countFields; j++)
                {
                    // Add each column of this field to the FK definition
                    // TODO [CORE-2584] Implement this - presumably we want to find the PK cols of "table"
                }
            }
        }

        return fk;
    }

    /**
     * Accessor for the expected candidate keys for this table.
     * @return The expected candidate keys.
     **/
    protected List getExpectedCandidateKeys()
    {
        assertIsInitialized();

        // The candidate keys required by the basic table
        List candidateKeys = super.getExpectedCandidateKeys();

        // Add on any user-required candidate keys for the fields
        Set fieldNumbersSet = fieldMappingsMap.keySet();
        Iterator iter = fieldNumbersSet.iterator();
        while (iter.hasNext())
        {
            AbstractPropertyMetaData fmd=(AbstractPropertyMetaData) iter.next();
            JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmd);
            if (fieldMapping instanceof EmbeddedPCMapping)
            {
                // Add indexes for fields of this embedded PC object
                EmbeddedPCMapping embMapping = (EmbeddedPCMapping)fieldMapping;
                for (int i=0;i<embMapping.getNumberOfJavaTypeMappings();i++)
                {
                    JavaTypeMapping embFieldMapping = embMapping.getJavaTypeMapping(i);
                    UniqueMetaData umd = embFieldMapping.getFieldMetaData().getUniqueMetaData();
                    if (umd != null)
                    {
                        CandidateKey ck = TableUtils.getCandidateKeyForField(this, umd, embFieldMapping);
                        if (ck != null)
                        {
                            candidateKeys.add(ck);
                        }
                    }
                }
            }
            else
            {
                // Add any required candidate key for this field
                UniqueMetaData umd = fmd.getUniqueMetaData();
                if (umd != null)
                {
                    CandidateKey ck = TableUtils.getCandidateKeyForField(this, umd, fieldMapping);
                    if (ck != null)
                    {
                        candidateKeys.add(ck);
                    }
                }
            }
        }

        // Add on any user-required candidate keys for the class(es) as a whole (composite keys)
        Iterator cmdIter = managedClassMetaData.iterator();
        while (cmdIter.hasNext())
        {
            ClassMetaData thisCmd = (ClassMetaData)cmdIter.next();
            UniqueMetaData[] classCKs = thisCmd.getUniqueMetaData();
            if (classCKs != null)
            {
                for (int i=0;i<classCKs.length;i++)
                {
                    CandidateKey ck = getCandidateKeyForUniqueMetaData(classCKs[i]);
                    if (ck != null)
                    {
                        candidateKeys.add(ck);
                    }
                }
            }
        }

        return candidateKeys;
    }

    /**
     * Convenience method to convert a UniqueMetaData into a CandidateKey.
     * @param umd The Unique MetaData
     * @return The Candidate Key
     */
    private CandidateKey getCandidateKeyForUniqueMetaData(UniqueMetaData umd)
    {
        CandidateKey ck = new CandidateKey(this);

        // Set the key name if required
        if (umd.getName() != null)
        {
            ck.setName(umd.getName());
        }

        // Class-level index so use its column definition
        ColumnMetaData[] colmds = umd.getColumnMetaData();
        AbstractPropertyMetaData[] fmds = umd.getFieldMetaData();
        // a). Columns specified directly
        if (colmds != null && colmds.length > 0)
        {
            for (int i=0;i<colmds.length;i++)
            {
                DatastoreIdentifier colName = storeMgr.getIdentifierFactory().newDatastoreFieldIdentifier(colmds[i].getName());
                Column col = (Column)columnsByName.get(colName);
                if (col == null)
                {
                    JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMS.Table.CandidateKey.ColumnNotFound", 
                        toString(), ck.getName(), colmds[i].getName()));
                    break;
                }
                else
                {
                    ck.addDatastoreField(col);
                }
            }
        }
        // b). Columns specified using fields
        else if (fmds != null && fmds.length > 0)
        {
            for (int i=0;i<fmds.length;i++)
            {
                // Find the metadata for the actual field with the same name as this "unique" field
                AbstractPropertyMetaData realFmd = getFieldMetaData(fmds[i].getName());
                JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(realFmd);
                int countFields = fieldMapping.getNumberOfDatastoreFields();
                for (int j=0; j<countFields; j++)
                {
                    ck.addDatastoreField(fieldMapping.getDataStoreMapping(j).getDatastoreField());
                }
            }
        }
        else
        {
            // We can't have an index of no columns
            JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMS.Table.CandidateKey.NoColumnsSpecified", 
                toString(), ck.getName()));
            return null;
        }

        return ck;
    }

    /**
     * Accessor for the primary key for this table. Overrides the method in TableImpl
     * to add on any specification of PK name in the metadata.
     * @return The primary key.
     */
    public PrimaryKey getPrimaryKey()
    {
        PrimaryKey pk = super.getPrimaryKey();
        PrimaryKeyMetaData pkmd = cmd.getPrimaryKeyMetaData();
        if (pkmd != null && pkmd.getName() != null)
        {
            pk.setName(pkmd.getName());
        }

        return pk;
    }

    /**
     * Accessor for the CREATE statements for this table.
     * Creates this table followed by all secondary tables (if any).
     * @param props Properties for creating the table
     * @return the SQL statements to be executed for creation
     */
    protected List getSQLCreateStatements(Properties props)
    {
        List stmts;

        // Create statements for this table
        Properties tableProps = null;
        if (createStatementDDL != null)
        {
            // User has specified the DDL
            stmts = new ArrayList();
            StringTokenizer tokens = new StringTokenizer(createStatementDDL, ";");
            while (tokens.hasMoreTokens())
            {
                stmts.add(tokens.nextToken());
            }
        }
        else
        {
            // JPOX creates the DDL
            if (cmd.getExtensions() != null)
            {
                tableProps = new Properties();
                ExtensionMetaData[] emds = cmd.getExtensions();
                for (int i=0;i<emds.length;i++)
                {
                    if (emds[i].getVendorName().equalsIgnoreCase(ExtendableMetaData.JPOX_VENDOR_NAME))
                    {
                        tableProps.put(emds[i].getKey(), emds[i].getValue());
                    }
                }
            }
            stmts = super.getSQLCreateStatements(tableProps);
        }

        // Create statements for any secondary tables
        // Since the secondary tables are managed by us, we need to create their table
        if (secondaryTables != null)
        {
            Set secondaryTableNames = secondaryTables.keySet();
            Iterator iter = secondaryTableNames.iterator();
            while (iter.hasNext())
            {
                SecondaryTable secTable = (SecondaryTable)secondaryTables.get(iter.next());
                stmts.addAll(secTable.getSQLCreateStatements(tableProps)); // Use same tableProps as the primary table
            }
        }

        // Unique constraints creation
        stmts.addAll(getSQLAddUniqueConstraintsStatements());

        return stmts;
    }

    /**
     * Accessor for the DROP statements for this table.
     * Drops all secondary tables (if any) followed by the table itself.
     * @return List of statements
     */
    protected List getSQLDropStatements()
    {
        assertIsInitialized();
        ArrayList stmts = new ArrayList();

        // Drop any secondary tables
        if (secondaryTables != null)
        {
            Set secondaryTableNames = secondaryTables.keySet();
            Iterator iter = secondaryTableNames.iterator();
            while (iter.hasNext())
            {
                SecondaryTable secTable = (SecondaryTable)secondaryTables.get(iter.next());
                stmts.addAll(secTable.getSQLDropStatements());
            }
        }

        // Drop this table
        stmts.add(dba.getDropTableStatement(this));

        return stmts;
    }

    private List getSQLAddUniqueConstraintsStatements()
    {
        ArrayList stmts = new ArrayList();
        Iterator cks = candidateKeysByMapField.values().iterator();
        int ckNum = 0;

        RDBMSIdentifierFactory idFactory = (RDBMSIdentifierFactory)storeMgr.getIdentifierFactory();
        while (cks.hasNext())
        {
            DatastoreIdentifier ckName = idFactory.newCandidateKeyIdentifier(this, ++ckNum);
            CandidateKey ck = (CandidateKey)cks.next();
            ck.setName(ckName.getIdentifier());

            stmts.add(dba.getAddCandidateKeyStatement(ck, idFactory));
        }

        return stmts;
    }

    /**
     * Method to initialise unique constraints for 1-N Map using FK.
     * @param ownerfmd FieldMetaData for the field with the map in the owner class
     */
    private void initializeFKMapUniqueConstraints(AbstractPropertyMetaData ownerfmd)
    {
        // Load "mapped-by"
        AbstractPropertyMetaData mfmd = null;
        String map_field_name = ownerfmd.getMappedBy(); // Field in our class that maps back to the owner class
        if (map_field_name != null)
        {
            // Bidirectional
            mfmd = cmd.getField(map_field_name);
            if (mfmd == null)
            {
                // Field not in primary class so may be in subclass so check all managed classes
                Iterator cmdIter = managedClassMetaData.iterator();
                while (cmdIter.hasNext())
                {
                    ClassMetaData managedCmd = (ClassMetaData)cmdIter.next();
                    mfmd = managedCmd.getField(map_field_name);
                    if (mfmd != null)
                    {
                        break;
                    }
                }
            }
            if (mfmd == null)
            {
                // "mapped-by" refers to a field in our class that doesnt exist!
                throw new JDOUserException(LOCALISER.msg("RDBMS.Table.RelationshipMappedByFieldNotFound",
                    map_field_name, cmd.getFullClassName(), ownerfmd.getFullFieldName()));                       
            }

            if (mfmd != null)
            {
                if (ownerfmd.getJoinMetaData() == null)
                {
                    // Load field of key in value
                    if (ownerfmd.getKeyMetaData() != null && ownerfmd.getKeyMetaData().getMappedBy() != null)
                    {
                        // Key field is stored in the value table
                        AbstractPropertyMetaData kmd = null;
                        String key_field_name = ownerfmd.getKeyMetaData().getMappedBy();
                        if (key_field_name != null)
                        {
                            kmd = cmd.getField(key_field_name);
                        }
                        if (kmd == null)
                        {
                            // Field not in primary class so may be in subclass so check all managed classes
                            Iterator cmdIter = managedClassMetaData.iterator();
                            while (cmdIter.hasNext())
                            {
                                ClassMetaData managedCmd = (ClassMetaData)cmdIter.next();
                                kmd = managedCmd.getField(key_field_name);
                                if (kmd != null)
                                {
                                    break;
                                }
                            }
                        }
                        if (kmd == null)
                        {
                            throw new ClassDefinitionException(LOCALISER.msg("RDBMS.Table.InvalidMapKeyField", 
                                mfmd.getFullFieldName(), key_field_name));
                        }

                        JavaTypeMapping ownerMapping = getFieldMapping(map_field_name);
                        JavaTypeMapping keyMapping = getFieldMapping(kmd.getName());

                        if (dba.supportsNullsInCandidateKeys() || (!ownerMapping.isNullable() && !keyMapping.isNullable()))
                        {
                            // If the owner and key fields are represented in this table then we can impose
                            // a unique constraint on them. If the key field is in a superclass then we
                            // cannot do this so just omit it.
                            if (keyMapping.getDatastoreContainer() == this &&
                                ownerMapping.getDatastoreContainer() == this)
                            {
                                CandidateKey ck = new CandidateKey(this);

                                // This HashSet is to avoid duplicate adding of columns.
                                HashSet addedColumns = new HashSet();

                                // Add columns for the owner field
                                int countOwnerFields = ownerMapping.getNumberOfDatastoreFields();
                                for (int i = 0; i < countOwnerFields; i++)
                                {
                                    Column col = (Column) ownerMapping.getDataStoreMapping(i).getDatastoreField();
                                    addedColumns.add(col);
                                    ck.addDatastoreField(col);
                                }

                                // Add columns for the key field
                                int countKeyFields = keyMapping.getNumberOfDatastoreFields();
                                for (int i = 0; i < countKeyFields; i++)
                                {
                                    Column col = (Column) keyMapping.getDataStoreMapping(i).getDatastoreField();
                                    if (!addedColumns.contains(col))
                                    {
                                        addedColumns.add(col);
                                        ck.addDatastoreField(col);
                                    }
                                    else
                                    {
                                        JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMS.Table.KeyTypeMapColumnOverwriteValueTypeMapColumn", 
                                            ownerfmd.getName()));
                                    }
                                }

                                if (candidateKeysByMapField.put(mfmd, ck) != null)
                                {
                                    // We have multiple "mapped-by" coming to this field so give a warning that this may potentially
                                    // cause problems. For example if they have the key field defined here for 2 different relations
                                    // so you may get keys/values appearing in the other relation that shouldn't be.
                                    // Logged as a WARNING for now. 
                                    // If there is a situation where this should throw an exception, please update this AND COMMENT WHY.
                                    JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMS.Table.DuplicateMapOwnerField", 
                                        mfmd.getFullFieldName(), ownerfmd.getFullFieldName()));
                                }
                            }
                        }
                    }
                    else if (ownerfmd.getValueMetaData() != null && ownerfmd.getValueMetaData().getMappedBy() != null)
                    {
                        // Value field is stored in the key table
                        AbstractPropertyMetaData vmd = null;
                        String value_field_name = ownerfmd.getValueMetaData().getMappedBy();
                        if (value_field_name != null)
                        {
                            vmd = cmd.getField(value_field_name);
                        }
                        if (vmd == null)
                        {
                            throw new ClassDefinitionException(LOCALISER.msg("RDBMS.Table.InvalidMapValueField", mfmd));
                        }

                        JavaTypeMapping ownerMapping = getFieldMapping(map_field_name);
                        JavaTypeMapping valueMapping = getFieldMapping(vmd.getName());

                        if (dba.supportsNullsInCandidateKeys() || (!ownerMapping.isNullable() && !valueMapping.isNullable()))
                        {
                            // If the owner and value fields are represented in this table then we can impose
                            // a unique constraint on them. If the value field is in a superclass then we
                            // cannot do this so just omit it.
                            if (valueMapping.getDatastoreContainer() == this &&
                                ownerMapping.getDatastoreContainer() == this)
                            {
                                CandidateKey ck = new CandidateKey(this);

                                // This HashSet is to avoid duplicate adding of columns.
                                HashSet addedColumns = new HashSet();

                                // Add columns for the owner field
                                int countOwnerFields = ownerMapping.getNumberOfDatastoreFields();
                                for (int i = 0; i < countOwnerFields; i++)
                                {
                                    Column col = (Column) ownerMapping.getDataStoreMapping(i).getDatastoreField();
                                    addedColumns.add(col);
                                    ck.addDatastoreField(col);
                                }

                                // Add columns for the value field
                                int countValueFields = valueMapping.getNumberOfDatastoreFields();
                                for (int i = 0; i < countValueFields; i++)
                                {
                                    Column col = (Column) valueMapping.getDataStoreMapping(i).getDatastoreField();
                                    if (!addedColumns.contains(col))
                                    {
                                        addedColumns.add(col);
                                        ck.addDatastoreField(col);
                                    }
                                    else
                                    {
                                        JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMS.Table.ValueTypeMapColumnOverwriteKeyTypeMapColumn", 
                                            ownerfmd.getName()));
                                    }
                                }

                                if (candidateKeysByMapField.put(mfmd, ck) != null)
                                {
                                    // We have multiple "mapped-by" coming to this field so give a warning that this may potentially
                                    // cause problems. For example if they have the key field defined here for 2 different relations
                                    // so you may get keys/values appearing in the other relation that shouldn't be.
                                    // Logged as a WARNING for now. 
                                    // If there is a situation where this should throw an exception, please update this AND COMMENT WHY.
                                    JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMS.Table.DuplicateMapOwnerField",
                                        mfmd.getFullFieldName(), ownerfmd.getFullFieldName()));
                                }
                            }
                        }
                    }
                    else
                    {
                        // We can only have either the key stored in the value or the value stored in the key but nothing else!
                        throw new ClassDefinitionException(LOCALISER.msg("RDBMS.Table.MissingMapKeyValueField", ownerfmd.getFullFieldName()));
                    }
                }
            }
        }
    }

    /**
     * Initialize the ID Mapping
     */
    private void initializeIDMapping()
    {
        if (idMapping != null)
        {
            return;
        }

        final PersistenceCapableMapping mapping = new PersistenceCapableMapping();
        mapping.initialize(getStoreManager().getDatastoreAdapter(), cmd.getFullClassName());
        if (getIdentityType() == IdentityType.DATASTORE)
        {
            mapping.addJavaTypeMapping(datastoreIDMapping);
        }
        else if (getIdentityType() == IdentityType.APPLICATION)
        {
            for (int i = 0; i < pkMappings.length; i++)
            {
                mapping.addJavaTypeMapping(pkMappings[i]);
            }
        }
        else
        {
            // Nothing to do for nondurable since no identity
        }

        idMapping = mapping;        
    }
    
    /**
     * Accessor for a mapping for the ID (PersistenceCapable) for this table.
     * @return The (PersistenceCapable) ID mapping.
     **/
    public JavaTypeMapping getIDMapping()
    {
        return idMapping;
    }

    /**
     * Accessor for all of the order mappings (used by FK Lists, Collections, Arrays)
     * @return The mappings for the order columns.
     **/
    private HashMap getExternalOrderMappings()
    {
        if (externalOrderMappings == null)
        {       
            externalOrderMappings = new HashMap();
        }
        return externalOrderMappings;
    }

    /**
     * Accessor for all of the external FK mappings.
     * @return The mappings for external FKs
     **/
    private HashMap getExternalFkMappings()
    {
        if (externalFkMappings == null)
        {       
            externalFkMappings = new HashMap();
        }
        return externalFkMappings;
    }

    /**
     * Accessor for an external mapping for the specified field of the required type.
     * @param fmd MetaData for the field
     * @param mappingType Type of mapping
     * @return The (external) mapping
     */
    public JavaTypeMapping getExternalMapping(AbstractPropertyMetaData fmd, int mappingType)
    {
        if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_FK)
        {
            return (JavaTypeMapping) getExternalFkMappings().get(fmd);
        }
        else if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_FK_DISCRIM)
        {
            return (JavaTypeMapping) getExternalFkDiscriminatorMappings().get(fmd);
        }
        else if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_INDEX)
        {
            return (JavaTypeMapping) getExternalOrderMappings().get(fmd);
        }
        else
        {
            return null;
        }
    }

    /**
     * Accessor for the MetaData for the (owner) field that an external mapping corresponds to.
     * @param mapping The mapping
     * @param mappingType The mapping type
     * @return The fieldMetaData
     */
    public AbstractPropertyMetaData getMetaDataForExternalMapping(JavaTypeMapping mapping, int mappingType)
    {
        if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_FK)
        {
            Set entries = getExternalFkMappings().entrySet();
            Iterator iter = entries.iterator();
            while (iter.hasNext())
            {
                Map.Entry entry = (Map.Entry)iter.next();
                if (entry.getValue() == mapping)
                {
                    return (AbstractPropertyMetaData)entry.getKey();
                }
            }
        }
        else if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_FK_DISCRIM)
        {
            Set entries = getExternalFkDiscriminatorMappings().entrySet();
            Iterator iter = entries.iterator();
            while (iter.hasNext())
            {
                Map.Entry entry = (Map.Entry)iter.next();
                if (entry.getValue() == mapping)
                {
                    return (AbstractPropertyMetaData)entry.getKey();
                }
            }
        }
        else if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_INDEX)
        {
            Set entries = getExternalOrderMappings().entrySet();
            Iterator iter = entries.iterator();
            while (iter.hasNext())
            {
                Map.Entry entry = (Map.Entry)iter.next();
                if (entry.getValue() == mapping)
                {
                    return (AbstractPropertyMetaData)entry.getKey();
                }
            }
        }
        return null;
    }

    /**
     * Accessor for all of the external FK discriminator mappings.
     * @return The mappings for external FKs
     **/
    private HashMap getExternalFkDiscriminatorMappings()
    {
        if (externalFkDiscriminatorMappings == null)
        {       
            externalFkDiscriminatorMappings = new HashMap();
        }
        return externalFkDiscriminatorMappings;
    }

    /**
     * Accessor for the field mapping for the specified field.
     * The field can be managed by a supertable of this table.
     * @param fmd Field MetaData for this field
     * @return the Mapping for the field
     */
    public JavaTypeMapping getFieldMapping(AbstractPropertyMetaData fmd)
    {
        if (fmd == null)
        {
            return null;
        }

        if (fmd instanceof PropertyMetaData)
        {
            // "persistent interfaces" cant do proper lookups here presumably something to do with having PropertyMetaData in one
            // place and FieldMetaData in others so just lookup on the field name
            return getFieldMapping(fmd.getName());
        }

        if (fmd.isPrimaryKey())
        {
            assertIsPKInitialized();
        }
        else
        {
            assertIsInitialized();
        }

        // Check if we manage this field
        JavaTypeMapping m = (JavaTypeMapping)fieldMappingsMap.get(fmd);
        if (m != null)
        {
            return m;
        }

        // Check supertable
        int ifc = cmd.getNoOfInheritedManagedFields();
        if (fmd.getAbsoluteFieldNumber() < ifc)
        {
            if (supertable != null)
            {
                m = supertable.getFieldMapping(fmd);
                if (m != null)
                {
                    return m;
                }
            }
        }

        // Check secondary tables
        if (secondaryTables != null)
        {
            Collection secTables = secondaryTables.values();
            Iterator iter = secTables.iterator();
            while (iter.hasNext())
            {
                SecondaryTable secTable = (SecondaryTable)iter.next();
                m = secTable.getFieldMapping(fmd);
                if (m != null)
                {
                    return m;
                }
            }
        }

        return null;
    }

    /**
     * Accessor for the field mapping for the named field.
     * The field may exist in a parent table or a secondary table.
     * Throws a NoSuchPersistentFieldException if the field name is not found.
     * TODO Use of this is discouraged since the fieldName is not fully qualified
     * and if a superclass-table inheritance is used we could have 2 fields of that name here.
     * @param fieldName Name of field
     * @return The field mapping.
     * @throws NoSuchPersistentFieldException Thrown when the field is not found
     */
    public JavaTypeMapping getFieldMapping(String fieldName)
    {
        assertIsInitialized();
        AbstractPropertyMetaData fmd = getFieldMetaData(fieldName);
        JavaTypeMapping m = getFieldMapping(fmd);
        if (m == null)
        {
            throw new NoSuchPersistentFieldException(cmd.getFullClassName(), fieldName);
        }
        return m;
    }

    /**
     * Acessor for the FieldMetaData for the field with the specified name.
     * Searches the MetaData of all classes managed by this table.
     * Doesn't allow for cases where the table manages subclasses with the same field name.
     * In that case you should use the equivalent method passing FieldMetaData.
     * TODO Support subclasses with fields of the same name
     * TODO Use of this is discouraged since the fieldName is not fully qualified
     * and if a superclass-table inheritance is used we could have 2 fields of that name here.
     * @param fieldName the field name
     * @return the FieldMetaData
     */
    AbstractPropertyMetaData getFieldMetaData(String fieldName)
    {
        // TODO Dont search "cmd" since it is included in "managedClassMetaData"
        AbstractPropertyMetaData fmd = cmd.getField(fieldName);
        if (fmd == null)
        {
            Iterator iter = managedClassMetaData.iterator();
            while (iter.hasNext())
            {
                AbstractClassMetaData theCmd = (AbstractClassMetaData) iter.next();
                final AbstractPropertyMetaData foundFmd = theCmd.getField(fieldName);
                if (foundFmd != null)
                {
                    if (fmd != null && (!fmd.toString().equalsIgnoreCase(foundFmd.toString()) || fmd.getType() != foundFmd.getType()))
                    {
                        final String errMsg = "Table " + getIdentifier() + 
                        " manages at least 2 subclasses that both define a field \"" + fieldName + "\", " + 
                        "and the fields' metadata is different or they have different type! That means you can get e.g. wrong fetch results.";
                        JPOXLogger.RDBMS.error(errMsg);
                        throw new JDOFatalInternalException(errMsg);
                    }
                    fmd = foundFmd;
                }
            }
        }
        return fmd;
    }

    /**
     * Utility to throw an exception if the object is not PersistenceCapable.
     * @param sm The StateManager for the object
     **/ 
    void assertPCClass(StateManager sm)
    {
        Class c = sm.getObject().getClass();

        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(cmd.getFullClassName(),c))
        {
            throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.Table.ClassObjectInfo",cmd.getFullClassName(),c));
        }
    }

    /**
     * Method to fetch an insert into this table. Will insert any superclass
     * records as well (recursive).
     * @param sm StateManager for the object to insert.
     **/
    public void insert(StateManager sm)
    {
        assertIsInitialized();
        assertPCClass(sm);

        if (supertable != null)
        {
            supertable.insert(sm);
        }

        InsertRequest req = storeMgr.getInsertRequest(this, sm.getObject().getClass(), sm.getPersistenceManager().getClassLoaderResolver());
        req.execute(sm);

        // Insert any secondary tables
        if (secondaryTables != null)
        {
            Collection tables = secondaryTables.values();
            Iterator tablesIter = tables.iterator();
            while (tablesIter.hasNext())
            {
                SecondaryTable table = (SecondaryTable)tablesIter.next();
                table.insert(sm);
            }
        }
    }

    /**
     * Method to fetch an object in this table. 
     * Will fetch any superclass records as well (recursive).
     * @param sm StateManager for the object to fetch.
     * @param fieldMetaData MetaDat afor the fields to be fetched
     */
    public void fetch(StateManager sm, AbstractPropertyMetaData[] fieldMetaData)
    {
        assertIsInitialized();
        assertPCClass(sm);

        if (supertable != null)
        {
            supertable.fetch(sm, fieldMetaData);
        }

        FetchRequest req = storeMgr.getFetchRequest(this, fieldMetaData, sm.getObject().getClass(), 
            sm.getPersistenceManager().getClassLoaderResolver());
        req.execute(sm);
    }

    /**
     * Method to update an object in this table. Will update any superclass
     * records as well (recursive).
     * @param sm StateManager for the object to update.
     * @param fieldMetaData MetaData for the fields to be updated
     */
    public void update(StateManager sm, AbstractPropertyMetaData[] fieldMetaData)
    {
        assertIsInitialized();
        assertPCClass(sm);

        if (supertable != null)
        {
            supertable.update(sm, fieldMetaData);
        }

        UpdateRequest req = storeMgr.getUpdateRequest(this, fieldMetaData, sm.getObject().getClass(), 
            sm.getPersistenceManager().getClassLoaderResolver());
        req.execute(sm);

        // Update any secondary tables
        if (secondaryTables != null)
        {
            Collection tables = secondaryTables.values();
            Iterator tablesIter = tables.iterator();
            while (tablesIter.hasNext())
            {
                SecondaryTable table = (SecondaryTable)tablesIter.next();
                table.update(sm, fieldMetaData);
            }
        }
    }

    /**
     * Method to delete an object from this table. Will delete any superclass
     * records as well (recursive).
     * @param sm StateManager for the object to delete.
     **/
    public void delete(StateManager sm)
    {
        assertIsInitialized();
        assertPCClass(sm);

        // Delete any secondary tables
        if (secondaryTables != null)
        {
            Collection tables = secondaryTables.values();
            Iterator tablesIter = tables.iterator();
            while (tablesIter.hasNext())
            {
                SecondaryTable table = (SecondaryTable)tablesIter.next();
                table.delete(sm);
            }
        }

        DeleteRequest req = storeMgr.getDeleteRequest(this, sm.getObject().getClass(), sm.getPersistenceManager().getClassLoaderResolver());
        req.execute(sm);

        if (supertable != null)
        {
            supertable.delete(sm);
        }
    }

    /**
     * Locates this object in the datastore.
     * Returns normally if it is present, or throws an exception if not found.
     * @param sm The StateManager for the object to be found 
     * @throws javax.jdo.JDOObjectNotFoundException If the instance does not exist in
     * the datastore
     */
    public void locate(StateManager sm)
    {
        assertIsInitialized();
        assertPCClass(sm);

        LocateRequest req = storeMgr.getLocateRequest(this, sm.getObject().getClass().getName());
        req.execute(sm);
    }

    /**
     * Delete dependent objects of a persistent object.
     * @param sm StateManager for the object being deleted.
     * @param fieldMetaData MetaData for the fields to delete.
     * @since 1.1
     */
    public void deleteDependent(StateManager sm, AbstractPropertyMetaData[] fieldMetaData)
    {
        assertIsInitialized();
        assertPCClass(sm);
 
        for (int i = 0; i < fieldMetaData.length; ++i)
        {
            JavaTypeMapping m = getFieldMapping(fieldMetaData[i]);
            if (m instanceof MappingCallbacks)
            {
               ((MappingCallbacks)m).deleteDependent(sm);
            }
            else if (m instanceof PersistenceCapableMapping && fieldMetaData[i].isDependent())
            {
                ((PersistenceCapableMapping)m).deleteDependent(sm);
            }
            else if (m instanceof ReferenceMapping && fieldMetaData[i].isDependent())
            {
                ((ReferenceMapping)m).deleteDependent(sm);
            }
        }
    }

    /**
     * Adds an ordering column to the element table (this) in inverse list relationships.
     * Used to store the position of the element in the List.
     * If the &lt;order&gt; provides a mapped-by, this will return the existing column mapping.
     * @param fmd The AbstractPropertyMetaData for the column to map to
     * @return The Mapping for the order column
     */
    private JavaTypeMapping addOrderColumn(AbstractPropertyMetaData fmd, ClassLoaderResolver clr)
    {
        Class indexType = Integer.class;
        JavaTypeMapping indexMapping = new IndexMapping();
        indexMapping.initialize(dba, indexType.getName());
        IdentifierFactory idFactory = storeMgr.getIdentifierFactory();
        DatastoreIdentifier indexColumnName = null;
        ColumnMetaData colmd = null;

        // Allow for any user definition in OrderMetaData
        OrderMetaData omd = fmd.getOrderMetaData();
        if (omd != null)
        {
            colmd = (omd.getColumnMetaData() != null && omd.getColumnMetaData().length > 0 ? omd.getColumnMetaData()[0] : null);
            if (omd.getMappedBy() != null)
            {
                // User has defined ordering using the column(s) of an existing field.
                state = TABLE_STATE_INITIALIZED; // Not adding anything so just set table back to "initialised"
                JavaTypeMapping orderMapping = getFieldMapping(omd.getMappedBy());
                if (orderMapping == null)
                {
                    throw new JDOUserException(LOCALISER.msg("RDBMS.Table.OrderMappingMappedByFieldNotFound", 
                        fmd.getFullFieldName(), omd.getMappedBy()));
                }
                if (!(orderMapping instanceof IntegerMapping) && !(orderMapping instanceof LongMapping))
                {
                    throw new JDOUserException(LOCALISER.msg("RDBMS.Table.OrderMappingFieldInvalid", 
                        fmd.getFullFieldName(), omd.getMappedBy()));
                }
                return orderMapping;
            }

            String colName = null;
            if (omd.getColumnMetaData() != null && omd.getColumnMetaData().length > 0 && omd.getColumnMetaData()[0].getName() != null)
            {
                // User-defined name so create an identifier using it
                colName = omd.getColumnMetaData()[0].getName();
                indexColumnName = idFactory.newDatastoreFieldIdentifier(colName);
            }
        }
        if (indexColumnName == null)
        {
            // No name defined so generate one
            indexColumnName = ((RDBMSIdentifierFactory)idFactory).newForeignKeyFieldIdentifier(fmd, null, null, 
                storeMgr.getPMFContext().getTypeManager().isDefaultEmbeddedType(indexType), DatastoreField.ROLE_INDEX);
        }

        DatastoreField column = addDatastoreField(indexType.getName(), indexColumnName, indexMapping, colmd);
        if (colmd == null || (colmd != null && !colmd.isAllowsNullSet()) ||
            (colmd != null && colmd.isAllowsNullSet() && colmd.isAllowsNull()))
        {
            // User either wants it nullable, or havent specified anything, so make it nullable
            column.setNullable();
        }

        storeMgr.getMappingManager().createDatastoreMapping(indexMapping, storeMgr, column, indexType.getName());

        return indexMapping;
    }

    /**
     * Callbacks is used for inverse relationships to run some operation
     * in the target table. The operation is creation of columns, indexes, or whatever needed. 
     */
    private static class CallBack
    {
        final AbstractPropertyMetaData fmd;

        /**
         * Default constructor
         * @param fmd The FieldMetaData
         */
        public CallBack(AbstractPropertyMetaData fmd)
        {
            this.fmd = fmd;
        }
    }
    
    /**
     * Provide the mappings to the consumer for all primary-key fields mapped to
     * this table.
     * @param consumer Consumer for the mappings
     */
    public void providePrimaryKeyMappings(MappingConsumer consumer)
    {
        consumer.preConsumeMapping(highestFieldNumber + 1);

        if (pkMappings != null)
        {
            // Application identity
            int[] primaryKeyFieldNumbers = cmd.getPrimaryKeyFieldNumbers();
            for (int i=0;i<pkMappings.length;i++)
            {
                // Make the assumption that the pkMappings are in the same order as the absolute field numbers
                AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(primaryKeyFieldNumbers[i]);
                consumer.consumeMapping(pkMappings[i], fmd);
            }
        }
        else
        {
            // Datastore identity
            int[] primaryKeyFieldNumbers = cmd.getPrimaryKeyFieldNumbers();
            int countPkFields = cmd.getNoOfPrimaryKeyFields();
            for (int i = 0; i < countPkFields; i++)
            {
                AbstractPropertyMetaData pkfmd = cmd.getManagedFieldAbsolute(primaryKeyFieldNumbers[i]);
                consumer.consumeMapping(getFieldMapping(pkfmd), pkfmd);
            }
        }
    }

    /**
     * Provide the mappings to the consumer for all external fields mapped to this table
     * of the specified type
     * @param consumer Consumer for the mappings
     * @param mappingType Type of external mapping
     */
    final public void provideExternalMappings(MappingConsumer consumer, int mappingType)
    {
        if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_FK && externalFkMappings != null)
        {
            consumer.preConsumeMapping(highestFieldNumber + 1);
            
            Set fieldNumbersSet = externalFkMappings.keySet();
            Iterator iter = fieldNumbersSet.iterator();
            while (iter.hasNext())
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData) iter.next();
                JavaTypeMapping fieldMapping = (JavaTypeMapping)externalFkMappings.get(fmd);
                if (fieldMapping != null)
                {
                    consumer.consumeMapping(fieldMapping, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK);
                }
            }
        }
        else if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_FK_DISCRIM && externalFkDiscriminatorMappings != null)
        {
            consumer.preConsumeMapping(highestFieldNumber + 1);
            
            Set fieldNumbersSet = externalFkDiscriminatorMappings.keySet();
            Iterator iter = fieldNumbersSet.iterator();
            while (iter.hasNext())
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData) iter.next();
                JavaTypeMapping fieldMapping = (JavaTypeMapping)externalFkDiscriminatorMappings.get(fmd);
                if (fieldMapping != null)
                {
                    consumer.consumeMapping(fieldMapping, MappingConsumer.MAPPING_TYPE_EXTERNAL_FK_DISCRIM);
                }
            }
        }
        else if (mappingType == MappingConsumer.MAPPING_TYPE_EXTERNAL_INDEX && externalOrderMappings != null)
        {
            consumer.preConsumeMapping(highestFieldNumber + 1);
            
            Set fieldNumbersSet = externalOrderMappings.keySet();
            Iterator iter = fieldNumbersSet.iterator();
            while (iter.hasNext())
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData) iter.next();
                JavaTypeMapping fieldMapping = (JavaTypeMapping)externalOrderMappings.get(fmd);
                if (fieldMapping != null)
                {
                    consumer.consumeMapping(fieldMapping, MappingConsumer.MAPPING_TYPE_EXTERNAL_INDEX);
                }
            }
        }
    }

    /**
     * Provide the mappings to the consumer for all absolute field Numbers in this table
     * that are container in the fieldNumbers parameter.
     * @param consumer Consumer for the mappings
     * @param fieldMetaData MetaData for the fields to provide mappings for
     * @param includeSecondaryTables Whether to provide fields in secondary tables
     */
    public void provideMappingsForFields(MappingConsumer consumer, AbstractPropertyMetaData[] fieldMetaData, boolean includeSecondaryTables)
    {
        super.provideMappingsForFields(consumer, fieldMetaData, true);
        if (includeSecondaryTables && secondaryTables != null)
        {
            Collection secTables = secondaryTables.values();
            Iterator iter = secTables.iterator();
            while (iter.hasNext())
            {
                SecondaryTable secTable = (SecondaryTable)iter.next();
                secTable.provideMappingsForFields(consumer, fieldMetaData, false);
            }
        }
    }

    /**
     * Method to validate the constraints of this table.
     * @param conn Connection to use in validation
     * @param autoCreate Whether to auto create the constraints
     * @param autoCreateErrors Whether to log a warning only on errors during "auto create"
     * @param clr The ClassLoaderResolver
     * @return Whether the DB was modified
     * @throws SQLException Thrown when an error occurs in validation
     */
    public boolean validateConstraints(Connection conn, boolean autoCreate, Collection autoCreateErrors, ClassLoaderResolver clr)
    throws SQLException
    {
        boolean modified = false;
        if (super.validateConstraints(conn, autoCreate, autoCreateErrors, clr))
        {
            modified = true;
        }

        // Validate our secondary tables since we manage them
        if (secondaryTables != null)
        {
            Collection secTables = secondaryTables.values();
            Iterator iter = secTables.iterator();
            while (iter.hasNext())
            {
                SecondaryTable secTable = (SecondaryTable)iter.next();
                if (secTable.validateConstraints(conn, autoCreate, autoCreateErrors, clr))
                {
                    modified = true;
                }
            }
        }

        return modified;
    }
}