/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Andy Jefferson - replaced TableMetadata with SQLIdentifier and java name
2004 Marco Schulze - added advance-check via TypeManager.isSupportedType(...)
2005 Andy Jefferson - added support for embedded PC elements
2005 Andy Jefferson - added <primary-key> support
     ...
 **********************************************************************/
package org.jpox.store.rdbms.table;

import org.jpox.ClassLoaderResolver;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.PrimaryKeyMetaData;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.sqlidentifier.RDBMSIdentifierFactory;

/**
 * Representation of a join table for a List/array.
 * The join table is used in the following situations :-
 * <UL>
 * <LI><B>List&lt;Simple&gt;</B> - a list of simple elements.
 * For example a List of Strings. The elements are stored in their own
 * column in this join table. There is no separate element table</LI>
 * <LI><B>List&lt;PersistenceCapable&gt;</B> - a list of
 * PersistenceCapable elements with the elements stored in their own table.
 * A FK is added to this table to get to the element table</LI>
 * <LI><B>List&lt;PersistenceCapable&gt; (embedded)</B> - a list of
 * PersistenceCapable elements with the elements embedded in this table.
 * A surrogate key is added to form part of the (composite) PK.</LI>
 * <LI><B>Collection&lt;PersistenceCapable&gt; (serialised)</B> - a collection
 * of PersistenceCapable elements with the elements serialised into a single column
 * in this table. A surrogate key is added to form part of the (composite) PK.</LI>
 * </UL>
 * 
 * @version $Revision: 1.45 $
 */
public class ListTable extends CollectionTable
{
    /**
     * Constructor.
     * @param tableName Identifier name of the table
     * @param fmd MetaData for the field of the owner
     * @param storeMgr The Store Manager managing these tables.
     */
    public ListTable(DatastoreIdentifier tableName, 
                     AbstractPropertyMetaData fmd,
                     RDBMSManager storeMgr)
    {
        super(tableName, fmd, storeMgr);
    }

    /**
     * Method to initialise the table definition.
     * @param clr The ClassLoaderResolver
     */
    public void initialize(ClassLoaderResolver clr)
    {
        super.initialize(clr);

        PrimaryKeyMetaData pkmd = (fmd.getJoinMetaData() != null ? fmd.getJoinMetaData().getPrimaryKeyMetaData() : null);
        boolean pkColsSpecified = (pkmd != null ? pkmd.getColumnMetaData() != null : false);
        boolean pkRequired = requiresPrimaryKey();

        if (isSerialisedElementPC() || isEmbeddedElementPC())
        {
            // Element = PC(embedded), PC(serialised)

            // Add field(s) for element
            elementMapping = dba.getMappingManager().getMapping(this, fmd, dba, clr, MappingManager.MAPPING_COLLECTION_ELEMENT);
        }
        else
        {
            // Element = PC, Non-PC(embedded)

            // Create the element mapping, taking a serialised mapping if the element is embedded
            CollectionMetaData colmd = fmd.getCollection();
            elementMapping = ColumnCreator.createColumnsForJoinTables(clr.classForName(colmd.getElementType()), fmd,
                fmd.getElementMetaData() == null ? null : fmd.getElementMetaData().getColumnMetaData(), storeMgr, this,
                false, true, colmd.isSerializedElement(), false, DatastoreField.ROLE_COLLECTION_ELEMENT, clr, null, false);
        }

        // Add the list ordering index
        ColumnMetaData colmd = null;
        if (fmd.getOrderMetaData() != null && fmd.getOrderMetaData().getColumnMetaData() != null &&
            fmd.getOrderMetaData().getColumnMetaData().length > 0)
        {
            // Specified column name
            colmd = fmd.getOrderMetaData().getColumnMetaData()[0];
        }
        else
        {
            // No column name so use default
            DatastoreIdentifier id = ((RDBMSIdentifierFactory)storeMgr.getIdentifierFactory()).newIndexFieldIdentifier();
            colmd = new ColumnMetaData(fmd, id.getIdentifier());
        }
        orderMapping = dba.getMapping(int.class, storeMgr); // JDO2 spec [18.5] order column is assumed to be "int"
        ColumnCreator.createAdapterColumn(orderMapping, storeMgr, clr, this, colmd, pkRequired && !pkColsSpecified);

        if (pkRequired && pkColsSpecified)
        {
            // Apply the users PK specification
            applyUserPrimaryKeySpecification(pkmd);
        }

        state = TABLE_STATE_INITIALIZED;
    }
}