/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Andy Jefferson - added comments
2003 Andy Jefferson - replaced TableMetadata with identifier
2004 Erik Bengtson - reads the columnOptions from the elementtable to
                    create a column with the same type
2004 Marco Schulze - added advance-check via TypeManager.isSupportedType(...)
2005 Andy Jefferson - only create ADPT column when necessary
2005 Andy Jefferson - added embedded PC element capability
2005 Andy Jefferson - added <primary-key> support
    ...
 **********************************************************************/
package org.jpox.store.rdbms.table;

import org.jpox.ClassLoaderResolver;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.ElementMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.PrimaryKeyMetaData;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.rdbms.RDBMSManager;

/**
 * Representation of a join table for a Set/Collection.
 * The join table is used in the following situations :-
 * <UL>
 * <LI><B>Collection&lt;Simple&gt;</B> - a collection/set of simple elements.
 * For example a Collection of Strings. The elements are stored in their own
 * column in this join table. There is no separate element table</LI>
 * <LI><B>Collection&lt;PersistenceCapable&gt;</B> - a collection/set of
 * PersistenceCapable elements with the elements stored in their own table.
 * A FK is added to this table to get to the element table</LI>
 * <LI><B>Collection&lt;PersistenceCapable&gt; (embedded)</B> - a collection
 * of PersistenceCapable elements with the elements embedded in this table.
 * A surrogate key is added to form part of the (composite) PK.</LI>
 * <LI><B>Collection&lt;PersistenceCapable&gt; (serialised)</B> - a collection
 * of PersistenceCapable elements with the elements serialised into a single column
 * in this table. A surrogate key is added to form part of the (composite) PK.</LI>
 * </UL>
 * 
 * @version $Revision: 1.53 $
 */
public class SetTable extends CollectionTable
{
    /**
     * Constructor.
     * @param tableName Identifier name of the table
     * @param fmd MetaData for the field of the owner
     * @param storeMgr The Store Manager managing these tables.
     */
    public SetTable(DatastoreIdentifier tableName,
                    AbstractPropertyMetaData fmd,
                    RDBMSManager storeMgr)
    {
        super(tableName, fmd, storeMgr);
    }

    /**
     * Method to initialise the table definition.
     * @param clr The ClassLoaderResolver
     */
    public void initialize(ClassLoaderResolver clr)
    {
        super.initialize(clr);

        PrimaryKeyMetaData pkmd = (fmd.getJoinMetaData() != null ? fmd.getJoinMetaData().getPrimaryKeyMetaData() : null);
        boolean pkColsSpecified = (pkmd != null ? pkmd.getColumnMetaData() != null : false);
        boolean pkRequired = requiresPrimaryKey();

        if (isSerialisedElementPC() || isEmbeddedElementPC())
        {
            // Element = PC(embedded), PC(serialised)
            if (fmd.getOrderMetaData() != null)
            {
                ColumnMetaData colmd = null;
                if (fmd.getOrderMetaData().getColumnMetaData() != null)
                {
                    colmd = fmd.getOrderMetaData().getColumnMetaData()[0];
                }
                orderMapping = dba.getMapping(int.class, storeMgr); // JDO2 spec [18.5] order column is assumed to be "int"
                ColumnCreator.createAdapterColumn(orderMapping, storeMgr, clr, this, colmd, pkRequired);
            }
            else if (pkRequired && !pkColsSpecified &&
                !(isEmbeddedElementPC() && fmd.getCollection().getElementClassMetaData().getIdentityType() == IdentityType.APPLICATION))
            {
                // No available PK specification but we need one, so add adapter index to form part of PK since no other suitable field
                ColumnMetaData colmd = null;
                if (fmd.hasExtension("adapter-column-name"))
                {
                    colmd = new ColumnMetaData(fmd, fmd.getValueForExtension("adapter-column-name"));
                }
                orderMapping = dba.getMapping(int.class, storeMgr); // JDO2 spec [18.5] order column is assumed to be "int"
                ColumnCreator.createAdapterColumn(orderMapping, storeMgr, clr, this, colmd, pkRequired);
            }

            // Add field(s) for element
            elementMapping = dba.getMappingManager().getMapping(this, fmd, dba, clr, MappingManager.MAPPING_COLLECTION_ELEMENT);

            if (pkRequired && !pkColsSpecified && fmd.getOrderMetaData() == null &&
                isEmbeddedElementPC() && fmd.getCollection().getElementClassMetaData().getIdentityType() == IdentityType.APPLICATION)
            {
                // No user PK definition, and no order column and we have the element app id fields so assign them to be part of PK
                int numberOfKeyColumns = elementMapping.getNumberOfDatastoreFields();
                for (int i=0;i<numberOfKeyColumns;i++)
                {
                    if (elementMapping.getDataStoreMapping(i).getJavaTypeMapping().getFieldMetaData().isPrimaryKey());
                    {
                        elementMapping.getDataStoreMapping(i).getDatastoreField().setAsPrimaryKey();
                    }
                }
            }
        }
        else
        {
            // Element = PC, Non-PC(embedded)
            if (fmd.getOrderMetaData() != null)
            {
                ColumnMetaData colmd = null;
                if (fmd.getOrderMetaData().getColumnMetaData() != null && fmd.getOrderMetaData().getColumnMetaData().length > 0)
                {
                    colmd = fmd.getOrderMetaData().getColumnMetaData()[0];
                }
                orderMapping = dba.getMapping(int.class, storeMgr); // JDO2 spec [18.5] order column is assumed to be "int"
                ColumnCreator.createAdapterColumn(orderMapping, storeMgr, clr, this, colmd, pkRequired);
            }

            ColumnMetaData[] columnMetaData = null;
            AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
            ElementMetaData elemmd = fmd.getElementMetaData();
            if (elemmd != null && elemmd.getColumnMetaData() != null && elemmd.getColumnMetaData().length > 0)
            {
                // Column mappings defined at this side (1-N, M-N)
                columnMetaData = elemmd.getColumnMetaData();
            }
            else if (relatedFmd != null && relatedFmd.getJoinMetaData() != null && 
                relatedFmd.getJoinMetaData().getColumnMetaData() != null && relatedFmd.getJoinMetaData().getColumnMetaData().length > 0)
            {
                // Column mappings defined at other side (M-N)
                columnMetaData = relatedFmd.getJoinMetaData().getColumnMetaData();
            }

            boolean adapterColumnMandatory = false;
            if (orderMapping == null)
            {
                // Collection allows duplicates so make the adapter column mandatory
                if (fmd.getTypeName().equals("java.util.Collection") && !pkColsSpecified)
                {
                    // TODO Extend this to allow other types that allow dups (Bags, etc)
                    adapterColumnMandatory = true;
                }
                orderMapping = dba.getMapping(int.class, storeMgr); // JDO2 spec [18.5] order column is assumed to be "int"
            }

            // Create the element mapping
            CollectionMetaData colmd = fmd.getCollection();
            elementMapping = ColumnCreator.createColumnsForJoinTables(clr.classForName(getElementType()), fmd, 
                columnMetaData, storeMgr, this,
                (pkRequired && !pkColsSpecified && fmd.getOrderMetaData() == null), false, colmd.isSerializedElement(), false,
                DatastoreField.ROLE_COLLECTION_ELEMENT, clr, orderMapping, adapterColumnMandatory);
            if (orderMapping.getNumberOfDatastoreFields() < 1)
            {
                orderMapping = null;
            }
        }

        if (pkRequired && pkColsSpecified)
        {
            // Apply the users PK specification
            applyUserPrimaryKeySpecification(pkmd);
        }

        state = TABLE_STATE_INITIALIZED;
    }
}