/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - added patch from Jakob Braeuchi for FLOAT to REAL mapping
2007 Andy Jefferson - added handling of BLOB/LONGVARBINARY and CLOB/LONGVARCHAR
    ...
**********************************************************************/
package org.jpox.store.rdbms.typeinfo;

import java.sql.ResultSet;
import java.sql.Types;

import org.jpox.store.rdbms.columninfo.ColumnInfo;

/**
 * Represents the metadata of a MySQL data type.
 *
 * @version $Revision: 1.4.2.1 $
 */
public class MySQLTypeInfo extends TypeInfo
{
    /**
     * Constructs a type information object from the current row of the given
     * result set. The {@link ResultSet} object passed must have been obtained
     * from a call to DatabaseMetaData.getTypeInfo().
     * <p>This method only retrieves the values from the current row; the caller
     * is required to advance to the next row with {@link ResultSet#next}.
     * @param rs The result set returned from DatabaseMetaData.getTypeInfo().
     */
    public MySQLTypeInfo(ResultSet rs)
    {
        super(rs);

        if (typeName.equalsIgnoreCase("FLOAT"))
        {
            // MySQL maps FLOAT to JDBC type REAL, so reset it
            dataType = Types.FLOAT;
        }
        else if (typeName.equalsIgnoreCase("CHAR"))
        {
            // Use the BINARY qualifier to specify case-sensitive treatment of CHAR.
            // The M in (M) will get replaced by the actual precision in Column.getSQLDefinition().
            typeName = "CHAR(M) BINARY";
            createParams = "";
        }
        else if (typeName.equalsIgnoreCase("VARCHAR"))
        {
            // Use the BINARY qualifier to specify case-sensitive treatment of VARCHAR.
            // The M in (M) will get replaced by the actual precision in Column.getSQLDefinition().
            typeName = "VARCHAR(M) BINARY";
            createParams = "";
        }
        else if (typeName.equalsIgnoreCase("LONG VARCHAR") ||
                typeName.equalsIgnoreCase("BLOB") ||
                typeName.equalsIgnoreCase("MEDIUMBLOB") ||
                typeName.equalsIgnoreCase("LONGBLOB") ||
                typeName.equalsIgnoreCase("MEDIUMTEXT") ||
                typeName.equalsIgnoreCase("LONGTEXT") ||
                typeName.equalsIgnoreCase("TEXT"))
        {
            // Some MySQL types don't allow precision specification i.e "LONG VARCHAR(...)" is illegal
            allowsPrecisionSpec = false;
        }
    }

    /**
     * Utility to check the compatibility of this type with the supplied Column
     * type.
     * @param colInfo The Column type
     * @return Whether they are compatible
     **/
    public boolean isCompatibleWith(ColumnInfo colInfo)
    {
        if (super.isCompatibleWith(colInfo))
        {
            return true;
        }

        if (isStringType(dataType) && isStringType(colInfo.getDataType()))
        {
            return true;
        }

        if (dataType == Types.BIT)
        {
            // Columns created as BIT come back as TINYINT(1)
            return colInfo.getDataType() == Types.TINYINT && colInfo.getColumnSize() == 1;
        }

        if ((dataType == Types.BLOB && colInfo.getDataType() == Types.LONGVARBINARY) ||
            (dataType == Types.LONGVARBINARY && colInfo.getDataType() == Types.BLOB))
        {
            // JPOX provides BLOB since MySQL JDBC is too lazy to. They then often return it as opposite type
            return true;
        }
        if ((dataType == Types.CLOB && colInfo.getDataType() == Types.LONGVARCHAR) ||
        		(dataType == Types.LONGVARCHAR && colInfo.getDataType() == Types.CLOB))
        {
            // JPOX provides CLOB since MySQL JDBC is too lazy to. They then often return it as opposite type
            return true;
        }

        return false;
    }

    /**
     * Tests whether or not the given JDBC type is a MySQL "string" type.
     * <p>
     * MySQL likes to interchange CHAR and VARCHAR at its own discretion, and
     * automatically upgrades types to bigger types if necessary.
     * In addition, we use the BINARY qualifier on CHAR and VARCHAR to get
     * case-sensitive treatment.
     * Taken together it means we really can't distinguish one string/binary
     * type from another so we treat them all as one big happy string family.
     */
    private static boolean isStringType(int type)
    {
        switch (type)
        {
            case Types.CHAR:
            case Types.VARCHAR:
            case Types.LONGVARCHAR:
            case Types.BINARY:
            case Types.VARBINARY:
            case Types.LONGVARBINARY:
                return true;

            default:
                return false;
        }
    }
}