/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.scostore;

import java.util.Collection;
import java.util.Iterator;

import org.jpox.StateManager;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.ResultObjectFactory;

/**
 * Interface representation of the backing store for a Collection.
 *
 * @version $Revision: 1.16 $
 **/
public interface CollectionStore extends Store
{
    // --------------------------- Accessor Methods ----------------------------
 
    /**
     * Accessor for the element type in the collection.
     * @return The element type.
     **/
    String getElementType();

    /**
     * Accessor for whether the store utilises an order mapping.
     * An order mapping is used to allow for ordering of elements or to allow duplicates.
     * @return Whether it uses an order mapping.
     */
    boolean hasOrderMapping();

    /**
     * Method to update en embedded element in the collection.
     * @param sm State Manager of the owner
     * @param element The element
     * @param fieldNumber Field to update in the element
     * @param value The new value for the field
     * @return Whether the element was modified
     */
    boolean updateEmbeddedElement(StateManager sm, Object element, int fieldNumber, Object value);

    // -------------------------- Collection Methods ---------------------------
 
    /**
     * Accessor for an iterator for the collection.
     * @param sm State Manager for the collection. 
     * @return Iterator for the collection.
     **/
    Iterator iterator(StateManager sm);

    /**
     * Accessor for the size of the collection.
     * @param sm State Manager for the collection. 
     * @return The size of the collection.
     **/
    int size(StateManager sm);

    /**
     * Method to check if an element exists in the collection.
     * @param sm State Manager for the collection. 
     * @param element Element to check
     * @return Whether the element exists in the collection.
     **/
    boolean contains(StateManager sm, Object element);

    /**
     * Method to add an element to the collection.
     * @param sm State Manager for the collection. 
     * @param element Element to add
     * @return Whether the element was added ok
     **/
    boolean add(StateManager sm, Object element);

    /**
     * Method to add a collection of elements to the collection.
     * @param sm State Manager for the collection. 
     * @param elements Elements to add
     * @return Whether the elements were added ok
     **/
    boolean addAll(StateManager sm, Collection elements);

    /**
     * Method to remove an element from the collection.
     * @param sm State Manager for the collection. 
     * @param element Element to remove
     * @return Whether the element was removed ok
     **/
    boolean remove(StateManager sm, Object element);

    /**
     * Method to remove a collection of elements from the collection.
     * @param sm State Manager for the collection. 
     * @param elements Element to remove
     * @return Whether the elements were removed ok
     **/
    boolean removeAll(StateManager sm, Collection elements);

    /**
     * Method to clear the collection.
     * @param sm State Manager for the collection. 
     **/
    void clear(StateManager sm);

    // -------------------------- Query Methods --------------------------------
 
    /**
     * Method to return a new Query Statement containing the candidate class.
     * @param sm State Manager for the collection.
     * @param candidateClass Class of the elements.
     * @return The Query Statement.
     **/
    QueryExpression newQueryStatement(StateManager sm, String candidateClass);

    /**
     * Method to return an object factory for processing query statements.
     * @param sm State Manager for the collection.
     * @param stmt The Query Statement.
     * @param ignoreCache Whether to ignore the cache.
     * @param useFetchPlan whether to use the fetch plan to retrieve fields in the same query
     * @return The Persistent object factory.
     **/
    ResultObjectFactory newResultObjectFactory(StateManager sm,
                                                 QueryExpression stmt,
                                                 boolean ignoreCache,
                                                 boolean useFetchPlan);

    /**
     * Create a subquery for the given query that joins a SetStore
     * element table to the owner table. This subquery can subsequently be used
     * in an EXISTS expression to determine whether a Collection is empty. 
     * @param stmt The Query Statement.
     * @param ownerMapping the mapping for the owner. 
     * @param ownerTe Table Expression for the owner
     * @param collRangeVar   The range variable for the "Collection" table. 
     * @return  A subquery for the given query that joins a SetStore
     *          element table to the owner table. 
     */ 
    QueryExpression getExistsSubquery(QueryExpression stmt,
            				JavaTypeMapping ownerMapping,
                            LogicSetExpression ownerTe,
                            DatastoreIdentifier collRangeVar);

    /**
     * Create a subquery for the size of the collection.
     * @param stmt The Query Statement
     * @param ownerMapping Mapping for the owner
     * @param ownerTe Table Expression for the owner
     * @param collRangeVar Range variable for the owner
     * @return The subquery
     */
    QueryExpression getSizeSubquery(QueryExpression stmt,
                            JavaTypeMapping ownerMapping,
                            LogicSetExpression ownerTe,
                            DatastoreIdentifier collRangeVar);

    /**
     * Method used in queries when contains() has been invoked.
     * @param stmt The Query Statement
     * @param parentStmt the parent Query Statement. If there is no parent, <code>parentStmt</code> must be equals to <code>stmt</code> 
     * @param ownerMapping the mapping for the owner. 
     * @param ownerTe Table Expression for the owner
     * @param collRangeVar   The range variable for the "Collection" table. 
     * @param filteredElementType The Class Type for the filtered element
     * @param elmExpr The Expression for the element
     * @param elementRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the element table.
     * @return expression to the join
     **/
    ScalarExpression joinElementsTo(QueryExpression stmt,
                            QueryExpression parentStmt,
                            JavaTypeMapping ownerMapping,
                            LogicSetExpression ownerTe,
                            DatastoreIdentifier collRangeVar,
                            Class filteredElementType,
                            ScalarExpression elmExpr,
                            DatastoreIdentifier elementRangeVar);
}